Package: ukbflow
Title: Streamlined Workflow for UK Biobank Data Extraction, Analysis,
        and Visualization
Version: 0.3.4
Authors@R: 
    person("Yibin", "Zhou", email = "evanzhou.bio@gmail.com", role = c("aut", "cre"),
           comment = c(ORCID = "0009-0009-4600-8175"))
Description: Provides a streamlined workflow for UK Biobank cloud-based
    analysis on the Research Analysis Platform (RAP). Includes tools for
    phenotype extraction and decoding, variable derivation, survival and
    association analysis, genetic risk score computation, and
    publication-quality visualization. For details on the UK Biobank
    resource, see Bycroft et al. (2018) <doi:10.1038/s41586-018-0579-z>.
License: MIT + file LICENSE
URL: https://github.com/evanbio/ukbflow,
        https://evanbio.github.io/ukbflow/
BugReports: https://github.com/evanbio/ukbflow/issues
Depends: R (>= 4.1)
Imports: broom, cli, curl, data.table, dplyr, forestploter, gt,
        gtsummary, jsonlite, processx, rlang, survival, tidyselect,
        tools
Encoding: UTF-8
RoxygenNote: 7.3.2
Suggests: knitr, pagedown, pROC, rmarkdown, testthat (>= 3.0.0),
        mockery, webshot2, withr
VignetteBuilder: knitr, rmarkdown
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-08 11:19:45 UTC; 84241
Author: Yibin Zhou [aut, cre] (ORCID: <https://orcid.org/0009-0009-4600-8175>)
Maintainer: Yibin Zhou <evanzhou.bio@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-08 19:10:02 UTC
Built: R 4.6.0; ; 2026-04-25 14:00:22 UTC; unix

# Generated by roxygen2: do not edit by hand

export(assoc_competing)
export(assoc_cox)
export(assoc_coxph)
export(assoc_coxph_zph)
export(assoc_fg)
export(assoc_lag)
export(assoc_linear)
export(assoc_lm)
export(assoc_logistic)
export(assoc_logit)
export(assoc_sub)
export(assoc_subgroup)
export(assoc_tr)
export(assoc_trend)
export(assoc_zph)
export(auth_list_projects)
export(auth_login)
export(auth_logout)
export(auth_select_project)
export(auth_status)
export(decode_names)
export(decode_values)
export(derive_age)
export(derive_cancer_registry)
export(derive_case)
export(derive_covariate)
export(derive_cut)
export(derive_death_registry)
export(derive_first_occurrence)
export(derive_followup)
export(derive_hes)
export(derive_icd10)
export(derive_missing)
export(derive_selfreport)
export(derive_timing)
export(extract_batch)
export(extract_ls)
export(extract_pheno)
export(fetch_field)
export(fetch_file)
export(fetch_ls)
export(fetch_metadata)
export(fetch_tree)
export(fetch_url)
export(grs_bgen2pgen)
export(grs_check)
export(grs_score)
export(grs_standardize)
export(grs_validate)
export(grs_zscore)
export(job_ls)
export(job_path)
export(job_result)
export(job_status)
export(job_wait)
export(ops_na)
export(ops_set_safe_cols)
export(ops_setup)
export(ops_snapshot)
export(ops_snapshot_cols)
export(ops_snapshot_diff)
export(ops_snapshot_remove)
export(ops_toy)
export(ops_withdraw)
export(plot_forest)
export(plot_tableone)
import(data.table)
importFrom(broom,tidy)
importFrom(dplyr,any_of)
importFrom(forestploter,add_border)
importFrom(forestploter,edit_plot)
importFrom(forestploter,forest)
importFrom(forestploter,forest_theme)
importFrom(grid,gpar)
importFrom(grid,unit)
importFrom(gt,cell_borders)
importFrom(gt,cell_fill)
importFrom(gt,cells_body)
importFrom(gt,cells_column_labels)
importFrom(gt,cols_width)
importFrom(gt,gtsave)
importFrom(gt,opt_css)
importFrom(gt,px)
importFrom(gt,tab_options)
importFrom(gt,tab_style)
importFrom(gtsummary,add_overall)
importFrom(gtsummary,add_p)
importFrom(gtsummary,as_gt)
importFrom(gtsummary,bold_labels)
importFrom(gtsummary,bold_p)
importFrom(gtsummary,modify_header)
importFrom(gtsummary,modify_table_body)
importFrom(gtsummary,modify_table_styling)
importFrom(gtsummary,style_pvalue)
importFrom(gtsummary,tbl_summary)
importFrom(stats,anova)
importFrom(stats,confint)
importFrom(stats,relevel)
importFrom(stats,rnorm)
importFrom(stats,runif)
importFrom(stats,setNames)
importFrom(survival,Surv)
importFrom(survival,concordance)
importFrom(survival,cox.zph)
importFrom(survival,coxph)
importFrom(survival,finegray)
importFrom(utils,object.size)

<div align="center">

<img src="man/figures/logo.png" width="160" alt="ukbflow logo" />

# ukbflow

### *RAP-Native R Workflow for UK Biobank Analysis*

[![CRAN status](https://www.r-pkg.org/badges/version/ukbflow)](https://CRAN.R-project.org/package=ukbflow)
[![R-CMD-check](https://github.com/evanbio/ukbflow/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/evanbio/ukbflow/actions/workflows/R-CMD-check.yaml)
[![Codecov](https://codecov.io/gh/evanbio/ukbflow/branch/main/graph/badge.svg)](https://app.codecov.io/gh/evanbio/ukbflow?branch=main)
[![Lifecycle](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)

[📚 Documentation](https://evanbio.github.io/ukbflow/) •
[🚀 Get Started](https://evanbio.github.io/ukbflow/articles/get-started.html) •
[💬 Issues](https://github.com/evanbio/ukbflow/issues) •
[🤝 Contributing](https://github.com/evanbio/ukbflow/blob/main/CONTRIBUTING.md)

**Languages:** English | [简体中文](https://github.com/evanbio/ukbflow/blob/main/README_zh.md)

</div>

---

> [!NOTE]
> 🎉 **2026-04 — ukbflow is now available on CRAN!** Install with `install.packages("ukbflow")`.

## Overview

**ukbflow** provides a streamlined, RAP-native R workflow for UK Biobank analysis — from phenotype extraction and disease derivation to association analysis and publication-quality figures.

> **UK Biobank Data Policy (2024+)**: Individual-level data must remain within the RAP environment. Only summary-level outputs may be downloaded locally. All `ukbflow` functions are designed with this constraint in mind.

```r
library(ukbflow)

# Simulate UKB-style data locally (on RAP: replace with extract_batch() + job_wait())
data <- ops_toy(n = 5000, seed = 2026) |>
  derive_missing()

# Derive lung cancer outcome (ICD-10 C34) and follow-up time
data <- data |>
  derive_icd10(name = "lung", icd10 = "C34",
               source = c("cancer_registry", "hes")) |>
  derive_followup(name        = "lung",
                  event_col   = "lung_icd10_date",
                  baseline_col = "p53_i0",
                  censor_date  = as.Date("2022-10-31"),
                  death_col    = "p40000_i0")

# Define exposure: ever vs. never smoker
data[, smoking_ever := factor(
  ifelse(p20116_i0 == "Never", "Never", "Ever"),
  levels = c("Never", "Ever")
)]

# Cox regression: smoking → lung cancer (3-model adjustment)
res <- assoc_coxph(data,
  outcome_col  = "lung_icd10",
  time_col     = "lung_followup_years",
  exposure_col = "smoking_ever",
  covariates   = c("p21022", "p31", "p22189"))

# Forest plot
res_df <- as.data.frame(res)
plot_forest(
  data      = res_df,
  est       = res_df$HR,
  lower     = res_df$CI_lower,
  upper     = res_df$CI_upper,
  ci_column = 2L
)
```

---

## Installation

```r
# From CRAN (recommended)
install.packages("ukbflow")

# Latest development version from GitHub
pak::pkg_install("evanbio/ukbflow")

# or
remotes::install_github("evanbio/ukbflow")
```

**Requirements:** R ≥ 4.1 · [dxpy](https://documentation.dnanexus.com/downloads) (dx-toolkit, required for RAP interaction)

```bash
pip install dxpy
```

---

## Core Features

| Layer | Key Functions | Description |
|---|---|---|
| **Connection** | `auth_login`, `auth_select_project` | Authenticate to RAP via dx-toolkit |
| **Data Access** | `fetch_metadata`, `extract_batch`, `job_wait` | Retrieve phenotype data from UKB dataset on RAP |
| **Data Processing** | `decode_names`, `decode_values`, `derive_icd10`, `derive_followup`, `derive_case` | Harmonize multi-source records; derive analysis-ready cohort |
| **Association Analysis** | `assoc_coxph`, `assoc_logistic`, `assoc_subgroup` | Three-model adjustment; subgroup & trend analysis |
| **Genomic Scoring** | `grs_bgen2pgen`, `grs_score`, `grs_standardize` | Distributed plink2 scoring on RAP worker nodes |
| **Visualization** | `plot_forest`, `plot_tableone` | Publication-ready figures & tables |
| **Utilities** | `ops_setup`, `ops_toy`, `ops_na`, `ops_snapshot`, `ops_withdraw` | Environment check, synthetic data, pipeline diagnostics, and cohort management |

---

## Function Reference

<details>
<summary><b>Auth & Fetch</b></summary>

- `auth_login()`, `auth_status()`, `auth_logout()`, `auth_list_projects()`, `auth_select_project()` — RAP authentication
- `fetch_ls()`, `fetch_tree()`, `fetch_url()`, `fetch_file()` — RAP file system
- `fetch_metadata()`, `fetch_field()` — UKB metadata shortcuts

</details>

<details>
<summary><b>Extract & Decode</b></summary>

- `extract_ls()`, `extract_pheno()`, `extract_batch()` — phenotype extraction
- `decode_values()` — integer codes → human-readable labels
- `decode_names()` — field IDs → snake_case column names

</details>

<details>
<summary><b>Job Monitoring</b></summary>

- `job_status()` — query job status by ID
- `job_wait()` — block until job completes (with timeout)
- `job_path()` — get output path of a completed job
- `job_result()` — retrieve job result object
- `job_ls()` — list recent jobs

</details>

<details>
<summary><b>Derive — Phenotypes</b></summary>

- `derive_missing()` — handle "Do not know" / "Prefer not to answer"
- `derive_covariate()` — type conversion + summary
- `derive_cut()` — bin continuous variables
- `derive_selfreport()` — self-reported disease status + date
- `derive_hes()` — HES inpatient ICD-10
- `derive_first_occurrence()` — First Occurrence fields
- `derive_cancer_registry()` — cancer registry
- `derive_death_registry()` — death registry
- `derive_icd10()` — combine sources (wrapper)
- `derive_case()` — merge self-report + ICD-10

</details>

<details>
<summary><b>Derive — Survival</b></summary>

- `derive_timing()` — prevalent vs. incident classification
- `derive_age()` — age at event
- `derive_followup()` — follow-up end date and duration

</details>

<details>
<summary><b>Association Analysis</b></summary>

- `assoc_coxph()` / `assoc_cox()` — Cox proportional hazards (HR)
- `assoc_logistic()` / `assoc_logit()` — logistic regression (OR)
- `assoc_linear()` / `assoc_lm()` — linear regression (β)
- `assoc_coxph_zph()` — proportional hazards assumption test
- `assoc_subgroup()` — stratified analysis + interaction LRT
- `assoc_trend()` — dose-response trend + p_trend
- `assoc_competing()` — Fine-Gray competing risks (SHR)
- `assoc_lag()` — lagged exposure sensitivity analysis

</details>

<details>
<summary><b>Visualisation</b></summary>

- `plot_forest()` — forest plot (PNG / PDF / JPG / TIFF, 300 dpi)
- `plot_tableone()` — Table 1 (DOCX / HTML / PDF / PNG)

</details>

<details>
<summary><b>Utilities &amp; Diagnostics</b></summary>

- `ops_setup()` — environment health check (dx CLI, RAP auth, R packages)
- `ops_toy()` — generate synthetic UKB-like data for development and testing
- `ops_na()` — summarise missing values (NA and `""`) across all columns
- `ops_snapshot()` — record pipeline checkpoints and track dataset changes
- `ops_snapshot_cols()` — retrieve column list from a saved snapshot
- `ops_snapshot_diff()` — compare columns between two snapshots
- `ops_snapshot_remove()` — remove columns added after a given snapshot
- `ops_set_safe_cols()` — define protected columns that ops_snapshot_remove will not drop
- `ops_withdraw()` — exclude UKB withdrawn participants from a cohort

</details>

<details>
<summary><b>GRS Pipeline</b></summary>

- `grs_check()` — validate SNP weights file
- `grs_bgen2pgen()` — convert BGEN → PGEN on RAP (submits cloud jobs)
- `grs_score()` — score GRS across chromosomes with plink2
- `grs_standardize()` / `grs_zscore()` — Z-score standardisation
- `grs_validate()` — OR/HR per SD, high vs low, trend, AUC/C-index

</details>

---

## Documentation

Full vignettes and function reference:

**[https://evanbio.github.io/ukbflow/](https://evanbio.github.io/ukbflow/)**

---

## Contributing

Bug reports, feature requests, and pull requests are welcome. See [CONTRIBUTING.md](https://github.com/evanbio/ukbflow/blob/main/CONTRIBUTING.md).

---

## License

MIT License © 2026 [Yibin Zhou](mailto:evanzhou.bio@gmail.com)

---

<div align="center">

**Made with ❤️ by [Yibin Zhou](https://github.com/evanbio)**

[⬆ Back to Top](#ukbflow)

</div>

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = FALSE
)

## ----auth---------------------------------------------------------------------
# library(ukbflow)
# 
# auth_login()
# auth_select_project("project-XXXXXXXXXXXX")

## ----fetch-ls-----------------------------------------------------------------
# # List project root
# fetch_ls()
# #>                  name   type    size            modified
# #> 1  Showcase metadata folder    <NA>                <NA>
# #> 2             results folder    <NA>                <NA>
# #> 3        analysis.log   file  4.2 KB 2024-11-01 10:22:03
# 
# # List a specific folder
# fetch_ls("Showcase metadata/")
# #>          name  type     size            modified
# #> 1   field.tsv  file  12.3 MB 2024-10-15 08:01:44
# #> 2 esimpint.tsv  file   3.1 MB 2024-10-15 08:01:50
# 
# # Filter by type
# fetch_ls("results/", type = "file")
# 
# # Filter by name pattern
# fetch_ls("results/", pattern = "\\.csv$")

## ----fetch-tree---------------------------------------------------------------
# # Top-level overview
# fetch_tree()
# 
# # Drill into a subfolder
# fetch_tree("results/", max_depth = 2)

## ----fetch-url----------------------------------------------------------------
# # Single file
# fetch_url("Showcase metadata/field.tsv")
# 
# # Entire folder (returns a named character vector)
# fetch_url("Showcase metadata/", duration = "7d")

## ----fetch-file---------------------------------------------------------------
# # Download a single file
# fetch_file("Showcase metadata/field.tsv", dest_dir = "data/")
# 
# # Download an entire folder
# fetch_file("Showcase metadata/", dest_dir = "data/metadata/")
# 
# # Resume an interrupted download
# fetch_file("results/summary_stats.csv", dest_dir = "data/", resume = TRUE)

## ----fetch-meta---------------------------------------------------------------
# # Download all Showcase metadata files (field.tsv, encoding.tsv, etc.)
# fetch_metadata(dest_dir = "data/metadata")
# 
# # Download the field dictionary only
# fetch_field(dest_dir = "data/metadata")

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = FALSE
)

## ----install------------------------------------------------------------------
# # From CRAN (recommended)
# install.packages("ukbflow")
# 
# # Latest development version from GitHub
# pak::pkg_install("evanbio/ukbflow")

## ----load-data----------------------------------------------------------------
# library(ukbflow)
# 
# df <- ops_toy()   # synthetic UKB-like cohort, no RAP connection needed
# 
# # On RAP, replace with:
# # auth_login()
# # auth_select_project("project-XXXXXXXXXXXX")
# # df <- extract_pheno(c(31, 21022, 53, 20116)) |>
# #   decode_values() |>
# #   decode_names()

## ----derive-------------------------------------------------------------------
# df <- df |>
#   derive_missing() |>                                               # recode "Prefer not to answer" → NA
#   derive_selfreport(name = "t2dm", regex = "diabetes",           # T2DM self-report
#                     field = "noncancer") |>
#   derive_icd10(name = "t2dm", icd10 = "E11", source = "hes") |> # T2DM from HES
#   derive_case(name = "t2dm") |>                                  # → t2dm_status, t2dm_date
#   derive_followup(name         = "t2dm",
#                   event_col    = "t2dm_date",
#                   baseline_col = "p53_i0",                          # assessment centre date
#                   censor_date  = as.Date("2022-06-01"))

## ----assoc--------------------------------------------------------------------
# res <- assoc_coxph(
#   data         = df,
#   outcome_coll  = "t2dm_status",
#   time_col     = "t2dm_followup_years",
#   exposure_col = "p21001_i0",   # BMI (continuous)
#   covariates   = c("p21022",    # age_at_recruitment
#                    "p31")       # sex
# )

## ----plot---------------------------------------------------------------------
# # Forest plot — see vignette("plot") for full usage
# res_df <- as.data.frame(res)
# plot_forest(
#   data      = res_df,
#   est       = res_df$HR,
#   lower     = res_df$CI_lower,
#   upper     = res_df$CI_upper,
#   ci_column = 7L   # res_df has 6 cols before HR; CI graphic goes here
# )
# 
# # Table 1
# plot_tableone(
#   data   = as.data.frame(df),
#   vars   = c("p21022",     # age_at_recruitment
#              "p31",        # sex
#              "p21001_i0"), # bmi
#   strata = "t2dm_status"
# )

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = FALSE
)

## ----setup-data---------------------------------------------------------------
# library(ukbflow)
# 
# # ops_toy(scenario = "association") returns a pre-derived analysis-ready table:
# # covariates already as factors, bmi_cat / tdi_cat binned, and two outcomes
# # (dm_* and htn_*) with status / date / timing / followup columns in place.
# dt <- ops_toy(scenario = "association")
# dt <- dt[dm_timing != 1L]   # incident analysis: exclude prevalent DM cases

## ----assoc-coxph--------------------------------------------------------------
# # Crude + age-sex adjusted (automatic); p21022 and p31 auto-detected
# res <- assoc_coxph(
#   data         = dt,
#   outcome_col  = "dm_status",
#   time_col     = "dm_followup_years",
#   exposure_col = c("p20116_i0", "bmi_cat")
# )

## ----assoc-coxph-full---------------------------------------------------------
# # Add a Fully adjusted model
# res <- assoc_coxph(
#   data         = dt,
#   outcome_col  = "dm_status",
#   time_col     = "dm_followup_years",
#   exposure_col = "p20116_i0",
#   covariates   = c("bmi_cat", "tdi_cat", "p1558_i0",
#                    paste0("p22009_a", 1:10))
# )

## ----assoc-logistic-----------------------------------------------------------
# res <- assoc_logistic(
#   data         = dt,
#   outcome_col  = "dm_status",
#   exposure_col = c("p20116_i0", "bmi_cat"),
#   covariates   = c("tdi_cat", "p1558_i0", paste0("p22009_a", 1:10))
# )

## ----assoc-logistic-profile---------------------------------------------------
# res <- assoc_logistic(
#   data         = dt,
#   outcome_col  = "dm_status",
#   exposure_col = "grs_bmi",
#   ci_method    = "profile"   # slower but more accurate for sparse data
# )

## ----assoc-linear-------------------------------------------------------------
# # Continuous outcome (BMI); smoking and GRS as exposures
# res <- assoc_linear(
#   data         = dt,
#   outcome_col  = "p21001_i0",
#   exposure_col = c("p20116_i0", "grs_bmi"),
#   covariates   = c("tdi_cat", "p1558_i0", paste0("p22009_a", 1:10))
# )

## ----assoc-zph----------------------------------------------------------------
# zph <- assoc_coxph_zph(
#   data         = dt,
#   outcome_col  = "dm_status",
#   time_col     = "dm_followup_years",
#   exposure_col = c("p20116_i0", "bmi_cat"),
#   covariates   = c("tdi_cat", "p1558_i0")
# )
# 
# # Identify any violations
# zph[ph_satisfied == FALSE]

## ----assoc-subgroup-----------------------------------------------------------
# # Subgroup by sex; p31 is automatically excluded from within-stratum models
# res <- assoc_subgroup(
#   data         = dt,
#   outcome_col  = "dm_status",
#   time_col     = "dm_followup_years",
#   exposure_col = "p20116_i0",
#   by           = "p31",
#   method       = "coxph",
#   covariates   = c("p21022", "bmi_cat", "tdi_cat")
# )

## ----assoc-trend-prep---------------------------------------------------------
# # assoc_trend() requires an ordered factor; make bmi_cat ordered in-place
# dt[, bmi_cat := factor(bmi_cat, levels = levels(bmi_cat), ordered = TRUE)]

## ----assoc-trend--------------------------------------------------------------
# res <- assoc_trend(
#   data         = dt,
#   outcome_col  = "dm_status",
#   time_col     = "dm_followup_years",
#   exposure_col = "bmi_cat",
#   method       = "coxph",
#   covariates   = c("p21022", "p31", "tdi_cat", "p20116_i0")
# )

## ----assoc-trend-scores-------------------------------------------------------
# res <- assoc_trend(
#   data         = dt,
#   outcome_col  = "dm_status",
#   time_col     = "dm_followup_years",
#   exposure_col = "bmi_cat",
#   method       = "coxph",
#   covariates   = c("p21022", "p31", "tdi_cat", "p20116_i0"),
#   scores       = c(17, 22, 27, 35)   # approximate median BMI per category
# )

## ----assoc-competing-a--------------------------------------------------------
# # Construct a combined event column: 0 = censored, 1 = DM, 2 = HTN (competing)
# dt_cr <- dt[htn_timing != 1L]   # also exclude prevalent HTN
# dt_cr[, event_type := data.table::fcase(
#   dm_timing  == 2L, 1L,
#   htn_timing == 2L, 2L,
#   default          = 0L
# )]
# 
# res <- assoc_competing(
#   data         = dt_cr,
#   outcome_col  = "event_type",       # 0 = censored, 1 = DM, 2 = HTN
#   time_col     = "dm_followup_years",
#   exposure_col = "p20116_i0",
#   event_val    = 1L,
#   compete_val  = 2L,
#   covariates   = c("bmi_cat", "tdi_cat")
# )

## ----assoc-competing-b--------------------------------------------------------
# res <- assoc_competing(
#   data         = dt_cr,
#   outcome_col  = "dm_status",        # primary event
#   time_col     = "dm_followup_years",
#   exposure_col = c("p20116_i0", "bmi_cat"),
#   compete_col  = "htn_status",       # competing event
#   covariates   = c("tdi_cat", "p1558_i0")
# )

## ----downstream---------------------------------------------------------------
# res <- assoc_coxph(
#   data         = dt,
#   outcome_col  = "dm_status",
#   time_col     = "dm_followup_years",
#   exposure_col = "p20116_i0",
#   covariates   = c("bmi_cat", "tdi_cat", "p1558_i0")
# )
# 
# # Pass directly to plot_forest()
# plot_forest(as.data.frame(res))
# 
# # Filter to a single model
# res[model == "Fully adjusted"]
# 
# # Export
# data.table::fwrite(res, "assoc_results.csv")

## ----assoc-lag----------------------------------------------------------------
# res <- assoc_lag(
#   data         = dt,
#   outcome_col  = "dm_status",
#   time_col     = "dm_followup_years",
#   exposure_col = "p20116_i0",
#   lag_years    = c(0, 1, 2),   # 0 = full cohort reference
#   covariates   = c("bmi_cat", "tdi_cat", "p1558_i0")
# )

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = FALSE
)

## ----setup-data---------------------------------------------------------------
# library(ukbflow)
# 
# # Build on the derive pipeline from vignette("derive")
# df <- ops_toy(n = 500)
# df <- derive_missing(df)
# df <- derive_covariate(df, as_factor = c("p31", "p20116_i0"))
# df <- derive_selfreport(df, name = "dm", regex = "type 2 diabetes")
# df <- derive_icd10(df, name = "dm", icd10 = "E11", source = c("hes", "death"))
# df <- derive_case(df, name = "dm")

## ----derive-timing------------------------------------------------------------
# # Uses {name}_status and {name}_date by default
# df <- derive_timing(df, name = "dm", baseline_col = "p53_i0")

## ----derive-timing-explicit---------------------------------------------------
# df <- derive_timing(df,
#   name         = "dm",
#   status_col   = "dm_status",
#   date_col     = "dm_date",
#   baseline_col = "p53_i0"
# )

## ----derive-age---------------------------------------------------------------
# # Auto-detects {name}_date and {name}_status; produces age_at_{name} column.
# df <- derive_age(df,
#   name         = "dm",
#   baseline_col = "p53_i0",
#   age_col      = "p21022"
# )

## ----derive-age-explicit------------------------------------------------------
# df <- derive_age(df,
#   name         = "dm",
#   baseline_col = "p53_i0",
#   age_col      = "p21022",
#   date_cols    = c(dm = "dm_date"),
#   status_cols  = c(dm = "dm_status")
# )

## ----derive-followup----------------------------------------------------------
# df <- derive_followup(df,
#   name         = "dm",
#   event_col    = "dm_date",
#   baseline_col = "p53_i0",
#   censor_date  = as.Date("2022-10-31"),   # set to your study's cut-off date
#   death_col    = "p40000_i0",
#   lost_col     = FALSE                    # not available in ops_toy
# )

## ----derive-followup-nodeath--------------------------------------------------
# df <- derive_followup(df,
#   name         = "dm",
#   event_col    = "dm_date",
#   baseline_col = "p53_i0",
#   censor_date  = as.Date("2022-10-31"),
#   death_col    = FALSE,
#   lost_col     = FALSE
# )

## ----cox-example--------------------------------------------------------------
# library(survival)
# 
# # Incident analysis: exclude prevalent cases and those with undetermined timing
# df_incident <- df[dm_timing != 1L]
# 
# fit <- coxph(
#   Surv(dm_followup_years, dm_status) ~
#     p20116_i0 + p21022 + p31 + p1558_i0,
#   data = df_incident
# )
# summary(fit)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = FALSE
)

## ----grs-check-local----------------------------------------------------------
# library(ukbflow)
# 
# # Local: weights file on your machine
# w <- grs_check("weights.csv", dest = "weights_clean.txt")
# #> Read 'weights.csv': 312 rows, 5 columns.
# #> ✔ No NA values.
# #> ✔ No duplicate SNPs.
# #> ✔ All SNP IDs match rs[0-9]+ format.
# #> ✔ All effect alleles are A/T/C/G.
# #> Beta summary:
# #>   Range     : -0.412 to 0.589
# #>   Mean |beta|: 0.183
# #>   Positive  : 187 (59.9%)
# #>   Negative  : 125 (40.1%)
# #>   Zero      : 0
# #> ✔ Weights file passed checks: 312 SNPs ready for UKB RAP.
# #> ✔ Saved: 'weights_clean.txt'

## ----grs-check-rap------------------------------------------------------------
# # On RAP (RStudio) -- use /mnt/project/ paths directly
# w <- grs_check(
#   file = "/mnt/project/weights/weights.csv",
#   dest = "/mnt/project/weights/weights_clean.txt"
# )

## ----grs-bgen2pgen-test-------------------------------------------------------
# # Test on the smallest chromosome first
# ids <- grs_bgen2pgen(chr = 22, dest = "/pgen/", priority = "high")
# #> Uploading 'grs_bgen2pgen_std.R' to RAP ...
# #> ✔ Uploaded: '/grs_bgen2pgen_std.R'
# #> Submitting 1 job(s) -- mem2_ssd1_v2_x4 / priority: high
# #> ✔ chr22 -> 'job-Gxxxxxxxxxxxxxxxxxxxxxxxx'
# #> ✔ 1/1 job(s) submitted. Monitor with job_ls().

## ----grs-bgen2pgen-all--------------------------------------------------------
# # Full run: small chromosomes on standard, large on upgraded instance
# ids_small <- grs_bgen2pgen(chr = 17:22, dest = "/pgen/")
# ids_large <- grs_bgen2pgen(chr = 1:16,  dest = "/pgen/", instance = "large")
# 
# # Monitor progress (job_wait() takes one job ID at a time)
# job_ls()
# for (id in c(ids_small, ids_large)) job_wait(id)

## ----grs-score----------------------------------------------------------------
# ids <- grs_score(
#   file     = c(
#     grs_a = "weights/grs_a_weights.txt",
#     grs_b = "weights/grs_b_weights.txt"
#   ),
#   pgen_dir = "/mnt/project/pgen",
#   dest     = "/grs/",
#   priority = "high"
# )
# #> ── Uploading 2 weight file(s) to RAP ────────────────────────────────────────
# #> Uploading 'grs_a_weights.txt' ...
# #> ✔ Uploaded: '/grs_a_weights.txt'
# #> Uploading 'grs_b_weights.txt' ...
# #> ✔ Uploaded: '/grs_b_weights.txt'
# #> Submitting 2 job(s) -- mem2_ssd1_v2_x4 / priority: high
# #> ✔ grs_a -> 'job-Gxxxxxxxxxxxxxxxxxxxxxxxx'
# #> ✔ grs_b -> 'job-Gyyyyyyyyyyyyyyyyyyyyyyyy'
# #> ✔ 2/2 job(s) submitted. Monitor with job_ls().
# 
# job_wait(ids)

## ----grs-score-rap------------------------------------------------------------
# # On RAP: weights already at /mnt/project/grs_a_weights.txt
# ids <- grs_score(
#   file     = c(grs_a = "/mnt/project/grs_a_weights.txt"),
#   pgen_dir = "/mnt/project/pgen",
#   dest     = "/grs/"
# )
# #> ℹ grs_a_weights.txt already at RAP root, skipping upload.

## ----grs-standardize----------------------------------------------------------
# # Auto-detect all columns containing "grs" (case-insensitive)
# cohort <- grs_standardize(cohort)
# #> Auto-detected 2 GRS column(s): 'GRS_a', 'GRS_b'
# #> ✔ GRS_a -> GRS_a_z  [mean=0.0000, sd=1.0000]
# #> ✔ GRS_b -> GRS_b_z  [mean=0.0000, sd=1.0000]

## ----grs-standardize-explicit-------------------------------------------------
# # Or specify columns explicitly
# cohort <- grs_standardize(cohort, grs_cols = c("GRS_a", "GRS_b"))

## ----grs-validate-logistic----------------------------------------------------
# res <- grs_validate(
#   data        = cohort,
#   grs_cols    = c("GRS_a_z", "GRS_b_z"),
#   outcome_col = "outcome"
# )
# #> ── Creating GRS groups ───────────────────────────────────────────────────────
# #> ── Effect per SD (OR) ───────────────────────────────────────────────────────
# #> ── High vs Low ──────────────────────────────────────────────────────────────
# #> ── Trend test ───────────────────────────────────────────────────────────────
# #> ── AUC ──────────────────────────────────────────────────────────────────────
# #> ✔ Validation complete.
# 
# res$per_sd
# res$discrimination

## ----grs-validate-cox---------------------------------------------------------
# res <- grs_validate(
#   data        = cohort,
#   grs_cols    = c("GRS_a_z", "GRS_b_z"),
#   outcome_col = "outcome",
#   time_col    = "followup_years",
#   covariates  = c("age", "sex", paste0("pc", 1:10))
# )
# 
# res$per_sd          # HR per SD x model
# res$high_vs_low     # HR: top 20% vs bottom 20%
# res$trend           # p-trend across Q1–Q4
# res$discrimination  # C-index with 95% CI

## ----grs-full-pipeline--------------------------------------------------------
# library(ukbflow)
# 
# # 1. Validate weights (local)
# grs_check("weights.csv", dest = "weights_clean.txt")
# 
# # 2. Convert BGEN -> PGEN on RAP (submit jobs)
# ids_std  <- grs_bgen2pgen(chr = 17:22, dest = "/pgen/", maf = 0.01)
# ids_lrg  <- grs_bgen2pgen(chr = 1:16,  dest = "/pgen/", maf = 0.01, instance = "large")
# for (id in c(ids_std, ids_lrg)) job_wait(id)
# 
# # 3. Score GRS on RAP (submit jobs)
# score_ids <- grs_score(
#   file     = c(grs_a = "weights_clean.txt"),
#   pgen_dir = "/mnt/project/pgen",
#   maf      = 0.01,   # must match grs_bgen2pgen()
#   dest     = "/grs/"
# )
# job_wait(score_ids)
# 
# # 4. Download score CSV from RAP
# fetch_file("/grs/GRS_a_scores.csv", dest = "GRS_a_scores.csv")
# 
# # 5. Merge into analysis dataset and standardise
# # cohort: your analysis data.table with IID and phenotype columns
# scores <- data.table::fread("GRS_a_scores.csv")   # columns: IID, GRS_a
# cohort <- scores[cohort, on = "IID"]               # right-join: keep all cohort rows
# cohort <- grs_standardize(cohort, grs_cols = "GRS_a")
# 
# # 6. Validate
# res <- grs_validate(
#   data        = cohort,
#   grs_cols    = "GRS_a_z",
#   outcome_col = "outcome",
#   time_col    = "followup_years",
#   covariates  = c("age", "sex", paste0("pc", 1:10))
# )
# 
# res$per_sd
# res$discrimination

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = FALSE
)

## ----load-data----------------------------------------------------------------
# library(ukbflow)
# 
# data <- ops_toy(n = 500000, seed = 2026)
# #> ✔ ops_toy: 500000 participants | 75 columns | scenario = "cohort" | seed = 2026

## ----decode-names-------------------------------------------------------------
# data <- decode_names(data)
# #> Field dictionary not cached - calling `extract_ls()` to populate it.
# #> Using dataset: "XXXXXXXXXXXXXXXXXXXXXXX"
# #> Fetching approved fields... (cached after first call)
# #> 29951 fields available. Assign to a variable or use pattern= to search.
# #> ✔ Renamed 68 columns.
# #> ! 5 column names longer than 60 characters - consider renaming manually:
# #> • interpolated_year_when_non_cancer_illness_first_diagnosed_i0_a0
# #> • interpolated_year_when_non_cancer_illness_first_diagnosed_i0_a1
# #> • interpolated_year_when_non_cancer_illness_first_diagnosed_i0_a2
# #> • interpolated_year_when_non_cancer_illness_first_diagnosed_i0_a3
# #> • interpolated_year_when_non_cancer_illness_first_diagnosed_i0_a4

## ----derive-missing-----------------------------------------------------------
# data <- derive_missing(data)
# #> ✔ derive_missing: replaced 279650 values across 3 columns (action = "na").

## ----derive-covariate---------------------------------------------------------
# data <- derive_covariate(
#   data,
#   as_factor = c(
#     "p31",        # sex
#     "p20116_i0",  # smoking_status_i0
#     "p1558_i0",   # alcohol_intake_frequency_i0
#     "p54_i0"      # uk_biobank_assessment_centre_i0
#   )
# )
# #> ── Factor ───────────────────────────────────────────────────────────────────
# #> sex [2 levels]
# #>   Female: n=270191 (54%)  |  Male: n=229809 (46%)  |  <NA>: n=0 (0%)
# #> smoking_status_i0 [3 levels]
# #>   Current: n=69867 (14%)  |  Never: n=260458 (52.1%)  |  Previous: n=154871 (31%)  |  <NA>: n=14804 (3%)
# #> ! alcohol_intake_frequency_i0: 6 levels > max_levels (5), consider collapsing categories.
# #>   Daily or almost daily: n=40113 (8%)  |  Never: n=45207 (9%)  |  ...  |  <NA>: n=14973 (3%)
# #> ! uk_biobank_assessment_centre_i0: 10 levels > max_levels (5), consider collapsing categories.
# #>   Birmingham: n=49709 (9.9%)  |  Bristol: n=49832 (10%)  |  ...  |  Sheffield: n=50293 (10.1%)

## ----derive-cut---------------------------------------------------------------
# data <- derive_cut(
#   data,
#   col    = "p21001_i0",                              # body_mass_index_bmi_i0
#   n      = 4,
#   breaks = c(18.5, 25, 30),
#   labels = c("Underweight", "Normal", "Overweight", "Obese"),
#   name   = "bmi_cat"
# )
# #> ── Source: body_mass_index_bmi_i0 ───────────────────────────────────────────
# #>   mean=26.21, median=26.19, sd=5.48, Q1=22.48, Q3=29.9, NA=0% (n=0)
# #> ── New column: bmi_cat ──────────────────────────────────────────────────────
# #> bmi_cat [4 levels]
# #>   Underweight: n=40238 (8%)  |  Normal: n=166710 (33.3%)
# #>   Overweight:  n=170794 (34.2%)  |  Obese: n=122258 (24.5%)  |  <NA>: n=0 (0%)
# 
# data <- derive_cut(
#   data,
#   col    = "p22189",                                 # townsend_deprivation_index_at_recruitment
#   n      = 4,
#   labels = c("Q1 (least deprived)", "Q2", "Q3", "Q4 (most deprived)"),
#   name   = "tdi_cat"
# )
# #> ── Source: townsend_deprivation_index_at_recruitment ────────────────────────
# #>   mean=-1.25, median=-1.3, sd=3.1, Q1=-3.46, Q3=0.86, NA=0% (n=0)
# #> ── New column: tdi_cat ──────────────────────────────────────────────────────
# #> tdi_cat [4 levels]
# #>   Q1 (least deprived): n=125290 (25.1%)  |  Q2: n=124831 (25%)
# #>   Q3: n=125102 (25%)  |  Q4 (most deprived): n=124777 (25%)  |  <NA>: n=0 (0%)

## ----derive-selfreport--------------------------------------------------------
# data <- derive_selfreport(
#   data,
#   name  = "lung_cancer",
#   regex = "lung cancer",
#   field = "cancer"
# )
# #> ✔ derive_selfreport (lung_cancer): 6700 cases, 6700 with dates.

## ----derive-icd10-------------------------------------------------------------
# data <- derive_icd10(
#   data,
#   name      = "lung",
#   icd10     = "^C3[34]",
#   match     = "regex",
#   source    = "cancer_registry",
#   behaviour = 3L
# )
# #> ✔ derive_cancer_registry (lung): 2449 cases, 2449 with date.
# #> ✔ derive_icd10 (lung): 2449 cases across 1 source, 2449 with date.

## ----derive-case--------------------------------------------------------------
# data <- derive_case(
#   data,
#   name                = "lung",
#   selfreport_col      = "lung_cancer_selfreport",
#   selfreport_date_col = "lung_cancer_selfreport_date"
# )
# #> ✔ derive_case (lung): 9110 cases, 9110 with date.
# #> ℹ   Both sources (lung_icd10 & lung_cancer_selfreport): 39

## ----derive-timing------------------------------------------------------------
# data <- derive_timing(data, name = "lung", baseline_col = "p53_i0")  # date_of_attending_assessment_centre_i0
# #> ✔ derive_timing (lung_timing):
# #> ℹ   0 (no disease): 490890
# #> ℹ   1 (prevalent):  5056
# #> ℹ   2 (incident):   4054
# #> ℹ   NA (no date):   0

## ----derive-followup----------------------------------------------------------
# data <- derive_followup(
#   data,
#   name         = "lung",
#   event_col    = "lung_date",
#   baseline_col = "p53_i0",         # date_of_attending_assessment_centre_i0
#   censor_date  = as.Date("2022-10-31"),
#   death_col    = "p40000_i0",      # date_of_death_i0
#   lost_col     = FALSE             # lost-to-follow-up not available in this dataset
# )
# #> ✔ derive_followup (lung):
# #> ℹ   lung_followup_end: 500000 / 500000 non-missing
# #> ℹ   lung_followup_years: mean=13.71, median=14.09, range=[0, 16.83]

## ----exposure-----------------------------------------------------------------
# data[, smoking_ever := factor(
#   ifelse(p20116_i0 == "Never", "Never", "Ever"),
#   levels = c("Never", "Ever")   # Never = reference
# )]

## ----snapshot-raw-------------------------------------------------------------
# ops_snapshot(data, label = "raw")
# #> ── snapshot: raw ────────────────────────────────────────────────────────────
# #>   rows      500,000
# #>   cols           89
# #>   NA cols        59
# #>   size       293 MB

## ----excl-prevalent-----------------------------------------------------------
# data <- data[lung_timing != 1 | is.na(lung_timing)]
# ops_snapshot(data, label = "after excluding prevalent cases")
# #> ── snapshot: after excluding prevalent cases ─────────────────────────────────
# #>   rows      494,944  (-5,056)
# #>   cols           89  (= 0)
# #>   NA cols        58  (-1)
# #>   size       290.1 MB  (-2.95 MB)

## ----excl-na------------------------------------------------------------------
# data <- data[!is.na(smoking_ever)  &
#              !is.na(p31)           &   # sex
#              !is.na(p21022)        &   # age_at_recruitment
#              !is.na(bmi_cat)       &   # bmi category
#              !is.na(p1558_i0)      &   # alcohol_intake_frequency_i0
#              !is.na(tdi_cat)       &   # townsend deprivation category
#              !is.na(p54_i0)        &   # assessment_centre
#              !is.na(p22009_a1)     &   # PC1
#              !is.na(p22009_a2)     &   # PC2
#              !is.na(p22009_a3)     &   # PC3
#              !is.na(p22009_a4)     &   # PC4
#              !is.na(p22009_a5)     &   # PC5
#              !is.na(p22009_a6)     &   # PC6
#              !is.na(p22009_a7)     &   # PC7
#              !is.na(p22009_a8)     &   # PC8
#              !is.na(p22009_a9)     &   # PC9
#              !is.na(p22009_a10)]       # PC10
# ops_snapshot(data, label = "after excluding missing covariates")
# #> ── snapshot: after excluding missing covariates ──────────────────────────────
# #>   rows      465,937  (-29,007)
# #>   cols           89  (= 0)
# #>   NA cols        55  (-3)
# #>   size       273.3 MB  (-16.75 MB)

## ----snapshot-history---------------------------------------------------------
# ops_snapshot()
# #> ── ops_snapshot history ─────────────────────────────────────────────────────
# #>    idx                              label timestamp   nrow  ncol n_na_cols size_mb
# #>  1:  1                                raw  15:35:24 500000    89        59  293.03
# #>  2:  2    after excluding prevalent cases  15:35:51 494944    89        58  290.08
# #>  3:  3 after excluding missing covariates  15:36:18 465937    89        55  273.33

## ----cohort-overview----------------------------------------------------------
# # Exposure distribution
# data[, .N, by = smoking_ever]
# 
# #>    smoking_ever      N
# #>          <fctr>  <int>
# #>  1:        Ever 215893
# #>  2:       Never 250044
# 
# # Outcome: incident cases and timing
# data[, .N, by = lung_timing]
# #>    lung_timing      N
# #>          <int>  <int>
# #>  1:          0 462110
# #>  2:          2   3827
# 
# # Follow-up time in years
# data[, .(mean   = round(mean(lung_followup_years),   2),
#          median = round(median(lung_followup_years), 2),
#          min    = round(min(lung_followup_years),    2),
#          max    = round(max(lung_followup_years),    2))]
# #>     mean median min    max
# #>    <num>  <num> <num> <num>
# #>  1: 13.71  14.09     0 16.83

## ----assoc-coxph--------------------------------------------------------------
# res <- assoc_coxph(
#   data         = data,
#   outcome_col  = "lung_status",
#   time_col     = "lung_followup_years",
#   exposure_col = "smoking_ever",
#   covariates   = c("p21022",              # age_at_recruitment
#                    "p31",                 # sex
#                    "bmi_cat",
#                    "tdi_cat",
#                    "p1558_i0",            # alcohol_intake_frequency_i0
#                    "p54_i0",              # uk_biobank_assessment_centre_i0
#                    paste0("p22009_a", 1:10))  # genetic PCs 1-10
# )
# #> ℹ outcome_col lung_status: logical detected, converting TRUE/FALSE -> 1/0
# #> ── assoc_coxph ──────────────────────────────────────────────────────────────
# #> ℹ 1 exposure x 3 models = 3 Cox regressions
# #> ℹ Input cohort: 465937 participants (n/n_events/person_years reflect each model's actual analysis set)
# #> ── smoking_ever ─────────────────────────────────────────────────────────────
# #>   ✔ Unadjusted             | smoking_everEver: HR 0.99 (0.93-1.06), p = 0.834
# #>   ✔ Age and sex adjusted   | smoking_everEver: HR 0.99 (0.93-1.06), p = 0.838
# #>   ✔ Fully adjusted         | smoking_everEver: HR 0.99 (0.93-1.06), p = 0.829
# #> ✔ Done: 3 result rows across 1 exposure and 3 models.

## ----res-print----------------------------------------------------------------
# print(res)
# #>        exposure             term                model      n n_events person_years        HR  CI_lower
# #>          <char>           <char>                <ord>  <int>    <num>        <num>     <num>     <num>
# #>  1: smoking_ever smoking_everEver           Unadjusted 465937     3827      6388195 0.9932076 0.9320517
# #>  2: smoking_ever smoking_everEver Age and sex adjusted 465937     3827      6388195 0.9933735 0.9322072
# #>  3: smoking_ever smoking_everEver       Fully adjusted 465937     3827      6388195 0.9930029 0.9318576
# #>     CI_upper   p_value         HR_label
# #>        <num>     <num>           <char>
# #>  1: 1.058376 0.8335146 0.99 (0.93-1.06)
# #>  2: 1.058553 0.8375372 0.99 (0.93-1.06)
# #>  3: 1.058160 0.8285626 0.99 (0.93-1.06)
# 
# class(res)
# #> [1] "data.table" "data.frame"

## ----res-df-------------------------------------------------------------------
# res_df <- as.data.frame(res)

## ----forest-plot--------------------------------------------------------------
# p <- plot_forest(
#   data       = res_df,
#   est        = res_df$HR,
#   lower      = res_df$CI_lower,
#   upper      = res_df$CI_upper,
#   ci_column  = 2L,          # forest plot rendered in column 2 (HR_label)
#   p_cols     = "p_value",
#   ref_line   = 1,
#   xlim       = c(0, 2.0),
#   ticks_at   = c(0, 0.5, 1.0, 1.5, 2.0)
# )
# 
# plot(p)

## ----forest-plot-pub----------------------------------------------------------
# p2 <- plot_forest(
#   data       = res_df[, c("model", "n_events", "n", "p_value")],
#   est        = res_df$HR,
#   lower      = res_df$CI_lower,
#   upper      = res_df$CI_upper,
#   ci_column  = 4L,
#   p_cols     = "p_value",
#   ref_line   = 1,
#   xlim       = c(0, 2.0),
#   ticks_at   = c(0.5, 0.75, 1.0, 1.25, 1.5),
#   header     = c("Model", "Cases", "N", "", "HR (95% CI)", "P value")
# )
# 
# plot(p2)

## ----forest-plot-header-------------------------------------------------------
# res_pub <- rbind(
#   data.frame(model = "Ever vs. Never", HR_label = "", p_value = NA,
#              HR = NA, CI_lower = NA, CI_upper = NA, stringsAsFactors = FALSE),
#   res_df[, c("model", "HR_label", "p_value", "HR", "CI_lower", "CI_upper")]
# )
# 
# p3 <- plot_forest(
#   data       = res_pub[, c("model", "p_value")],
#   est        = res_pub$HR,
#   lower      = res_pub$CI_lower,
#   upper      = res_pub$CI_upper,
#   ci_column  = 2L,
#   p_cols     = "p_value",
#   ref_line   = 1,
#   xlim       = c(0, 2.0),
#   ticks_at   = c(0, 0.5, 1.0, 1.5, 2.0),
#   indent     = c(0L, 1L, 1L, 1L),
#   bold_label = c(TRUE, FALSE, FALSE, FALSE),
#   header     = c("Model", "", "HR (95% CI)", "P value")
# )
# 
# plot(p3)

## ----tableone-----------------------------------------------------------------
# t1 <- plot_tableone(
#   data    = as.data.frame(data),
#   vars    = c("p21022", "p31", "bmi_cat", "tdi_cat", "p1558_i0"),
#   strata  = "smoking_ever",
#   label   = list(
#     p21022   ~ "Age at recruitment (years)",
#     p31      ~ "Sex",
#     bmi_cat  ~ "BMI category",
#     tdi_cat  ~ "Townsend deprivation index",
#     p1558_i0 ~ "Alcohol intake frequency"
#   ),
#   add_p   = TRUE,
#   save    = FALSE
# )
# 
# print(t1)

## ----session-info-------------------------------------------------------------
# sessionInfo()
# #> R version 4.5.1 (2025-06-13 ucrt)
# #> Platform: x86_64-w64-mingw32/x64
# #> Running under: Windows 11 x64 (build 26200)
# #>
# #> Matrix products: default
# #>   LAPACK version 3.12.1
# #>
# #> locale:
# #> [1] LC_COLLATE=Chinese (Simplified)_China.utf8  LC_CTYPE=Chinese (Simplified)_China.utf8
# #> [3] LC_MONETARY=Chinese (Simplified)_China.utf8 LC_NUMERIC=C
# #> [5] LC_TIME=Chinese (Simplified)_China.utf8
# #>
# #> time zone: Asia/Shanghai
# #> tzcode source: internal
# #>
# #> attached base packages:
# #> [1] stats     graphics  grDevices utils     datasets  methods   base
# #>
# #> other attached packages:
# #> [1] ukbflow_0.3.0  testthat_3.2.3
# #>
# #> loaded via a namespace (and not attached):
# #>  [1] gt_1.0.0            sass_0.4.10         tidyr_1.3.1         generics_0.1.4      gtsummary_2.4.0
# #>  [6] xml2_1.3.8          lattice_0.22-7      digest_0.6.37       magrittr_2.0.3      evaluate_1.0.5
# #> [11] grid_4.5.1          cards_0.7.0         pkgload_1.4.0       fastmap_1.2.0       rprojroot_2.1.1
# #> [16] jsonlite_2.0.0      Matrix_1.7-3        processx_3.8.6      pkgbuild_1.4.8      sessioninfo_1.2.3
# #> [21] backports_1.5.0     cardx_0.3.0         brio_1.1.5          survival_3.8-3      ps_1.9.1
# #> [26] gridExtra_2.3       purrr_1.0.4         cli_3.6.4           forestploter_1.1.3  rlang_1.1.6
# #> [31] litedown_0.7        commonmark_2.0.0    ellipsis_0.3.2      splines_4.5.1       remotes_2.5.0
# #> [36] withr_3.0.2         cachem_1.1.0        devtools_2.4.5.9000 tools_4.5.1         memoise_2.0.1
# #> [41] dplyr_1.1.4.9000    broom_1.0.10        curl_7.0.0          vctrs_0.6.5         R6_2.6.1
# #> [46] lifecycle_1.0.4     fs_1.6.6            usethis_3.2.1       pkgconfig_2.0.3     desc_1.4.3
# #> [51] pillar_1.11.1       gtable_0.3.6        data.table_1.17.0   glue_1.8.0          xfun_0.52
# #> [56] tibble_3.2.1        tidyselect_1.2.1    rstudioapi_0.17.1   knitr_1.50          htmltools_0.5.8.1
# #> [61] rmarkdown_2.29      compiler_4.5.1      markdown_2.0

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = FALSE
)

## ----auth---------------------------------------------------------------------
# library(ukbflow)
# 
# auth_login()
# auth_select_project("project-XXXXXXXXXXXX")

## ----extract-ls---------------------------------------------------------------
# # List all approved fields (cached after first call)
# extract_ls()
# 
# # Search by keyword
# extract_ls(pattern = "cancer")
# extract_ls(pattern = "p31|p53|p21022")
# 
# # Force refresh after switching projects or datasets
# extract_ls(refresh = TRUE)

## ----extract-batch------------------------------------------------------------
# # Submit extraction job
# job_id <- extract_batch(c(31, 53, 21022, 22189))
# 
# # Custom output name
# job_id <- extract_batch(
#   field_id = c(31, 53, 21022, 22189),
#   file     = "ukb_demographics"
# )
# 
# # High priority (faster queue, higher cost)
# job_id <- extract_batch(
#   field_id = c(31, 53, 21022, 22189),
#   priority = "high"
# )

## ----job-monitor--------------------------------------------------------------
# job_status(job_id)        # check progress
# job_path(job_id)          # get cloud file path once complete
# job_result(job_id)        # read result as data.table (inside RAP only)

## ----extract-pheno------------------------------------------------------------
# df <- extract_pheno(c(31, 53, 21022))

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = FALSE
)

## ----forest-minimal-----------------------------------------------------------
# library(ukbflow)
# 
# df <- data.frame(
#   item      = c("Exposure vs. control", "Unadjusted", "Fully adjusted"),
#   `Cases/N` = c("", "89 / 4 521", "89 / 4 521"),
#   p_value   = c(NA_real_, 0.001, 0.006),
#   check.names = FALSE
# )
# 
# p <- plot_forest(
#   data      = df,
#   est       = c(NA,   1.52, 1.43),
#   lower     = c(NA,   1.18, 1.11),
#   upper     = c(NA,   1.96, 1.85),
#   ci_column = 3L,
#   indent    = c(0L,   1L,   1L),
#   p_cols    = "p_value",
#   xlim      = c(0.5,  3.0)
# )
# plot(p)

## ----forest-from-assoc--------------------------------------------------------
# dt  <- ops_toy(scenario = "association")
# dt  <- dt[dm_timing != 1L]
# 
# res <- assoc_coxph(
#   data         = dt,
#   outcome_col  = "dm_status",
#   time_col     = "dm_followup_years",
#   exposure_col = "p20116_i0",
#   covariates   = c("bmi_cat", "tdi_cat", "p1558_i0")
# )
# res <- as.data.frame(res)
# 
# # Reshape: one row per model, label column first
# df2 <- data.frame(
#   item    = c("Smoking status", as.character(res$model)),
#   `N`     = c("", paste0(res$n, " / ", res$n_events)),
#   p_value = c(NA_real_, res$p_value),
#   check.names = FALSE
# )
# 
# p <- plot_forest(
#   data      = df2,
#   est       = c(NA,   res$HR),
#   lower     = c(NA,   res$CI_lower),
#   upper     = c(NA,   res$CI_upper),
#   ci_column = 3L,
#   indent    = c(0L,   rep(1L, nrow(res))),
#   p_cols    = "p_value",
#   xlim      = c(0.5,  2.5)
# )
# plot(p)

## ----forest-ci----------------------------------------------------------------
# # uses df, est, lower, upper from the minimal example above
# p <- plot_forest(
#   data      = df,
#   est       = est, lower = lower, upper = upper,
#   ci_column = 3L,
#   ci_col    = c("grey50", "steelblue", "steelblue"),  # per-row colours
#   ci_sizes  = 0.5,       # point size
#   ci_Theight = 0.15,     # cap height
#   ref_line  = 1,         # reference line (use 0 for beta coefficients)
#   xlim      = c(0.2, 5), ticks_at = c(0.5, 1, 2, 3)
# )

## ----forest-indent------------------------------------------------------------
# # indent = 0 → bold parent row; indent >= 1 → indented sub-row (plain)
# p <- plot_forest(
#   data       = df,
#   est        = est, lower = lower, upper = upper,
#   ci_column  = 3L,
#   indent     = c(0L, 1L, 1L),        # parent + 2 sub-rows
#   bold_label = c(TRUE, FALSE, FALSE)  # explicit control (overrides indent default)
# )

## ----forest-pval--------------------------------------------------------------
# # p_cols: column names in data that contain raw numeric p-values.
# # Values < 10^(-p_digits) are displayed as e.g. "<0.001".
# # bold_p = TRUE bolds all p < p_threshold (default 0.05).
# p <- plot_forest(
#   data        = df,
#   est         = est, lower = lower, upper = upper,
#   ci_column   = 3L,
#   p_cols      = "p_value",
#   p_digits    = 3L,
#   bold_p      = TRUE,
#   p_threshold = 0.05
# )

## ----forest-header------------------------------------------------------------
# # data has 3 columns → final table has 5 columns (original 3 + gap_ci + OR label)
# # Layout with ci_column = 3L: item | Cases/N | gap_ci | OR (95% CI) | p_value
# p <- plot_forest(
#   data      = df,
#   est       = est, lower = lower, upper = upper,
#   ci_column = 3L,
#   header    = c("Comparison", "Cases / N", "", "HR (95% CI)", "P-value")
#   #             col 1          col 2        gap  OR label       col 5
# )

## ----forest-align-------------------------------------------------------------
# p <- plot_forest(
#   data      = df,
#   est       = est, lower = lower, upper = upper,
#   ci_column = 3L,
#   align     = c(-1L, 0L, 0L, 0L, 1L)   # label left | Cases/N centre | gap | OR centre | p right
# )

## ----forest-style-------------------------------------------------------------
# p <- plot_forest(
#   data       = df,
#   est        = est, lower = lower, upper = upper,
#   ci_column  = 3L,
#   background = "zebra",       # "zebra" | "bold_label" | "none"
#   bg_col     = "#F0F0F0",     # shading colour
#   border     = "three_line",  # "three_line" | "none"
#   border_width = 3            # scalar or length-3 vector (top / mid / bottom)
# )

## ----forest-save--------------------------------------------------------------
# # uses df, est, lower, upper from the minimal example above
# p <- plot_forest(
#   data        = df,
#   est         = est, lower = lower, upper = upper,
#   ci_column   = 3L,
#   row_height  = NULL,   # auto (8 / 12 / 10 / 15 mm); or scalar/vector
#   col_width   = NULL,   # auto (rounds up to nearest 5 mm)
#   save        = TRUE,
#   dest        = "forest_main",   # extension ignored; all 4 formats saved
#   save_width  = 20,              # cm
#   save_height = NULL             # auto: nrow(data) * 0.9 + 3 cm
# )

## ----tableone-minimal---------------------------------------------------------
# library(gtsummary)
# data(trial)   # built-in gtsummary dataset
# 
# plot_tableone(
#   data   = trial,
#   vars   = c("age", "marker", "grade"),
#   strata = "trt",
#   save   = FALSE
# )

## ----tableone-full------------------------------------------------------------
# plot_tableone(
#   data    = trial,
#   vars    = c("age", "marker", "grade", "stage"),
#   strata  = "trt",
#   label   = list(age ~ "Age (years)", marker ~ "Marker level (ng/mL)"),
#   add_p   = TRUE,    # Wilcoxon / chi-squared p-values; formatted as <0.001
#   add_smd = TRUE,
#   overall = TRUE,
#   dest    = "table1",
#   save    = TRUE
# )

## ----tableone-types-----------------------------------------------------------
# dt <- as.data.frame(ops_toy(scenario = "association"))
# 
# plot_tableone(
#   data      = dt,
#   vars      = c("p21022", "p21001_i0", "p31", "p20116_i0"),
#   strata    = "dm_status",
#   type      = list(p21022 = "continuous2"),   # show median + IQR
#   statistic = list(
#     all_continuous()  ~ "{mean} ({sd})",
#     all_categorical() ~ "{n} ({p}%)"
#   ),
#   digits    = list(p21022 ~ 1, p21001_i0 ~ 1),
#   missing   = "ifany",   # show missing counts when present
#   save      = FALSE
# )

## ----tableone-smd-------------------------------------------------------------
# plot_tableone(
#   data    = dt,
#   vars    = c("p21022", "p21001_i0", "p31"),
#   strata  = "dm_status",
#   add_smd = TRUE,
#   save    = FALSE
# )

## ----tableone-exclude---------------------------------------------------------
# plot_tableone(
#   data           = dt,
#   vars           = c("p31", "p20116_i0"),
#   strata         = "dm_status",
#   exclude_labels = "Never",   # e.g. remove reference category from display
#   save           = FALSE
# )

# =============================================================================
# auth.R — Authentication & connection to DNAnexus RAP
# =============================================================================


#' Login to DNAnexus with a token
#'
#' Authenticates with the DNAnexus Research Analysis Platform using an API
#' token. Equivalent to running `dx login --token` on the command line.
#'
#' @param token (character) DNAnexus API token. If NULL, reads from the
#'   environment variable `DX_API_TOKEN`.
#'
#' @return Invisible TRUE on success.
#' @export
#'
#' @examples
#' \dontrun{
#' # Supply token directly
#' auth_login(token = "your_token_here")
#'
#' # Or store token in ~/.Renviron (recommended):
#' # usethis::edit_r_environ()
#' # Add: DX_API_TOKEN=your_token_here
#' # Save and restart R, then call:
#' auth_login()
#' }
auth_login <- function(token = NULL) {
  if (is.null(token)) {
    token <- Sys.getenv("DX_API_TOKEN")
    if (!nzchar(token)) {
      stop(
        "No token provided. Supply a token or set DX_API_TOKEN environment variable.",
        call. = FALSE
      )
    }
  } else {
    .assert_scalar_string(token)
  }

  result <- .dx_run(c("login", "--token", token, "--noprojects"))

  if (!result$success) {
    stop("Login failed: ", result$stderr, call. = FALSE)
  }

  # Reason: dx login may return status 0 even with invalid token if a session
  # already exists. Verify by calling whoami to confirm actual authentication.
  verify <- .dx_run(c("whoami"))
  if (!verify$success) {
    stop("Login failed: token invalid or expired.", call. = FALSE)
  }

  cli::cli_alert_success("Logged in to DNAnexus as: {.val {verify$stdout}}")
  invisible(TRUE)
}


#' Check current DNAnexus authentication status
#'
#' Returns the current logged-in user and selected project.
#'
#' @return A named list with `user` and `project`.
#' @export
#'
#' @examples
#' \dontrun{
#' auth_status()
#' }
auth_status <- function() {
  whoami <- .dx_run(c("whoami"))

  if (!whoami$success) {
    stop("Not logged in. Run auth_login() first.", call. = FALSE)
  }

  status <- list(
    user    = whoami$stdout,
    project = .dx_get_project_id()
  )

  cli::cli_inform(c(
    "*" = "User:    {.val {status$user}}",
    "*" = "Project: {.val {if (is.na(status$project)) 'None selected' else status$project}}"
  ))

  invisible(status)
}


#' Logout from DNAnexus
#'
#' Invalidates the current DNAnexus session on the remote platform. The local
#' token file is not removed but becomes invalid. A new token must be generated
#' from the DNAnexus platform before calling [auth_login()] again.
#'
#' @return Invisible TRUE on success.
#' @export
#'
#' @examples
#' \dontrun{
#' auth_logout()
#' }
auth_logout <- function() {
  result <- .dx_run(c("logout"))

  if (!result$success) {
    stop("Logout failed: ", result$stderr, call. = FALSE)
  }

  cli::cli_alert_success("Logged out from DNAnexus.")
  invisible(TRUE)
}


#' List available DNAnexus projects
#'
#' Returns a list of all projects accessible to the current user.
#'
#' @return A character vector of project names and IDs.
#' @export
#'
#' @examples
#' \dontrun{
#' auth_list_projects()
#' }
auth_list_projects <- function() {
  result <- .dx_run(c("find", "projects", "--level", "VIEW"))

  if (!result$success) {
    stop("Failed to list projects: ", result$stderr, call. = FALSE)
  }

  if (!nzchar(result$stdout)) {
    cli::cli_inform("No projects found.")
    return(invisible(character(0)))
  }

  projects <- strsplit(result$stdout, "\n")[[1]]
  cli::cli_inform(paste(projects, collapse = "\n"))
  invisible(projects)
}


#' Select a DNAnexus project
#'
#' Switches the active project context on the DNAnexus platform. Only project
#' IDs (e.g. `"project-XXXXXXXXXXXX"`) are accepted. Run
#' [auth_list_projects()] to find your project ID.
#'
#' @param project (character) Project ID in the form `"project-XXXXXXXXXXXX"`.
#'
#' @return Invisible TRUE on success.
#' @export
#'
#' @examples
#' \dontrun{
#' auth_select_project("project-XXXXXXXXXXXX")
#' }
auth_select_project <- function(project) {
  if (missing(project)) {
    stop(
      "Please provide a project ID. Run auth_list_projects() to see available projects.",
      call. = FALSE
    )
  }
  .assert_scalar_string(project)

  # Reason: only accept project IDs (project-XXXX) to avoid name/ID ambiguity;
  # names are not unique and can cause silent mismatches.
  if (!grepl("^project-[A-Za-z0-9]+$", project)) {
    stop(
      "Invalid project ID format. Expected 'project-XXXXXXXXXXXX'. ",
      "Run auth_list_projects() to see available IDs.",
      call. = FALSE
    )
  }

  result <- .dx_run(c("select", project))

  if (!result$success) {
    stop("Failed to select project: ", result$stderr, call. = FALSE)
  }

  # Reason: verify project was actually switched
  confirmed <- .dx_get_project_id()

  if (is.na(confirmed) || !identical(confirmed, project)) {
    stop(
      "Project selection failed: expected ", project,
      ", got ", ifelse(is.na(confirmed), "NA", confirmed),
      call. = FALSE
    )
  }

  cli::cli_alert_success("Project selected: {.val {confirmed}}")
  invisible(TRUE)
}

# =============================================================================
# test-integration-auth.R — Integration tests for auth_ series
# Requires real dx-toolkit, token, and network connection
# Run manually before release: devtools::test(filter = "integration-auth")
# =============================================================================

skip_on_cran()

# ===========================================================================
# auth_login()
# ===========================================================================

test_that("auth_login() successfully logs in with a valid token", {
  token <- .skip_if_no_rap()
  expect_message(auth_login(token = token), "Logged in to DNAnexus")
})


# ===========================================================================
# auth_status()
# ===========================================================================

test_that("auth_status() returns a list with user and project fields", {
  .skip_if_no_rap()
  result <- suppressMessages(auth_status())
  expect_type(result, "list")
  expect_named(result, c("user", "project"))
  expect_true(nzchar(result$user))
})

# ===========================================================================
# auth_list_projects()
# ===========================================================================

test_that("auth_list_projects() returns a non-empty character vector", {
  .skip_if_no_rap()
  result <- suppressMessages(auth_list_projects())
  expect_type(result, "character")
  expect_gt(length(result), 0)
})

# ===========================================================================
# auth_select_project()
# ===========================================================================

test_that("auth_select_project() successfully selects a valid project", {
  .skip_if_no_rap()
  projects <- suppressMessages(auth_list_projects())
  project_id <- regmatches(projects[1], regexpr("project-[A-Za-z0-9]+", projects[1]))
  expect_message(auth_select_project(project_id), "Project selected")
})

test_that("auth_select_project() switches to the correct project (confirmed via auth_status)", {
  .skip_if_no_rap()
  projects <- suppressMessages(auth_list_projects())
  project_id <- regmatches(projects[1], regexpr("project-[A-Za-z0-9]+", projects[1]))
  suppressMessages(auth_select_project(project_id))
  status <- suppressMessages(auth_status())
  expect_equal(status$project, project_id)
})

test_that("auth_select_project() throws error for a non-existent project ID", {
  .skip_if_no_rap()
  expect_error(auth_select_project("project-DOESNOTEXIST000"), "Failed to select project")
})

# =============================================================================
# test-auth.R — Unit tests for auth_ series (mock-based, no network required)
# =============================================================================
.skip_if_no_mockery()

# Helper: build a fake .dx_run() result
.fake_dx <- function(stdout = "", stderr = "", status = 0) {
  list(
    stdout  = stdout,
    stderr  = stderr,
    status  = status,
    success = status == 0
  )
}

# ===========================================================================
# auth_login()
# ===========================================================================

test_that("auth_login() throws error when token is NULL and env var is unset", {
  withr::with_envvar(c(DX_API_TOKEN = ""), {
    expect_error(auth_login(token = NULL), "DX_API_TOKEN")
  })
})

test_that("auth_login() throws error when token is an empty string", {
  expect_error(auth_login(token = ""), "single non-empty string")
})

test_that("auth_login() uses DX_API_TOKEN env var when token argument is NULL", {
  withr::with_envvar(c(DX_API_TOKEN = "fake_token"), {
    mockery::stub(auth_login, ".dx_run", function(args, ...) {
      if (args[1] == "login") return(.fake_dx(status = 0))
      if (args[1] == "whoami") return(.fake_dx(stdout = "TestUser"))
    })
    expect_message(auth_login(token = NULL), "TestUser")
  })
})

test_that("auth_login() throws error when token is NA", {
  expect_error(auth_login(token = NA_character_), "single non-empty string")
})

test_that("auth_login() throws error when token is a length-2 vector", {
  expect_error(auth_login(token = c("tok1", "tok2")), "single non-empty string")
})

test_that("auth_login() throws error when token is non-character", {
  expect_error(auth_login(token = 123), "single non-empty string")
})

test_that("auth_login() throws error when dx login returns non-zero status", {
  mockery::stub(auth_login, ".dx_run", function(args, ...) {
    .fake_dx(stderr = "Invalid auth token", status = 1)
  })
  expect_error(auth_login(token = "bad_token"), "Login failed")
})

test_that("auth_login() throws error when whoami fails after login", {
  mockery::stub(auth_login, ".dx_run", function(args, ...) {
    if (args[1] == "login") return(.fake_dx(status = 0))
    if (args[1] == "whoami") return(.fake_dx(stderr = "Not logged in", status = 1))
  })
  expect_error(auth_login(token = "bad_token"), "token invalid or expired")
})

test_that("auth_login() returns TRUE invisibly on success", {
  mockery::stub(auth_login, ".dx_run", function(args, ...) {
    if (args[1] == "login") return(.fake_dx(status = 0))
    if (args[1] == "whoami") return(.fake_dx(stdout = "TestUser"))
  })
  expect_invisible(auth_login(token = "valid_token"))
})

# ===========================================================================
# auth_status()
# ===========================================================================

test_that("auth_status() throws error when not logged in", {
  mockery::stub(auth_status, ".dx_run", function(...) .fake_dx(stderr = "Not logged in", status = 1))
  expect_error(auth_status(), "Not logged in")
})

test_that("auth_status() returns NA project when no project is selected", {
  mockery::stub(auth_status, ".dx_run", function(args, ...) {
    if (args[1] == "whoami") return(.fake_dx(stdout = "TestUser"))
    if (args[1] == "env")    return(.fake_dx(stdout = "export DX_APISERVER_HOST=api.dnanexus.com"))
  })
  mockery::stub(auth_status, ".dx_get_project_id", function(...) NA_character_)
  result <- suppressMessages(auth_status())
  expect_true(is.na(result$project))
})

test_that("auth_status() returns correct user and project when logged in", {
  mockery::stub(auth_status, ".dx_run", function(...) .fake_dx(stdout = "TestUser"))
  mockery::stub(auth_status, ".dx_get_project_id", function(...) "project-XXXXXXXXXXXX")
  result <- suppressMessages(auth_status())
  expect_equal(result$user, "TestUser")
  expect_equal(result$project, "project-XXXXXXXXXXXX")
})

# ===========================================================================
# auth_logout()
# ===========================================================================

test_that("auth_logout() throws error when dx logout fails", {
  mockery::stub(auth_logout, ".dx_run", function(...) .fake_dx(stderr = "Not logged in", status = 1))
  expect_error(auth_logout(), "Logout failed")
})

test_that("auth_logout() returns TRUE invisibly on success", {
  mockery::stub(auth_logout, ".dx_run", function(...) .fake_dx(status = 0))
  expect_invisible(auth_logout())
})

# ===========================================================================
# auth_list_projects()
# ===========================================================================

test_that("auth_list_projects() throws error when dx find fails", {
  mockery::stub(auth_list_projects, ".dx_run", function(...) .fake_dx(stderr = "Auth error", status = 1))
  expect_error(auth_list_projects(), "Failed to list projects")
})

test_that("auth_list_projects() returns empty vector when no projects found", {
  mockery::stub(auth_list_projects, ".dx_run", function(...) .fake_dx(stdout = "", status = 0))
  result <- suppressMessages(auth_list_projects())
  expect_equal(result, character(0))
})

test_that("auth_list_projects() returns character vector of projects", {
  fake_output <- paste(
    "project-AAA : Project A (ADMINISTER)",
    "project-BBB : Project B (VIEW)",
    sep = "\n"
  )
  mockery::stub(auth_list_projects, ".dx_run", function(...) .fake_dx(stdout = fake_output, status = 0))
  result <- suppressMessages(auth_list_projects())
  expect_length(result, 2)
  expect_true(grepl("project-AAA", result[1]))
})

# ===========================================================================
# auth_select_project()
# ===========================================================================

test_that("auth_select_project() throws error when project argument is missing", {
  expect_error(auth_select_project(), "project ID")
})

test_that("auth_select_project() throws error when project is empty string", {
  expect_error(auth_select_project(""), "single non-empty string")
})

test_that("auth_select_project() throws error when project is NA", {
  expect_error(auth_select_project(NA_character_), "single non-empty string")
})

test_that("auth_select_project() throws error when project is a length-2 vector", {
  expect_error(auth_select_project(c("project-AAA", "project-BBB")), "single non-empty string")
})

test_that("auth_select_project() throws error when project is non-character", {
  expect_error(auth_select_project(123), "single non-empty string")
})

test_that("auth_select_project() throws error when dx select fails", {
  mockery::stub(auth_select_project, ".dx_run", function(...) .fake_dx(stderr = "Not found", status = 1))
  expect_error(auth_select_project("project-XXXXXXXXXXXX"), "Failed to select project")
})

test_that("auth_select_project() throws error when project context is not updated (NA)", {
  mockery::stub(auth_select_project, ".dx_run", function(...) .fake_dx(status = 0))
  mockery::stub(auth_select_project, ".dx_get_project_id", function(...) NA_character_)
  expect_error(auth_select_project("project-XXXXXXXXXXXX"), "Project selection failed")
})

test_that("auth_select_project() throws error when confirmed project differs from requested", {
  mockery::stub(auth_select_project, ".dx_run", function(...) .fake_dx(status = 0))
  mockery::stub(auth_select_project, ".dx_get_project_id", function(...) "project-YYYYYYYYYYYY")
  expect_error(auth_select_project("project-XXXXXXXXXXXX"), "Project selection failed")
})

test_that("auth_select_project() returns TRUE invisibly on success", {
  mockery::stub(auth_select_project, ".dx_run", function(...) .fake_dx(status = 0))
  mockery::stub(auth_select_project, ".dx_get_project_id", function(...) "project-XXXXXXXXXXXX")
  expect_invisible(auth_select_project("project-XXXXXXXXXXXX"))
})

# =============================================================================
# utils_derive.R — internal helpers for derive_ series
# =============================================================================


# UKB labels that carry informative missingness (refusal / uncertainty).
# These are converted to NA or "Unknown" depending on the action parameter.
# "" (empty string) is handled separately — always converted to NA.
.ukb_informative_missing <- c(
  "Do not know",
  "Prefer not to answer",
  "Prefer not to say"
)


# Print a cli summary line for a numeric column.
# Used by derive_covariate() and derive_cut().
.cli_summarise_numeric <- function(x, col, n_rows) {
  x_valid  <- x[!is.na(x)]
  na_n     <- sum(is.na(x))
  na_pct   <- round(na_n / n_rows * 100, 1L)
  mean_v   <- if (length(x_valid) > 0L) round(mean(x_valid),                    2L) else NA
  median_v <- if (length(x_valid) > 0L) round(stats::median(x_valid),            2L) else NA
  sd_v     <- if (length(x_valid) > 1L) round(stats::sd(x_valid),                2L) else NA
  q1_v     <- if (length(x_valid) > 0L) round(stats::quantile(x_valid, 0.25),    2L) else NA
  q3_v     <- if (length(x_valid) > 0L) round(stats::quantile(x_valid, 0.75),    2L) else NA
  cli::cli_text(
    "  {.field {col}}: mean={mean_v}, median={median_v}, sd={sd_v}, Q1={q1_v}, Q3={q3_v}, NA={na_pct}% (n={na_n})"
  )
}


# Print a cli summary block for a factor column and warn if high-cardinality.
# Used by derive_covariate() and derive_cut().
.cli_summarise_factor <- function(f, col, n_rows, max_levels = 5L) {
  n_lvls <- nlevels(f)

  if (n_lvls > max_levels) {
    cli::cli_alert_warning(
      "{.field {col}}: {n_lvls} levels > max_levels ({max_levels}), consider collapsing categories."
    )
  }

  cli::cli_text("  {.field {col}} [{n_lvls} level{?s}]")

  # Reason: include NA so user sees missing proportion alongside valid levels
  tab <- table(f, useNA = "always")
  for (i in seq_along(tab)) {
    lbl   <- names(tab)[i]
    lbl   <- if (is.na(lbl)) "<NA>" else lbl
    count <- tab[[i]]
    pct   <- round(count / n_rows * 100, 1L)
    cli::cli_text("    {lbl}: n={count} ({pct}%)")
  }
}


# Parse UKB decimal year-month format (e.g. 2005.58 → 2005-07-01) to IDate.
# Returns NA_integer_ (IDate) for missing, zero, or negative values.
# Uses IDate (integer-backed) so fcoalesce() works without type coercion.
# Used by derive_selfreport() to convert p20008/p20006 date fields.
.parse_ukb_year_month <- function(ym) {
  ym    <- suppressWarnings(as.numeric(ym))
  valid <- !is.na(ym) & ym > 0
  out   <- rep(NA_integer_, length(ym))
  if (any(valid)) {
    yr  <- floor(ym[valid])
    mo  <- pmax(1L, pmin(12L, round((ym[valid] - yr) * 12)))
    out[valid] <- as.integer(
      data.table::as.IDate(sprintf("%04d-%02d-01", as.integer(yr), as.integer(mo)))
    )
  }
  structure(out, class = c("IDate", "Date"))
}


# Detect a single First Occurrence column for a given UKB field ID.
# First Occurrence fields (p131xxx) are single-column, no _i/_a suffix.
# Strategy:
#   1. Cache available → title → snake_case base → grep names(data)
#   2. Fallback → raw "p{field_id}" pattern (works before decode_names)
# Returns NULL if not found.
.detect_fo_col <- function(data, field_id) {
  fields_df <- tryCatch(suppressMessages(extract_ls()), error = function(e) NULL)

  if (!is.null(fields_df)) {
    idx <- grep(paste0("p", field_id), fields_df$field_name, fixed = TRUE)[1L]
    if (!is.na(idx)) {
      base <- trimws(sub("\\s*\\|.*$", "", fields_df$title[idx]))
      base <- tolower(base)
      base <- gsub("[^a-z0-9]+", "_", base)
      base <- gsub("^_+|_+$",    "",  base)
      cols <- grep(base, names(data), value = TRUE, fixed = TRUE)
      if (length(cols) > 0L) return(cols[1L])
    }
  }

  # Fallback: raw field ID pattern (before decode_names)
  cols <- grep(paste0("^p", field_id, "($|_)"), names(data), value = TRUE)
  if (length(cols) > 0L) return(cols[1L])

  NULL
}


# Detect columns in data that belong to a given UKB field ID.
# Uses extract_ls() cache: looks up field title → converts base to snake_case
# pattern → greps column names. Falls back to raw "p{id}_i" pattern when
# the cache is unavailable or decode_names() has not been run.
.detect_cols_by_field <- function(data, field_id) {
  fields_df <- tryCatch(suppressMessages(extract_ls()), error = function(e) NULL)

  if (!is.null(fields_df)) {
    # Find any entry for this field ID in the dictionary
    idx <- grep(paste0("p", field_id, "_"), fields_df$field_name, fixed = TRUE)[1L]
    if (!is.na(idx)) {
      # Extract base title (before first "|") and convert to snake_case
      base <- trimws(sub("\\s*\\|.*$", "", fields_df$title[idx]))
      base <- tolower(base)
      base <- gsub("[^a-z0-9]+", "_", base)
      base <- gsub("^_+|_+$",    "",  base)
      cols <- grep(base, names(data), value = TRUE, fixed = TRUE)
      if (length(cols) > 0L) return(cols)
    }
  }

  # Fallback: raw field ID pattern (works before decode_names)
  grep(paste0("^p", field_id, "($|_)"), names(data), value = TRUE)
}

# =============================================================================
# zzz.R — Package startup and initialisation
# =============================================================================


# ── .onAttach ─────────────────────────────────────────────────────────────────

.onAttach <- function(libname, pkgname) {
  if (!interactive()) return()
  version <- utils::packageVersion(pkgname)

  # Reason: cli_inform() routes through message(), so users can suppress with
  # suppressPackageStartupMessages(). cli_text() uses cat() and cannot be suppressed.
  cli::cli_inform(c(
    "i" = "Welcome to {.pkg {pkgname}} v{version}",
    "i" = "Streamlined workflow for UK Biobank data analysis",
    "i" = "Run {.run ops_setup()} to check your environment"
  ))
}


# ── R CMD check — data.table NSE ──────────────────────────────────────────────

utils::globalVariables(c(".SD", ".N", ".I", ".GRP", ".BY", ".EACHI"))

# =============================================================================
# utils_grs.R - Internal helpers for grs_ series
# =============================================================================


# Generate the plink2 driver R script executed on RAP via Swiss Army Knife.
# Faithfully mirrors scripts 06 (standard) and 13 (large), differing only
# in n_threads and plink_memory which are injected at generation time.
#
# CRITICAL: output_dir MUST be "/home/dnanexus/out/out/" - Swiss Army Knife
# auto-uploads everything in that path to the job --destination on RAP.
.bgen2pgen_script <- function(n_threads, plink_memory, maf = 0.01) {
  maf_label <- sprintf("%03d", round(maf * 100))
  c(
    "#!/usr/bin/env Rscript",
    "# Auto-generated by ukbflow::grs_bgen2pgen()",
    "args <- commandArgs(trailingOnly = TRUE)",
    'if (length(args) == 0L) stop("Usage: Rscript grs_bgen2pgen.R <chr>")',
    "chr <- as.integer(args[1L])",
    'if (is.na(chr) || chr < 1L || chr > 22L) stop("chr must be 1-22")',
    "",
    "suppressPackageStartupMessages(library(cli))",
    "",
    'cli_h1("BGEN -> PGEN: Chromosome {chr}")',
    "",
    "# --- Paths (fixed UKB RAP layout) ---",
    'geno_dir    <- "/mnt/project/Bulk/Imputation/UKB imputation from genotype"',
    'bgen_file   <- file.path(geno_dir, paste0("ukb22828_c", chr, "_b0_v3.bgen"))',
    'sample_file <- file.path(geno_dir, "ukb22828_c1_b0_v3.sample")',
    "",
    "# CRITICAL: Swiss Army Knife auto-uploads /home/dnanexus/out/out/ to --destination",
    'output_dir <- "/home/dnanexus/out/out"',
    paste0('out_prefix <- file.path(output_dir, paste0("chr", chr, "_maf', maf_label, '"))'),
    "dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)",
    "",
    paste0("n_threads    <- ", n_threads, "L"),
    paste0("plink_memory <- ", plink_memory, "L"),
    paste0("maf_filter   <- ", maf),
    "",
    "# --- Input checks ---",
    'if (!file.exists(bgen_file)) {',
    '  cli_alert_danger("BGEN not found: {bgen_file}")',
    '  quit(save = "no", status = 1L)',
    "}",
    'if (!file.exists(sample_file)) {',
    '  cli_alert_danger("Sample file not found: {sample_file}")',
    '  quit(save = "no", status = 1L)',
    "}",
    "",
    "# --- Skip if already done ---",
    'if (file.exists(paste0(out_prefix, ".pgen"))) {',
    '  cli_alert_warning("PGEN already exists, skipping.")',
    '  quit(save = "no", status = 0L)',
    "}",
    "",
    "# --- Install / locate plink2 ---",
    'plink_bin <- file.path(Sys.getenv("HOME"), "plink2")',
    'if (!file.exists(plink_bin)) {',
    '  cli_alert_info("Installing plink2 ...")',
    "  # Pinned version - matches original analysis environment on UKB RAP",
    '  url      <- "https://s3.amazonaws.com/plink2-assets/alpha5/plink2_linux_x86_64_20240105.zip"',
    '  dest_zip <- file.path(tempdir(), "plink2.zip")',
    '  download.file(url, dest_zip, quiet = FALSE, mode = "wb")',
    '  unzip(dest_zip, exdir = Sys.getenv("HOME"))',
    '  system(sprintf("chmod +x %s", shQuote(plink_bin)))',
    '  file.remove(dest_zip)',
    '  cli_alert_success("plink2 installed")',
    "} else {",
    '  cli_alert_success("plink2 found")',
    "}",
    "ver <- system(sprintf(\"%s --version\", shQuote(plink_bin)), intern = TRUE)",
    "cli_alert_info(\"Version: {ver[1L]}\")",
    "",
    "# --- plink2: BGEN ->  PGEN ---",
    "plink_cmd <- sprintf(",
    '  "%s --bgen %s ref-first --sample %s --maf %s --make-pgen --out %s --threads %d --memory %d",',
    "  shQuote(plink_bin), shQuote(bgen_file), shQuote(sample_file),",
    "  maf_filter, shQuote(out_prefix), n_threads, plink_memory",
    ")",
    'cli_alert_info("Running plink2 ...")',
    "t0     <- Sys.time()",
    "status <- system(plink_cmd)",
    'elapsed <- round(difftime(Sys.time(), t0, units = "mins"), 1L)',
    "",
    "# --- Check result ---",
    'pgen_file <- paste0(out_prefix, ".pgen")',
    "if (file.exists(pgen_file) && status == 0L) {",
    '  sz <- round(file.size(pgen_file) / 1024^3, 2)',
    '  cli_alert_success("Done in {elapsed} min -- {pgen_file} ({sz} GB)")',
    '  log_file <- paste0(out_prefix, ".log")',
    "  if (file.exists(log_file)) {",
    "    ll    <- readLines(log_file)",
    '    vline <- grep("variants remaining", ll, value = TRUE)',
    '    if (length(vline) > 0L) cli_alert_info("{vline[length(vline)]}")',
    "  }",
    '  quit(save = "no", status = 0L)',
    "} else {",
    '  cli_alert_danger("plink2 failed -- check {out_prefix}.log")',
    '  quit(save = "no", status = 1L)',
    "}"
  )
}


# Generate the plink2 scoring driver R script executed on RAP via Swiss Army Knife.
# n_threads and plink_memory are injected at generation time (same pattern as
# .bgen2pgen_script).  The script accepts three command-line args:
#   <weight_file>   basename of the weight file (available at /mnt/project/<name>)
#   <score_name>    GRS column name written in the output CSV (e.g. GRS_gwas_cscc)
#   <pgen_dir>      directory of PGEN files on RAP
#
# Output: /home/dnanexus/out/out/<score_name>_scores.csv
# (Swiss Army Knife auto-uploads that path to the job --destination on RAP.)
.grs_score_script <- function(n_threads, plink_memory, maf = 0.01) {
  maf_label <- sprintf("%03d", round(maf * 100))
  c(
    "#!/usr/bin/env Rscript",
    "# Auto-generated by ukbflow::grs_score()",
    "args <- commandArgs(trailingOnly = TRUE)",
    'if (length(args) < 3L) stop("Usage: Rscript grs_score_driver.R <weight_file> <score_name> <pgen_dir>")',
    "weight_file <- args[1L]",
    "score_name  <- args[2L]",
    "pgen_dir    <- args[3L]",
    "",
    "suppressPackageStartupMessages({",
    "  library(data.table)",
    "  library(cli)",
    "})",
    "",
    'cli_h1("GRS scoring: {score_name}")',
    "",
    "# --- Paths ---",
    'weight_path <- file.path("/mnt/project", weight_file)',
    'output_dir  <- "/home/dnanexus/out/out"',
    "dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)",
    "",
    paste0("n_threads    <- ", n_threads, "L"),
    paste0("plink_memory <- ", plink_memory, "L"),
    "chromosomes  <- 1:22",
    "",
    "# --- Check weight file ---",
    'if (!file.exists(weight_path)) {',
    '  cli_alert_danger("Weight file not found: {weight_path}")',
    '  quit(save = "no", status = 1L)',
    "}",
    "",
    "# --- Load + standardise weights (take first 3 cols: ID, A1, WEIGHT) ---",
    "w <- fread(weight_path, data.table = TRUE)",
    'if (ncol(w) < 3L) stop("Weight file must have >= 3 columns: ID, A1, WEIGHT")',
    "w <- w[, 1:3]",
    'setnames(w, c("ID", "A1", "WEIGHT"))',
    'cli_alert_success("Loaded {nrow(w)} SNPs")',
    "",
    "# Write standardised weight and variant-list temp files for plink2",
    'wf_tmp <- file.path(tempdir(), "weights_std.txt")',
    'vl_tmp <- file.path(tempdir(), "variants.txt")',
    'fwrite(w, wf_tmp, sep = "\\t")',
    "fwrite(w[, .(ID)], vl_tmp, col.names = FALSE)",
    "",
    "# --- Install / locate plink2 (pinned to alpha5/20240105) ---",
    'plink_bin <- file.path(Sys.getenv("HOME"), "plink2")',
    'if (!file.exists(plink_bin)) {',
    '  cli_alert_info("Installing plink2 ...")',
    '  url      <- "https://s3.amazonaws.com/plink2-assets/alpha5/plink2_linux_x86_64_20240105.zip"',
    '  dest_zip <- file.path(tempdir(), "plink2.zip")',
    '  download.file(url, dest_zip, quiet = FALSE, mode = "wb")',
    '  unzip(dest_zip, exdir = Sys.getenv("HOME"))',
    '  system(sprintf("chmod +x %s", shQuote(plink_bin)))',
    '  file.remove(dest_zip)',
    '  cli_alert_success("plink2 installed")',
    "} else {",
    '  cli_alert_success("plink2 found")',
    "}",
    'ver <- system(sprintf("%s --version", shQuote(plink_bin)), intern = TRUE)',
    'cli_alert_info("plink2 version: {ver[1L]}")',
    "",
    "# --- Score per chromosome ---",
    "chr_results <- list()",
    "failed_chrs <- integer(0L)",
    "",
    "for (chr in chromosomes) {",
    paste0('  pgen_prefix <- file.path(pgen_dir, paste0("chr", chr, "_maf', maf_label, '"))'),
    '  out_prefix  <- file.path(tempdir(), paste0(score_name, "_chr", chr))',
    "",
    "  plink_cmd <- sprintf(",
    '    "%s --pfile %s --extract %s --score %s 1 2 3 header cols=+scoresums --out %s --threads %d --memory %d",',
    "    shQuote(plink_bin), shQuote(pgen_prefix), shQuote(vl_tmp),",
    "    shQuote(wf_tmp), shQuote(out_prefix), n_threads, plink_memory",
    "  )",
    "",
    "  status     <- system(plink_cmd, ignore.stdout = TRUE, ignore.stderr = TRUE)",
    '  score_file <- paste0(out_prefix, ".sscore")',
    "",
    "  if (status == 0L && file.exists(score_file)) {",
    "    chr_dt <- fread(score_file, data.table = TRUE)",
    '    setnames(chr_dt, gsub("^#", "", names(chr_dt)))',
    '    if (nrow(chr_dt) > 0L && "IID" %in% names(chr_dt) && "SCORE1_SUM" %in% names(chr_dt)) {',
    '      chr_results[[as.character(chr)]] <- chr_dt[, .(IID, SCORE1_SUM)]',
    '      cli_alert_success("chr{chr}: {format(nrow(chr_dt), big.mark=\',\')} individuals")',
    "    } else {",
    "      failed_chrs <- c(failed_chrs, chr)",
    '      cli_alert_warning("chr{chr}: empty result (no matching SNPs?)")',
    "    }",
    "  } else {",
    "    failed_chrs <- c(failed_chrs, chr)",
    '    cli_alert_warning("chr{chr}: plink2 failed (exit status {status})")',
    "  }",
    "}",
    "",
    'if (length(chr_results) == 0L) stop("All chromosomes failed for ", score_name)',
    'cli_alert_info("{length(chr_results)}/22 chromosomes succeeded; {length(failed_chrs)} failed")',
    "",
    "# --- Merge chromosomes: sum SCORE1_SUM across all successful chromosomes ---",
    "# nxt[merged, ...] is a right-join on merged; NA score_next (chr absent) -> 0",
    "merged <- chr_results[[1L]]",
    'setnames(merged, "SCORE1_SUM", "score_sum")',
    "",
    "if (length(chr_results) > 1L) {",
    "  for (k in 2:length(chr_results)) {",
    "    nxt <- chr_results[[k]]",
    '    setnames(nxt, "SCORE1_SUM", "score_next")',
    '    merged <- nxt[merged, on = "IID"]',
    "    merged[is.na(score_next), score_next := 0]",
    "    merged[, score_sum := score_sum + score_next]",
    "    merged[, score_next := NULL]",
    "  }",
    "}",
    "",
    'setnames(merged, "score_sum", score_name)',
    "",
    "# --- Summary ---",
    "v <- merged[[score_name]]",
    'cli_alert_success("N={format(nrow(merged), big.mark=\',\')}, mean={round(mean(v), 4)}, sd={round(sd(v), 4)}, range=[{round(min(v), 4)}, {round(max(v), 4)}]")',
    "",
    "# --- Save to Swiss Army Knife auto-upload path ---",
    'out_file <- file.path(output_dir, paste0(score_name, "_scores.csv"))',
    "fwrite(merged, out_file)",
    'cli_alert_success("Saved: {out_file}")',
    "",
    'quit(save = "no", status = 0L)'
  )
}

# =============================================================================
# assoc.R - association analysis for UKB survival data
# =============================================================================


#' Cox proportional hazards association analysis
#'
#' Fits one or more Cox models for each exposure variable and returns a tidy
#' result table suitable for downstream forest plots. By default, two standard
#' adjustment models are always included alongside any user-specified model:
#'
#' \itemize{
#'   \item \strong{Unadjusted} - no covariates (crude).
#'   \item \strong{Age and sex adjusted} - age + sex auto-detected from the
#'     data via UKB field IDs (21022 and 31). Skipped with a warning if either
#'     column cannot be found.
#'   \item \strong{Fully adjusted} - the covariates supplied via the
#'     \code{covariates} argument. Only run when \code{covariates} is non-NULL.
#' }
#'
#' \strong{Outcome coding}: \code{outcome_col} may be \code{logical}
#' (\code{TRUE}/\code{FALSE}) or integer/numeric (\code{0}/\code{1}).
#' Logical values are converted to integer internally.
#'
#' \strong{Exposure types supported}:
#' \itemize{
#'   \item \emph{Binary} - \code{0}/\code{1} or \code{TRUE}/\code{FALSE};
#'     produces one \code{term} row per model.
#'   \item \emph{Factor} - produces one \code{term} row per non-reference level.
#'   \item \emph{Numeric} (continuous) - produces one \code{term} row per model.
#' }
#'
#' @param data (data.frame or data.table) Analysis dataset. Must contain all
#'   columns referenced by \code{outcome_col}, \code{time_col}, and
#'   \code{exposure_col}.
#' @param outcome_col (character) Name of the event indicator column.
#'   Accepts \code{logical} (\code{TRUE}/\code{FALSE}) or numeric/integer
#'   (\code{0}/\code{1}).
#' @param time_col (character) Name of the follow-up time column (numeric,
#'   in consistent units, e.g. years).
#' @param exposure_col (character) One or more exposure variable names.
#'   Each variable is analysed separately; results are stacked row-wise.
#' @param covariates (character or NULL) Additional covariate column names for
#'   the \strong{Fully adjusted} model (e.g.
#'   \code{c("tdi", "smoking", paste0("pc", 1:10))}). When \code{NULL}
#'   (default), the Fully adjusted model is not run.
#' @param base (logical) If \code{TRUE} (default), always include the
#'   \strong{Unadjusted} and \strong{Age and sex adjusted} models in addition
#'   to any user-specified \code{covariates} model. Set to \code{FALSE} to run
#'   only the Fully adjusted model (requires \code{covariates} to be non-NULL).
#' @param strata (character or NULL) Optional stratification variable.
#'   Passed to \code{survival::strata()} in the Cox formula.
#' @param conf_level (numeric) Confidence level for hazard ratio intervals.
#'   Default: \code{0.95}.
#'
#' @return A \code{data.table} with one row per exposure \eqn{\times} term
#'   \eqn{\times} model combination, and the following columns:
#'   \describe{
#'     \item{\code{exposure}}{Exposure variable name.}
#'     \item{\code{term}}{Coefficient name as returned by \code{coxph} (e.g.
#'       \code{"bmi_categoryObese"} for a factor, or the variable name itself
#'       for numeric/binary exposures).}
#'     \item{\code{model}}{Ordered factor: \code{Unadjusted} <
#'       \code{Age and sex adjusted} < \code{Fully adjusted}.}
#'     \item{\code{n}}{Number of participants included in the model (after
#'       \code{NA} removal).}
#'     \item{\code{n_events}}{Number of events in the model's analysis set
#'       (after \code{NA} removal).}
#'     \item{\code{person_years}}{Total person-years of follow-up in the
#'       model's analysis set (rounded, after \code{NA} removal).}
#'     \item{\code{HR}}{Hazard ratio (point estimate).}
#'     \item{\code{CI_lower}}{Lower bound of the confidence interval.}
#'     \item{\code{CI_upper}}{Upper bound of the confidence interval.}
#'     \item{\code{p_value}}{Wald test p-value.}
#'     \item{\code{HR_label}}{Formatted string, e.g.
#'       \code{"1.23 (1.05-1.44)"}.}
#'   }
#'
#' @importFrom survival coxph Surv
#' @export
#'
#' @examples
#' dt <- ops_toy(scenario = "association", n = 500)
#' dt <- dt[dm_timing != 1L]
#'
#' res <- assoc_coxph(
#'   data         = dt,
#'   outcome_col  = "dm_status",
#'   time_col     = "dm_followup_years",
#'   exposure_col = "p20116_i0",
#'   covariates   = c("bmi_cat", "tdi_cat"),
#'   base         = FALSE
#' )
#'
#' # Fully adjusted only (skip Unadjusted + Age-sex)
#' res <- assoc_coxph(
#'   data         = dt,
#'   outcome_col  = "dm_status",
#'   time_col     = "dm_followup_years",
#'   exposure_col = "p20116_i0",
#'   covariates   = c("p21022", "p31", "bmi_cat", "tdi_cat"),
#'   base         = FALSE
#' )
assoc_coxph <- function(data,
                         outcome_col,
                         time_col,
                         exposure_col,
                         covariates  = NULL,
                         base        = TRUE,
                         strata      = NULL,
                         conf_level  = 0.95) {

  # Input validation
  .assert_data_frame(data)
  .assert_has_cols(data, c(outcome_col, time_col, exposure_col, covariates, strata))
  .assert_flag(base)

  .assert_not_null_if_false(base, covariates)

  .assert_proportion(conf_level)

  # Prepare working copy: normalise outcome + logical exposures
  dt <- data.table::copy(data.table::as.data.table(data))

  dt[, .ukb_event := .normalise_event(dt[[outcome_col]], outcome_col)]
  .normalise_logical_exposures(dt, exposure_col)

  # Build model list: name -> covariates vector (NULL = crude)
  model_list <- list()

  if (base) {
    # Reason: single-bracket + list(NULL) is required to store a NULL value in a
    # named list. Double-bracket assignment [["key"]] <- NULL silently deletes
    # the key instead of setting it, which would drop the Unadjusted model.
    model_list["Unadjusted"] <- list(NULL)

    age_col <- .detect_age_col(dt)
    sex_col <- .detect_sex_col(dt)

    if (is.null(age_col) || is.null(sex_col)) {
      cli::cli_alert_warning(
        paste0(
          "Age and sex adjusted model skipped: ",
          if (is.null(age_col)) "age column (UKB field 21022) not found" else "",
          if (is.null(age_col) && is.null(sex_col)) " and " else "",
          if (is.null(sex_col)) "sex column (UKB field 31) not found" else "",
          "."
        )
      )
    } else {
      model_list[["Age and sex adjusted"]] <- c(age_col, sex_col)
    }
  }

  if (!is.null(covariates) && length(covariates) > 0L) {
    model_list[["Fully adjusted"]] <- covariates
  }

  # Run models: each exposure x each model
  n_models    <- length(model_list)
  n_exposures <- length(exposure_col)

  cli::cli_h1("assoc_coxph")
  cli::cli_alert_info(
    "{n_exposures} exposure{?s} x {n_models} model{?s} = {n_exposures * n_models} Cox regression{?s}"
  )
  cli::cli_alert_info(
    "Input cohort: {nrow(dt)} participants (n/n_events/person_years reflect each model's actual analysis set)"
  )

  results <- vector("list", n_exposures * n_models)
  idx      <- 1L

  for (exp in exposure_col) {

    cli::cli_h2("{.field {exp}}")

    for (model_label in names(model_list)) {

      covs <- model_list[[model_label]]
      res  <- .run_one_cox_model(
        dt           = dt,
        time_col     = time_col,
        exposure     = exp,
        covariates   = covs,
        strata       = strata,
        model_label  = model_label,
        conf_level   = conf_level
      )

      if (!is.null(res)) {
        # Print each term result
        for (i in seq_len(nrow(res))) {
          cli::cli_alert_success(
            "  {model_label} | {res$term[i]}: HR {res$HR_label[i]}, p = {format.pval(res$p_value[i], digits = 3L)}"
          )
        }
        results[[idx]] <- res
      }
      idx <- idx + 1L
    }
  }

  # Combine and return
  out <- data.table::rbindlist(Filter(Negate(is.null), results))

  if (nrow(out) == 0L) {
    cli::cli_alert_warning("No results returned - check model warnings above.")
    return(out)
  }

  # Ordered factor for downstream plot ordering
  model_levels <- c("Unadjusted", "Age and sex adjusted", "Fully adjusted")
  out[, model := factor(model, levels = model_levels, ordered = TRUE)]

  cli::cli_alert_success(
    "Done: {nrow(out)} result row{?s} across {uniqueN(out$exposure)} exposure{?s} and {uniqueN(out$model)} model{?s}."
  )

  out[]
}


#' @rdname assoc_coxph
#' @export
assoc_cox <- assoc_coxph


#' Logistic regression association analysis
#'
#' Fits one or more logistic regression models for each exposure variable and
#' returns a tidy result table suitable for downstream forest plots. By default,
#' two standard adjustment models are always included:
#'
#' \itemize{
#'   \item \strong{Unadjusted} - no covariates (crude).
#'   \item \strong{Age and sex adjusted} - age + sex auto-detected from the
#'     data via UKB field IDs (21022 and 31). Skipped with a warning if either
#'     column cannot be found.
#'   \item \strong{Fully adjusted} - the covariates supplied via the
#'     \code{covariates} argument. Only run when \code{covariates} is non-NULL.
#' }
#'
#' \strong{Outcome coding}: \code{outcome_col} may be \code{logical}
#' (\code{TRUE}/\code{FALSE}) or integer/numeric (\code{0}/\code{1}).
#' Logical values are converted to integer internally.
#'
#' \strong{CI methods}:
#' \itemize{
#'   \item \code{"wald"} (default) - fast, appropriate for large UKB samples.
#'   \item \code{"profile"} - profile likelihood CI via \code{confint.glm()};
#'     slower but more accurate for small or sparse data.
#' }
#'
#' @param data (data.frame or data.table) Analysis dataset.
#' @param outcome_col (character) Binary outcome column (\code{0}/\code{1} or
#'   \code{TRUE}/\code{FALSE}).
#' @param exposure_col (character) One or more exposure variable names.
#' @param covariates (character or NULL) Covariate column names for the
#'   \strong{Fully adjusted} model. Default: \code{NULL}.
#' @param base (logical) Include \strong{Unadjusted} and \strong{Age and sex
#'   adjusted} models. Default: \code{TRUE}.
#' @param ci_method (character) CI calculation method: \code{"wald"} (default)
#'   or \code{"profile"}.
#' @param conf_level (numeric) Confidence level. Default: \code{0.95}.
#'
#' @return A \code{data.table} with one row per exposure \eqn{\times} term
#'   \eqn{\times} model combination, and columns:
#'   \describe{
#'     \item{\code{exposure}}{Exposure variable name.}
#'     \item{\code{term}}{Coefficient name (e.g. \code{"bmi_categoryObese"}).}
#'     \item{\code{model}}{Ordered factor: \code{Unadjusted} <
#'       \code{Age and sex adjusted} < \code{Fully adjusted}.}
#'     \item{\code{n}}{Participants in model (after NA removal).}
#'     \item{\code{n_cases}}{Number of cases (outcome = 1) in model.}
#'     \item{\code{OR}}{Odds ratio (point estimate).}
#'     \item{\code{CI_lower}}{Lower confidence bound.}
#'     \item{\code{CI_upper}}{Upper confidence bound.}
#'     \item{\code{p_value}}{Wald test p-value.}
#'     \item{\code{OR_label}}{Formatted string, e.g. \code{"1.23 (1.05-1.44)"}.}
#'   }
#'
#' @export
#'
#' @examples
#' dt <- ops_toy(scenario = "association", n = 500)
#'
#' res <- assoc_logistic(
#'   data         = dt,
#'   outcome_col  = "dm_status",
#'   exposure_col = "p20116_i0",
#'   covariates   = c("bmi_cat", "tdi_cat"),
#'   base         = FALSE
#' )
assoc_logistic <- function(data,
                            outcome_col,
                            exposure_col,
                            covariates = NULL,
                            base       = TRUE,
                            ci_method  = c("wald", "profile"),
                            conf_level = 0.95) {

  ci_method <- match.arg(ci_method)

  # Input validation
  .assert_data_frame(data)
  .assert_has_cols(data, c(outcome_col, exposure_col, covariates))
  .assert_flag(base)

  .assert_not_null_if_false(base, covariates)

  .assert_proportion(conf_level)

  # Prepare working copy
  dt <- data.table::copy(data.table::as.data.table(data))

  dt[, .ukb_event := .normalise_event(dt[[outcome_col]], outcome_col)]
  .normalise_logical_exposures(dt, exposure_col)

  # Build model list
  model_list <- list()

  if (base) {
    model_list["Unadjusted"] <- list(NULL)

    age_col <- .detect_age_col(dt)
    sex_col <- .detect_sex_col(dt)

    if (is.null(age_col) || is.null(sex_col)) {
      cli::cli_alert_warning(
        paste0(
          "Age and sex adjusted model skipped: ",
          if (is.null(age_col)) "age column (UKB field 21022) not found" else "",
          if (is.null(age_col) && is.null(sex_col)) " and " else "",
          if (is.null(sex_col)) "sex column (UKB field 31) not found" else "",
          "."
        )
      )
    } else {
      model_list[["Age and sex adjusted"]] <- c(age_col, sex_col)
    }
  }

  if (!is.null(covariates) && length(covariates) > 0L) {
    model_list[["Fully adjusted"]] <- covariates
  }

  # Run models: each exposure x each model
  n_models    <- length(model_list)
  n_exposures <- length(exposure_col)

  cli::cli_h1("assoc_logistic")
  cli::cli_alert_info(
    "{n_exposures} exposure{?s} x {n_models} model{?s} = {n_exposures * n_models} logistic regression{?s}"
  )
  cli::cli_alert_info(
    "Input cohort: {nrow(dt)} participants | CI method: {ci_method} (n/n_cases reflect each model's actual analysis set)"
  )

  results <- vector("list", n_exposures * n_models)
  idx      <- 1L

  for (exp in exposure_col) {

    cli::cli_h2("{.field {exp}}")

    for (model_label in names(model_list)) {

      covs <- model_list[[model_label]]
      res  <- .run_one_logistic_model(
        dt          = dt,
        exposure    = exp,
        covariates  = covs,
        model_label = model_label,
        ci_method   = ci_method,
        conf_level  = conf_level
      )

      if (!is.null(res)) {
        for (i in seq_len(nrow(res))) {
          cli::cli_alert_success(
            "  {model_label} | {res$term[i]}: OR {res$OR_label[i]}, p = {format.pval(res$p_value[i], digits = 3L)}"
          )
        }
        results[[idx]] <- res
      }
      idx <- idx + 1L
    }
  }

  # Combine and return
  out <- data.table::rbindlist(Filter(Negate(is.null), results))

  if (nrow(out) == 0L) {
    cli::cli_alert_warning("No results returned - check model warnings above.")
    return(out)
  }

  model_levels <- c("Unadjusted", "Age and sex adjusted", "Fully adjusted")
  out[, model := factor(model, levels = model_levels, ordered = TRUE)]

  cli::cli_alert_success(
    "Done: {nrow(out)} result row{?s} across {uniqueN(out$exposure)} exposure{?s} and {uniqueN(out$model)} model{?s}."
  )

  out[]
}


#' @rdname assoc_logistic
#' @export
assoc_logit <- assoc_logistic


#' Linear regression association analysis
#'
#' Fits one or more linear regression models for each exposure variable and
#' returns a tidy result table. By default, two standard adjustment models are
#' always included:
#'
#' \itemize{
#'   \item \strong{Unadjusted} - no covariates (crude).
#'   \item \strong{Age and sex adjusted} - age + sex auto-detected from the
#'     data via UKB field IDs (21022 and 31). Skipped with a warning if either
#'     column cannot be found.
#'   \item \strong{Fully adjusted} - the covariates supplied via the
#'     \code{covariates} argument. Only run when \code{covariates} is non-NULL.
#' }
#'
#' \strong{Outcome}: intended for continuous numeric variables. Passing a
#' binary (0/1) or logical column is permitted (linear probability model) but
#' will trigger a warning recommending \code{\link{assoc_logistic}} instead.
#'
#' \strong{CI method}: based on the t-distribution via \code{confint.lm()},
#' which is exact under the normal linear model assumption. There is no
#' \code{ci_method} argument (unlike \code{\link{assoc_logistic}}) as profile
#' likelihood does not apply to \code{lm}.
#'
#' \strong{SE column}: the standard error of \eqn{\beta} is included to
#' support downstream meta-analysis and GWAS-style summary statistics.
#'
#' @param data (data.frame or data.table) Analysis dataset.
#' @param outcome_col (character) Name of the continuous numeric outcome column.
#' @param exposure_col (character) One or more exposure variable names.
#' @param covariates (character or NULL) Covariate column names for the
#'   \strong{Fully adjusted} model. Default: \code{NULL}.
#' @param base (logical) Include \strong{Unadjusted} and \strong{Age and sex
#'   adjusted} models. Default: \code{TRUE}.
#' @param conf_level (numeric) Confidence level. Default: \code{0.95}.
#'
#' @return A \code{data.table} with one row per exposure \eqn{\times} term
#'   \eqn{\times} model combination, and columns:
#'   \describe{
#'     \item{\code{exposure}}{Exposure variable name.}
#'     \item{\code{term}}{Coefficient name (e.g. \code{"bmi_categoryObese"}).}
#'     \item{\code{model}}{Ordered factor: \code{Unadjusted} <
#'       \code{Age and sex adjusted} < \code{Fully adjusted}.}
#'     \item{\code{n}}{Participants in model (after NA removal).}
#'     \item{\code{beta}}{Regression coefficient (\eqn{\beta}).}
#'     \item{\code{se}}{Standard error of \eqn{\beta}.}
#'     \item{\code{CI_lower}}{Lower confidence bound.}
#'     \item{\code{CI_upper}}{Upper confidence bound.}
#'     \item{\code{p_value}}{t-test p-value.}
#'     \item{\code{beta_label}}{Formatted string, e.g.
#'       \code{"0.23 (0.05-0.41)"}.}
#'   }
#'
#' @export
#'
#' @examples
#' dt <- ops_toy(scenario = "association", n = 500)
#'
#' res <- assoc_linear(
#'   data         = dt,
#'   outcome_col  = "p21001_i0",
#'   exposure_col = "p20116_i0",
#'   covariates   = c("bmi_cat", "tdi_cat"),
#'   base         = FALSE
#' )
assoc_linear <- function(data,
                          outcome_col,
                          exposure_col,
                          covariates = NULL,
                          base       = TRUE,
                          conf_level = 0.95) {

  # Input validation
  .assert_data_frame(data)
  .assert_has_cols(data, c(outcome_col, exposure_col, covariates))
  .assert_flag(base)

  .assert_not_null_if_false(base, covariates)

  .assert_proportion(conf_level)

  # Warn if outcome looks binary - logistic regression is more appropriate
  outcome_vec <- data[[outcome_col]]
  if (is.logical(outcome_vec) ||
      (is.numeric(outcome_vec) &&
       all(stats::na.omit(unique(outcome_vec)) %in% c(0, 1)))) {
    cli::cli_alert_warning(
      "outcome_col {.field {outcome_col}} appears binary (0/1 or logical). Consider {.fn assoc_logistic} instead."
    )
  }

  if (!is.numeric(outcome_vec) && !is.logical(outcome_vec)) {
    cli::cli_abort(
      "outcome_col {.field {outcome_col}} must be numeric. Found: {class(outcome_vec)[1]}.",
      call = NULL
    )
  }

  # Prepare working copy: normalise logical exposures only
  dt <- data.table::copy(data.table::as.data.table(data))
  .normalise_logical_exposures(dt, exposure_col)

  # Build model list
  model_list <- list()

  if (base) {
    model_list["Unadjusted"] <- list(NULL)

    age_col <- .detect_age_col(dt)
    sex_col <- .detect_sex_col(dt)

    if (is.null(age_col) || is.null(sex_col)) {
      cli::cli_alert_warning(
        paste0(
          "Age and sex adjusted model skipped: ",
          if (is.null(age_col)) "age column (UKB field 21022) not found" else "",
          if (is.null(age_col) && is.null(sex_col)) " and " else "",
          if (is.null(sex_col)) "sex column (UKB field 31) not found" else "",
          "."
        )
      )
    } else {
      model_list[["Age and sex adjusted"]] <- c(age_col, sex_col)
    }
  }

  if (!is.null(covariates) && length(covariates) > 0L) {
    model_list[["Fully adjusted"]] <- covariates
  }

  # Run models: each exposure x each model
  n_models    <- length(model_list)
  n_exposures <- length(exposure_col)

  cli::cli_h1("assoc_linear")
  cli::cli_alert_info(
    "{n_exposures} exposure{?s} x {n_models} model{?s} = {n_exposures * n_models} linear regression{?s}"
  )
  cli::cli_alert_info(
    "Input cohort: {nrow(dt)} participants (n reflects each model's actual analysis set)"
  )

  results <- vector("list", n_exposures * n_models)
  idx      <- 1L

  for (exp in exposure_col) {

    cli::cli_h2("{.field {exp}}")

    for (model_label in names(model_list)) {

      covs <- model_list[[model_label]]
      res  <- .run_one_linear_model(
        dt          = dt,
        outcome_col = outcome_col,
        exposure    = exp,
        covariates  = covs,
        model_label = model_label,
        conf_level  = conf_level
      )

      if (!is.null(res)) {
        for (i in seq_len(nrow(res))) {
          cli::cli_alert_success(
            "  {model_label} | {res$term[i]}: beta {res$beta_label[i]}, p = {format.pval(res$p_value[i], digits = 3L)}"
          )
        }
        results[[idx]] <- res
      }
      idx <- idx + 1L
    }
  }

  # Combine and return
  out <- data.table::rbindlist(Filter(Negate(is.null), results))

  if (nrow(out) == 0L) {
    cli::cli_alert_warning("No results returned - check model warnings above.")
    return(out)
  }

  model_levels <- c("Unadjusted", "Age and sex adjusted", "Fully adjusted")
  out[, model := factor(model, levels = model_levels, ordered = TRUE)]

  cli::cli_alert_success(
    "Done: {nrow(out)} result row{?s} across {uniqueN(out$exposure)} exposure{?s} and {uniqueN(out$model)} model{?s}."
  )

  out[]
}


#' @rdname assoc_linear
#' @export
assoc_lm <- assoc_linear


#' Proportional hazards assumption test for Cox regression
#'
#' Tests the proportional hazards (PH) assumption using Schoenfeld residuals
#' via \code{\link[survival]{cox.zph}}. Re-fits the same models as
#' \code{\link{assoc_coxph}} (same interface) and returns a tidy result table
#' with term-level and global test statistics.
#'
#' A non-significant p-value (p > 0.05) indicates the PH assumption is
#' satisfied for that term. The global test (\code{global_p}) reflects the
#' overall PH assumption for the whole model.
#'
#' @param data (data.frame or data.table) Analysis dataset.
#' @param outcome_col (character) Binary event indicator (\code{0}/\code{1}
#'   or \code{TRUE}/\code{FALSE}).
#' @param time_col (character) Follow-up time column name.
#' @param exposure_col (character) One or more exposure variable names.
#' @param covariates (character or NULL) Covariates for the Fully adjusted
#'   model. Default: \code{NULL}.
#' @param base (logical) Include Unadjusted and Age and sex adjusted models.
#'   Default: \code{TRUE}.
#' @param strata (character or NULL) Optional stratification variable.
#'
#' @return A \code{data.table} with one row per exposure \eqn{\times} term
#'   \eqn{\times} model combination, and columns:
#'   \describe{
#'     \item{\code{exposure}}{Exposure variable name.}
#'     \item{\code{term}}{Coefficient name.}
#'     \item{\code{model}}{Ordered factor: \code{Unadjusted} <
#'       \code{Age and sex adjusted} < \code{Fully adjusted}.}
#'     \item{\code{chisq}}{Schoenfeld residual chi-squared statistic.}
#'     \item{\code{df}}{Degrees of freedom.}
#'     \item{\code{p_value}}{P-value for the PH test (term-level).}
#'     \item{\code{ph_satisfied}}{Logical; \code{TRUE} if \code{p_value > 0.05}.}
#'     \item{\code{global_chisq}}{Global chi-squared for the whole model.}
#'     \item{\code{global_df}}{Global degrees of freedom.}
#'     \item{\code{global_p}}{Global p-value for the whole model.}
#'   }
#'
#' @importFrom survival cox.zph
#' @export
#'
#' @examples
#' dt <- ops_toy(scenario = "association", n = 500)
#' dt <- dt[dm_timing != 1L]
#'
#' zph <- assoc_coxph_zph(
#'   data         = dt,
#'   outcome_col  = "dm_status",
#'   time_col     = "dm_followup_years",
#'   exposure_col = "p20116_i0",
#'   covariates   = c("bmi_cat", "tdi_cat"),
#'   base         = FALSE
#' )
#' zph[ph_satisfied == FALSE]
assoc_coxph_zph <- function(data,
                              outcome_col,
                              time_col,
                              exposure_col,
                              covariates = NULL,
                              base       = TRUE,
                              strata     = NULL) {

  # Input validation
  .assert_data_frame(data)
  .assert_has_cols(data, c(outcome_col, time_col, exposure_col, covariates, strata))
  .assert_flag(base)

  .assert_not_null_if_false(base, covariates)

  # Prepare working copy
  dt <- data.table::copy(data.table::as.data.table(data))
  dt[, .ukb_event := .normalise_event(dt[[outcome_col]], outcome_col)]
  .normalise_logical_exposures(dt, exposure_col)

  # Build model list (identical logic to assoc_coxph)
  model_list <- list()

  if (base) {
    model_list["Unadjusted"] <- list(NULL)

    age_col <- .detect_age_col(dt)
    sex_col <- .detect_sex_col(dt)

    if (is.null(age_col) || is.null(sex_col)) {
      cli::cli_alert_warning(
        paste0(
          "Age and sex adjusted model skipped: ",
          if (is.null(age_col)) "age column (UKB field 21022) not found" else "",
          if (is.null(age_col) && is.null(sex_col)) " and " else "",
          if (is.null(sex_col)) "sex column (UKB field 31) not found" else "",
          "."
        )
      )
    } else {
      model_list[["Age and sex adjusted"]] <- c(age_col, sex_col)
    }
  }

  if (!is.null(covariates) && length(covariates) > 0L) {
    model_list[["Fully adjusted"]] <- covariates
  }

  # Run zph tests: each exposure x each model
  n_models    <- length(model_list)
  n_exposures <- length(exposure_col)

  cli::cli_h1("assoc_coxph_zph")
  cli::cli_alert_info(
    "{n_exposures} exposure{?s} x {n_models} model{?s} = {n_exposures * n_models} PH assumption test{?s}"
  )

  results <- vector("list", n_exposures * n_models)
  idx      <- 1L

  for (exp in exposure_col) {

    cli::cli_h2("{.field {exp}}")

    for (model_label in names(model_list)) {

      covs <- model_list[[model_label]]
      res  <- .run_one_zph_test(
        dt          = dt,
        time_col    = time_col,
        exposure    = exp,
        covariates  = covs,
        strata      = strata,
        model_label = model_label
      )

      if (!is.null(res)) {
        for (i in seq_len(nrow(res))) {
          status <- if (res$ph_satisfied[i]) "[OK] satisfied" else "[!!] VIOLATED"
          cli::cli_alert_success(
            "  {model_label} | {res$term[i]}: chisq = {round(res$chisq[i], 3)}, p = {format.pval(res$p_value[i], digits = 3L)} {status}"
          )
        }
        cli::cli_alert_info(
          "  Global: chisq = {round(res$global_chisq[1], 3)}, p = {format.pval(res$global_p[1], digits = 3L)}"
        )
        results[[idx]] <- res
      }
      idx <- idx + 1L
    }
  }

  # Combine and return
  out <- data.table::rbindlist(Filter(Negate(is.null), results))

  if (nrow(out) == 0L) {
    cli::cli_alert_warning("No results returned - check model warnings above.")
    return(out)
  }

  model_levels <- c("Unadjusted", "Age and sex adjusted", "Fully adjusted")
  out[, model := factor(model, levels = model_levels, ordered = TRUE)]

  n_violated <- sum(!out$ph_satisfied)
  if (n_violated == 0L) {
    cli::cli_alert_success(
      "Done: all {nrow(out)} term{?s} satisfy the PH assumption."
    )
  } else {
    cli::cli_alert_warning(
      "Done: {n_violated}/{nrow(out)} term{?s} violate the PH assumption (p <= 0.05) -- consider {.code strata()} or time-varying effects."
    )
  }

  out[]
}


#' @rdname assoc_coxph_zph
#' @export
assoc_zph <- assoc_coxph_zph


#' Subgroup association analysis with optional interaction test
#'
#' Stratifies the dataset by a single grouping variable (\code{by}) and runs
#' the specified association model (\code{coxph}, \code{logistic}, or
#' \code{linear}) within each subgroup. Unlike \code{\link{assoc_coxph}} and
#' its siblings, there is no automatic age-and-sex-adjusted model: at the
#' subgroup level the variable that defines the stratum would have zero
#' variance, making the auto-detected adjustment meaningless. Accordingly,
#' this function does not accept a \code{base} argument. Instead, two
#' models are available:
#'
#' \itemize{
#'   \item \strong{Unadjusted} - always run (no covariates).
#'   \item \strong{Fully adjusted} - run when \code{covariates} is non-NULL.
#'     Users are responsible for excluding the \code{by} variable from
#'     \code{covariates} (a warning is issued if it is included).
#' }
#'
#' \strong{Interaction test}: when \code{interaction = TRUE} (default), a
#' likelihood ratio test (LRT) for the exposure \eqn{\times} by interaction is
#' computed on the \emph{full} dataset for each exposure \eqn{\times} model
#' combination and appended as \code{p_interaction}. LRT is preferred over
#' Wald because it handles factor, binary, and continuous \code{by} variables
#' uniformly without requiring the user to recode the \code{by} variable.
#'
#' @param data (data.frame or data.table) Analysis dataset.
#' @param outcome_col (character) Outcome column name.
#' @param time_col (character or NULL) Follow-up time column (required when
#'   \code{method = "coxph"}).
#' @param exposure_col (character) One or more exposure variable names.
#' @param by (character) Single stratification variable name. Its unique
#'   non-NA values (or factor levels, in order) define the subgroups.
#'   Should be a categorical or binary variable with a small number of
#'   levels (e.g. sex, smoking status). Continuous variables are technically
#'   permitted and the interaction LRT will still run, but per-unique-value
#'   subgrouping is rarely meaningful in practice — use a pre-categorised
#'   version (e.g. via \code{\link{derive_cut}}) instead.
#' @param method (character) Regression method: \code{"coxph"} (default),
#'   \code{"logistic"}, or \code{"linear"}.
#' @param covariates (character or NULL) Covariate column names for the Fully
#'   adjusted model. When \code{NULL}, only the Unadjusted model is run.
#' @param interaction (logical) Compute the LRT p-value for the exposure
#'   \eqn{\times} by interaction on the full dataset. Default: \code{TRUE}.
#' @param conf_level (numeric) Confidence level. Default: \code{0.95}.
#'
#' @return A \code{data.table} with one row per subgroup level \eqn{\times}
#'   exposure \eqn{\times} term \eqn{\times} model, containing:
#'   \describe{
#'     \item{\code{subgroup}}{Name of the \code{by} variable.}
#'     \item{\code{subgroup_level}}{Factor: level of the \code{by} variable
#'       (ordered by original level / sort order).}
#'     \item{\code{exposure}}{Exposure variable name.}
#'     \item{\code{term}}{Coefficient name from the fitted model.}
#'     \item{\code{model}}{Ordered factor: \code{Unadjusted} <
#'       \code{Fully adjusted}.}
#'     \item{\code{n}}{Participants in model (after NA removal).}
#'     \item{...}{Effect estimate columns: \code{HR}/\code{OR}/\code{beta},
#'       \code{CI_lower}, \code{CI_upper}, \code{p_value}, and a formatted
#'       label column. Cox models additionally include \code{n_events} and
#'       \code{person_years}; logistic models include \code{n_cases}; linear
#'       models include \code{se}.}
#'     \item{\code{p_interaction}}{LRT p-value for the exposure \eqn{\times}
#'       by interaction on the full dataset. Shared across all subgroup levels
#'       for the same exposure \eqn{\times} model. \code{NA} when the
#'       interaction model fails. Only present when
#'       \code{interaction = TRUE}.}
#'   }
#'
#' @importFrom survival coxph Surv
#' @export
#'
#' @examples
#' dt <- ops_toy(scenario = "association", n = 500)
#' dt <- dt[dm_timing != 1L]
#'
#' res <- assoc_subgroup(
#'   data         = dt,
#'   outcome_col  = "dm_status",
#'   time_col     = "dm_followup_years",
#'   exposure_col = "p20116_i0",
#'   by           = "p31",
#'   method       = "coxph",
#'   covariates   = c("bmi_cat", "tdi_cat")
#' )
assoc_subgroup <- function(data,
                            outcome_col,
                            time_col    = NULL,
                            exposure_col,
                            by,
                            method      = c("coxph", "logistic", "linear"),
                            covariates  = NULL,
                            interaction = TRUE,
                            conf_level  = 0.95) {

  method <- match.arg(method)

  # Input validation
  .assert_data_frame(data)
  .assert_scalar_string(by)
  .assert_flag(interaction)
  .assert_has_cols(data, c(outcome_col, time_col, exposure_col, by, covariates))

  if (method == "coxph" && is.null(time_col)) {
    cli::cli_abort("{.arg time_col} is required when {.arg method = 'coxph'}.", call = NULL)
  }
  .assert_proportion(conf_level)

  # Warn if by is also listed as a covariate (collinearity within subgroups)
  if (!is.null(covariates) && by %in% covariates) {
    cli::cli_alert_warning(
      "{.arg by} variable {.field {by}} is also in {.arg covariates} -- collinearity risk within each subgroup."
    )
  }

  # Prepare working copy: normalise outcome + logical exposures
  dt <- data.table::copy(data.table::as.data.table(data))

  if (method %in% c("coxph", "logistic")) {
    dt[, .ukb_event := .normalise_event(dt[[outcome_col]], outcome_col)]
  }
  .normalise_logical_exposures(dt, exposure_col)

  # Build model list (no age/sex auto-detection at subgroup level)
  model_list <- list()
  model_list["Unadjusted"] <- list(NULL)
  if (!is.null(covariates) && length(covariates) > 0L) {
    model_list[["Fully adjusted"]] <- covariates
  }

  # Subgroup levels (preserve factor order; sort otherwise)
  by_vec    <- dt[[by]]
  levels_by <- if (is.factor(by_vec)) {
    levels(by_vec)
  } else {
    sort(unique(stats::na.omit(as.character(by_vec))))
  }

  n_levels    <- length(levels_by)
  n_models    <- length(model_list)
  n_exposures <- length(exposure_col)

  cli::cli_h1("assoc_subgroup")
  cli::cli_alert_info(
    "{n_exposures} exposure{?s} x {n_models} model{?s} x {n_levels} subgroup{?s} ({.field {by}})"
  )

  # Interaction LRT on full data (one p-value per exposure x model)
  interaction_dt <- NULL
  if (interaction) {
    cli::cli_alert_info(
      "Computing interaction LRT (exposure x {.field {by}}) on full data ..."
    )
    inter_rows <- vector("list", n_exposures * n_models)
    k <- 1L

    for (exp in exposure_col) {
      for (model_label in names(model_list)) {
        covs  <- model_list[[model_label]]
        p_int <- .run_one_interaction_lrt(
          dt          = dt,
          method      = method,
          outcome_col = outcome_col,
          time_col    = time_col,
          exposure    = exp,
          by          = by,
          covariates  = covs,
          model_label = model_label
        )
        p_fmt <- if (is.na(p_int)) "NA" else format.pval(p_int, digits = 3L)
        cli::cli_alert_info(
          "  {model_label} | {.field {exp}}: p_interaction = {p_fmt}"
        )
        inter_rows[[k]] <- data.table::data.table(
          exposure      = exp,
          model         = model_label,
          p_interaction = p_int
        )
        k <- k + 1L
      }
    }
    interaction_dt <- data.table::rbindlist(inter_rows)
  }

  # Subgroup loop: filter -> run models -> collect results
  all_results <- vector("list", n_levels * n_exposures * n_models)
  idx <- 1L

  for (lv in levels_by) {

    sub_dt <- dt[as.character(dt[[by]]) == lv]
    cli::cli_h2("{.field {by}} = {lv}  (n = {nrow(sub_dt)})")

    # Warn if events are sparse for this subgroup
    if (method %in% c("coxph", "logistic")) {
      n_ev <- sum(sub_dt$.ukb_event, na.rm = TRUE)
      if (n_ev < 10L) {
        cli::cli_alert_warning(
          "  {.field {by}} = {lv}: only {n_ev} event{?s} -- results may be unstable."
        )
      }
    }

    for (exp in exposure_col) {

      cli::cli_h3("{.field {exp}}")

      for (model_label in names(model_list)) {

        covs      <- model_list[[model_label]]
        warn_label <- sprintf("[%s=%s] %s", by, lv, model_label)

        res <- switch(method,
          coxph = .run_one_cox_model(
            dt          = sub_dt,
            time_col    = time_col,
            exposure    = exp,
            covariates  = covs,
            strata      = NULL,
            model_label = warn_label,
            conf_level  = conf_level
          ),
          logistic = .run_one_logistic_model(
            dt          = sub_dt,
            exposure    = exp,
            covariates  = covs,
            model_label = warn_label,
            ci_method   = "wald",
            conf_level  = conf_level
          ),
          linear = .run_one_linear_model(
            dt          = sub_dt,
            outcome_col = outcome_col,
            exposure    = exp,
            covariates  = covs,
            model_label = warn_label,
            conf_level  = conf_level
          )
        )

        if (!is.null(res)) {
          # Overwrite model column with clean label (helpers store warn_label)
          res[, model := model_label]
          res[, `:=`(subgroup = by, subgroup_level = lv)]

          # Print per-term results
          for (i in seq_len(nrow(res))) {
            est_label <- switch(method,
              coxph    = sprintf("HR %s",       res$HR_label[i]),
              logistic = sprintf("OR %s",       res$OR_label[i]),
              linear   = sprintf("beta %s",   res$beta_label[i])
            )
            cli::cli_alert_success(
              "  {model_label} | {res$term[i]}: {est_label}, p = {format.pval(res$p_value[i], digits = 3L)}"
            )
          }
          all_results[[idx]] <- res
        }
        idx <- idx + 1L
      }
    }
  }

  # Combine and finalise
  out <- data.table::rbindlist(Filter(Negate(is.null), all_results))

  if (nrow(out) == 0L) {
    cli::cli_alert_warning("No results returned -- check model warnings above.")
    return(out)
  }

  # Merge p_interaction via update-by-reference join (model still character)
  if (!is.null(interaction_dt)) {
    out[interaction_dt, on = c("exposure", "model"),
        p_interaction := i.p_interaction]
  }

  # Ordered factor for model
  model_levels   <- c("Unadjusted", "Fully adjusted")
  present_levels <- intersect(model_levels, as.character(unique(out$model)))
  out[, model := factor(model, levels = present_levels, ordered = TRUE)]

  # Factor for subgroup_level (preserve original level order)
  out[, subgroup_level := factor(subgroup_level, levels = levels_by)]

  # Put subgroup identifiers first
  front_cols <- c("subgroup", "subgroup_level", "exposure", "term", "model")
  other_cols <- setdiff(names(out), front_cols)
  data.table::setcolorder(out, c(front_cols, other_cols))

  cli::cli_alert_success(
    "Done: {nrow(out)} result row{?s} across {data.table::uniqueN(out$exposure)} exposure{?s}, {data.table::uniqueN(out$model)} model{?s}, {data.table::uniqueN(out$subgroup_level)} subgroup{?s}."
  )

  out[]
}


#' @rdname assoc_subgroup
#' @export
assoc_sub <- assoc_subgroup


#' Dose-response trend analysis
#'
#' Fits categorical and trend models simultaneously for each ordered-factor
#' exposure, returning per-category effect estimates alongside a p-value for
#' linear trend. Two models are run internally per exposure \eqn{\times}
#' adjustment combination:
#'
#' \enumerate{
#'   \item \strong{Categorical model} - exposure treated as a factor; produces
#'     one row per non-reference level (HR / OR / \eqn{\beta} vs reference).
#'   \item \strong{Trend model} - exposure recoded as numeric scores (default:
#'     0, 1, 2, \ldots); produces the per-score-unit effect estimate
#'     (\code{*_per_score}) and \code{p_trend}.
#' }
#'
#' Both results are merged: the output contains a reference row (effect = 1 /
#' 0, CI = NA) followed by non-reference rows, with \code{*_per_score} and
#' \code{p_trend} appended as shared columns (same value within each
#' exposure \eqn{\times} model combination).
#'
#' \strong{Scores}: By default levels are scored 0, 1, 2, \ldots so the
#' reference group = 0 and each step = 1 unit. Supply \code{scores} to use
#' meaningful units (e.g. median years per category) - only \code{p_trend} and
#' the per-score estimate change; per-category HRs are unaffected.
#'
#' \strong{Adjustment models}: follows the same logic as
#' \code{\link{assoc_coxph}} - Unadjusted and Age-and-sex-adjusted models are
#' included by default (\code{base = TRUE}); a Fully adjusted model is added
#' when \code{covariates} is non-NULL.
#'
#' @param data (data.frame or data.table) Analysis dataset.
#' @param outcome_col (character) Outcome column name.
#' @param time_col (character or NULL) Follow-up time column (required when
#'   \code{method = "coxph"}).
#' @param exposure_col (character) One or more exposure column names. Each
#'   must be a \code{factor} (ordered or unordered). The first level is used
#'   as the reference group.
#' @param method (character) Regression method: \code{"coxph"} (default),
#'   \code{"logistic"}, or \code{"linear"}.
#' @param covariates (character or NULL) Covariates for the Fully adjusted
#'   model. Default: \code{NULL}.
#' @param base (logical) Include Unadjusted and Age-and-sex-adjusted models.
#'   Default: \code{TRUE}.
#' @param scores (numeric or NULL) Numeric scores assigned to factor levels in
#'   level order. Length must equal \code{nlevels} of every exposure.
#'   \code{NULL} (default) uses \code{0, 1, 2, \ldots}
#' @param conf_level (numeric) Confidence level. Default: \code{0.95}.
#'
#' @return A \code{data.table} with one row per exposure \eqn{\times} level
#'   \eqn{\times} model, containing:
#'   \describe{
#'     \item{\code{exposure}}{Exposure variable name.}
#'     \item{\code{level}}{Factor level (ordered factor preserving original
#'       level order).}
#'     \item{\code{term}}{Coefficient name as returned by the model (reference
#'       row uses \code{paste0(exposure, ref_level)}).}
#'     \item{\code{model}}{Ordered factor: \code{Unadjusted} <
#'       \code{Age and sex adjusted} < \code{Fully adjusted}.}
#'     \item{\code{n}}{Participants in model (after NA removal).}
#'     \item{...}{Effect estimate columns from the categorical model:
#'       \code{HR}/\code{OR}/\code{beta}, \code{CI_lower}, \code{CI_upper},
#'       \code{p_value}, and a formatted label. Reference row has
#'       \code{HR = 1} / \code{beta = 0} and \code{NA} for CI and p.
#'       Cox models additionally include \code{n_events} and
#'       \code{person_years}; logistic includes \code{n_cases}; linear
#'       includes \code{se}.}
#'     \item{\code{HR_per_score} / \code{OR_per_score} /
#'       \code{beta_per_score}}{Per-score-unit effect estimate from the trend
#'       model. Shared across all levels within the same exposure \eqn{\times}
#'       model.}
#'     \item{\code{HR_per_score_label} / \code{OR_per_score_label} /
#'       \code{beta_per_score_label}}{Formatted string for the per-score
#'       estimate.}
#'     \item{\code{p_trend}}{P-value for linear trend from the trend model.
#'       Shared across all levels within the same exposure \eqn{\times}
#'       model.}
#'   }
#'
#' @importFrom survival coxph Surv
#' @export
#'
#' @examples
#' dt <- ops_toy(scenario = "association", n = 500)
#' dt <- dt[dm_timing != 1L]
#'
#' res <- assoc_trend(
#'   data         = dt,
#'   outcome_col  = "dm_status",
#'   time_col     = "dm_followup_years",
#'   exposure_col = "bmi_cat",
#'   method       = "coxph",
#'   covariates   = c("tdi_cat", "p20116_i0"),
#'   base         = FALSE
#' )
assoc_trend <- function(data,
                         outcome_col,
                         time_col    = NULL,
                         exposure_col,
                         method      = c("coxph", "logistic", "linear"),
                         covariates  = NULL,
                         base        = TRUE,
                         scores      = NULL,
                         conf_level  = 0.95) {

  method <- match.arg(method)

  # Input validation
  .assert_data_frame(data)
  .assert_flag(base)
  .assert_has_cols(data, c(outcome_col, time_col, exposure_col, covariates))

  if (method == "coxph" && is.null(time_col)) {
    cli::cli_abort("{.arg time_col} is required when {.arg method = 'coxph'}.", call = NULL)
  }
  .assert_proportion(conf_level)
  .assert_not_null_if_false(base, covariates)

  # All exposures must be factors
  non_factor <- exposure_col[!vapply(exposure_col,
                                      function(e) is.factor(data[[e]]),
                                      logical(1L))]
  if (length(non_factor) > 0L) {
    cli::cli_abort(
      "exposure_col must be a factor. Non-factor column{?s}: {.field {non_factor}}. Use {.fn factor} or {.fn derive_cut} first.",
      call = NULL
    )
  }

  # Warn if any exposure is unordered
  unordered <- exposure_col[!vapply(exposure_col,
                                     function(e) is.ordered(data[[e]]),
                                     logical(1L))]
  if (length(unordered) > 0L) {
    cli::cli_alert_warning(
      "Exposure{?s} {.field {unordered}} {?is/are} not an ordered factor -- levels will be scored 0, 1, 2, ... (equal spacing assumed)."
    )
  }

  # Validate custom scores
  if (!is.null(scores)) {
    nlv_all <- vapply(exposure_col,
                      function(e) nlevels(data[[e]]), integer(1L))
    bad <- exposure_col[nlv_all != length(scores)]
    if (length(bad) > 0L) {
      cli::cli_abort(
        "{.arg scores} has length {length(scores)} but exposure{?s} {.field {bad}} ha{?s/ve} {nlv_all[exposure_col %in% bad]} level{?s}. Length must equal nlevels.",
        call = NULL
      )
    }
  }

  # Prepare working copy
  dt <- data.table::copy(data.table::as.data.table(data))

  if (method %in% c("coxph", "logistic")) {
    dt[, .ukb_event := .normalise_event(dt[[outcome_col]], outcome_col)]
  }

  # Convert ordered factors to unordered (preserve level order) so that the
  # categorical model uses treatment contrasts and produces term names like
  # "exposureMild" rather than polynomial contrasts ("exposure.L", ".Q").
  # The trend model uses a separate numeric column, so this does not affect
  # p_trend or HR_per_score.
  for (exp in exposure_col) {
    if (is.ordered(dt[[exp]])) {
      lv_tmp <- levels(dt[[exp]])
      dt[, (exp) := factor(get(exp), levels = lv_tmp, ordered = FALSE)]
    }
  }

  # Build model list (same logic as assoc_coxph)
  model_list <- list()

  if (base) {
    model_list["Unadjusted"] <- list(NULL)

    age_col <- .detect_age_col(dt)
    sex_col <- .detect_sex_col(dt)

    if (is.null(age_col) || is.null(sex_col)) {
      cli::cli_alert_warning(
        paste0(
          "Age and sex adjusted model skipped: ",
          if (is.null(age_col)) "age column (UKB field 21022) not found" else "",
          if (is.null(age_col) && is.null(sex_col)) " and " else "",
          if (is.null(sex_col)) "sex column (UKB field 31) not found" else "",
          "."
        )
      )
    } else {
      model_list[["Age and sex adjusted"]] <- c(age_col, sex_col)
    }
  }

  if (!is.null(covariates) && length(covariates) > 0L) {
    model_list[["Fully adjusted"]] <- covariates
  }

  # Run analyses: each exposure x each model
  n_models    <- length(model_list)
  n_exposures <- length(exposure_col)

  cli::cli_h1("assoc_trend")
  cli::cli_alert_info(
    "{n_exposures} exposure{?s} x {n_models} model{?s} (categorical + trend model per combination)"
  )

  all_results <- vector("list", n_exposures * n_models)
  idx <- 1L

  for (exp in exposure_col) {

    cli::cli_h2("{.field {exp}}")

    lv   <- levels(dt[[exp]])
    n_lv <- length(lv)

    # Determine scores for this exposure
    exp_scores <- if (!is.null(scores)) scores else seq(0L, n_lv - 1L)

    cli::cli_alert_info(
      "Levels: {paste(lv, collapse = ' -> ')} | Scores: {paste(exp_scores, collapse = ', ')}"
    )

    # Add numeric trend score column: map factor integer index (1-based) → score
    # Reason: as.integer(factor) gives 1-based index; exp_scores is 1-indexed in R
    dt[, .ukb_trend_score := exp_scores[as.integer(get(exp))]]

    for (model_label in names(model_list)) {

      covs <- model_list[[model_label]]

      # ---- Categorical model (factor exposure → per-category estimates) ----
      res_cat <- switch(method,
        coxph = .run_one_cox_model(
          dt, time_col, exp, covs, NULL, model_label, conf_level
        ),
        logistic = .run_one_logistic_model(
          dt, exp, covs, model_label, "wald", conf_level
        ),
        linear = .run_one_linear_model(
          dt, outcome_col, exp, covs, model_label, conf_level
        )
      )

      # ---- Trend model (numeric score → per-unit estimate + p_trend) ----
      res_trend <- switch(method,
        coxph = .run_one_cox_model(
          dt, time_col, ".ukb_trend_score", covs, NULL,
          paste0(model_label, " [trend]"), conf_level
        ),
        logistic = .run_one_logistic_model(
          dt, ".ukb_trend_score", covs,
          paste0(model_label, " [trend]"), "wald", conf_level
        ),
        linear = .run_one_linear_model(
          dt, outcome_col, ".ukb_trend_score", covs,
          paste0(model_label, " [trend]"), conf_level
        )
      )

      if (is.null(res_cat) || is.null(res_trend)) {
        idx <- idx + 1L
        next
      }

      # ---- Build reference row + extract trend statistics ----
      p_trend <- res_trend$p_value[1L]

      if (method == "coxph") {

        est_per   <- res_trend$HR[1L]
        per_label <- sprintf("%.2f (%.2f-%.2f)",
                             est_per, res_trend$CI_lower[1L], res_trend$CI_upper[1L])

        ref_row <- data.table::data.table(
          exposure     = exp,
          term         = paste0(exp, lv[1L]),
          model        = model_label,
          n            = res_cat$n[1L],
          n_events     = res_cat$n_events[1L],
          person_years = res_cat$person_years[1L],
          HR           = 1.00,
          CI_lower     = NA_real_,
          CI_upper     = NA_real_,
          p_value      = NA_real_,
          HR_label     = "1.00 (ref)"
        )
        res_cat[, `:=`(HR_per_score       = est_per,
                        HR_per_score_label = per_label,
                        p_trend            = p_trend)]
        ref_row[, `:=`(HR_per_score       = est_per,
                        HR_per_score_label = per_label,
                        p_trend            = p_trend)]

        per_str <- sprintf("HR_per_score = %s", per_label)

      } else if (method == "logistic") {

        est_per   <- res_trend$OR[1L]
        per_label <- sprintf("%.2f (%.2f-%.2f)",
                             est_per, res_trend$CI_lower[1L], res_trend$CI_upper[1L])

        ref_row <- data.table::data.table(
          exposure = exp,
          term     = paste0(exp, lv[1L]),
          model    = model_label,
          n        = res_cat$n[1L],
          n_cases  = res_cat$n_cases[1L],
          OR       = 1.00,
          CI_lower = NA_real_,
          CI_upper = NA_real_,
          p_value  = NA_real_,
          OR_label = "1.00 (ref)"
        )
        res_cat[, `:=`(OR_per_score       = est_per,
                        OR_per_score_label = per_label,
                        p_trend            = p_trend)]
        ref_row[, `:=`(OR_per_score       = est_per,
                        OR_per_score_label = per_label,
                        p_trend            = p_trend)]

        per_str <- sprintf("OR_per_score = %s", per_label)

      } else {  # linear

        est_per   <- res_trend$beta[1L]
        per_label <- sprintf("%.2f (%.2f-%.2f)",
                             est_per, res_trend$CI_lower[1L], res_trend$CI_upper[1L])

        ref_row <- data.table::data.table(
          exposure   = exp,
          term       = paste0(exp, lv[1L]),
          model      = model_label,
          n          = res_cat$n[1L],
          beta       = 0.00,
          se         = NA_real_,
          CI_lower   = NA_real_,
          CI_upper   = NA_real_,
          p_value    = NA_real_,
          beta_label = "0.00 (ref)"
        )
        res_cat[, `:=`(beta_per_score       = est_per,
                        beta_per_score_label = per_label,
                        p_trend              = p_trend)]
        ref_row[, `:=`(beta_per_score       = est_per,
                        beta_per_score_label = per_label,
                        p_trend              = p_trend)]

        per_str <- sprintf("beta_per_score = %s", per_label)
      }

      # ---- Combine reference row + non-reference rows ----
      combined <- data.table::rbindlist(list(ref_row, res_cat), fill = TRUE)

      # Add level column: strip the exposure prefix from term to get level name
      # Reason: coxph term names are paste0(exposure, level); reference row
      # follows the same convention, so stripping the prefix recovers the level.
      term_to_lv <- stats::setNames(lv, paste0(exp, lv))
      combined[, level := factor(term_to_lv[term], levels = lv)]

      # Print per-term progress
      cli::cli_alert_info("  {model_label} | {combined$term[1L]}: 1.00 (ref)")
      for (i in seq(2L, nrow(combined))) {
        est_str <- switch(method,
          coxph    = sprintf("HR %s",     combined$HR_label[i]),
          logistic = sprintf("OR %s",     combined$OR_label[i]),
          linear   = sprintf("beta %s", combined$beta_label[i])
        )
        cli::cli_alert_success(
          "  {model_label} | {combined$term[i]}: {est_str}, p = {format.pval(combined$p_value[i], digits = 3L)}"
        )
      }
      cli::cli_alert_info(
        "  {model_label} | trend: {per_str}, p_trend = {format.pval(p_trend, digits = 3L)}"
      )

      all_results[[idx]] <- combined
      idx <- idx + 1L
    }

    # Remove temporary score column before next exposure
    dt[, .ukb_trend_score := NULL]
  }

  # Combine and return
  out <- data.table::rbindlist(Filter(Negate(is.null), all_results), fill = TRUE)

  if (nrow(out) == 0L) {
    cli::cli_alert_warning("No results returned -- check model warnings above.")
    return(out)
  }

  # Ordered factor for model
  model_levels   <- c("Unadjusted", "Age and sex adjusted", "Fully adjusted")
  present_levels <- intersect(model_levels, as.character(unique(out$model)))
  out[, model := factor(model, levels = present_levels, ordered = TRUE)]

  # Column ordering: identifiers first, p_trend last
  front_cols <- c("exposure", "level", "term", "model")
  end_cols   <- intersect("p_trend", names(out))
  mid_cols   <- setdiff(names(out), c(front_cols, end_cols))
  data.table::setcolorder(out, c(front_cols, mid_cols, end_cols))

  cli::cli_alert_success(
    "Done: {nrow(out)} result row{?s} across {data.table::uniqueN(out$exposure)} exposure{?s} and {data.table::uniqueN(out$model)} model{?s}."
  )

  out[]
}


#' @rdname assoc_trend
#' @export
assoc_tr <- assoc_trend


# =============================================================================
# assoc_competing / assoc_fg - Fine-Gray competing risks analysis
# =============================================================================


#' Fine-Gray competing risks association analysis
#'
#' Fits a Fine-Gray subdistribution hazard model (via
#' \code{survival::finegray()} + weighted \code{coxph()}) for each exposure
#' variable and returns a tidy result table with subdistribution hazard ratios
#' (SHR).
#'
#' Two input modes are supported depending on how the outcome is coded in your
#' dataset:
#'
#' \describe{
#'   \item{\strong{Mode A - single multi-value column}}{
#'     \code{compete_col = NULL} (default). \code{outcome_col} contains all
#'     event codes in one column (e.g. \code{0}/\code{1}/\code{2}/\code{3}).
#'     Use \code{event_val} and \code{compete_val} to identify the event of
#'     interest and the competing event; all other values are treated as
#'     censored. Example: UKB \code{censoring_type} where 1 = event, 2 = death
#'     (competing), 0/3 = censored.
#'   }
#'   \item{\strong{Mode B - dual binary columns}}{
#'     \code{compete_col} is the name of a separate 0/1 column for the
#'     competing event. \code{outcome_col} is a 0/1 column for the primary
#'     event. When both are 1 for the same participant, the primary event takes
#'     priority. Example: \code{outcome_col = "copd_status"},
#'     \code{compete_col = "death_status"}.
#'   }
#' }
#'
#' Internally both modes are converted to a three-level factor
#' \code{c("censor", "event", "compete")} before being passed to
#' \code{finegray()}.
#'
#' Three adjustment models are produced (where data allow):
#' \itemize{
#'   \item \strong{Unadjusted} - always included.
#'   \item \strong{Age and sex adjusted} - when \code{base = TRUE} and
#'     age/sex columns are detected.
#'   \item \strong{Fully adjusted} - when \code{covariates} is non-NULL.
#' }
#'
#' @param data (data.frame or data.table) Analysis dataset.
#' @param outcome_col (character) Primary event column. In Mode A: a
#'   multi-value column (any integer or character codes). In Mode B: a 0/1
#'   binary column.
#' @param time_col (character) Follow-up time column (numeric, e.g. years).
#' @param exposure_col (character) One or more exposure variable names.
#' @param compete_col (character or NULL) Mode B only: name of the 0/1
#'   competing event column. When \code{NULL} (default), Mode A is used.
#' @param event_val Scalar value in \code{outcome_col} indicating the primary
#'   event (Mode A only). Default: \code{1L}.
#' @param compete_val Scalar value in \code{outcome_col} indicating the
#'   competing event (Mode A only). Default: \code{2L}.
#' @param covariates (character or NULL) Covariate names for the Fully
#'   adjusted model. When \code{NULL}, only Unadjusted (and Age/sex adjusted
#'   if \code{base = TRUE}) are run.
#' @param base (logical) Whether to auto-detect age and sex columns and include
#'   an Age and sex adjusted model. Default: \code{TRUE}.
#' @param conf_level (numeric) Confidence level for SHR intervals.
#'   Default: \code{0.95}.
#'
#' @return A \code{data.table} with one row per exposure \eqn{\times} term
#'   \eqn{\times} model combination:
#'   \describe{
#'     \item{\code{exposure}}{Exposure variable name.}
#'     \item{\code{term}}{Coefficient name as returned by \code{coxph}.}
#'     \item{\code{model}}{Ordered factor: \code{Unadjusted} <
#'       \code{Age and sex adjusted} < \code{Fully adjusted}.}
#'     \item{\code{n}}{Participants in the model (after NA removal).}
#'     \item{\code{n_events}}{Primary events in the analysis set.}
#'     \item{\code{n_compete}}{Competing events in the analysis set.}
#'     \item{\code{SHR}}{Subdistribution hazard ratio.}
#'     \item{\code{CI_lower}}{Lower CI bound.}
#'     \item{\code{CI_upper}}{Upper CI bound.}
#'     \item{\code{p_value}}{Robust z-test p-value from weighted Cox.}
#'     \item{\code{SHR_label}}{Formatted string, e.g. \code{"1.23 (1.05-1.44)"}.}
#'   }
#'
#' @importFrom survival finegray coxph Surv
#' @export
#'
#' @examples
#' \donttest{
#' dt <- ops_toy(scenario = "association", n = 500)
#' dt <- dt[dm_timing != 1L & htn_timing != 1L]
#'
#' res <- assoc_competing(
#'   data         = dt,
#'   outcome_col  = "dm_status",
#'   time_col     = "dm_followup_years",
#'   exposure_col = "p20116_i0",
#'   compete_col  = "htn_status",
#'   covariates   = c("bmi_cat", "tdi_cat")
#' )
#' }
assoc_competing <- function(data,
                             outcome_col,
                             time_col,
                             exposure_col,
                             compete_col = NULL,
                             event_val   = 1L,
                             compete_val = 2L,
                             covariates  = NULL,
                             base        = TRUE,
                             conf_level  = 0.95) {

  # Input validation
  .assert_data_frame(data)
  .assert_has_cols(data, c(outcome_col, time_col, exposure_col, compete_col, covariates))
  .assert_flag(base)

  .assert_proportion(conf_level)

  # Normalise logical exposures → integer (avoids "t2d_tfTRUE" term names)
  dt <- data.table::copy(data.table::as.data.table(data))
  .normalise_logical_exposures(dt, exposure_col)

  # Build .fg_status: factor(censor / event / compete)
  if (is.null(compete_col)) {
    # Mode A: single multi-value column
    raw <- dt[[outcome_col]]
    fg_vec <- data.table::fcase(
      raw == event_val,   "event",
      raw == compete_val, "compete",
      default = "censor"
    )
    cli::cli_alert_info(
      "Mode A: {.field {outcome_col}} -> event={event_val}, compete={compete_val}, rest=censor"
    )
  } else {
    # Mode B: dual binary columns; primary event takes priority
    ev  <- suppressWarnings(as.integer(dt[[outcome_col]]))
    cmp <- suppressWarnings(as.integer(dt[[compete_col]]))
    fg_vec <- data.table::fcase(
      ev  == 1L, "event",
      cmp == 1L, "compete",
      default = "censor"
    )
    cli::cli_alert_info(
      "Mode B: {.field {outcome_col}} (event) + {.field {compete_col}} (compete)"
    )
  }

  dt[, .fg_status := factor(fg_vec, levels = c("censor", "event", "compete"))]

  n_event_total   <- sum(dt$.fg_status == "event",   na.rm = TRUE)
  n_compete_total <- sum(dt$.fg_status == "compete",  na.rm = TRUE)
  cli::cli_alert_info(
    "Events: {n_event_total}, Competing: {n_compete_total}, Censored: {sum(dt$.fg_status == 'censor', na.rm=TRUE)}"
  )

  # Build model list (Unadjusted / Age and sex adjusted / Fully adjusted)
  models <- list(list(label = "Unadjusted", covs = NULL))

  if (isTRUE(base)) {
    age_col <- .detect_age_col(dt)
    sex_col <- .detect_sex_col(dt)

    if (is.null(age_col) || is.null(sex_col)) {
      cli::cli_alert_warning(
        paste0(
          "Age and sex adjusted model skipped: ",
          if (is.null(age_col)) "age column (UKB field 21022) not found" else "",
          if (is.null(age_col) && is.null(sex_col)) " and " else "",
          if (is.null(sex_col)) "sex column (UKB field 31) not found" else "",
          "."
        )
      )
    } else {
      models <- c(models, list(list(label = "Age and sex adjusted",
                                    covs  = c(age_col, sex_col))))
    }
  }

  if (!is.null(covariates)) {
    models <- c(models, list(list(label = "Fully adjusted", covs = covariates)))
  }

  # Loop over exposures x models
  all_results <- list()

  for (exp in exposure_col) {
    cli::cli_alert_info("Exposure: {.field {exp}}")

    for (m in models) {
      cli::cli_alert_info("  Model: {m$label}")
      res <- .run_one_fg_model(
        dt          = dt,
        time_col    = time_col,
        exposure    = exp,
        covariates  = m$covs,
        model_label = m$label,
        conf_level  = conf_level
      )
      all_results <- c(all_results, list(res))
    }
  }

  # Combine and return
  out <- data.table::rbindlist(Filter(Negate(is.null), all_results), fill = TRUE)

  if (nrow(out) == 0L) {
    cli::cli_alert_warning("No results returned -- check model warnings above.")
    return(out)
  }

  model_levels   <- c("Unadjusted", "Age and sex adjusted", "Fully adjusted")
  present_levels <- intersect(model_levels, as.character(unique(out$model)))
  out[, model := factor(model, levels = present_levels, ordered = TRUE)]

  cli::cli_alert_success(
    "Done: {nrow(out)} result row{?s} across {data.table::uniqueN(out$exposure)} exposure{?s} and {data.table::uniqueN(out$model)} model{?s}."
  )

  out[]
}


#' @rdname assoc_competing
#' @export
assoc_fg <- assoc_competing


# =============================================================================
# assoc_lag - Cox lag (landmark) sensitivity analysis
# =============================================================================


#' Cox regression lag sensitivity analysis
#'
#' Runs Cox proportional hazards models at one or more lag periods to assess
#' whether associations are robust to the exclusion of early events. For each
#' lag, participants whose follow-up time is less than \code{lag_years} are
#' removed from the analysis dataset; follow-up time is kept on its original
#' scale (not shifted). This mirrors the approach used in UK Biobank sensitivity
#' analyses to address reverse causation and detection bias.
#'
#' Setting \code{lag_years = 0} (or including \code{0} in the vector) runs the
#' model on the full unfiltered cohort, providing a reference against which
#' lagged results can be compared.
#'
#' The same three adjustment models produced by \code{\link{assoc_coxph}} are
#' available here (\strong{Unadjusted}, \strong{Age and sex adjusted},
#' \strong{Fully adjusted}).
#'
#' @param data (data.frame or data.table) Analysis dataset.
#' @param outcome_col (character) Event indicator column (0/1 or logical).
#' @param time_col (character) Follow-up time column (numeric, e.g. years).
#' @param exposure_col (character) One or more exposure variable names.
#' @param lag_years (numeric) One or more lag periods in the same units as
#'   \code{time_col}. Default: \code{c(1, 2)}. Use \code{0} to include the
#'   unfiltered full-cohort result as a reference.
#' @param covariates (character or NULL) Covariates for the Fully adjusted
#'   model. When \code{NULL}, only Unadjusted (and Age and sex adjusted if
#'   \code{base = TRUE}) are run.
#' @param base (logical) Auto-detect age and sex and include an Age and sex
#'   adjusted model. Default: \code{TRUE}.
#' @param strata (character or NULL) Optional stratification variable passed to
#'   \code{survival::strata()}.
#' @param conf_level (numeric) Confidence level for HR intervals.
#'   Default: \code{0.95}.
#'
#' @return A \code{data.table} with one row per lag \eqn{\times} exposure
#'   \eqn{\times} term \eqn{\times} model combination, containing all columns
#'   produced by \code{\link{assoc_coxph}} plus:
#'   \describe{
#'     \item{\code{lag_years}}{The lag period applied (numeric).}
#'     \item{\code{n_excluded}}{Number of participants excluded because their
#'       follow-up time was less than \code{lag_years}.}
#'   }
#'   \code{lag_years} and \code{n_excluded} are placed immediately after
#'   \code{model} in the column order.
#'
#' @importFrom survival coxph Surv
#' @export
#'
#' @examples
#' dt <- ops_toy(scenario = "association", n = 500)
#' dt <- dt[dm_timing != 1L]
#'
#' res <- assoc_lag(
#'   data         = dt,
#'   outcome_col  = "dm_status",
#'   time_col     = "dm_followup_years",
#'   exposure_col = "p20116_i0",
#'   lag_years    = c(0, 1, 2),
#'   covariates   = c("bmi_cat", "tdi_cat"),
#'   base         = FALSE
#' )
#' res[, .(lag_years, n, n_excluded, HR, CI_lower, CI_upper, p_value)]
assoc_lag <- function(data,
                       outcome_col,
                       time_col,
                       exposure_col,
                       lag_years  = c(1, 2),
                       covariates = NULL,
                       base       = TRUE,
                       strata     = NULL,
                       conf_level = 0.95) {

  # Input validation
  .assert_data_frame(data)
  .assert_has_cols(data, c(outcome_col, time_col, exposure_col, covariates, strata))
  .assert_flag(base)
  .assert_not_null_if_false(base, covariates)

  if (!is.numeric(lag_years) || any(lag_years < 0)) {
    cli::cli_abort("{.arg lag_years} must be a non-negative numeric vector.", call = NULL)
  }

  .assert_proportion(conf_level)

  lag_years <- sort(unique(lag_years))

  # Prepare full working copy
  dt <- data.table::copy(data.table::as.data.table(data))
  dt[, .ukb_event := .normalise_event(dt[[outcome_col]], outcome_col)]
  .normalise_logical_exposures(dt, exposure_col)

  n_full <- nrow(dt)

  # Build model list (same logic as assoc_coxph)
  model_list <- list()

  if (base) {
    model_list["Unadjusted"] <- list(NULL)

    age_col <- .detect_age_col(dt)
    sex_col <- .detect_sex_col(dt)

    if (is.null(age_col) || is.null(sex_col)) {
      cli::cli_alert_warning(
        paste0(
          "Age and sex adjusted model skipped: ",
          if (is.null(age_col)) "age column not found" else "",
          if (is.null(age_col) && is.null(sex_col)) " and " else "",
          if (is.null(sex_col)) "sex column not found" else "",
          "."
        )
      )
    } else {
      model_list[["Age and sex adjusted"]] <- c(age_col, sex_col)
    }
  }

  if (!is.null(covariates) && length(covariates) > 0L) {
    model_list[["Fully adjusted"]] <- covariates
  }

  # Loop: lag x exposure x model
  cli::cli_h1("assoc_lag")
  cli::cli_alert_info(
    "{length(lag_years)} lag period{?s} x {length(exposure_col)} exposure{?s} x {length(model_list)} model{?s}"
  )

  all_results <- list()

  for (lag in lag_years) {

    cli::cli_h2("Lag: {lag} year{?s}")

    # Filter: exclude participants with follow-up < lag
    sub <- if (lag == 0) {
      dt
    } else {
      dt[dt[[time_col]] >= lag, ]
    }

    n_excluded <- n_full - nrow(sub)
    n_events   <- sum(sub$.ukb_event, na.rm = TRUE)

    cli::cli_alert_info(
      "Excluded (time < {lag} yr): {n_excluded} -- remaining: {nrow(sub)}, events: {n_events}"
    )

    for (exp in exposure_col) {
      cli::cli_h2("{.field {exp}}")

      for (model_label in names(model_list)) {

        res <- .run_one_cox_model(
          dt          = sub,
          time_col    = time_col,
          exposure    = exp,
          covariates  = model_list[[model_label]],
          strata      = strata,
          model_label = model_label,
          conf_level  = conf_level
        )

        if (!is.null(res)) {
          res[, `:=`(lag_years = lag, n_excluded = n_excluded)]
        }

        all_results <- c(all_results, list(res))
      }
    }
  }

  # Combine and return
  out <- data.table::rbindlist(Filter(Negate(is.null), all_results), fill = TRUE)

  if (nrow(out) == 0L) {
    cli::cli_alert_warning("No results returned -- check model warnings above.")
    return(out)
  }

  # Ordered factor for model
  model_levels   <- c("Unadjusted", "Age and sex adjusted", "Fully adjusted")
  present_levels <- intersect(model_levels, as.character(unique(out$model)))
  out[, model := factor(model, levels = present_levels, ordered = TRUE)]

  # Column order: identifiers first, lag_years + n_excluded after model
  front_cols <- c("exposure", "term", "model", "lag_years", "n_excluded")
  mid_cols   <- setdiff(names(out), front_cols)
  data.table::setcolorder(out, c(front_cols, mid_cols))

  cli::cli_alert_success(
    "Done: {nrow(out)} result row{?s} across {data.table::uniqueN(out$lag_years)} lag period{?s}, {data.table::uniqueN(out$exposure)} exposure{?s}, and {data.table::uniqueN(out$model)} model{?s}."
  )

  out[]
}

# =============================================================================
# test-assoc.R — Unit tests for assoc_ series (simulated data, no real files)
# =============================================================================


# ===========================================================================
# Shared helpers
# ===========================================================================

.fake_assoc_dt <- function(n = 300, seed = 42) {
  set.seed(seed)
  dt <- data.table::data.table(
    eid                = seq_len(n),
    followup_years     = round(runif(n, 1, 15), 2),
    copd_01            = rbinom(n, 1, 0.10),
    t2d_01              = rbinom(n, 1, 0.20),
    t2d_tf              = as.logical(rbinom(n, 1, 0.20)),
    bmi_cat            = factor(
      sample(c("Normal", "Overweight", "Obese"), n, TRUE, c(0.4, 0.35, 0.25)),
      levels = c("Normal", "Overweight", "Obese")
    ),
    bmi_num            = round(rnorm(n, 27, 5), 1),
    age_at_recruitment = round(rnorm(n, 57, 8), 1),
    sex                = factor(sample(c("Male", "Female"), n, TRUE)),
    tdi                = rnorm(n, -1, 3),
    smoking            = factor(sample(c("Never", "Previous", "Current"), n, TRUE))
  )
  # Inject some NA into tdi for listwise deletion tests
  dt[sample(n, 20L), tdi := NA_real_]
  dt
}

.fake_competing_dt <- function(n = 300, seed = 42) {
  dt <- .fake_assoc_dt(n = n, seed = seed)
  set.seed(seed + 1L)
  dt[, death_01 := rbinom(n, 1, 0.05)]
  # Prevent simultaneous primary + competing event
  dt[copd_01 == 1L & death_01 == 1L, death_01 := 0L]
  dt[, censoring_type := data.table::fcase(
    copd_01  == 1L, 1L,
    death_01 == 1L, 2L,
    default  = 0L
  )]
  dt
}


# ===========================================================================
# assoc_coxph()
# ===========================================================================

test_that("assoc_coxph() aborts on non-data.frame input", {
  expect_error(
    assoc_coxph("not a df", "copd_01", "followup_years", "t2d_01"),
    "data.frame"
  )
})

test_that("assoc_coxph() aborts on missing column", {
  dt <- .fake_assoc_dt()
  expect_error(
    suppressMessages(assoc_coxph(dt, "nonexistent", "followup_years", "t2d_01")),
    "missing"
  )
})

test_that("assoc_coxph() aborts when base=FALSE and no covariates", {
  dt <- .fake_assoc_dt()
  expect_error(
    suppressMessages(assoc_coxph(dt, "copd_01", "followup_years", "t2d_01",
                                 base = FALSE)),
    "covariates"
  )
})

test_that("assoc_coxph() returns data.table", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_coxph(dt, "copd_01", "followup_years", "t2d_01")
  )
  expect_true(data.table::is.data.table(res))
})

test_that("assoc_coxph() output has required columns", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_coxph(dt, "copd_01", "followup_years", "t2d_01")
  )
  expect_true(all(c("exposure", "term", "model", "n", "n_events",
                    "person_years", "HR", "CI_lower", "CI_upper",
                    "p_value", "HR_label") %in% names(res)))
})

test_that("assoc_coxph() produces Unadjusted and Age-sex models by default", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_coxph(dt, "copd_01", "followup_years", "t2d_01")
  )
  expect_true("Unadjusted" %in% levels(res$model))
  expect_true("Age and sex adjusted" %in% levels(res$model))
})

test_that("assoc_coxph() model is ordered factor", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_coxph(dt, "copd_01", "followup_years", "t2d_01")
  )
  expect_true(is.ordered(res$model))
})

test_that("assoc_coxph() HR > 0 and CI_lower < HR < CI_upper", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_coxph(dt, "copd_01", "followup_years", "t2d_01")
  )
  expect_true(all(res$HR > 0))
  expect_true(all(res$CI_lower < res$HR))
  expect_true(all(res$HR < res$CI_upper))
})

test_that("assoc_coxph() p_value in [0, 1]", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_coxph(dt, "copd_01", "followup_years", "t2d_01")
  )
  expect_true(all(res$p_value >= 0 & res$p_value <= 1))
})

test_that("assoc_coxph() binary term name matches exposure name (no suffix)", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_coxph(dt, "copd_01", "followup_years", "t2d_01")
  )
  expect_true(all(res$term == "t2d_01"))
})

test_that("assoc_coxph() logical exposure converts to integer (no TRUE suffix)", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_coxph(dt, "copd_01", "followup_years", "t2d_tf")
  )
  expect_true(all(res$term == "t2d_tf"))
})

test_that("assoc_coxph() factor exposure produces n_levels - 1 term rows per model", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_coxph(dt, "copd_01", "followup_years", "bmi_cat")
  )
  n_per_model <- nrow(res[res$model == "Unadjusted", ])
  expect_equal(n_per_model, 2L)  # 3 levels → 2 non-reference terms
})

test_that("assoc_coxph() Fully adjusted model included when covariates given", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_coxph(dt, "copd_01", "followup_years", "t2d_01",
                covariates = c("tdi", "smoking"))
  )
  expect_true("Fully adjusted" %in% levels(res$model))
})

test_that("assoc_coxph() Fully adjusted n <= Unadjusted n (listwise deletion)", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_coxph(dt, "copd_01", "followup_years", "t2d_01",
                covariates = c("tdi", "smoking"))
  )
  n_full <- res[res$model == "Fully adjusted", "n"][[1L]]
  n_unadj <- res[res$model == "Unadjusted", "n"][[1L]]
  expect_lte(n_full, n_unadj)
})

test_that("assoc_coxph() multiple exposures stack rows correctly", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_coxph(dt, "copd_01", "followup_years", c("t2d_01", "bmi_num"))
  )
  expect_equal(length(unique(res$exposure)), 2L)
})

test_that("assoc_cox() alias matches assoc_coxph()", {
  dt   <- .fake_assoc_dt()
  res1 <- suppressMessages(assoc_coxph(dt, "copd_01", "followup_years", "t2d_01"))
  res2 <- suppressMessages(assoc_cox(dt,   "copd_01", "followup_years", "t2d_01"))
  expect_equal(res1$HR, res2$HR)
})


# ===========================================================================
# assoc_logistic()
# ===========================================================================

test_that("assoc_logistic() aborts on non-data.frame input", {
  expect_error(
    assoc_logistic("not a df", "copd_01", "t2d_01"),
    "data.frame"
  )
})

test_that("assoc_logistic() aborts on missing column", {
  dt <- .fake_assoc_dt()
  expect_error(
    suppressMessages(assoc_logistic(dt, "nonexistent", "t2d_01")),
    "missing"
  )
})

test_that("assoc_logistic() returns data.table with OR columns", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(assoc_logistic(dt, "copd_01", "t2d_01"))
  expect_true(data.table::is.data.table(res))
  expect_true(all(c("OR", "CI_lower", "CI_upper", "p_value", "OR_label") %in%
                    names(res)))
})

test_that("assoc_logistic() OR > 0 and CI_lower < OR < CI_upper", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(assoc_logistic(dt, "copd_01", "t2d_01"))
  expect_true(all(res$OR > 0))
  expect_true(all(res$CI_lower < res$OR & res$OR < res$CI_upper))
})

test_that("assoc_logistic() aborts on non-binary outcome", {
  dt <- .fake_assoc_dt()
  dt[, bmi_int := as.integer(bmi_num)]
  expect_error(
    suppressMessages(assoc_logistic(dt, "bmi_int", "t2d_01")),
    "0/1"
  )
})

test_that("assoc_logit() alias matches assoc_logistic()", {
  dt   <- .fake_assoc_dt()
  res1 <- suppressMessages(assoc_logistic(dt, "copd_01", "t2d_01"))
  res2 <- suppressMessages(assoc_logit(dt,    "copd_01", "t2d_01"))
  expect_equal(res1$OR, res2$OR)
})


# ===========================================================================
# assoc_linear()
# ===========================================================================

test_that("assoc_linear() aborts on non-data.frame input", {
  expect_error(
    assoc_linear("not a df", "bmi_num", "t2d_01"),
    "data.frame"
  )
})

test_that("assoc_linear() returns data.table with beta/se columns", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(assoc_linear(dt, "bmi_num", "t2d_01"))
  expect_true(data.table::is.data.table(res))
  expect_true(all(c("beta", "se", "CI_lower", "CI_upper",
                    "p_value", "beta_label") %in% names(res)))
})

test_that("assoc_linear() se > 0", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(assoc_linear(dt, "bmi_num", "t2d_01"))
  expect_true(all(res$se > 0))
})

test_that("assoc_linear() p_value in [0, 1]", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(assoc_linear(dt, "bmi_num", "t2d_01"))
  expect_true(all(res$p_value >= 0 & res$p_value <= 1))
})

test_that("assoc_lm() alias matches assoc_linear()", {
  dt   <- .fake_assoc_dt()
  res1 <- suppressMessages(assoc_linear(dt, "bmi_num", "t2d_01"))
  res2 <- suppressMessages(assoc_lm(dt,     "bmi_num", "t2d_01"))
  expect_equal(res1$beta, res2$beta)
})


# ===========================================================================
# assoc_coxph_zph()
# ===========================================================================

test_that("assoc_coxph_zph() aborts on missing column", {
  dt <- .fake_assoc_dt()
  expect_error(
    suppressMessages(
      assoc_coxph_zph(dt, "nonexistent", "followup_years", "t2d_01")
    ),
    "missing"
  )
})

test_that("assoc_coxph_zph() returns data.table with zph columns", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_coxph_zph(dt, "copd_01", "followup_years", "t2d_01")
  )
  expect_true(data.table::is.data.table(res))
  expect_true(all(c("term", "chisq", "df", "p_value",
                    "ph_satisfied", "global_p") %in% names(res)))
})

test_that("assoc_coxph_zph() p_value in [0, 1]", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_coxph_zph(dt, "copd_01", "followup_years", "t2d_01")
  )
  expect_true(all(res$p_value >= 0 & res$p_value <= 1, na.rm = TRUE))
})

test_that("assoc_zph() alias matches assoc_coxph_zph()", {
  dt   <- .fake_assoc_dt()
  res1 <- suppressMessages(assoc_coxph_zph(dt, "copd_01", "followup_years", "t2d_01"))
  res2 <- suppressMessages(assoc_zph(dt,        "copd_01", "followup_years", "t2d_01"))
  expect_equal(res1$p_value, res2$p_value)
})


# ===========================================================================
# assoc_subgroup()
# ===========================================================================

test_that("assoc_subgroup() aborts on missing by column", {
  dt <- .fake_assoc_dt()
  expect_error(
    suppressMessages(
      assoc_subgroup(dt, "copd_01", "followup_years", "t2d_01", by = "nonexistent")
    ),
    "missing"
  )
})

test_that("assoc_subgroup() aborts when by is not length 1", {
  dt <- .fake_assoc_dt()
  expect_error(
    suppressMessages(
      assoc_subgroup(dt, "copd_01", "followup_years", "t2d_01",
                     by = c("sex", "smoking"))
    ),
    "single"
  )
})

test_that("assoc_subgroup() returns data.table with subgroup_level column", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_subgroup(dt, "copd_01", "followup_years", "t2d_01", by = "sex")
  )
  expect_true(data.table::is.data.table(res))
  expect_true("subgroup_level" %in% names(res))
})

test_that("assoc_subgroup() produces one result row per subgroup level", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_subgroup(dt, "copd_01", "followup_years", "t2d_01", by = "sex")
  )
  expect_equal(length(unique(res$subgroup_level)), 2L)  # Male / Female
})

test_that("assoc_subgroup() contains p_interaction column", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_subgroup(dt, "copd_01", "followup_years", "t2d_01", by = "sex")
  )
  expect_true("p_interaction" %in% names(res))
})

test_that("assoc_subgroup() p_interaction is shared across subgroup levels", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_subgroup(dt, "copd_01", "followup_years", "t2d_01", by = "sex",
                   covariates = "tdi")
  )
  # Same model → same p_interaction value for all levels
  pi_vals <- unique(res[res$model == "Fully adjusted", "p_interaction"])
  expect_equal(nrow(pi_vals), 1L)
})

test_that("assoc_sub() alias matches assoc_subgroup()", {
  dt   <- .fake_assoc_dt()
  res1 <- suppressMessages(
    assoc_subgroup(dt, "copd_01", "followup_years", "t2d_01", by = "sex"))
  res2 <- suppressMessages(
    assoc_sub(dt, "copd_01", "followup_years", "t2d_01", by = "sex"))
  expect_equal(res1$HR, res2$HR)
})


# ===========================================================================
# assoc_trend()
# ===========================================================================

test_that("assoc_trend() aborts when exposure is not factor", {
  dt <- .fake_assoc_dt()
  expect_error(
    suppressMessages(
      assoc_trend(dt, "copd_01", "followup_years", "t2d_01", method = "coxph")
    ),
    "factor"
  )
})

test_that("assoc_trend() aborts when scores length != nlevels", {
  dt <- .fake_assoc_dt()
  expect_error(
    suppressMessages(
      assoc_trend(dt, "copd_01", "followup_years", "bmi_cat",
                  method = "coxph", scores = c(0, 1))  # 3 levels, 2 scores
    ),
    "length|scores"
  )
})

test_that("assoc_trend() returns data.table with level and p_trend columns", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_trend(dt, "copd_01", "followup_years", "bmi_cat", method = "coxph")
  )
  expect_true(data.table::is.data.table(res))
  expect_true("level"   %in% names(res))
  expect_true("p_trend" %in% names(res))
})

test_that("assoc_trend() reference row has HR=1 and NA CI", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_trend(dt, "copd_01", "followup_years", "bmi_cat", method = "coxph")
  )
  ref <- res[res$level == "Normal" & res$model == "Unadjusted", ]
  expect_equal(ref$HR[1L], 1.0)
  expect_true(is.na(ref$CI_lower[1L]))
})

test_that("assoc_trend() p_trend is last column", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_trend(dt, "copd_01", "followup_years", "bmi_cat", method = "coxph")
  )
  expect_equal(names(res)[ncol(res)], "p_trend")
})

test_that("assoc_trend() level is factor", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_trend(dt, "copd_01", "followup_years", "bmi_cat", method = "coxph")
  )
  expect_true(is.factor(res$level))
})

test_that("assoc_trend() HR_per_score differs with custom scores", {
  dt    <- .fake_assoc_dt()
  res_a <- suppressMessages(
    assoc_trend(dt, "copd_01", "followup_years", "bmi_cat",
                method = "coxph", scores = c(0, 1, 2))
  )
  res_b <- suppressMessages(
    assoc_trend(dt, "copd_01", "followup_years", "bmi_cat",
                method = "coxph", scores = c(0, 5, 10))
  )
  # Larger score interval → smaller per-unit HR
  expect_false(isTRUE(all.equal(
    res_a[res_a$model == "Unadjusted", "HR_per_score"][[1L]][1L],
    res_b[res_b$model == "Unadjusted", "HR_per_score"][[1L]][1L]
  )))
})

test_that("assoc_tr() alias matches assoc_trend()", {
  dt   <- .fake_assoc_dt()
  res1 <- suppressMessages(
    assoc_trend(dt, "copd_01", "followup_years", "bmi_cat", method = "coxph"))
  res2 <- suppressMessages(
    assoc_tr(dt,   "copd_01", "followup_years", "bmi_cat", method = "coxph"))
  expect_equal(res1$HR, res2$HR)
})


# ===========================================================================
# assoc_competing()
# ===========================================================================

test_that("assoc_competing() aborts on missing column", {
  dt <- .fake_competing_dt()
  expect_error(
    suppressMessages(
      assoc_competing(dt, "nonexistent", "followup_years", "t2d_01")
    ),
    "missing"
  )
})

test_that("assoc_competing() returns data.table with SHR columns", {
  dt  <- .fake_competing_dt()
  res <- suppressMessages(
    assoc_competing(dt, "censoring_type", "followup_years", "t2d_01",
                    event_val = 1L, compete_val = 2L)
  )
  expect_true(data.table::is.data.table(res))
  expect_true(all(c("SHR", "CI_lower", "CI_upper",
                    "p_value", "SHR_label", "n_compete") %in% names(res)))
})

test_that("assoc_competing() has no HR column (SHR not HR)", {
  dt  <- .fake_competing_dt()
  res <- suppressMessages(
    assoc_competing(dt, "censoring_type", "followup_years", "t2d_01",
                    event_val = 1L, compete_val = 2L)
  )
  expect_false("HR" %in% names(res))
})

test_that("assoc_competing() SHR > 0 and CI_lower < SHR < CI_upper", {
  dt  <- .fake_competing_dt()
  res <- suppressMessages(
    assoc_competing(dt, "censoring_type", "followup_years", "t2d_01",
                    event_val = 1L, compete_val = 2L)
  )
  expect_true(all(res$SHR > 0))
  expect_true(all(res$CI_lower < res$SHR & res$SHR < res$CI_upper))
})

test_that("assoc_competing() Mode B produces same n_events as Mode A", {
  dt  <- .fake_competing_dt()
  res_a <- suppressMessages(
    assoc_competing(dt, "censoring_type", "followup_years", "t2d_01",
                    event_val = 1L, compete_val = 2L)
  )
  res_b <- suppressMessages(
    assoc_competing(dt, "copd_01", "followup_years", "t2d_01",
                    compete_col = "death_01")
  )
  expect_equal(
    res_a[res_a$model == "Unadjusted", "n_events"][[1L]],
    res_b[res_b$model == "Unadjusted", "n_events"][[1L]]
  )
})

test_that("assoc_competing() logical exposure normalised (no TRUE suffix in term)", {
  dt  <- .fake_competing_dt()
  res <- suppressMessages(
    assoc_competing(dt, "censoring_type", "followup_years", "t2d_tf",
                    event_val = 1L, compete_val = 2L)
  )
  expect_true(all(res$term == "t2d_tf"))
})

test_that("assoc_competing() Fully adjusted n <= Unadjusted n", {
  dt  <- .fake_competing_dt()
  res <- suppressMessages(
    assoc_competing(dt, "censoring_type", "followup_years", "t2d_01",
                    event_val  = 1L, compete_val = 2L,
                    covariates = c("tdi", "smoking"))
  )
  n_full  <- res[res$model == "Fully adjusted", "n"][[1L]]
  n_unadj <- res[res$model == "Unadjusted",     "n"][[1L]]
  expect_lte(n_full, n_unadj)
})

test_that("assoc_competing() skips Age and sex adjusted when age/sex columns absent", {
  dt <- .fake_competing_dt()
  # Remove both age and sex columns so auto-detection fails
  dt[, c("age_at_recruitment", "sex") := NULL]
  res <- suppressMessages(
    assoc_competing(dt, "censoring_type", "followup_years", "t2d_01",
                    event_val = 1L, compete_val = 2L)
  )
  expect_false("Age and sex adjusted" %in% as.character(unique(res$model)))
})

test_that("assoc_fg() alias matches assoc_competing()", {
  dt   <- .fake_competing_dt()
  res1 <- suppressMessages(
    assoc_competing(dt, "censoring_type", "followup_years", "t2d_01",
                    event_val = 1L, compete_val = 2L))
  res2 <- suppressMessages(
    assoc_fg(dt, "censoring_type", "followup_years", "t2d_01",
             event_val = 1L, compete_val = 2L))
  expect_equal(res1$SHR, res2$SHR)
})


# ===========================================================================
# assoc_lag()
# ===========================================================================

test_that("assoc_lag() aborts on missing column", {
  dt <- .fake_assoc_dt()
  expect_error(
    suppressMessages(
      assoc_lag(dt, "nonexistent", "followup_years", "t2d_01")
    ),
    "missing"
  )
})

test_that("assoc_lag() aborts when base=FALSE and no covariates", {
  dt <- .fake_assoc_dt()
  expect_error(
    suppressMessages(
      assoc_lag(dt, "copd_01", "followup_years", "t2d_01", base = FALSE)
    ),
    "covariates"
  )
})

test_that("assoc_lag() aborts on negative lag_years", {
  dt <- .fake_assoc_dt()
  expect_error(
    suppressMessages(
      assoc_lag(dt, "copd_01", "followup_years", "t2d_01", lag_years = c(-1, 2))
    ),
    "non-negative"
  )
})

test_that("assoc_lag() returns data.table with lag_years and n_excluded", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_lag(dt, "copd_01", "followup_years", "t2d_01", lag_years = c(1, 2))
  )
  expect_true(data.table::is.data.table(res))
  expect_true("lag_years"  %in% names(res))
  expect_true("n_excluded" %in% names(res))
})

test_that("assoc_lag() produces one result set per lag value", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_lag(dt, "copd_01", "followup_years", "t2d_01", lag_years = c(1, 2))
  )
  expect_equal(sort(unique(res$lag_years)), c(1, 2))
})

test_that("assoc_lag() lag=0 excludes nobody (n_excluded = 0)", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_lag(dt, "copd_01", "followup_years", "t2d_01", lag_years = c(0, 2))
  )
  expect_equal(res[res$lag_years == 0, "n_excluded"][[1L]][1L], 0L)
})

test_that("assoc_lag() larger lag excludes more participants", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_lag(dt, "copd_01", "followup_years", "t2d_01", lag_years = c(2, 5))
  )
  n_ex_2 <- res[res$lag_years == 2, "n_excluded"][[1L]][1L]
  n_ex_5 <- res[res$lag_years == 5, "n_excluded"][[1L]][1L]
  expect_gte(n_ex_5, n_ex_2)
})

test_that("assoc_lag() larger lag gives smaller or equal n", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_lag(dt, "copd_01", "followup_years", "t2d_01", lag_years = c(2, 5))
  )
  n_2 <- res[res$lag_years == 2 & res$model == "Unadjusted", "n"][[1L]]
  n_5 <- res[res$lag_years == 5 & res$model == "Unadjusted", "n"][[1L]]
  expect_lte(n_5, n_2)
})

test_that("assoc_lag() HR > 0 across all lags", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_lag(dt, "copd_01", "followup_years", "t2d_01", lag_years = c(1, 2))
  )
  expect_true(all(res$HR > 0))
})

test_that("assoc_lag() multiple exposures produce correct row count", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_lag(dt, "copd_01", "followup_years",
              exposure_col = c("t2d_01", "bmi_num"),
              lag_years    = c(1, 2))
  )
  # 2 lag × 2 exposure × 2 model (Unadjusted + Age-sex) = 8 rows
  expect_equal(nrow(res), 8L)
})

test_that("assoc_lag() lag_years and n_excluded placed after model column", {
  dt  <- .fake_assoc_dt()
  res <- suppressMessages(
    assoc_lag(dt, "copd_01", "followup_years", "t2d_01", lag_years = 1)
  )
  col_pos <- match(c("model", "lag_years", "n_excluded"), names(res))
  expect_true(col_pos[2L] == col_pos[1L] + 1L)
  expect_true(col_pos[3L] == col_pos[1L] + 2L)
})

# =============================================================================
# test-integration-assoc.R — Integration tests for assoc_ series
# Uses a larger simulated dataset (n=2000) to test real model behaviour.
# Run manually before release: devtools::test(filter = "integration-assoc")
# =============================================================================

# ===========================================================================
# Shared fixture (n=2000, realistic UKB-like distributions)
# ===========================================================================

skip_on_cran()
RNGkind("Mersenne-Twister", "Inversion", "Rejection")
set.seed(2024L)
N <- 2000L

DT <- data.table::data.table(
  eid                = seq_len(N),
  followup_years     = round(runif(N, 0.5, 15), 2),
  copd_01            = rbinom(N, 1, 0.08),
  t2d_01              = rbinom(N, 1, 0.15),
  t2d_severity        = factor(
    sample(c("No AD", "Mild", "Moderate/Severe"), N, TRUE, c(0.70, 0.20, 0.10)),
    levels = c("No AD", "Mild", "Moderate/Severe")
  ),
  bmi_cat            = factor(
    sample(c("Normal", "Overweight", "Obese"), N, TRUE, c(0.40, 0.35, 0.25)),
    levels = c("Normal", "Overweight", "Obese")
  ),
  bmi_num            = round(rnorm(N, 27, 5), 1),
  age_at_recruitment = round(rnorm(N, 57, 8), 1),
  sex                = factor(sample(c("Male", "Female"), N, TRUE)),
  tdi                = rnorm(N, -1, 3),
  smoking            = factor(sample(c("Never", "Previous", "Current"), N, TRUE)),
  ethnicity          = factor(sample(c("White", "Asian", "Black", "Other"),
                                     N, TRUE, c(0.85, 0.07, 0.04, 0.04)))
)
# Inject NA into tdi to test listwise deletion
DT[sample(N, 150L), tdi := NA_real_]

# Competing risks columns
set.seed(2025L)
DT[, death_01      := rbinom(N, 1, 0.05)]
DT[copd_01 == 1L & death_01 == 1L, death_01 := 0L]
DT[, censoring_type := data.table::fcase(
  copd_01  == 1L, 1L,
  death_01 == 1L, 2L,
  default  = 0L
)]

COVS_FULL <- c("tdi", "smoking", "ethnicity")


# ===========================================================================
# assoc_coxph() — integration
# ===========================================================================

test_that("assoc_coxph() runs on n=2000 without error", {
  expect_no_error(suppressMessages(
    assoc_coxph(DT, "copd_01", "followup_years", "t2d_01",
                covariates = COVS_FULL)
  ))
})

test_that("assoc_coxph() 3-model output is numerically coherent", {
  res <- suppressMessages(
    assoc_coxph(DT, "copd_01", "followup_years", "t2d_01",
                covariates = COVS_FULL)
  )
  expect_equal(uniqueN(res$model), 3L)
  expect_true(all(res$HR > 0))
  expect_true(all(res$CI_lower < res$HR))
  expect_true(all(res$HR < res$CI_upper))
  expect_true(all(res$p_value >= 0 & res$p_value <= 1))
})

test_that("assoc_coxph() factor exposure returns n_levels - 1 terms per model", {
  res <- suppressMessages(
    assoc_coxph(DT, "copd_01", "followup_years", "bmi_cat")
  )
  # Drop empty factor levels before counting (ordered factor may retain all levels)
  per_model <- table(droplevels(res$model))
  expect_true(all(per_model == 2L))   # 3 levels → 2 non-reference terms
})

test_that("assoc_coxph() HR_label format matches 'x.xx (x.xx-x.xx)'", {
  res <- suppressMessages(
    assoc_coxph(DT, "copd_01", "followup_years", "t2d_01")
  )
  expect_true(all(grepl("^\\d+\\.\\d+ \\(\\d+\\.\\d+", res$HR_label)))
})

test_that("assoc_coxph() person_years is positive and finite", {
  res <- suppressMessages(
    assoc_coxph(DT, "copd_01", "followup_years", "t2d_01")
  )
  expect_true(all(res$person_years > 0))
  expect_true(all(is.finite(res$person_years)))
})

test_that("assoc_coxph() n_events consistent with data", {
  res <- suppressMessages(
    assoc_coxph(DT, "copd_01", "followup_years", "t2d_01")
  )
  # Unadjusted model uses all rows (no covariate NA)
  expected_events <- sum(DT$copd_01)
  expect_equal(res[res$model == "Unadjusted", "n_events"][[1L]], expected_events)
})

test_that("assoc_coxph() multiple exposures correct row count", {
  res <- suppressMessages(
    assoc_coxph(DT, "copd_01", "followup_years", c("t2d_01", "bmi_num", "bmi_cat"))
  )
  # t2d_01: 1 term, bmi_num: 1 term, bmi_cat: 2 terms → 4 per model × 2 models = 8
  expect_equal(nrow(res), 8L)
})


# ===========================================================================
# assoc_logistic() — integration
# ===========================================================================

test_that("assoc_logistic() runs on n=2000 without error", {
  expect_no_error(suppressMessages(
    assoc_logistic(DT, "copd_01", "t2d_01", covariates = COVS_FULL)
  ))
})

test_that("assoc_logistic() OR > 0, CI spans OR, p in [0,1]", {
  res <- suppressMessages(
    assoc_logistic(DT, "copd_01", "t2d_01", covariates = COVS_FULL)
  )
  expect_true(all(res$OR > 0))
  expect_true(all(res$CI_lower < res$OR & res$OR < res$CI_upper))
  expect_true(all(res$p_value >= 0 & res$p_value <= 1))
})


# ===========================================================================
# assoc_linear() — integration
# ===========================================================================

test_that("assoc_linear() runs on n=2000 without error", {
  expect_no_error(suppressMessages(
    assoc_linear(DT, "bmi_num", "t2d_01", covariates = COVS_FULL)
  ))
})

test_that("assoc_linear() se > 0 and p in [0,1]", {
  res <- suppressMessages(
    assoc_linear(DT, "bmi_num", "t2d_01", covariates = COVS_FULL)
  )
  expect_true(all(res$se > 0))
  expect_true(all(res$p_value >= 0 & res$p_value <= 1))
})


# ===========================================================================
# assoc_coxph_zph() — integration
# ===========================================================================

test_that("assoc_coxph_zph() runs without error and returns ph_satisfied", {
  res <- suppressMessages(
    assoc_coxph_zph(DT, "copd_01", "followup_years", "t2d_01")
  )
  expect_true("ph_satisfied" %in% names(res))
  expect_true(is.logical(res$ph_satisfied))
})


# ===========================================================================
# assoc_subgroup() — integration
# ===========================================================================

test_that("assoc_subgroup() runs on n=2000 by sex without error", {
  expect_no_error(suppressMessages(
    assoc_subgroup(DT, "copd_01", "followup_years", "t2d_01",
                   by = "sex", covariates = COVS_FULL)
  ))
})

test_that("assoc_subgroup() p_interaction is numeric scalar in [0,1]", {
  res <- suppressMessages(
    assoc_subgroup(DT, "copd_01", "followup_years", "t2d_01",
                   by = "sex", covariates = COVS_FULL)
  )
  pi <- unique(stats::na.omit(res$p_interaction))
  expect_true(length(pi) >= 1L)
  expect_true(all(pi >= 0 & pi <= 1))
})

test_that("assoc_subgroup() subgroup n sums to <= total n", {
  res <- suppressMessages(
    assoc_subgroup(DT, "copd_01", "followup_years", "t2d_01", by = "sex")
  )
  total_n <- sum(res[res$model == "Unadjusted", "n"][[1L]])
  expect_lte(total_n, nrow(DT))
})

test_that("assoc_subgroup() works with 3-level by variable", {
  res <- suppressMessages(
    assoc_subgroup(DT, "copd_01", "followup_years", "t2d_01", by = "smoking")
  )
  expect_equal(uniqueN(res$subgroup_level), 3L)
})


# ===========================================================================
# assoc_trend() — integration
# ===========================================================================

test_that("assoc_trend() coxph on n=2000 returns reference row + non-ref rows", {
  res <- suppressMessages(
    assoc_trend(DT, "copd_01", "followup_years", "t2d_severity",
                method = "coxph")
  )
  ref_rows <- res[res$level == "No AD" & res$model == "Unadjusted", ]
  expect_equal(nrow(ref_rows), 1L)
  expect_equal(ref_rows$HR[1L], 1.0)
  expect_true(is.na(ref_rows$CI_lower[1L]))
})

test_that("assoc_trend() p_trend is numeric and in [0,1]", {
  res <- suppressMessages(
    assoc_trend(DT, "copd_01", "followup_years", "t2d_severity",
                method = "coxph")
  )
  pt <- stats::na.omit(unique(res$p_trend))
  expect_true(all(pt >= 0 & pt <= 1))
})

test_that("assoc_trend() logistic method returns OR_per_score", {
  res <- suppressMessages(
    assoc_trend(DT, "copd_01", "followup_years", "t2d_severity",
                method = "logistic")
  )
  expect_true("OR_per_score" %in% names(res))
})

test_that("assoc_trend() linear method returns beta_per_score", {
  res <- suppressMessages(
    assoc_trend(DT, "bmi_num", "followup_years", "t2d_severity",
                method = "linear")
  )
  expect_true("beta_per_score" %in% names(res))
})


# ===========================================================================
# assoc_competing() — integration
# ===========================================================================

test_that("assoc_competing() Mode A runs on n=2000 without error", {
  expect_no_error(suppressMessages(
    assoc_competing(DT, "censoring_type", "followup_years", "t2d_01",
                    event_val = 1L, compete_val = 2L,
                    covariates = COVS_FULL)
  ))
})

test_that("assoc_competing() Mode B runs without error", {
  expect_no_error(suppressMessages(
    assoc_competing(DT, "copd_01", "followup_years", "t2d_01",
                    compete_col = "death_01",
                    covariates  = COVS_FULL)
  ))
})

test_that("assoc_competing() SHR > 0, CI spans SHR, p in [0,1]", {
  res <- suppressMessages(
    assoc_competing(DT, "censoring_type", "followup_years", "t2d_01",
                    event_val = 1L, compete_val = 2L,
                    covariates = COVS_FULL)
  )
  expect_true(all(res$SHR > 0))
  expect_true(all(res$CI_lower < res$SHR & res$SHR < res$CI_upper))
  expect_true(all(res$p_value >= 0 & res$p_value <= 1))
})

test_that("assoc_competing() n_compete matches expected competing event count", {
  res <- suppressMessages(
    assoc_competing(DT, "censoring_type", "followup_years", "t2d_01",
                    event_val = 1L, compete_val = 2L)
  )
  expected_compete <- sum(DT$death_01)
  # n_compete in model may be <= raw count (complete cases only)
  expect_lte(res[res$model == "Unadjusted", "n_compete"][[1L]], expected_compete + 1L)
})

test_that("assoc_competing() Fully adjusted n <= Unadjusted n", {
  res <- suppressMessages(
    assoc_competing(DT, "censoring_type", "followup_years", "t2d_01",
                    event_val = 1L, compete_val = 2L,
                    covariates = COVS_FULL)
  )
  n_full  <- res[res$model == "Fully adjusted", "n"][[1L]]
  n_unadj <- res[res$model == "Unadjusted",     "n"][[1L]]
  expect_lte(n_full, n_unadj)
})

test_that("assoc_competing() factor exposure produces correct term names", {
  res <- suppressMessages(
    assoc_competing(DT, "censoring_type", "followup_years", "bmi_cat",
                    event_val = 1L, compete_val = 2L)
  )
  expect_true("bmi_catOverweight" %in% res$term)
  expect_true("bmi_catObese"      %in% res$term)
})


# ===========================================================================
# assoc_lag() — integration
# ===========================================================================

test_that("assoc_lag() runs on n=2000 with lag=c(0,1,2) without error", {
  expect_no_error(suppressMessages(
    assoc_lag(DT, "copd_01", "followup_years", "t2d_01",
              lag_years  = c(0, 1, 2),
              covariates = COVS_FULL)
  ))
})

test_that("assoc_lag() lag=0 n equals total cohort size", {
  res <- suppressMessages(
    assoc_lag(DT, "copd_01", "followup_years", "t2d_01",
              lag_years = c(0, 2))
  )
  n_lag0 <- res[res$lag_years == 0 & res$model == "Unadjusted", "n"][[1L]]
  expect_equal(n_lag0, nrow(DT))
})

test_that("assoc_lag() HR positive and CI ordered for all lags", {
  res <- suppressMessages(
    assoc_lag(DT, "copd_01", "followup_years", "t2d_01",
              lag_years = c(1, 2, 5))
  )
  expect_true(all(res$HR > 0))
  expect_true(all(res$CI_lower < res$HR & res$HR < res$CI_upper))
})

test_that("assoc_lag() n_excluded monotonically increases with lag", {
  res <- suppressMessages(
    assoc_lag(DT, "copd_01", "followup_years", "t2d_01",
              lag_years = c(1, 2, 5))
  )
  excl <- unique(res[, c("lag_years", "n_excluded")])
  data.table::setorder(excl, lag_years)
  expect_true(all(diff(excl$n_excluded) >= 0))
})

test_that("assoc_lag() 3 lags × 2 exposures × 3 models = 18 rows", {
  res <- suppressMessages(
    assoc_lag(DT, "copd_01", "followup_years",
              exposure_col = c("t2d_01", "bmi_num"),
              lag_years    = c(1, 2, 5),
              covariates   = COVS_FULL)
  )
  expect_equal(nrow(res), 18L)
})

test_that("assoc_lag() column order: lag_years directly after model", {
  res <- suppressMessages(
    assoc_lag(DT, "copd_01", "followup_years", "t2d_01", lag_years = 2)
  )
  pos_model    <- which(names(res) == "model")
  pos_lag      <- which(names(res) == "lag_years")
  pos_excluded <- which(names(res) == "n_excluded")
  expect_equal(pos_lag,      pos_model + 1L)
  expect_equal(pos_excluded, pos_model + 2L)
})

# =============================================================================
# job.R — Monitor and retrieve DNAnexus table-exporter job results
# =============================================================================


#' Check the current state of a DNAnexus job
#'
#' Returns the current state of a job submitted by \code{extract_batch()}.
#' For failed jobs, the failure message is attached as an attribute.
#'
#' @param job_id (character) Job ID returned by \code{extract_batch()},
#'   e.g. \code{"job-XXXX"}.
#'
#' @return A named character string — the job state. Possible values:
#'   \describe{
#'     \item{\code{"idle"}}{Queued, waiting to be scheduled.}
#'     \item{\code{"runnable"}}{Resources being allocated.}
#'     \item{\code{"running"}}{Actively executing.}
#'     \item{\code{"done"}}{Completed successfully.}
#'     \item{\code{"failed"}}{Failed; see \code{attr(result, "failure_message")}.}
#'     \item{\code{"terminated"}}{Manually terminated.}
#'   }
#' @export
#'
#' @examples
#' \dontrun{
#' job_id <- extract_batch(c(31, 53, 21022))
#' job_status(job_id)
#'
#' s <- job_status(job_id)
#' if (s == "failed") cli::cli_inform(attr(s, "failure_message"))
#' }
job_status <- function(job_id) {
  .assert_job_id(job_id)

  desc  <- .dx_job_describe(job_id)
  state <- desc$state

  result <- structure(state, names = job_id)

  if (identical(state, "failed")) {
    msg <- if (!is.null(desc$failureMessage)) desc$failureMessage else
           if (!is.null(desc$failureReason))  desc$failureReason  else
           "(no message)"
    attr(result, "failure_message") <- msg
  }

  result
}


#' Wait for a DNAnexus job to finish
#'
#' Polls \code{job_status()} at regular intervals until the job reaches a
#' terminal state (\code{"done"}, \code{"failed"}, or \code{"terminated"}).
#' Stops with an error if the job fails, is terminated, or times out.
#'
#' @param job_id (character) Job ID returned by \code{extract_batch()}.
#' @param interval (integer) Polling interval in seconds. Default: \code{30}.
#' @param timeout (numeric) Maximum wait time in seconds. Default: \code{Inf}
#'   (wait indefinitely). On UKB RAP, jobs can stay in \code{"runnable"} for
#'   several hours during peak times — set a finite value (e.g. \code{7200})
#'   only if you need a hard deadline.
#' @param verbose (logical) Print state and elapsed time at each poll.
#'   Default: \code{TRUE}.
#'
#' @return Invisibly returns the final state string (\code{"done"}).
#' @export
#'
#' @examples
#' \dontrun{
#' job_id <- extract_batch(c(31, 53, 21022))
#' job_wait(job_id)
#'
#' # Read result immediately after completion (RAP only)
#' job_wait(job_id)
#' df <- job_result(job_id)
#' }
job_wait <- function(job_id, interval = 30, timeout = Inf, verbose = TRUE) {
  .assert_job_id(job_id)

  terminal <- c("done", "failed", "terminated")
  start    <- proc.time()[["elapsed"]]

  repeat {
    elapsed <- as.integer(proc.time()[["elapsed"]] - start)

    # Reason: guard with is.finite() so timeout = Inf never triggers this branch
    if (is.finite(timeout) && elapsed >= timeout) {
      cli::cli_abort("Timed out after {timeout} seconds waiting for {job_id}.", call = NULL)
    }

    desc  <- .dx_job_describe(job_id)
    state <- desc$state
    hms   <- sprintf(
      "%02d:%02d:%02d",
      elapsed %/% 3600L, (elapsed %% 3600L) %/% 60L, elapsed %% 60L
    )

    if (verbose) {
      cli::cli_inform("[{hms}] {job_id} - {state}")
    }

    if (state %in% terminal) break

    Sys.sleep(interval)
  }

  if (state == "failed") {
    msg <- if (!is.null(desc$failureMessage)) desc$failureMessage else
           if (!is.null(desc$failureReason))  desc$failureReason  else
           "(no message)"
    cli::cli_abort("Job {job_id} failed: {msg}", call = NULL)
  }

  if (state == "terminated") {
    cli::cli_abort("Job {job_id} was terminated.", call = NULL)
  }

  invisible(state)
}


#' Get the RAP file path of a completed DNAnexus job output
#'
#' Returns the absolute \code{/mnt/project/} path of the CSV produced by
#' \code{extract_batch()}. Use this to read the file directly on the RAP
#' without downloading.
#'
#' @param job_id (character) Job ID returned by \code{extract_batch()}.
#'
#' @return A character string — the absolute path to the output CSV under
#'   \code{/mnt/project/}.
#' @export
#'
#' @examples
#' \dontrun{
#' path <- job_path(job_id)
#' df   <- data.table::fread(path)
#' }
job_path <- function(job_id) {
  .assert_job_id(job_id)

  desc  <- .dx_job_describe(job_id)
  state <- desc$state

  if (state != "done") {
    cli::cli_abort(
      "Job {job_id} is '{state}', not 'done'. Use {.fn job_wait} to wait for completion.",
      call = NULL
    )
  }

  file_id <- .dx_job_output_id(desc)
  .dx_file_path(file_id)
}


#' Load the result of a completed DNAnexus job into R
#'
#' Reads the output CSV produced by \code{extract_batch()} directly from RAP
#' project storage and returns a \code{data.table}. Must be run inside the
#' RAP environment.
#'
#' @param job_id (character) Job ID returned by \code{extract_batch()}.
#'
#' @return A \code{data.table} with one row per participant.
#' @export
#'
#' @examples
#' \dontrun{
#' job_id <- extract_batch(c(31, 53, 21022))
#' job_wait(job_id)
#' df <- job_result(job_id)
#' }
job_result <- function(job_id) {
  .assert_on_rap()

  .assert_job_id(job_id)

  path <- job_path(job_id)
  cli::cli_inform("Reading {.file {path}}")

  # Reason: integer64 = "double" avoids bit64 dependency; UKB eids are
  # 7-digit integers, well within double precision range
  dt <- data.table::fread(path, data.table = TRUE, integer64 = "double")
  cli::cli_inform("{nrow(dt)} rows x {ncol(dt)} cols (incl. eid)")
  dt
}


#' List recent DNAnexus jobs in the current project
#'
#' Returns a summary of the most recent jobs, optionally filtered by state.
#' Useful for quickly reviewing which jobs have completed, failed, or are
#' still running.
#'
#' @param n (integer) Maximum number of recent jobs to return. Must be a
#'   single positive integer. Default: \code{20}.
#' @param state (character) Filter by state(s). Must be \code{NULL} or a
#'   character vector of valid states:
#'   \code{"idle"}, \code{"runnable"}, \code{"running"}, \code{"done"},
#'   \code{"failed"}, \code{"terminated"}.
#'   Default: \code{NULL} (return all).
#'
#' @return A data.frame with columns:
#'   \describe{
#'     \item{job_id}{Job ID string, e.g. \code{"job-XXXX"}.}
#'     \item{name}{Job name (typically \code{"Table exporter"}).}
#'     \item{state}{Current job state.}
#'     \item{created}{Job creation time (\code{POSIXct}).}
#'     \item{runtime}{Runtime string (e.g. \code{"0:04:36"}), \code{NA} if
#'       still running.}
#'   }
#' @export
#'
#' @examples
#' \dontrun{
#' job_ls()
#' job_ls(n = 5)
#' job_ls(state = "failed")
#' job_ls(state = c("done", "failed"))
#' }
job_ls <- function(n = 20, state = NULL) {
  n     <- .assert_count(n)
  state <- .assert_choices(
    state,
    c("idle", "runnable", "running", "done", "failed", "terminated")
  )

  result <- .dx_find_jobs_raw(n)
  if (!result$success) {
    cli::cli_abort("Failed to list jobs: {result$stderr}", call = NULL)
  }

  df <- .dx_parse_jobs(result$stdout)

  if (!is.null(state) && nrow(df) > 0) {
    df <- df[df$state %in% state, , drop = FALSE]
  }

  df
}

# =============================================================================
# test-job.R — Unit tests for job_ series (mock-based, no network required)
# =============================================================================
.skip_if_no_mockery()

# Helper: build a fake .dx_run() result
.fake_dx <- function(stdout = "", stderr = "", status = 0) {
  list(stdout = stdout, stderr = stderr, status = status, success = status == 0)
}

# Fake describe list — done job
.fake_desc_done <- function() {
  list(
    state          = "done",
    runInput       = list(output = "ad_pheno"),
    output         = list(csv = list(list(`$dnanexus_link` = "file-XXXX"))),
    failureMessage = NULL,
    failureReason  = NULL
  )
}

# Fake describe list — failed job
.fake_desc_failed <- function() {
  list(
    state          = "failed",
    runInput       = list(output = "ad_pheno"),
    output         = NULL,
    failureMessage = "AppError: invalid field names",
    failureReason  = "AppError"
  )
}

# Fake describe list — running job
.fake_desc_running <- function() {
  list(
    state          = "running",
    runInput       = list(output = "ad_pheno"),
    output         = list(),
    failureMessage = NULL,
    failureReason  = NULL
  )
}

# Fake dx find jobs stdout (3 real jobs + trailing "More results" line)
.fake_find_jobs_stdout <- function() {
  paste(
    "* Table exporter (table-exporter:main) (done) job-AAAAAAAAAAAA",
    "  user001 2026-03-03 17:15:21 (runtime 0:04:36)",
    "* Table exporter (table-exporter:main) (failed) job-BBBBBBBBBBBB",
    "  user001 2026-03-03 16:57:07 (runtime 0:03:25)",
    "* Table exporter (table-exporter:main) (running) job-CCCCCCCCCCCC",
    "  user001 2026-03-03 18:27:02",
    "* More results not shown; use -n to increase number of results or",
    "  --created-before to show older results",
    sep = "\n"
  )
}

# ===========================================================================
# .dx_job_output_id() — pure function, no mocking needed
# ===========================================================================

test_that(".dx_job_output_id() returns file ID from done describe", {
  id <- ukbflow:::.dx_job_output_id(.fake_desc_done())
  expect_equal(id, "file-XXXX")
})

test_that(".dx_job_output_id() stops when output is NULL (failed job)", {
  expect_error(
    ukbflow:::.dx_job_output_id(.fake_desc_failed()),
    "no output CSV"
  )
})

test_that(".dx_job_output_id() stops when output$csv is empty list", {
  desc <- .fake_desc_done()
  desc$output$csv <- list()
  expect_error(ukbflow:::.dx_job_output_id(desc), "no output CSV")
})

# ===========================================================================
# .dx_parse_jobs() — pure function, no mocking needed
# ===========================================================================

test_that(".dx_parse_jobs() returns empty data.frame for empty stdout", {
  result <- ukbflow:::.dx_parse_jobs("")
  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 0)
  expect_named(result, c("job_id", "name", "state", "created", "runtime"))
})

test_that(".dx_parse_jobs() parses done/failed/running jobs correctly", {
  result <- ukbflow:::.dx_parse_jobs(.fake_find_jobs_stdout())
  expect_equal(nrow(result), 3)
  expect_equal(result$job_id,  c("job-AAAAAAAAAAAA", "job-BBBBBBBBBBBB", "job-CCCCCCCCCCCC"))
  expect_equal(result$state,   c("done", "failed", "running"))
})

test_that(".dx_parse_jobs() skips 'More results not shown' line", {
  result <- ukbflow:::.dx_parse_jobs(.fake_find_jobs_stdout())
  expect_equal(nrow(result), 3)
  expect_false(any(grepl("More results", result$job_id)))
})

test_that(".dx_parse_jobs() sets NA runtime for running job", {
  result <- ukbflow:::.dx_parse_jobs(.fake_find_jobs_stdout())
  expect_true(is.na(result$runtime[result$state == "running"]))
  expect_equal(result$runtime[result$state == "done"], "0:04:36")
})

test_that(".dx_parse_jobs() returns POSIXct created column", {
  result <- ukbflow:::.dx_parse_jobs(.fake_find_jobs_stdout())
  expect_s3_class(result$created, "POSIXct")
})

test_that(".dx_parse_jobs() returns correct column names", {
  result <- ukbflow:::.dx_parse_jobs(.fake_find_jobs_stdout())
  expect_named(result, c("job_id", "name", "state", "created", "runtime"))
})

# ===========================================================================
# job_status()
# ===========================================================================

test_that("job_status() returns named character for done job", {
  mockery::stub(job_status, ".dx_job_describe", function(...) .fake_desc_done())
  result <- job_status("job-XXXX")
  expect_type(result, "character")
  expect_equal(result[[1]], "done")
  expect_equal(names(result), "job-XXXX")
})

test_that("job_status() returns 'failed' with failure_message attr", {
  mockery::stub(job_status, ".dx_job_describe", function(...) .fake_desc_failed())
  result <- job_status("job-XXXX")
  expect_equal(result[[1]], "failed")
  expect_equal(attr(result, "failure_message"), "AppError: invalid field names")
})

test_that("job_status() failure_message falls back to failureReason", {
  desc <- .fake_desc_failed()
  desc$failureMessage <- NULL
  mockery::stub(job_status, ".dx_job_describe", function(...) desc)
  result <- job_status("job-XXXX")
  expect_equal(attr(result, "failure_message"), "AppError")
})

test_that("job_status() returns NULL failure_message for done job", {
  mockery::stub(job_status, ".dx_job_describe", function(...) .fake_desc_done())
  result <- job_status("job-XXXX")
  expect_null(attr(result, "failure_message"))
})

test_that("job_status() stops on invalid job_id format", {
  expect_error(job_status("notajob"), "job-XXXX")
})

test_that("job_status() stops on NA job_id", {
  expect_error(job_status(NA_character_))
})

test_that("job_status() stops on length > 1 job_id", {
  expect_error(job_status(c("job-AAA", "job-BBB")))
})

# ===========================================================================
# job_wait()
# ===========================================================================

test_that("job_wait() returns invisibly when job is already done", {
  mockery::stub(job_wait, ".dx_job_describe", function(...) .fake_desc_done())
  expect_invisible(suppressMessages(job_wait("job-XXXX", verbose = FALSE)))
  result <- suppressMessages(job_wait("job-XXXX", verbose = FALSE))
  expect_equal(result, "done")
})

test_that("job_wait() stops with message when job failed", {
  mockery::stub(job_wait, ".dx_job_describe", function(...) .fake_desc_failed())
  expect_error(
    suppressMessages(job_wait("job-XXXX", verbose = FALSE)),
    "AppError: invalid field names"
  )
})

test_that("job_wait() stops when job is terminated", {
  desc_terminated <- .fake_desc_done()
  desc_terminated$state <- "terminated"
  mockery::stub(job_wait, ".dx_job_describe", function(...) desc_terminated)
  expect_error(
    suppressMessages(job_wait("job-XXXX", verbose = FALSE)),
    "terminated"
  )
})

test_that("job_wait() stops on timeout", {
  mockery::stub(job_wait, ".dx_job_describe", function(...) .fake_desc_running())
  expect_error(
    suppressMessages(job_wait("job-XXXX", timeout = 0, verbose = FALSE)),
    "Timed out"
  )
})

test_that("job_wait() stops on invalid job_id format", {
  expect_error(job_wait("notajob"), "job-XXXX")
})


# ===========================================================================
# job_path()
# ===========================================================================

test_that("job_path() stops on invalid job_id format", {
  expect_error(job_path("notajob"), "job-XXXX")
})

test_that("job_path() stops when job is not done", {
  mockery::stub(job_path, ".dx_job_describe", function(...) .fake_desc_running())
  expect_error(suppressMessages(job_path("job-XXXX")), "not 'done'")
})

test_that("job_path() stops when job failed", {
  mockery::stub(job_path, ".dx_job_describe", function(...) .fake_desc_failed())
  expect_error(suppressMessages(job_path("job-XXXX")), "not 'done'")
})

test_that("job_path() returns /mnt/project/ path for done job", {
  mockery::stub(job_path, ".dx_job_describe", function(...) .fake_desc_done())
  mockery::stub(job_path, ".dx_file_path",
                function(...) "/mnt/project/ad_pheno.csv")
  result <- job_path("job-XXXX")
  expect_equal(result, "/mnt/project/ad_pheno.csv")
})

# ===========================================================================
# job_result()
# ===========================================================================

test_that("job_result() stops when not on RAP", {
  mockery::stub(job_result, ".assert_on_rap", function() cli::cli_abort("This function must be run inside the RAP environment."))
  expect_error(job_result("job-XXXX"), "RAP environment")
})

test_that("job_result() stops on invalid job_id format", {
  mockery::stub(job_result, ".assert_on_rap", function() invisible(NULL))
  expect_error(job_result("notajob"), "job-XXXX")
})

test_that("job_result() stops when job is not done", {
  mockery::stub(job_result, ".assert_on_rap", function() invisible(NULL))
  mockery::stub(job_result, "job_path", function(...) {
    stop("Job job-XXXX is 'running', not 'done'.")
  })
  expect_error(suppressMessages(job_result("job-XXXX")), "not 'done'")
})

test_that("job_result() returns a data.table when on RAP", {
  tmp <- tempfile(fileext = ".csv")
  write.csv(data.frame(eid = c(1L, 2L), p31 = c(0L, 1L)), tmp, row.names = FALSE)

  mockery::stub(job_result, ".assert_on_rap", function() invisible(NULL))
  mockery::stub(job_result, "job_path", function(...) tmp)

  result <- suppressMessages(job_result("job-XXXX"))
  expect_true(data.table::is.data.table(result))
  expect_equal(nrow(result), 2L)
})

# ===========================================================================
# job_ls()
# ===========================================================================

test_that("job_ls() returns a data.frame with correct columns", {
  mockery::stub(job_ls, ".dx_find_jobs_raw",
                function(...) .fake_dx(stdout = .fake_find_jobs_stdout()))
  result <- job_ls()
  expect_s3_class(result, "data.frame")
  expect_named(result, c("job_id", "name", "state", "created", "runtime"))
})

test_that("job_ls() returns all jobs when state = NULL", {
  mockery::stub(job_ls, ".dx_find_jobs_raw",
                function(...) .fake_dx(stdout = .fake_find_jobs_stdout()))
  result <- job_ls()
  expect_equal(nrow(result), 3L)
})

test_that("job_ls() filters by state correctly", {
  mockery::stub(job_ls, ".dx_find_jobs_raw",
                function(...) .fake_dx(stdout = .fake_find_jobs_stdout()))
  result <- job_ls(state = "done")
  expect_equal(nrow(result), 1L)
  expect_true(all(result$state == "done"))
})

test_that("job_ls() filters by multiple states", {
  mockery::stub(job_ls, ".dx_find_jobs_raw",
                function(...) .fake_dx(stdout = .fake_find_jobs_stdout()))
  result <- job_ls(state = c("done", "failed"))
  expect_equal(nrow(result), 2L)
  expect_true(all(result$state %in% c("done", "failed")))
})

test_that("job_ls() returns empty data.frame when no jobs match state filter", {
  mockery::stub(job_ls, ".dx_find_jobs_raw",
                function(...) .fake_dx(stdout = .fake_find_jobs_stdout()))
  result <- job_ls(state = "terminated")
  expect_equal(nrow(result), 0L)
})

test_that("job_ls() stops when dx find jobs fails", {
  mockery::stub(job_ls, ".dx_find_jobs_raw",
                function(...) .fake_dx(stderr = "Not logged in", status = 1))
  expect_error(job_ls(), "Failed to list jobs")
})

# ---------------------------------------------------------------------------
# job_ls() — n validation
# ---------------------------------------------------------------------------

test_that("job_ls() stops when n is character", {
  expect_error(job_ls(n = "five"), "positive integer")
})

test_that("job_ls() stops when n is 0", {
  expect_error(job_ls(n = 0), "positive integer")
})

test_that("job_ls() stops when n is negative", {
  expect_error(job_ls(n = -1), "positive integer")
})

test_that("job_ls() stops when n is Inf", {
  expect_error(job_ls(n = Inf), "positive integer")
})

test_that("job_ls() stops when n is non-whole number", {
  expect_error(job_ls(n = 1.5), "positive integer")
})

# ---------------------------------------------------------------------------
# job_ls() — state validation
# ---------------------------------------------------------------------------

test_that("job_ls() stops when state is non-character", {
  expect_error(job_ls(state = TRUE), "character vector")
})

test_that("job_ls() stops when state contains an invalid value", {
  expect_error(job_ls(state = "oops"), "oops")
})

test_that("job_ls() stops when state contains mix of valid and invalid values", {
  expect_error(job_ls(state = c("done", "oops")), "oops")
})

# =============================================================================
# test-integration-job.R — Integration tests for job_ series
# Requires real dx-toolkit, token, and network connection
# Run manually before release: devtools::test(filter = "integration-job")
# =============================================================================

skip_on_cran()

# Known job IDs from this project (used as stable test fixtures)
# State is immutable once terminal — safe to hardcode
JOB_DONE   <- "job-J6YBJ7QJ0QQ82Kz0J52x2jGp"   # done, ukb_pheno_20260303_171503
JOB_FAILED <- "job-J6YB4jjJ0QQ3fVJ0yPqG3b1K"   # failed, CRLF error


# ===========================================================================
# job_ls()
# ===========================================================================

test_that("job_ls() returns a data.frame with correct columns", {
  .skip_if_no_rap()
  result <- job_ls(n = 5)
  expect_s3_class(result, "data.frame")
  expect_named(result, c("job_id", "name", "state", "created", "runtime"))
})

test_that("job_ls() returns at least one job", {
  .skip_if_no_rap()
  result <- job_ls(n = 5)
  expect_gt(nrow(result), 0)
})

test_that("job_ls() job_id values follow job-XXXX format", {
  .skip_if_no_rap()
  result <- job_ls(n = 5)
  expect_true(all(grepl("^job-", result$job_id)))
})

test_that("job_ls() state column contains only valid states", {
  .skip_if_no_rap()
  valid_states <- c("idle", "runnable", "running", "done", "failed", "terminated")
  result <- job_ls(n = 20)
  expect_true(all(result$state %in% valid_states))
})

test_that("job_ls() created column is POSIXct", {
  .skip_if_no_rap()
  result <- job_ls(n = 5)
  expect_s3_class(result$created, "POSIXct")
})

test_that("job_ls() state filter returns only matching rows", {
  .skip_if_no_rap()
  result <- job_ls(n = 20, state = "done")
  if (nrow(result) > 0) {
    expect_true(all(result$state == "done"))
  } else {
    skip("No done jobs found to test filter.")
  }
})


# ===========================================================================
# job_status()
# ===========================================================================

test_that("job_status() returns 'done' for known done job", {
  .skip_if_no_rap()
  result <- job_status(JOB_DONE)
  expect_equal(result[[1]], "done")
})

test_that("job_status() returns named character with job_id as name", {
  .skip_if_no_rap()
  result <- job_status(JOB_DONE)
  expect_type(result, "character")
  expect_equal(names(result), JOB_DONE)
})

test_that("job_status() returns 'failed' for known failed job", {
  .skip_if_no_rap()
  result <- job_status(JOB_FAILED)
  expect_equal(result[[1]], "failed")
})

test_that("job_status() attaches failure_message for failed job", {
  .skip_if_no_rap()
  result <- job_status(JOB_FAILED)
  expect_false(is.null(attr(result, "failure_message")))
  expect_true(nzchar(attr(result, "failure_message")))
})


# ===========================================================================
# job_wait()
# ===========================================================================

test_that("job_wait() returns 'done' immediately for already-done job", {
  .skip_if_no_rap()
  result <- suppressMessages(job_wait(JOB_DONE, verbose = FALSE))
  expect_equal(result, "done")
})

test_that("job_wait() stops with error message for known failed job", {
  .skip_if_no_rap()
  expect_error(
    suppressMessages(job_wait(JOB_FAILED, verbose = FALSE)),
    "failed"
  )
})


# ===========================================================================
# job_path()
# ===========================================================================

test_that("job_path() returns a /mnt/project/ path for known done job", {
  skip("job_path() requires real RAP environment with valid job ID")
  .skip_if_no_rap()
  result <- job_path(JOB_DONE)
  expect_type(result, "character")
  expect_true(startsWith(result, "/mnt/project/"))
  expect_true(endsWith(result, ".csv"))
})

test_that("job_path() stops for known failed job", {
  .skip_if_no_rap()
  expect_error(suppressMessages(job_path(JOB_FAILED)), "not 'done'")
})


# ===========================================================================
# job_result()
# ===========================================================================

# Note: job_result() requires RAP environment and a job whose output file
# still exists on the project. Use an active job for full integration testing:
#
#   job_id <- extract_batch(c(31, 21022), file = "integration_test")
#   job_wait(job_id)
#   df <- job_result(job_id)
#   # clean up: dx rm <output_file_id>

# =============================================================================
# derive.R - analysis-ready variable preparation for UKB data
# =============================================================================


#' Handle informative missing labels in UKB decoded data
#'
#' After \code{\link{decode_values}} converts categorical codes to character
#' labels, some values represent meaningful non-response rather than true data:
#' \code{"Do not know"}, \code{"Prefer not to answer"}, and
#' \code{"Prefer not to say"}. This function either converts them to \code{NA}
#' (for complete-case analysis) or retains them as \code{"Unknown"} (to
#' preserve the informative missingness as a model category).
#'
#' Empty strings (\code{""}) are always converted to \code{NA} regardless of
#' \code{action}, as they carry no informational content.
#'
#' Only character columns are modified; numeric, integer, Date, and logical
#' columns are silently skipped.
#'
#' \strong{data.table pass-by-reference}: when the input is a
#' \code{data.table}, modifications are made in-place via
#' \code{\link[data.table]{set}}. The returned object and the original variable
#' point to the same memory. If you need to preserve the original, pass
#' \code{data.table::copy(data)} instead.
#'
#' @param data (data.frame or data.table) Decoded UKB data, typically output
#'   of \code{\link{decode_values}} followed by \code{\link{decode_names}}.
#' @param cols (tidyselect) Columns to process. Default: \code{everything()}
#'   (all columns). Non-character columns in the selection are silently skipped.
#' @param action (character) One of \code{"na"} (default) or \code{"unknown"}.
#'   \itemize{
#'     \item \code{"na"}: convert all informative missing labels to \code{NA}.
#'     \item \code{"unknown"}: convert informative missing labels to
#'       \code{"Unknown"}, preserving them as a distinct model category.
#'   }
#'   Empty strings are always converted to \code{NA} regardless of this
#'   parameter.
#' @param extra_labels (character or NULL) Additional labels to treat as
#'   informative missing, appended to the built-in list. Default: \code{NULL}.
#'
#' @return The input \code{data} with missing labels replaced in-place.
#'   Always returns a \code{data.table}.
#' @export
#'
#' @examples
#' dt <- ops_toy(n = 100)
#' derive_missing(dt)
#'
#' # Retain informative non-response as a model category
#' derive_missing(dt, action = "unknown")
#'
#' # Add a custom label
#' derive_missing(dt, extra_labels = "Not applicable")
derive_missing <- function(data,
                           cols         = tidyselect::everything(),
                           action       = c("na", "unknown"),
                           extra_labels = NULL) {

  .assert_data_frame(data)

  action <- match.arg(action)

  if (!is.null(extra_labels)) {
    .assert_character(extra_labels)
    extra_labels <- extra_labels[nchar(trimws(extra_labels)) > 0L]
  }

  # Convert to data.table in-place (no copy for data.table input)
  if (!data.table::is.data.table(data)) {
    data <- data.table::as.data.table(data)
  }

  # Resolve tidyselect column indices, then keep only character columns
  selected_idx  <- tidyselect::eval_select(rlang::enquo(cols), data)
  selected_cols <- names(selected_idx)
  char_cols     <- selected_cols[vapply(data[, selected_cols, with = FALSE],
                                        is.character, logical(1L))]

  if (length(char_cols) == 0L) {
    cli::cli_alert_info("No character columns found in selection - nothing to do.")
    return(invisible(data))
  }

  # Build label sets
  informative_labels <- c(.ukb_informative_missing, extra_labels)
  replace_value      <- if (action == "na") NA_character_ else "Unknown"

  n_cols_affected  <- 0L
  n_vals_replaced  <- 0L

  for (col in char_cols) {

    x          <- data[[col]]
    empty_idx  <- which(x == "")
    inform_idx <- which(x %in% informative_labels)

    n_replaced <- length(empty_idx) + length(inform_idx)
    if (n_replaced == 0L) next

    # "" always -> NA; informative labels -> NA or "Unknown" per action
    data.table::set(data, i = empty_idx,  j = col, value = NA_character_)
    data.table::set(data, i = inform_idx, j = col, value = replace_value)

    n_cols_affected <- n_cols_affected + 1L
    n_vals_replaced <- n_vals_replaced + n_replaced
  }

  cli::cli_alert_success(
    "derive_missing: replaced {n_vals_replaced} value{?s} across {n_cols_affected} column{?s} (action = {.val {action}})."
  )

  data
}


#' Prepare UKB covariates for analysis
#'
#' Converts decoded UKB columns to analysis-ready types: character-encoded
#' numeric fields to \code{numeric}, and categorical fields to \code{factor}.
#' Prints a concise summary for each converted column - mean / median / SD /
#' missing rate for numeric columns, and level counts for factor columns - so
#' you can verify distributions without leaving the pipeline.
#'
#' \strong{data.table pass-by-reference}: when the input is a
#' \code{data.table}, modifications are made in-place. Pass
#' \code{data.table::copy(data)} to preserve the original.
#'
#' @param data (data.frame or data.table) UKB data, typically output of
#'   \code{\link{derive_missing}}.
#' @param as_numeric (character or NULL) Column names to convert to
#'   \code{numeric}. Values that cannot be coerced (e.g. residual text) become
#'   \code{NA} with a warning. Default: \code{NULL}.
#' @param as_factor (character or NULL) Column names to convert to
#'   \code{factor}. Default levels are the sorted unique non-NA values unless
#'   overridden by \code{factor_levels}. Default: \code{NULL}.
#' @param factor_levels (named list or NULL) Custom level ordering for specific
#'   factor columns. Names must match entries in \code{as_factor}; values are
#'   character vectors of levels in the desired order (first level = reference
#'   group in regression). Columns not listed use default ordering.
#'   Default: \code{NULL}.
#' @param max_levels (integer) Factor columns with more levels than this
#'   threshold trigger a warning suggesting the user consider collapsing
#'   categories. Default: \code{5L}.
#'
#' @return The input \code{data} with converted columns. Always returns a
#'   \code{data.table}.
#' @export
#'
#' @examples
#' dt <- ops_toy(n = 100)
#' dt <- derive_missing(dt)
#' # messy_label is a character column with numeric-looking values ("-1", "999")
#' # mixed with non-parseable strings — demonstrates coercion and NA warnings
#' derive_covariate(dt,
#'   as_numeric = "messy_label",
#'   as_factor  = c("p31", "p20116_i0"),
#'   factor_levels = list(
#'     p20116_i0 = c("Never", "Previous", "Current")
#'   )
#' )
derive_covariate <- function(data,
                             as_numeric    = NULL,
                             as_factor     = NULL,
                             factor_levels = NULL,
                             max_levels    = 5L) {

  .assert_data_frame(data)
  if (!is.null(factor_levels) && !is.list(factor_levels)) {
    cli::cli_abort("{.arg factor_levels} must be a named list or NULL.", call = NULL)
  }
  if (!is.null(factor_levels) && is.null(names(factor_levels))) {
    cli::cli_abort("{.arg factor_levels} must be a named list (names = column names).", call = NULL)
  }

  if (!data.table::is.data.table(data)) {
    data <- data.table::as.data.table(data)
  }

  n_rows <- nrow(data)

  # ── as_numeric ──────────────────────────────────────────────────────────────
  if (!is.null(as_numeric)) {

    bad        <- setdiff(as_numeric, names(data))
    as_numeric <- intersect(as_numeric, names(data))
    if (length(bad) > 0L) {
      cli::cli_alert_warning("as_numeric: columns not found, skipped: {.val {bad}}")
    }

    if (length(as_numeric) > 0L) {
      cli::cli_rule(left = "Numeric")

      for (col in as_numeric) {
        na_before <- sum(is.na(data[[col]]))
        data.table::set(data, j = col,
                        value = suppressWarnings(as.numeric(data[[col]])))
        na_new <- sum(is.na(data[[col]])) - na_before
        if (na_new > 0L) {
          cli::cli_alert_warning(
            "{.field {col}}: {na_new} value{?s} coerced to NA."
          )
        }
        .cli_summarise_numeric(data[[col]], col, n_rows)
      }
    }
  }

  # ── as_factor ───────────────────────────────────────────────────────────────
  if (!is.null(as_factor)) {

    bad       <- setdiff(as_factor, names(data))
    as_factor <- intersect(as_factor, names(data))
    if (length(bad) > 0L) {
      cli::cli_alert_warning("as_factor: columns not found, skipped: {.val {bad}}")
    }

    if (length(as_factor) > 0L) {
      cli::cli_rule(left = "Factor")

      for (col in as_factor) {
        lvls  <- factor_levels[[col]]   # NULL -> default sorted levels
        f_val <- if (is.null(lvls)) factor(data[[col]]) else factor(data[[col]], levels = lvls)
        data.table::set(data, j = col, value = f_val)
        .cli_summarise_factor(data[[col]], col, n_rows, max_levels)
      }
    }
  }

  data
}


#' Cut a continuous UKB variable into quantile-based or custom groups
#'
#' Creates a new factor column by binning a continuous variable into \code{n}
#' groups. When \code{breaks} is omitted, group boundaries are derived from
#' quantiles of the observed data (equal-frequency binning). When
#' \code{breaks} is supplied, those values are used as interior cut points.
#'
#' Before binning, a numeric summary (mean, median, SD, Q1, Q3, missing rate)
#' is printed for the source column. After binning, the group distribution is
#' printed via an internal summary helper.
#'
#' Only one column can be processed per call; loop over columns explicitly
#' when binning multiple variables.
#'
#' \strong{data.table pass-by-reference}: when the input is a
#' \code{data.table}, the new column is added in-place. Pass
#' \code{data.table::copy(data)} to preserve the original.
#'
#' @param data (data.frame or data.table) UKB data.
#' @param col (character) Name of the source numeric column.
#' @param n (integer) Number of groups. Any integer \code{>= 2} is accepted.
#'   Commonly used values are \code{2}, \code{3}, \code{4}, \code{5}.
#' @param breaks (numeric vector or NULL) Interior cut points; length must
#'   equal \code{n - 1}. When \code{NULL} (default), quantile-based equal-
#'   frequency boundaries are computed automatically.
#' @param labels (character vector or NULL) Group labels of length \code{n}.
#'   Defaults to \code{"Q1"}, \code{"Q2"}, \ldots, \code{"Qn"}.
#' @param name (character or NULL) Name for the new column. Defaults to
#'   \code{"{col}_bi"} / \code{"{col}_tri"} / \code{"{col}_quad"} /
#'   \code{"{col}_quin"} for \code{n} = 2 / 3 / 4 / 5; other values of
#'   \code{n} produce \code{"{col}_g{n}"}.
#'
#' @return The input \code{data} with one new factor column appended. Always
#'   returns a \code{data.table}.
#' @export
#'
#' @examples
#' dt <- ops_toy(n = 100)
#' derive_cut(dt, col = "p21001_i0", n = 3)
#'
#' # Custom breaks and labels
#' derive_cut(dt, col = "p21001_i0", n = 3,
#'            breaks = c(25, 30),
#'            labels = c("<25", "25-30", "30+"),
#'            name   = "bmi_group")
derive_cut <- function(data,
                       col,
                       n,
                       breaks = NULL,
                       labels = NULL,
                       name   = NULL) {

  .assert_data_frame(data)
  .assert_has_cols(data, col)
  n <- .assert_count_min(n, min = 2L)

  if (!is.null(breaks)) {
    if (!is.numeric(breaks) || length(breaks) != n - 1L) {
      cli::cli_abort("{.arg breaks} must be a numeric vector of length n - 1 ({n - 1L}).", call = NULL)
    }
  }
  if (!is.null(labels) && length(labels) != n) {
    cli::cli_abort("{.arg labels} must be a character vector of length n ({n}).", call = NULL)
  }

  if (!data.table::is.data.table(data)) {
    data <- data.table::as.data.table(data)
  }

  n_rows <- nrow(data)
  x      <- data[[col]]

  if (!is.numeric(x)) {
    cli::cli_abort("Column {.val {col}} must be numeric.", call = NULL)
  }

  # Default name based on n
  if (is.null(name)) {
    suffix_map <- c("2" = "_bi", "3" = "_tri", "4" = "_quad", "5" = "_quin")
    suffix <- suffix_map[as.character(n)]
    name   <- if (!is.na(suffix)) paste0(col, suffix) else paste0(col, "_g", n)
  }

  # Default labels: Q1, Q2, ..., Qn
  if (is.null(labels)) labels <- paste0("Q", seq_len(n))

  # ── Print source column numeric summary ────────────────────────────────────
  cli::cli_rule(left = sprintf("Source: %s", col))
  .cli_summarise_numeric(x, col, n_rows)

  # ── Compute cut points ─────────────────────────────────────────────────────
  if (is.null(breaks)) {
    # Reason: quantile-based equal-frequency binning; include.lowest ensures
    # the minimum value falls into the first group
    cut_breaks <- stats::quantile(x, probs = seq(0, 1, length.out = n + 1L),
                                  na.rm = TRUE)
    cut_breaks <- unique(cut_breaks)   # collapse duplicate boundaries from ties
    n_actual <- length(cut_breaks) - 1L
    if (n_actual < n) {
      cli::cli_alert_warning(
        "Ties at quantile boundaries reduced the number of distinct groups (requested {n}, got {n_actual})."
      )
      # Reason: labels must match the actual number of groups after tie
      # collapsing; truncate to avoid cut() error
      labels <- labels[seq_len(n_actual)]
    }
  } else {
    cut_breaks <- c(-Inf, breaks, Inf)
  }

  f_val <- base::cut(x, breaks = cut_breaks, labels = labels,
                     include.lowest = TRUE, right = TRUE)

  # ── Add new column and print factor summary ─────────────────────────────────
  data.table::set(data, j = name, value = f_val)

  cli::cli_rule(left = sprintf("New column: %s", name))
  .cli_summarise_factor(data[[name]], name, n_rows, max_levels = n)

  data
}


#' Define a self-reported phenotype from UKB touchscreen data
#'
#' Searches UKB self-reported illness fields across all instances and arrays,
#' matches records against a user-supplied regex, parses the associated
#' year-month date, and appends two columns to the data:
#' \code{{name}_selfreport} (logical) and \code{{name}_selfreport_date}
#' (\code{IDate}, earliest matching instance).
#'
#' The function auto-detects the relevant columns using the
#' \code{extract_ls()} field dictionary cache (populated by
#' \code{extract_ls()} or \code{extract_pheno()}). The three \code{_cols}
#' parameters let you override auto-detection when column names have been
#' customised (e.g. after \code{decode_names()}).
#'
#' \strong{Field mapping by \code{field}:}
#' \itemize{
#'   \item \code{"noncancer"}: disease text = p20002, date = p20008.
#'   \item \code{"cancer"}:    disease text = p20001, date = p20006.
#' }
#' Baseline visit date (p53) is used as a fallback when no specific
#' diagnosis date is recorded.
#'
#' \strong{data.table pass-by-reference}: new columns are added in-place.
#' Pass \code{data.table::copy(data)} to preserve the original.
#'
#' @param data (data.frame or data.table) UKB data containing self-report
#'   fields.
#' @param name (character) Output column name prefix, e.g. \code{"disease"} or
#'   \code{"outcome"}.
#' @param regex (character) Regular expression matched against disease text
#'   values (after \code{tolower()}), e.g. \code{"^diabetes$"}.
#' @param field (character) Self-report field type: \code{"noncancer"}
#'   (p20002 / p20008) or \code{"cancer"} (p20001 / p20006).
#' @param ignore_case (logical) Should regex matching ignore case?
#'   Default: \code{TRUE}.
#' @param disease_cols (character or NULL) Column name(s) containing disease
#'   text (p20002 or p20001). \code{NULL} = auto-detect via
#'   \code{extract_ls()} cache.
#' @param date_cols (character or NULL) Column name(s) containing the
#'   self-report year-month date (p20008 or p20006). \code{NULL} =
#'   auto-detect.
#' @param visit_cols (character or NULL) Column name(s) containing the
#'   baseline assessment date (p53). \code{NULL} = auto-detect.
#'
#' @return The input \code{data} with two new columns appended in-place:
#'   \code{{name}_selfreport} (logical) and
#'   \code{{name}_selfreport_date} (IDate). Always returns a
#'   \code{data.table}.
#' @export
#'
#' @examples
#' \donttest{
#' dt <- ops_toy(n = 100)
#' # disease_cols / date_cols / visit_cols can be omitted — auto-detected from
#' # column names matching p20002_* / p20008_* / p53_* respectively
#' derive_selfreport(dt, name = "htn", regex = "hypertension",
#'                   field = "noncancer")
#' }
derive_selfreport <- function(data,
                              name,
                              regex,
                              field        = c("noncancer", "cancer"),
                              ignore_case  = TRUE,
                              disease_cols = NULL,
                              date_cols    = NULL,
                              visit_cols   = NULL) {

  .assert_data_frame(data)
  .assert_scalar_string(name)
  .assert_scalar_string(regex)

  field <- match.arg(field)

  if (!data.table::is.data.table(data)) {
    data <- data.table::as.data.table(data)
  }

  # ── Field ID mapping ───────────────────────────────────────────────────────
  fid_disease <- if (field == "noncancer") 20002L else 20001L
  fid_date    <- if (field == "noncancer") 20008L else 20006L
  fid_visit   <- 53L

  # ── Auto-detect columns ────────────────────────────────────────────────────
  if (is.null(disease_cols)) disease_cols <- .detect_cols_by_field(data, fid_disease)
  if (is.null(date_cols))    date_cols    <- .detect_cols_by_field(data, fid_date)
  if (is.null(visit_cols))   visit_cols   <- .detect_cols_by_field(data, fid_visit)

  status_col <- paste0(name, "_selfreport")
  date_col   <- paste0(name, "_selfreport_date")

  if (length(disease_cols) == 0L) {
    cli::cli_alert_warning(
      "derive_selfreport: no disease columns for field {fid_disease}. Supply disease_cols manually."
    )
    data[, (status_col) := FALSE]
    data[, (date_col)   := data.table::as.IDate(NA_character_)]
    return(invisible(data))
  }
  if (length(date_cols)  == 0L) cli::cli_alert_warning(
    "derive_selfreport: no date columns for field {fid_date}. Baseline (p53) used."
  )
  if (length(visit_cols) == 0L) cli::cli_alert_warning(
    "derive_selfreport: no visit columns for field {fid_visit}."
  )

  # ── Detect instances (e.g. "_i0_", "_i1_") ────────────────────────────────
  inst_raw  <- regmatches(disease_cols,
                          regexpr("_i(\\d+)_", disease_cols, perl = TRUE))
  instances <- sort(unique(sub("_i(\\d+)_", "\\1", inst_raw, perl = TRUE)))

  results <- vector("list", length(instances))
  names(results) <- instances

  for (inst in instances) {

    d_cols <- sort(grep(paste0("_i", inst, "_"), disease_cols, value = TRUE))
    t_cols <- sort(grep(paste0("_i", inst, "_"), date_cols,    value = TRUE))
    v_col  <- grep(paste0("_i", inst, "$"),      visit_cols,   value = TRUE)
    v_col  <- if (length(v_col) > 0L) v_col[1L] else NULL

    if (length(d_cols) == 0L) next

    id_cols <- if (!is.null(v_col)) c("eid", v_col) else "eid"

    # ── Melt disease text columns ────────────────────────────────────────────
    # Reason: two separate simple melts (disease + date) then join by array_idx.
    # Avoids list-melt entirely - list-melt with mixed-type sparse UKB array
    # columns is unreliable across data.table versions (may strip class or
    # ignore value.name). Simple melt on a single pre-typed column group is
    # always well-behaved.
    sub_d <- data.table::copy(data[, c(id_cols, d_cols), with = FALSE])
    for (.c in d_cols) {
      data.table::set(sub_d, j = .c, value = as.character(sub_d[[.c]]))
    }

    long_d <- data.table::melt(
      sub_d,
      id.vars       = id_cols,
      measure.vars  = d_cols,
      variable.name = "array_idx",
      value.name    = "text"
    )
    long_d[, array_idx := as.integer(array_idx)]

    # Filter: drop empty/NA, lowercase, apply regex
    long_d <- long_d[!is.na(text) & nchar(as.character(text)) > 0L]
    if (nrow(long_d) == 0L) next
    long_d[, text := tolower(as.character(text))]
    long_d <- long_d[grepl(regex, text, ignore.case = ignore_case, perl = TRUE)]
    if (nrow(long_d) == 0L) next

    # ── Melt date columns, join onto matched disease rows ────────────────────
    if (length(t_cols) > 0L) {
      sub_t <- data.table::copy(data[, c("eid", t_cols), with = FALSE])
      for (.c in t_cols) {
        data.table::set(sub_t, j = .c,
                        value = suppressWarnings(as.numeric(sub_t[[.c]])))
      }

      long_t <- data.table::melt(
        sub_t,
        id.vars       = "eid",
        measure.vars  = t_cols,
        variable.name = "array_idx",
        value.name    = "year_month"
      )
      long_t[, array_idx  := as.integer(array_idx)]
      long_t[, year_month := suppressWarnings(as.numeric(year_month))]

      # Left join: attach year_month to matched disease rows by (eid, array_idx)
      long_d <- long_t[long_d, on = c("eid", "array_idx")]
    } else {
      long_d[, year_month := NA_real_]
    }

    # ── Baseline date (IDate; integer-backed for fcoalesce compatibility) ────
    if (!is.null(v_col) && v_col %in% names(long_d)) {
      data.table::setnames(long_d, v_col, "baseline_date")
      long_d[, baseline_date := data.table::as.IDate(as.character(baseline_date))]
    } else {
      long_d[, baseline_date := data.table::as.IDate(NA_character_)]
    }

    # ── Parse year_month → IDate, fallback to baseline ──────────────────────
    # Both parsed_date (.parse_ukb_year_month) and baseline_date are IDate
    # (integer-backed), so fcoalesce() works without type coercion.
    long_d[, parsed_date   := .parse_ukb_year_month(year_month)]
    long_d[, instance_date := data.table::fcoalesce(parsed_date, baseline_date)]

    results[[inst]] <- long_d[, .(eid, instance_date)]
  }

  # ── Combine across instances ───────────────────────────────────────────────
  results <- Filter(Negate(is.null), results)

  if (length(results) == 0L) {
    data[, (status_col) := FALSE]
    data[, (date_col)   := data.table::as.IDate(NA_character_)]
    cli::cli_alert_warning("derive_selfreport ({name}): 0 cases found.")
    return(invisible(data))
  }

  combined <- data.table::rbindlist(results, fill = TRUE)

  # Status: in-place logical flag
  data[, (status_col) := eid %in% unique(combined$eid)]

  # Earliest date per eid - update-on-join (zero-copy, truly in-place)
  dates <- combined[!is.na(instance_date),
                    .(date = min(instance_date, na.rm = TRUE)),
                    by = eid]
  data[, (date_col) := data.table::as.IDate(NA_character_)]
  data[dates, on = "eid", (date_col) := i.date]

  n_cases <- sum(data[[status_col]])
  n_dated  <- sum(!is.na(data[[date_col]]))
  cli::cli_alert_success(
    "derive_selfreport ({name}): {n_cases} case{?s}, {n_dated} with date{?s}."
  )

  data
}


#' Derive a binary disease flag from UKB First Occurrence fields
#'
#' UKB pre-computes the earliest recorded date for hundreds of ICD-10 chapters
#' and categories as \emph{First Occurrence} fields (\code{p131xxx}).  Each
#' field contains a single date per participant; no array or instance depth is
#' involved.  This function reads that date, converts it to \code{IDate},
#' and writes two analysis-ready columns:
#'
#' \describe{
#'   \item{\code{{name}_fo_date}}{Earliest First Occurrence date (\code{IDate}).
#'     Values that cannot be coerced to a valid date (e.g. UKB error codes)
#'     are silently set to \code{NA}.}
#'   \item{\code{{name}_fo}}{Logical flag derived from
#'     \code{{name}_fo_date}: \code{TRUE} if and only if a valid date exists.
#'     This guarantees that every positive case has a usable date - essential
#'     for time-to-event and prevalent/incident classification.}
#' }
#'
#' \strong{Column detection}: the function locates the source column
#' automatically from \code{field}, handling both the raw format used by
#' \code{\link{extract_pheno}} (\code{participant.p131666}) and the
#' snake_case format produced by \code{\link{decode_names}}
#' (\code{date_e11_first_reported_type_2_diabetes}).  Supply \code{col}
#' to override auto-detection.
#'
#' \strong{data.table pass-by-reference}: when the input is a
#' \code{data.table}, new columns are added in-place via \code{:=}.
#' The returned object and the original variable point to the same memory.
#'
#' @param data (data.frame or data.table) UKB phenotype data.
#' @param name (character) Output column prefix, e.g. \code{"disease"} produces
#'   \code{disease_fo} and \code{disease_fo_date}.
#' @param field (integer or character) UKB field ID of the First Occurrence
#'   field, e.g. \code{131666} for E11 (type 2 diabetes).
#' @param col (character or NULL) Name of the source column in \code{data}.
#'   When \code{NULL} (default) the column is detected automatically from
#'   \code{field}.
#'
#' @return The input \code{data} (invisibly) with two new columns added
#'   in-place: \code{{name}_fo} (logical) and \code{{name}_fo_date} (IDate).
#' @export
#'
#' @examples
#' dt <- ops_toy(n = 100)
#' derive_first_occurrence(dt, name = "outcome", field = 131742L, col = "p131742")
derive_first_occurrence <- function(data, name, field, col = NULL) {

  .assert_data_frame(data)
  if (!data.table::is.data.table(data)) data.table::setDT(data)

  status_col <- paste0(name, "_fo")
  date_col   <- paste0(name, "_fo_date")

  # ── Locate source column ──────────────────────────────────────────────────
  if (!is.null(col)) {
    src_col <- col
    .assert_has_cols(data, src_col)
  } else {
    src_col <- .detect_fo_col(data, field)
    if (is.null(src_col)) {
      cli::cli_abort(
        "Cannot find First Occurrence column for field {.val {field}}. Specify {.arg col} directly or ensure {.fn extract_ls} cache is populated.",
        call = NULL
      )
    }
  }

  # ── Convert to IDate ──────────────────────────────────────────────────────
  # Reason: some p131xxx values are UKB error codes that cannot be coerced to
  # a valid date; as.IDate(as.character(...)) converts them to NA naturally.
  # Dates that predate birth or carry other clinical implausibility are also
  # surfaced as NA at this step (as.IDate returns NA for unparseable strings).
  data[, (date_col) := data.table::as.IDate(as.character(get(src_col)))]

  # ── Derive logical flag FROM date (never from raw column) ─────────────────
  # Reason: status must imply a usable date for time-to-event analysis.
  # A case without a valid date is analytically unusable.
  data[, (status_col) := !is.na(get(date_col))]

  n_cases <- sum(data[[status_col]], na.rm = TRUE)
  cli::cli_alert_success(
    "derive_first_occurrence ({name}): {n_cases} case{?s} with valid date."
  )

  invisible(data)
}


#' Derive a binary disease flag from UKB HES inpatient diagnoses
#'
#' Hospital Episode Statistics (HES) inpatient records store ICD-10 diagnosis
#' codes in field \code{p41270} (single JSON-array column on UKB RAP) and
#' corresponding first-diagnosis dates in field \code{p41280}
#' (\code{p41280_a0}, \code{p41280_a1}, \ldots).  The array index in
#' \code{p41270} and \code{p41280} are aligned: the \emph{N}-th code in the
#' JSON array corresponds to \code{p41280_aN} (date of first in-patient
#' diagnosis for that code).
#'
#' \describe{
#'   \item{\code{{name}_hes}}{Logical flag: \code{TRUE} if any HES record
#'     contains a matching ICD-10 code.}
#'   \item{\code{{name}_hes_date}}{Earliest first-diagnosis date across all
#'     matching codes (\code{IDate}).  \code{NA} if no date is available.}
#' }
#'
#' @param data (data.frame or data.table) UKB phenotype data containing HES
#'   fields (\code{p41270} and \code{p41280_a*}).
#' @param name (character) Output column prefix, e.g. \code{"disease"} produces
#'   \code{disease_hes} and \code{disease_hes_date}.
#' @param icd10 (character) ICD-10 code(s) to match.  For \code{"prefix"} and
#'   \code{"exact"}, supply a vector such as \code{c("L20", "L21")}.  For
#'   \code{"regex"}, supply a single regex string.
#' @param match (character) Matching strategy: \code{"prefix"} (default)
#'   matches any code starting with the supplied string; \code{"exact"}
#'   requires a full match; \code{"regex"} uses \code{icd10} directly.
#' @param disease_cols (character or NULL) Name of the \code{p41270} column.
#'   \code{NULL} = auto-detect.
#' @param date_cols (character or NULL) Names of \code{p41280_a*} columns.
#'   \code{NULL} = auto-detect.
#'
#' @return The input \code{data} with two new columns added in-place:
#'   \code{{name}_hes} (logical) and \code{{name}_hes_date} (IDate).
#'   Always returns a \code{data.table}.
#' @export
#'
#' @examples
#' \donttest{
#' dt <- ops_toy(n = 100)
#' derive_hes(dt, name = "htn",      icd10 = "I10")
#' derive_hes(dt, name = "diabetes", icd10 = c("E10", "E11"))
#' derive_hes(dt, name = "asthma",   icd10 = "^J4", match = "regex")
#' }
derive_hes <- function(data,
                       name,
                       icd10,
                       match        = c("prefix", "exact", "regex"),
                       disease_cols = NULL,
                       date_cols    = NULL) {

  .assert_data_frame(data)
  .assert_scalar_string(name)
  if (!data.table::is.data.table(data)) data <- data.table::as.data.table(data)

  match      <- match.arg(match)
  status_col <- paste0(name, "_hes")
  date_col   <- paste0(name, "_hes_date")

  # ── Auto-detect columns ────────────────────────────────────────────────────
  # p41270: single JSON-array column (no _i/_a suffix on RAP)
  if (is.null(disease_cols)) disease_cols <- .detect_fo_col(data, 41270L)

  # p41280: wide date columns p41280_a0, p41280_a1, ...
  if (is.null(date_cols)) {
    date_cols <- .detect_cols_by_field(data, 41280L)
    # Fallback: RAP format has no _i, only _a
    if (length(date_cols) == 0L) {
      date_cols <- grep("p41280", names(data), value = TRUE, fixed = TRUE)
      date_cols <- setdiff(date_cols, "eid")
    }
  }

  if (is.null(disease_cols) || !disease_cols %in% names(data)) {
    cli::cli_alert_warning(
      "derive_hes: ICD-10 code column (p41270) not found. Supply disease_cols manually."
    )
    data[, (status_col) := FALSE]
    data[, (date_col)   := data.table::as.IDate(NA_character_)]
    return(invisible(data))
  }
  if (length(date_cols) == 0L) {
    cli::cli_alert_warning("derive_hes: no date columns (p41280_a*) found.")
  }

  # ── Build ICD-10 match patterns ───────────────────────────────────────────
  pattern <- switch(match,
    prefix = paste0("^(", paste(icd10, collapse = "|"), ")"),
    exact  = paste0("^(", paste(icd10, collapse = "|"), ")$"),
    regex  = icd10[1L]
  )
  # Reason: pre_pattern is anchor-free so it can match codes embedded inside
  # the raw JSON string (e.g. '"L209"' inside '["R55","L209",...]').
  # The anchored pattern is applied later on individual parsed codes.
  pre_pattern <- switch(match,
    prefix = paste(icd10, collapse = "|"),
    exact  = paste(icd10, collapse = "|"),
    regex  = icd10[1L]
  )

  # ── Pre-filter: coarse grepl on raw JSON string before strsplit ───────────
  # Reason: ~90%+ of participants have no matching code; scanning the raw
  # string is O(n_chars) and avoids the expensive strsplit + expand step for
  # the vast majority of rows.
  sub <- data[
    !is.na(get(disease_cols)) &
    nchar(as.character(get(disease_cols))) > 2L &
    grepl(pre_pattern, as.character(get(disease_cols)),
          ignore.case = TRUE, perl = TRUE),
    .(eid, raw = as.character(get(disease_cols)))
  ]

  if (nrow(sub) == 0L) {
    data[, (status_col) := FALSE]
    data[, (date_col)   := data.table::as.IDate(NA_character_)]
    cli::cli_alert_warning("derive_hes ({name}): 0 cases found.")
    return(invisible(data))
  }

  # ── Parse JSON array → long table (eid, array_idx, icd_code) ─────────────
  # Vectorized: strip brackets, split by "," separator, clean quotes
  sub[, raw_clean := gsub('^\\[|\\]$', '', raw)]
  codes_list <- strsplit(sub$raw_clean, '","')
  codes_list <- lapply(codes_list, function(x) gsub('"', '', x))
  codes_list <- lapply(codes_list, function(x) x[nchar(trimws(x)) > 0L])

  long_code <- data.table::data.table(
    eid       = rep(sub$eid, lengths(codes_list)),
    array_idx = unlist(lapply(codes_list, function(x) seq_along(x) - 1L)),
    icd_code  = unlist(codes_list)
  )

  # ── Exact filter on parsed codes (anchored pattern) ───────────────────────
  long_code <- long_code[
    grepl(pattern, icd_code, ignore.case = TRUE, perl = TRUE)
  ]

  if (nrow(long_code) == 0L) {
    data[, (status_col) := FALSE]
    data[, (date_col)   := data.table::as.IDate(NA_character_)]
    cli::cli_alert_warning("derive_hes ({name}): 0 cases found.")
    return(invisible(data))
  }

  # Status: any matching code
  data[, (status_col) := eid %in% unique(long_code$eid)]

  # ── Melt p41280 → long, extract array index, join on (eid, array_idx) ─────
  if (length(date_cols) > 0L) {
    matched_eids <- unique(long_code$eid)

    # Reason: only melt date columns for matched eids - avoids building a
    # 500k × 259 long table; na.rm = TRUE drops empty cells at C level.
    sub_t <- data.table::copy(
      data[eid %in% matched_eids, c("eid", date_cols), with = FALSE]
    )
    for (.c in date_cols) {
      data.table::set(sub_t, j = .c,
                      value = data.table::as.IDate(as.character(sub_t[[.c]])))
    }
    long_date <- data.table::melt(sub_t,
                                   id.vars       = "eid",
                                   measure.vars  = date_cols,
                                   variable.name = "col_name",
                                   value.name    = "diag_date",
                                   na.rm         = TRUE)
    # Extract 0-based array index from column name: p41280_a0 → 0
    long_date[, array_idx := as.integer(
      gsub(".*_a(\\d+)$", "\\1", as.character(col_name), perl = TRUE)
    )]
    long_date[, col_name := NULL]

    # Inner join: matched codes × dates by (eid, array_idx)
    matched <- long_date[long_code, on = c("eid", "array_idx"), nomatch = 0L]

    dates <- matched[!is.na(diag_date),
                     .(date = min(diag_date, na.rm = TRUE)),
                     by = eid]
    data[, (date_col) := data.table::as.IDate(NA_character_)]
    data[dates, on = "eid", (date_col) := i.date]
  } else {
    data[, (date_col) := data.table::as.IDate(NA_character_)]
  }

  n_cases <- sum(data[[status_col]])
  n_dated  <- sum(!is.na(data[[date_col]]))
  cli::cli_alert_success(
    "derive_hes ({name}): {n_cases} case{?s}, {n_dated} with date."
  )

  data
}


#' Derive a binary disease flag from UKB cancer registry
#'
#' The UK Biobank cancer registry links each participant's cancer diagnoses
#' from national cancer registries.  Each diagnosis is stored as a separate
#' instance with four parallel fields: ICD-10 code (\code{p40006}), histology
#' code (\code{p40011}), behaviour code (\code{p40012}), and diagnosis date
#' (\code{p40005}).  Unlike HES or self-report data, each instance holds
#' exactly one record - there is no array (\code{a*}) dimension.
#'
#' All three filter arguments (\code{icd10}, \code{histology},
#' \code{behaviour}) are applied with AND logic: a record must satisfy
#' every non-\code{NULL} filter to be counted.  For OR conditions (e.g.\
#' D04 \emph{or} C44 with specific histology), call the function twice and
#' combine the resulting columns downstream.
#'
#' \describe{
#'   \item{\code{{name}_cancer}}{Logical flag: \code{TRUE} if any cancer
#'     registry record satisfies all supplied filters.}
#'   \item{\code{{name}_cancer_date}}{Earliest matching diagnosis date
#'     (\code{IDate}).}
#' }
#'
#' @param data (data.frame or data.table) UKB phenotype data containing
#'   cancer registry fields.
#' @param name (character) Output column prefix, e.g. \code{"outcome_invasive"}
#'   produces \code{outcome_invasive_cancer} and
#'   \code{outcome_invasive_cancer_date}.
#' @param icd10 (character or NULL) Regular expression matched against the
#'   ICD-10 code column (\code{p40006}).  \code{NULL} = no ICD-10 filter.
#'   Examples: \code{"^C44"}, \code{"^(C44|D04)"}.
#' @param histology (integer vector or NULL) Histology codes to retain
#'   (\code{p40011}).  \code{NULL} = no histology filter.
#'   Example: \code{c(8070:8078, 8083, 8084)}.
#' @param behaviour (integer vector or NULL) Behaviour codes to retain
#'   (\code{p40012}).  \code{NULL} = no behaviour filter.
#'   Typical values: \code{3L} (invasive / malignant),
#'   \code{2L} (in situ).
#' @param code_cols (character or NULL) Names of ICD-10 code columns
#'   (\code{p40006_i*}).  \code{NULL} = auto-detect via field 40006.
#' @param hist_cols (character or NULL) Names of histology columns
#'   (\code{p40011_i*}).  \code{NULL} = auto-detect via field 40011.
#' @param behv_cols (character or NULL) Names of behaviour columns
#'   (\code{p40012_i*}).  \code{NULL} = auto-detect via field 40012.
#' @param date_cols (character or NULL) Names of diagnosis date columns
#'   (\code{p40005_i*}).  \code{NULL} = auto-detect via field 40005.
#'
#' @return The input \code{data} with two new columns added in-place:
#'   \code{{name}_cancer} (logical) and \code{{name}_cancer_date} (IDate).
#'   Always returns a \code{data.table}.
#' @export
#'
#' @examples
#' \donttest{
#' dt <- ops_toy(n = 100)
#' derive_cancer_registry(dt, name = "skin",     icd10 = "^C44")
#' derive_cancer_registry(dt, name = "invasive", icd10 = "^C", behaviour = 3L)
#' }
derive_cancer_registry <- function(data,
                                   name,
                                   icd10     = NULL,
                                   histology = NULL,
                                   behaviour = NULL,
                                   code_cols = NULL,
                                   hist_cols = NULL,
                                   behv_cols = NULL,
                                   date_cols = NULL) {

  .assert_data_frame(data)
  .assert_scalar_string(name)
  if (!is.null(icd10) && (!is.character(icd10) || length(icd10) != 1L)) {
    cli::cli_abort("{.arg icd10} must be a single regex string or NULL.", call = NULL)
  }
  if (!data.table::is.data.table(data)) data <- data.table::as.data.table(data)

  status_col <- paste0(name, "_cancer")
  date_col   <- paste0(name, "_cancer_date")

  # ── Auto-detect columns ────────────────────────────────────────────────────
  if (is.null(code_cols)) code_cols <- .detect_cols_by_field(data, 40006L)
  if (is.null(hist_cols)) hist_cols <- .detect_cols_by_field(data, 40011L)
  if (is.null(behv_cols)) behv_cols <- .detect_cols_by_field(data, 40012L)
  if (is.null(date_cols)) date_cols <- .detect_cols_by_field(data, 40005L)

  if (length(code_cols) == 0L) {
    cli::cli_alert_warning(
      "derive_cancer_registry: no ICD-10 code columns (p40006_i*) found. Supply code_cols manually."
    )
    data[, (status_col) := FALSE]
    data[, (date_col)   := data.table::as.IDate(NA_character_)]
    return(invisible(data))
  }

  # ── Extract instance indices ───────────────────────────────────────────────
  inst_raw  <- regmatches(code_cols,
                          regexpr("_i(\\d+)", code_cols, perl = TRUE))
  instances <- as.character(
    sort(as.integer(unique(sub("_i", "", inst_raw, fixed = TRUE))))
  )

  # ── Build long table: one row per (eid, instance) ─────────────────────────
  # Reason: cancer registry has no array (_a*) dimension - each instance is
  # a single record per field. rbindlist + lapply avoids melt entirely and
  # handles sparse / all-NA instances cleanly via fill = TRUE.
  long <- data.table::rbindlist(lapply(instances, function(inst) {

    c_col <- grep(paste0("_i", inst, "$"), code_cols, value = TRUE)
    if (length(c_col) != 1L) return(NULL)

    h_col <- grep(paste0("_i", inst, "$"), hist_cols, value = TRUE)
    b_col <- grep(paste0("_i", inst, "$"), behv_cols, value = TRUE)
    d_col <- grep(paste0("_i", inst, "$"), date_cols, value = TRUE)

    # Reason: pre-filter to rows with a non-empty ICD code before extracting
    # other columns - most instances are NA for most participants, so this
    # drops ~95% of rows at the source and keeps the long table small.
    data[
      !is.na(get(c_col)) & nchar(as.character(get(c_col))) > 0L,
      .(
        eid,
        icd_code  = as.character(get(c_col)),
        hist_code = if (length(h_col) == 1L)
                      suppressWarnings(as.integer(get(h_col)))
                    else
                      NA_integer_,
        behv_code = if (length(b_col) == 1L)
                      suppressWarnings(as.integer(get(b_col)))
                    else
                      NA_integer_,
        diag_date = if (length(d_col) == 1L)
                      data.table::as.IDate(as.character(get(d_col)))
                    else
                      data.table::as.IDate(NA_character_)
      )
    ]
  }), fill = TRUE)

  if (nrow(long) == 0L) {
    data[, (status_col) := FALSE]
    data[, (date_col)   := data.table::as.IDate(NA_character_)]
    cli::cli_alert_warning("derive_cancer_registry ({name}): 0 records found.")
    return(invisible(data))
  }

  # ── Multi-dimensional AND filter ───────────────────────────────────────────
  # Reason: column names (hist_code, behv_code) are distinct from parameter
  # names (histology, behaviour), so %in% comparisons are unambiguous.
  if (!is.null(icd10)) {
    long <- long[grepl(icd10, icd_code, ignore.case = TRUE, perl = TRUE)]
  }
  if (!is.null(histology)) {
    long <- long[!is.na(hist_code) & hist_code %in% histology]
  }
  if (!is.null(behaviour)) {
    long <- long[!is.na(behv_code) & behv_code %in% behaviour]
  }

  if (nrow(long) == 0L) {
    data[, (status_col) := FALSE]
    data[, (date_col)   := data.table::as.IDate(NA_character_)]
    cli::cli_alert_warning("derive_cancer_registry ({name}): 0 cases after filtering.")
    return(invisible(data))
  }

  # ── Write status and earliest date back to data ────────────────────────────
  data[, (status_col) := eid %in% long[, unique(eid)]]

  dates <- long[!is.na(diag_date),
                .(date = min(diag_date, na.rm = TRUE)),
                by = eid]
  data[, (date_col) := data.table::as.IDate(NA_character_)]
  data[dates, on = "eid", (date_col) := i.date]

  n_cases <- sum(data[[status_col]])
  n_dated  <- sum(!is.na(data[[date_col]]))
  cli::cli_alert_success(
    "derive_cancer_registry ({name}): {n_cases} case{?s}, {n_dated} with date."
  )

  data
}


#' Derive a binary disease flag from UKB death registry
#'
#' Death registry records store the underlying (primary) cause of death in
#' field \code{p40001} and contributory (secondary) causes in field
#' \code{p40002}, both coded in ICD-10.  The date of death is in field
#' \code{p40000}.  All three fields have an instance dimension (\code{i0},
#' \code{i1}) reflecting potential amendments; \code{p40002} additionally
#' has an array dimension (\code{a0}, \code{a1}, \ldots).
#'
#' \describe{
#'   \item{\code{{name}_death}}{Logical flag: \code{TRUE} if any death
#'     registry record (primary or secondary cause) contains a matching
#'     ICD-10 code.}
#'   \item{\code{{name}_death_date}}{Earliest death date across matching
#'     instances (\code{IDate}).  Note: this is the \emph{date of death},
#'     not onset date.}
#' }
#'
#' @param data (data.frame or data.table) UKB phenotype data containing
#'   death registry fields.
#' @param name (character) Output column prefix, e.g. \code{"disease"} produces
#'   \code{disease_death} and \code{disease_death_date}.
#' @param icd10 (character) ICD-10 code(s) to match.  For \code{"prefix"}
#'   and \code{"exact"}, supply a vector such as \code{c("L20", "L21")}.
#'   For \code{"regex"}, supply a single regex string.
#' @param match (character) Matching strategy: \code{"prefix"} (default),
#'   \code{"exact"}, or \code{"regex"}.
#' @param primary_cols (character or NULL) Names of primary cause columns
#'   (\code{p40001_i*}).  \code{NULL} = auto-detect via field 40001.
#' @param secondary_cols (character or NULL) Names of secondary cause
#'   columns (\code{p40002_i*_a*}).  \code{NULL} = auto-detect via field
#'   40002.
#' @param date_cols (character or NULL) Names of death date columns
#'   (\code{p40000_i*}).  \code{NULL} = auto-detect via field 40000.
#'
#' @return The input \code{data} with two new columns added in-place:
#'   \code{{name}_death} (logical) and \code{{name}_death_date} (IDate).
#'   Always returns a \code{data.table}.
#' @export
#'
#' @examples
#' \donttest{
#' dt <- ops_toy(n = 100)
#' derive_death_registry(dt, name = "mi",   icd10 = "I21")
#' derive_death_registry(dt, name = "lung", icd10 = "C34")
#' }
derive_death_registry <- function(data,
                                  name,
                                  icd10,
                                  match          = c("prefix", "exact", "regex"),
                                  primary_cols   = NULL,
                                  secondary_cols = NULL,
                                  date_cols      = NULL) {

  .assert_data_frame(data)
  .assert_scalar_string(name)
  if (!data.table::is.data.table(data)) data <- data.table::as.data.table(data)

  match      <- match.arg(match)
  status_col <- paste0(name, "_death")
  date_col   <- paste0(name, "_death_date")

  # ── Auto-detect columns ────────────────────────────────────────────────────
  if (is.null(primary_cols))   primary_cols   <- .detect_cols_by_field(data, 40001L)
  if (is.null(secondary_cols)) secondary_cols <- .detect_cols_by_field(data, 40002L)
  if (is.null(date_cols))      date_cols      <- .detect_cols_by_field(data, 40000L)

  if (length(primary_cols) == 0L && length(secondary_cols) == 0L) {
    cli::cli_alert_warning(
      "derive_death_registry: no death cause columns found. Supply primary_cols / secondary_cols manually."
    )
    data[, (status_col) := FALSE]
    data[, (date_col)   := data.table::as.IDate(NA_character_)]
    return(invisible(data))
  }

  # ── Build ICD-10 match pattern ─────────────────────────────────────────────
  pattern <- switch(match,
    prefix = paste0("^(", paste(icd10, collapse = "|"), ")"),
    exact  = paste0("^(", paste(icd10, collapse = "|"), ")$"),
    regex  = icd10[1L]
  )

  # ── Detect instances from primary cause columns ────────────────────────────
  inst_src  <- if (length(primary_cols) > 0L) primary_cols else secondary_cols
  inst_raw  <- regmatches(inst_src, regexpr("_i(\\d+)", inst_src, perl = TRUE))
  instances <- sort(unique(sub("_i", "", inst_raw, fixed = TRUE)))

  results <- vector("list", length(instances))
  names(results) <- instances

  for (inst in instances) {

    matched_eids <- integer(0L)

    # ── ① Primary cause: single column per instance, direct grepl ─────────
    p_col <- grep(paste0("_i", inst, "$"), primary_cols, value = TRUE)
    if (length(p_col) == 1L) {
      # Reason: pure data.table [i, j] avoids copying the column to a
      # separate vector - zero extra allocation.
      hits <- data[
        !is.na(get(p_col)) &
        grepl(pattern, as.character(get(p_col)), ignore.case = TRUE, perl = TRUE),
        eid
      ]
      matched_eids <- union(matched_eids, hits)
    }

    # ── ② Secondary causes: melt instance array columns, then grepl ───────
    s_cols <- grep(paste0("_i", inst, "_a"), secondary_cols, value = TRUE)
    if (length(s_cols) > 0L) {
      sub_s <- data.table::copy(data[, c("eid", s_cols), with = FALSE])
      for (.c in s_cols) {
        data.table::set(sub_s, j = .c, value = as.character(sub_s[[.c]]))
      }
      long_s <- data.table::melt(sub_s,
                                  id.vars       = "eid",
                                  measure.vars  = s_cols,
                                  variable.name = "array_idx",
                                  value.name    = "icd_code",
                                  na.rm         = TRUE)
      long_s <- long_s[
        nchar(as.character(icd_code)) > 0L &
        grepl(pattern, icd_code, ignore.case = TRUE, perl = TRUE)
      ]
      matched_eids <- union(matched_eids, unique(long_s$eid))
    }

    if (length(matched_eids) == 0L) next

    # ── ③ Death date: p40000_iX (instance level) ──────────────────────────
    d_col <- grep(paste0("_i", inst, "$"), date_cols, value = TRUE)
    if (length(d_col) == 1L) {
      # Reason: filter to matched eids first, then parse dates - avoids
      # converting the full 500k-row date column for a handful of matches.
      death_dates <- data[
        eid %in% matched_eids,
        .(eid, death_date = data.table::as.IDate(as.character(get(d_col))))
      ]
    } else {
      death_dates <- data.table::data.table(
        eid        = matched_eids,
        death_date = data.table::as.IDate(NA_character_)
      )
    }

    results[[inst]] <- death_dates
  }

  # ── Combine across instances ───────────────────────────────────────────────
  results <- Filter(Negate(is.null), results)

  if (length(results) == 0L) {
    data[, (status_col) := FALSE]
    data[, (date_col)   := data.table::as.IDate(NA_character_)]
    cli::cli_alert_warning("derive_death_registry ({name}): 0 cases found.")
    return(invisible(data))
  }

  combined <- data.table::rbindlist(results, fill = TRUE)

  # Status: any matching death cause record
  data[, (status_col) := eid %in% unique(combined$eid)]

  # Earliest death date per eid - update-on-join (zero-copy)
  dates <- combined[!is.na(death_date),
                    .(date = min(death_date, na.rm = TRUE)),
                    by = eid]
  data[, (date_col) := data.table::as.IDate(NA_character_)]
  data[dates, on = "eid", (date_col) := i.date]

  n_cases <- sum(data[[status_col]])
  n_dated  <- sum(!is.na(data[[date_col]]))
  cli::cli_alert_success(
    "derive_death_registry ({name}): {n_cases} case{?s}, {n_dated} with date."
  )

  data
}


#' Derive a unified ICD-10 disease flag across multiple UKB data sources
#'
#' A high-level wrapper that calls one or more of \code{\link{derive_hes}},
#' \code{\link{derive_death_registry}}, \code{\link{derive_first_occurrence}},
#' and \code{\link{derive_cancer_registry}} according to the \code{source}
#' argument, then combines their results into a single status flag and
#' earliest-date column.
#'
#' All intermediate source columns (\code{{name}_hes}, \code{{name}_death},
#' \code{{name}_fo}, \code{{name}_cancer} and their \code{_date} counterparts)
#' are retained in \code{data} so that per-source contributions remain
#' traceable.
#'
#' \describe{
#'   \item{\code{{name}_icd10}}{Logical flag: \code{TRUE} if any selected
#'     source contains a matching record.}
#'   \item{\code{{name}_icd10_date}}{Earliest matching date across all
#'     selected sources (\code{IDate}).}
#' }
#'
#' @param data (data.frame or data.table) UKB phenotype data.
#' @param name (character) Output column prefix, e.g. \code{"disease"} produces
#'   \code{disease_icd10} and \code{disease_icd10_date}, plus intermediate columns
#'   such as \code{disease_hes}, \code{disease_hes_date}, etc.
#' @param icd10 (character) ICD-10 code(s) to match.  For \code{"prefix"}
#'   and \code{"exact"}, supply a vector such as \code{c("L20", "L21")}.
#'   For \code{"regex"}, supply a single regex string.  When
#'   \code{"cancer_registry"} is included in \code{source}, \code{icd10}
#'   and \code{match} are automatically converted to a regex and passed to
#'   \code{\link{derive_cancer_registry}}.
#' @param source (character) One or more of \code{"hes"},
#'   \code{"death"}, \code{"first_occurrence"}, \code{"cancer_registry"}.
#'   Defaults to all four.
#' @param match (character) Matching strategy passed to \code{derive_hes}
#'   and \code{derive_death_registry}: \code{"prefix"} (default),
#'   \code{"exact"}, or \code{"regex"}.
#' @param fo_field (integer or character or NULL) UKB field ID for the
#'   First Occurrence column (e.g. \code{131666L} for E11).  Required when
#'   \code{"first_occurrence"} is in \code{source} and \code{fo_col} is
#'   \code{NULL}.
#' @param fo_col (character or NULL) Column name of the First Occurrence
#'   field in \code{data}.  Alternative to \code{fo_field}.
#' @param histology (integer vector or NULL) Passed to
#'   \code{\link{derive_cancer_registry}}.  Ignored for other sources.
#' @param behaviour (integer vector or NULL) Passed to
#'   \code{\link{derive_cancer_registry}}.  Ignored for other sources.
#' @param hes_code_col (character or NULL) Passed as \code{disease_cols}
#'   to \code{\link{derive_hes}}.
#' @param hes_date_cols (character or NULL) Passed as \code{date_cols}
#'   to \code{\link{derive_hes}}.
#' @param primary_cols (character or NULL) Passed to
#'   \code{\link{derive_death_registry}}.
#' @param secondary_cols (character or NULL) Passed to
#'   \code{\link{derive_death_registry}}.
#' @param death_date_cols (character or NULL) Passed as \code{date_cols}
#'   to \code{\link{derive_death_registry}}.
#' @param cr_code_cols (character or NULL) Passed as \code{code_cols}
#'   to \code{\link{derive_cancer_registry}}.
#' @param cr_hist_cols (character or NULL) Passed as \code{hist_cols}
#'   to \code{\link{derive_cancer_registry}}.
#' @param cr_behv_cols (character or NULL) Passed as \code{behv_cols}
#'   to \code{\link{derive_cancer_registry}}.
#' @param cr_date_cols (character or NULL) Passed as \code{date_cols}
#'   to \code{\link{derive_cancer_registry}}.
#'
#' @return The input \code{data} with \code{{name}_icd10} (logical) and
#'   \code{{name}_icd10_date} (IDate) added in-place, plus all intermediate
#'   source columns.  Always returns a \code{data.table}.
#' @export
#'
#' @examples
#' \donttest{
#' dt <- ops_toy(n = 100)
#' derive_icd10(dt, name = "htn",
#'              icd10  = "I10",
#'              source = c("hes", "death"))
#' derive_icd10(dt, name = "mi",
#'              icd10  = "I21",
#'              source = c("hes", "death", "first_occurrence"),
#'              fo_col = "p131742")
#' }
derive_icd10 <- function(data,
                         name,
                         icd10,
                         source          = c("hes", "death",
                                             "first_occurrence",
                                             "cancer_registry"),
                         match           = c("prefix", "exact", "regex"),
                         fo_field        = NULL,
                         fo_col          = NULL,
                         histology       = NULL,
                         behaviour       = NULL,
                         hes_code_col    = NULL,
                         hes_date_cols   = NULL,
                         primary_cols    = NULL,
                         secondary_cols  = NULL,
                         death_date_cols = NULL,
                         cr_code_cols    = NULL,
                         cr_hist_cols    = NULL,
                         cr_behv_cols    = NULL,
                         cr_date_cols    = NULL) {

  .assert_data_frame(data)
  .assert_scalar_string(name)
  if (!data.table::is.data.table(data)) data <- data.table::as.data.table(data)

  source     <- match.arg(source, several.ok = TRUE)
  match      <- match.arg(match)
  status_col <- paste0(name, "_icd10")
  date_col   <- paste0(name, "_icd10_date")

  # ── HES ───────────────────────────────────────────────────────────────────
  if ("hes" %in% source) {
    data <- derive_hes(data, name = name, icd10 = icd10, match = match,
                       disease_cols = hes_code_col,
                       date_cols    = hes_date_cols)
  }

  # ── Death registry ─────────────────────────────────────────────────────────
  if ("death" %in% source) {
    data <- derive_death_registry(data, name = name, icd10 = icd10,
                                  match          = match,
                                  primary_cols   = primary_cols,
                                  secondary_cols = secondary_cols,
                                  date_cols      = death_date_cols)
  }

  # ── First Occurrence ───────────────────────────────────────────────────────
  if ("first_occurrence" %in% source) {
    if (is.null(fo_field) && is.null(fo_col)) {
      cli::cli_alert_warning(
        "derive_icd10: 'first_occurrence' selected but fo_field and fo_col are both NULL - skipping this source."
      )
    } else {
      data <- derive_first_occurrence(data, name = name,
                                      field = fo_field, col = fo_col)
    }
  }

  # ── Cancer registry ────────────────────────────────────────────────────────
  if ("cancer_registry" %in% source) {
    # Reason: derive_cancer_registry takes a single regex string, not a vector
    # + match mode. Convert here so the caller only needs one set of params.
    icd10_regex <- switch(match,
      prefix = paste0("^(", paste(icd10, collapse = "|"), ")"),
      exact  = paste0("^(", paste(icd10, collapse = "|"), ")$"),
      regex  = icd10[1L]
    )
    data <- derive_cancer_registry(data, name = name, icd10 = icd10_regex,
                                   histology = histology,
                                   behaviour = behaviour,
                                   code_cols = cr_code_cols,
                                   hist_cols = cr_hist_cols,
                                   behv_cols = cr_behv_cols,
                                   date_cols = cr_date_cols)
  }

  # ── Collect active intermediate columns ────────────────────────────────────
  fo_ran <- "first_occurrence" %in% source &&
            (!is.null(fo_field) || !is.null(fo_col))

  active_status <- intersect(
    c(if ("hes"              %in% source) paste0(name, "_hes"),
      if ("death"            %in% source) paste0(name, "_death"),
      if (fo_ran)                         paste0(name, "_fo"),
      if ("cancer_registry"  %in% source) paste0(name, "_cancer")),
    names(data)
  )
  active_dates <- intersect(
    c(if ("hes"              %in% source) paste0(name, "_hes_date"),
      if ("death"            %in% source) paste0(name, "_death_date"),
      if (fo_ran)                         paste0(name, "_fo_date"),
      if ("cancer_registry"  %in% source) paste0(name, "_cancer_date")),
    names(data)
  )

  # ── Combine: OR for status, pmin for date ──────────────────────────────────
  data[, (status_col) := as.logical(Reduce(`|`, lapply(active_status, function(col) data[[col]])))]

  if (length(active_dates) > 0L) {
    date_list <- lapply(active_dates, function(col) data[[col]])
    data[, (date_col) := data.table::as.IDate(
      do.call(pmin, c(date_list, list(na.rm = TRUE)))
    )]
  } else {
    data[, (date_col) := data.table::as.IDate(NA_character_)]
  }

  n_cases  <- sum(data[[status_col]], na.rm = TRUE)
  n_dated  <- sum(!is.na(data[[date_col]]))
  n_src    <- length(active_status)
  cli::cli_alert_success(
    "derive_icd10 ({name}): {n_cases} case{?s} across {n_src} source{?s}, {n_dated} with date."
  )

  data
}


#' Combine self-report and ICD-10 sources into a unified case definition
#'
#' Takes the self-report flag and ICD-10 flag produced by
#' \code{\link{derive_selfreport}} and \code{\link{derive_icd10}} (or any
#' pair of logical columns) and merges them into a single logical case status
#' and earliest date.
#'
#' \describe{
#'   \item{\code{{name}_status}}{Logical: \code{TRUE} if positive in at
#'     least one source.}
#'   \item{\code{{name}_date}}{Earliest diagnosis/report date across both
#'     sources (\code{IDate}).}
#' }
#'
#' @param data (data.frame or data.table) UKB phenotype data.
#' @param name (character) Column prefix used both to locate the default
#'   input columns and to name the output columns.  Defaults:
#'   \code{{name}_icd10}, \code{{name}_selfreport},
#'   \code{{name}_icd10_date}, \code{{name}_selfreport_date}.
#' @param icd10_col (character or NULL) Name of the ICD-10 status column.
#'   \code{NULL} = \code{paste0(name, "_icd10")}.
#' @param selfreport_col (character or NULL) Name of the self-report status
#'   column.  \code{NULL} = \code{paste0(name, "_selfreport")}.
#' @param icd10_date_col (character or NULL) Name of the ICD-10 date column.
#'   \code{NULL} = \code{paste0(name, "_icd10_date")}.
#' @param selfreport_date_col (character or NULL) Name of the self-report date
#'   column.  \code{NULL} = \code{paste0(name, "_selfreport_date")}.
#'
#' @return The input \code{data} with two new columns added in-place:
#'   \code{{name}_status} (logical) and \code{{name}_date} (IDate).
#'   Always returns a \code{data.table}.
#' @export
#'
#' @examples
#' \donttest{
#' dt <- ops_toy(n = 100)
#' derive_selfreport(dt, name = "htn", regex = "hypertension",
#'                   field        = "noncancer",
#'                   disease_cols = paste0("p20002_i0_a", 0:4),
#'                   date_cols    = paste0("p20008_i0_a", 0:4),
#'                   visit_cols   = "p53_i0")
#' derive_icd10(dt, name = "htn", icd10 = "I10", source = c("hes", "death"))
#' derive_case(dt, name = "htn")
#' }
derive_case <- function(data,
                        name,
                        icd10_col           = NULL,
                        selfreport_col      = NULL,
                        icd10_date_col      = NULL,
                        selfreport_date_col = NULL) {

  .assert_data_frame(data)
  .assert_scalar_string(name)
  if (!data.table::is.data.table(data)) data <- data.table::as.data.table(data)

  # ── Resolve column names ───────────────────────────────────────────────────
  icd10_col           <- icd10_col           %||% paste0(name, "_icd10")
  selfreport_col      <- selfreport_col      %||% paste0(name, "_selfreport")
  icd10_date_col      <- icd10_date_col      %||% paste0(name, "_icd10_date")
  selfreport_date_col <- selfreport_date_col %||% paste0(name, "_selfreport_date")

  status_col <- paste0(name, "_status")
  date_col   <- paste0(name, "_date")

  # ── Validate presence ──────────────────────────────────────────────────────
  status_cols_present <- intersect(c(icd10_col, selfreport_col), names(data))
  if (length(status_cols_present) == 0L) {
    cli::cli_abort(
      "derive_case: neither {.val {icd10_col}} nor {.val {selfreport_col}} found in data. Supply column names explicitly.",
      call = NULL
    )
  }
  if (length(status_cols_present) < 2L) {
    cli::cli_alert_warning(
      "derive_case ({name}): only one status column found ({status_cols_present[1L]}), combining with itself."
    )
  }

  date_cols_present <- intersect(c(icd10_date_col, selfreport_date_col), names(data))

  # ── Status: OR across available status columns → logical ──────────────────
  data[, (status_col) := as.logical(Reduce(`|`, lapply(status_cols_present, function(col) data[[col]])))]

  # ── Date: earliest across available date columns ───────────────────────────
  if (length(date_cols_present) > 0L) {
    date_list <- lapply(date_cols_present, function(col) data[[col]])
    data[, (date_col) := data.table::as.IDate(
      do.call(pmin, c(date_list, list(na.rm = TRUE)))
    )]
  } else {
    data[, (date_col) := data.table::as.IDate(NA_character_)]
  }

  # ── CLI summary ────────────────────────────────────────────────────────────
  n_cases <- sum(data[[status_col]], na.rm = TRUE)
  n_dated  <- sum(!is.na(data[[date_col]]))
  cli::cli_alert_success(
    "derive_case ({name}): {n_cases} case{?s}, {n_dated} with date."
  )
  if (length(status_cols_present) == 2L) {
    n_both <- sum(data[[status_cols_present[1L]]] & data[[status_cols_present[2L]]],
                  na.rm = TRUE)
    cli::cli_alert_info(
      "  Both sources ({icd10_col} & {selfreport_col}): {n_both}"
    )
  }

  data
}


#' Compute age at event for one or more UKB outcomes
#'
#' For each name in \code{name}, adds one column \code{age_at_{name}} (numeric,
#' years) computed as:
#' \deqn{age\_at\_event = age\_col + (event\_date - baseline\_date) / 365.25}
#'
#' The value is \code{NA} for participants who did not experience the event
#' (status is \code{FALSE} / \code{0}) or who lack an event date.
#'
#' \strong{Auto-detection per name} (when \code{date_cols} / \code{status_cols}
#' are \code{NULL}):
#' \itemize{
#'   \item \code{date_col}   - looked up as \code{{name}_date}.
#'   \item \code{status_col} - looked up first as \code{{name}_status}, then
#'     as \code{{name}} (logical column); if neither exists all rows with a
#'     non-\code{NA} date are treated as cases.
#' }
#'
#' \strong{data.table pass-by-reference}: new columns are added in-place.
#'
#' @param data (data.frame or data.table) UKB phenotype data.
#' @param name (character) One or more output prefixes, e.g.
#'   \code{c("disease", "disease_icd10", "outcome")}. Each produces \code{age_at_{name}}.
#' @param baseline_col (character) Name of the baseline date column
#'   (e.g. \code{"date_baseline"}).
#' @param age_col (character) Name of the age-at-baseline column
#'   (e.g. \code{"age_recruitment"}).
#' @param date_cols (character or NULL) Named character vector mapping each
#'   name to its event date column, e.g.
#'   \code{c(disease = "disease_date", outcome = "outcome_date")}. \code{NULL} (default)
#'   triggers auto-detection as \code{{name}_date}.
#' @param status_cols (character or NULL) Named character vector mapping each
#'   name to its status column. \code{NULL} (default) triggers auto-detection.
#'
#' @return The input \code{data} with one new \code{age_at_{name}} column
#'   per entry in \code{name}, added in-place.
#' @export
#'
#' @examples
#' dt <- ops_toy(scenario = "association", n = 100)
#' derive_age(dt, name = "dm", baseline_col = "p53_i0", age_col = "p21022")
derive_age <- function(data,
                       name,
                       baseline_col,
                       age_col,
                       date_cols   = NULL,
                       status_cols = NULL) {

  .assert_data_frame(data)
  if (!is.character(name) || length(name) == 0L) {
    cli::cli_abort("{.arg name} must be a non-empty character vector.", call = NULL)
  }
  if (!data.table::is.data.table(data)) data <- data.table::as.data.table(data)

  .assert_has_cols(data, c(baseline_col, age_col))

  # ── Process each name ──────────────────────────────────────────────────────
  for (nm in name) {

    out_col <- paste0("age_at_", nm)

    # ── Resolve date column ──────────────────────────────────────────────────
    date_col <- if (!is.null(date_cols) && nm %in% names(date_cols)) date_cols[[nm]] else NULL
    date_col <- date_col %||% paste0(nm, "_date")

    if (!date_col %in% names(data)) {
      cli::cli_alert_warning("derive_age: date column '{date_col}' not found - skipping {nm}.")
      next
    }

    # ── Resolve status column (optional guard) ───────────────────────────────
    # Reason: status may be integer (0/1) or logical; try {name}_status first,
    # then bare {name}, then fall back to NA-date guard only.
    status_col <- if (!is.null(status_cols) && nm %in% names(status_cols)) status_cols[[nm]] else NULL
    if (is.null(status_col)) {
      if (paste0(nm, "_status") %in% names(data)) {
        status_col <- paste0(nm, "_status")
      } else if (nm %in% names(data) && is.logical(data[[nm]])) {
        status_col <- nm
      }
    }

    # ── Compute age at event in-place ────────────────────────────────────────
    # Condition: participant is a case AND has a recorded event date.
    # When no status column exists, compute for all rows with non-NA date.
    if (!is.null(status_col)) {
      s <- data[[status_col]]
      is_case <- if (is.logical(s)) s else s == 1L
      data[, (out_col) := data.table::fifelse(
        is_case & !is.na(get(date_col)),
        get(age_col) + as.numeric(
          data.table::as.IDate(get(date_col)) -
          data.table::as.IDate(get(baseline_col))
        ) / 365.25,
        NA_real_
      )]
    } else {
      data[, (out_col) := data.table::fifelse(
        !is.na(get(date_col)),
        get(age_col) + as.numeric(
          data.table::as.IDate(get(date_col)) -
          data.table::as.IDate(get(baseline_col))
        ) / 365.25,
        NA_real_
      )]
    }

    # ── CLI summary ──────────────────────────────────────────────────────────
    v <- data[[out_col]]
    n_cases <- sum(!is.na(v))
    if (n_cases > 0L) {
      cli::cli_alert_info(
        "  {out_col}: n={n_cases}, median={round(median(v, na.rm=TRUE), 1)}, range=[{round(min(v, na.rm=TRUE), 1)}, {round(max(v, na.rm=TRUE), 1)}]"
      )
    } else {
      cli::cli_alert_info("  {out_col}: n=0, all NA.")
    }
  }

  cli::cli_alert_success("derive_age: {length(name)} event{?s} processed.")
  data
}


#' Compute follow-up end date and follow-up time for survival analysis
#'
#' Adds two columns to \code{data}:
#' \itemize{
#'   \item \code{{name}_followup_end} (IDate) - the earliest of the outcome
#'     event date, death date, lost-to-follow-up date, and the administrative
#'     censoring date.
#'   \item \code{{name}_followup_years} (numeric) - time in years from
#'     \code{baseline_col} to \code{{name}_followup_end}.
#' }
#'
#' \strong{data.table pass-by-reference}: when the input is a
#' \code{data.table}, new columns are added in-place via \code{:=}.
#'
#' @param data (data.frame or data.table) UKB phenotype data.
#' @param name (character) Output column prefix, e.g. \code{"outcome"} produces
#'   \code{outcome_followup_end} and \code{outcome_followup_years}.
#' @param event_col (character) Name of the outcome event date column
#'   (e.g. \code{"outcome_date"}).
#' @param baseline_col (character) Name of the baseline date column
#'   (e.g. \code{"date_baseline"}).
#' @param censor_date (Date or character) Scalar administrative censoring date,
#'   e.g. \code{as.Date("2022-06-01")}.  A character string in
#'   \code{"YYYY-MM-DD"} format is also accepted.
#' @param death_col (character or NULL) Name of the death date column
#'   (UKB field 40000).  \code{NULL} (default) triggers auto-detection via
#'   the \code{\link{extract_ls}} cache; pass \code{FALSE} to explicitly
#'   disable death as a competing end-point.
#' @param lost_col (character or NULL) Name of the lost-to-follow-up date
#'   column (UKB field 191).  \code{NULL} (default) triggers auto-detection;
#'   pass \code{FALSE} to explicitly disable.
#'
#' @return The input \code{data} with two new columns added in-place:
#'   \code{{name}_followup_end} (IDate) and \code{{name}_followup_years}
#'   (numeric).
#' @export
#'
#' @examples
#' \donttest{
#' dt <- ops_toy(n = 100)
#' derive_hes(dt, name = "htn", icd10 = "I10")
#' derive_followup(dt,
#'   name         = "htn_hes",
#'   event_col    = "htn_hes_date",
#'   baseline_col = "p53_i0",
#'   censor_date  = as.Date("2022-06-01"),
#'   death_col    = "p40000_i0",
#'   lost_col     = FALSE)
#' }
derive_followup <- function(data,
                            name,
                            event_col,
                            baseline_col,
                            censor_date,
                            death_col = NULL,
                            lost_col  = NULL) {

  .assert_data_frame(data)
  .assert_scalar_string(name)
  if (!data.table::is.data.table(data)) data <- data.table::as.data.table(data)

  # Reason: FALSE is a documented sentinel meaning "disable this endpoint".
  # Convert to NULL early so all downstream NULL-checks work correctly.
  if (isFALSE(death_col)) death_col <- NULL
  if (isFALSE(lost_col))  lost_col  <- NULL

  # ── Auto-detect death / lost-to-follow-up columns from extract_ls() cache ──
  # Reason: UKB field 40000 = date of death, field 191 = date lost to follow-up.
  # When the user does not supply these, try to locate them automatically so
  # the function works out-of-the-box after extract_pheno() + decode_names().
  if (is.null(death_col)) death_col <- .detect_fo_col(data, 40000L)
  if (is.null(lost_col))  lost_col  <- .detect_fo_col(data, 191L)

  # ── Validate required columns ──────────────────────────────────────────────
  .assert_has_cols(data, c(event_col, baseline_col))
  if (!is.null(death_col)) .assert_has_cols(data, death_col)
  if (!is.null(lost_col))  .assert_has_cols(data, lost_col)

  censor_date <- data.table::as.IDate(censor_date)

  # ── Output column names ────────────────────────────────────────────────────
  end_col   <- paste0(name, "_followup_end")
  years_col <- paste0(name, "_followup_years")

  # ── Collect date columns for pmin ─────────────────────────────────────────
  # Reason: using .SDcols keeps column references inside data.table's j-scope;
  # valid_date_cols collects only columns that are actually present.
  valid_date_cols <- event_col
  if (!is.null(death_col)) valid_date_cols <- c(valid_date_cols, death_col)
  if (!is.null(lost_col))  valid_date_cols <- c(valid_date_cols, lost_col)

  # Reason: coerce all date columns to IDate before pmin. Real RAP data and
  # ops_toy both store date columns as character strings; mixing character with
  # IDate in pmin causes IDate to be cast to its integer representation
  # (e.g. "18628"), which charToDate() cannot parse.
  for (col in valid_date_cols) {
    if (!inherits(data[[col]], "IDate")) {
      data[, (col) := data.table::as.IDate(as.character(get(col)))]
    }
  }

  # ── Compute follow-up end (earliest competing date) ────────────────────────
  data[, (end_col) := data.table::as.IDate(
    do.call(pmin, c(.SD, list(censor_date), list(na.rm = TRUE)))
  ), .SDcols = valid_date_cols]

  # ── Compute follow-up time in years ───────────────────────────────────────
  data[, (years_col) := as.numeric(get(end_col) - data.table::as.IDate(get(baseline_col))) / 365.25]

  # Reason: prevalent cases have event_date < baseline, yielding negative follow-up
  # time. Set to NA — negative survival time is not meaningful for Cox/logistic models.
  data[!is.na(get(years_col)) & get(years_col) <= 0, (years_col) := NA_real_]

  # ── CLI summary ────────────────────────────────────────────────────────────
  yrs <- data[[years_col]]
  cli::cli_alert_success("derive_followup ({name}):")
  cli::cli_alert_info("  {end_col}: {sum(!is.na(data[[end_col]]))} / {nrow(data)} non-missing")
  cli::cli_alert_info("  {years_col}: mean={round(mean(yrs, na.rm=TRUE), 2)}, median={round(median(yrs, na.rm=TRUE), 2)}, range=[{round(min(yrs, na.rm=TRUE), 2)}, {round(max(yrs, na.rm=TRUE), 2)}]")

  data
}


#' Classify disease timing relative to UKB baseline assessment
#'
#' Assigns each participant an integer timing category based on whether their
#' disease date falls before or after the baseline visit date:
#'
#' \describe{
#'   \item{\code{0}}{No disease (\code{status_col} is \code{FALSE}).}
#'   \item{\code{1}}{Prevalent - disease date on or before baseline.}
#'   \item{\code{2}}{Incident - disease date strictly after baseline.}
#'   \item{\code{NA}}{Case with missing date; timing cannot be determined.}
#' }
#'
#' Call once per timing variable needed (e.g. once for the combined case,
#' once per individual source).
#'
#' @param data (data.frame or data.table) UKB phenotype data.
#' @param name (character) Output column prefix.  The new column is named
#'   \code{{name}_timing}.  Also used to derive default \code{status_col}
#'   and \code{date_col} when those are \code{NULL}.
#' @param baseline_col (character) Name of the baseline date column in
#'   \code{data} (e.g. \code{"date_baseline"} or \code{"p53_i0"}).
#' @param status_col (character or NULL) Name of the logical disease flag.
#'   \code{NULL} = \code{paste0(name, "_status")}.
#' @param date_col (character or NULL) Name of the disease date column
#'   (\code{IDate} or \code{Date}).
#'   \code{NULL} = \code{paste0(name, "_date")}.
#'
#' @return The input \code{data} with one new integer column
#'   \code{{name}_timing} (0/1/2/\code{NA}) added in-place.
#'   Always returns a \code{data.table}.
#' @export
#'
#' @examples
#' \donttest{
#' dt <- ops_toy(n = 100)
#' derive_hes(dt, name = "htn", icd10 = "I10")
#' derive_timing(dt, name = "htn_hes",
#'               status_col   = "htn_hes",
#'               date_col     = "htn_hes_date",
#'               baseline_col = "p53_i0")
#' }
derive_timing <- function(data,
                          name,
                          baseline_col,
                          status_col = NULL,
                          date_col   = NULL) {

  .assert_data_frame(data)
  .assert_scalar_string(name)
  if (!data.table::is.data.table(data)) data <- data.table::as.data.table(data)

  # ── Resolve column names ───────────────────────────────────────────────────
  status_col  <- status_col  %||% paste0(name, "_status")
  date_col    <- date_col    %||% paste0(name, "_date")
  timing_col  <- paste0(name, "_timing")

  .assert_has_cols(data, c(status_col, date_col, baseline_col))

  # ── Classify timing with fcase (data.table's vectorised case_when) ─────────
  # Reason: fcase evaluates conditions sequentially and short-circuits,
  # making it faster than nested ifelse and cleaner than case_when.
  data[, (timing_col) := data.table::fcase(
    !get(status_col),                                                          0L,
    get(status_col) & !is.na(get(date_col)) &
      get(date_col) <= data.table::as.IDate(get(baseline_col)),               1L,
    get(status_col) & !is.na(get(date_col)) &
      get(date_col) >  data.table::as.IDate(get(baseline_col)),               2L,
    default = NA_integer_
  )]

  # ── CLI summary ────────────────────────────────────────────────────────────
  tv <- data[[timing_col]]
  cli::cli_alert_success("derive_timing ({name}_timing):")
  cli::cli_alert_info("  0 (no disease): {sum(tv == 0L, na.rm = TRUE)}")
  cli::cli_alert_info("  1 (prevalent):  {sum(tv == 1L, na.rm = TRUE)}")
  cli::cli_alert_info("  2 (incident):   {sum(tv == 2L, na.rm = TRUE)}")
  cli::cli_alert_info("  NA (no date):   {sum(is.na(tv))}")

  data
}

# =============================================================================
# test-derive.R — Unit tests for derive_ series
# Primary fixture: ops_toy(n = 300, seed = 42) — UKB-like structure, no network
# Hand-crafted helpers only where exact row-level values must be asserted
# =============================================================================

# File-level fixture (created once, copied per test)
DT <- suppressMessages(ops_toy(n = 300L, seed = 42L))

# ── Hand-crafted helpers (exact-value tests) ──────────────────────────────────

.timing_dt <- function() {
  data.table::data.table(
    eid           = 1:5,
    t2d_status    = c(FALSE, TRUE,  TRUE,  TRUE,  FALSE),
    t2d_date      = data.table::as.IDate(
      c(NA, "2010-01-01", "2018-06-01", NA, NA)),
    date_baseline = data.table::as.IDate("2015-01-01")
  )
}

.followup_dt <- function() {
  data.table::data.table(
    eid           = 1:4,
    t2d_date      = data.table::as.IDate(c("2018-01-01", NA, "2023-01-01", NA)),
    date_baseline = data.table::as.IDate("2008-01-01"),
    date_death    = data.table::as.IDate(c(NA, "2020-06-01", NA, NA)),
    date_lost     = data.table::as.IDate(c(NA, NA, NA, "2019-03-01"))
  )
}

.age_dt <- function() {
  data.table::data.table(
    eid             = 1:4,
    t2d_status      = c(1L, 1L, 0L, 1L),
    t2d_date        = data.table::as.IDate(
      c("2010-01-01", "2018-06-01", NA, "2016-03-15")),
    cvd_status      = c(1L, 0L, 1L, 0L),
    cvd_date        = data.table::as.IDate(
      c("2019-05-01", NA, "2021-11-01", NA)),
    date_baseline   = data.table::as.IDate("2008-01-01"),
    age_recruitment = c(55.0, 62.0, 48.0, 70.0)
  )
}

.case_dt <- function() {
  data.table::data.table(
    eid                 = 1:5,
    t2d_icd10          = c(TRUE,  FALSE, TRUE,  FALSE, TRUE),
    t2d_selfreport     = c(FALSE, TRUE,  TRUE,  FALSE, FALSE),
    t2d_icd10_date     = data.table::as.IDate(
      c("2010-01-01", NA, "2015-06-01", NA, "2018-03-01")),
    t2d_selfreport_date = data.table::as.IDate(
      c(NA, "2012-05-01", "2013-01-01", NA, NA))
  )
}


# =============================================================================
# derive_missing()
# =============================================================================

test_that("derive_missing() stops on non-data.frame input", {
  expect_error(derive_missing("not a df"), "data.frame")
})

test_that("derive_missing() stops on invalid extra_labels type", {
  expect_error(derive_missing(data.table::copy(DT), extra_labels = 123), "character")
})

test_that("derive_missing() action='na': removes all three built-in refusal labels", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(derive_missing(d, action = "na"))
  char_cols <- names(result)[vapply(result, is.character, logical(1L))]
  refusal   <- c("Do not know", "Prefer not to answer", "Prefer not to say")
  for (col in char_cols) {
    expect_false(any(result[[col]] %in% refusal, na.rm = TRUE),
                 label = paste0("refusal label still present in ", col))
  }
})

test_that("derive_missing() always converts empty string to NA regardless of action", {
  d_na  <- data.table::copy(DT)
  d_unk <- data.table::copy(DT)
  r_na  <- suppressMessages(derive_missing(d_na,  action = "na"))
  r_unk <- suppressMessages(derive_missing(d_unk, action = "unknown"))
  char_cols <- names(r_na)[vapply(r_na, is.character, logical(1L))]
  for (col in char_cols) {
    expect_false(any(r_na[[col]]  == "", na.rm = TRUE))
    expect_false(any(r_unk[[col]] == "", na.rm = TRUE))
  }
})

test_that("derive_missing() action='unknown': converts refusal labels to 'Unknown'", {
  d <- data.table::data.table(
    eid   = 1:4,
    smoke = c("Never", "Prefer not to answer", "Do not know", "Current")
  )
  result <- suppressMessages(derive_missing(d, action = "unknown"))
  expect_equal(result$smoke[2], "Unknown")
  expect_equal(result$smoke[3], "Unknown")
  expect_equal(result$smoke[1], "Never")    # valid label untouched
})

test_that("derive_missing() preserves valid labels and skips numeric columns", {
  d      <- data.table::copy(DT)
  bmi_before <- d$p21001_i0
  result <- suppressMessages(derive_missing(d, action = "na"))
  expect_equal(result$p21001_i0, bmi_before)
})

test_that("derive_missing() extra_labels treated as informative missing", {
  d <- data.table::data.table(
    eid   = 1:3,
    smoke = c("Never", "Not applicable", "Current")
  )
  result <- suppressMessages(
    derive_missing(d, extra_labels = "Not applicable", action = "na")
  )
  expect_true(is.na(result$smoke[2]))
  expect_equal(result$smoke[1], "Never")
})

test_that("derive_missing() returns data.table", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(derive_missing(d))
  expect_true(data.table::is.data.table(result))
})


# =============================================================================
# derive_covariate()
# =============================================================================

test_that("derive_covariate() stops on non-data.frame input", {
  expect_error(derive_covariate("not a df"), "data.frame")
})

test_that("derive_covariate() stops on non-list factor_levels", {
  expect_error(
    suppressMessages(derive_covariate(data.table::copy(DT), factor_levels = "bad")),
    "named list"
  )
})

test_that("derive_covariate() converts character column to numeric", {
  d <- data.table::data.table(eid = 1:3, bmi_char = c("22.5", "30.1", "28.4"))
  result <- suppressMessages(derive_covariate(d, as_numeric = "bmi_char"))
  expect_true(is.numeric(result$bmi_char))
  expect_equal(result$bmi_char, c(22.5, 30.1, 28.4))
})

test_that("derive_covariate() converts smoking column to factor using ops_toy data", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(derive_covariate(d, as_factor = "p20116_i0"))
  expect_true(is.factor(result$p20116_i0))
})

test_that("derive_covariate() respects custom factor_levels order", {
  d <- data.table::copy(DT)
  result <- suppressMessages(
    derive_covariate(d, as_factor = "p20116_i0",
                     factor_levels = list(
                       p20116_i0 = c("Never", "Previous", "Current", "Prefer not to answer")
                     ))
  )
  expect_equal(levels(result$p20116_i0)[1], "Never")
  expect_equal(levels(result$p20116_i0)[3], "Current")
})

test_that("derive_covariate() skips missing column with warning", {
  d <- data.table::copy(DT)
  expect_message(
    derive_covariate(d, as_numeric = "nonexistent_col"),
    regexp = "not found|skipped",
    ignore.case = TRUE
  )
})

test_that("derive_covariate() returns data.table", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(derive_covariate(d, as_factor = "p21000_i0"))
  expect_true(data.table::is.data.table(result))
})


# =============================================================================
# derive_cut()
# =============================================================================

test_that("derive_cut() stops on non-data.frame input", {
  expect_error(derive_cut("not a df", col = "p21001_i0", n = 3), "data.frame")
})

test_that("derive_cut() stops when col not found", {
  expect_error(
    suppressMessages(derive_cut(data.table::copy(DT), col = "nonexistent", n = 3)),
    "missing column"
  )
})

test_that("derive_cut() stops when n < 2", {
  expect_error(
    suppressMessages(derive_cut(data.table::copy(DT), col = "p21001_i0", n = 1)),
    "2"
  )
})

test_that("derive_cut() stops when breaks length != n - 1", {
  expect_error(
    suppressMessages(derive_cut(data.table::copy(DT), col = "p21001_i0",
                                n = 3, breaks = c(25))),
    "length"
  )
})

test_that("derive_cut() stops when labels length != n", {
  expect_error(
    suppressMessages(derive_cut(data.table::copy(DT), col = "p21001_i0",
                                n = 3, labels = c("low", "high"))),
    "length"
  )
})

test_that("derive_cut() uses correct default name suffix for n = 2/3/4/5", {
  for (spec in list(list(2L, "_bi"), list(3L, "_tri"),
                    list(4L, "_quad"), list(5L, "_quin"))) {
    d   <- data.table::copy(DT)
    res <- suppressMessages(derive_cut(d, col = "p21001_i0", n = spec[[1]]))
    expect_true(paste0("p21001_i0", spec[[2]]) %in% names(res),
                label = paste0("suffix ", spec[[2]], " missing for n=", spec[[1]]))
  }
})

test_that("derive_cut() produces factor with n levels (quantile-based)", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(derive_cut(d, col = "p21001_i0", n = 3))
  expect_true(is.factor(result$p21001_i0_tri))
  expect_equal(nlevels(result$p21001_i0_tri), 3L)
})

test_that("derive_cut() respects custom breaks, labels, and name", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(
    derive_cut(d, col = "p21001_i0", n = 3,
               breaks = c(25, 30),
               labels = c("Normal", "Overweight", "Obese"),
               name   = "bmi_who")
  )
  expect_true("bmi_who" %in% names(result))
  expect_equal(levels(result$bmi_who), c("Normal", "Overweight", "Obese"))
})

test_that("derive_cut() returns data.table", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(derive_cut(d, col = "p21001_i0", n = 3))
  expect_true(data.table::is.data.table(result))
})


# =============================================================================
# derive_selfreport()
# =============================================================================

test_that("derive_selfreport() stops on non-data.frame input", {
  expect_error(
    derive_selfreport("not a df", name = "t2d",
                      regex = "diabetes", field = "noncancer"),
    "data.frame"
  )
})

test_that("derive_selfreport() stops on non-scalar name", {
  expect_error(
    suppressMessages(
      derive_selfreport(data.table::copy(DT), name = c("t2d", "cvd"),
                        regex = "diabetes", field = "noncancer")
    ),
    "single non-empty string"
  )
})

test_that("derive_selfreport() adds {name}_selfreport and {name}_selfreport_date", {
  d      <- data.table::copy(DT)
  sr_d   <- grep("^p20002_i0_a", names(d), value = TRUE)
  sr_t   <- grep("^p20008_i0_a", names(d), value = TRUE)
  result <- suppressMessages(
    derive_selfreport(d, name = "t2d", regex = "type 2 diabetes",
                      field        = "noncancer",
                      disease_cols = sr_d,
                      date_cols    = sr_t,
                      visit_cols   = "p53_i0")
  )
  expect_true("t2d_selfreport"      %in% names(result))
  expect_true("t2d_selfreport_date" %in% names(result))
})

test_that("derive_selfreport() date column is IDate", {
  d    <- data.table::copy(DT)
  sr_d <- grep("^p20002_i0_a", names(d), value = TRUE)
  sr_t <- grep("^p20008_i0_a", names(d), value = TRUE)
  result <- suppressMessages(
    derive_selfreport(d, name = "t2d", regex = "type 2 diabetes",
                      field        = "noncancer",
                      disease_cols = sr_d,
                      date_cols    = sr_t,
                      visit_cols   = "p53_i0")
  )
  expect_true(inherits(result$t2d_selfreport_date, "Date"))
})

test_that("derive_selfreport() status is logical", {
  d    <- data.table::copy(DT)
  sr_d <- grep("^p20002_i0_a", names(d), value = TRUE)
  sr_t <- grep("^p20008_i0_a", names(d), value = TRUE)
  result <- suppressMessages(
    derive_selfreport(d, name = "t2d", regex = "type 2 diabetes",
                      field        = "noncancer",
                      disease_cols = sr_d,
                      date_cols    = sr_t,
                      visit_cols   = "p53_i0")
  )
  expect_true(is.logical(result$t2d_selfreport))
})

test_that("derive_selfreport() cases have non-NA date, non-cases have NA date", {
  d    <- data.table::copy(DT)
  sr_d <- grep("^p20002_i0_a", names(d), value = TRUE)
  sr_t <- grep("^p20008_i0_a", names(d), value = TRUE)
  result <- suppressMessages(
    derive_selfreport(d, name = "t2d", regex = "type 2 diabetes",
                      field        = "noncancer",
                      disease_cols = sr_d,
                      date_cols    = sr_t,
                      visit_cols   = "p53_i0")
  )
  # All non-cases must have NA date
  expect_true(all(is.na(result[t2d_selfreport == FALSE, t2d_selfreport_date])))
})

test_that("derive_selfreport() returns FALSE flag when no regex match", {
  d    <- data.table::copy(DT)
  sr_d <- grep("^p20002_i0_a", names(d), value = TRUE)
  sr_t <- grep("^p20008_i0_a", names(d), value = TRUE)
  result <- suppressMessages(
    derive_selfreport(d, name = "rare_disease", regex = "^xyzzy_not_real$",
                      field        = "noncancer",
                      disease_cols = sr_d,
                      date_cols    = sr_t,
                      visit_cols   = "p53_i0")
  )
  expect_true(all(!result$rare_disease_selfreport))
})

test_that("derive_selfreport() returns data.table", {
  d    <- data.table::copy(DT)
  sr_d <- grep("^p20002_i0_a", names(d), value = TRUE)
  sr_t <- grep("^p20008_i0_a", names(d), value = TRUE)
  result <- suppressMessages(
    derive_selfreport(d, name = "t2d", regex = "type 2 diabetes",
                      field        = "noncancer",
                      disease_cols = sr_d,
                      date_cols    = sr_t,
                      visit_cols   = "p53_i0")
  )
  expect_true(data.table::is.data.table(result))
})


# =============================================================================
# derive_first_occurrence()
# =============================================================================

test_that("derive_first_occurrence() stops on non-data.frame input", {
  expect_error(
    derive_first_occurrence("not a df", name = "t2d", field = 131742L),
    "data.frame"
  )
})

test_that("derive_first_occurrence() adds {name}_fo and {name}_fo_date", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(
    derive_first_occurrence(d, name = "t2d", field = 131742L, col = "p131742")
  )
  expect_true("t2d_fo"      %in% names(result))
  expect_true("t2d_fo_date" %in% names(result))
})

test_that("derive_first_occurrence() status is TRUE only when date is valid", {
  d <- data.table::data.table(
    eid     = 1:4,
    p131742 = c("2010-01-01", NA, "bad-date", "2015-06-01")
  )
  result <- suppressMessages(
    derive_first_occurrence(d, name = "t2d", field = 131742L, col = "p131742")
  )
  expect_true(result$t2d_fo[1])
  expect_false(result$t2d_fo[2])   # NA
  expect_false(result$t2d_fo[3])   # unparseable → NA → FALSE
  expect_true(result$t2d_fo[4])
})

test_that("derive_first_occurrence() date column is IDate", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(
    derive_first_occurrence(d, name = "t2d", field = 131742L, col = "p131742")
  )
  expect_true(inherits(result$t2d_fo_date, "Date"))
})

test_that("derive_first_occurrence() aborts when col missing and no cache", {
  d <- data.table::data.table(eid = 1:3, x = 1:3)
  expect_error(
    suppressMessages(derive_first_occurrence(d, name = "t2d", field = 999999L)),
    "Cannot find|not found",
    ignore.case = TRUE
  )
})

test_that("derive_first_occurrence() returns data.table", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(
    derive_first_occurrence(d, name = "t2d", field = 131742L, col = "p131742")
  )
  expect_true(data.table::is.data.table(result))
})


# =============================================================================
# derive_hes()
# =============================================================================

test_that("derive_hes() stops on non-data.frame input", {
  expect_error(derive_hes("not a df", name = "t2d", icd10 = "E11"), "data.frame")
})

test_that("derive_hes() adds {name}_hes and {name}_hes_date columns", {
  d        <- data.table::copy(DT)
  date_cols <- grep("^p41280_a", names(d), value = TRUE)
  result   <- suppressMessages(
    derive_hes(d, name = "t2d", icd10 = "E11",
               disease_cols = "p41270",
               date_cols    = date_cols)
  )
  expect_true("t2d_hes"      %in% names(result))
  expect_true("t2d_hes_date" %in% names(result))
})

test_that("derive_hes() status is logical", {
  d        <- data.table::copy(DT)
  date_cols <- grep("^p41280_a", names(d), value = TRUE)
  result   <- suppressMessages(
    derive_hes(d, name = "t2d", icd10 = "E11",
               disease_cols = "p41270",
               date_cols    = date_cols)
  )
  expect_true(is.logical(result$t2d_hes))
})

test_that("derive_hes() prefix match finds at least some cases in ops_toy data", {
  d        <- data.table::copy(DT)
  date_cols <- grep("^p41280_a", names(d), value = TRUE)
  result   <- suppressMessages(
    derive_hes(d, name = "t2d", icd10 = "E11", match = "prefix",
               disease_cols = "p41270",
               date_cols    = date_cols)
  )
  # ops_toy generates E11 in HES with ~35% having records, multiple ICD codes
  expect_true(sum(result$t2d_hes) > 0L)
})

test_that("derive_hes() date is IDate and NA for non-cases", {
  d        <- data.table::copy(DT)
  date_cols <- grep("^p41280_a", names(d), value = TRUE)
  result   <- suppressMessages(
    derive_hes(d, name = "t2d", icd10 = "E11",
               disease_cols = "p41270",
               date_cols    = date_cols)
  )
  expect_true(inherits(result$t2d_hes_date, "Date"))
  expect_true(all(is.na(result[t2d_hes == FALSE, t2d_hes_date])))
})

test_that("derive_hes() exact match returns subset of prefix match cases", {
  d        <- data.table::copy(DT)
  date_cols <- grep("^p41280_a", names(d), value = TRUE)
  r_prefix  <- suppressMessages(
    derive_hes(data.table::copy(d), name = "t2d", icd10 = "E", match = "prefix",
               disease_cols = "p41270", date_cols = date_cols)
  )
  r_exact   <- suppressMessages(
    derive_hes(data.table::copy(d), name = "t2d", icd10 = "E11", match = "exact",
               disease_cols = "p41270", date_cols = date_cols)
  )
  expect_true(sum(r_exact$t2d_hes) <= sum(r_prefix$t2d_hes))
})

test_that("derive_hes() returns data.table", {
  d        <- data.table::copy(DT)
  date_cols <- grep("^p41280_a", names(d), value = TRUE)
  result   <- suppressMessages(
    derive_hes(d, name = "t2d", icd10 = "E11",
               disease_cols = "p41270",
               date_cols    = date_cols)
  )
  expect_true(data.table::is.data.table(result))
})


# =============================================================================
# derive_cancer_registry()
# =============================================================================

test_that("derive_cancer_registry() stops on non-data.frame input", {
  expect_error(
    derive_cancer_registry("not a df", name = "skin", icd10 = "^C44"),
    "data.frame"
  )
})

test_that("derive_cancer_registry() stops on non-scalar icd10 regex", {
  d <- data.table::copy(DT)
  expect_error(
    suppressMessages(
      derive_cancer_registry(d, name = "skin", icd10 = c("^C44", "^C50"))
    ),
    "single regex"
  )
})

test_that("derive_cancer_registry() adds {name}_cancer and {name}_cancer_date", {
  d    <- data.table::copy(DT)
  code <- grep("^p40006_i", names(d), value = TRUE)
  hist <- grep("^p40011_i", names(d), value = TRUE)
  behv <- grep("^p40012_i", names(d), value = TRUE)
  date <- grep("^p40005_i", names(d), value = TRUE)
  result <- suppressMessages(
    derive_cancer_registry(d, name = "skin", icd10 = "^C44",
                           code_cols = code, hist_cols = hist,
                           behv_cols = behv, date_cols = date)
  )
  expect_true("skin_cancer"      %in% names(result))
  expect_true("skin_cancer_date" %in% names(result))
})

test_that("derive_cancer_registry() status is logical and date is IDate", {
  d    <- data.table::copy(DT)
  code <- grep("^p40006_i", names(d), value = TRUE)
  hist <- grep("^p40011_i", names(d), value = TRUE)
  behv <- grep("^p40012_i", names(d), value = TRUE)
  date <- grep("^p40005_i", names(d), value = TRUE)
  result <- suppressMessages(
    derive_cancer_registry(d, name = "skin", icd10 = "^C44",
                           code_cols = code, hist_cols = hist,
                           behv_cols = behv, date_cols = date)
  )
  expect_true(is.logical(result$skin_cancer))
  expect_true(inherits(result$skin_cancer_date, "Date"))
})

test_that("derive_cancer_registry() behaviour filter narrows case count", {
  d    <- data.table::copy(DT)
  code <- grep("^p40006_i", names(d), value = TRUE)
  hist <- grep("^p40011_i", names(d), value = TRUE)
  behv <- grep("^p40012_i", names(d), value = TRUE)
  date <- grep("^p40005_i", names(d), value = TRUE)
  r_all     <- suppressMessages(
    derive_cancer_registry(data.table::copy(d), name = "skin",  icd10 = "^C44",
                           code_cols = code, hist_cols = hist,
                           behv_cols = behv, date_cols = date)
  )
  r_malign  <- suppressMessages(
    derive_cancer_registry(data.table::copy(d), name = "skin2", icd10 = "^C44",
                           behaviour = 3L,
                           code_cols = code, hist_cols = hist,
                           behv_cols = behv, date_cols = date)
  )
  expect_true(sum(r_malign$skin2_cancer) <= sum(r_all$skin_cancer))
})

test_that("derive_cancer_registry() non-cases have NA date", {
  d    <- data.table::copy(DT)
  code <- grep("^p40006_i", names(d), value = TRUE)
  hist <- grep("^p40011_i", names(d), value = TRUE)
  behv <- grep("^p40012_i", names(d), value = TRUE)
  date <- grep("^p40005_i", names(d), value = TRUE)
  result <- suppressMessages(
    derive_cancer_registry(d, name = "skin", icd10 = "^C44",
                           code_cols = code, hist_cols = hist,
                           behv_cols = behv, date_cols = date)
  )
  expect_true(all(is.na(result[skin_cancer == FALSE, skin_cancer_date])))
})

test_that("derive_cancer_registry() returns data.table", {
  d    <- data.table::copy(DT)
  code <- grep("^p40006_i", names(d), value = TRUE)
  result <- suppressMessages(
    derive_cancer_registry(d, name = "skin", icd10 = "^C44", code_cols = code)
  )
  expect_true(data.table::is.data.table(result))
})


# =============================================================================
# derive_death_registry()
# =============================================================================

test_that("derive_death_registry() stops on non-data.frame input", {
  expect_error(
    derive_death_registry("not a df", name = "cvd", icd10 = "I21"),
    "data.frame"
  )
})

test_that("derive_death_registry() adds {name}_death and {name}_death_date", {
  d    <- data.table::copy(DT)
  pri  <- grep("^p40001_i", names(d), value = TRUE)
  sec  <- grep("^p40002_i", names(d), value = TRUE)
  dts  <- grep("^p40000_i", names(d), value = TRUE)
  result <- suppressMessages(
    derive_death_registry(d, name = "cvd", icd10 = "I",
                          primary_cols   = pri,
                          secondary_cols = sec,
                          date_cols      = dts)
  )
  expect_true("cvd_death"      %in% names(result))
  expect_true("cvd_death_date" %in% names(result))
})

test_that("derive_death_registry() status is logical and date is IDate", {
  d    <- data.table::copy(DT)
  pri  <- grep("^p40001_i", names(d), value = TRUE)
  sec  <- grep("^p40002_i", names(d), value = TRUE)
  dts  <- grep("^p40000_i", names(d), value = TRUE)
  result <- suppressMessages(
    derive_death_registry(d, name = "cvd", icd10 = "I",
                          primary_cols   = pri,
                          secondary_cols = sec,
                          date_cols      = dts)
  )
  expect_true(is.logical(result$cvd_death))
  expect_true(inherits(result$cvd_death_date, "Date"))
})

test_that("derive_death_registry() exact match is subset of prefix match", {
  d   <- data.table::copy(DT)
  pri <- grep("^p40001_i", names(d), value = TRUE)
  sec <- grep("^p40002_i", names(d), value = TRUE)
  dts <- grep("^p40000_i", names(d), value = TRUE)
  r_prefix <- suppressMessages(
    derive_death_registry(data.table::copy(d), name = "cvd",  icd10 = "I",
                          match = "prefix",
                          primary_cols = pri, secondary_cols = sec, date_cols = dts)
  )
  r_exact  <- suppressMessages(
    derive_death_registry(data.table::copy(d), name = "cvd2", icd10 = "I25.9",
                          match = "exact",
                          primary_cols = pri, secondary_cols = sec, date_cols = dts)
  )
  expect_true(sum(r_exact$cvd2_death) <= sum(r_prefix$cvd_death))
})

test_that("derive_death_registry() non-cases have NA date", {
  d   <- data.table::copy(DT)
  pri <- grep("^p40001_i", names(d), value = TRUE)
  sec <- grep("^p40002_i", names(d), value = TRUE)
  dts <- grep("^p40000_i", names(d), value = TRUE)
  result <- suppressMessages(
    derive_death_registry(d, name = "cvd", icd10 = "I",
                          primary_cols = pri, secondary_cols = sec, date_cols = dts)
  )
  expect_true(all(is.na(result[cvd_death == FALSE, cvd_death_date])))
})

test_that("derive_death_registry() returns data.table", {
  d   <- data.table::copy(DT)
  pri <- grep("^p40001_i", names(d), value = TRUE)
  sec <- grep("^p40002_i", names(d), value = TRUE)
  dts <- grep("^p40000_i", names(d), value = TRUE)
  result <- suppressMessages(
    derive_death_registry(d, name = "cvd", icd10 = "I",
                          primary_cols = pri, secondary_cols = sec, date_cols = dts)
  )
  expect_true(data.table::is.data.table(result))
})


# =============================================================================
# derive_icd10()
# =============================================================================

test_that("derive_icd10() adds {name}_icd10 and {name}_icd10_date", {
  d        <- data.table::copy(DT)
  date_cols <- grep("^p41280_a", names(d), value = TRUE)
  pri       <- grep("^p40001_i", names(d), value = TRUE)
  sec       <- grep("^p40002_i", names(d), value = TRUE)
  dts       <- grep("^p40000_i", names(d), value = TRUE)
  result <- suppressMessages(
    derive_icd10(d, name = "t2d", icd10 = "E11",
                 source          = c("hes", "death"),
                 hes_code_col    = "p41270",
                 hes_date_cols   = date_cols,
                 primary_cols    = pri,
                 secondary_cols  = sec,
                 death_date_cols = dts)
  )
  expect_true("t2d_icd10"      %in% names(result))
  expect_true("t2d_icd10_date" %in% names(result))
})

test_that("derive_icd10() retains intermediate source columns", {
  d        <- data.table::copy(DT)
  date_cols <- grep("^p41280_a", names(d), value = TRUE)
  result <- suppressMessages(
    derive_icd10(d, name = "t2d", icd10 = "E11",
                 source       = "hes",
                 hes_code_col = "p41270",
                 hes_date_cols = date_cols)
  )
  expect_true("t2d_hes"      %in% names(result))
  expect_true("t2d_hes_date" %in% names(result))
})

test_that("derive_icd10() combined count >= any single source count", {
  d        <- data.table::copy(DT)
  date_cols <- grep("^p41280_a", names(d), value = TRUE)
  pri       <- grep("^p40001_i", names(d), value = TRUE)
  sec       <- grep("^p40002_i", names(d), value = TRUE)
  dts       <- grep("^p40000_i", names(d), value = TRUE)
  result <- suppressMessages(
    derive_icd10(d, name = "t2d", icd10 = "E11",
                 source          = c("hes", "death"),
                 hes_code_col    = "p41270",
                 hes_date_cols   = date_cols,
                 primary_cols    = pri,
                 secondary_cols  = sec,
                 death_date_cols = dts)
  )
  expect_true(sum(result$t2d_icd10) >= sum(result$t2d_hes))
  expect_true(sum(result$t2d_icd10) >= sum(result$t2d_death))
})

test_that("derive_icd10() warns when first_occurrence selected without fo_field/fo_col", {
  d        <- data.table::copy(DT)
  date_cols <- grep("^p41280_a", names(d), value = TRUE)
  expect_message(
    derive_icd10(d, name = "t2d", icd10 = "E11",
                 source        = c("hes", "first_occurrence"),
                 hes_code_col  = "p41270",
                 hes_date_cols = date_cols),
    regexp = "fo_field|fo_col|skipping",
    ignore.case = TRUE
  )
})

test_that("derive_icd10() returns data.table", {
  d        <- data.table::copy(DT)
  date_cols <- grep("^p41280_a", names(d), value = TRUE)
  result   <- suppressMessages(
    derive_icd10(d, name = "t2d", icd10 = "E11",
                 source       = "hes",
                 hes_code_col = "p41270",
                 hes_date_cols = date_cols)
  )
  expect_true(data.table::is.data.table(result))
})


# =============================================================================
# derive_case()
# =============================================================================

test_that("derive_case() stops on non-data.frame input", {
  expect_error(derive_case("not a df", name = "t2d"), "data.frame")
})

test_that("derive_case() aborts when neither status column found", {
  d <- data.table::data.table(eid = 1:3, x = 1:3)
  expect_error(
    suppressMessages(derive_case(d, name = "t2d")),
    "not found|found in data",
    ignore.case = TRUE
  )
})

test_that("derive_case() adds {name}_status and {name}_date", {
  d      <- .case_dt()
  result <- suppressMessages(derive_case(d, name = "t2d"))
  expect_true("t2d_status" %in% names(result))
  expect_true("t2d_date"   %in% names(result))
})

test_that("derive_case() OR-combines status correctly", {
  d      <- .case_dt()
  result <- suppressMessages(derive_case(d, name = "t2d"))
  expect_true(result$t2d_status[1])    # icd10 only
  expect_true(result$t2d_status[2])    # selfreport only
  expect_true(result$t2d_status[3])    # both
  expect_false(result$t2d_status[4])   # neither
  expect_true(result$t2d_status[5])    # icd10 only
})

test_that("derive_case() takes earliest date across sources", {
  d      <- .case_dt()
  result <- suppressMessages(derive_case(d, name = "t2d"))
  # Row 3: icd10_date=2015-06-01, selfreport_date=2013-01-01 → 2013-01-01
  expect_equal(result$t2d_date[3], data.table::as.IDate("2013-01-01"))
  # Row 1: icd10_date=2010-01-01, selfreport_date=NA → 2010-01-01
  expect_equal(result$t2d_date[1], data.table::as.IDate("2010-01-01"))
})

test_that("derive_case() returns data.table", {
  d      <- .case_dt()
  result <- suppressMessages(derive_case(d, name = "t2d"))
  expect_true(data.table::is.data.table(result))
})


# =============================================================================
# derive_age()
# =============================================================================

test_that("derive_age() stops on non-data.frame input", {
  expect_error(
    derive_age("not a df", name = "t2d",
               baseline_col = "date_baseline", age_col = "age_recruitment"),
    "data.frame"
  )
})

test_that("derive_age() stops on empty name", {
  d <- .age_dt()
  expect_error(
    derive_age(d, name = character(0),
               baseline_col = "date_baseline", age_col = "age_recruitment"),
    "non-empty"
  )
})

test_that("derive_age() stops when baseline_col missing", {
  d <- .age_dt()
  expect_error(
    derive_age(d, name = "t2d",
               baseline_col = "nonexistent", age_col = "age_recruitment"),
    "nonexistent"
  )
})

test_that("derive_age() adds age_at_{name} column", {
  d      <- .age_dt()
  result <- suppressMessages(
    derive_age(d, name = "t2d",
               baseline_col = "date_baseline", age_col = "age_recruitment")
  )
  expect_true("age_at_t2d" %in% names(result))
})

test_that("derive_age() computes correct age for integer-status case", {
  d      <- .age_dt()
  result <- suppressMessages(
    derive_age(d, name = "t2d",
               baseline_col = "date_baseline", age_col = "age_recruitment")
  )
  # Row 1: age=55, date=2010-01-01, baseline=2008-01-01
  expected <- 55 + as.numeric(
    data.table::as.IDate("2010-01-01") - data.table::as.IDate("2008-01-01")
  ) / 365.25
  expect_equal(result$age_at_t2d[1], expected)
})

test_that("derive_age() returns NA for non-cases (status = 0)", {
  d      <- .age_dt()
  result <- suppressMessages(
    derive_age(d, name = "t2d",
               baseline_col = "date_baseline", age_col = "age_recruitment")
  )
  expect_true(is.na(result$age_at_t2d[3]))
})

test_that("derive_age() handles logical status column", {
  d <- data.table::data.table(
    eid             = 1:3,
    t2d             = c(TRUE, FALSE, TRUE),
    t2d_date        = data.table::as.IDate(c("2012-06-01", NA, "2017-09-01")),
    date_baseline   = data.table::as.IDate("2010-01-01"),
    age_recruitment = c(50.0, 60.0, 55.0)
  )
  result <- suppressMessages(
    derive_age(d, name = "t2d",
               baseline_col = "date_baseline", age_col = "age_recruitment")
  )
  expect_false(is.na(result$age_at_t2d[1]))
  expect_true(is.na(result$age_at_t2d[2]))   # FALSE → NA
})

test_that("derive_age() processes multiple names in one call", {
  d      <- .age_dt()
  result <- suppressMessages(
    derive_age(d, name = c("t2d", "cvd"),
               baseline_col = "date_baseline", age_col = "age_recruitment")
  )
  expect_true("age_at_t2d" %in% names(result))
  expect_true("age_at_cvd" %in% names(result))
})

test_that("derive_age() warns and skips when date column not found", {
  d <- .age_dt()
  expect_message(
    derive_age(d, name = "unknown_outcome",
               baseline_col = "date_baseline", age_col = "age_recruitment"),
    regexp = "not found|skipping",
    ignore.case = TRUE
  )
  expect_false("age_at_unknown_outcome" %in% names(d))
})

test_that("derive_age() returns data.table", {
  d      <- .age_dt()
  result <- suppressMessages(
    derive_age(d, name = "t2d",
               baseline_col = "date_baseline", age_col = "age_recruitment")
  )
  expect_true(data.table::is.data.table(result))
})


# =============================================================================
# derive_followup()
# =============================================================================

test_that("derive_followup() stops on non-data.frame input", {
  expect_error(
    derive_followup("not a df", name = "t2d", event_col = "t2d_date",
                    baseline_col = "date_baseline", censor_date = as.Date("2022-06-01")),
    "data.frame"
  )
})

test_that("derive_followup() stops when event_col missing", {
  d <- .followup_dt()
  expect_error(
    suppressMessages(
      derive_followup(d, name = "t2d", event_col = "nonexistent",
                      baseline_col = "date_baseline", censor_date = as.Date("2022-06-01"))
    ),
    "nonexistent"
  )
})

test_that("derive_followup() adds {name}_followup_end and {name}_followup_years", {
  d      <- .followup_dt()
  result <- suppressMessages(
    derive_followup(d, name = "t2d", event_col = "t2d_date",
                    baseline_col = "date_baseline", censor_date = as.Date("2022-06-01"))
  )
  expect_true("t2d_followup_end"   %in% names(result))
  expect_true("t2d_followup_years" %in% names(result))
})

test_that("derive_followup() end = event_date when event precedes censor", {
  d      <- .followup_dt()
  result <- suppressMessages(
    derive_followup(d, name = "t2d", event_col = "t2d_date",
                    baseline_col = "date_baseline", censor_date = as.Date("2022-06-01"),
                    death_col = "date_death", lost_col = "date_lost")
  )
  # Row 1: t2d 2018-01-01 < censor
  expect_equal(result$t2d_followup_end[1], data.table::as.IDate("2018-01-01"))
})

test_that("derive_followup() end = censor_date when event exceeds censor", {
  d      <- .followup_dt()
  result <- suppressMessages(
    derive_followup(d, name = "t2d", event_col = "t2d_date",
                    baseline_col = "date_baseline", censor_date = as.Date("2022-06-01"),
                    death_col = "date_death", lost_col = "date_lost")
  )
  # Row 3: t2d 2023-01-01 > censor
  expect_equal(result$t2d_followup_end[3], data.table::as.IDate("2022-06-01"))
})

test_that("derive_followup() end = death_date when death precedes censor", {
  d      <- .followup_dt()
  result <- suppressMessages(
    derive_followup(d, name = "t2d", event_col = "t2d_date",
                    baseline_col = "date_baseline", censor_date = as.Date("2022-06-01"),
                    death_col = "date_death", lost_col = "date_lost")
  )
  # Row 2: no t2d, death 2020-06-01 < censor
  expect_equal(result$t2d_followup_end[2], data.table::as.IDate("2020-06-01"))
})

test_that("derive_followup() end = lost_date when lost precedes censor", {
  d      <- .followup_dt()
  result <- suppressMessages(
    derive_followup(d, name = "t2d", event_col = "t2d_date",
                    baseline_col = "date_baseline", censor_date = as.Date("2022-06-01"),
                    death_col = "date_death", lost_col = "date_lost")
  )
  # Row 4: no t2d, no death, lost 2019-03-01 < censor
  expect_equal(result$t2d_followup_end[4], data.table::as.IDate("2019-03-01"))
})

test_that("derive_followup() followup_years computed correctly", {
  d      <- .followup_dt()
  result <- suppressMessages(
    derive_followup(d, name = "t2d", event_col = "t2d_date",
                    baseline_col = "date_baseline", censor_date = as.Date("2022-06-01"))
  )
  expected <- as.numeric(
    data.table::as.IDate("2018-01-01") - data.table::as.IDate("2008-01-01")
  ) / 365.25
  expect_equal(result$t2d_followup_years[1], expected)
})

test_that("derive_followup() followup_years all positive and finite", {
  d      <- .followup_dt()
  result <- suppressMessages(
    derive_followup(d, name = "t2d", event_col = "t2d_date",
                    baseline_col = "date_baseline", censor_date = as.Date("2022-06-01"))
  )
  yrs <- result$t2d_followup_years
  expect_true(all(yrs > 0, na.rm = TRUE))
  expect_true(all(is.finite(yrs)))
})

test_that("derive_followup() accepts character censor_date", {
  d <- .followup_dt()
  expect_no_error(
    suppressMessages(
      derive_followup(d, name = "t2d", event_col = "t2d_date",
                      baseline_col = "date_baseline", censor_date = "2022-06-01")
    )
  )
})

test_that("derive_followup() returns data.table", {
  d      <- .followup_dt()
  result <- suppressMessages(
    derive_followup(d, name = "t2d", event_col = "t2d_date",
                    baseline_col = "date_baseline", censor_date = as.Date("2022-06-01"))
  )
  expect_true(data.table::is.data.table(result))
})


# =============================================================================
# derive_timing()
# =============================================================================

test_that("derive_timing() stops on non-data.frame input", {
  expect_error(
    derive_timing("not a df", name = "t2d", baseline_col = "date_baseline"),
    "data.frame"
  )
})

test_that("derive_timing() stops when required column missing", {
  d <- .timing_dt()
  expect_error(
    suppressMessages(derive_timing(d, name = "t2d", baseline_col = "nonexistent")),
    "nonexistent"
  )
})

test_that("derive_timing() adds {name}_timing column", {
  d      <- .timing_dt()
  result <- suppressMessages(
    derive_timing(d, name = "t2d", baseline_col = "date_baseline")
  )
  expect_true("t2d_timing" %in% names(result))
})

test_that("derive_timing() codes 0 for no-disease rows", {
  d      <- .timing_dt()
  result <- suppressMessages(
    derive_timing(d, name = "t2d", baseline_col = "date_baseline")
  )
  expect_equal(result$t2d_timing[1], 0L)
  expect_equal(result$t2d_timing[5], 0L)
})

test_that("derive_timing() codes 1 for prevalent (date <= baseline)", {
  d      <- .timing_dt()
  result <- suppressMessages(
    derive_timing(d, name = "t2d", baseline_col = "date_baseline")
  )
  # Row 2: t2d_date 2010-01-01 < baseline 2015-01-01 → prevalent
  expect_equal(result$t2d_timing[2], 1L)
})

test_that("derive_timing() codes 2 for incident (date > baseline)", {
  d      <- .timing_dt()
  result <- suppressMessages(
    derive_timing(d, name = "t2d", baseline_col = "date_baseline")
  )
  # Row 3: t2d_date 2018-06-01 > baseline 2015-01-01 → incident
  expect_equal(result$t2d_timing[3], 2L)
})

test_that("derive_timing() codes NA when case has no date", {
  d      <- .timing_dt()
  result <- suppressMessages(
    derive_timing(d, name = "t2d", baseline_col = "date_baseline")
  )
  # Row 4: status TRUE but date NA
  expect_true(is.na(result$t2d_timing[4]))
})

test_that("derive_timing() respects explicit status_col and date_col overrides", {
  d <- data.table::data.table(
    eid           = 1:2,
    my_status     = c(TRUE, FALSE),
    my_date       = data.table::as.IDate(c("2012-01-01", NA)),
    date_baseline = data.table::as.IDate("2015-01-01")
  )
  result <- suppressMessages(
    derive_timing(d, name = "x",
                  status_col   = "my_status",
                  date_col     = "my_date",
                  baseline_col = "date_baseline")
  )
  expect_equal(result$x_timing[1], 1L)   # prevalent
  expect_equal(result$x_timing[2], 0L)   # no disease
})

test_that("derive_timing() returns data.table", {
  d      <- .timing_dt()
  result <- suppressMessages(
    derive_timing(d, name = "t2d", baseline_col = "date_baseline")
  )
  expect_true(data.table::is.data.table(result))
})

# =============================================================================
# test-integration-derive.R — Integration tests for derive_ series
# Fixture: ops_toy(n = 2000, seed = 42) — realistic UKB-like structure
# Runs the full derive pipeline for t2d (Type 2 Diabetes) and cvd (CVD).
# Run manually before release: devtools::test(filter = "integration-derive")
# =============================================================================


# =============================================================================
# Shared fixture
# =============================================================================

skip_on_cran()
DT      <- suppressMessages(ops_toy(n = 2000L, seed = 42L))
CENSOR  <- as.Date("2022-06-01")

SR_D    <- grep("^p20002_i0_a", names(DT), value = TRUE)   # selfreport disease
SR_T    <- grep("^p20008_i0_a", names(DT), value = TRUE)   # selfreport dates
HES_DT  <- grep("^p41280_a",   names(DT), value = TRUE)    # HES dates
CR_CODE <- grep("^p40006_i",   names(DT), value = TRUE)    # cancer ICD-10
CR_HIST <- grep("^p40011_i",   names(DT), value = TRUE)    # cancer histology
CR_BEHV <- grep("^p40012_i",   names(DT), value = TRUE)    # cancer behaviour
CR_DATE <- grep("^p40005_i",   names(DT), value = TRUE)    # cancer date
DR_PRI  <- grep("^p40001_i",   names(DT), value = TRUE)    # death primary
DR_SEC  <- grep("^p40002_i",   names(DT), value = TRUE)    # death secondary
DR_DT   <- grep("^p40000_i",   names(DT), value = TRUE)    # death date


# =============================================================================
# derive_missing() — full data
# =============================================================================

test_that("derive_missing() runs on n=2000 without error", {
  d <- data.table::copy(DT)
  expect_no_error(suppressMessages(derive_missing(d, action = "na")))
})

test_that("derive_missing() removes all built-in refusal labels from character columns", {
  d       <- data.table::copy(DT)
  result  <- suppressMessages(derive_missing(d, action = "na"))
  refusal <- c("Do not know", "Prefer not to answer", "Prefer not to say")
  char_cols <- names(result)[vapply(result, is.character, logical(1L))]
  for (col in char_cols) {
    expect_false(any(result[[col]] %in% refusal, na.rm = TRUE),
                 label = paste0("refusal label still in ", col))
  }
})

test_that("derive_missing() action='unknown' produces 'Unknown', not NA, for refusals", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(derive_missing(d, action = "unknown"))
  char_cols <- names(result)[vapply(result, is.character, logical(1L))]
  for (col in char_cols) {
    expect_false(any(c("Do not know", "Prefer not to answer",
                       "Prefer not to say") %in% result[[col]], na.rm = TRUE),
                 label = paste0("raw refusal label still in ", col))
  }
})

test_that("derive_missing() does not alter numeric columns", {
  d_orig <- data.table::copy(DT)
  d      <- data.table::copy(DT)
  suppressMessages(derive_missing(d, action = "na"))
  expect_equal(d$p21001_i0, d_orig$p21001_i0)
  expect_equal(d$p22189,    d_orig$p22189)
})


# =============================================================================
# derive_covariate() — full data
# =============================================================================

test_that("derive_covariate() converts smoking to ordered factor", {
  d      <- data.table::copy(DT)
  suppressMessages(derive_missing(d, action = "na"))
  result <- suppressMessages(
    derive_covariate(d, as_factor = "p20116_i0",
                     factor_levels = list(
                       p20116_i0 = c("Never", "Previous", "Current")
                     ))
  )
  expect_true(is.factor(result$p20116_i0))
  expect_equal(levels(result$p20116_i0)[1], "Never")
})

test_that("derive_covariate() converts BMI to numeric when supplied as character", {
  d <- data.table::data.table(
    eid = 1:5,
    bmi = c("22.5", "30.1", "28.4", "19.8", "35.0")
  )
  result <- suppressMessages(derive_covariate(d, as_numeric = "bmi"))
  expect_true(is.numeric(result$bmi))
})

test_that("derive_covariate() warns for high-cardinality factor column", {
  d <- data.table::copy(DT)
  # p54_i0 (assessment centre) has 10 levels > default max_levels = 5
  expect_message(
    derive_covariate(d, as_factor = "p54_i0"),
    regexp = "levels|max_levels",
    ignore.case = TRUE
  )
})


# =============================================================================
# derive_cut() — full data
# =============================================================================

test_that("derive_cut() bins BMI into tertiles on n=2000", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(derive_cut(d, col = "p21001_i0", n = 3))
  expect_true("p21001_i0_tri" %in% names(result))
  expect_equal(nlevels(result$p21001_i0_tri), 3L)
  expect_equal(sum(is.na(result$p21001_i0_tri)), 0L)  # no NA from valid BMI
})

test_that("derive_cut() bins age into quartiles correctly", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(derive_cut(d, col = "p21022", n = 4))
  expect_equal(nlevels(result$p21022_quad), 4L)
})

test_that("derive_cut() WHO BMI categories with custom breaks", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(
    derive_cut(d, col = "p21001_i0", n = 3,
               breaks = c(25, 30),
               labels = c("Normal", "Overweight", "Obese"),
               name   = "bmi_who")
  )
  expect_equal(levels(result$bmi_who), c("Normal", "Overweight", "Obese"))
  expect_true(all(!is.na(result$bmi_who)))
})

test_that("derive_cut() preserves NA values in output", {
  d <- data.table::copy(DT)
  d[sample(.N, 50L), p21001_i0 := NA_real_]
  result <- suppressMessages(derive_cut(d, col = "p21001_i0", n = 3))
  expect_true(sum(is.na(result$p21001_i0_tri)) == 50L)
})


# =============================================================================
# derive_selfreport() — t2d via noncancer field
# =============================================================================

test_that("derive_selfreport() finds t2d cases in ops_toy noncancer data", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(
    derive_selfreport(d, name = "t2d", regex = "type 2 diabetes",
                      field        = "noncancer",
                      disease_cols = SR_D,
                      date_cols    = SR_T,
                      visit_cols   = "p53_i0")
  )
  expect_true(sum(result$t2d_selfreport) > 0L)
})

test_that("derive_selfreport() non-cases have NA date", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(
    derive_selfreport(d, name = "t2d", regex = "type 2 diabetes",
                      field        = "noncancer",
                      disease_cols = SR_D,
                      date_cols    = SR_T,
                      visit_cols   = "p53_i0")
  )
  expect_true(all(is.na(result[t2d_selfreport == FALSE, t2d_selfreport_date])))
})

test_that("derive_selfreport() hypertension yields more cases than t2d (higher prevalence)", {
  d     <- data.table::copy(DT)
  r_t2d <- suppressMessages(
    derive_selfreport(d, name = "t2d", regex = "type 2 diabetes",
                      field        = "noncancer",
                      disease_cols = SR_D, date_cols = SR_T, visit_cols = "p53_i0")
  )
  r_htn <- suppressMessages(
    derive_selfreport(d, name = "htn", regex = "hypertension",
                      field        = "noncancer",
                      disease_cols = SR_D, date_cols = SR_T, visit_cols = "p53_i0")
  )
  # Both should have at least some cases; hypertension fill rate ~30% vs t2d ~18%
  expect_true(sum(r_t2d$t2d_selfreport) > 0L)
  expect_true(sum(r_htn$htn_selfreport) > 0L)
})


# =============================================================================
# derive_first_occurrence() — using p131742
# =============================================================================

test_that("derive_first_occurrence() runs on ops_toy p131742 without error", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(
    derive_first_occurrence(d, name = "t2d_fo", field = 131742L, col = "p131742")
  )
  expect_true("t2d_fo_fo"      %in% names(result))
  expect_true("t2d_fo_fo_date" %in% names(result))
})

test_that("derive_first_occurrence() ~8% case rate matches ops_toy generation", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(
    derive_first_occurrence(d, name = "t2d_fo", field = 131742L, col = "p131742")
  )
  pct <- mean(result$t2d_fo_fo) * 100
  # ops_toy generates ~8% positive; with seed=42 and n=2000, allow 5-12%
  expect_true(pct >= 4 && pct <= 15,
              label = paste0("FO case rate ", round(pct, 1), "% outside expected 4-15%"))
})

test_that("derive_first_occurrence() every positive case has a valid date", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(
    derive_first_occurrence(d, name = "t2d_fo", field = 131742L, col = "p131742")
  )
  expect_true(all(!is.na(result[t2d_fo_fo == TRUE, t2d_fo_fo_date])))
})


# =============================================================================
# derive_hes() — E11 (type 2 diabetes) and I (cardiovascular)
# =============================================================================

test_that("derive_hes() finds E11 cases in ops_toy HES data", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(
    derive_hes(d, name = "t2d", icd10 = "E11",
               disease_cols = "p41270", date_cols = HES_DT)
  )
  expect_true(sum(result$t2d_hes) > 0L)
})

test_that("derive_hes() date never exceeds 2022-12-31 (ops_toy generation bound)", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(
    derive_hes(d, name = "t2d", icd10 = "E11",
               disease_cols = "p41270", date_cols = HES_DT)
  )
  max_date <- max(result$t2d_hes_date, na.rm = TRUE)
  expect_true(max_date <= data.table::as.IDate("2022-12-31"))
})

test_that("derive_hes() cardiovascular prefix 'I' finds cases across I10/I25/I48", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(
    derive_hes(d, name = "cvd", icd10 = "I", match = "prefix",
               disease_cols = "p41270", date_cols = HES_DT)
  )
  expect_true(sum(result$cvd_hes) > 0L)
})

test_that("derive_hes() combined count hes >= any single prefix subgroup", {
  d      <- data.table::copy(DT)
  r_all  <- suppressMessages(
    derive_hes(data.table::copy(d), name = "cvd",  icd10 = "I",   match = "prefix",
               disease_cols = "p41270", date_cols = HES_DT)
  )
  r_i25  <- suppressMessages(
    derive_hes(data.table::copy(d), name = "ihd",  icd10 = "I25", match = "prefix",
               disease_cols = "p41270", date_cols = HES_DT)
  )
  expect_true(sum(r_all$cvd_hes) >= sum(r_i25$ihd_hes))
})


# =============================================================================
# derive_cancer_registry() — C44 (non-melanoma skin cancer)
# =============================================================================

test_that("derive_cancer_registry() finds C44 cases in ops_toy data", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(
    derive_cancer_registry(d, name = "skin", icd10 = "^C44",
                           code_cols = CR_CODE, hist_cols = CR_HIST,
                           behv_cols = CR_BEHV, date_cols = CR_DATE)
  )
  expect_true(sum(result$skin_cancer) > 0L)
})

test_that("derive_cancer_registry() malignant filter (behaviour=3) reduces case count", {
  d        <- data.table::copy(DT)
  r_all    <- suppressMessages(
    derive_cancer_registry(data.table::copy(d), name = "skin",    icd10 = "^C44",
                           code_cols = CR_CODE, hist_cols = CR_HIST,
                           behv_cols = CR_BEHV, date_cols = CR_DATE)
  )
  r_malign <- suppressMessages(
    derive_cancer_registry(data.table::copy(d), name = "skin_inv", icd10 = "^C44",
                           behaviour = 3L,
                           code_cols = CR_CODE, hist_cols = CR_HIST,
                           behv_cols = CR_BEHV, date_cols = CR_DATE)
  )
  expect_true(sum(r_malign$skin_inv_cancer) <= sum(r_all$skin_cancer))
})

test_that("derive_cancer_registry() date is within ops_toy cancer range (1990-2020)", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(
    derive_cancer_registry(d, name = "skin", icd10 = "^C44",
                           code_cols = CR_CODE, hist_cols = CR_HIST,
                           behv_cols = CR_BEHV, date_cols = CR_DATE)
  )
  dates <- result$skin_cancer_date[!is.na(result$skin_cancer_date)]
  if (length(dates) > 0L) {
    expect_true(all(dates >= data.table::as.IDate("1990-01-01")))
    expect_true(all(dates <= data.table::as.IDate("2020-12-31")))
  }
})


# =============================================================================
# derive_death_registry() — I (cardiovascular) and C (cancer)
# =============================================================================

test_that("derive_death_registry() finds CVD deaths in ops_toy data", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(
    derive_death_registry(d, name = "cvd", icd10 = "I",
                          primary_cols   = DR_PRI,
                          secondary_cols = DR_SEC,
                          date_cols      = DR_DT)
  )
  expect_true(sum(result$cvd_death) > 0L)
})

test_that("derive_death_registry() death rate ~10% matches ops_toy generation", {
  d      <- data.table::copy(DT)
  # Any cause of death (prefix match on all codes)
  result <- suppressMessages(
    derive_death_registry(d, name = "any", icd10 = ".",
                          match          = "regex",
                          primary_cols   = DR_PRI,
                          secondary_cols = DR_SEC,
                          date_cols      = DR_DT)
  )
  pct <- mean(result$any_death) * 100
  # ops_toy death_flag = runif < 0.10; allow 7-14% with n=2000, seed=42
  expect_true(pct >= 5 && pct <= 18,
              label = paste0("death rate ", round(pct, 1), "% outside expected 5-18%"))
})

test_that("derive_death_registry() death dates within ops_toy range (2011-2023)", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(
    derive_death_registry(d, name = "cvd", icd10 = "I",
                          primary_cols   = DR_PRI,
                          secondary_cols = DR_SEC,
                          date_cols      = DR_DT)
  )
  dates <- result$cvd_death_date[!is.na(result$cvd_death_date)]
  if (length(dates) > 0L) {
    expect_true(all(dates >= data.table::as.IDate("2011-01-01")))
    expect_true(all(dates <= data.table::as.IDate("2023-12-31")))
  }
})


# =============================================================================
# derive_icd10() — t2d (HES + death) and cvd (HES + death + first_occurrence)
# =============================================================================

test_that("derive_icd10() t2d (hes + death) OR-combines both sources", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(
    derive_icd10(d, name = "t2d", icd10 = "E11",
                 source          = c("hes", "death"),
                 hes_code_col    = "p41270",
                 hes_date_cols   = HES_DT,
                 primary_cols    = DR_PRI,
                 secondary_cols  = DR_SEC,
                 death_date_cols = DR_DT)
  )
  # Combined must be >= each source alone
  expect_true(sum(result$t2d_icd10) >= sum(result$t2d_hes))
  expect_true(sum(result$t2d_icd10) >= sum(result$t2d_death))
})


test_that("derive_icd10() with first_occurrence adds fo columns", {
  d      <- data.table::copy(DT)
  result <- suppressMessages(
    derive_icd10(d, name = "t2d", icd10 = "E11",
                 source          = c("hes", "first_occurrence"),
                 hes_code_col    = "p41270",
                 hes_date_cols   = HES_DT,
                 fo_col          = "p131742")
  )
  expect_true("t2d_fo"   %in% names(result))
  expect_true("t2d_icd10" %in% names(result))
})


# =============================================================================
# derive_case() — t2d (icd10 + selfreport)
# =============================================================================

test_that("derive_case() merges icd10 + selfreport into unified case", {
  d <- data.table::copy(DT)

  # Step 1: derive component flags
  suppressMessages({
    d <- derive_selfreport(d, name = "t2d", regex = "type 2 diabetes",
                           field        = "noncancer",
                           disease_cols = SR_D, date_cols = SR_T,
                           visit_cols   = "p53_i0")
    d <- derive_icd10(d, name = "t2d", icd10 = "E11",
                      source       = "hes",
                      hes_code_col = "p41270",
                      hes_date_cols = HES_DT)
    d <- derive_case(d, name = "t2d")
  })

  expect_true("t2d_status" %in% names(d))
  expect_true("t2d_date"   %in% names(d))
})

test_that("derive_case() t2d_status >= t2d_icd10 and >= t2d_selfreport (OR logic)", {
  d <- data.table::copy(DT)
  suppressMessages({
    d <- derive_selfreport(d, name = "t2d", regex = "type 2 diabetes",
                           field        = "noncancer",
                           disease_cols = SR_D, date_cols = SR_T,
                           visit_cols   = "p53_i0")
    d <- derive_icd10(d, name = "t2d", icd10 = "E11",
                      source       = "hes",
                      hes_code_col = "p41270",
                      hes_date_cols = HES_DT)
    d <- derive_case(d, name = "t2d")
  })
  expect_true(sum(d$t2d_status) >= sum(d$t2d_icd10))
  expect_true(sum(d$t2d_status) >= sum(d$t2d_selfreport))
})

test_that("derive_case() t2d_date <= min of component dates where both present", {
  d <- data.table::copy(DT)
  suppressMessages({
    d <- derive_selfreport(d, name = "t2d", regex = "type 2 diabetes",
                           field        = "noncancer",
                           disease_cols = SR_D, date_cols = SR_T,
                           visit_cols   = "p53_i0")
    d <- derive_icd10(d, name = "t2d", icd10 = "E11",
                      source       = "hes",
                      hes_code_col = "p41270",
                      hes_date_cols = HES_DT)
    d <- derive_case(d, name = "t2d")
  })
  both <- d[t2d_icd10 == TRUE & t2d_selfreport == TRUE &
            !is.na(t2d_icd10_date) & !is.na(t2d_selfreport_date)]
  if (nrow(both) > 0L) {
    expected <- pmin(both$t2d_icd10_date, both$t2d_selfreport_date, na.rm = TRUE)
    expect_equal(both$t2d_date, data.table::as.IDate(expected))
  }
})


# =============================================================================
# derive_timing() — t2d (incident / prevalent classification)
# =============================================================================

test_that("derive_timing() adds t2d_timing with values in {0,1,2,NA}", {
  d <- data.table::copy(DT)
  suppressMessages({
    d <- derive_icd10(d, name = "t2d", icd10 = "E11",
                      source       = "hes",
                      hes_code_col = "p41270",
                      hes_date_cols = HES_DT)
    d <- derive_case(d, name = "t2d")
    d <- derive_timing(d, name = "t2d", baseline_col = "p53_i0")
  })
  expect_true("t2d_timing" %in% names(d))
  valid_vals <- c(0L, 1L, 2L, NA_integer_)
  expect_true(all(d$t2d_timing %in% valid_vals))
})

test_that("derive_timing() non-cases coded 0, incident coded 2 (post-baseline)", {
  d <- data.table::copy(DT)
  suppressMessages({
    d <- derive_icd10(d, name = "t2d", icd10 = "E11",
                      source       = "hes",
                      hes_code_col = "p41270",
                      hes_date_cols = HES_DT)
    d <- derive_case(d, name = "t2d")
    d <- derive_timing(d, name = "t2d", baseline_col = "p53_i0")
  })
  expect_true(all(d[t2d_status == FALSE, t2d_timing] == 0L, na.rm = TRUE))
  incident <- d[!is.na(t2d_timing) & t2d_timing == 2L]
  if (nrow(incident) > 0L) {
    bl <- data.table::as.IDate(incident$p53_i0)
    expect_true(all(incident$t2d_date > bl, na.rm = TRUE))
  }
})


# =============================================================================
# derive_age() — age at t2d onset and cvd onset
# =============================================================================

test_that("derive_age() computes plausible age_at_t2d (20-110 years)", {
  d <- data.table::copy(DT)
  suppressMessages({
    d <- derive_icd10(d, name = "t2d", icd10 = "E11",
                      source       = "hes",
                      hes_code_col = "p41270",
                      hes_date_cols = HES_DT)
    d <- derive_case(d, name = "t2d")
    d <- derive_age(d, name = "t2d",
                    baseline_col = "p53_i0",
                    age_col      = "p21022")
  })
  v <- d$age_at_t2d
  expect_true(all(v >= 20 & v <= 110, na.rm = TRUE))
})

test_that("derive_age() returns NA for non-cases", {
  d <- data.table::copy(DT)
  suppressMessages({
    d <- derive_icd10(d, name = "t2d", icd10 = "E11",
                      source       = "hes",
                      hes_code_col = "p41270",
                      hes_date_cols = HES_DT)
    d <- derive_case(d, name = "t2d")
    d <- derive_age(d, name = "t2d",
                    baseline_col = "p53_i0",
                    age_col      = "p21022")
  })
  expect_true(all(is.na(d[t2d_status == FALSE, age_at_t2d])))
})


# =============================================================================
# derive_followup() — survival variables for t2d
# =============================================================================

test_that("derive_followup() followup_end never exceeds censor_date", {
  d <- data.table::copy(DT)
  suppressMessages({
    d <- derive_icd10(d, name = "t2d", icd10 = "E11",
                      source       = "hes",
                      hes_code_col = "p41270",
                      hes_date_cols = HES_DT)
    d <- derive_case(d, name = "t2d")
    d <- derive_followup(d, name = "t2d",
                         event_col    = "t2d_date",
                         baseline_col = "p53_i0",
                         censor_date  = CENSOR,
                         death_col    = FALSE,
                         lost_col     = FALSE)
  })
  expect_true(all(
    d$t2d_followup_end <= data.table::as.IDate(CENSOR), na.rm = TRUE
  ))
})

test_that("derive_followup() followup_years all positive", {
  d <- data.table::copy(DT)
  suppressMessages({
    d <- derive_icd10(d, name = "t2d", icd10 = "E11",
                      source       = "hes",
                      hes_code_col = "p41270",
                      hes_date_cols = HES_DT)
    d <- derive_case(d, name = "t2d")
    d <- derive_followup(d, name = "t2d",
                         event_col    = "t2d_date",
                         baseline_col = "p53_i0",
                         censor_date  = CENSOR,
                         death_col    = FALSE,
                         lost_col     = FALSE)
  })
  expect_true(all(d$t2d_followup_years > 0, na.rm = TRUE))
})

test_that("derive_followup() with death competing endpoint reduces some follow-up times", {
  d <- data.table::copy(DT)
  suppressMessages({
    d <- derive_icd10(d, name = "t2d", icd10 = "E11",
                      source       = "hes",
                      hes_code_col = "p41270",
                      hes_date_cols = HES_DT)
    d <- derive_case(d, name = "t2d")
    # death date from p40000_i0
    d[, death_date := data.table::as.IDate(as.character(p40000_i0))]
    r_no_death <- suppressMessages(
      derive_followup(data.table::copy(d), name = "t2d",
                      event_col = "t2d_date", baseline_col = "p53_i0",
                      censor_date = CENSOR, death_col = FALSE, lost_col = FALSE)
    )
    r_with_death <- suppressMessages(
      derive_followup(data.table::copy(d), name = "t2d",
                      event_col = "t2d_date", baseline_col = "p53_i0",
                      censor_date = CENSOR, death_col = "death_date", lost_col = FALSE)
    )
  })
  # With death as competing event, some follow-up ends are earlier or equal
  end_no_death   <- r_no_death$t2d_followup_end
  end_with_death <- r_with_death$t2d_followup_end
  expect_true(all(end_with_death <= end_no_death, na.rm = TRUE))
})


# =============================================================================
# End-to-end pipeline: t2d full workflow
# =============================================================================

test_that("full t2d pipeline runs without error and produces analysis-ready data", {
  d <- data.table::copy(DT)

  suppressMessages({
    # 1. Clean missing
    d <- derive_missing(d, action = "na")

    # 2. Covariates
    d <- derive_covariate(d,
      as_factor  = c("p20116_i0", "p21000_i0"),
      as_numeric = "p22189",
      factor_levels = list(
        p20116_i0 = c("Never", "Previous", "Current")
      )
    )

    # 3. BMI tertiles
    d <- derive_cut(d, col = "p21001_i0", n = 3, name = "bmi_tri")

    # 4. T2D self-report
    d <- derive_selfreport(d, name = "t2d", regex = "type 2 diabetes",
                           field        = "noncancer",
                           disease_cols = SR_D, date_cols = SR_T,
                           visit_cols   = "p53_i0")

    # 5. T2D ICD-10 (HES only for speed)
    d <- derive_icd10(d, name = "t2d", icd10 = "E11",
                      source       = "hes",
                      hes_code_col = "p41270",
                      hes_date_cols = HES_DT)

    # 6. Unified case
    d <- derive_case(d, name = "t2d")

    # 7. Timing
    d <- derive_timing(d, name = "t2d", baseline_col = "p53_i0")

    # 8. Age at onset
    d <- derive_age(d, name = "t2d", baseline_col = "p53_i0", age_col = "p21022")

    # 9. Follow-up
    d <- derive_followup(d, name = "t2d",
                         event_col    = "t2d_date",
                         baseline_col = "p53_i0",
                         censor_date  = CENSOR,
                         death_col    = FALSE,
                         lost_col     = FALSE)
  })

  # Structural checks
  expected_cols <- c(
    "t2d_selfreport", "t2d_selfreport_date",
    "t2d_icd10",      "t2d_icd10_date",
    "t2d_status",     "t2d_date",
    "t2d_timing",     "age_at_t2d",
    "t2d_followup_end", "t2d_followup_years",
    "bmi_tri"
  )
  for (col in expected_cols) {
    expect_true(col %in% names(d), label = paste0("missing column: ", col))
  }

  # Quality checks
  expect_true(all(d$t2d_followup_years > 0, na.rm = TRUE))
  expect_true(all(d$t2d_timing %in% c(0L, 1L, 2L, NA_integer_)))
  expect_true(sum(d$t2d_status) >= sum(d$t2d_icd10))
  expect_true(sum(d$t2d_status) >= sum(d$t2d_selfreport))
  expect_true(data.table::is.data.table(d))
})

# =============================================================================
# ops.R — Operational utilities for ukbflow
# =============================================================================


#' Check the ukbflow operating environment
#'
#' Runs a four-block health check covering the dx CLI, dxpy (Python), RAP
#' authentication, and R package dependencies. Designed to be the first
#' function a new user runs after installation.
#'
#' The function is **read-only**: it never modifies system state, installs
#' packages, or authenticates. Auth failures are reported as warnings, not
#' errors, because the check itself does not require a live RAP connection.
#'
#' @param check_dx   (logical) Check dx CLI installation (dxpy is implied by dx).
#' @param check_auth (logical) Check RAP login status.
#' @param check_deps (logical) Check R package dependencies.
#' @param verbose    (logical) Print the formatted report. Set to `FALSE` for
#'   programmatic use (results are still returned invisibly).
#'
#' @return An invisible named list with elements `dx`, `dxpy`, `auth`, `deps`,
#'   and `summary`. Each element reflects the result of its respective check
#'   block and can be inspected programmatically.
#'
#' @export
#'
#' @examples
#' ops_setup(check_dx = FALSE, check_auth = FALSE)
#'
#' result <- ops_setup(check_dx = FALSE, check_auth = FALSE, verbose = FALSE)
#' result$summary$fail == 0
ops_setup <- function(
    check_dx   = TRUE,
    check_auth = TRUE,
    check_deps = TRUE,
    verbose    = TRUE
) {
  results <- list()

  r_ver <- paste0(R.version$major, ".", R.version$minor)

  if (verbose) {
    cli::cli_rule(left = "ukbflow environment check")
    cli::cli_inform(c(
      "i" = "ukbflow {utils::packageVersion('ukbflow')}  |  R {r_ver}  |  {Sys.Date()}"
    ))
  }

  # dx-toolkit
  if (check_dx) {
    if (verbose) cli::cli_rule(left = "1. dx-toolkit")

    dx <- .ops_check_dx()
    results$dx <- dx

    if (verbose) {
      if (dx$ok) {
        # Reason: dx CLI is a dxpy entry point — if dx works, dxpy is present.
        cli::cli_alert_success("dx: {.path {dx$path}}  ({dx$version})")
      } else {
        cli::cli_alert_danger(
          "dx not found. Install dxpy: {.code pip install dxpy}"
        )
      }
    }
  }

  # RAP authentication
  if (check_auth) {
    if (verbose) cli::cli_rule(left = "2. RAP authentication")

    auth <- .ops_check_auth()
    results$auth <- auth

    if (verbose) {
      if (auth$logged_in) {
        cli::cli_alert_success("user: {.val {auth$user}}")
        if (!is.na(auth$project)) {
          cli::cli_alert_success("project: {.val {auth$project}}")
        } else {
          cli::cli_alert_warning(
            "No project selected. Run {.fn auth_select_project}"
          )
        }
      } else if (!check_dx || results$dx$ok) {
        # Only warn about auth if dx itself is available
        cli::cli_alert_warning(
          "Not logged in. Run {.fn auth_login}"
        )
      } else {
        cli::cli_alert_warning("Skipped (dx not available).")
      }
    }
  }

  # R package dependencies
  if (check_deps) {
    if (verbose) cli::cli_rule(left = "3. R packages")

    deps <- .ops_check_deps()
    results$deps <- deps

    if (verbose) {
      for (d in deps) {
        if (d$installed) {
          cli::cli_alert_success(
            "{.pkg {d$package}}  {d$version}  [{d$group}]"
          )
        } else if (d$required) {
          cli::cli_alert_danger(
            "{.pkg {d$package}}  not installed  [{d$group}]"
          )
        } else {
          cli::cli_alert_warning(
            "{.pkg {d$package}}  not installed  [{d$group}] (optional)"
          )
        }
      }
    }
  }

  # ── Summary ───────────────────────────────────────────────────────────────
  n_pass <- 0L
  n_warn <- 0L
  n_fail <- 0L

  if (check_dx) {
    if (results$dx$ok) n_pass <- n_pass + 1L else n_fail <- n_fail + 1L
  }
  if (check_auth) {
    if (results$auth$logged_in) {
      n_pass <- n_pass + 1L
    } else {
      # Auth failure = warning, not a hard fail (tools may still be usable)
      n_warn <- n_warn + 1L
    }
  }
  if (check_deps) {
    for (d in results$deps) {
      if (d$installed) {
        n_pass <- n_pass + 1L
      } else if (d$required) {
        n_fail <- n_fail + 1L
      } else {
        n_warn <- n_warn + 1L
      }
    }
  }

  results$summary <- list(pass = n_pass, warn = n_warn, fail = n_fail)

  if (verbose) {
    cli::cli_rule()
    msg <- character(0)
    msg <- c(msg, "v" = "{n_pass} passed")
    if (n_warn > 0L) msg <- c(msg, "!" = "{n_warn} optional / warning")
    if (n_fail > 0L) msg <- c(msg, "x" = "{n_fail} required missing")
    cli::cli_inform(msg)

    # Suggest install command for missing required R packages
    if (check_deps && n_fail > 0L) {
      missing_pkgs <- Filter(
        function(d) !d$installed && d$required, results$deps
      )
      if (length(missing_pkgs) > 0L) {
        pkg_str <- paste0(
          '"', paste(vapply(missing_pkgs, `[[`, "", "package"), collapse = '", "'), '"'
        )
        cli::cli_inform(c(
          "i" = "Fix with: {.code install.packages(c({pkg_str}))}"
        ))
      }
    }
  }

  invisible(results)
}


#' Generate toy UKB-like data for testing and development
#'
#' Creates a small, synthetic dataset that mimics the structure of UK Biobank
#' phenotype data on the RAP. Useful for developing and testing `derive_*`,
#' `assoc_*`, and `plot_*` functions without requiring real UKB data access.
#'
#' This dataset is entirely synthetic. Column names follow RAP conventions
#' (e.g. `p41270`, `p20002_i0_a0`).
#'
#' @param scenario (character) Data structure to generate:
#'   - `"cohort"`: wide participant-level table with raw UKB field columns for
#'     the full `derive_*` → `assoc_*` → `plot_*` pipeline.
#'   - `"association"`: analysis-ready table with covariates already as factors,
#'     BMI/TDI binned, and two pre-derived disease outcomes (`dm_*`, `htn_*`)
#'     including status, date, timing, and follow-up columns. Use this for
#'     `assoc_*` examples and testing without running the derive pipeline.
#'   - `"forest"`: association results table matching `assoc_coxph()` output,
#'     for testing `plot_forest()`. `n` = number of exposures (default 8).
#' @param n (integer) Number of participants (or exposures for `"forest"`).
#'   Default `1000L` for `"cohort"`, `2000L` for `"association"`, `8L` for
#'   `"forest"`.
#' @param seed (integer or NULL) Random seed for reproducibility. Pass `NULL`
#'   for a different dataset on every call. Default `42L`.
#'
#' @return A `data.table` with UKB-style column names. See Details for the
#'   columns included in each scenario.
#'
#' @details
#' ## scenario = "cohort"
#' Includes the following column groups:
#' - **Demographics**: `eid`, `p31`, `p34`, `p53_i0`, `p21022`
#' - **Covariates**: `p21001_i0`, `p20116_i0`, `p1558_i0`, `p21000_i0`,
#'   `p22189`, `p54_i0`
#' - **Genetic PCs**: `p22009_a1` – `p22009_a10`
#' - **Self-report disease**: `p20002_i0_a0–a4`, `p20008_i0_a0–a4`
#' - **Self-report cancer**: `p20001_i0_a0–a4`, `p20006_i0_a0–a4`
#' - **HES**: `p41270` (JSON array), `p41280_a0–a8`
#' - **Cancer registry**: `p40006_i0–i2`, `p40011_i0–i2`, `p40012_i0–i2`,
#'   `p40005_i0–i2`
#' - **Death registry**: `p40001_i0`, `p40002_i0_a0–a2`, `p40000_i0`
#' - **First occurrence**: `p131742`
#' - **GRS**: `grs_bmi`, `grs_raw`, `grs_finngen`
#' - **Messy columns**: `messy_allna`, `messy_empty`, `messy_label`
#'
#' ## scenario = "association"
#' Analysis-ready table. All derive inputs (raw arrays, HES JSON, registry
#' fields) are omitted; derive outputs are pre-computed with internally
#' consistent relationships:
#' - **Demographics**: `eid`, `p31` (factor), `p53_i0` (IDate), `p21022`
#' - **Covariates**: `p21001_i0`, `bmi_cat` (factor, derived from `p21001_i0`),
#'   `p20116_i0` (factor), `p1558_i0` (factor), `p21000_i0` (factor),
#'   `p22189`, `tdi_cat` (factor, derived from `p22189` quartiles), `p54_i0`
#'   (factor)
#' - **Genetic PCs**: `p22009_a1` – `p22009_a10`
#' - **GRS**: `grs_bmi` (continuous exposure)
#' - **DM outcome**: `dm_status`, `dm_date`, `dm_timing`, `dm_followup_end`,
#'   `dm_followup_years` (type 2 diabetes, ~12% prevalence)
#' - **HTN outcome**: `htn_status`, `htn_date`, `htn_timing`,
#'   `htn_followup_end`, `htn_followup_years` (hypertension, ~28% prevalence)
#'
#' Internal relationships guaranteed:
#' - `bmi_cat` is cut from `p21001_i0` (breaks 18.5 / 25 / 30)
#' - `tdi_cat` is cut from `p22189` quartiles
#' - `dm_date` is `NA` iff `dm_status = FALSE`
#' - `dm_timing`: 0 = no disease, 1 = prevalent, 2 = incident, `NA` = no date
#' - `dm_followup_years` is `NA` for prevalent cases (`dm_timing == 1`)
#'
#' @export
#'
#' @examples
#' # cohort: raw UKB-style columns, feed into derive pipeline
#' dt <- ops_toy(n = 100)
#' dt <- derive_missing(dt)
#'
#' # association: analysis-ready, feed directly into assoc_* functions
#' dt <- ops_toy(scenario = "association", n = 500)
#' dt <- dt[dm_timing != 1L]   # exclude prevalent cases
#'
#' # forest: results table for plot_forest()
#' dt <- ops_toy(scenario = "forest")
ops_toy <- function(
    scenario = "cohort",
    n        = 1000L,
    seed     = 42L
) {
  scenario <- match.arg(scenario, choices = c("cohort", "association", "forest"))

  # Reason: sensible defaults differ by scenario
  if (missing(n)) {
    n <- switch(scenario,
      cohort      = 1000L,
      association = 2000L,
      forest      = 8L
    )
  }
  .assert_count_min(n, 1L)
  n <- as.integer(n)
  if (!is.null(seed)) set.seed(as.integer(seed))

  dt <- switch(scenario,
    cohort      = .ops_toy_cohort(n),
    association = .ops_toy_association(n),
    forest      = .ops_toy_forest(n)
  )

  seed_info <- if (!is.null(seed)) paste0(" | seed = ", seed) else ""
  unit <- if (scenario == "forest") "rows" else "participants"
  cli::cli_alert_success(
    "ops_toy: {nrow(dt)} {unit} | {ncol(dt)} columns | scenario = {.val {scenario}}{seed_info}"
  )

  dt
}


#' Summarise missing values by column
#'
#' Scans each column of a data.frame or data.table and returns the count and
#' percentage of missing values. Results are sorted by missingness in
#' descending order. Columns above 10\% are flagged in red; those between 0\%
#' and 10\% in yellow. A summary block is always printed regardless of
#' `threshold`.
#'
#' **Missing value definition**: a value is counted as missing if it is `NA`
#' *or* an empty string (`""`). Empty strings are treated as missing because
#' UKB exports frequently use `""` as a placeholder for absent text values.
#' This means `n_na` and `pct_na` reflect *effective* missingness, not just
#' `is.na()`. Numeric and logical columns are not affected (they cannot hold
#' `""`).
#'
#' @param data A data.frame or data.table to scan.
#' @param threshold (numeric) Columns with `pct_na <= threshold` are silenced
#'   from the per-column CLI output. The summary block is always shown.
#'   Default `0`: every column with any missing value is listed.
#' @param verbose (logical) Print the CLI report. Default `TRUE`.
#'
#' @return An invisible data.table with columns `column`, `n_na`, and `pct_na`,
#'   sorted by `pct_na` descending. `n_na` counts both `NA` and `""`.
#'   Always contains all columns regardless of `threshold` (which only affects
#'   CLI output).
#'
#' @export
#'
#' @examples
#' dt <- ops_toy(n = 100)
#'
#' # Show all columns with any missing value
#' ops_na(dt)
#'
#' # Only list columns with > 10% missing in the CLI output
#' ops_na(dt, threshold = 10)
#'
#' # Programmatic use — retrieve result silently
#' result <- ops_na(dt, verbose = FALSE)
#' result[pct_na > 50]
ops_na <- function(data, threshold = 0, verbose = TRUE) {

  # ── Validation ──────────────────────────────────────────────────────────────
  .assert_data_frame(data)
  if (nrow(data) == 0L)
    cli::cli_abort("{.arg data} has 0 rows.", call = NULL)
  if (!is.numeric(threshold) || length(threshold) != 1L ||
      is.na(threshold) || threshold < 0 || threshold >= 100)
    cli::cli_abort("{.arg threshold} must be a single numeric value in [0, 100).", call = NULL)
  .assert_flag(verbose)

  n_row <- nrow(data)
  n_col <- ncol(data)
  dt    <- data.table::as.data.table(data)

  # ── Count NA + "" per column ──────────────────────────────────────────────
  # Reason: vapply over data.table columns is fast; is.character() guard avoids
  # unnecessary coercion overhead on numeric/integer/logical columns.
  n_na_vec <- vapply(dt, function(col) {
    sum(is.na(col) | (is.character(col) & !is.na(col) & col == ""))
  }, integer(1L))

  result <- data.table::data.table(
    column = names(n_na_vec),
    n_na   = n_na_vec,
    pct_na = round(n_na_vec / n_row * 100, 2)
  )
  data.table::setorder(result, -pct_na)

  # ── Summary counts (all columns, unaffected by threshold) ─────────────────
  n_high     <- sum(result$pct_na >= 10)
  n_mid      <- sum(result$pct_na > 0 & result$pct_na < 10)
  n_complete <- sum(result$pct_na == 0)

  # ── CLI output ─────────────────────────────────────────────────────────────
  if (verbose) {
    cli::cli_rule(left = "ops_na")
    cli::cli_inform(c(
      "i" = "{n_row} rows | {n_col} columns | threshold = {threshold}%"
    ))

    # Per-column lines: only for pct_na > threshold
    to_print <- result[pct_na > threshold]

    if (nrow(to_print) > 0L) {
      # Reason: pre-format widths so columns align neatly
      max_col_w <- max(nchar(to_print$column))
      max_n_w   <- nchar(as.character(n_row))

      for (i in seq_len(nrow(to_print))) {
        col_name <- formatC(to_print$column[i], width = max_col_w, flag = "-")
        n_str    <- formatC(to_print$n_na[i],   width = max_n_w,   format = "d")
        pct_str  <- formatC(to_print$pct_na[i], width = 6, format = "f", digits = 2)
        line     <- "{col_name}  {n_str} / {n_row}  ({pct_str}%)"

        if (to_print$pct_na[i] >= 10) {
          cli::cli_alert_danger(line)
        } else {
          cli::cli_alert_warning(line)
        }
      }
    }

    # Summary block — always shown
    cli::cli_rule()
    if (n_high > 0L) cli::cli_inform(c("x" = "{n_high} column{?s} \u2265 10% missing"))
    if (n_mid  > 0L) cli::cli_inform(c("!" = "{n_mid} column{?s} > 0% and < 10% missing"))
    cli::cli_inform(c("v" = "{n_complete} column{?s} complete (0% missing)"))
  }

  invisible(result)
}


#' Record and review dataset pipeline snapshots
#'
#' Captures a lightweight summary of a data.frame at a given pipeline stage
#' and stores it in the session cache. Subsequent calls automatically compute
#' deltas against the previous snapshot, making it easy to track how data
#' changes through `derive_*`, `assoc_*`, and other processing steps.
#'
#' @param data A data.frame or data.table to snapshot. Pass `NULL` (or omit)
#'   to print the full snapshot history without recording a new entry.
#' @param label (character) A short name for this pipeline stage, e.g.
#'   `"raw"`, `"after_derive_missing"`. Defaults to `"snapshot_N"` where N
#'   is the sequential index.
#' @param reset (logical) If `TRUE`, clears the entire snapshot history and
#'   returns invisibly. Default `FALSE`.
#' @param verbose (logical) Print the CLI report. Default `TRUE`.
#' @param check_na (logical) Whether to count columns with any `NA` or blank
#'   string values and include the delta in the report. Set to `FALSE` to skip
#'   the NA scan (useful for large datasets or when NA tracking is not needed).
#'   Default `TRUE`.
#'
#' @return When `data` is supplied, returns the new snapshot row invisibly
#'   (a one-row data.table). When called with no `data`, returns the full
#'   history data.table invisibly.
#'
#' @export
#'
#' @examples
#' dt <- ops_toy(n = 100)
#' ops_snapshot(dt, label = "raw")
#'
#' dt <- derive_missing(dt)
#' ops_snapshot(dt, label = "after_derive_missing")
#'
#' # View full history
#' ops_snapshot()
#'
#' # Reset history
#' ops_snapshot(reset = TRUE)
ops_snapshot <- function(data = NULL, label = NULL, reset = FALSE, verbose = TRUE,
                         check_na = TRUE) {

  # ── Validation ──────────────────────────────────────────────────────────────
  .assert_flag(reset)
  .assert_flag(verbose)
  .assert_flag(check_na)

  # ── Reset ───────────────────────────────────────────────────────────────────
  if (reset) {
    .ukbflow_cache$snapshots     <- NULL
    .ukbflow_cache$snapshot_cols <- NULL
    if (verbose) cli::cli_alert_success("Snapshot history cleared.")
    return(invisible(NULL))
  }

  # ── View history (no data supplied) ─────────────────────────────────────────
  if (is.null(data)) {
    history <- .ukbflow_cache$snapshots
    if (is.null(history) || nrow(history) == 0L) {
      cli::cli_alert_warning("No snapshots recorded yet.")
      return(invisible(NULL))
    }
    if (verbose) {
      cli::cli_rule(left = "ops_snapshot history")
      print(history)
      cli::cli_rule()
    }
    return(invisible(history))
  }

  # ── Record new snapshot ──────────────────────────────────────────────────────
  .assert_data_frame(data)
  if (!is.null(label)) .assert_scalar_string(label)

  history <- .ukbflow_cache$snapshots
  idx     <- if (is.null(history)) 1L else nrow(history) + 1L

  if (is.null(label)) label <- paste0("snapshot_", idx)

  # Reason: count NA + "" consistently with ops_na(); skippable via check_na = FALSE
  if (check_na) {
    dt_tmp    <- data.table::as.data.table(data)
    n_na_cols <- sum(vapply(dt_tmp, function(col) {
      any(is.na(col) | (is.character(col) & !is.na(col) & col == ""))
    }, logical(1L)))
  } else {
    n_na_cols <- NA_integer_
  }

  size_mb <- round(as.numeric(object.size(data)) / 1024^2, 2)

  new_row <- data.table::data.table(
    idx       = idx,
    label     = label,
    timestamp = format(Sys.time(), "%H:%M:%S"),
    nrow      = nrow(data),
    ncol      = ncol(data),
    n_na_cols = n_na_cols,
    size_mb   = size_mb
  )

  # Append to cache
  .ukbflow_cache$snapshots <- if (is.null(history)) {
    new_row
  } else {
    data.table::rbindlist(list(history, new_row))
  }

  # Store column names separately (negligible overhead, enables diff/drop workflows)
  if (is.null(.ukbflow_cache$snapshot_cols)) .ukbflow_cache$snapshot_cols <- list()
  .ukbflow_cache$snapshot_cols[[label]] <- data.table::copy(colnames(data))

  # ── CLI output ───────────────────────────────────────────────────────────────
  if (verbose) {
    prev <- if (idx > 1L) history[nrow(history)] else NULL

    cli::cli_rule(left = "snapshot: {label}")

    .fmt_delta <- function(curr, prev_val, unit = "") {
      curr_str <- paste0(formatC(curr, format = "fg", big.mark = ","), unit)
      if (is.null(prev_val) || is.na(prev_val) || is.na(curr)) return(curr_str)
      delta <- curr - prev_val
      delta_str <- if (delta == 0) {
        "(= 0)"
      } else if (delta > 0) {
        paste0("(+", formatC(delta, format = "fg", big.mark = ","), unit, ")")
      } else {
        paste0("(", formatC(delta, format = "fg", big.mark = ","), unit, ")")
      }
      paste0(curr_str, "  ", delta_str)
    }

    s_row  <- .fmt_delta(nrow(data),  prev$nrow,      "")
    s_col  <- .fmt_delta(ncol(data),  prev$ncol,      "")
    s_na   <- if (check_na) .fmt_delta(n_na_cols, prev$n_na_cols, "") else "(skipped)"
    s_size <- .fmt_delta(size_mb,     prev$size_mb,   " MB")

    # Reason: individual cli_inform calls avoid duplicate-key merging in cli
    cli::cli_inform("  rows      {s_row}")
    cli::cli_inform("  cols      {s_col}")
    cli::cli_inform("  NA cols   {s_na}")
    cli::cli_inform("  size      {s_size}")
    cli::cli_rule()
  }

  invisible(new_row)
}


#' Retrieve column names recorded at a snapshot
#'
#' Returns the column names stored by a previous \code{\link{ops_snapshot}}
#' call, optionally excluding columns you wish to keep.
#'
#' @param label (character) Snapshot label passed to \code{ops_snapshot()}.
#' @param keep (character or NULL) Column names to exclude from the returned
#'   vector (i.e. columns to retain in the data even if they were present at
#'   that snapshot). Default \code{NULL}.
#'
#' @return A character vector of column names.
#' @export
#'
#' @examples
#' dt <- ops_toy(n = 100)
#' ops_snapshot(dt, label = "raw")
#' ops_snapshot_cols("raw")
#' ops_snapshot_cols("raw", keep = "eid")
ops_snapshot_cols <- function(label, keep = NULL) {

  .assert_scalar_string(label)
  if (!is.null(keep)) .assert_character(keep)

  cols <- .ukbflow_cache$snapshot_cols[[label]]
  if (is.null(cols))
    cli::cli_abort("No snapshot found with label {.val {label}}.", call = NULL)

  # Always protect built-in safe cols + user-registered safe cols + explicit keep
  builtin_safe <- c("eid", "sex", "age", "age_at_recruitment")
  user_safe    <- .ukbflow_cache$safe_cols
  protected    <- unique(c(builtin_safe, user_safe, keep))

  setdiff(cols, protected)
}


#' Register additional safe columns protected from snapshot-based drops
#'
#' Adds column names to the session-level safe list. Columns in this list are
#' automatically excluded when \code{\link{ops_snapshot_cols}} is used to
#' build a drop vector, in addition to the built-in protected columns
#' (\code{"eid"}, \code{"sex"}, \code{"age"}, \code{"age_at_recruitment"}).
#'
#' @param cols (character) One or more column names to protect.
#' @param reset (logical) If \code{TRUE}, clear the current user-registered
#'   safe list before adding. Default \code{FALSE}.
#'
#' @return Invisibly returns the updated safe cols vector.
#' @export
#'
#' @examples
#' ops_set_safe_cols(c("date_baseline", "townsend_index"))
#' ops_set_safe_cols(reset = TRUE)  # clear user-registered safe cols
ops_set_safe_cols <- function(cols = NULL, reset = FALSE) {

  .assert_flag(reset)

  if (reset) {
    .ukbflow_cache$safe_cols <- NULL
    cli::cli_alert_success("User-registered safe cols cleared.")
  }

  if (!is.null(cols)) {
    if (!is.character(cols))
      cli::cli_abort("{.arg cols} must be a character vector.", call = NULL)
    .ukbflow_cache$safe_cols <- unique(c(.ukbflow_cache$safe_cols, cols))
    cli::cli_alert_success("Safe cols registered: {.val {cols}}")
  }

  invisible(.ukbflow_cache$safe_cols)
}


#' Remove raw source columns recorded at a snapshot
#'
#' Drops columns that were present at snapshot \code{from} from \code{data},
#' while automatically protecting built-in safe columns
#' (\code{"eid"}, \code{"sex"}, \code{"age"}, \code{"age_at_recruitment"}) and
#' any user-registered safe columns set via \code{\link{ops_set_safe_cols}}.
#' Columns that no longer exist in \code{data} are silently skipped.
#'
#' @param data A data.frame or data.table.
#' @param from (character) Label of the snapshot whose columns should be
#'   dropped (typically \code{"raw"}).
#' @param keep (character or NULL) Additional column names to protect beyond
#'   the built-in and user-registered safe cols. Default \code{NULL}.
#' @param verbose (logical) Print a summary of dropped columns. Default
#'   \code{TRUE}.
#'
#' @return A \code{data.table} with the specified columns removed. For
#'   \code{data.table} input the operation is performed by reference (in-place);
#'   for \code{data.frame} input the data is first converted to a new
#'   \code{data.table} — the original \code{data.frame} is not modified.
#' @export
#'
#' @examples
#' dt <- ops_toy(n = 100)
#' ops_snapshot(dt, label = "raw")
#' dt <- derive_missing(dt)
#' ops_snapshot(dt, label = "derived")
#' ops_snapshot_diff("raw", "derived")
#' dt <- ops_snapshot_remove(dt, from = "raw")
ops_snapshot_remove <- function(data, from, keep = NULL, verbose = TRUE) {

  .assert_data_frame(data)
  .assert_scalar_string(from)

  if (!data.table::is.data.table(data)) data <- data.table::as.data.table(data)

  # Build protected set
  builtin_safe <- c("eid", "sex", "age", "age_at_recruitment")
  user_safe    <- .ukbflow_cache$safe_cols
  protected    <- unique(c(builtin_safe, user_safe, keep))

  # Columns to drop: in snapshot but not protected and still present in data
  snap_cols    <- ops_snapshot_cols(from, keep = keep)  # excludes built-in safe + user safe + keep
  cols_to_drop <- intersect(snap_cols, names(data))

  if (length(cols_to_drop) == 0L) {
    if (verbose) cli::cli_alert_info("ops_snapshot_remove: no columns to drop.")
    return(invisible(data))
  }

  data[, (cols_to_drop) := NULL]

  if (verbose) {
    cli::cli_alert_success(
      "ops_snapshot_remove: dropped {length(cols_to_drop)} raw column{?s}, {ncol(data)} remaining."
    )
  }

  invisible(data)
}


#' Compare column names between two snapshots
#'
#' Returns lists of columns added and removed between two recorded snapshots.
#'
#' @param label1 (character) Label of the earlier snapshot.
#' @param label2 (character) Label of the later snapshot.
#'
#' @return A named list with two character vectors: \code{added} (columns
#'   present in \code{label2} but not \code{label1}) and \code{removed}
#'   (columns present in \code{label1} but not \code{label2}).
#' @export
#'
#' @examples
#' dt <- ops_toy(n = 100)
#' ops_snapshot(dt, label = "raw")
#' dt <- derive_missing(dt)
#' ops_snapshot(dt, label = "derived")
#' ops_snapshot_diff("raw", "derived")
#' # $added   — newly derived columns
#' # $removed — columns dropped between snapshots
ops_snapshot_diff <- function(label1, label2) {

  .assert_scalar_string(label1)
  .assert_scalar_string(label2)

  cols1 <- .ukbflow_cache$snapshot_cols[[label1]]
  cols2 <- .ukbflow_cache$snapshot_cols[[label2]]

  if (is.null(cols1))
    cli::cli_abort("No snapshot found with label {.val {label1}}.", call = NULL)
  if (is.null(cols2))
    cli::cli_abort("No snapshot found with label {.val {label2}}.", call = NULL)

  result <- list(
    added   = setdiff(cols2, cols1),
    removed = setdiff(cols1, cols2)
  )

  cli::cli_inform("Columns added ({length(result$added)}):   {.val {result$added}}")
  cli::cli_inform("Columns removed ({length(result$removed)}): {.val {result$removed}}")

  invisible(result)
}


#' Exclude withdrawn participants from a dataset
#'
#' Reads a UK Biobank withdrawal list (a headerless single-column CSV of
#' anonymised participant IDs) and removes the corresponding rows from
#' `data`. A pair of [ops_snapshot()] calls is made automatically so the
#' before/after row counts are recorded in the session snapshot history.
#'
#' @param data A data.frame or data.table containing a participant ID column.
#' @param file (character) Path to the UKB withdrawal CSV file. The file must
#'   be a single-column, **header-free** CSV as supplied by UK Biobank
#'   (e.g. `w854944_20260310.csv`).
#' @param eid_col (character) Name of the participant ID column in `data`.
#'   Default `"eid"`.
#' @param verbose (logical) Print the CLI report. Default `TRUE`.
#'
#' @return A data.table with withdrawn participants removed.
#'
#' @export
#'
#' @examples
#' dt <- ops_toy(n = 100)
#' withdraw_file <- tempfile(fileext = ".csv")
#' writeLines(as.character(dt$eid[1:5]), withdraw_file)
#' dt <- ops_withdraw(dt, file = withdraw_file)
ops_withdraw <- function(data, file, eid_col = "eid", verbose = TRUE) {

  # ── Validation ──────────────────────────────────────────────────────────────
  .assert_data_frame(data)
  .assert_scalar_string(file)
  if (!file.exists(file))
    cli::cli_abort("Withdrawal file not found: {.path {file}}", call = NULL)
  .assert_scalar_string(eid_col)
  if (!eid_col %in% names(data))
    cli::cli_abort("Column {.val {eid_col}} not found in {.arg data}.", call = NULL)
  .assert_flag(verbose)

  dt <- data.table::as.data.table(data)

  # ── Read withdrawal list ─────────────────────────────────────────────────────
  # Reason: UKB withdrawal files are headerless single-column CSVs; fread
  # auto-detects integer type so no coercion is needed.
  withdraw_ids <- data.table::fread(file, header = FALSE)[[1L]]
  n_withdraw   <- length(withdraw_ids)

  # ── Snapshot before ──────────────────────────────────────────────────────────
  ops_snapshot(dt, label = "before_withdraw", verbose = verbose, check_na = FALSE)

  # ── Exclude ──────────────────────────────────────────────────────────────────
  n_found <- sum(dt[[eid_col]] %in% withdraw_ids)
  dt      <- dt[!get(eid_col) %in% withdraw_ids]

  # ── Snapshot after ───────────────────────────────────────────────────────────
  ops_snapshot(dt, label = "after_withdraw", verbose = verbose, check_na = FALSE)

  # ── Extra CLI summary ────────────────────────────────────────────────────────
  if (verbose) {
    fname <- basename(file)
    cli::cli_inform(c(
      "i" = "Withdrawal file: {.file {fname}} ({n_withdraw} IDs)",
      "x" = "Excluded: {n_found} participant{?s} found in data",
      "v" = "Remaining: {format(nrow(dt), big.mark = ',')} participants"
    ))
  }

  dt
}

# =============================================================================
# test-ops.R — Unit tests for ops_setup() and ops_toy()
#   Focus: input validation, output structure, reproducibility
#   No network, no dx-toolkit required
# =============================================================================


# ===========================================================================
# ops_toy() — input validation
# ===========================================================================

test_that("ops_toy() stops when scenario is invalid", {
  expect_error(ops_toy(scenario = "invalid"), "arg")
})

test_that("ops_toy() stops when n < 1", {
  expect_error(suppressMessages(ops_toy(n = 0)), "integer")
})

test_that("ops_toy() stops when n is negative", {
  expect_error(suppressMessages(ops_toy(n = -5)), "integer")
})

test_that("ops_toy() rejects non-integer numeric n", {
  expect_error(suppressMessages(ops_toy(n = 10.9, seed = 1)), "integer")
})


# ===========================================================================
# ops_toy(scenario = "cohort") — output structure
# ===========================================================================

test_that("ops_toy() cohort returns a data.table", {
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  expect_true(data.table::is.data.table(dt))
})

test_that("ops_toy() cohort returns correct number of rows", {
  dt <- suppressMessages(ops_toy(n = 200, seed = 1))
  expect_equal(nrow(dt), 200L)
})

test_that("ops_toy() cohort has eid column", {
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  expect_true("eid" %in% names(dt))
})

test_that("ops_toy() cohort has demographics columns", {
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  expect_true(all(c("p31", "p34", "p53_i0", "p21022") %in% names(dt)))
})

test_that("ops_toy() cohort has covariate columns", {
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  expect_true(all(c("p21001_i0", "p20116_i0", "p1558_i0",
                     "p21000_i0", "p22189", "p54_i0") %in% names(dt)))
})

test_that("ops_toy() cohort has 10 genetic PC columns", {
  dt   <- suppressMessages(ops_toy(n = 50, seed = 1))
  pcs  <- grep("^p22009_a", names(dt), value = TRUE)
  expect_equal(length(pcs), 10L)
})

test_that("ops_toy() cohort has self-report disease columns (p20002_i0_a0~a4)", {
  dt   <- suppressMessages(ops_toy(n = 50, seed = 1))
  cols <- paste0("p20002_i0_a", 0:4)
  expect_true(all(cols %in% names(dt)))
})

test_that("ops_toy() cohort has self-report date columns (p20008_i0_a0~a4)", {
  dt   <- suppressMessages(ops_toy(n = 50, seed = 1))
  cols <- paste0("p20008_i0_a", 0:4)
  expect_true(all(cols %in% names(dt)))
})

test_that("ops_toy() cohort has HES JSON column p41270", {
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  expect_true("p41270" %in% names(dt))
})

test_that("ops_toy() cohort has HES date columns (p41280_a0~a8)", {
  dt   <- suppressMessages(ops_toy(n = 50, seed = 1))
  cols <- paste0("p41280_a", 0:8)
  expect_true(all(cols %in% names(dt)))
})

test_that("ops_toy() cohort has self-report cancer columns (p20001_i0_a0~a4 and p20006_i0_a0~a4)", {
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  expect_true(all(paste0("p20001_i0_a", 0:4) %in% names(dt)))
  expect_true(all(paste0("p20006_i0_a", 0:4) %in% names(dt)))
})

test_that("ops_toy() cohort has cancer registry columns (3 instances, icd/hist/behv/date)", {
  dt   <- suppressMessages(ops_toy(n = 50, seed = 1))
  cols <- c(paste0("p40006_i", 0:2), paste0("p40011_i", 0:2),
            paste0("p40012_i", 0:2), paste0("p40005_i", 0:2))
  expect_true(all(cols %in% names(dt)))
})

test_that("ops_toy() cohort has death registry columns (primary + 3 secondary + date)", {
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  expect_true(all(c("p40001_i0", "p40000_i0",
                    "p40002_i0_a0", "p40002_i0_a1", "p40002_i0_a2") %in% names(dt)))
})

test_that("ops_toy() cohort has first occurrence column p131742", {
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  expect_true("p131742" %in% names(dt))
})

test_that("ops_toy() cohort has GRS columns", {
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  expect_true(all(c("grs_bmi", "grs_raw", "grs_finngen") %in% names(dt)))
})

test_that("ops_toy() cohort has messy quality-testing columns", {
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  expect_true(all(c("messy_allna", "messy_empty", "messy_label") %in% names(dt)))
})

test_that("ops_toy() cohort messy_allna is entirely NA", {
  dt <- suppressMessages(ops_toy(n = 100, seed = 1))
  expect_true(all(is.na(dt$messy_allna)))
})

test_that("ops_toy() cohort p41270 non-NA values are valid JSON arrays", {
  dt   <- suppressMessages(ops_toy(n = 100, seed = 1))
  vals <- dt$p41270[!is.na(dt$p41270)]
  if (length(vals) > 0L) {
    expect_true(all(grepl('^\\[".+\\"]$', vals)))
  }
})

test_that("ops_toy() cohort eid values are unique", {
  dt <- suppressMessages(ops_toy(n = 200, seed = 1))
  expect_equal(length(unique(dt$eid)), 200L)
})

test_that("ops_toy() cohort BMI values are within plausible range", {
  dt <- suppressMessages(ops_toy(n = 500, seed = 1))
  expect_true(all(dt$p21001_i0 >= 12 & dt$p21001_i0 <= 65))
})


# ===========================================================================
# ops_toy(scenario = "forest") — output structure
# ===========================================================================

test_that("ops_toy() forest returns a data.table", {
  dt <- suppressMessages(ops_toy(scenario = "forest", seed = 1))
  expect_true(data.table::is.data.table(dt))
})

test_that("ops_toy() forest has required assoc_coxph-style columns", {
  dt   <- suppressMessages(ops_toy(scenario = "forest", seed = 1))
  cols <- c("exposure", "term", "model", "n", "n_events",
            "person_years", "HR", "CI_lower", "CI_upper",
            "p_value", "HR_label")
  expect_true(all(cols %in% names(dt)))
})

test_that("ops_toy() forest model column is an ordered factor", {
  dt <- suppressMessages(ops_toy(scenario = "forest", seed = 1))
  expect_true(is.factor(dt$model))
  expect_true(is.ordered(dt$model))
})

test_that("ops_toy() forest model factor has 3 levels", {
  dt <- suppressMessages(ops_toy(scenario = "forest", seed = 1))
  expect_equal(nlevels(dt$model), 3L)
})

test_that("ops_toy() forest default n=8 produces 8 exposures × 3 models = 24 rows", {
  dt <- suppressMessages(ops_toy(scenario = "forest", seed = 1))
  expect_equal(nrow(dt), 24L)
})

test_that("ops_toy() forest n=3 produces 3 exposures × 3 models = 9 rows", {
  dt <- suppressMessages(ops_toy(scenario = "forest", n = 3, seed = 1))
  expect_equal(nrow(dt), 9L)
})

test_that("ops_toy() forest HR values are positive", {
  dt <- suppressMessages(ops_toy(scenario = "forest", seed = 1))
  expect_true(all(dt$HR > 0))
  expect_true(all(dt$CI_lower > 0))
  expect_true(all(dt$CI_upper > 0))
})

test_that("ops_toy() forest CI_lower < HR < CI_upper for all rows", {
  dt <- suppressMessages(ops_toy(scenario = "forest", seed = 1))
  expect_true(all(dt$CI_lower < dt$HR))
  expect_true(all(dt$HR < dt$CI_upper))
})

test_that("ops_toy() forest p_value is in [0, 1]", {
  dt <- suppressMessages(ops_toy(scenario = "forest", seed = 1))
  expect_true(all(dt$p_value >= 0 & dt$p_value <= 1))
})

test_that("ops_toy() forest HR_label matches 'x.xx (x.xx-x.xx)' pattern", {
  dt <- suppressMessages(ops_toy(scenario = "forest", seed = 1))
  expect_true(all(grepl("^[0-9.]+\\s\\([0-9.]+-[0-9.]+\\)$", dt$HR_label)))
})


# ===========================================================================
# ops_toy() — reproducibility and seed handling
# ===========================================================================

test_that("ops_toy() produces identical output with same seed", {
  dt1 <- suppressMessages(ops_toy(n = 100, seed = 42))
  dt2 <- suppressMessages(ops_toy(n = 100, seed = 42))
  expect_identical(dt1, dt2)
})

test_that("ops_toy() produces different output with different seeds", {
  dt1 <- suppressMessages(ops_toy(n = 100, seed = 1))
  dt2 <- suppressMessages(ops_toy(n = 100, seed = 2))
  expect_false(identical(dt1$p21001_i0, dt2$p21001_i0))
})

test_that("ops_toy() with seed=NULL does not error", {
  expect_no_error(suppressMessages(ops_toy(n = 50, seed = NULL)))
})


# ===========================================================================
# ops_setup() — output structure (no dx / no RAP required)
# ===========================================================================

test_that("ops_setup() returns an invisible list", {
  result <- suppressMessages(ops_setup(verbose = FALSE))
  expect_true(is.list(result))
})

test_that("ops_setup() result has a summary element", {
  result <- suppressMessages(ops_setup(verbose = FALSE))
  expect_true("summary" %in% names(result))
})

test_that("ops_setup() summary has pass / warn / fail fields", {
  result  <- suppressMessages(ops_setup(verbose = FALSE))
  summary <- result$summary
  expect_true(all(c("pass", "warn", "fail") %in% names(summary)))
})

test_that("ops_setup() summary counts are non-negative integers", {
  result  <- suppressMessages(ops_setup(verbose = FALSE))
  summary <- result$summary
  expect_true(summary$pass >= 0L)
  expect_true(summary$warn >= 0L)
  expect_true(summary$fail >= 0L)
})

test_that("ops_setup() check_deps=FALSE omits deps from result", {
  result <- suppressMessages(ops_setup(check_deps = FALSE, verbose = FALSE))
  expect_null(result$deps)
})

test_that("ops_setup() check_dx=FALSE omits dx from result", {
  result <- suppressMessages(ops_setup(check_dx = FALSE, verbose = FALSE))
  expect_null(result$dx)
})

test_that("ops_setup() check_auth=FALSE omits auth from result", {
  result <- suppressMessages(ops_setup(check_auth = FALSE, verbose = FALSE))
  expect_null(result$auth)
})

test_that("ops_setup() deps result is a list of named lists", {
  result <- suppressMessages(
    ops_setup(check_dx = FALSE, check_auth = FALSE, verbose = FALSE)
  )
  deps <- result$deps
  expect_true(is.list(deps))
  expect_true(all(vapply(deps, function(d) "package" %in% names(d), logical(1))))
  expect_true(all(vapply(deps, function(d) "installed" %in% names(d), logical(1))))
})

test_that("ops_setup() known installed packages (cli, data.table) show installed=TRUE", {
  result <- suppressMessages(
    ops_setup(check_dx = FALSE, check_auth = FALSE, verbose = FALSE)
  )
  pkgs <- vapply(result$deps, `[[`, "", "package")
  installed <- vapply(result$deps, `[[`, TRUE, "installed")

  cli_ok        <- installed[pkgs == "cli"]
  datatable_ok  <- installed[pkgs == "data.table"]

  expect_true(length(cli_ok) > 0L && isTRUE(cli_ok))
  expect_true(length(datatable_ok) > 0L && isTRUE(datatable_ok))
})

test_that("ops_setup() verbose=FALSE produces no messages", {
  expect_no_message(
    ops_setup(check_dx = FALSE, check_auth = FALSE, verbose = FALSE)
  )
})


# ===========================================================================
# ops_setup() — branch behaviour via stubbed internal helpers
# ===========================================================================

test_that("ops_setup() dx not found → summary$fail >= 1", {
  local_mocked_bindings(
    .ops_check_dx   = function() list(ok = FALSE, path = NA_character_, version = NA_character_),
    .ops_check_auth = function() list(ok = FALSE, logged_in = FALSE,
                                      user = NA_character_, project = NA_character_),
    .package = "ukbflow"
  )
  result <- suppressMessages(ops_setup(check_deps = FALSE, verbose = FALSE))
  expect_gte(result$summary$fail, 1L)
})

test_that("ops_setup() dx found + auth ok → summary$pass >= 2, fail == 0", {
  local_mocked_bindings(
    .ops_check_dx   = function() list(ok = TRUE, path = "/usr/bin/dx", version = "0.350.0"),
    .ops_check_auth = function() list(ok = TRUE, logged_in = TRUE,
                                      user = "testuser", project = "project-abc123"),
    .package = "ukbflow"
  )
  result <- suppressMessages(ops_setup(check_deps = FALSE, verbose = FALSE))
  expect_gte(result$summary$pass, 2L)
  expect_equal(result$summary$fail, 0L)
})

test_that("ops_setup() dx found but not logged in → summary$warn >= 1, fail == 0", {
  local_mocked_bindings(
    .ops_check_dx   = function() list(ok = TRUE, path = "/usr/bin/dx", version = "0.350.0"),
    .ops_check_auth = function() list(ok = FALSE, logged_in = FALSE,
                                      user = NA_character_, project = NA_character_),
    .package = "ukbflow"
  )
  result <- suppressMessages(ops_setup(check_deps = FALSE, verbose = FALSE))
  expect_gte(result$summary$warn, 1L)
  expect_equal(result$summary$fail, 0L)   # auth failure is warn, not fail
})

test_that("ops_setup() auth ok but no project selected → summary$pass for auth, warning message", {
  local_mocked_bindings(
    .ops_check_dx   = function() list(ok = TRUE, path = "/usr/bin/dx", version = "0.350.0"),
    .ops_check_auth = function() list(ok = TRUE, logged_in = TRUE,
                                      user = "testuser", project = NA_character_),
    .package = "ukbflow"
  )
  result <- suppressMessages(ops_setup(check_deps = FALSE, verbose = FALSE))
  expect_gte(result$summary$pass, 2L)   # dx + auth both pass
})

test_that("ops_setup() summary pass + warn + fail sum matches checked items", {
  local_mocked_bindings(
    .ops_check_dx   = function() list(ok = TRUE, path = "/usr/bin/dx", version = "0.350.0"),
    .ops_check_auth = function() list(ok = TRUE, logged_in = TRUE,
                                      user = "testuser", project = "project-abc123"),
    .package = "ukbflow"
  )
  result <- suppressMessages(ops_setup(check_deps = FALSE, verbose = FALSE))
  s <- result$summary
  # With check_deps=FALSE: dx (pass) + auth (pass) = 2 items checked
  expect_equal(s$pass + s$warn + s$fail, 2L)
})


# ===========================================================================
# .ops_check_deps() — internal helper
# ===========================================================================

test_that(".ops_check_deps() returns a list", {
  result <- ukbflow:::.ops_check_deps()
  expect_true(is.list(result))
})

test_that(".ops_check_deps() each entry has required fields", {
  result <- ukbflow:::.ops_check_deps()
  for (d in result) {
    expect_true(all(c("package", "required", "group", "installed", "version") %in% names(d)))
  }
})

test_that(".ops_check_deps() installed is logical for each entry", {
  result <- ukbflow:::.ops_check_deps()
  for (d in result) {
    expect_true(is.logical(d$installed))
  }
})

test_that(".ops_check_deps() version is character or NA for each entry", {
  result <- ukbflow:::.ops_check_deps()
  for (d in result) {
    expect_true(is.na(d$version) || is.character(d$version))
  }
})


# ===========================================================================
# ops_na() — input validation
# ===========================================================================

test_that("ops_na() aborts on non-data.frame input", {
  expect_error(ops_na("not a df"), "data.frame")
})

test_that("ops_na() aborts on 0-row data", {
  expect_error(ops_na(data.frame(x = integer(0))), "0 rows")
})

test_that("ops_na() aborts when threshold < 0", {
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  expect_error(suppressMessages(ops_na(dt, threshold = -1)), "threshold")
})

test_that("ops_na() aborts when threshold >= 100", {
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  expect_error(suppressMessages(ops_na(dt, threshold = 100)), "threshold")
})

test_that("ops_na() aborts on invalid verbose", {
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  expect_error(suppressMessages(ops_na(dt, verbose = "yes")), "verbose")
})


# ===========================================================================
# ops_na() — output structure
# ===========================================================================

test_that("ops_na() returns a data.table invisibly", {
  dt     <- suppressMessages(ops_toy(n = 50, seed = 1))
  result <- suppressMessages(ops_na(dt, verbose = FALSE))
  expect_true(data.table::is.data.table(result))
})

test_that("ops_na() result has columns: column, n_na, pct_na", {
  dt     <- suppressMessages(ops_toy(n = 50, seed = 1))
  result <- suppressMessages(ops_na(dt, verbose = FALSE))
  expect_true(all(c("column", "n_na", "pct_na") %in% names(result)))
})

test_that("ops_na() returns one row per column in data", {
  dt     <- suppressMessages(ops_toy(n = 50, seed = 1))
  result <- suppressMessages(ops_na(dt, verbose = FALSE))
  expect_equal(nrow(result), ncol(dt))
})

test_that("ops_na() result is sorted by pct_na descending", {
  dt     <- suppressMessages(ops_toy(n = 50, seed = 1))
  result <- suppressMessages(ops_na(dt, verbose = FALSE))
  expect_true(all(diff(result$pct_na) <= 0))
})

test_that("ops_na() pct_na values are in [0, 100]", {
  dt     <- suppressMessages(ops_toy(n = 50, seed = 1))
  result <- suppressMessages(ops_na(dt, verbose = FALSE))
  expect_true(all(result$pct_na >= 0 & result$pct_na <= 100))
})

test_that("ops_na() n_na is non-negative integer", {
  dt     <- suppressMessages(ops_toy(n = 50, seed = 1))
  result <- suppressMessages(ops_na(dt, verbose = FALSE))
  expect_true(all(result$n_na >= 0L))
  expect_true(is.integer(result$n_na))
})

test_that("ops_na() messy_allna column shows 100% missing", {
  dt     <- suppressMessages(ops_toy(n = 100, seed = 1))
  result <- suppressMessages(ops_na(dt, verbose = FALSE))
  expect_equal(result[column == "messy_allna", pct_na], 100)
})

test_that("ops_na() complete numeric column shows 0% missing", {
  dt     <- suppressMessages(ops_toy(n = 100, seed = 1))
  result <- suppressMessages(ops_na(dt, verbose = FALSE))
  expect_equal(result[column == "p21001_i0", pct_na], 0)
})

test_that("ops_na() counts empty strings as missing", {
  dt     <- suppressMessages(ops_toy(n = 200, seed = 1))
  result <- suppressMessages(ops_na(dt, verbose = FALSE))
  # messy_empty contains "" and NA — both counted as missing
  expect_gt(result[column == "messy_empty", n_na], 0L)
})

test_that("ops_na() threshold does not affect returned data.table", {
  dt      <- suppressMessages(ops_toy(n = 100, seed = 1))
  res0    <- suppressMessages(ops_na(dt, threshold = 0,  verbose = FALSE))
  res50   <- suppressMessages(ops_na(dt, threshold = 50, verbose = FALSE))
  expect_equal(nrow(res0), nrow(res50))
  expect_identical(res0, res50)
})

test_that("ops_na() verbose=FALSE produces no messages", {
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  expect_no_message(ops_na(dt, verbose = FALSE))
})


# ===========================================================================
# ops_snapshot() — input validation
# ===========================================================================

test_that("ops_snapshot() aborts on non-data.frame data", {
  expect_error(ops_snapshot("not a df"), "data.frame")
})

test_that("ops_snapshot() aborts on invalid label", {
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  expect_error(suppressMessages(ops_snapshot(dt, label = 123)), "label")
})

test_that("ops_snapshot() aborts on empty string label", {
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  expect_error(suppressMessages(ops_snapshot(dt, label = "")), "label")
})

test_that("ops_snapshot() aborts on invalid verbose", {
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  expect_error(suppressMessages(ops_snapshot(dt, verbose = "yes")), "verbose")
})


# ===========================================================================
# ops_snapshot() — record and history
# ===========================================================================

test_that("ops_snapshot() reset clears history", {
  suppressMessages(ops_snapshot(reset = TRUE))
  suppressMessages(ops_snapshot(suppressMessages(ops_toy(n = 50, seed = 1)),
                                label = "x", verbose = FALSE))
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  history <- suppressMessages(ops_snapshot(verbose = FALSE))
  expect_null(history)
})

test_that("ops_snapshot() records a new row each call", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  suppressMessages(ops_snapshot(dt, label = "s1", verbose = FALSE))
  suppressMessages(ops_snapshot(dt, label = "s2", verbose = FALSE))
  history <- suppressMessages(ops_snapshot(verbose = FALSE))
  expect_equal(nrow(history), 2L)
})

test_that("ops_snapshot() result has expected columns", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  row <- suppressMessages(ops_snapshot(dt, label = "test", verbose = FALSE))
  expect_true(all(c("idx", "label", "timestamp",
                    "nrow", "ncol", "n_na_cols", "size_mb") %in% names(row)))
})

test_that("ops_snapshot() auto-labels as snapshot_N when label omitted", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt  <- suppressMessages(ops_toy(n = 50, seed = 1))
  row <- suppressMessages(ops_snapshot(dt, verbose = FALSE))
  expect_equal(row$label, "snapshot_1")
})

test_that("ops_snapshot() nrow and ncol match input data", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt  <- suppressMessages(ops_toy(n = 200, seed = 1))
  row <- suppressMessages(ops_snapshot(dt, label = "chk", verbose = FALSE))
  expect_equal(row$nrow, 200L)
  expect_equal(row$ncol, ncol(dt))
})

test_that("ops_snapshot() idx increments sequentially", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  suppressMessages(ops_snapshot(dt, label = "a", verbose = FALSE))
  suppressMessages(ops_snapshot(dt, label = "b", verbose = FALSE))
  suppressMessages(ops_snapshot(dt, label = "c", verbose = FALSE))
  history <- suppressMessages(ops_snapshot(verbose = FALSE))
  expect_equal(history$idx, 1:3)
})

test_that("ops_snapshot() verbose=FALSE produces no messages when recording", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  expect_no_message(ops_snapshot(dt, label = "quiet", verbose = FALSE))
})

test_that("ops_snapshot() warns when no history and no data supplied", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  expect_message(ops_snapshot(verbose = TRUE), "No snapshots")
})


# ===========================================================================
# ops_withdraw() — input validation
# ===========================================================================

test_that("ops_withdraw() aborts on non-data.frame input", {
  wfile <- withr::local_tempfile(fileext = ".csv")
  writeLines("1234567", wfile)
  expect_error(ops_withdraw("not a df", file = wfile), "data.frame")
})

test_that("ops_withdraw() aborts when file does not exist", {
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  expect_error(ops_withdraw(dt, file = "nonexistent.csv"), "not found")
})

test_that("ops_withdraw() aborts when eid_col not in data", {
  dt    <- suppressMessages(ops_toy(n = 50, seed = 1))
  wfile <- withr::local_tempfile(fileext = ".csv")
  writeLines("1234567", wfile)
  expect_error(
    suppressMessages(ops_withdraw(dt, file = wfile, eid_col = "wrong_col")),
    "wrong_col"
  )
})

test_that("ops_withdraw() aborts on invalid verbose", {
  dt    <- suppressMessages(ops_toy(n = 50, seed = 1))
  wfile <- withr::local_tempfile(fileext = ".csv")
  writeLines("1234567", wfile)
  expect_error(
    suppressMessages(ops_withdraw(dt, file = wfile, verbose = "yes")),
    "verbose"
  )
})


# ===========================================================================
# ops_withdraw() — expected use
# ===========================================================================

test_that("ops_withdraw() returns a data.table", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt    <- suppressMessages(ops_toy(n = 100, seed = 1))
  wfile <- withr::local_tempfile(fileext = ".csv")
  writeLines("9999999", wfile)  # eid not in toy data
  result <- suppressMessages(ops_withdraw(dt, file = wfile, verbose = FALSE))
  expect_true(data.table::is.data.table(result))
})

test_that("ops_withdraw() removes 0 rows when no eid matches", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt    <- suppressMessages(ops_toy(n = 100, seed = 1))
  wfile <- withr::local_tempfile(fileext = ".csv")
  writeLines("9999999", wfile)
  result <- suppressMessages(ops_withdraw(dt, file = wfile, verbose = FALSE))
  expect_equal(nrow(result), 100L)
})

test_that("ops_withdraw() removes matched eids correctly", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt         <- suppressMessages(ops_toy(n = 100, seed = 1))
  inject_ids <- dt$eid[1:3]
  wfile      <- withr::local_tempfile(fileext = ".csv")
  writeLines(as.character(inject_ids), wfile)
  result <- suppressMessages(ops_withdraw(dt, file = wfile, verbose = FALSE))
  expect_equal(nrow(result), 97L)
  expect_false(any(result$eid %in% inject_ids))
})

test_that("ops_withdraw() works with custom eid_col", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt <- suppressMessages(ops_toy(n = 100, seed = 1))
  data.table::setnames(dt, "eid", "participant_id")
  inject_ids <- dt$participant_id[1:2]
  wfile      <- withr::local_tempfile(fileext = ".csv")
  writeLines(as.character(inject_ids), wfile)
  result <- suppressMessages(
    ops_withdraw(dt, file = wfile, eid_col = "participant_id", verbose = FALSE)
  )
  expect_equal(nrow(result), 98L)
  expect_false(any(result$participant_id %in% inject_ids))
})

test_that("ops_withdraw() records two snapshots in session history", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt    <- suppressMessages(ops_toy(n = 100, seed = 1))
  wfile <- withr::local_tempfile(fileext = ".csv")
  writeLines("9999999", wfile)
  suppressMessages(ops_withdraw(dt, file = wfile, verbose = FALSE))
  history <- suppressMessages(ops_snapshot(verbose = FALSE))
  expect_equal(nrow(history), 2L)
  expect_equal(history$label, c("before_withdraw", "after_withdraw"))
})

test_that("ops_withdraw() verbose=FALSE produces no messages", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt    <- suppressMessages(ops_toy(n = 50, seed = 1))
  wfile <- withr::local_tempfile(fileext = ".csv")
  writeLines("9999999", wfile)
  expect_no_message(ops_withdraw(dt, file = wfile, verbose = FALSE))
})


# ===========================================================================
# ops_snapshot() — check_na = FALSE
# ===========================================================================

test_that("ops_snapshot() check_na=FALSE sets n_na_cols to NA", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt  <- suppressMessages(ops_toy(n = 50, seed = 1))
  row <- suppressMessages(
    ops_snapshot(dt, label = "nochk", verbose = FALSE, check_na = FALSE)
  )
  expect_true(is.na(row$n_na_cols))
})


# ===========================================================================
# ops_snapshot_cols() — input validation and expected use
# ===========================================================================

test_that("ops_snapshot_cols() aborts on non-string label", {
  expect_error(ops_snapshot_cols(123), "label")
})

test_that("ops_snapshot_cols() aborts when label not found in cache", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  expect_error(ops_snapshot_cols("nonexistent"), "nonexistent")
})

test_that("ops_snapshot_cols() returns a character vector", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  suppressMessages(ops_snapshot(dt, label = "raw", verbose = FALSE))
  cols <- ops_snapshot_cols("raw")
  expect_type(cols, "character")
  expect_gt(length(cols), 0L)
})

test_that("ops_snapshot_cols() always excludes built-in safe col eid", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  suppressMessages(ops_snapshot(dt, label = "raw", verbose = FALSE))
  expect_false("eid" %in% ops_snapshot_cols("raw"))
})

test_that("ops_snapshot_cols() keep excludes additional columns", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  suppressMessages(ops_snapshot(dt, label = "raw", verbose = FALSE))
  cols <- ops_snapshot_cols("raw", keep = "p31")
  expect_false("p31" %in% cols)
})

test_that("ops_snapshot_cols() keep=NULL returns all non-protected columns", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  suppressMessages(ops_snapshot(dt, label = "raw", verbose = FALSE))
  cols_no_keep   <- ops_snapshot_cols("raw")
  cols_with_keep <- ops_snapshot_cols("raw", keep = "p31")
  expect_gt(length(cols_no_keep), length(cols_with_keep))
})


# ===========================================================================
# ops_snapshot_diff() — input validation and expected use
# ===========================================================================

test_that("ops_snapshot_diff() aborts on non-string label1", {
  expect_error(suppressMessages(ops_snapshot_diff(123, "b")), "label1")
})

test_that("ops_snapshot_diff() aborts on non-string label2", {
  expect_error(suppressMessages(ops_snapshot_diff("a", 123)), "label2")
})

test_that("ops_snapshot_diff() aborts when label not found in cache", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  expect_error(suppressMessages(ops_snapshot_diff("raw", "derived")), "raw")
})

test_that("ops_snapshot_diff() returns list with added and removed", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt1 <- suppressMessages(ops_toy(n = 50, seed = 1))
  suppressMessages(ops_snapshot(dt1, label = "before", verbose = FALSE))
  dt2 <- data.table::copy(dt1)
  dt2[, derived_col := 1L]
  dt2[, p31 := NULL]
  suppressMessages(ops_snapshot(dt2, label = "after", verbose = FALSE))
  result <- suppressMessages(ops_snapshot_diff("before", "after"))
  expect_true("derived_col" %in% result$added)
  expect_true("p31" %in% result$removed)
})

test_that("ops_snapshot_diff() reports empty vectors when snapshots are identical", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  suppressMessages(ops_snapshot(dt, label = "s1", verbose = FALSE))
  suppressMessages(ops_snapshot(dt, label = "s2", verbose = FALSE))
  result <- suppressMessages(ops_snapshot_diff("s1", "s2"))
  expect_equal(length(result$added),   0L)
  expect_equal(length(result$removed), 0L)
})


# ===========================================================================
# ops_snapshot_remove() — input validation and expected use
# ===========================================================================

test_that("ops_snapshot_remove() aborts on non-data.frame input", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  suppressMessages(ops_snapshot(dt, label = "raw", verbose = FALSE))
  expect_error(ops_snapshot_remove("not a df", from = "raw"), "data.frame")
})

test_that("ops_snapshot_remove() aborts on non-string from", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  suppressMessages(ops_snapshot(dt, label = "raw", verbose = FALSE))
  expect_error(ops_snapshot_remove(dt, from = 123), "from")
})

test_that("ops_snapshot_remove() returns a data.table", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  suppressMessages(ops_snapshot(dt, label = "raw", verbose = FALSE))
  dt2 <- data.table::copy(dt)
  dt2[, derived_col := 1L]
  result <- suppressMessages(ops_snapshot_remove(dt2, from = "raw", verbose = FALSE))
  expect_true(data.table::is.data.table(result))
})

test_that("ops_snapshot_remove() drops raw columns and retains derived columns", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  suppressMessages(ops_snapshot(dt, label = "raw", verbose = FALSE))
  dt2 <- data.table::copy(dt)
  dt2[, derived_col := 1L]
  result <- suppressMessages(ops_snapshot_remove(dt2, from = "raw", verbose = FALSE))
  expect_true("derived_col" %in% names(result))
  expect_false("p31" %in% names(result))
})

test_that("ops_snapshot_remove() always protects eid", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  suppressMessages(ops_snapshot(dt, label = "raw", verbose = FALSE))
  result <- suppressMessages(ops_snapshot_remove(
    data.table::copy(dt), from = "raw", verbose = FALSE
  ))
  expect_true("eid" %in% names(result))
})

test_that("ops_snapshot_remove() keep protects additional columns", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt <- suppressMessages(ops_toy(n = 50, seed = 1))
  suppressMessages(ops_snapshot(dt, label = "raw", verbose = FALSE))
  result <- suppressMessages(ops_snapshot_remove(
    data.table::copy(dt), from = "raw", keep = "p31", verbose = FALSE
  ))
  expect_true("p31" %in% names(result))
})

test_that("ops_snapshot_remove() accepts data.frame input and returns data.table without modifying original", {
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  df <- as.data.frame(suppressMessages(ops_toy(n = 50, seed = 1)))
  suppressMessages(ops_snapshot(df, label = "raw", verbose = FALSE))
  result <- suppressMessages(ops_snapshot_remove(df, from = "raw", verbose = FALSE))
  expect_true(data.table::is.data.table(result))
  expect_false(data.table::is.data.table(df))
})


# ===========================================================================
# ops_set_safe_cols() — input validation and expected use
# ===========================================================================

test_that("ops_set_safe_cols() aborts on non-character cols", {
  expect_error(ops_set_safe_cols(cols = 123), "cols")
})

test_that("ops_set_safe_cols() aborts on invalid reset", {
  expect_error(ops_set_safe_cols(reset = "yes"), "reset")
})

test_that("ops_set_safe_cols() registered cols are protected in ops_snapshot_cols()", {
  suppressMessages(ops_set_safe_cols(reset = TRUE))
  suppressMessages(ops_set_safe_cols(cols = "my_safe_col"))
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt <- data.table::data.table(eid = 1L, my_safe_col = 1L, other_col = 1L)
  suppressMessages(ops_snapshot(dt, label = "raw", verbose = FALSE))
  cols <- ops_snapshot_cols("raw")
  expect_false("my_safe_col" %in% cols)
  expect_true("other_col"  %in% cols)
  suppressMessages(ops_set_safe_cols(reset = TRUE))
})

test_that("ops_set_safe_cols() reset clears registered cols", {
  suppressMessages(ops_set_safe_cols(cols = "tmp_col"))
  suppressMessages(ops_set_safe_cols(reset = TRUE))
  suppressMessages(ops_snapshot(reset = TRUE, verbose = FALSE))
  dt <- data.table::data.table(eid = 1L, tmp_col = 1L)
  suppressMessages(ops_snapshot(dt, label = "raw", verbose = FALSE))
  cols <- ops_snapshot_cols("raw")
  expect_true("tmp_col" %in% cols)
})

# =============================================================================
# utils.R — Package-wide internal helpers
# =============================================================================


# Null-coalescing operator: return lhs unless it is NULL, then return rhs.
`%||%` <- function(lhs, rhs) if (!is.null(lhs)) lhs else rhs


#' Assert that an argument is a single non-empty string
#'
#' @keywords internal
#' @noRd
.assert_scalar_string <- function(x, arg = deparse(substitute(x))) {
  if (!is.character(x) || length(x) != 1L || is.na(x) || !nzchar(x)) {
    cli::cli_abort("`{arg}` must be a single non-empty string.", call = NULL)
  }
  invisible(x)
}


#' Assert that an argument is a valid vector of integer-valued field IDs
#'
#' Rejects non-numeric input, non-whole numbers, empty vectors, and values
#' that are NA, Inf, or NaN. Returns deduplicated integer values ready for use.
#'
#' @param x The argument to check.
#' @param arg Name of the argument (for error messages).
#' @return Integer vector of unique values.
#'
#' @keywords internal
#' @noRd
.assert_integer_ids <- function(x, arg = deparse(substitute(x))) {
  if (!is.numeric(x) || length(x) == 0L) {
    cli::cli_abort("`{arg}` must be a non-empty numeric vector.", call = NULL)
  }
  if (anyNA(x) || any(is.infinite(x)) || any(is.nan(x))) {
    cli::cli_abort(
      "`{arg}` must not contain NA, Inf, or NaN values.",
      call = NULL
    )
  }
  if (any(x != floor(x))) {
    cli::cli_abort(
      "`{arg}` must contain whole numbers only (e.g. 31, not 31.7).",
      call = NULL
    )
  }
  as.integer(unique(x))
}


#' Assert that an argument is a valid DNAnexus job ID
#'
#' @param job_id The argument to check.
#' @param arg Name of the argument (for error messages).
#' @return Invisibly returns \code{job_id}.
#'
#' @keywords internal
#' @noRd
.assert_job_id <- function(job_id, arg = deparse(substitute(job_id))) {
  .assert_scalar_string(job_id, arg)
  if (!grepl("^job-", job_id)) {
    cli::cli_abort(
      "{.arg {arg}} must be a {.code job-XXXX} string, got {.val {job_id}}.",
      call = NULL
    )
  }
  invisible(job_id)
}


#' Assert that an argument is a single positive integer (count parameter)
#'
#' @param x The argument to check.
#' @param arg Name of the argument (for error messages).
#' @return Invisibly returns \code{as.integer(x)}.
#'
#' @keywords internal
#' @noRd
.assert_count <- function(x, arg = deparse(substitute(x))) {
  # Reason: is.finite() required because floor(Inf) == Inf, so without it
  # Inf would silently pass the floor check
  if (!is.numeric(x) || length(x) != 1L || is.na(x) ||
      !is.finite(x) || x < 1L || x != floor(x)) {
    cli::cli_abort("{.arg {arg}} must be a single positive integer.", call = NULL)
  }
  invisible(as.integer(x))
}


#' Assert that an argument is a single integer >= a specified minimum
#'
#' @param x The argument to check.
#' @param min Minimum allowed value (inclusive).
#' @param arg Name of the argument (for error messages).
#' @return Invisibly returns \code{as.integer(x)}.
#'
#' @keywords internal
#' @noRd
.assert_count_min <- function(x, min, arg = deparse(substitute(x))) {
  if (!is.numeric(x) || length(x) != 1L || is.na(x) ||
      !is.finite(x) || x < min || x != floor(x)) {
    cli::cli_abort("{.arg {arg}} must be a single integer >= {min}.", call = NULL)
  }
  invisible(as.integer(x))
}


#' Assert that an argument is NULL or a character vector of allowed values
#'
#' @param x The argument to check. \code{NULL} is always allowed.
#' @param choices Character vector of valid values.
#' @param arg Name of the argument (for error messages).
#' @return Invisibly returns \code{x}.
#'
#' @keywords internal
#' @noRd
.assert_choices <- function(x, choices, arg = deparse(substitute(x))) {
  if (is.null(x)) return(invisible(NULL))
  if (!is.character(x)) {
    cli::cli_abort(
      "{.arg {arg}} must be NULL or a character vector.", call = NULL
    )
  }
  # Reason: character(0) is allowed — downstream %in% returns all FALSE,
  # giving an empty result, which is predictable and not harmful
  bad <- setdiff(x, choices)
  if (length(bad) > 0) {
    cli::cli_abort(
      c("Invalid {.arg {arg}} value{?s}: {.val {bad}}.",
        "i" = "Valid values: {.val {choices}}."),
      call = NULL
    )
  }
  invisible(x)
}


#' Assert that an argument is a character vector
#'
#' @keywords internal
#' @noRd
.assert_character <- function(x, arg = deparse(substitute(x))) {
  if (!is.character(x)) {
    cli::cli_abort("{.arg {arg}} must be a character vector.", call = NULL)
  }
  invisible(x)
}


#' Assert that an argument is a data.frame
#'
#' @keywords internal
#' @noRd
.assert_data_frame <- function(x, arg = deparse(substitute(x))) {
  if (!is.data.frame(x)) {
    cli::cli_abort("{.arg {arg}} must be a data.frame.", call = NULL)
  }
  invisible(x)
}


#' Assert that an argument is a data.table
#'
#' @keywords internal
#' @noRd
.assert_data_table <- function(x, arg = deparse(substitute(x))) {
  if (!data.table::is.data.table(x)) {
    cli::cli_abort("{.arg {arg}} must be a data.table.", call = NULL)
  }
  invisible(x)
}


#' Assert that an argument is a logical vector
#'
#' @keywords internal
#' @noRd
.assert_logical <- function(x, arg = deparse(substitute(x))) {
  if (!is.logical(x)) {
    cli::cli_abort("{.arg {arg}} must be a logical vector.", call = NULL)
  }
  invisible(x)
}


#' Assert that an argument is a single non-NA logical (flag)
#'
#' @keywords internal
#' @noRd
.assert_flag <- function(x, arg = deparse(substitute(x))) {
  if (!is.logical(x) || length(x) != 1L || is.na(x)) {
    cli::cli_abort("{.arg {arg}} must be a single TRUE or FALSE.", call = NULL)
  }
  invisible(x)
}


#' Assert that a file path exists
#'
#' @keywords internal
#' @noRd
.assert_file_exists <- function(x, arg = deparse(substitute(x))) {
  if (!file.exists(x)) {
    cli::cli_abort("File not found: {.path {x}}", call = NULL)
  }
  invisible(x)
}


#' Assert that a value is non-NULL when a flag is FALSE
#'
#' Commonly used to enforce "when \code{base = FALSE}, \code{covariates} must
#' be supplied" contracts.
#'
#' @keywords internal
#' @noRd
.assert_not_null_if_false <- function(flag, value,
                                      flag_name  = deparse(substitute(flag)),
                                      value_name = deparse(substitute(value))) {
  if (!isTRUE(flag) && is.null(value)) {
    cli::cli_abort(
      "When {.arg {flag_name}} = FALSE, {.arg {value_name}} must be supplied.",
      call = NULL
    )
  }
  invisible(value)
}


#' Assert that a value is non-NULL when a flag is TRUE
#'
#' Commonly used to enforce "when \code{save = TRUE}, \code{dest} must
#' be supplied" contracts.
#'
#' @keywords internal
#' @noRd
.assert_not_null_if_true <- function(flag, value,
                                     flag_name  = deparse(substitute(flag)),
                                     value_name = deparse(substitute(value))) {
  if (isTRUE(flag) && is.null(value)) {
    cli::cli_abort(
      "When {.arg {flag_name}} = TRUE, {.arg {value_name}} must be supplied.",
      call = NULL
    )
  }
  invisible(value)
}


#' Assert that an argument is a numeric scalar strictly between 0 and 1
#'
#' @keywords internal
#' @noRd
.assert_proportion <- function(x, arg = deparse(substitute(x))) {
  if (!is.numeric(x) || length(x) != 1L || is.na(x) ||
      !is.finite(x) || x <= 0 || x >= 1) {
    cli::cli_abort("{.arg {arg}} must be a single number strictly between 0 and 1.", call = NULL)
  }
  invisible(x)
}


#' Assert that a vector has the expected length
#'
#' @keywords internal
#' @noRd
.assert_length_n <- function(x, n, arg = deparse(substitute(x))) {
  if (length(x) != n) {
    cli::cli_abort("{.arg {arg}} must have length {n}.", call = NULL)
  }
  invisible(x)
}


#' Assert that required columns are present in a data.frame
#'
#' Reports all missing columns in a single error rather than stopping at the
#' first, so the user can fix everything in one pass.
#'
#' @keywords internal
#' @noRd
.assert_has_cols <- function(data, cols, arg = deparse(substitute(data))) {
  missing <- setdiff(cols, names(data))
  if (length(missing) > 0L) {
    cli::cli_abort(
      "{.arg {arg}} is missing column{?s}: {.val {missing}}.",
      call = NULL
    )
  }
  invisible(data)
}


#' @keywords internal
#' @noRd
.assert_on_rap <- function() {
  if (!.is_on_rap())
    cli::cli_abort("This function must be run inside the RAP environment.", call = NULL)
  invisible(TRUE)
}

# =============================================================================
# test-utils.R — Unit tests for package-wide internal helpers (R/utils.R)
# =============================================================================


# ===========================================================================
# .assert_scalar_string()
# ===========================================================================

test_that(".assert_scalar_string() accepts a valid scalar string", {
  expect_invisible(ukbflow:::.assert_scalar_string("abc"))
})

test_that(".assert_scalar_string() rejects an empty string", {
  expect_error(ukbflow:::.assert_scalar_string(""), "single non-empty string")
})

test_that(".assert_scalar_string() rejects NA_character_", {
  expect_error(ukbflow:::.assert_scalar_string(NA_character_), "single non-empty string")
})

test_that(".assert_scalar_string() rejects non-character input", {
  expect_error(ukbflow:::.assert_scalar_string(1), "single non-empty string")
})

test_that(".assert_scalar_string() rejects length > 1 vector", {
  expect_error(ukbflow:::.assert_scalar_string(c("a", "b")), "single non-empty string")
})

test_that(".assert_scalar_string() error message includes the argument name", {
  my_token <- NA_character_
  expect_error(ukbflow:::.assert_scalar_string(my_token), "`my_token`")
})


# ===========================================================================
# .assert_integer_ids()
# ===========================================================================

test_that(".assert_integer_ids() accepts a valid integer vector and returns integers", {
  result <- ukbflow:::.assert_integer_ids(c(31, 53, 22189))
  expect_equal(result, c(31L, 53L, 22189L))
})

test_that(".assert_integer_ids() deduplicates values", {
  result <- ukbflow:::.assert_integer_ids(c(31, 31, 53))
  expect_equal(result, c(31L, 53L))
})

test_that(".assert_integer_ids() rejects empty vector", {
  expect_error(ukbflow:::.assert_integer_ids(integer(0)), "non-empty numeric")
})

test_that(".assert_integer_ids() rejects non-numeric input", {
  expect_error(ukbflow:::.assert_integer_ids("p31"), "non-empty numeric")
})

test_that(".assert_integer_ids() rejects NA values", {
  expect_error(ukbflow:::.assert_integer_ids(c(31, NA)), "NA")
})

test_that(".assert_integer_ids() rejects Inf values", {
  expect_error(ukbflow:::.assert_integer_ids(c(31, Inf)), "Inf")
})

test_that(".assert_integer_ids() rejects NaN values", {
  expect_error(ukbflow:::.assert_integer_ids(c(31, NaN)), "NaN")
})

test_that(".assert_integer_ids() rejects decimal values", {
  expect_error(ukbflow:::.assert_integer_ids(31.7), "whole numbers")
})

test_that(".assert_integer_ids() error message includes the argument name", {
  bad_ids <- c(31, NA)
  expect_error(ukbflow:::.assert_integer_ids(bad_ids), "bad_ids")
})


# ===========================================================================
# .assert_job_id()
# ===========================================================================

test_that(".assert_job_id() accepts a valid job-XXXX string", {
  expect_invisible(ukbflow:::.assert_job_id("job-XXXX"))
})

test_that(".assert_job_id() returns the job_id invisibly", {
  result <- ukbflow:::.assert_job_id("job-ABC123")
  expect_equal(result, "job-ABC123")
})

test_that(".assert_job_id() rejects wrong prefix", {
  expect_error(ukbflow:::.assert_job_id("notajob"), "job-XXXX")
})

test_that(".assert_job_id() rejects NA_character_", {
  expect_error(ukbflow:::.assert_job_id(NA_character_))
})

test_that(".assert_job_id() rejects length > 1 vector", {
  expect_error(ukbflow:::.assert_job_id(c("job-AAA", "job-BBB")))
})

test_that(".assert_job_id() rejects non-character input", {
  expect_error(ukbflow:::.assert_job_id(123))
})

test_that(".assert_job_id() rejects empty string", {
  expect_error(ukbflow:::.assert_job_id(""))
})

test_that(".assert_job_id() error message includes the argument name", {
  bad_id <- "notajob"
  expect_error(ukbflow:::.assert_job_id(bad_id), "bad_id")
})


# ===========================================================================
# .assert_count()
# ===========================================================================

test_that(".assert_count() accepts a valid positive integer", {
  expect_invisible(ukbflow:::.assert_count(20))
})

test_that(".assert_count() returns as.integer invisibly", {
  result <- ukbflow:::.assert_count(5)
  expect_equal(result, 5L)
  expect_type(result, "integer")
})

test_that(".assert_count() accepts 1", {
  expect_invisible(ukbflow:::.assert_count(1))
})

test_that(".assert_count() rejects 0", {
  expect_error(ukbflow:::.assert_count(0), "positive integer")
})

test_that(".assert_count() rejects negative value", {
  expect_error(ukbflow:::.assert_count(-5), "positive integer")
})

test_that(".assert_count() rejects Inf", {
  expect_error(ukbflow:::.assert_count(Inf), "positive integer")
})

test_that(".assert_count() rejects NA", {
  expect_error(ukbflow:::.assert_count(NA_real_), "positive integer")
})

test_that(".assert_count() rejects decimal value", {
  expect_error(ukbflow:::.assert_count(1.5), "positive integer")
})

test_that(".assert_count() rejects character input", {
  expect_error(ukbflow:::.assert_count("five"), "positive integer")
})

test_that(".assert_count() rejects length > 1 vector", {
  expect_error(ukbflow:::.assert_count(c(5, 10)), "positive integer")
})

test_that(".assert_count() error message includes the argument name", {
  bad_n <- -1
  expect_error(ukbflow:::.assert_count(bad_n), "bad_n")
})


# ===========================================================================
# .assert_choices()
# ===========================================================================

test_that(".assert_choices() accepts NULL and returns NULL invisibly", {
  result <- ukbflow:::.assert_choices(NULL, c("a", "b"))
  expect_null(result)
})

test_that(".assert_choices() accepts a valid single choice", {
  expect_invisible(ukbflow:::.assert_choices("a", c("a", "b", "c")))
})

test_that(".assert_choices() accepts valid multiple choices", {
  expect_invisible(ukbflow:::.assert_choices(c("a", "b"), c("a", "b", "c")))
})

test_that(".assert_choices() accepts character(0)", {
  expect_invisible(ukbflow:::.assert_choices(character(0), c("a", "b")))
})

test_that(".assert_choices() rejects non-character input", {
  expect_error(ukbflow:::.assert_choices(TRUE, c("a", "b")), "character vector")
})

test_that(".assert_choices() rejects invalid value", {
  expect_error(ukbflow:::.assert_choices("oops", c("a", "b")), "oops")
})

test_that(".assert_choices() error message lists valid values", {
  expect_error(ukbflow:::.assert_choices("oops", c("a", "b")), "a")
})

test_that(".assert_choices() rejects mix of valid and invalid values", {
  expect_error(ukbflow:::.assert_choices(c("a", "oops"), c("a", "b")), "oops")
})

test_that(".assert_choices() error message includes the argument name", {
  bad_state <- "oops"
  expect_error(ukbflow:::.assert_choices(bad_state, c("a", "b")), "bad_state")
})


# ===========================================================================
# %||%  (null-coalescing operator)
# ===========================================================================

test_that("%||% returns lhs when lhs is not NULL", {
  expect_equal("hello" %||% "fallback", "hello")
})

test_that("%||% returns rhs when lhs is NULL", {
  expect_equal(NULL %||% "fallback", "fallback")
})

test_that("%||% returns rhs when lhs is NULL, rhs is numeric", {
  expect_equal(NULL %||% 42L, 42L)
})

test_that("%||% returns lhs FALSE (not NULL) unchanged", {
  expect_equal(FALSE %||% TRUE, FALSE)
})

test_that("%||% returns lhs NA (not NULL) unchanged", {
  expect_equal(NA %||% "fallback", NA)
})


# ===========================================================================
# .assert_count_min()
# ===========================================================================

test_that(".assert_count_min() accepts integer equal to min", {
  expect_invisible(ukbflow:::.assert_count_min(2L, min = 2L))
})

test_that(".assert_count_min() returns as.integer invisibly", {
  result <- ukbflow:::.assert_count_min(3, min = 2L)
  expect_equal(result, 3L)
  expect_type(result, "integer")
})

test_that(".assert_count_min() accepts integer greater than min", {
  expect_invisible(ukbflow:::.assert_count_min(5L, min = 2L))
})

test_that(".assert_count_min() rejects value below min", {
  expect_error(ukbflow:::.assert_count_min(1L, min = 2L), ">= 2")
})

test_that(".assert_count_min() rejects decimal value", {
  expect_error(ukbflow:::.assert_count_min(2.5, min = 2L), ">= 2")
})

test_that(".assert_count_min() rejects NA", {
  expect_error(ukbflow:::.assert_count_min(NA_real_, min = 1L), ">= 1")
})

test_that(".assert_count_min() rejects Inf", {
  expect_error(ukbflow:::.assert_count_min(Inf, min = 1L), ">= 1")
})

test_that(".assert_count_min() rejects length > 1 vector", {
  expect_error(ukbflow:::.assert_count_min(c(2L, 3L), min = 2L), ">= 2")
})

test_that(".assert_count_min() error message includes the argument name", {
  bad_n <- 1L
  expect_error(ukbflow:::.assert_count_min(bad_n, min = 2L), "bad_n")
})


# ===========================================================================
# .assert_character()
# ===========================================================================

test_that(".assert_character() accepts a character scalar", {
  expect_invisible(ukbflow:::.assert_character("hello"))
})

test_that(".assert_character() accepts a character vector", {
  expect_invisible(ukbflow:::.assert_character(c("a", "b", "c")))
})

test_that(".assert_character() accepts character(0)", {
  expect_invisible(ukbflow:::.assert_character(character(0)))
})

test_that(".assert_character() rejects numeric input", {
  expect_error(ukbflow:::.assert_character(1L), "character vector")
})

test_that(".assert_character() rejects logical input", {
  expect_error(ukbflow:::.assert_character(TRUE), "character vector")
})

test_that(".assert_character() rejects NULL", {
  expect_error(ukbflow:::.assert_character(NULL), "character vector")
})

test_that(".assert_character() error message includes the argument name", {
  bad_labels <- 123L
  expect_error(ukbflow:::.assert_character(bad_labels), "bad_labels")
})


# ===========================================================================
# .assert_data_frame()
# ===========================================================================

test_that(".assert_data_frame() accepts a data.frame", {
  expect_invisible(ukbflow:::.assert_data_frame(data.frame(x = 1)))
})

test_that(".assert_data_frame() accepts a data.table (is also a data.frame)", {
  expect_invisible(
    ukbflow:::.assert_data_frame(data.table::data.table(x = 1))
  )
})

test_that(".assert_data_frame() rejects a plain list", {
  expect_error(ukbflow:::.assert_data_frame(list(x = 1)), "data.frame")
})

test_that(".assert_data_frame() rejects a character string", {
  expect_error(ukbflow:::.assert_data_frame("not a df"), "data.frame")
})

test_that(".assert_data_frame() rejects NULL", {
  expect_error(ukbflow:::.assert_data_frame(NULL), "data.frame")
})

test_that(".assert_data_frame() error message includes the argument name", {
  bad_data <- "oops"
  expect_error(ukbflow:::.assert_data_frame(bad_data), "bad_data")
})


# ===========================================================================
# .assert_data_table()
# ===========================================================================

test_that(".assert_data_table() accepts a data.table", {
  expect_invisible(
    ukbflow:::.assert_data_table(data.table::data.table(x = 1))
  )
})

test_that(".assert_data_table() rejects a plain data.frame", {
  expect_error(
    ukbflow:::.assert_data_table(data.frame(x = 1)), "data.table"
  )
})

test_that(".assert_data_table() rejects a list", {
  expect_error(ukbflow:::.assert_data_table(list(x = 1)), "data.table")
})

test_that(".assert_data_table() rejects NULL", {
  expect_error(ukbflow:::.assert_data_table(NULL), "data.table")
})

test_that(".assert_data_table() error message includes the argument name", {
  bad_dt <- data.frame(x = 1)
  expect_error(ukbflow:::.assert_data_table(bad_dt), "bad_dt")
})


# ===========================================================================
# .assert_has_cols()
# ===========================================================================

test_that(".assert_has_cols() passes when all columns present", {
  dt <- data.table::data.table(a = 1, b = 2, c = 3)
  expect_invisible(ukbflow:::.assert_has_cols(dt, c("a", "b")))
})

test_that(".assert_has_cols() returns data invisibly on success", {
  dt     <- data.table::data.table(a = 1, b = 2)
  result <- ukbflow:::.assert_has_cols(dt, "a")
  expect_identical(result, dt)
})

test_that(".assert_has_cols() aborts with single missing column name", {
  dt <- data.table::data.table(a = 1, b = 2)
  expect_error(ukbflow:::.assert_has_cols(dt, "missing_col"), "missing_col")
})

test_that(".assert_has_cols() reports all missing columns in one error", {
  dt  <- data.table::data.table(a = 1)
  err <- tryCatch(
    ukbflow:::.assert_has_cols(dt, c("b", "c")),
    error = function(e) conditionMessage(e)
  )
  expect_true(grepl("b", err))
  expect_true(grepl("c", err))
})

test_that(".assert_has_cols() passes with zero columns to check", {
  dt <- data.table::data.table(a = 1)
  expect_invisible(ukbflow:::.assert_has_cols(dt, character(0)))
})

test_that(".assert_has_cols() error message includes the data argument name", {
  my_data <- data.frame(x = 1)
  expect_error(ukbflow:::.assert_has_cols(my_data, "missing"), "my_data")
})


# ===========================================================================
# .assert_logical()
# ===========================================================================

test_that(".assert_logical() accepts TRUE", {
  expect_invisible(ukbflow:::.assert_logical(TRUE))
})

test_that(".assert_logical() accepts FALSE", {
  expect_invisible(ukbflow:::.assert_logical(FALSE))
})

test_that(".assert_logical() accepts a logical vector", {
  expect_invisible(ukbflow:::.assert_logical(c(TRUE, FALSE, TRUE)))
})

test_that(".assert_logical() accepts logical(0)", {
  expect_invisible(ukbflow:::.assert_logical(logical(0)))
})

test_that(".assert_logical() rejects character input", {
  expect_error(ukbflow:::.assert_logical("yes"), "logical")
})

test_that(".assert_logical() rejects numeric input", {
  expect_error(ukbflow:::.assert_logical(1L), "logical")
})

test_that(".assert_logical() rejects NULL", {
  expect_error(ukbflow:::.assert_logical(NULL), "logical")
})

test_that(".assert_logical() error message includes the argument name", {
  bad_flag <- "yes"
  expect_error(ukbflow:::.assert_logical(bad_flag), "bad_flag")
})


# ===========================================================================
# .assert_length_n()
# ===========================================================================

test_that(".assert_length_n() accepts vector of correct length", {
  expect_invisible(ukbflow:::.assert_length_n(c(1, 2, 3), 3L))
})

test_that(".assert_length_n() returns input invisibly", {
  x      <- c("a", "b")
  result <- ukbflow:::.assert_length_n(x, 2L)
  expect_identical(result, x)
})

test_that(".assert_length_n() accepts length-1 vector when n = 1", {
  expect_invisible(ukbflow:::.assert_length_n("only", 1L))
})

test_that(".assert_length_n() accepts empty vector when n = 0", {
  expect_invisible(ukbflow:::.assert_length_n(character(0), 0L))
})

test_that(".assert_length_n() rejects vector that is too short", {
  expect_error(ukbflow:::.assert_length_n(c(1, 2), 3L), "length 3")
})

test_that(".assert_length_n() rejects vector that is too long", {
  expect_error(ukbflow:::.assert_length_n(c(1, 2, 3, 4), 3L), "length 3")
})

test_that(".assert_length_n() error message includes the argument name", {
  bad_vec <- c(1, 2)
  expect_error(ukbflow:::.assert_length_n(bad_vec, 5L), "bad_vec")
})


# ===========================================================================
# .assert_flag()
# ===========================================================================

test_that(".assert_flag() accepts TRUE", {
  expect_invisible(ukbflow:::.assert_flag(TRUE))
})

test_that(".assert_flag() accepts FALSE", {
  expect_invisible(ukbflow:::.assert_flag(FALSE))
})

test_that(".assert_flag() rejects NA", {
  expect_error(ukbflow:::.assert_flag(NA), "TRUE or FALSE")
})

test_that(".assert_flag() rejects character", {
  expect_error(ukbflow:::.assert_flag("yes"), "TRUE or FALSE")
})

test_that(".assert_flag() rejects numeric", {
  expect_error(ukbflow:::.assert_flag(1L), "TRUE or FALSE")
})

test_that(".assert_flag() rejects length-2 logical", {
  expect_error(ukbflow:::.assert_flag(c(TRUE, FALSE)), "TRUE or FALSE")
})

test_that(".assert_flag() error message includes argument name", {
  my_flag <- "bad"
  expect_error(ukbflow:::.assert_flag(my_flag), "my_flag")
})


# ===========================================================================
# .assert_file_exists()
# ===========================================================================

test_that(".assert_file_exists() accepts an existing file", {
  tmp <- tempfile()
  writeLines("x", tmp)
  on.exit(unlink(tmp))
  expect_invisible(ukbflow:::.assert_file_exists(tmp))
})

test_that(".assert_file_exists() returns path invisibly", {
  tmp <- tempfile()
  writeLines("x", tmp)
  on.exit(unlink(tmp))
  result <- ukbflow:::.assert_file_exists(tmp)
  expect_equal(result, tmp)
})

test_that(".assert_file_exists() rejects a non-existent path", {
  expect_error(
    ukbflow:::.assert_file_exists("/no/such/file.csv"),
    "not found|File"
  )
})

test_that(".assert_file_exists() error message includes the path", {
  expect_error(
    ukbflow:::.assert_file_exists("/no/such/file.csv"),
    "/no/such/file.csv"
  )
})


# ===========================================================================
# .assert_not_null_if_false()
# ===========================================================================

test_that(".assert_not_null_if_false() passes when flag is TRUE and value is NULL", {
  expect_invisible(ukbflow:::.assert_not_null_if_false(TRUE, NULL))
})

test_that(".assert_not_null_if_false() passes when flag is FALSE and value is non-NULL", {
  expect_invisible(ukbflow:::.assert_not_null_if_false(FALSE, "covariates"))
})

test_that(".assert_not_null_if_false() returns value invisibly", {
  result <- ukbflow:::.assert_not_null_if_false(FALSE, "covariates")
  expect_equal(result, "covariates")
})

test_that(".assert_not_null_if_false() aborts when flag is FALSE and value is NULL", {
  expect_error(
    ukbflow:::.assert_not_null_if_false(FALSE, NULL),
    "must be supplied"
  )
})

test_that(".assert_not_null_if_false() error message includes flag and value names", {
  use_base <- FALSE
  covariates <- NULL
  expect_error(
    ukbflow:::.assert_not_null_if_false(use_base, covariates),
    "use_base"
  )
  expect_error(
    ukbflow:::.assert_not_null_if_false(use_base, covariates),
    "covariates"
  )
})


# ===========================================================================
# .assert_proportion()
# ===========================================================================

test_that(".assert_proportion() accepts a valid proportion", {
  expect_invisible(ukbflow:::.assert_proportion(0.05))
})

test_that(".assert_proportion() returns value invisibly", {
  result <- ukbflow:::.assert_proportion(0.5)
  expect_equal(result, 0.5)
})

test_that(".assert_proportion() rejects 0 (boundary, not strictly > 0)", {
  expect_error(ukbflow:::.assert_proportion(0), "strictly between 0 and 1")
})

test_that(".assert_proportion() rejects 1 (boundary, not strictly < 1)", {
  expect_error(ukbflow:::.assert_proportion(1), "strictly between 0 and 1")
})

test_that(".assert_proportion() rejects negative value", {
  expect_error(ukbflow:::.assert_proportion(-0.1), "strictly between 0 and 1")
})

test_that(".assert_proportion() rejects value > 1", {
  expect_error(ukbflow:::.assert_proportion(1.5), "strictly between 0 and 1")
})

test_that(".assert_proportion() rejects NA", {
  expect_error(ukbflow:::.assert_proportion(NA_real_), "strictly between 0 and 1")
})

test_that(".assert_proportion() rejects Inf", {
  expect_error(ukbflow:::.assert_proportion(Inf), "strictly between 0 and 1")
})

test_that(".assert_proportion() rejects character input", {
  expect_error(ukbflow:::.assert_proportion("0.05"), "strictly between 0 and 1")
})

test_that(".assert_proportion() rejects length > 1 vector", {
  expect_error(ukbflow:::.assert_proportion(c(0.1, 0.2)), "strictly between 0 and 1")
})

test_that(".assert_proportion() error message includes the argument name", {
  bad_p <- 1.5
  expect_error(ukbflow:::.assert_proportion(bad_p), "bad_p")
})

# =============================================================================
# utils_decode.R — Internal helpers for decode_ series
# =============================================================================


#' Convert a UKB field title to a snake_case column name
#'
#' Parses the UKB field title format (e.g.
#' \code{"Date of attending assessment centre | Instance 0"}) into a
#' compact snake_case identifier (e.g.
#' \code{"date_of_attending_assessment_centre_i0"}).
#'
#' @param title (character) Single UKB field title string.
#' @return (character) Snake_case column name.
#'
#' @keywords internal
#' @noRd
.title_to_snake <- function(title) {

  # Extract "| Instance X" → suffix "_iX"
  inst_m <- regmatches(title, regexpr("\\|\\s*Instance\\s+(\\d+)", title, perl = TRUE))
  inst_suffix <- if (length(inst_m) > 0 && nchar(inst_m) > 0) {
    paste0("_i", sub(".*?(\\d+)$", "\\1", inst_m, perl = TRUE))
  } else {
    ""
  }

  # Extract "| Array X" → suffix "_aX"
  arr_m <- regmatches(title, regexpr("\\|\\s*Array\\s+(\\d+)", title, perl = TRUE))
  arr_suffix <- if (length(arr_m) > 0 && nchar(arr_m) > 0) {
    paste0("_a", sub(".*?(\\d+)$", "\\1", arr_m, perl = TRUE))
  } else {
    ""
  }

  # Extract base label (everything before the first "|")
  base <- trimws(sub("\\s*\\|.*$", "", title))

  # Convert base to snake_case
  # Reason: if base somehow contains only special characters the two gsub()
  # calls below would yield ""; this is unreachable in practice because
  # .build_name_map() only calls here after a successful match() in fields_df
  base <- tolower(base)
  base <- gsub("[^a-z0-9]+", "_", base)
  base <- gsub("^_+|_+$", "", base)   # strip leading/trailing underscores

  paste0(base, inst_suffix, arr_suffix)
}


#' Extract the UKB field ID integer from a column name
#'
#' @param col_name (character) Column name such as \code{"participant.p31"},
#'   \code{"p31"}, or \code{"p53_i0_a1"}.
#' @return Integer field ID, or \code{NA_integer_} if none found.
#'
#' @keywords internal
#' @noRd
.extract_field_id <- function(col_name) {
  # Reason: anchor on literal "p" followed by digits; strips any prefix
  # (participant.) and any suffix (_i0, _a1, etc.)
  m <- regmatches(col_name, regexpr("(?<![a-z])p(\\d+)", col_name, perl = TRUE))
  if (length(m) == 0 || nchar(m) == 0) return(NA_integer_)
  as.integer(sub("p", "", m, fixed = TRUE))
}


#' Load and cache field.tsv metadata
#'
#' Reads \code{field.tsv} from \code{metadata_dir}, retaining only the columns
#' needed for value decoding (\code{field_id}, \code{value_type},
#' \code{encoding_id}). Result is cached in \code{.ukbflow_cache$field_meta}.
#'
#' @param metadata_dir (character) Directory containing \code{field.tsv}.
#' @return data.frame with columns \code{field_id}, \code{value_type},
#'   \code{encoding_id}.
#'
#' @keywords internal
#' @noRd
.load_field_meta <- function(metadata_dir) {
  if (!is.null(.ukbflow_cache$field_meta)) return(.ukbflow_cache$field_meta)

  path <- file.path(metadata_dir, "field.tsv")
  if (!file.exists(path)) {
    cli::cli_abort(
      c("{.file field.tsv} not found in {.path {metadata_dir}}.",
        "i" = "Run {.code fetch_file(\"Showcase metadata/field.tsv\", dest_dir = \"{metadata_dir}\")} first."),
      call = NULL
    )
  }

  # Reason: select only 3 columns to keep memory footprint small (~3.5 MB raw)
  df <- data.table::fread(
    path,
    select      = c("field_id", "value_type", "encoding_id"),
    data.table  = FALSE
  )
  .ukbflow_cache$field_meta <- df
  df
}


#' Load and cache esimpint.tsv encoding table
#'
#' Reads \code{esimpint.tsv} from \code{metadata_dir} and caches it in
#' \code{.ukbflow_cache$esimpint}. This table maps
#' \code{encoding_id + value → meaning} for simple integer-encoded categorical
#' fields (UKB \code{value_type} 21 and 22).
#'
#' @param metadata_dir (character) Directory containing \code{esimpint.tsv}.
#' @return data.frame with columns \code{encoding_id}, \code{value},
#'   \code{meaning}.
#'
#' @keywords internal
#' @noRd
.load_esimpint <- function(metadata_dir) {
  if (!is.null(.ukbflow_cache$esimpint)) return(.ukbflow_cache$esimpint)

  path <- file.path(metadata_dir, "esimpint.tsv")
  if (!file.exists(path)) {
    cli::cli_abort(
      c("{.file esimpint.tsv} not found in {.path {metadata_dir}}.",
        "i" = "Run {.code fetch_file(\"Showcase metadata/esimpint.tsv\", dest_dir = \"{metadata_dir}\")} first."),
      call = NULL
    )
  }

  df <- data.table::fread(
    path,
    select     = c("encoding_id", "value", "meaning"),
    data.table = FALSE
  )
  .ukbflow_cache$esimpint <- df
  df
}


#' Build a mapping from raw UKB column names to snake_case names
#'
#' Handles both \code{extract_pheno()} output (\code{participant.pXXXX})
#' and \code{job_result()} output (\code{pXXXX_iX}).
#'
#' @param col_names (character) Column names from the input data.
#' @param fields_df (data.frame) Output of \code{extract_ls()}, with columns
#'   \code{field_name} and \code{title}.
#' @return Character vector (same length as \code{col_names}) of new
#'   snake_case column names.
#'
#' @keywords internal
#' @noRd
.build_name_map <- function(col_names, fields_df) {
  vapply(col_names, function(col) {

    # Special case: participant ID column always stays "eid"
    if (col %in% c("eid", "participant.eid")) return("eid")

    # Normalise lookup key: ensure "participant." prefix for dictionary lookup
    lookup_key <- if (grepl("^participant\\.", col, perl = TRUE)) {
      col
    } else {
      paste0("participant.", col)
    }

    idx <- match(lookup_key, fields_df$field_name)

    if (is.na(idx)) {
      # Reason: fallback preserves column rather than silently dropping it;
      # should not occur if extract_ls() cache is fully populated
      return(sub("^participant\\.", "", col))
    }

    .title_to_snake(fields_df$title[idx])

  }, character(1), USE.NAMES = FALSE)
}

# =============================================================================
# test-decode.R — Unit tests for decode_ series (no network, no real files)
# =============================================================================

# ===========================================================================
# Shared helpers
# ===========================================================================

.fake_fields_decode <- function() {
  data.frame(
    field_name = c(
      "participant.eid",
      "participant.p31",
      "participant.p53_i0", "participant.p53_i1",
      "participant.p22189",
      "participant.p20002_i0_a0",
      "participant.p22009_a1"
    ),
    title = c(
      "Participant ID",
      "Sex",
      "Date of attending assessment centre | Instance 0",
      "Date of attending assessment centre | Instance 1",
      "Townsend deprivation index at recruitment",
      "Non-cancer illness code, self-reported | Instance 0 | Array 0",
      "Genetic principal components | Array 1"
    ),
    stringsAsFactors = FALSE
  )
}

.fake_field_meta <- function() {
  data.frame(
    field_id    = c(31L, 21000L, 20116L, 21022L, 22189L, 53L),
    value_type  = c(21L, 21L,    21L,    31L,    31L,    51L),
    encoding_id = c(9L,  1001L,  90L,    0L,     0L,     0L),
    stringsAsFactors = FALSE
  )
}

.fake_esimpint <- function() {
  data.frame(
    encoding_id = c(9L,  9L,    90L,     90L,        90L),
    value       = c(0L,  1L,    0L,      1L,          2L),
    meaning     = c("Female", "Male", "Never", "Previous", "Current"),
    stringsAsFactors = FALSE
  )
}

.set_decode_cache <- function() {
  # Reason: fields cache is a per-dataset named list; dataset cache avoids
  # dx ls network call so decode_names() -> extract_ls() is a pure memory path
  .ukbflow_cache$dataset    <- "fake_dataset"
  .ukbflow_cache$fields     <- list("fake_dataset" = .fake_fields_decode())
  .ukbflow_cache$field_meta <- .fake_field_meta()
  .ukbflow_cache$esimpint   <- .fake_esimpint()
}

.clear_decode_cache <- function() {
  .ukbflow_cache$dataset    <- NULL
  .ukbflow_cache$fields     <- NULL
  .ukbflow_cache$field_meta <- NULL
  .ukbflow_cache$esimpint   <- NULL
}


# ===========================================================================
# .title_to_snake() — pure function
# ===========================================================================

test_that(".title_to_snake() handles simple title", {
  expect_equal(ukbflow:::.title_to_snake("Sex"), "sex")
})

test_that(".title_to_snake() converts spaces and special chars to underscores", {
  expect_equal(
    ukbflow:::.title_to_snake("Townsend deprivation index at recruitment"),
    "townsend_deprivation_index_at_recruitment"
  )
})

test_that(".title_to_snake() appends instance suffix", {
  result <- ukbflow:::.title_to_snake(
    "Date of attending assessment centre | Instance 0"
  )
  expect_true(endsWith(result, "_i0"))
  expect_true(startsWith(result, "date_of_attending"))
})

test_that(".title_to_snake() appends array suffix", {
  result <- ukbflow:::.title_to_snake("Genetic principal components | Array 1")
  expect_true(endsWith(result, "_a1"))
  expect_true(startsWith(result, "genetic_principal"))
})

test_that(".title_to_snake() handles both instance and array suffixes", {
  result <- ukbflow:::.title_to_snake(
    "Non-cancer illness code, self-reported | Instance 0 | Array 1"
  )
  expect_true(grepl("_i0_a1$", result))
})

test_that(".title_to_snake() strips leading and trailing underscores from base", {
  # Titles starting with special chars should not produce leading underscores
  result <- ukbflow:::.title_to_snake("Age at recruitment")
  expect_false(startsWith(result, "_"))
  expect_false(endsWith(result, "_"))
})


# ===========================================================================
# .extract_field_id() — pure function
# ===========================================================================

test_that(".extract_field_id() extracts field_id from participant.pXXXX", {
  expect_equal(ukbflow:::.extract_field_id("participant.p31"), 31L)
})

test_that(".extract_field_id() extracts field_id from pXXXX", {
  expect_equal(ukbflow:::.extract_field_id("p31"), 31L)
})

test_that(".extract_field_id() extracts field_id from pXXXX_iX_aX", {
  expect_equal(ukbflow:::.extract_field_id("p20002_i0_a1"), 20002L)
})

test_that(".extract_field_id() returns NA for eid column", {
  expect_true(is.na(ukbflow:::.extract_field_id("eid")))
})

test_that(".extract_field_id() returns NA for non-UKB column name", {
  expect_true(is.na(ukbflow:::.extract_field_id("my_column")))
})


# ===========================================================================
# .build_name_map() — pure function
# ===========================================================================

test_that(".build_name_map() maps participant.eid to eid", {
  result <- ukbflow:::.build_name_map("participant.eid", .fake_fields_decode())
  expect_equal(result, "eid")
})

test_that(".build_name_map() maps participant.p31 to sex", {
  result <- ukbflow:::.build_name_map("participant.p31", .fake_fields_decode())
  expect_equal(result, "sex")
})

test_that(".build_name_map() maps p31 (no prefix) to sex", {
  result <- ukbflow:::.build_name_map("p31", .fake_fields_decode())
  expect_equal(result, "sex")
})

test_that(".build_name_map() appends _i0 suffix for instance fields", {
  result <- ukbflow:::.build_name_map("participant.p53_i0", .fake_fields_decode())
  expect_true(endsWith(result, "_i0"))
})

test_that(".build_name_map() appends _a1 suffix for array fields", {
  result <- ukbflow:::.build_name_map("participant.p22009_a1", .fake_fields_decode())
  expect_true(endsWith(result, "_a1"))
})

test_that(".build_name_map() appends _i0_a0 for instance+array fields", {
  result <- ukbflow:::.build_name_map("participant.p20002_i0_a0", .fake_fields_decode())
  expect_true(grepl("_i0_a0$", result))
})

test_that(".build_name_map() falls back to stripped name for unknown column", {
  result <- ukbflow:::.build_name_map("participant.p99999", .fake_fields_decode())
  expect_equal(result, "p99999")
})


# ===========================================================================
# decode_names() — user-facing function
# ===========================================================================

test_that("decode_names() stops on non-data.frame input", {
  expect_error(decode_names("not a df"), "data.frame")
})

test_that("decode_names() stops on invalid max_nchar", {
  .set_decode_cache()
  on.exit(.clear_decode_cache())
  df <- data.frame(participant.p31 = 1L)
  expect_error(decode_names(df, max_nchar = 0), "positive integer")
})

test_that("decode_names() renames participant.eid to eid", {
  .set_decode_cache()
  on.exit(.clear_decode_cache())
  df <- data.frame(`participant.eid` = 1L, check.names = FALSE)
  result <- suppressMessages(decode_names(df))
  expect_true("eid" %in% names(result))
})

test_that("decode_names() renames p31 to sex", {
  .set_decode_cache()
  on.exit(.clear_decode_cache())
  df <- data.frame(`participant.eid` = 1L, `participant.p31` = 1L,
                   check.names = FALSE)
  result <- suppressMessages(decode_names(df))
  expect_true("sex" %in% names(result))
  expect_false("participant.p31" %in% names(result))
})

test_that("decode_names() works with extract_batch format (no prefix)", {
  .set_decode_cache()
  on.exit(.clear_decode_cache())
  df <- data.frame(eid = 1L, p31 = 1L)
  result <- suppressMessages(decode_names(df))
  expect_true("eid"  %in% names(result))
  expect_true("sex"  %in% names(result))
})

test_that("decode_names() preserves data.table class", {
  .set_decode_cache()
  on.exit(.clear_decode_cache())
  dt <- data.table::data.table(`participant.eid` = 1L, `participant.p31` = 1L)
  result <- suppressMessages(decode_names(dt))
  expect_true(data.table::is.data.table(result))
})

test_that("decode_names() warns when a renamed column exceeds max_nchar", {
  .set_decode_cache()
  on.exit(.clear_decode_cache())
  # p53_i0 → "date_of_attending_assessment_centre_i0" (>30 chars)
  df <- data.frame(`participant.p53_i0` = "2008-01-01", check.names = FALSE)
  expect_message(
    decode_names(df, max_nchar = 30),
    "longer than 30"
  )
})

test_that("decode_names() does not warn within max_nchar limit", {
  .set_decode_cache()
  on.exit(.clear_decode_cache())
  df <- data.frame(`participant.p31` = 1L, check.names = FALSE)
  # "sex" is 3 chars — no warning expected
  expect_no_warning(suppressMessages(decode_names(df, max_nchar = 60)))
})

test_that("decode_names() produces correct output when dataset and fields are cached", {
  # Reason: with both $dataset and $fields populated, decode_names() → extract_ls()
  # is a pure memory path — verify the end result is correct, not the mechanism
  .set_decode_cache()
  on.exit(.clear_decode_cache())

  df <- data.frame(`participant.eid` = 1L, `participant.p31` = 1L,
                   check.names = FALSE)
  result <- suppressMessages(decode_names(df))
  expect_true("eid" %in% names(result))
  expect_true("sex" %in% names(result))
})

test_that("decode_names() handles duplicate snake_case names with make.unique()", {
  # Two columns that would produce the same snake_case name
  fields_dup <- data.frame(
    field_name = c("participant.p99_i0", "participant.p99_i0_dup"),
    title      = c("Some field | Instance 0", "Some field | Instance 0"),
    stringsAsFactors = FALSE
  )
  .ukbflow_cache$dataset <- "fake_dataset"
  .ukbflow_cache$fields  <- list("fake_dataset" = fields_dup)
  on.exit(.clear_decode_cache())

  df <- data.frame(p99_i0 = 1L, p99_i0_dup = 2L)
  result <- suppressMessages(decode_names(df))
  expect_equal(length(unique(names(result))), 2L)
})


# ===========================================================================
# decode_values() — user-facing function
# ===========================================================================

test_that("decode_values() stops on non-data.frame input", {
  expect_error(decode_values("not a df"), "data.frame")
})

test_that("decode_values() stops on non-string metadata_dir", {
  df <- data.frame(p31 = 1L)
  expect_error(decode_values(df, metadata_dir = 123))
})

test_that("decode_values() stops on NA metadata_dir", {
  df <- data.frame(p31 = 1L)
  expect_error(decode_values(df, metadata_dir = NA_character_))
})

test_that("decode_values() stops on length > 1 metadata_dir", {
  df <- data.frame(p31 = 1L)
  expect_error(decode_values(df, metadata_dir = c("a/", "b/")))
})

test_that("decode_values() warns when no UKB field columns found", {
  .set_decode_cache()
  on.exit(.clear_decode_cache())
  df <- data.frame(sex = c("Male", "Female"))
  expect_message(decode_values(df), "no UKB field ID columns detected",
                 ignore.case = TRUE)
})

test_that("decode_values() decodes p31: 0 → Female, 1 → Male", {
  .set_decode_cache()
  on.exit(.clear_decode_cache())
  df <- data.frame(p31 = c(0L, 1L, 1L, 0L))
  result <- suppressMessages(decode_values(df))
  expect_equal(result$p31, c("Female", "Male", "Male", "Female"))
})

test_that("decode_values() decodes p20116 (smoking): 0/1/2 → labels", {
  .set_decode_cache()
  on.exit(.clear_decode_cache())
  df <- data.frame(p20116_i0 = c(0L, 1L, 2L))
  result <- suppressMessages(decode_values(df))
  expect_equal(result$p20116_i0, c("Never", "Previous", "Current"))
})

test_that("decode_values() leaves continuous column unchanged", {
  .set_decode_cache()
  on.exit(.clear_decode_cache())
  df <- data.frame(p31 = 1L, p22189 = -3.94)
  result <- suppressMessages(decode_values(df))
  expect_equal(result$p22189, -3.94)
  expect_true(is.numeric(result$p22189))
})

test_that("decode_values() leaves already-decoded character column unchanged", {
  .set_decode_cache()
  on.exit(.clear_decode_cache())
  df <- data.frame(p31 = c("Female", "Male"), stringsAsFactors = FALSE)
  result <- suppressMessages(decode_values(df))
  # Should remain unchanged — already character
  expect_equal(result$p31, c("Female", "Male"))
})

test_that("decode_values() returns NA for codes absent from encoding table", {
  .set_decode_cache()
  on.exit(.clear_decode_cache())
  df <- data.frame(p31 = c(0L, 1L, 99L))   # 99 is not in encoding 9
  result <- suppressMessages(decode_values(df))
  expect_true(is.na(result$p31[3]))
})

test_that("decode_values() preserves data.table class", {
  .set_decode_cache()
  on.exit(.clear_decode_cache())
  dt <- data.table::data.table(p31 = c(0L, 1L))
  result <- suppressMessages(decode_values(dt))
  expect_true(data.table::is.data.table(result))
})

test_that("decode_values() leaves non-categorical field (p53 date) unchanged", {
  .set_decode_cache()
  on.exit(.clear_decode_cache())
  df <- data.frame(p31 = 1L, p53_i0 = 13956L)
  result <- suppressMessages(decode_values(df))
  expect_equal(result$p53_i0, 13956L)
})

test_that("decode_values() errors with clear message when field.tsv missing", {
  .clear_decode_cache()
  on.exit(.clear_decode_cache())
  df <- data.frame(p31 = 1L)
  expect_error(
    suppressMessages(decode_values(df, metadata_dir = tempdir())),
    "field.tsv"
  )
})

test_that("decode_values() errors with clear message when esimpint.tsv missing", {
  .clear_decode_cache()
  on.exit(.clear_decode_cache())

  # Provide field.tsv in a temp dir but not esimpint.tsv
  tmp <- tempdir()
  field_path <- file.path(tmp, "field.tsv")
  data.table::fwrite(.fake_field_meta(), field_path, sep = "\t")
  on.exit(unlink(field_path), add = TRUE)

  df <- data.frame(p31 = 1L)
  expect_error(
    suppressMessages(decode_values(df, metadata_dir = tmp)),
    "esimpint.tsv"
  )
})

# =============================================================================
# utils_plot.R — internal helpers for plot_forest() and plot_tableone()
# =============================================================================


# -----------------------------------------------------------------------------
# Theme
# -----------------------------------------------------------------------------

# Build forest theme. theme must be a recognised preset string or a
# forest_theme object. Unrecognised strings warn and fall back to "default".
.fp_theme <- function(theme, ci_Theight = 0.2) {
  if (inherits(theme, "forest_theme")) return(theme)

  presets <- list(
    default = list(
      base_size    = 12,
      base_family  = "sans",
      ci_pch       = 15,
      ci_lty       = 1,
      ci_lwd       = 3,
      ci_col       = "black",
      ci_alpha     = 1,
      ci_fill      = "black",
      ci_Theight   = ci_Theight,
      refline_gp   = grid::gpar(lwd = 2, lty = "dashed", col = "grey20"),
      xaxis_gp     = grid::gpar(fontsize = 12, fontfamily = "sans"),
      arrow_type   = "open",
      arrow_length = 0.1,
      arrow_gp     = grid::gpar(fontsize = 12, fontfamily = "sans", lwd = 2),
      xlab_adjust  = "center",
      xlab_gp      = grid::gpar(fontsize = 10, fontfamily = "sans", fontface = "plain")
    )
  )

  if (!is.character(theme) || !theme %in% names(presets)) {
    cli::cli_warn(
      c("Unknown theme {.val {theme}}, falling back to {.val {'default'}}.",
        "i" = "Available: {.val {names(presets)}}")
    )
    theme <- "default"
  }

  do.call(forestploter::forest_theme, presets[[theme]])
}


# -----------------------------------------------------------------------------
# Data pre-processing
# -----------------------------------------------------------------------------

# Apply indent, format p_cols, insert gap_ci + OR label at ci_column.
# Returns list(data = final_df, p_col_idxs = integer vector of p col positions
#              in final_df).
.fp_build_data <- function(data, est, lower, upper, indent,
                           p_cols, p_digits, ci_digits, ci_sep, ci_column) {
  n         <- nrow(data)
  ncol_orig <- ncol(data)

  # 1. Indent: prepend non-breaking spaces to item (col 1).
  # Reason: regular spaces are often collapsed by grid text rendering;
  # \u00a0 (non-breaking space) preserves visual indentation reliably.
  data[[1L]] <- paste0(strrep("\u00a0\u00a0", as.integer(indent)), data[[1L]])

  # 2. Format p_cols: numeric → character, cache originals for bold logic
  p_numeric <- list()
  if (!is.null(p_cols)) {
    thresh_str <- formatC(10^(-p_digits), digits = p_digits, format = "f")
    for (col in p_cols) {
      pv             <- as.numeric(data[[col]])
      p_numeric[[col]] <- pv
      data[[col]]    <- ifelse(
        is.na(pv), "",
        ifelse(pv < 10^(-p_digits),
               paste0("<", thresh_str),
               formatC(pv, digits = p_digits, format = "f"))
      )
    }
  }

  # 3. Generate OR (95% CI) label from est/lower/upper
  fmt <- function(x) formatC(x, digits = ci_digits, format = "f")
  ci_label <- ifelse(
    is.na(est), "",
    paste0(fmt(est), " (", fmt(lower), ci_sep, fmt(upper), ")")
  )

  # 4. Build final table: left cols | gap_ci | OR (95% CI) | right cols
  if (ci_column <= 1L)
    cli::cli_abort("{.arg ci_column} must be >= 2.", call = NULL)
  if (ci_column > ncol_orig + 1L)
    cli::cli_abort(
      "{.arg ci_column} ({ci_column}) exceeds ncol(data)+1 ({ncol_orig + 1L}).",
      call = NULL
    )

  gap_df <- data.frame(gap_ci        = strrep(" ", 20L), stringsAsFactors = FALSE)
  lab_df <- data.frame(`OR (95% CI)` = ci_label,         stringsAsFactors = FALSE,
                       check.names   = FALSE)

  left  <- data[, seq_len(ci_column - 1L), drop = FALSE]
  right <- if (ci_column <= ncol_orig)
    data[, seq(ci_column, ncol_orig), drop = FALSE]
  else
    NULL

  data_r <- if (is.null(right)) cbind(left, gap_df, lab_df)
            else                 cbind(left, gap_df, lab_df, right)

  # 5. Map original p_cols indices to final positions (+2 if >= ci_column)
  p_col_idxs <- NULL
  if (!is.null(p_cols)) {
    orig_idxs  <- match(p_cols, names(data))
    p_col_idxs <- ifelse(orig_idxs >= ci_column, orig_idxs + 2L, orig_idxs)
  }

  list(data = data_r, p_col_idxs = p_col_idxs, p_numeric = p_numeric)
}


# -----------------------------------------------------------------------------
# Alignment
# -----------------------------------------------------------------------------

# align: integer vector, length = ncol(final data).
#   -1 = left, 0 = center, 1 = right.
# NULL = auto: col 1 left, everything else center.
.fp_align <- function(p, n_cols, align) {
  if (is.null(align)) align <- c(-1L, rep(0L, n_cols - 1L))

  hjust_map <- c("-1" = 0, "0" = 0.5, "1" = 1)
  x_map     <- c("-1" = 0, "0" = 0.5, "1" = 1)

  for (j in seq_len(n_cols)) {
    key  <- as.character(align[j])
    hj   <- grid::unit(hjust_map[[key]], "npc")
    xpos <- grid::unit(x_map[[key]], "npc")
    for (part in c("body", "header")) {
      p <- forestploter::edit_plot(p, col = j, which = "text", part = part,
                                   hjust = hj, x = xpos)
    }
  }
  p
}


# -----------------------------------------------------------------------------
# Bold label (item column)
# -----------------------------------------------------------------------------

# bold_label: logical vector length n_rows. TRUE rows → bold item column.
.fp_bold_label <- function(p, bold_label) {
  rows <- which(bold_label)
  if (length(rows) == 0L) return(p)
  forestploter::edit_plot(p, row = rows, col = 1L, which = "text",
                          gp = grid::gpar(fontface = "bold"))
}


# -----------------------------------------------------------------------------
# Bold p values
# -----------------------------------------------------------------------------

# p_cols: column names in original data.
# p_col_idxs: their indices in final data_r.
# p_numeric: list of original numeric p vectors (named by p_cols).
# bold_p: TRUE / FALSE / logical vector (length n_rows).
# p_threshold: scalar; only used when bold_p is TRUE.
.fp_bold_p <- function(p, p_cols, p_col_idxs, p_numeric,
                       bold_p, p_threshold, n_rows) {
  if (isFALSE(bold_p)) return(p)

  for (j in seq_along(p_cols)) {
    pv      <- p_numeric[[p_cols[j]]]
    col_idx <- p_col_idxs[j]

    bold_rows <- which(bold_p & !is.na(pv) & pv < p_threshold)

    for (r in bold_rows) {
      p <- forestploter::edit_plot(p, row = r, col = col_idx, which = "text",
                                   gp = grid::gpar(fontface = "bold"))
    }
  }
  p
}


# -----------------------------------------------------------------------------
# CI colours
# -----------------------------------------------------------------------------

# ci_col: scalar or vector (length n_rows). Rows with NA est or NA ci_col skipped.
.fp_ci_colors <- function(p, ci_col, ci_column, est, n_rows) {
  if (length(ci_col) == 1L) ci_col <- rep(ci_col, n_rows)
  for (i in seq_len(n_rows)) {
    if (!is.na(est[i]) && !is.na(ci_col[i])) {
      p <- forestploter::edit_plot(p, row = i, col = ci_column, which = "ci",
                                   gp = grid::gpar(fill = ci_col[i],
                                                   col  = ci_col[i]))
    }
  }
  p
}


# -----------------------------------------------------------------------------
# Background
# -----------------------------------------------------------------------------

# background: "zebra" | "bold_label" | "none"
# bg_col: scalar (used for shaded rows) or vector (length n_rows, overrides all).
.fp_background <- function(p, n_rows, n_cols, background, bold_label, bg_col) {
  if (background == "none") return(p)

  all_cols <- seq_len(n_cols)

  # Determine per-row fill colour
  if (length(bg_col) == n_rows) {
    # User supplied per-row vector: direct override
    fill <- bg_col
  } else if (background == "zebra") {
    fill <- ifelse(seq_len(n_rows) %% 2L == 0L, bg_col, "white")
  } else if (background == "bold_label") {
    fill <- ifelse(bold_label, bg_col, "white")
  } else {
    cli::cli_abort(
      c("Unknown {.arg background} value {.val {background}}.",
        "i" = "Must be one of: {.val {c('zebra', 'bold_label', 'none')}}"),
      call = NULL
    )
  }

  for (i in seq_len(n_rows)) {
    p <- forestploter::edit_plot(p, row = i, col = all_cols,
                                 which = "background",
                                 gp    = grid::gpar(fill = fill[i], col = NA))
  }
  p
}


# -----------------------------------------------------------------------------
# Borders
# -----------------------------------------------------------------------------

# border: "three_line" | "none"
# border_width: scalar (all lines same) or length-3 vector
#   (top-of-header, bottom-of-header, bottom-of-body).
.fp_borders <- function(p, n_rows, border, border_width) {
  if (border == "none") return(p)

  if (length(border_width) == 1L) border_width <- rep(border_width, 3L)

  p <- forestploter::add_border(p, part = "header", row = 1L, where = "top",
                                gp = grid::gpar(lwd = border_width[1L]))
  p <- forestploter::add_border(p, part = "header", row = 1L, where = "bottom",
                                gp = grid::gpar(lwd = border_width[2L]))
  p <- forestploter::add_border(p, part = "body",   row = n_rows, where = "bottom",
                                gp = grid::gpar(lwd = border_width[3L]))
  p
}


# -----------------------------------------------------------------------------
# Layout (heights & widths)
# -----------------------------------------------------------------------------

# row_height / col_width: NULL (auto) | scalar | vector.
#
# Auto height: [1]=8mm top, [2]=12mm header, [3..n-1]=10mm rows, [n]=15mm bottom
# Auto width:  each col → ceiling((w+5)/5)*5  (adjustment in [5,10) mm)
.fp_layout <- function(p, row_height, col_width) {
  # --- heights ---
  nh      <- length(p$heights)
  if (!is.null(row_height)) {
    # scalar or vector supplied by user
    if (length(row_height) == 1L) row_height <- rep(row_height, nh)
    for (i in seq_len(nh))
      p$heights[i] <- grid::unit(row_height[i], "mm")
  } else {
    # auto
    p$heights[1L] <- grid::unit(8,  "mm")
    p$heights[2L] <- grid::unit(12, "mm")
    if (nh > 3L)
      for (i in seq(3L, nh - 1L))
        p$heights[i] <- grid::unit(10, "mm")
    p$heights[nh] <- grid::unit(15, "mm")
  }

  # --- widths ---
  nw        <- length(p$widths)
  default_w <- round(grid::convertWidth(p$widths, "mm", valueOnly = TRUE), 1)

  if (!is.null(col_width)) {
    if (length(col_width) == 1L) col_width <- rep(col_width, nw)
    for (i in seq_len(nw))
      p$widths[i] <- grid::unit(col_width[i], "mm")
  } else {
    # auto: round up so adjustment is in [5, 10) mm
    for (i in seq_len(nw)) {
      w   <- default_w[i]
      new <- ceiling((w + 5) / 5) * 5
      p$widths[i] <- grid::unit(new, "mm")
    }
  }

  p
}


# =============================================================================
# Table 1 helpers (.t1_*)
# =============================================================================


# Build a gtsummary tbl_summary with bold labels.
# All args NULL-safe: only passed to tbl_summary when non-NULL.
.t1_build_tbl <- function(data, vars, strata, type, label, statistic,
                           digits, percent, missing) {
  sub  <- if (is.null(strata)) data[, vars, drop = FALSE]
          else                  data[, c(vars, strata), drop = FALSE]

  args <- list(data = sub, by = strata, percent = percent, missing = missing)
  if (!is.null(type))      args$type      <- type
  if (!is.null(label))     args$label     <- label
  if (!is.null(statistic)) args$statistic <- statistic
  if (!is.null(digits))    args$digits    <- digits

  tbl <- do.call(gtsummary::tbl_summary, args)
  gtsummary::bold_labels(tbl)
}


# Add p-value column: 3 dp / <0.001 cutoff, italic header, bold sig p.
.t1_add_p <- function(tbl) {
  tbl <- gtsummary::add_p(
    tbl,
    pvalue_fun = function(x) gtsummary::style_pvalue(x, digits = 3)
  )
  tbl <- gtsummary::bold_p(tbl)
  # Italic *P* via gtsummary markdown syntax
  tbl <- gtsummary::modify_header(tbl, p.value ~ "***P*-value**")
  tbl
}


# Add SMD column to tbl_summary.
# Continuous: Cohen's d. Categorical: RMSD. Shown on label rows only.
.t1_add_smd <- function(tbl, data, vars, strata) {
  smd_vals <- .t1_compute_smd(data, vars, strata)
  smd_map  <- stats::setNames(smd_vals$smd_fmt, smd_vals$variable)

  tbl <- gtsummary::modify_table_body(tbl, function(tb) {
    tb$smd <- ifelse(
      tb$row_type == "label",
      smd_map[tb$variable],
      NA_character_
    )
    tb
  })

  # Register column so as_gt() renders it
  tbl <- gtsummary::modify_table_styling(
    tbl,
    columns = "smd",
    label   = "**SMD**",
    align   = "center",
    hide    = FALSE
  )

  tbl <- gtsummary::modify_table_body(tbl, function(tb) {
    dplyr::relocate(tb, "smd", .before = "p.value")
  })

  tbl
}


# Compute SMD for each variable.
# 2 groups  → direct formula.
# 3+ groups → mean of all pairwise values.
.t1_compute_smd <- function(data, vars, strata) {
  grp  <- factor(data[[strata]])
  levs <- levels(grp)

  rows <- lapply(vars, function(v) {
    x <- data[[v]]
    val <- tryCatch(
      if (is.numeric(x)) .t1_smd_continuous(x, grp, levs)
      else               .t1_smd_categorical(x, grp, levs),
      error = function(e) NA_real_
    )
    data.frame(
      variable = v,
      smd_fmt  = if (!is.na(val)) formatC(val, digits = 3, format = "f")
                 else             NA_character_,
      stringsAsFactors = FALSE
    )
  })
  do.call(rbind, rows)
}


# Cohen's d for continuous variable (pooled-SD formula).
.t1_smd_continuous <- function(x, grp, levs) {
  pairs <- utils::combn(levs, 2L, simplify = FALSE)
  smds  <- vapply(pairs, function(p) {
    x1 <- x[grp == p[1L] & !is.na(x)]
    x2 <- x[grp == p[2L] & !is.na(x)]
    # Require at least 2 observations per group; var() on n<2 returns NA + warning
    if (length(x1) < 2L || length(x2) < 2L) return(NA_real_)
    ps <- sqrt((stats::var(x1) + stats::var(x2)) / 2)
    if (ps == 0) return(NA_real_)
    abs(mean(x1) - mean(x2)) / ps
  }, numeric(1L))
  mean(smds, na.rm = TRUE)
}


# RMSD of group proportions for categorical variable.
.t1_smd_categorical <- function(x, grp, levs) {
  cats  <- levels(factor(x))
  if (length(cats) < 2L) return(NA_real_)
  pairs <- utils::combn(levs, 2L, simplify = FALSE)
  smds  <- vapply(pairs, function(p) {
    p1 <- prop.table(table(factor(x[grp == p[1L]], levels = cats)))
    p2 <- prop.table(table(factor(x[grp == p[2L]], levels = cats)))
    sqrt(mean((as.numeric(p1) - as.numeric(p2))^2))
  }, numeric(1L))
  mean(smds, na.rm = TRUE)
}


# Apply visual theme (currently only "lancet") to a gt table.
# Block-based row shading (by variable group), header shading, three-line
# borders, font size, and column widths.
# tbl: original gtsummary object (needed to read row_type for block IDs).
.t1_apply_theme <- function(gt_tbl, tbl, theme, label_width, stat_width,
                              pvalue_width, row_height) {
  valid_themes <- "lancet"
  if (!theme %in% valid_themes) {
    cli::cli_warn("Unknown theme {.val {theme}}; using {.val {'lancet'}}.")
    theme <- "lancet"
  }

  # --- block-based row shading -----------------------------------------------
  # Reason: each variable occupies 1-N rows (label + levels); coloring by
  # block (not row index) keeps a variable's rows visually grouped.
  tbl_data  <- tbl$table_body
  label_idx <- which(tbl_data$row_type == "label")

  if (length(label_idx) > 0L) {
    block_id    <- rep(seq_along(label_idx),
                       times = diff(c(label_idx, nrow(tbl_data) + 1L)))
    fill_colors <- ifelse(block_id %% 2L == 1L, "#f8e8e7", "#FFFFFF")

    pink_rows  <- which(fill_colors == "#f8e8e7")
    white_rows <- which(fill_colors == "#FFFFFF")

    if (length(pink_rows) > 0L)
      gt_tbl <- gt::tab_style(gt_tbl,
        style     = gt::cell_fill(color = "#f8e8e7"),
        locations = gt::cells_body(rows = pink_rows))

    if (length(white_rows) > 0L)
      gt_tbl <- gt::tab_style(gt_tbl,
        style     = gt::cell_fill(color = "#FFFFFF"),
        locations = gt::cells_body(rows = white_rows))
  }

  # --- header row: same #f8e8e7 background -----------------------------------
  gt_tbl <- gt::tab_style(gt_tbl,
    style     = gt::cell_fill(color = "#f8e8e7"),
    locations = gt::cells_column_labels())

  # --- three-line borders (via tab_style, higher specificity than tab_options)
  n_body <- nrow(tbl_data)
  # Line 1: top of header
  gt_tbl <- gt::tab_style(gt_tbl,
    style     = gt::cell_borders(sides = "top",    color = "black", weight = gt::px(3)),
    locations = gt::cells_column_labels())
  # Line 2: bottom of header
  gt_tbl <- gt::tab_style(gt_tbl,
    style     = gt::cell_borders(sides = "bottom", color = "black", weight = gt::px(2)),
    locations = gt::cells_column_labels())
  # Line 3: bottom of last body row
  gt_tbl <- gt::tab_style(gt_tbl,
    style     = gt::cell_borders(sides = "bottom", color = "black", weight = gt::px(3)),
    locations = gt::cells_body(rows = n_body))

  # Remove all other horizontal rules
  gt_tbl <- gt::tab_options(gt_tbl,
    table_body.hlines.style   = "none",
    table.border.bottom.style = "none",
    table.border.top.style    = "none",
    table.font.size           = gt::px(15),
    data_row.padding          = gt::px(row_height),
    column_labels.padding     = gt::px(12))

  # --- column widths ---------------------------------------------------------
  # Reason: gt::cols_width() evaluates formula RHS via NSE and may not resolve
  # local function variables from within a package. rlang::inject() + !!
  # substitutes the actual integer values before gt sees the call.
  gt_tbl <- rlang::inject(
    gt::cols_width(
      gt_tbl,
      dplyr::any_of("label")             ~ gt::px(!!label_width),
      dplyr::any_of(c("p.value", "smd")) ~ gt::px(!!pvalue_width),
      gt::everything()                   ~ gt::px(!!stat_width)
    )
  )

  gt_tbl
}


# Save table to word / html / pdf / png.
#
# DOCX/HTML : gt::gtsave() directly — preserves gt styling.
# PDF       : gt → temp HTML → pagedown::chrome_print(printBackground=TRUE).
# PNG       : gt → temp HTML → webshot2::webshot(selector=".gt_table").
#
# Reason: browsers strip backgrounds by default when printing/rendering.
# CSS `print-color-adjust: exact` + printBackground=TRUE forces preservation.
.t1_save <- function(gt_tbl, dest, png_scale = 2, pdf_width = NULL, pdf_height = NULL) {
  base <- tools::file_path_sans_ext(dest)

  # Inject CSS for background/border preservation in PDF & PNG rendering.
  # Applied to a separate copy so the returned gt object stays clean.
  gt_print <- gt_tbl |>
    gt::opt_css(css = "
      @media print {
        body, td, th {
          -webkit-print-color-adjust: exact !important;
          print-color-adjust: exact !important;
          color-adjust: exact !important;
        }
        table, td, th { border-color: inherit !important; }
      }
      @media screen {
        body, td, th {
          -webkit-print-color-adjust: exact !important;
          print-color-adjust: exact !important;
        }
      }
    ")

  # DOCX ----------------------------------------------------------------------
  tryCatch({
    out <- paste0(base, ".docx")
    gt::gtsave(gt_tbl, filename = out)
    cli::cli_alert_success("Saved: {.path {out}}")
  }, error = function(e) {
    cli::cli_alert_danger("Failed to save .docx: {conditionMessage(e)}")
  })

  # HTML ----------------------------------------------------------------------
  tryCatch({
    out <- paste0(base, ".html")
    gt::gtsave(gt_tbl, filename = out)
    cli::cli_alert_success("Saved: {.path {out}}")
  }, error = function(e) {
    cli::cli_alert_danger("Failed to save .html: {conditionMessage(e)}")
  })

  # PDF + PNG share the same HTML source — write once, reuse twice.
  tmp_html <- tempfile(fileext = ".html")
  on.exit(unlink(tmp_html), add = TRUE)
  html_ready <- tryCatch({
    gt::gtsave(gt_print, filename = tmp_html)
    TRUE
  }, error = function(e) {
    cli::cli_alert_danger("Failed to build temp HTML for PDF/PNG: {conditionMessage(e)}")
    FALSE
  })

  # PDF (HTML → pagedown::chrome_print) ---------------------------------------
  if (html_ready) tryCatch({
    if (!requireNamespace("pagedown", quietly = TRUE))
      cli::cli_abort("Package {.pkg pagedown} is required for PDF export.", call = NULL)
    out <- paste0(base, ".pdf")
    pdf_opts <- list(
      printBackground     = TRUE,
      preferCSSPageSize   = TRUE,
      displayHeaderFooter = FALSE
    )
    if (!is.null(pdf_width))  pdf_opts$paperWidth  <- pdf_width
    if (!is.null(pdf_height)) pdf_opts$paperHeight <- pdf_height
    pagedown::chrome_print(
      input      = tmp_html,
      output     = out,
      format     = "pdf",
      options    = pdf_opts,
      extra_args = c("--no-sandbox", "--disable-dev-shm-usage"),
      verbose    = 0
    )
    cli::cli_alert_success("Saved: {.path {out}}")
  }, error = function(e) {
    cli::cli_alert_danger("Failed to save .pdf: {conditionMessage(e)}")
  })

  # PNG (HTML → webshot2::webshot) --------------------------------------------
  if (html_ready) tryCatch({
    if (!requireNamespace("webshot2", quietly = TRUE))
      cli::cli_abort("Package {.pkg webshot2} is required for PNG export.", call = NULL)
    out <- paste0(base, ".png")
    webshot2::webshot(
      url      = tmp_html,
      file     = out,
      zoom     = png_scale,
      selector = ".gt_table",
      expand   = 10
    )
    cli::cli_alert_success("Saved: {.path {out}}")
  }, error = function(e) {
    cli::cli_alert_danger("Failed to save .png: {conditionMessage(e)}")
  })

  invisible(NULL)
}


# =============================================================================
# Forest plot save helper (.fp_save) — below
# =============================================================================


# dest: file path (with or without extension; extension stripped and ignored).
# Saves png / pdf / jpg / tiff at 300 dpi, white background, always overwrite.
.fp_save <- function(p, dest, save_width, save_height) {
  base  <- tools::file_path_sans_ext(dest)
  exts  <- c("png", "pdf", "jpg", "tiff")
  w_in  <- save_width  / 2.54
  h_in  <- save_height / 2.54

  for (ext in exts) {
    out <- paste0(base, ".", ext)
    tryCatch({
      if (ext == "pdf") {
        grDevices::pdf(out, width = w_in, height = h_in, bg = "white")
      } else if (ext == "png") {
        grDevices::png(out, width = save_width, height = save_height,
                       units = "cm", res = 300, bg = "white")
      } else if (ext == "jpg") {
        grDevices::jpeg(out, width = save_width, height = save_height,
                        units = "cm", res = 300, bg = "white", quality = 95)
      } else if (ext == "tiff") {
        grDevices::tiff(out, width = save_width, height = save_height,
                        units = "cm", res = 300, bg = "white")
      }
      grid::grid.newpage()
      grid::grid.draw(p)
      grDevices::dev.off()
      cli::cli_alert_success("Saved: {.path {out}}")
    }, error = function(e) {
      try(grDevices::dev.off(), silent = TRUE)
      cli::cli_alert_danger("Failed to save {.path {out}}: {conditionMessage(e)}")
    })
  }
  invisible(NULL)
}

# =============================================================================
# test-plot.R — Unit tests for plot_forest() and plot_tableone()
#   Focus: input validation + pure internal helpers (no graphics device needed)
# =============================================================================


# ===========================================================================
# Shared helpers
# ===========================================================================

.fake_forest_df <- function() {
  data.frame(
    item    = c("AD vs control", "Crude", "Adjusted"),
    cases_n = c("", "89/4521", "89/4521"),
    stringsAsFactors = FALSE
  )
}

.fake_t1_df <- function(n = 120, seed = 42) {
  set.seed(seed)
  data.frame(
    age = round(rnorm(n, 57, 8), 1),
    sex = factor(sample(c("Male", "Female"), n, TRUE)),
    bmi = round(rnorm(n, 27, 5), 1),
    trt = factor(sample(c("Drug A", "Drug B"), n, TRUE)),
    stringsAsFactors = FALSE
  )
}


# ===========================================================================
# plot_forest() — input validation
# ===========================================================================

test_that("plot_forest() aborts on non-data.frame input", {
  expect_error(plot_forest("not a df", 1, 0.5, 2), "data.frame")
})

test_that("plot_forest() aborts when est length != nrow(data)", {
  df <- .fake_forest_df()
  expect_error(
    plot_forest(df, est = c(1, 2), lower = c(0.5, 1), upper = c(2, 3)),
    "length 3"
  )
})

test_that("plot_forest() aborts when lower length != nrow(data)", {
  df <- .fake_forest_df()
  expect_error(
    plot_forest(df,
                est   = c(NA, 1.52, 1.43),
                lower = c(0.5, 1.18),          # wrong length
                upper = c(NA, 1.96, 1.85)),
    "length 3"
  )
})

test_that("plot_forest() aborts when upper length != nrow(data)", {
  df <- .fake_forest_df()
  expect_error(
    plot_forest(df,
                est   = c(NA, 1.52, 1.43),
                lower = c(NA, 1.18, 1.11),
                upper = c(1.96, 1.85)),         # wrong length
    "length 3"
  )
})

test_that("plot_forest() aborts when indent length != nrow(data)", {
  df <- .fake_forest_df()
  expect_error(
    plot_forest(df,
                est    = c(NA, 1.52, 1.43),
                lower  = c(NA, 1.18, 1.11),
                upper  = c(NA, 1.96, 1.85),
                indent = c(0L, 1L)),             # length 2, need 3
    "length"
  )
})

test_that("plot_forest() aborts when bold_label length != nrow(data)", {
  df <- .fake_forest_df()
  expect_error(
    plot_forest(df,
                est        = c(NA, 1.52, 1.43),
                lower      = c(NA, 1.18, 1.11),
                upper      = c(NA, 1.96, 1.85),
                bold_label = c(TRUE, FALSE)),    # length 2, need 3
    "length"
  )
})

test_that("plot_forest() aborts when ci_col is wrong vector length", {
  df <- .fake_forest_df()
  expect_error(
    plot_forest(df,
                est    = c(NA, 1.52, 1.43),
                lower  = c(NA, 1.18, 1.11),
                upper  = c(NA, 1.96, 1.85),
                ci_col = c("black", "red")),     # length 2, need 1 or 3
    "length"
  )
})

test_that("plot_forest() aborts when p_cols not in data", {
  df <- .fake_forest_df()
  expect_error(
    plot_forest(df,
                est    = c(NA, 1.52, 1.43),
                lower  = c(NA, 1.18, 1.11),
                upper  = c(NA, 1.96, 1.85),
                p_cols = "nonexistent"),
    "nonexistent"
  )
})

test_that("plot_forest() aborts when save=TRUE and dest=NULL", {
  df <- .fake_forest_df()
  expect_error(
    plot_forest(df,
                est   = c(NA, 1.52, 1.43),
                lower = c(NA, 1.18, 1.11),
                upper = c(NA, 1.96, 1.85),
                save  = TRUE, dest = NULL),
    "dest"
  )
})

test_that("plot_forest() aborts when bold_p is not logical", {
  df <- .fake_forest_df()
  expect_error(
    plot_forest(df,
                est    = c(NA, 1.52, 1.43),
                lower  = c(NA, 1.18, 1.11),
                upper  = c(NA, 1.96, 1.85),
                bold_p = "yes"),
    "logical"
  )
})

test_that("plot_forest() aborts when ci_column < 2", {
  df <- .fake_forest_df()
  expect_error(
    plot_forest(df,
                est       = c(NA, 1.52, 1.43),
                lower     = c(NA, 1.18, 1.11),
                upper     = c(NA, 1.96, 1.85),
                ci_column = 1L),
    "ci_column"
  )
})

test_that("plot_forest() aborts when ci_column > ncol(data) + 1", {
  df <- .fake_forest_df()   # 2 cols
  expect_error(
    plot_forest(df,
                est       = c(NA, 1.52, 1.43),
                lower     = c(NA, 1.18, 1.11),
                upper     = c(NA, 1.96, 1.85),
                ci_column = 4L),  # max allowed = 3
    "ci_column"
  )
})

test_that("plot_forest() aborts when header length != ncol(data) + 2", {
  df <- .fake_forest_df()   # 2 orig cols → nc_final = 4
  expect_error(
    plot_forest(df,
                est    = c(NA, 1.52, 1.43),
                lower  = c(NA, 1.18, 1.11),
                upper  = c(NA, 1.96, 1.85),
                header = c("Label", "Gap")),  # need 4
    "length"
  )
})

test_that("plot_forest() aborts when align length != ncol(data) + 2", {
  df <- .fake_forest_df()   # nc_final = 4
  expect_error(
    plot_forest(df,
                est   = c(NA, 1.52, 1.43),
                lower = c(NA, 1.18, 1.11),
                upper = c(NA, 1.96, 1.85),
                align = c(-1L, 0L)),  # need 4
    "length"
  )
})


# ===========================================================================
# .fp_build_data() — pure data processing helper
# ===========================================================================

test_that(".fp_build_data() inserts gap_ci and OR (95% CI) columns", {
  df  <- data.frame(item = c("A", "B"), stringsAsFactors = FALSE)
  out <- ukbflow:::.fp_build_data(
    df, c(1.2, NA), c(0.9, NA), c(1.6, NA),
    indent = c(0L, 0L),
    p_cols = NULL, p_digits = 3L, ci_digits = 2L, ci_sep = ", ", ci_column = 2L
  )
  expect_true("gap_ci"       %in% names(out$data))
  expect_true("OR (95% CI)"  %in% names(out$data))
})

test_that(".fp_build_data() formats non-NA est as 'est (lower, upper)'", {
  df  <- data.frame(item = c("A", "B"), stringsAsFactors = FALSE)
  out <- ukbflow:::.fp_build_data(
    df, c(1.52, NA), c(1.18, NA), c(1.96, NA),
    indent = c(0L, 0L),
    p_cols = NULL, p_digits = 3L, ci_digits = 2L, ci_sep = ", ", ci_column = 2L
  )
  ci_col <- out$data[["OR (95% CI)"]]
  expect_equal(ci_col[1L], "1.52 (1.18, 1.96)")
  expect_equal(ci_col[2L], "")
})

test_that(".fp_build_data() applies indent as non-breaking spaces", {
  df  <- data.frame(item = c("Parent", "Child"), stringsAsFactors = FALSE)
  out <- ukbflow:::.fp_build_data(
    df, c(NA, 1.2), c(NA, 0.9), c(NA, 1.6),
    indent = c(0L, 1L),
    p_cols = NULL, p_digits = 3L, ci_digits = 2L, ci_sep = ", ", ci_column = 2L
  )
  expect_equal(out$data[[1L]][1L], "Parent")
  expect_true(startsWith(out$data[[1L]][2L], "\u00a0\u00a0"))
})

test_that(".fp_build_data() formats p_cols numeric to string and caches originals", {
  df  <- data.frame(item = c("A", "B"), p = c(0.031, NA_real_),
                    stringsAsFactors = FALSE)
  out <- ukbflow:::.fp_build_data(
    df, c(1, NA), c(0.5, NA), c(2, NA),
    indent = c(0L, 0L),
    p_cols = "p", p_digits = 3L, ci_digits = 2L, ci_sep = ", ", ci_column = 2L
  )
  expect_equal(out$data$p[1L], "0.031")
  expect_equal(out$data$p[2L], "")
  expect_equal(out$p_numeric$p[1L], 0.031)
})

test_that(".fp_build_data() clips p < 0.001 to '<0.001'", {
  df  <- data.frame(item = "A", p = 0.00001, stringsAsFactors = FALSE)
  out <- ukbflow:::.fp_build_data(
    df, 1.2, 0.9, 1.6,
    indent = 0L,
    p_cols = "p", p_digits = 3L, ci_digits = 2L, ci_sep = ", ", ci_column = 2L
  )
  expect_equal(out$data$p[1L], "<0.001")
})

test_that(".fp_build_data() aborts when ci_column < 2", {
  df <- data.frame(item = "A", stringsAsFactors = FALSE)
  expect_error(
    ukbflow:::.fp_build_data(
      df, 1, 0.5, 2, indent = 0L,
      p_cols = NULL, p_digits = 3L, ci_digits = 2L, ci_sep = ", ", ci_column = 1L
    ),
    ">= 2"
  )
})

test_that(".fp_build_data() maps p_col indices correctly when p_col is right of ci_column", {
  # data: item | n | p  →  ci_column = 2 → gap_ci inserted at 2
  # original p is col 3, after insert becomes col 5
  df  <- data.frame(item = "A", n = "10/100", p = 0.05, stringsAsFactors = FALSE)
  out <- ukbflow:::.fp_build_data(
    df, 1, 0.5, 2, indent = 0L,
    p_cols = "p", p_digits = 3L, ci_digits = 2L, ci_sep = ", ", ci_column = 2L
  )
  # orig p is col 3, >= ci_column (2) → idx = 3 + 2 = 5
  expect_equal(out$p_col_idxs, 5L)
})


# ===========================================================================
# .fp_theme() — theme builder
# ===========================================================================

test_that(".fp_theme() returns a list for 'default'", {
  th <- ukbflow:::.fp_theme("default")
  expect_true(is.list(th))
})

test_that(".fp_theme() result has expected ci element", {
  th <- ukbflow:::.fp_theme("default")
  # forest_theme stores CI settings under $ci
  expect_true(!is.null(th$ci) || !is.null(th$ci_col))
})

test_that(".fp_theme() warns on unknown preset and falls back to default", {
  expect_warning(
    ukbflow:::.fp_theme("unknown_theme"),
    "Unknown theme"
  )
})

test_that(".fp_theme() different ci_Theight values do not error", {
  expect_no_error(ukbflow:::.fp_theme("default", ci_Theight = 0.1))
  expect_no_error(ukbflow:::.fp_theme("default", ci_Theight = 0.5))
})


# ===========================================================================
# plot_tableone() — input validation
# ===========================================================================

test_that("plot_tableone() aborts on non-data.frame input", {
  expect_error(plot_tableone("not a df", vars = "age"), "data.frame")
})

test_that("plot_tableone() aborts when vars not in data", {
  df <- .fake_t1_df()
  expect_error(
    plot_tableone(df, vars = "nonexistent", save = FALSE),
    "missing"
  )
})

test_that("plot_tableone() aborts when strata not in data", {
  df <- .fake_t1_df()
  expect_error(
    plot_tableone(df, vars = "age", strata = "missing_col", save = FALSE),
    "missing_col"
  )
})

test_that("plot_tableone() aborts when save=TRUE and dest=NULL", {
  df <- .fake_t1_df()
  expect_error(
    plot_tableone(df, vars = "age", strata = "trt",
                  save = TRUE, dest = NULL),
    "dest"
  )
})

test_that("plot_tableone() warns and disables add_p when strata=NULL", {
  skip_on_cran()
  df <- .fake_t1_df()
  # cli::cli_warn() produces a warning, not a message
  expect_warning(
    plot_tableone(df, vars = "age", strata = NULL, add_p = TRUE, save = FALSE),
    "add_p"
  )
})

test_that("plot_tableone() warns and disables add_smd when strata=NULL", {
  skip_on_cran()
  df <- .fake_t1_df()
  # cli::cli_warn() produces a warning, not a message.
  # Disable add_p to avoid a second unrelated warning leaking out.
  expect_warning(
    plot_tableone(df, vars = "age", strata = NULL,
                  add_p = FALSE, add_smd = TRUE, save = FALSE),
    "add_smd"
  )
})

test_that("plot_tableone() aborts when exclude_labels is not character", {
  df <- .fake_t1_df()
  expect_error(
    plot_tableone(df, vars = "age", strata = "trt",
                  exclude_labels = 1L, save = FALSE),
    "character"
  )
})


# ===========================================================================
# .t1_smd_continuous() — Cohen's d helper
# ===========================================================================

test_that(".t1_smd_continuous() returns non-negative numeric for two groups", {
  set.seed(1)
  x   <- c(rnorm(50, mean = 0), rnorm(50, mean = 1))
  grp <- factor(rep(c("A", "B"), each = 50))
  val <- ukbflow:::.t1_smd_continuous(x, grp, levels(grp))
  expect_true(is.numeric(val))
  expect_gte(val, 0)
})

test_that(".t1_smd_continuous() returns ~0 when groups have same mean", {
  x   <- c(rep(5, 50), rep(5, 50))
  grp <- factor(rep(c("A", "B"), each = 50))
  val <- ukbflow:::.t1_smd_continuous(x, grp, levels(grp))
  # pooled SD = 0 → NA (division by zero guard), or 0 if identical
  expect_true(is.na(val) || val == 0)
})

test_that(".t1_smd_continuous() returns NA when a group has < 2 observations", {
  x   <- c(1, 2, 3)
  grp <- factor(c("A", "A", "B"))   # group B has only 1 obs
  val <- ukbflow:::.t1_smd_continuous(x, grp, levels(grp))
  expect_true(is.na(val))
})


# ===========================================================================
# .t1_smd_categorical() — RMSD helper
# ===========================================================================

test_that(".t1_smd_categorical() returns non-negative numeric", {
  x   <- factor(c("Yes", "No", "Yes", "Yes", "No", "No"))
  grp <- factor(c("A",   "A",  "A",   "B",   "B",  "B"))
  val <- ukbflow:::.t1_smd_categorical(x, grp, levels(grp))
  expect_true(is.numeric(val))
  expect_gte(val, 0)
})

test_that(".t1_smd_categorical() returns NA for single-category variable", {
  x   <- factor(c("Yes", "Yes", "Yes", "Yes"))
  grp <- factor(c("A",   "A",   "B",   "B"))
  val <- ukbflow:::.t1_smd_categorical(x, grp, levels(grp))
  expect_true(is.na(val))
})

test_that(".t1_smd_categorical() returns ~0 when proportions are identical", {
  x   <- factor(c("Yes", "No", "Yes", "No"))
  grp <- factor(c("A",   "A",  "B",   "B"))
  val <- ukbflow:::.t1_smd_categorical(x, grp, levels(grp))
  expect_lt(val, 1e-10)
})


# ===========================================================================
# .t1_compute_smd() — per-variable SMD dispatcher
# ===========================================================================

test_that(".t1_compute_smd() returns one row per variable", {
  set.seed(42)
  df  <- .fake_t1_df()
  res <- ukbflow:::.t1_compute_smd(df, vars = c("age", "sex"), strata = "trt")
  expect_equal(nrow(res), 2L)
  expect_equal(res$variable, c("age", "sex"))
})

test_that(".t1_compute_smd() smd_fmt is character or NA", {
  df  <- .fake_t1_df()
  res <- ukbflow:::.t1_compute_smd(df, vars = c("age", "bmi", "sex"), strata = "trt")
  expect_true(is.character(res$smd_fmt))
})

test_that(".t1_compute_smd() smd_fmt is formatted to 3 decimal places", {
  df  <- .fake_t1_df()
  res <- ukbflow:::.t1_compute_smd(df, vars = "age", strata = "trt")
  fmt <- res$smd_fmt[!is.na(res$smd_fmt)]
  if (length(fmt) > 0L) {
    # e.g. "0.123" — decimal part is exactly 3 digits
    expect_true(all(grepl("^\\d+\\.\\d{3}$", fmt)))
  }
})

# =============================================================================
# extract.R — Dataset field listing and phenotype data extraction
# =============================================================================


#' List all approved fields in the UKB dataset
#'
#' Returns a data.frame of all fields available for extraction in the current
#' UKB project dataset. Fields reflect what has been approved for your project
#' — not all UKB fields are present.
#'
#' Results are cached in the session after the first call. Subsequent calls
#' return instantly from cache. Use \code{refresh = TRUE} to force a new
#' network request (e.g. after switching projects).
#'
#' @param dataset (character) Dataset file name, e.g.
#'   \code{"app12345_20260101.dataset"}. Default: \code{NULL} (auto-detect).
#' @param pattern (character) Optional regex to filter results by
#'   \code{field_name} or \code{title}. Default: \code{NULL}.
#' @param refresh (logical) Force re-fetch from cloud, ignoring cache.
#'   Default: \code{FALSE}.
#'
#' @return A data.frame with columns:
#'   \describe{
#'     \item{field_name}{Full field name as used in extraction, e.g.
#'       \code{"participant.p31"}, \code{"participant.p53_i0"}.}
#'     \item{title}{Human-readable field description, e.g.
#'       \code{"Sex"}, \code{"Date of attending assessment centre | Instance 0"}.}
#'   }
#' @export
#'
#' @examples
#' \dontrun{
#' # List all approved fields
#' extract_ls()
#'
#' # Search by keyword
#' extract_ls(pattern = "cancer")
#' extract_ls(pattern = "p31|p53|p22009")
#'
#' # Force refresh after switching projects
#' extract_ls(refresh = TRUE)
#' }
extract_ls <- function(dataset = NULL, pattern = NULL, refresh = FALSE) {

  # Reason: auto-detect before cache check so cache is keyed by dataset name —
  # prevents returning stale fields after switching datasets or projects
  if (is.null(dataset)) {
    dataset <- .dx_find_dataset()
    cli::cli_inform("Using dataset: {.val {dataset}}")
  }

  # Return from per-dataset cache if available and refresh not requested
  if (!is.null(.ukbflow_cache$fields[[dataset]]) && !refresh) {
    df <- .ukbflow_cache$fields[[dataset]]
    if (!is.null(pattern)) {
      return(df[
        grepl(pattern, df$field_name, perl = TRUE) |
        grepl(pattern, df$title,      perl = TRUE, ignore.case = TRUE),
        , drop = FALSE
      ])
    }
    cli::cli_inform("{nrow(df)} fields available. Assign to a variable or use pattern= to search.")
    return(invisible(df))
  }

  cli::cli_inform("Fetching approved fields... (cached after first call)")

  result <- .dx_list_fields_raw(dataset)
  if (!result$success) {
    cli::cli_abort("Failed to list fields: {result$stderr}", call = NULL)
  }

  df <- .dx_parse_fields(result$stdout)

  # Store in per-dataset cache slot
  .ukbflow_cache$fields[[dataset]] <- df

  if (!is.null(pattern)) {
    df <- df[
      grepl(pattern, df$field_name, perl = TRUE) |
      grepl(pattern, df$title,      perl = TRUE, ignore.case = TRUE),
      , drop = FALSE
    ]
    return(df)
  }

  # Reason: returning all 29,000+ rows visibly floods the console;
  # return invisibly and show a summary message instead
  cli::cli_inform("{nrow(df)} fields available. Assign to a variable or use pattern= to search.")
  invisible(df)
}


#' Extract phenotype data from a UKB dataset
#'
#' Extracts phenotypic fields from the UKB Research Analysis Platform dataset
#' and returns a \code{data.table}. All instances and arrays are returned for
#' each requested field. Column names are kept as-is (e.g.
#' \code{participant.p53_i0}); use the \code{clean_} series for renaming.
#'
#' @param field_id (integer) Vector of UKB Field IDs to extract, e.g.
#'   \code{c(31, 53, 22189)}. \code{eid} is always included automatically.
#' @param dataset (character) Dataset file name. Default: \code{NULL}
#'   (auto-detect from project root).
#' @param timeout (integer) Extraction timeout in seconds. Default: \code{300}.
#'
#' @return A \code{data.table} with one row per participant. Column names
#'   follow the \code{participant.p<id>_i<n>_a<m>} convention.
#'   Fields not found are skipped with a warning.
#' @export
#'
#' @examples
#' \dontrun{
#' df <- extract_pheno(c(31, 53, 21022))
#' df <- extract_pheno(c(31, 53, 20002), dataset = "app12345_20260101.dataset")
#' }
extract_pheno <- function(field_id, dataset = NULL, timeout = 300) {

  .assert_on_rap()

  field_id <- .assert_integer_ids(field_id)

  dest <- tempfile(fileext = ".csv")
  on.exit(unlink(dest), add = TRUE)

  # Auto-detect dataset
  if (is.null(dataset)) {
    dataset <- .dx_find_dataset()
  }
  cli::cli_inform("Using dataset: {.val {dataset}}")

  # Get approved fields list (uses session cache after first call)
  fields_df <- extract_ls(dataset = dataset)

  # Match field IDs to exact column names
  match_result <- .dx_match_fields(field_id, fields_df)
  matched      <- match_result$matched
  unmatched    <- match_result$unmatched

  if (length(matched) == 0) {
    cli::cli_abort(
      "No matching fields found. Run extract_ls() to see available fields.",
      call = NULL
    )
  }

  # Report matched fields
  n_cols_total <- sum(vapply(matched, `[[`, integer(1), "n_cols"))
  cli::cli_inform(
    "Matched {length(matched)}/{length(field_id)} field{?s} ({n_cols_total} column{?s})"
  )

  # Reason: cap per-field listing to avoid console flood on large extractions
  show_all    <- length(matched) <= 10
  preview_n   <- if (show_all) length(matched) else 5

  for (m in matched[seq_len(preview_n)]) {
    col_label <- if (m$n_cols == 1) "1 col" else paste0(m$n_cols, " cols")
    cli::cli_inform("  p{m$field_id}  {m$title}  [{col_label}]")
  }
  if (!show_all) {
    cli::cli_inform(
      "  ... and {length(matched) - preview_n} more fields. Use extract_ls() to review all."
    )
  }

  # Warn about unmatched fields
  if (length(unmatched) > 0) {
    cli::cli_warn(
      "Field ID{?s} not found and skipped: {.val {unmatched}}"
    )
  }

  # Build final field list (eid always first)
  all_fields <- c(
    "participant.eid",
    unlist(lapply(matched, `[[`, "field_names"))
  )

  # Run extraction — output saved to disk
  cli::cli_inform("Extracting data...")
  result <- .dx_extract_run(dataset, all_fields, dest, timeout = timeout)

  if (!result$success) {
    cli::cli_abort("Extraction failed: {result$stderr}", call = NULL)
  }

  cli::cli_inform("Saved: {.file {dest}}")

  # Read with data.table for performance on large UKB files
  # Reason: integer64 = "double" avoids bit64 compatibility issues;
  # UKB eids are 7-digit integers, well within double precision range
  dt <- data.table::fread(dest, data.table = TRUE, integer64 = "double")

  cli::cli_inform(
    "{nrow(dt)} rows x {ncol(dt)} cols (incl. eid)"
  )

  dt
}


#' Submit a large-scale phenotype extraction job via table-exporter
#'
#' Submits an asynchronous table-exporter job on the DNAnexus Research
#' Analysis Platform for large-scale phenotype extraction. Use this instead
#' of \code{extract_pheno()} when extracting many fields (e.g. 50+).
#'
#' The job runs on the cloud and typically completes in 20-40 minutes.
#' Monitor progress and retrieve results using the \code{job_} series.
#'
#' @param field_id (integer) Vector of UKB Field IDs to extract.
#'   \code{eid} is always included automatically.
#' @param dataset (character) Dataset file name. Default: \code{NULL}
#'   (auto-detect from project root).
#' @param file (character) Output file name on the cloud (without extension),
#'   e.g. \code{"ad_cscc_pheno"}. Default: \code{NULL} (auto-generate as
#'   \code{"ukb_pheno_YYYYMMDD_HHMMSS"} to avoid same-day collisions).
#' @param instance_type (character) DNAnexus instance type, e.g.
#'   \code{"mem1_ssd1_v2_x16"}. Default: \code{NULL} (auto-select:
#'   \code{x4} for up to 20 cols, \code{x8} for up to 100 cols,
#'   \code{x16} for up to 500 cols, \code{x36} for more than 500 cols).
#' @param priority (character) Job scheduling priority. \code{"low"}
#'   (recommended, cheaper) or \code{"high"} (faster queue). Default:
#'   \code{"low"}.
#'
#' @return Invisibly returns the job ID string (e.g. \code{"job-XXXX"}).
#' @export
#'
#' @examples
#' \dontrun{
#' job_id <- extract_batch(core_field_ids)
#' job_id <- extract_batch(core_field_ids, file = "ad_cscc_pheno")
#' job_id <- extract_batch(core_field_ids, priority = "high")
#' # Monitor: job_status(job_id)
#' # Download: job_result(job_id, dest = "data/pheno.csv")
#' }
extract_batch <- function(field_id, dataset = NULL, file = NULL,
                          instance_type = NULL, priority = c("low", "high")) {
  priority <- match.arg(priority)

  field_id <- .assert_integer_ids(field_id)

  # Auto-detect dataset
  if (is.null(dataset)) {
    dataset <- .dx_find_dataset()
  }
  cli::cli_inform("Using dataset: {.val {dataset}}")

  # Get approved fields list (uses session cache)
  fields_df <- extract_ls(dataset = dataset)

  # Match field IDs to exact column names
  match_result <- .dx_match_fields(field_id, fields_df)
  matched      <- match_result$matched
  unmatched    <- match_result$unmatched

  if (length(matched) == 0) {
    cli::cli_abort(
      "No matching fields found. Run extract_ls() to see available fields.",
      call = NULL
    )
  }

  # Report matched fields
  n_cols_total <- sum(vapply(matched, `[[`, integer(1), "n_cols"))
  cli::cli_inform(
    "Matched {length(matched)}/{length(field_id)} field{?s} ({n_cols_total} column{?s})"
  )

  show_all  <- length(matched) <= 10
  preview_n <- if (show_all) length(matched) else 5
  for (m in matched[seq_len(preview_n)]) {
    col_label <- if (m$n_cols == 1) "1 col" else paste0(m$n_cols, " cols")
    cli::cli_inform("  p{m$field_id}  {m$title}  [{col_label}]")
  }
  if (!show_all) {
    cli::cli_inform(
      "  ... and {length(matched) - preview_n} more fields. Use extract_ls() to review all."
    )
  }
  if (length(unmatched) > 0) {
    cli::cli_warn("Field ID{?s} not found and skipped: {.val {unmatched}}")
  }

  # Build field names: strip participant. prefix (table-exporter format)
  # Reason: table-exporter expects "eid", "p31", "p53_i0" without entity prefix
  all_fields <- c(
    "eid",
    gsub("^participant\\.", "", unlist(lapply(matched, `[[`, "field_names")))
  )

  # Write and upload fields file
  fields_file <- tempfile(fileext = ".txt")
  on.exit(unlink(fields_file), add = TRUE)
  # Reason: open in binary mode to force LF-only line endings — on Windows
  # text-mode writeLines() produces CRLF, which table-exporter rejects as
  # "invalid characters in field names"
  con <- file(fields_file, open = "wb")
  writeLines(all_fields, con = con, sep = "\n")
  close(con)

  cli::cli_inform("Uploading field list...")
  file_id <- .dx_upload_file(fields_file)

  # Auto output name: add HMS to avoid same-day collisions
  output <- if (!is.null(file)) file else
    paste0("ukb_pheno_", format(Sys.time(), "%Y%m%d_%H%M%S"))

  # Auto instance type based on column count (four tiers)
  instance_type <- if (!is.null(instance_type)) instance_type else
    if (n_cols_total > 500) "mem1_ssd1_v2_x36" else
    if (n_cols_total > 100) "mem1_ssd1_v2_x16" else
    if (n_cols_total > 20)  "mem1_ssd1_v2_x8"  else
                            "mem1_ssd1_v2_x4"

  # Submit job
  cli::cli_inform("Submitting table-exporter job...")
  job_id <- .dx_run_table_exporter(dataset, file_id, output, instance_type,
                                    priority = priority)

  cli::cli_inform("Job submitted: {.val {job_id}}")
  cli::cli_inform("  Output  : {output}.csv  (cloud)")
  cli::cli_inform("  Instance: {instance_type}")

  invisible(job_id)
}

# =============================================================================
# test-integration-decode.R — Integration tests for decode_ series
# Requires real dx-toolkit, token, and network connection (decode_names),
# and local metadata files (decode_values).
# Run manually before release: devtools::test(filter = "integration-decode")
# =============================================================================

skip_on_cran()

# Download metadata files once for the decode_values() tests (RAP + token required)
META_DIR <- local({
  if (!nzchar(Sys.getenv("DX_API_TOKEN")) || !ukbflow:::.is_on_rap()) {
    NULL
  } else {
    d <- withr::local_tempdir()
    suppressMessages({
      fetch_file("Showcase metadata/field.tsv",    dest_dir = d, overwrite = TRUE)
      fetch_file("Showcase metadata/esimpint.tsv", dest_dir = d, overwrite = TRUE)
    })
    d
  }
})


# ===========================================================================
# decode_names() — requires live extract_ls() to populate cache
# ===========================================================================

test_that("decode_names() renames participant.p31 to sex via live cache", {
  .skip_if_no_rap()
  .ukbflow_cache$dataset <- NULL
  .ukbflow_cache$fields  <- NULL

  df <- data.frame(`participant.eid` = 1L, `participant.p31` = 1L,
                   check.names = FALSE)
  result <- suppressMessages(decode_names(df))

  expect_true("eid" %in% names(result))
  expect_true("sex" %in% names(result))
  expect_false("participant.p31" %in% names(result))
})

test_that("decode_names() renames all columns from extract_pheno() format", {
  .skip_if_no_rap()
  .ukbflow_cache$dataset <- NULL
  .ukbflow_cache$fields  <- NULL

  df <- data.frame(
    `participant.eid`    = 1L,
    `participant.p31`    = 1L,
    `participant.p21022` = 65L,
    `participant.p22189` = -3.94,
    check.names = FALSE
  )
  result <- suppressMessages(decode_names(df))

  expect_true("eid"    %in% names(result))
  expect_true("sex"    %in% names(result))
  expect_true("age_at_recruitment" %in% names(result))
  expect_true("townsend_deprivation_index_at_recruitment" %in% names(result))
})

test_that("decode_names() renames p53 instance columns with _i0.._i3 suffixes", {
  .skip_if_no_rap()
  .ukbflow_cache$dataset <- NULL
  .ukbflow_cache$fields  <- NULL

  df <- data.frame(
    `participant.eid`    = 1L,
    `participant.p53_i0` = "2008-01-01",
    `participant.p53_i1` = NA_character_,
    check.names = FALSE
  )
  result <- suppressMessages(decode_names(df))

  new_names <- names(result)
  expect_true(any(grepl("_i0$", new_names)))
  expect_true(any(grepl("_i1$", new_names)))
})

test_that("decode_names() warms session cache as side effect", {
  .skip_if_no_rap()
  .ukbflow_cache$dataset <- NULL
  .ukbflow_cache$fields  <- NULL

  df <- data.frame(`participant.p31` = 1L, check.names = FALSE)
  suppressMessages(decode_names(df))

  expect_false(is.null(.ukbflow_cache$fields))
  expect_gt(length(.ukbflow_cache$fields), 0L)                         # list has at least one dataset entry
  dataset_key <- names(.ukbflow_cache$fields)[1L]
  expect_gt(nrow(.ukbflow_cache$fields[[dataset_key]]), 0L)            # that entry has rows
})


# ===========================================================================
# decode_values() — requires local metadata files (no network after download)
# ===========================================================================

# Clear metadata cache before each test to ensure fresh reads
.clear_meta_cache <- function() {
  .ukbflow_cache$field_meta <- NULL
  .ukbflow_cache$esimpint   <- NULL
}

test_that("decode_values() decodes p31: 0 → Female, 1 → Male", {
  .skip_if_no_rap()
  if (is.null(META_DIR)) skip("Metadata files not available.")
  .clear_meta_cache()
  df <- data.frame(p31 = c(0L, 1L))
  result <- suppressMessages(decode_values(df, metadata_dir = META_DIR))
  expect_equal(result$p31, c("Female", "Male"))
})

test_that("decode_values() caches field_meta and esimpint after first call", {
  .skip_if_no_rap()
  if (is.null(META_DIR)) skip("Metadata files not available.")
  .clear_meta_cache()
  df <- data.frame(p31 = 1L)
  suppressMessages(decode_values(df, metadata_dir = META_DIR))
  expect_false(is.null(.ukbflow_cache$field_meta))
  expect_false(is.null(.ukbflow_cache$esimpint))
})

test_that("decode_values() leaves continuous field p22189 (Townsend) unchanged", {
  .skip_if_no_rap()
  if (is.null(META_DIR)) skip("Metadata files not available.")
  .clear_meta_cache()
  df <- data.frame(p31 = 1L, p22189 = -3.94)
  result <- suppressMessages(decode_values(df, metadata_dir = META_DIR))
  expect_equal(result$p22189, -3.94)
  expect_true(is.numeric(result$p22189))
})

test_that("decode_values() decodes smoking status p20116: 0/1/2 → labels", {
  .skip_if_no_rap()
  if (is.null(META_DIR)) skip("Metadata files not available.")
  .clear_meta_cache()
  df <- data.frame(p20116_i0 = c(0L, 1L, 2L))
  result <- suppressMessages(decode_values(df, metadata_dir = META_DIR))
  expect_true(all(result$p20116_i0 %in% c("Never", "Previous", "Current")))
})

test_that("decode_values() returns NA for code absent from encoding table", {
  .skip_if_no_rap()
  if (is.null(META_DIR)) skip("Metadata files not available.")
  .clear_meta_cache()
  df <- data.frame(p31 = c(0L, 1L, 99L))
  result <- suppressMessages(decode_values(df, metadata_dir = META_DIR))
  expect_true(is.na(result$p31[3]))
  expect_equal(result$p31[1], "Female")
})

test_that("decode_values() then decode_names() produces clean final output", {
  .skip_if_no_rap()
  if (is.null(META_DIR)) skip("Metadata files not available.")
  .clear_meta_cache()
  .ukbflow_cache$dataset <- NULL
  .ukbflow_cache$fields  <- NULL

  df <- data.frame(
    `participant.eid`    = c(1L, 2L),
    `participant.p31`    = c(1L, 0L),
    `participant.p21022` = c(65L, 52L),
    check.names = FALSE
  )

  result <- df |>
    decode_values(metadata_dir = META_DIR) |>
    suppressMessages() |>
    decode_names() |>
    suppressMessages()

  expect_true("eid"  %in% names(result))
  expect_true("sex"  %in% names(result))
  expect_true("age_at_recruitment" %in% names(result))
  expect_equal(result$sex, c("Male", "Female"))
})

# =============================================================================
# test-integration-extract.R — Integration tests for extract_ series
# Requires real dx-toolkit, token, and network connection
# Run manually before release: devtools::test(filter = "integration-extract")
# =============================================================================

skip_on_cran()

# ===========================================================================
# extract_ls()
# ===========================================================================

test_that("extract_ls() returns a data.frame with correct columns", {
  .skip_if_no_rap()
  result <- suppressMessages(extract_ls())
  expect_s3_class(result, "data.frame")
  expect_named(result, c("field_name", "title"))
})

test_that("extract_ls() returns at least one field", {
  .skip_if_no_rap()
  result <- suppressMessages(extract_ls())
  expect_gt(nrow(result), 0)
})

test_that("extract_ls() field_name values follow participant. format", {
  .skip_if_no_rap()
  result <- suppressMessages(extract_ls())
  expect_true(all(grepl("^participant\\.", result$field_name)))
})

test_that("extract_ls() title column has no fully-missing values", {
  .skip_if_no_rap()
  result <- suppressMessages(extract_ls())
  expect_lt(sum(is.na(result$title)), nrow(result))
})

test_that("extract_ls() populates per-dataset session cache after first call", {
  .skip_if_no_rap()
  .ukbflow_cache$fields <- NULL
  suppressMessages(extract_ls())
  expect_gt(length(.ukbflow_cache$fields), 0L)
})

test_that("extract_ls() returns from cache on second call (no re-fetch)", {
  .skip_if_no_rap()
  suppressMessages(extract_ls(refresh = TRUE))          # warm cache with real data
  dataset_key <- names(.ukbflow_cache$fields)[1]

  # Replace slot with sentinel — a re-fetch would overwrite it
  .ukbflow_cache$fields[[dataset_key]] <- data.frame(
    field_name = "sentinel", title = "sentinel", stringsAsFactors = FALSE
  )
  result <- suppressMessages(extract_ls())
  expect_equal(nrow(result), 1L)                        # served from (sentinel) cache

  .ukbflow_cache$fields <- NULL
  suppressMessages(extract_ls())                        # restore real cache
})

test_that("extract_ls() refresh = TRUE re-fetches and overwrites cache", {
  .skip_if_no_rap()
  suppressMessages(extract_ls(refresh = TRUE))
  dataset_key <- names(.ukbflow_cache$fields)[1]
  expect_gt(nrow(.ukbflow_cache$fields[[dataset_key]]), 1L)
})

test_that("extract_ls() pattern filtering returns a non-empty subset", {
  .skip_if_no_rap()
  result_all      <- suppressMessages(extract_ls())
  result_filtered <- extract_ls(pattern = "^participant\\.p31(_|$)")
  expect_lt(nrow(result_filtered), nrow(result_all))
  expect_gt(nrow(result_filtered), 0L)
  expect_true(all(grepl("participant\\.p31", result_filtered$field_name)))
})

test_that("extract_ls() pattern search is case-insensitive on title", {
  .skip_if_no_rap()
  result <- extract_ls(pattern = "SEX")
  expect_gt(nrow(result), 0L)
})

test_that("extract_ls() returns empty data.frame for unmatched pattern", {
  .skip_if_no_rap()
  result <- extract_ls(pattern = "xyzzy_no_such_field_12345")
  expect_equal(nrow(result), 0L)
})

# ===========================================================================
# extract_pheno() and extract_batch() — not integration-tested
#
# Both functions trigger actual cloud computation (extract_dataset / table-
# exporter jobs) that incur cost and take several minutes to complete.
# They are covered exhaustively by mock-based unit tests in test-extract.R.
# If a real smoke-test is needed, run the following interactively:
#
#   df <- extract_pheno(c(31, 21022))   # RAP only
#   job_id <- extract_batch(c(31, 21022), priority = "low")
# ===========================================================================

# =============================================================================
# test-extract.R — Unit tests for extract_ series (mock-based, no network)
# =============================================================================
.skip_if_no_mockery()

# Helper: build a fake .dx_run() result
.fake_dx <- function(stdout = "", stderr = "", status = 0) {
  list(stdout = stdout, stderr = stderr, status = status, success = status == 0)
}

# Fake fields data.frame (representative sample of all field types)
.fake_fields_df <- function() {
  data.frame(
    field_name = c(
      "participant.eid",
      "participant.p31",
      "participant.p53_i0", "participant.p53_i1",
      "participant.p53_i2", "participant.p53_i3",
      "participant.p22189",
      "participant.p20002_i0_a0", "participant.p20002_i0_a1",
      "participant.p22009_a1",    "participant.p22009_a2",
      "participant.p41202"
    ),
    title = c(
      "Participant ID",
      "Sex",
      "Date of attending assessment centre | Instance 0",
      "Date of attending assessment centre | Instance 1",
      "Date of attending assessment centre | Instance 2",
      "Date of attending assessment centre | Instance 3",
      "Townsend deprivation index at recruitment",
      "Non-cancer illness code, self-reported | Instance 0 | Array 0",
      "Non-cancer illness code, self-reported | Instance 0 | Array 1",
      "Genetic principal components | Array 1",
      "Genetic principal components | Array 2",
      "Diagnoses - main ICD10"
    ),
    stringsAsFactors = FALSE
  )
}

# Helper: populate per-dataset session cache with fake fields
.set_fake_cache <- function(dataset = "app12345.dataset") {
  if (is.null(.ukbflow_cache$fields)) .ukbflow_cache$fields <- list()
  .ukbflow_cache$fields[[dataset]] <- .fake_fields_df()
}

# Helper: clear session cache entirely
.clear_cache <- function() {
  .ukbflow_cache$fields <- NULL
}

# ===========================================================================
# .dx_parse_fields() — pure function, no mocking needed
# ===========================================================================

test_that(".dx_parse_fields() returns empty df for empty stdout", {
  result <- ukbflow:::.dx_parse_fields("")
  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 0)
  expect_named(result, c("field_name", "title"))
})

test_that(".dx_parse_fields() parses tab-separated output correctly", {
  stdout <- "participant.p31\tSex\nparticipant.p53_i0\tDate of attending | Instance 0\n"
  result <- ukbflow:::.dx_parse_fields(stdout)
  expect_equal(nrow(result), 2)
  expect_equal(result$field_name[1], "participant.p31")
  expect_equal(result$title[1], "Sex")
})

test_that(".dx_parse_fields() handles missing title gracefully", {
  stdout <- "participant.p31\n"
  result <- ukbflow:::.dx_parse_fields(stdout)
  expect_equal(nrow(result), 1)
  expect_true(is.na(result$title[1]))
})

# ===========================================================================
# .dx_match_fields() — pure function, no mocking needed
# ===========================================================================

test_that(".dx_match_fields() matches simple field exactly", {
  result <- ukbflow:::.dx_match_fields(31L, .fake_fields_df())
  expect_length(result$matched, 1)
  expect_equal(result$matched[[1]]$field_id, 31L)
  expect_equal(result$matched[[1]]$n_cols, 1L)
  expect_length(result$unmatched, 0)
})

test_that(".dx_match_fields() expands all instances for p53", {
  result <- ukbflow:::.dx_match_fields(53L, .fake_fields_df())
  expect_equal(result$matched[[1]]$n_cols, 4L)
  expect_length(result$matched[[1]]$field_names, 4L)
})

test_that(".dx_match_fields() does NOT match p31 to p310xx fields", {
  # Add a decoy field to the fake df
  df <- .fake_fields_df()
  df <- rbind(df, data.frame(
    field_name = "participant.p31000_i0",
    title = "Some other field",
    stringsAsFactors = FALSE
  ))
  result <- ukbflow:::.dx_match_fields(31L, df)
  matched_names <- result$matched[[1]]$field_names
  expect_false(any(grepl("p31000", matched_names)))
  expect_equal(result$matched[[1]]$n_cols, 1L)
})

test_that(".dx_match_fields() returns unmatched for missing field", {
  result <- ukbflow:::.dx_match_fields(c(31L, 189L), .fake_fields_df())
  expect_length(result$matched, 1)
  expect_equal(result$unmatched, 189L)
})

test_that(".dx_match_fields() strips title instance suffix correctly", {
  result <- ukbflow:::.dx_match_fields(53L, .fake_fields_df())
  expect_equal(result$matched[[1]]$title, "Date of attending assessment centre")
})

# ===========================================================================
# extract_ls()
# ===========================================================================

test_that("extract_ls() fetches fields and caches on first call", {
  .clear_cache()
  on.exit(.clear_cache())

  mockery::stub(extract_ls, ".dx_find_dataset",
                function() "app12345.dataset")
  mockery::stub(extract_ls, ".dx_list_fields_raw",
                function(dataset, ...) .fake_dx(
                  stdout = "participant.p31\tSex\nparticipant.p22189\tTownsend\n"
                ))

  result <- suppressMessages(extract_ls())
  expect_false(is.null(.ukbflow_cache$fields[["app12345.dataset"]]))
  expect_equal(nrow(.ukbflow_cache$fields[["app12345.dataset"]]), 2)
})

test_that("extract_ls() returns from cache on second call", {
  .clear_cache()
  on.exit(.clear_cache())

  call_count <- 0L
  mockery::stub(extract_ls, ".dx_find_dataset", function() "app12345.dataset")
  mockery::stub(extract_ls, ".dx_list_fields_raw", function(...) {
    call_count <<- call_count + 1L
    .fake_dx(stdout = "participant.p31\tSex\n")
  })

  suppressMessages(extract_ls())   # first call: fetches and caches
  suppressMessages(extract_ls())   # second call: served from cache
  expect_equal(call_count, 1L)
})

test_that("extract_ls() refresh = TRUE re-fetches from network", {
  .set_fake_cache()
  on.exit(.clear_cache())

  mockery::stub(extract_ls, ".dx_list_fields_raw",
                function(...) .fake_dx(
                  stdout = "participant.p31\tSex\n"
                ))
  mockery::stub(extract_ls, ".dx_find_dataset", function() "app12345.dataset")

  suppressMessages(extract_ls(refresh = TRUE))
  expect_equal(nrow(.ukbflow_cache$fields[["app12345.dataset"]]), 1)
})

test_that("extract_ls() pattern filters results", {
  .set_fake_cache()
  on.exit(.clear_cache())

  result <- extract_ls(dataset = "app12345.dataset", pattern = "p31")
  expect_true(all(grepl("p31", result$field_name) |
                    grepl("p31", result$title, ignore.case = TRUE)))
})

test_that("extract_ls() pattern search is case-insensitive on title", {
  .set_fake_cache()
  on.exit(.clear_cache())

  result <- extract_ls(dataset = "app12345.dataset", pattern = "SEX")
  expect_gt(nrow(result), 0)
})

test_that("extract_ls() returns invisible and message when no pattern", {
  .set_fake_cache()
  on.exit(.clear_cache())

  expect_message(res <- extract_ls(dataset = "app12345.dataset"), "fields available")
  expect_s3_class(res, "data.frame")
})

test_that("extract_ls() caches per dataset — different datasets don't share cache", {
  .clear_cache()
  on.exit(.clear_cache())

  call_count <- 0L
  mockery::stub(extract_ls, ".dx_list_fields_raw", function(...) {
    call_count <<- call_count + 1L
    .fake_dx(stdout = "participant.p31\tSex\n")
  })

  suppressMessages(extract_ls(dataset = "app11111.dataset"))
  suppressMessages(extract_ls(dataset = "app22222.dataset"))
  expect_equal(call_count, 2L)   # each dataset triggers its own fetch
  suppressMessages(extract_ls(dataset = "app11111.dataset"))
  expect_equal(call_count, 2L)   # cache hit — no new fetch
})

test_that("extract_ls() throws error when .dx_list_fields_raw fails", {
  .clear_cache()
  on.exit(.clear_cache())

  mockery::stub(extract_ls, ".dx_find_dataset", function() "app12345.dataset")
  mockery::stub(extract_ls, ".dx_list_fields_raw",
                function(...) .fake_dx(stderr = "Not logged in", status = 1))

  expect_error(suppressMessages(extract_ls()), "Failed to list fields")
})

# ===========================================================================
# extract_pheno()
# ===========================================================================

test_that("extract_pheno() stops when not on RAP", {
  mockery::stub(extract_pheno, ".assert_on_rap", function() cli::cli_abort("This function must be run inside the RAP environment."))
  expect_error(extract_pheno(31), "RAP environment")
})

test_that("extract_pheno() throws error on empty field_id", {
  mockery::stub(extract_pheno, ".assert_on_rap", function() invisible(NULL))
  expect_error(extract_pheno(integer(0)), "non-empty numeric")
})

test_that("extract_pheno() throws error on non-numeric field_id", {
  mockery::stub(extract_pheno, ".assert_on_rap", function() invisible(NULL))
  expect_error(extract_pheno("p31"), "non-empty numeric")
})

test_that("extract_pheno() throws error on NA field_id", {
  mockery::stub(extract_pheno, ".assert_on_rap", function() invisible(NULL))
  expect_error(extract_pheno(c(31, NA)), "NA")
})

test_that("extract_pheno() throws error on Inf field_id", {
  mockery::stub(extract_pheno, ".assert_on_rap", function() invisible(NULL))
  expect_error(extract_pheno(c(31, Inf)), "Inf")
})

test_that("extract_pheno() throws error on decimal field_id", {
  mockery::stub(extract_pheno, ".assert_on_rap", function() invisible(NULL))
  expect_error(extract_pheno(31.7), "whole numbers")
})

test_that("extract_pheno() throws error when no fields match", {
  .set_fake_cache()
  on.exit(.clear_cache())

  mockery::stub(extract_pheno, ".assert_on_rap", function() invisible(NULL))
  mockery::stub(extract_pheno, ".dx_find_dataset", function() "app12345.dataset")
  expect_error(
    suppressMessages(extract_pheno(999999)),
    "No matching fields found"
  )
})

test_that("extract_pheno() warns on unmatched field_id", {
  .set_fake_cache()
  on.exit(.clear_cache())

  mockery::stub(extract_pheno, ".assert_on_rap", function() invisible(NULL))
  mockery::stub(extract_pheno, ".dx_find_dataset", function() "app12345.dataset")
  mockery::stub(extract_pheno, ".dx_extract_run", function(dataset, fields, dest, ...) {
    write.csv(data.frame(`participant.eid` = 1L, `participant.p31` = 1L,
                          check.names = FALSE), dest, row.names = FALSE)
    .fake_dx()
  })

  expect_warning(
    suppressMessages(extract_pheno(c(31, 189))),
    "not found"
  )
})

test_that("extract_pheno() always includes eid as first field", {
  .set_fake_cache()
  on.exit(.clear_cache())

  received_fields <- NULL
  mockery::stub(extract_pheno, ".assert_on_rap", function() invisible(NULL))
  mockery::stub(extract_pheno, ".dx_find_dataset", function() "app12345.dataset")
  mockery::stub(extract_pheno, ".dx_extract_run",
                function(dataset, fields, dest, ...) {
                  received_fields <<- fields
                  write.csv(data.frame(`participant.eid` = 1L,
                                        check.names = FALSE), dest, row.names = FALSE)
                  .fake_dx()
                })

  suppressMessages(suppressWarnings(extract_pheno(31)))
  expect_equal(received_fields[1], "participant.eid")
})

test_that("extract_pheno() returns a data.table", {
  .set_fake_cache()
  on.exit(.clear_cache())

  mockery::stub(extract_pheno, ".assert_on_rap", function() invisible(NULL))
  mockery::stub(extract_pheno, ".dx_find_dataset", function() "app12345.dataset")
  mockery::stub(extract_pheno, ".dx_extract_run", function(dataset, fields, dest, ...) {
    write.csv(data.frame(`participant.eid` = c(1L, 2L), `participant.p31` = c(1L, 0L),
                          check.names = FALSE), dest, row.names = FALSE)
    .fake_dx()
  })

  result <- suppressMessages(extract_pheno(31))
  expect_true(data.table::is.data.table(result))
})

# ===========================================================================
# extract_batch()
# ===========================================================================

test_that("extract_batch() throws error on invalid field_id", {
  expect_error(extract_batch(character(0)), "non-empty numeric")
})

test_that("extract_batch() throws error on NA field_id", {
  expect_error(extract_batch(c(31, NA)), "NA")
})

test_that("extract_batch() throws error on Inf field_id", {
  expect_error(extract_batch(c(31, Inf)), "Inf")
})

test_that("extract_batch() throws error on decimal field_id", {
  expect_error(extract_batch(31.7), "whole numbers")
})

test_that("extract_batch() strips participant. prefix for table-exporter", {
  .set_fake_cache()
  on.exit(.clear_cache())

  received_fields <- NULL
  captured_path   <- NULL
  mockery::stub(extract_batch, ".dx_find_dataset", function() "app12345.dataset")
  # Reason: capture local_path here so .dx_run_table_exporter stub can read
  # the fields file before on.exit() deletes it
  mockery::stub(extract_batch, ".dx_upload_file", function(local_path) {
    captured_path <<- local_path
    "file-XXXX"
  })
  mockery::stub(extract_batch, ".dx_run_table_exporter",
                function(dataset, file_id, output, instance_type, priority) {
                  received_fields <<- readLines(captured_path)
                  "job-XXXX"
                })

  suppressMessages(extract_batch(31))
  expect_true(any(grepl("^eid$", received_fields)))
  expect_true(any(grepl("^p31$", received_fields)))
  expect_false(any(grepl("^participant\\.", received_fields)))
})

test_that("extract_batch() auto-selects x8 for small extractions", {
  .set_fake_cache()
  on.exit(.clear_cache())

  received_instance <- NULL
  mockery::stub(extract_batch, ".dx_find_dataset", function() "app12345.dataset")
  mockery::stub(extract_batch, ".dx_upload_file", function(...) "file-XXXX")
  mockery::stub(extract_batch, ".dx_run_table_exporter",
                function(dataset, file_id, output, instance_type, priority) {
                  received_instance <<- instance_type
                  "job-XXXX"
                })

  suppressMessages(extract_batch(c(31, 22189)))   # 2 cols → x4
  expect_equal(received_instance, "mem1_ssd1_v2_x4")
})

test_that("extract_batch() passes priority to table-exporter", {
  .set_fake_cache()
  on.exit(.clear_cache())

  received_priority <- NULL
  mockery::stub(extract_batch, ".dx_find_dataset", function() "app12345.dataset")
  mockery::stub(extract_batch, ".dx_upload_file", function(...) "file-XXXX")
  mockery::stub(extract_batch, ".dx_run_table_exporter",
                function(dataset, file_id, output, instance_type, priority) {
                  received_priority <<- priority
                  "job-XXXX"
                })

  suppressMessages(extract_batch(31, priority = "high"))
  expect_equal(received_priority, "high")
})

test_that("extract_batch() uses custom file name when provided", {
  .set_fake_cache()
  on.exit(.clear_cache())

  received_output <- NULL
  mockery::stub(extract_batch, ".dx_find_dataset", function() "app12345.dataset")
  mockery::stub(extract_batch, ".dx_upload_file", function(...) "file-XXXX")
  mockery::stub(extract_batch, ".dx_run_table_exporter",
                function(dataset, file_id, output, instance_type, priority) {
                  received_output <<- output
                  "job-XXXX"
                })

  suppressMessages(extract_batch(31, file = "my_cohort"))
  expect_equal(received_output, "my_cohort")
})

test_that("extract_batch() returns job_id invisibly", {
  .set_fake_cache()
  on.exit(.clear_cache())

  mockery::stub(extract_batch, ".dx_find_dataset", function() "app12345.dataset")
  mockery::stub(extract_batch, ".dx_upload_file", function(...) "file-XXXX")
  mockery::stub(extract_batch, ".dx_run_table_exporter",
                function(...) "job-XXXX")

  expect_invisible(suppressMessages(extract_batch(31)))
  result <- suppressMessages(extract_batch(31))
  expect_equal(result, "job-XXXX")
})

test_that("extract_batch() throws error when no fields match", {
  .set_fake_cache()
  on.exit(.clear_cache())

  mockery::stub(extract_batch, ".dx_find_dataset", function() "app12345.dataset")
  expect_error(
    suppressMessages(extract_batch(999999)),
    "No matching fields found"
  )
})

test_that("extract_batch() warns on unmatched field_ids", {
  .set_fake_cache()
  on.exit(.clear_cache())

  mockery::stub(extract_batch, ".dx_find_dataset", function() "app12345.dataset")
  mockery::stub(extract_batch, ".dx_upload_file", function(...) "file-XXXX")
  mockery::stub(extract_batch, ".dx_run_table_exporter", function(...) "job-XXXX")

  expect_warning(
    suppressMessages(extract_batch(c(31, 189))),
    "not found"
  )
})

test_that("extract_batch() auto-selects x8 for 21-100 cols", {
  # p53 with 25 instances → 25 cols, which falls in the x8 tier (21-100)
  big_df <- data.frame(
    field_name = paste0("participant.p53_i", 0:24),
    title      = paste0("Field 53 | Instance ", 0:24),
    stringsAsFactors = FALSE
  )
  .ukbflow_cache$fields <- list("app12345.dataset" = big_df)
  on.exit(.clear_cache())

  received_instance <- NULL
  mockery::stub(extract_batch, ".dx_find_dataset", function() "app12345.dataset")
  mockery::stub(extract_batch, ".dx_upload_file", function(...) "file-XXXX")
  mockery::stub(extract_batch, ".dx_run_table_exporter",
                function(dataset, file_id, output, instance_type, priority) {
                  received_instance <<- instance_type
                  "job-XXXX"
                })

  suppressMessages(extract_batch(53))
  expect_equal(received_instance, "mem1_ssd1_v2_x8")
})

test_that("extract_batch() auto-selects x16 for 101-500 cols", {
  # p53 with 110 instances → 110 cols, which falls in the x16 tier (101-500)
  huge_df <- data.frame(
    field_name = paste0("participant.p53_i", 0:109),
    title      = paste0("Field 53 | Instance ", 0:109),
    stringsAsFactors = FALSE
  )
  .ukbflow_cache$fields <- list("app12345.dataset" = huge_df)
  on.exit(.clear_cache())

  received_instance <- NULL
  mockery::stub(extract_batch, ".dx_find_dataset", function() "app12345.dataset")
  mockery::stub(extract_batch, ".dx_upload_file", function(...) "file-XXXX")
  mockery::stub(extract_batch, ".dx_run_table_exporter",
                function(dataset, file_id, output, instance_type, priority) {
                  received_instance <<- instance_type
                  "job-XXXX"
                })

  suppressMessages(extract_batch(53))
  expect_equal(received_instance, "mem1_ssd1_v2_x16")
})

test_that("extract_batch() auto-selects x36 for >500 cols", {
  # p53 with 510 instances → 510 cols, which falls in the x36 tier (>500)
  giant_df <- data.frame(
    field_name = paste0("participant.p53_i", 0:509),
    title      = paste0("Field 53 | Instance ", 0:509),
    stringsAsFactors = FALSE
  )
  .ukbflow_cache$fields <- list("app12345.dataset" = giant_df)
  on.exit(.clear_cache())

  received_instance <- NULL
  mockery::stub(extract_batch, ".dx_find_dataset", function() "app12345.dataset")
  mockery::stub(extract_batch, ".dx_upload_file", function(...) "file-XXXX")
  mockery::stub(extract_batch, ".dx_run_table_exporter",
                function(dataset, file_id, output, instance_type, priority) {
                  received_instance <<- instance_type
                  "job-XXXX"
                })

  suppressMessages(extract_batch(53))
  expect_equal(received_instance, "mem1_ssd1_v2_x36")
})

test_that("extract_batch() respects custom instance_type", {
  .set_fake_cache()
  on.exit(.clear_cache())

  received_instance <- NULL
  mockery::stub(extract_batch, ".dx_find_dataset", function() "app12345.dataset")
  mockery::stub(extract_batch, ".dx_upload_file", function(...) "file-XXXX")
  mockery::stub(extract_batch, ".dx_run_table_exporter",
                function(dataset, file_id, output, instance_type, priority) {
                  received_instance <<- instance_type
                  "job-XXXX"
                })

  suppressMessages(extract_batch(31, instance_type = "mem2_ssd1_v2_x4"))
  expect_equal(received_instance, "mem2_ssd1_v2_x4")
})

# ===========================================================================
# extract_pheno() — additional edge cases
# ===========================================================================

test_that("extract_pheno() stops when extraction command fails", {
  .set_fake_cache()
  on.exit(.clear_cache())

  mockery::stub(extract_pheno, ".assert_on_rap", function() invisible(NULL))
  mockery::stub(extract_pheno, ".dx_find_dataset", function() "app12345.dataset")
  mockery::stub(extract_pheno, ".dx_extract_run",
                function(...) .fake_dx(stderr = "connection timeout", status = 1))

  expect_error(
    suppressMessages(extract_pheno(31)),
    "Extraction failed"
  )
})

test_that("extract_pheno() deduplicates repeated field_ids", {
  .set_fake_cache()
  on.exit(.clear_cache())

  received_fields <- NULL
  mockery::stub(extract_pheno, ".assert_on_rap", function() invisible(NULL))
  mockery::stub(extract_pheno, ".dx_find_dataset", function() "app12345.dataset")
  mockery::stub(extract_pheno, ".dx_extract_run",
                function(dataset, fields, dest, ...) {
                  received_fields <<- fields
                  write.csv(data.frame(`participant.eid` = 1L,
                                        check.names = FALSE), dest, row.names = FALSE)
                  .fake_dx()
                })

  suppressMessages(extract_pheno(c(31, 31, 31)))
  # participant.eid + participant.p31 — p31 must appear exactly once
  expect_equal(sum(received_fields == "participant.p31"), 1L)
})

# =============================================================================
# grs.R - Genetic Risk Score utilities for ukbflow
# =============================================================================


#' Check and export a GRS weights file
#'
#' Reads a SNP weights file, validates its content, and writes a
#' plink2-compatible space-delimited output ready for upload to UKB RAP.
#'
#' The input file must contain at least the three columns below (additional
#' columns are ignored):
#' \describe{
#'   \item{\code{snp}}{SNP identifier, expected in \code{rs} + digits format.}
#'   \item{\code{effect_allele}}{Effect allele; must be one of A / T / C / G.}
#'   \item{\code{beta}}{Effect size (log-OR or beta coefficient); must be numeric.}
#' }
#'
#' Checks performed:
#' \itemize{
#'   \item Required columns present.
#'   \item No \code{NA} values in the three required columns.
#'   \item No duplicate \code{snp} identifiers.
#'   \item \code{snp} matches \code{rs[0-9]+} pattern (warning if not).
#'   \item \code{effect_allele} contains only A / T / C / G (warning if not).
#'   \item \code{beta} is numeric (error if not).
#' }
#'
#' @param file Character scalar. Path to the input weights file.
#'   Read via \code{data.table::fread} (format auto-detected; handles
#'   CSV, TSV, space-delimited, etc.).
#' @param dest Character scalar or \code{NULL}. Output path for the validated,
#'   space-delimited weights file. When \code{NULL} (default), no file is
#'   written and the validated \code{data.table} is returned invisibly only.
#'
#' @return A \code{data.table} with columns \code{snp}, \code{effect_allele},
#'   and \code{beta}, returned invisibly.
#'
#' @export
#'
#' @examples
#' tmp_in <- tempfile(fileext = ".csv")
#' weights <- data.frame(
#'   snp           = c("rs1234567", "rs2345678", "rs3456789"),
#'   effect_allele = c("A", "T", "G"),
#'   beta          = c(0.12, -0.05, 0.23)
#' )
#' write.csv(weights, tmp_in, row.names = FALSE)
#'
#' w <- grs_check(tmp_in)
#' w
#'
#' # Save validated weights to a file
#' tmp_out <- tempfile(fileext = ".txt")
#' grs_check(tmp_in, dest = tmp_out)
grs_check <- function(file, dest = NULL) {

  # Read
  .assert_file_exists(file)

  dt <- data.table::fread(file, data.table = TRUE)
  if (nrow(dt) == 0L)
    cli::cli_abort("File contains no data rows: {.path {file}}", call = NULL)
  cli::cli_inform("Read {.path {file}}: {nrow(dt)} rows, {ncol(dt)} columns.")
  # Normalise effect_allele to uppercase to tolerate lowercase input
  if ("effect_allele" %in% names(dt))
    dt[, effect_allele := toupper(effect_allele)]

  # Required columns
  required    <- c("snp", "effect_allele", "beta")
  missing_cols <- setdiff(required, names(dt))
  if (length(missing_cols) > 0L)
    cli::cli_abort(
      c("Required column(s) not found: {.val {missing_cols}}",
        "i" = "File must contain: {.val {required}}")
    )

  w <- dt[, ..required]

  # NA check
  n_na <- sum(!stats::complete.cases(w))
  if (n_na > 0L)
    cli::cli_abort("{n_na} row(s) have NA in required columns - remove before proceeding.", call = NULL)
  cli::cli_alert_success("No NA values.")

  # Duplicate SNPs
  n_dup <- sum(duplicated(w$snp))
  if (n_dup > 0L)
    cli::cli_abort("{n_dup} duplicate SNP ID(s) found - each SNP must appear once.", call = NULL)
  cli::cli_alert_success("No duplicate SNPs.")

  # SNP format: rs + digits
  n_bad_rs <- sum(!grepl("^rs[0-9]+$", w$snp))
  if (n_bad_rs > 0L)
    cli::cli_warn(
      c("{n_bad_rs} SNP ID(s) do not match {.code rs[0-9]+} format.",
        "i" = "plink2 accepts non-rsID formats but verify alignment with target genotype data.")
    )
  else
    cli::cli_alert_success("All SNP IDs match rs[0-9]+ format.")

  # Effect allele: A / T / C / G only
  n_bad_al <- sum(!w$effect_allele %in% c("A", "T", "C", "G"))
  if (n_bad_al > 0L)
    cli::cli_warn(
      c("{n_bad_al} effect allele(s) outside A/T/C/G.",
        "i" = "Indels or multi-character alleles may not score correctly in plink2.")
    )
  else
    cli::cli_alert_success("All effect alleles are A/T/C/G.")

  # Beta: numeric
  if (!is.numeric(w$beta))
    cli::cli_abort("{.field beta} must be numeric (found {.cls {class(w$beta)}}).", call = NULL)

  # Beta summary
  n_pos  <- sum(w$beta >  0)
  n_neg  <- sum(w$beta <  0)
  n_zero <- sum(w$beta == 0)

  cli::cli_inform(c(
    "Beta summary:",
    " " = "Range     : {round(min(w$beta), 4)} to {round(max(w$beta), 4)}",
    " " = "Mean |beta|: {round(mean(abs(w$beta)), 4)}",
    " " = "Positive  : {n_pos} ({round(100 * n_pos / nrow(w), 1)}%)",
    " " = "Negative  : {n_neg} ({round(100 * n_neg / nrow(w), 1)}%)",
    " " = "Zero      : {n_zero}"
  ))

  cli::cli_alert_success(
    "Weights file passed checks: {.strong {nrow(w)} SNPs} ready for UKB RAP."
  )

  # Write plink2-compatible output (space-delimited, no quotes)
  if (!is.null(dest)) {
    data.table::fwrite(w, dest, sep = " ", quote = FALSE)
    cli::cli_alert_success("Saved: {.path {dest}}")
  }

  invisible(w)
}


#' Convert UKB imputed BGEN files to PGEN on RAP
#'
#' Submits one Swiss Army Knife job per chromosome to the DNAnexus Research
#' Analysis Platform, each converting a UKB imputed BGEN file to PGEN format
#' with a MAF > 0.01 filter applied via plink2. Jobs run in parallel across
#' chromosomes.
#'
#' The function auto-generates the plink2 driver script, uploads it once to
#' the RAP project root (\code{/}) on RAP, then loops over \code{chr} submitting
#' one job per chromosome. A 500 ms pause between submissions prevents API
#' rate-limiting.
#'
#' \strong{Output path is critical.} The driver script writes plink2 output to
#' \code{/home/dnanexus/out/out/} - the fixed path that Swiss Army Knife
#' auto-uploads to \code{dest} on completion. Output files per chromosome:
#' \code{chr\{N\}_maf001.pgen/.pvar/.psam/.log}.
#'
#' \strong{Instance types:}
#' \describe{
#'   \item{\code{"standard"}}{\code{mem2_ssd1_v2_x4}: 4 cores, 12 GB RAM.
#'     Suitable for smaller chromosomes (roughly chr 17–22).}
#'   \item{\code{"large"}}{\code{mem2_ssd2_v2_x8}: 8 cores, 28 GB RAM,
#'     640 GB SSD. Required for large chromosomes (roughly chr 1–16) where
#'     standard storage is insufficient.}
#' }
#'
#' @param chr Integer vector. Chromosomes to process. Default: \code{1:22}.
#' @param dest Character scalar. RAP destination path for output PGEN files
#'   (e.g. \code{"/pgen/"}). Must be specified explicitly.
#' @param maf Numeric scalar. Minor allele frequency filter passed to plink2
#'   \code{--maf}. Variants with MAF below this threshold are excluded.
#'   Default: \code{0.01}. Must be in \code{(0, 0.5)}.
#' @param instance Character scalar. Instance type preset: \code{"standard"}
#'   or \code{"large"}. See Details. Default: \code{"standard"}.
#' @param priority Character scalar. Job priority: \code{"low"} or
#'   \code{"high"}. Default: \code{"low"}.
#'
#' @return A character vector of job IDs (one per chromosome), returned
#'   invisibly. Failed submissions are \code{NA}. Use \code{\link{job_ls}},
#'   \code{\link{job_status}}, or \code{\link{job_wait}} to monitor progress.
#'
#' @export
#'
#' @examples
#' \dontrun{
#' # Test with chr22 first (smallest chromosome)
#' ids <- grs_bgen2pgen(chr = 22, dest = "/pgen/", priority = "high")
#'
#' # Small chromosomes - standard instance
#' ids_small <- grs_bgen2pgen(chr = 15:22, dest = "/pgen/")
#'
#' # Large chromosomes - upgrade instance to handle storage
#' ids_large <- grs_bgen2pgen(chr = 1:16, dest = "/pgen/", instance = "large")
#'
#' # Monitor
#' job_ls()
#' }
grs_bgen2pgen <- function(chr      = 1:22,
                           dest     = NULL,
                           maf      = 0.01,
                           instance = "standard",
                           priority = "low") {

  # Validate
  instance <- match.arg(instance, c("standard", "large"))
  priority <- match.arg(priority, c("low", "high"))

  chr <- as.integer(chr)
  if (any(is.na(chr)) || any(chr < 1L) || any(chr > 22L))
    cli::cli_abort("{.arg chr} must be integers between 1 and 22.", call = NULL)

  if (!is.numeric(maf) || length(maf) != 1L || maf <= 0 || maf >= 0.5)
    cli::cli_abort("{.arg maf} must be a single numeric value in (0, 0.5).", call = NULL)

  if (instance == "standard" && any(chr %in% 1:16))
    cli::cli_warn(c(
      "chr {.val {intersect(chr, 1:16)}} may exceed storage on {.val {'mem2_ssd1_v2_x4'}}.",
      "i" = "Consider {.code instance = \"large\"} (mem2_ssd2_v2_x8, 640 GB SSD) for these chromosomes."
    ))

  if (is.null(dest) || !nzchar(dest))
    cli::cli_abort("{.arg dest} must be specified (e.g. {.val {'/pgen/'}}).", call = NULL)

  # Instance config (mirrors scripts 06 and 13)
  if (instance == "standard") {
    instance_type <- "mem2_ssd1_v2_x4"
    n_threads     <- 4L
    plink_memory  <- 12000L
  } else {
    instance_type <- "mem2_ssd2_v2_x8"
    n_threads     <- 8L
    plink_memory  <- 28000L
  }

  # Verify project context
  project_id <- .dx_get_project_id()
  if (is.na(project_id) || !nzchar(project_id))
    cli::cli_abort("No project selected. Run {.fn auth_select_project} first.", call = NULL)

  # Generate driver script, write to tempfile, upload to RAP
  script_name   <- if (instance == "standard") "grs_bgen2pgen_std.R"
                   else                         "grs_bgen2pgen_lrg.R"
  script_remote <- paste0("/", script_name)

  tmp_script <- tempfile(fileext = ".R")
  on.exit(unlink(tmp_script), add = TRUE)
  writeLines(.bgen2pgen_script(n_threads, plink_memory, maf), tmp_script)

  existing <- tryCatch(fetch_ls(".", type = "file"), error = function(e) NULL)

  if (!is.null(existing) && script_name %in% existing$name) {
    cli::cli_alert_success("{.val {script_name}} already on RAP, skipping upload.")
  } else {
    cli::cli_inform("Uploading {.val {script_name}} to RAP ...")
    up <- .dx_run(
      c("upload", tmp_script, "--path", script_remote, "--brief", "--wait"),
      timeout = 120L
    )
    if (!up$success)
      cli::cli_abort("Script upload failed: {up$stderr}", call = NULL)
    cli::cli_alert_success("Uploaded: {.val {script_remote}}")
  }

  # Submit one Swiss Army Knife job per chromosome
  cli::cli_inform(
    "Submitting {length(chr)} job(s) -- {instance_type} / priority: {priority}"
  )

  job_ids <- character(length(chr))

  for (i in seq_along(chr)) {
    ch <- chr[i]

    # Mirrors dx download + Rscript pattern from scripts 07 / 09 / 12
    cmd <- paste0(
      "dx download ", project_id, ":", script_remote, " --overwrite",
      " && Rscript ", script_name, " ", ch
    )

    res <- .dx_run(c(
      "run", "swiss-army-knife",
      paste0("-icmd=", cmd),
      "--destination",   dest,
      "--instance-type", instance_type,
      "--priority",      priority,
      "--name",          paste0("pgen_chr", ch),
      "--brief",
      "--yes"
    ), timeout = 60L)

    if (res$success) {
      job_ids[i] <- trimws(res$stdout)
      cli::cli_alert_success("chr{ch} -> {.val {job_ids[i]}}")
    } else {
      job_ids[i] <- NA_character_
      cli::cli_alert_danger("chr{ch} submission failed: {res$stderr}")
    }

    # Avoid DNAnexus API rate limits
    if (i < length(chr)) Sys.sleep(0.5)
  }

  n_ok <- sum(!is.na(job_ids))
  cli::cli_alert_success(
    "{n_ok}/{length(chr)} job(s) submitted. Monitor with {.fn job_ls}."
  )

  invisible(job_ids)
}


#' Calculate genetic risk scores from PGEN files on RAP
#'
#' Uploads local SNP weight files to the RAP project root, then submits one
#' Swiss Army Knife job per GRS. Each job runs plink2 \code{--score} across
#' all 22 chromosomes and saves a single CSV to \code{dest} on completion.
#' Jobs run in parallel; use \code{\link{job_ls}} to monitor progress.
#'
#' Weight files should have three columns (any delimiter, header required):
#' \describe{
#'   \item{Column 1}{Variant ID (e.g. \code{rs} IDs).}
#'   \item{Column 2}{Effect allele (A1).}
#'   \item{Column 3}{Effect weight (beta / log-OR).}
#' }
#' This matches the output format of \code{\link{grs_check}}.
#'
#' \strong{Output per job:} \code{dest/<score_name>_scores.csv} with columns
#' \code{IID} and the GRS score (named \code{GRS_<name>}).
#'
#' \strong{Instance types:}
#' \describe{
#'   \item{\code{"standard"}}{\code{mem2_ssd1_v2_x4}: 4 cores, 12 GB RAM.}
#'   \item{\code{"large"}}{\code{mem2_ssd2_v2_x8}: 8 cores, 28 GB RAM.}
#' }
#'
#' @param file Named character vector of local weight file paths. Names become
#'   the GRS identifiers (output column = \code{GRS_<name>}).
#'   Example: \code{c(grs_a = "weights_a.txt")}.
#' @param pgen_dir Character scalar. Path to PGEN files on RAP
#'   (e.g. \code{"/mnt/project/pgen"}). Must be specified explicitly.
#' @param dest Character scalar. RAP destination path for output CSV files
#'   (e.g. \code{"/grs/"}). Must be specified explicitly.
#' @param maf Numeric scalar. MAF filter threshold used when locating PGEN
#'   files. Must match the value used in \code{\link{grs_bgen2pgen}}.
#'   Default: \code{0.01}.
#' @param instance Character scalar. Instance type preset: \code{"standard"}
#'   or \code{"large"}. Default: \code{"standard"}.
#' @param priority Character scalar. Job priority: \code{"low"} or
#'   \code{"high"}. Default: \code{"low"}.
#'
#' @return A named character vector of job IDs (one per GRS), returned
#'   invisibly. Failed submissions are \code{NA}. Use \code{\link{job_ls}} to
#'   monitor progress.
#'
#' @export
#'
#' @examples
#' \dontrun{
#' ids <- grs_score(
#'   file = c(
#'     grs_a = "weights/grs_a_weights.txt",
#'     grs_b = "weights/grs_b_weights.txt"
#'   ),
#'   pgen_dir = "/mnt/project/pgen",
#'   dest     = "/grs/",
#'   priority = "high"
#' )
#'
#' job_ls()
#' }
grs_score <- function(file,
                      pgen_dir = NULL,
                      dest     = NULL,
                      maf      = 0.01,
                      instance = "standard",
                      priority = "low") {

  # Validate arguments
  instance <- match.arg(instance, c("standard", "large"))
  priority <- match.arg(priority, c("low", "high"))

  if (!is.character(file))
    cli::cli_abort("{.arg file} must be a named character vector.", call = NULL)
  if (is.null(names(file)) || any(!nzchar(names(file))))
    cli::cli_abort("{.arg file} must be fully named (each entry needs a name).", call = NULL)
  if (any(duplicated(names(file))))
    cli::cli_abort(
      "Duplicate names in {.arg file}: {.val {names(file)[duplicated(names(file))]}}",
      call = NULL
    )

  missing_local <- file[!file.exists(file)]
  if (length(missing_local) > 0L)
    cli::cli_abort(c(
      "Local weight file(s) not found:",
      setNames(missing_local, rep("x", length(missing_local)))
    ), call = NULL)

  if (!is.numeric(maf) || length(maf) != 1L || maf <= 0 || maf >= 0.5)
    cli::cli_abort("{.arg maf} must be a single numeric value in (0, 0.5). Must match the value used in {.fn grs_bgen2pgen}.", call = NULL)

  if (is.null(pgen_dir) || !nzchar(pgen_dir))
    cli::cli_abort("{.arg pgen_dir} must be specified (e.g. {.val {'/mnt/project/pgen'}}).", call = NULL)

  if (is.null(dest) || !nzchar(dest))
    cli::cli_abort("{.arg dest} must be specified (e.g. {.val {'/grs/'}}).", call = NULL)

  # Instance config
  if (instance == "standard") {
    instance_type <- "mem2_ssd1_v2_x4"
    n_threads     <- 4L
    plink_memory  <- 12000L
  } else {
    instance_type <- "mem2_ssd2_v2_x8"
    n_threads     <- 8L
    plink_memory  <- 28000L
  }

  # Verify project context
  project_id <- .dx_get_project_id()
  if (is.na(project_id) || !nzchar(project_id))
    cli::cli_abort("No project selected. Run {.fn auth_select_project} first.", call = NULL)

  # Upload weight files to RAP root
  # Reason: SAK jobs access /mnt/project/<name>; uploading to root means the
  # driver script can read them at /mnt/project/<basename>.
  cli::cli_rule(left = sprintf("Uploading %d weight file(s) to RAP", length(file)))

  existing <- tryCatch(fetch_ls(".", type = "file"), error = function(e) NULL)
  existing_names <- if (!is.null(existing)) existing$name else character(0L)

  for (i in seq_along(file)) {
    fname       <- basename(file[[i]])
    remote_path <- paste0("/", fname)

    # If already at RAP root (e.g. /mnt/project/<fname>), skip delete + upload.
    # Reason: deleting before re-uploading would destroy the source file itself.
    if (.is_on_rap() && identical(file[[i]], paste0("/mnt/project/", fname))) {
      cli::cli_alert_info("{.val {fname}} already at RAP root, skipping upload.")
      next
    }

    if (fname %in% existing_names) {
      cli::cli_inform("Removing existing {.val {fname}} from RAP ...")
      rm_res <- .dx_run(c("rm", remote_path), timeout = 30L)
      if (!rm_res$success)
        cli::cli_abort("Failed to remove {.val {fname}}: {rm_res$stderr}", call = NULL)
    }

    cli::cli_inform("Uploading {.val {fname}} ...")
    up <- .dx_run(
      c("upload", file[[i]], "--path", remote_path, "--brief", "--wait"),
      timeout = 120L
    )
    if (!up$success)
      cli::cli_abort("Upload failed for {.val {fname}}: {up$stderr}", call = NULL)

    cli::cli_alert_success("Uploaded: {.val {remote_path}}")
  }

  # Generate driver script, upload once to RAP root
  script_name   <- if (instance == "standard") "grs_score_std.R"
                   else                         "grs_score_lrg.R"
  script_remote <- paste0("/", script_name)

  tmp_script <- tempfile(fileext = ".R")
  on.exit(unlink(tmp_script), add = TRUE)
  writeLines(.grs_score_script(n_threads, plink_memory, maf), tmp_script)

  # Re-fetch to account for any files added during weight uploads
  existing2 <- tryCatch(fetch_ls(".", type = "file"), error = function(e) NULL)

  if (!is.null(existing2) && script_name %in% existing2$name) {
    cli::cli_alert_success("{.val {script_name}} already on RAP, skipping upload.")
  } else {
    cli::cli_inform("Uploading {.val {script_name}} to RAP ...")
    up <- .dx_run(
      c("upload", tmp_script, "--path", script_remote, "--brief", "--wait"),
      timeout = 120L
    )
    if (!up$success)
      cli::cli_abort("Script upload failed: {up$stderr}", call = NULL)
    cli::cli_alert_success("Uploaded: {.val {script_remote}}")
  }

  # Submit one Swiss Army Knife job per GRS
  cli::cli_inform(
    "Submitting {length(file)} job(s) -- {instance_type} / priority: {priority}"
  )

  job_ids <- character(length(file))
  names(job_ids) <- names(file)

  for (i in seq_along(file)) {
    nm          <- names(file)[i]
    score_name  <- paste0("GRS_", nm)
    weight_fname <- basename(file[[i]])

    cmd <- paste0(
      "dx download ", project_id, ":", script_remote, " --overwrite",
      " && Rscript ", script_name,
      " ", weight_fname,
      " ", score_name,
      " ", pgen_dir
    )

    res <- .dx_run(c(
      "run", "swiss-army-knife",
      paste0("-icmd=", cmd),
      "--destination",   dest,
      "--instance-type", instance_type,
      "--priority",      priority,
      "--name",          paste0("grs_score_", nm),
      "--brief",
      "--yes"
    ), timeout = 60L)

    if (res$success) {
      job_ids[i] <- trimws(res$stdout)
      cli::cli_alert_success("{nm} -> {.val {job_ids[i]}}")
    } else {
      job_ids[i] <- NA_character_
      cli::cli_alert_danger("{nm} submission failed: {res$stderr}")
    }

    if (i < length(file)) Sys.sleep(0.5)
  }

  n_ok <- sum(!is.na(job_ids))
  cli::cli_alert_success(
    "{n_ok}/{length(file)} job(s) submitted. Monitor with {.fn job_ls}."
  )

  invisible(job_ids)
}


#' Standardise GRS columns by Z-score transformation
#'
#' Adds a \code{_z} column for every selected GRS column:
#' \code{z = (x - mean(x)) / sd(x)}.  The original columns are kept
#' unchanged.  \code{\link{grs_zscore}} is an alias for this function.
#'
#' @param data A \code{data.frame} or \code{data.table} containing at least
#'   one GRS column.
#' @param grs_cols Character vector of column names to standardise.
#'   If \code{NULL} (default), all columns whose names contain \code{"grs"}
#'   (case-insensitive) are selected automatically.
#'
#' @return The input \code{data} as a \code{data.table} with one additional
#'   \code{_z} column per GRS column appended after its source column.
#'
#' @export
#'
#' @examples
#' dt <- data.frame(
#'   IID   = 1:5,
#'   GRS_a = c(0.12, 0.34, 0.56, 0.23, 0.45),
#'   GRS_b = c(1.1,  0.9,  1.3,  0.8,  1.0)
#' )
#' grs_standardize(dt)
#' grs_zscore(dt)   # identical
grs_standardize <- function(data, grs_cols = NULL) {

  # Validate data
  .assert_data_frame(data)

  dt <- data.table::as.data.table(data)

  # Resolve target columns
  if (is.null(grs_cols)) {
    grs_cols <- names(dt)[grepl("grs", names(dt), ignore.case = TRUE)]
    if (length(grs_cols) == 0L)
      cli::cli_abort("No columns containing {.val grs} found. Supply column names via {.arg grs_cols}.", call = NULL)
    cli::cli_inform("Auto-detected {length(grs_cols)} GRS column(s): {.val {grs_cols}}")
  } else {
    missing_cols <- setdiff(grs_cols, names(dt))
    if (length(missing_cols) > 0L)
      cli::cli_abort("Column(s) not found in data: {.val {missing_cols}}", call = NULL)
  }

  # Z-score each column: add _z column immediately after source column
  for (col in grs_cols) {
    x       <- dt[[col]]
    mu      <- mean(x, na.rm = TRUE)
    sigma   <- stats::sd(x, na.rm = TRUE)

    if (sigma == 0)
      cli::cli_abort("{.field {col}} has zero variance - cannot standardise.", call = NULL)

    z_col   <- paste0(col, "_z")
    z_vals  <- (x - mu) / sigma

    # Insert _z column right after source column
    idx     <- which(names(dt) == col)
    dt      <- cbind(dt[, 1:idx, with = FALSE],
                     data.table::data.table(tmp__ = z_vals),
                     if (idx < ncol(dt)) dt[, (idx + 1L):ncol(dt), with = FALSE]
                     else NULL)
    data.table::setnames(dt, "tmp__", z_col)

    cli::cli_alert_success(
      "{col} -> {z_col}  [mean={round(mu, 4)}, sd={round(sigma, 4)}]"
    )
  }

  dt
}


#' @rdname grs_standardize
#' @export
grs_zscore <- grs_standardize


#' Validate GRS predictive performance
#'
#' For each GRS column, computes four sets of validation metrics:
#' \enumerate{
#'   \item \strong{Per SD} - OR (logistic) or HR (Cox) per 1-SD increase.
#'   \item \strong{High vs Low} - OR / HR comparing top 20\% vs bottom 20\%
#'     (extreme tertile grouping: Low / Mid / High).
#'   \item \strong{Trend test} - P-trend across quartiles (Q1–Q4).
#'   \item \strong{Discrimination} - AUC (logistic) or C-index (Cox).
#' }
#'
#' GRS grouping columns are created internally via \code{\link{derive_cut}}
#' and are not added to the user's data.  When \code{time_col} is
#' \code{NULL}, logistic regression is used throughout; when supplied, Cox
#' proportional hazards models are used.
#'
#' Models follow the same adjustment logic as \code{\link{assoc_logistic}} /
#' \code{\link{assoc_coxph}}: unadjusted and age-sex adjusted models are
#' always included; a fully adjusted model is added when \code{covariates}
#' is non-\code{NULL}.
#'
#' @param data A \code{data.frame} or \code{data.table}.
#' @param grs_cols Character vector of GRS column names to validate.
#'   If \code{NULL} (default), all columns whose names contain \code{"grs"}
#'   (case-insensitive) are selected automatically. All specified columns
#'   must be numeric.
#' @param outcome_col Character scalar. Name of the outcome column
#'   (\code{0}/\code{1} or \code{TRUE}/\code{FALSE}).
#' @param time_col Character scalar or \code{NULL}. Name of the follow-up
#'   time column. When \code{NULL} (default), logistic regression is used;
#'   when supplied, Cox regression is used.
#' @param covariates Character vector or \code{NULL}. Covariates for the
#'   fully adjusted model. When \code{NULL}, only unadjusted and age-sex
#'   adjusted models are run.
#'
#' @return A named \code{list} with four \code{data.table} elements:
#' \itemize{
#'   \item \code{per_sd}: OR / HR per 1-SD increase in GRS.
#'   \item \code{high_vs_low}: OR / HR for High vs Low extreme tertile.
#'   \item \code{trend}: P-trend across Q1–Q4 quartiles.
#'   \item \code{discrimination}: AUC (logistic) or C-index (Cox) with 95\% CI.
#' }
#'
#' @importFrom survival coxph Surv concordance
#' @export
#'
#' @examples
#' \donttest{
#' dt <- ops_toy(scenario = "association", n = 500)
#' dt <- grs_standardize(dt, grs_cols = "grs_bmi")
#'
#' res <- grs_validate(
#'   data        = dt,
#'   grs_cols    = "grs_bmi_z",
#'   outcome_col = "dm_status",
#'   time_col    = "dm_followup_years"
#' )
#' res$per_sd
#'
#' if (requireNamespace("pROC", quietly = TRUE)) {
#'   res_logit <- grs_validate(
#'     data        = dt,
#'     grs_cols    = "grs_bmi_z",
#'     outcome_col = "dm_status"
#'   )
#'   res_logit$discrimination
#' }
#' }
grs_validate <- function(data,
                         grs_cols    = NULL,
                         outcome_col,
                         time_col    = NULL,
                         covariates  = NULL) {

  # Validate inputs
  .assert_data_frame(data)

  # Work on a copy so derive_cut() in-place ops don't touch the user's data
  dt <- data.table::copy(data.table::as.data.table(data))

  if (is.null(grs_cols)) {
    grs_cols <- names(dt)[grepl("grs", names(dt), ignore.case = TRUE)]
    if (length(grs_cols) == 0L)
      cli::cli_abort("No GRS columns found. Supply column names via {.arg grs_cols}.", call = NULL)
    cli::cli_inform("Auto-detected {length(grs_cols)} GRS column(s): {.val {grs_cols}}")
  }

  missing_cols <- setdiff(c(grs_cols, outcome_col, time_col, covariates), names(dt))
  if (length(missing_cols) > 0L)
    cli::cli_abort("Column(s) not found in data: {.val {missing_cols}}", call = NULL)

  non_numeric <- grs_cols[!vapply(grs_cols, function(col) is.numeric(dt[[col]]), logical(1L))]
  if (length(non_numeric) > 0L)
    cli::cli_abort(
      "GRS column(s) must be numeric: {.val {non_numeric}}",
      call = NULL
    )

  is_cox <- !is.null(time_col)

  # Create Q4 (quartile) and E3 (20/60/20) grouping columns via derive_cut
  cli::cli_rule(left = "Creating GRS groups")

  quad_cols <- setNames(paste0(grs_cols, "_quad"), grs_cols)
  tri_cols  <- setNames(paste0(grs_cols, "_tri"),  grs_cols)

  for (col in grs_cols) {
    dt <- derive_cut(dt, col = col, n = 4L)

    e3_breaks <- stats::quantile(dt[[col]], probs = c(0.2, 0.8), na.rm = TRUE)
    dt <- derive_cut(dt, col = col, n = 3L,
                     breaks = e3_breaks,
                     labels = c("Low", "Mid", "High"))

    # Ensure "Low" is reference level for High vs Low contrast
    dt[, (tri_cols[col]) := relevel(factor(get(tri_cols[col])), ref = "Low")]
  }

  # Effect per SD
  cli::cli_rule(left = sprintf("Effect per SD (%s)", if (is_cox) "HR" else "OR"))

  per_sd <- if (is_cox) {
    assoc_coxph(dt,
                outcome_col  = outcome_col, time_col    = time_col,
                exposure_col = grs_cols,    covariates  = covariates)
  } else {
    assoc_logistic(dt,
                   outcome_col  = outcome_col,
                   exposure_col = grs_cols,   covariates = covariates)
  }

  # High vs Low (E3 grouping; keep only the High row from results)
  cli::cli_rule(left = "High vs Low")

  high_vs_low <- if (is_cox) {
    assoc_coxph(dt,
                outcome_col  = outcome_col,        time_col   = time_col,
                exposure_col = unname(tri_cols),   covariates = covariates)
  } else {
    assoc_logistic(dt,
                   outcome_col  = outcome_col,
                   exposure_col = unname(tri_cols), covariates = covariates)
  }
  high_vs_low <- high_vs_low[grepl("High$", term)]

  # Trend test across Q4 quartiles
  cli::cli_rule(left = "Trend test")

  trend <- assoc_trend(dt,
                       outcome_col  = outcome_col,
                       time_col     = time_col,
                       exposure_col = unname(quad_cols),
                       method       = if (is_cox) "coxph" else "logistic",
                       covariates   = covariates)

  # Discrimination: AUC (logistic) or C-index (Cox)
  cli::cli_rule(left = if (is_cox) "C-index" else "AUC")

  if (!is_cox && !requireNamespace("pROC", quietly = TRUE))
    cli::cli_abort(
      "Package {.pkg pROC} is required for AUC. Install with {.code install.packages('pROC')}.",
      call = NULL
    )

  disc_list <- lapply(grs_cols, function(col) {
    if (is_cox) {
      fit   <- survival::coxph(
        survival::Surv(dt[[time_col]], dt[[outcome_col]]) ~ dt[[col]]
      )
      conc  <- survival::concordance(fit)
      c_idx <- conc$concordance
      c_se  <- sqrt(conc$var)
      data.table::data.table(
        GRS      = col,
        C_index  = c_idx,
        CI_lower = c_idx - 1.96 * c_se,
        CI_upper = c_idx + 1.96 * c_se
      )
    } else {
      roc_obj <- pROC::roc(dt[[outcome_col]], dt[[col]], quiet = TRUE)
      auc_val <- as.numeric(pROC::auc(roc_obj))
      ci_obj  <- pROC::ci.auc(roc_obj)
      data.table::data.table(
        GRS      = col,
        AUC      = auc_val,
        CI_lower = ci_obj[1L],
        CI_upper = ci_obj[3L]
      )
    }
  })
  discrimination <- data.table::rbindlist(disc_list)

  cli::cli_alert_success("Validation complete.")

  list(
    per_sd         = per_sd,
    high_vs_low    = high_vs_low,
    trend          = trend,
    discrimination = discrimination
  )
}

# =============================================================================
# test-grs.R — Unit tests for grs_ series (no network, no RAP)
# =============================================================================


# ===========================================================================
# Shared helpers
# ===========================================================================

# Minimal valid weights data.table
.fake_weights_dt <- function() {
  data.table::data.table(
    snp           = paste0("rs", 1:5),
    effect_allele = c("A", "T", "C", "G", "A"),
    beta          = c(0.10, -0.20, 0.30, 0.05, -0.15)
  )
}

# Write a weights CSV to a temp path; returns the path
.write_weights_file <- function(path, dt = NULL) {
  if (is.null(dt)) dt <- .fake_weights_dt()
  data.table::fwrite(dt, path)
  path
}

# Simulated GRS cohort for standardize / validate
.fake_grs_dt <- function(n = 300, seed = 42) {
  set.seed(seed)
  data.table::data.table(
    IID              = seq_len(n),
    GRS_a_z    = rnorm(n),
    GRS_b_z = rnorm(n),
    outcome          = rbinom(n, 1, 0.20),
    followup_years   = round(runif(n, 1, 15), 2)
  )
}


# ===========================================================================
# grs_check()
# ===========================================================================

test_that("grs_check() aborts when file not found", {
  expect_error(grs_check("/no/such/file.txt"), "not found")
})

test_that("grs_check() aborts when required columns are missing", {
  withr::local_tempfile(fileext = ".csv") |> (function(path) {
    data.table::fwrite(data.table::data.table(x = 1:3, y = 1:3), path)
    expect_error(grs_check(path), "Required column")
  })()
})

test_that("grs_check() aborts when NA values are present", {
  withr::local_tempfile(fileext = ".csv") |> (function(path) {
    dt <- .fake_weights_dt()
    dt[2L, beta := NA_real_]
    .write_weights_file(path, dt)
    expect_error(grs_check(path), "NA")
  })()
})

test_that("grs_check() aborts on duplicate SNP IDs", {
  withr::local_tempfile(fileext = ".csv") |> (function(path) {
    dt <- .fake_weights_dt()
    dt[2L, snp := dt$snp[1L]]
    .write_weights_file(path, dt)
    expect_error(grs_check(path), "duplicate")
  })()
})

test_that("grs_check() aborts when beta is non-numeric", {
  withr::local_tempfile(fileext = ".csv") |> (function(path) {
    dt <- data.table::data.table(
      snp           = paste0("rs", 1:3),
      effect_allele = c("A", "T", "C"),
      beta          = c("high", "low", "medium")
    )
    .write_weights_file(path, dt)
    expect_error(grs_check(path), "[Bb]eta|numeric")
  })()
})

test_that("grs_check() warns on non-rs SNP IDs but does not abort", {
  withr::local_tempfile(fileext = ".csv") |> (function(path) {
    dt <- .fake_weights_dt()
    dt[1L, snp := "chr1:12345:A:G"]
    dest <- tempfile(fileext = ".txt")
    on.exit(unlink(dest), add = TRUE)
    .write_weights_file(path, dt)
    expect_warning(grs_check(path, dest = dest), "rs\\[0-9\\]\\+|rs.*format")
  })()
})

test_that("grs_check() warns on non-ATCG alleles but does not abort", {
  withr::local_tempfile(fileext = ".csv") |> (function(path) {
    dt <- .fake_weights_dt()
    dt[1L, effect_allele := "INDEL"]
    dest <- tempfile(fileext = ".txt")
    on.exit(unlink(dest), add = TRUE)
    .write_weights_file(path, dt)
    expect_warning(grs_check(path, dest = dest), "A/T/C/G|allele")
  })()
})

test_that("grs_check() returns a 3-column data.table and writes output file", {
  withr::local_tempfile(fileext = ".csv") |> (function(path) {
    .write_weights_file(path)
    dest <- tempfile(fileext = ".txt")
    on.exit(unlink(dest), add = TRUE)

    result <- grs_check(path, dest = dest)

    expect_s3_class(result, "data.table")
    expect_equal(ncol(result), 3L)
    expect_named(result, c("snp", "effect_allele", "beta"))
    expect_equal(nrow(result), 5L)
    expect_true(file.exists(dest))
  })()
})


# ===========================================================================
# grs_standardize() / grs_zscore()
# ===========================================================================

test_that("grs_standardize() aborts on non-data.frame input", {
  expect_error(grs_standardize("not a df"), "data.frame")
})

test_that("grs_standardize() aborts when manually specified column not found", {
  dt <- .fake_grs_dt()
  expect_error(grs_standardize(dt, grs_cols = "no_such_col"), "not found")
})

test_that("grs_standardize() aborts when no GRS columns auto-detected", {
  dt <- data.table::data.table(IID = 1:5, age = 50:54)
  expect_error(grs_standardize(dt), "No columns")
})

test_that("grs_standardize() aborts on zero-variance column", {
  dt <- data.table::data.table(IID = 1:5, GRS_flat = rep(1.0, 5))
  expect_error(grs_standardize(dt), "zero variance")
})

test_that("grs_standardize() produces mean ≈ 0 and SD ≈ 1 for _z columns", {
  dt  <- .fake_grs_dt()
  out <- grs_standardize(dt, grs_cols = c("GRS_a_z", "GRS_b_z"))

  for (col in c("GRS_a_z_z", "GRS_b_z_z")) {
    expect_true(col %in% names(out))
    expect_lt(abs(mean(out[[col]], na.rm = TRUE)), 1e-10)
    expect_lt(abs(sd(out[[col]], na.rm = TRUE) - 1), 1e-10)
  }
})

test_that("grs_standardize() inserts _z column immediately after source column", {
  dt  <- .fake_grs_dt()
  out <- grs_standardize(dt, grs_cols = "GRS_a_z")
  idx_src <- which(names(out) == "GRS_a_z")
  idx_z   <- which(names(out) == "GRS_a_z_z")
  expect_equal(idx_z, idx_src + 1L)
})

test_that("grs_standardize() preserves original columns unchanged", {
  dt  <- .fake_grs_dt()
  out <- grs_standardize(dt, grs_cols = "GRS_a_z")
  expect_equal(out$GRS_a_z, dt$GRS_a_z)
})

test_that("grs_standardize() works on plain data.frame input", {
  df  <- as.data.frame(.fake_grs_dt())
  out <- grs_standardize(df, grs_cols = "GRS_a_z")
  expect_s3_class(out, "data.table")
  expect_true("GRS_a_z_z" %in% names(out))
})

test_that("grs_zscore() produces identical result to grs_standardize()", {
  dt  <- .fake_grs_dt()
  r1  <- grs_standardize(dt, grs_cols = "GRS_a_z")
  r2  <- grs_zscore(dt,      grs_cols = "GRS_a_z")
  expect_equal(r1, r2)
})


# ===========================================================================
# grs_validate()
# ===========================================================================

test_that("grs_validate() aborts on non-data.frame input", {
  expect_error(grs_validate("x", outcome_col = "outcome"), "data.frame")
})

test_that("grs_validate() aborts when outcome_col is missing", {
  dt <- .fake_grs_dt()
  expect_error(grs_validate(dt, outcome_col = "no_such_col"), "not found")
})

test_that("grs_validate() aborts when grs_cols not found in data", {
  dt <- .fake_grs_dt()
  expect_error(
    grs_validate(dt, grs_cols = "no_such_grs", outcome_col = "outcome"),
    "not found"
  )
})

test_that("grs_validate() aborts when no GRS columns auto-detected", {
  dt <- data.table::data.table(IID = 1:10, outcome = rbinom(10, 1, 0.2))
  expect_error(grs_validate(dt, outcome_col = "outcome"), "No GRS")
})

test_that("grs_validate() aborts when grs_cols are non-numeric", {
  dt <- data.table::data.table(
    IID     = 1:10,
    GRS_x   = as.character(rnorm(10)),
    outcome = rbinom(10, 1, 0.3)
  )
  expect_error(
    grs_validate(dt, grs_cols = "GRS_x", outcome_col = "outcome"),
    "numeric"
  )
})

test_that("grs_validate() returns a list with 4 named elements (logistic)", {
  skip_if_not_installed("pROC")
  dt  <- .fake_grs_dt()
  res <- suppressWarnings(
    grs_validate(dt,
                 grs_cols    = c("GRS_a_z", "GRS_b_z"),
                 outcome_col = "outcome")
  )
  expect_type(res, "list")
  expect_named(res, c("per_sd", "high_vs_low", "trend", "discrimination"))
})

test_that("grs_validate() per_sd has one row per GRS (logistic, unadjusted)", {
  skip_if_not_installed("pROC")
  dt  <- .fake_grs_dt()
  res <- suppressWarnings(
    grs_validate(dt,
                 grs_cols    = c("GRS_a_z", "GRS_b_z"),
                 outcome_col = "outcome")
  )
  expect_s3_class(res$per_sd, "data.table")
  # At minimum one unadjusted row per GRS
  expect_gte(nrow(res$per_sd), 2L)
})

test_that("grs_validate() discrimination returns AUC column (logistic)", {
  skip_if_not_installed("pROC")
  dt  <- .fake_grs_dt()
  res <- suppressWarnings(
    grs_validate(dt,
                 grs_cols    = "GRS_a_z",
                 outcome_col = "outcome")
  )
  expect_true("AUC" %in% names(res$discrimination))
  expect_gte(res$discrimination$AUC, 0)
  expect_lte(res$discrimination$AUC, 1)
})

test_that("grs_validate() returns C-index column when time_col supplied (Cox)", {
  dt  <- .fake_grs_dt()
  res <- suppressWarnings(
    grs_validate(dt,
                 grs_cols    = "GRS_a_z",
                 outcome_col = "outcome",
                 time_col    = "followup_years")
  )
  expect_true("C_index" %in% names(res$discrimination))
  expect_gte(res$discrimination$C_index, 0)
  expect_lte(res$discrimination$C_index, 1)
})

test_that("grs_validate() does not modify the user's original data", {
  skip_if_not_installed("pROC")
  dt   <- .fake_grs_dt()
  orig <- data.table::copy(dt)
  suppressWarnings(
    grs_validate(dt,
                 grs_cols    = "GRS_a_z",
                 outcome_col = "outcome")
  )
  expect_equal(dt, orig)
})


# ===========================================================================
# grs_bgen2pgen() — argument validation (no RAP submission)
# ===========================================================================

test_that("grs_bgen2pgen() aborts when chr contains out-of-range values", {
  expect_error(grs_bgen2pgen(chr = 0),  "between 1 and 22")
  expect_error(grs_bgen2pgen(chr = 23), "between 1 and 22")
})

test_that("grs_bgen2pgen() aborts when chr contains NA", {
  expect_error(grs_bgen2pgen(chr = NA_integer_), "between 1 and 22")
})

test_that("grs_bgen2pgen() aborts when maf is out of range", {
  expect_error(grs_bgen2pgen(chr = 22, maf = 0),    "maf")
  expect_error(grs_bgen2pgen(chr = 22, maf = 0.5),  "maf")
  expect_error(grs_bgen2pgen(chr = 22, maf = -0.1), "maf")
})

test_that("grs_bgen2pgen() aborts when maf is not a scalar", {
  expect_error(grs_bgen2pgen(chr = 22, maf = c(0.01, 0.05)), "maf")
})

test_that("grs_bgen2pgen() aborts on invalid instance value", {
  expect_error(grs_bgen2pgen(chr = 22, instance = "xlarge"), "arg")
})

test_that("grs_bgen2pgen() aborts on invalid priority value", {
  expect_error(grs_bgen2pgen(chr = 22, priority = "medium"), "arg")
})

test_that("grs_bgen2pgen() warns when standard instance used for chr 1-16", {
  # Mock project check so the test never reaches dx run, regardless of auth state
  local_mocked_bindings(
    .dx_get_project_id = function() NA_character_,
    .package = "ukbflow"
  )
  expect_warning(
    tryCatch(
      grs_bgen2pgen(chr = 1, instance = "standard"),
      error = function(e) NULL
    ),
    "storage|mem2_ssd1|large"
  )
})


# ===========================================================================
# grs_score() — argument validation (no RAP submission)
# ===========================================================================

test_that("grs_score() aborts when file is not a character vector", {
  expect_error(grs_score(file = 123L), "character")
})

test_that("grs_score() aborts when file entries are unnamed", {
  expect_error(grs_score(file = c("weights.txt")), "named")
})

test_that("grs_score() aborts when file has partially missing names", {
  expect_error(grs_score(file = c(grs_a = "a.txt", "b.txt")), "named")
})

test_that("grs_score() aborts on duplicate names in file", {
  expect_error(
    grs_score(file = c(grs_a = "a.txt", grs_a = "b.txt")),
    "[Dd]uplicate"
  )
})

test_that("grs_score() aborts when local weight file does not exist", {
  expect_error(
    grs_score(file = c(grs_a = "/no/such/weights.txt")),
    "not found"
  )
})

test_that("grs_score() aborts when maf is out of range", {
  withr::local_tempfile(fileext = ".txt") |> (function(f) {
    file.create(f)
    expect_error(grs_score(file = c(grs_a = f), maf = 0),   "maf")
    expect_error(grs_score(file = c(grs_a = f), maf = 0.5), "maf")
  })()
})

test_that("grs_score() aborts on invalid instance value", {
  withr::local_tempfile(fileext = ".txt") |> (function(f) {
    file.create(f)
    expect_error(grs_score(file = c(grs_a = f), instance = "mega"), "arg")
  })()
})

test_that("grs_score() aborts on invalid priority value", {
  withr::local_tempfile(fileext = ".txt") |> (function(f) {
    file.create(f)
    expect_error(grs_score(file = c(grs_a = f), priority = "urgent"), "arg")
  })()
})

# =============================================================================
# test-integration-grs.R — Integration tests for grs_ series
# Fixture: ops_toy(n=2000, seed=42) + simulated GRS raw scores
# Mirrors the style of test-integration-derive.R.
# Run manually before release: devtools::test(filter = "integration-grs")
# =============================================================================

# =============================================================================
# Shared fixture
# =============================================================================

skip_on_cran()
DT <- suppressMessages(ops_toy(n = 2000L, seed = 42L))

# Simulate two raw GRS scores correlated with a latent signal
RNGkind("Mersenne-Twister", "Inversion", "Rejection")
set.seed(42L)
N          <- nrow(DT)
grs_signal <- rnorm(N)

DT[, GRS_a := grs_signal + rnorm(N, sd = 0.5)]
DT[, GRS_b := grs_signal + rnorm(N, sd = 0.8)]

# Binary outcome correlated with GRS signal (~15% event rate)
DT[, outcome := as.integer(
  plogis(-2.5 + 0.4 * grs_signal + rnorm(N, sd = 0.5)) > runif(N)
)]

# Follow-up time from baseline (p53_i0) to admin censor date
CENSOR <- as.Date("2022-06-01")
DT[, followup_years := as.numeric(
  difftime(CENSOR, as.Date(p53_i0), units = "days") / 365.25
)]

# Covariates available from ops_toy
COVS <- c("p21022", "p31", "p22189")  # age at recruitment, sex, TDI


# =============================================================================
# grs_check() — integration
# =============================================================================

test_that("grs_check() round-trips: writes and the output is plink2-ready", {
  input <- withr::local_tempfile(fileext = ".csv")
  dest  <- withr::local_tempfile(fileext = ".txt")

  data.table::fwrite(
    data.table::data.table(
      snp           = paste0("rs", 1:20),
      effect_allele = sample(c("A", "T", "C", "G"), 20, TRUE),
      beta          = rnorm(20, 0, 0.3)
    ),
    input
  )

  result <- suppressMessages(grs_check(input, dest = dest))

  expect_true(file.exists(dest))
  out_dt <- data.table::fread(dest)
  expect_equal(ncol(out_dt), 3L)
  expect_equal(nrow(out_dt), 20L)
  expect_equal(nrow(result), 20L)
  expect_true(is.numeric(result$beta))
})

test_that("grs_check() normalises lowercase effect_allele to uppercase", {
  input <- withr::local_tempfile(fileext = ".csv")
  dest  <- withr::local_tempfile(fileext = ".txt")

  data.table::fwrite(
    data.table::data.table(
      snp           = paste0("rs", 1:3),
      effect_allele = c("a", "t", "c"),
      beta          = c(0.1, 0.2, 0.3)
    ),
    input
  )

  result <- suppressMessages(grs_check(input, dest = dest))
  expect_true(all(result$effect_allele %in% c("A", "T", "C", "G")))
})


# =============================================================================
# grs_standardize() — integration with ops_toy GRS columns
# =============================================================================

test_that("grs_standardize() produces z-scores with mean≈0 and SD≈1 on ops_toy data", {
  out <- suppressMessages(
    grs_standardize(DT, grs_cols = c("GRS_a", "GRS_b"))
  )

  for (col in c("GRS_a_z", "GRS_b_z")) {
    v <- out[[col]]
    expect_lt(abs(mean(v, na.rm = TRUE)), 1e-10)
    expect_lt(abs(sd(v,   na.rm = TRUE) - 1), 1e-10)
  }
})

test_that("grs_standardize() auto-detects both GRS columns from ops_toy data", {
  out <- suppressMessages(grs_standardize(DT))
  expect_true("GRS_a_z" %in% names(out))
  expect_true("GRS_b_z" %in% names(out))
})

test_that("grs_standardize() adds exactly 2 columns for 2 GRS inputs", {
  out <- suppressMessages(
    grs_standardize(DT, grs_cols = c("GRS_a", "GRS_b"))
  )
  expect_equal(ncol(out), ncol(DT) + 2L)
})

test_that("grs_standardize() inserts _z column immediately after source column", {
  out     <- suppressMessages(grs_standardize(DT, grs_cols = "GRS_a"))
  idx_src <- which(names(out) == "GRS_a")
  idx_z   <- which(names(out) == "GRS_a_z")
  expect_equal(idx_z, idx_src + 1L)
})

test_that("grs_standardize() does not modify original GRS_a column in ops_toy", {
  orig_a <- DT$GRS_a
  suppressMessages(grs_standardize(DT, grs_cols = "GRS_a"))
  expect_equal(DT$GRS_a, orig_a)
})


# =============================================================================
# grs_validate() helpers — standardize first, then validate
# =============================================================================

# Pre-standardize GRS columns once for downstream validate tests
DT_Z <- suppressMessages(
  grs_standardize(DT, grs_cols = c("GRS_a", "GRS_b"))
)


# =============================================================================
# grs_validate() — logistic (time_col = NULL)
# =============================================================================

test_that("grs_validate() logistic runs without error on ops_toy n=2000", {
  skip_if_not_installed("pROC")
  expect_no_error(suppressWarnings(suppressMessages(
    grs_validate(DT_Z,
                 grs_cols    = c("GRS_a_z", "GRS_b_z"),
                 outcome_col = "outcome")
  )))
})

test_that("grs_validate() logistic per_sd has positive OR and valid CI", {
  skip_if_not_installed("pROC")
  res <- suppressWarnings(suppressMessages(
    grs_validate(DT_Z,
                 grs_cols    = c("GRS_a_z", "GRS_b_z"),
                 outcome_col = "outcome")
  ))
  expect_true(all(res$per_sd$OR > 0))
  expect_true(all(res$per_sd$CI_lower < res$per_sd$OR))
  expect_true(all(res$per_sd$OR       < res$per_sd$CI_upper))
})

test_that("grs_validate() logistic high_vs_low contains only High rows", {
  skip_if_not_installed("pROC")
  res <- suppressWarnings(suppressMessages(
    grs_validate(DT_Z,
                 grs_cols    = c("GRS_a_z", "GRS_b_z"),
                 outcome_col = "outcome")
  ))
  expect_true(all(grepl("High$", res$high_vs_low$term)))
})

test_that("grs_validate() logistic AUC is in [0.5, 1] with real GRS signal", {
  skip_if_not_installed("pROC")
  res <- suppressWarnings(suppressMessages(
    grs_validate(DT_Z,
                 grs_cols    = "GRS_a_z",
                 outcome_col = "outcome")
  ))
  expect_gte(res$discrimination$AUC, 0.5)
  expect_lte(res$discrimination$AUC, 1.0)
})

test_that("grs_validate() logistic with covariates adds fully adjusted model", {
  skip_if_not_installed("pROC")
  res <- suppressWarnings(suppressMessages(
    grs_validate(DT_Z,
                 grs_cols    = "GRS_a_z",
                 outcome_col = "outcome",
                 covariates  = COVS)
  ))
  expect_true("Fully adjusted" %in% as.character(res$per_sd$model))
})


# =============================================================================
# grs_validate() — Cox (time_col supplied)
# =============================================================================

test_that("grs_validate() Cox runs without error on ops_toy n=2000", {
  expect_no_error(suppressWarnings(suppressMessages(
    grs_validate(DT_Z,
                 grs_cols    = c("GRS_a_z", "GRS_b_z"),
                 outcome_col = "outcome",
                 time_col    = "followup_years")
  )))
})

test_that("grs_validate() Cox per_sd has positive HR and valid CI", {
  res <- suppressWarnings(suppressMessages(
    grs_validate(DT_Z,
                 grs_cols    = c("GRS_a_z", "GRS_b_z"),
                 outcome_col = "outcome",
                 time_col    = "followup_years")
  ))
  expect_true(all(res$per_sd$HR > 0))
  expect_true(all(res$per_sd$CI_lower < res$per_sd$HR))
  expect_true(all(res$per_sd$HR       < res$per_sd$CI_upper))
})

test_that("grs_validate() Cox C-index is in [0, 1] with CI bounds correct", {
  res <- suppressWarnings(suppressMessages(
    grs_validate(DT_Z,
                 grs_cols    = c("GRS_a_z", "GRS_b_z"),
                 outcome_col = "outcome",
                 time_col    = "followup_years")
  ))
  expect_true(all(res$discrimination$C_index >= 0))
  expect_true(all(res$discrimination$C_index <= 1))
  expect_true(all(res$discrimination$CI_lower < res$discrimination$C_index))
  expect_true(all(res$discrimination$C_index  < res$discrimination$CI_upper))
})

test_that("grs_validate() Cox C-index is above 0.5 with real GRS signal", {
  res <- suppressWarnings(suppressMessages(
    grs_validate(DT_Z,
                 grs_cols    = "GRS_a_z",
                 outcome_col = "outcome",
                 time_col    = "followup_years")
  ))
  expect_gte(res$discrimination$C_index, 0.5)
})

test_that("grs_validate() Cox trend data.table has one row per GRS × model", {
  res <- suppressWarnings(suppressMessages(
    grs_validate(DT_Z,
                 grs_cols    = c("GRS_a_z", "GRS_b_z"),
                 outcome_col = "outcome",
                 time_col    = "followup_years")
  ))
  expect_s3_class(res$trend, "data.table")
  expect_gte(nrow(res$trend), 2L)
})

test_that("grs_validate() Cox does not modify original ops_toy data", {
  orig <- data.table::copy(DT_Z)
  suppressWarnings(suppressMessages(
    grs_validate(DT_Z,
                 grs_cols    = "GRS_a_z",
                 outcome_col = "outcome",
                 time_col    = "followup_years")
  ))
  expect_equal(DT_Z, orig)
})


# =============================================================================
# End-to-end pipeline: ops_toy → GRS_a raw → standardize → validate (Cox)
# =============================================================================

test_that("full GRS pipeline ops_toy → standardize → Cox validate runs without error", {
  d <- data.table::copy(DT)

  suppressWarnings(suppressMessages({
    # Standardize raw GRS columns
    d <- grs_standardize(d, grs_cols = c("GRS_a", "GRS_b"))

    # Validate via Cox
    res <- grs_validate(d,
                        grs_cols    = c("GRS_a_z", "GRS_b_z"),
                        outcome_col = "outcome",
                        time_col    = "followup_years",
                        covariates  = COVS)
  }))

  expect_named(res, c("per_sd", "high_vs_low", "trend", "discrimination"))
  expect_true(all(res$per_sd$HR > 0))
  expect_true("C_index" %in% names(res$discrimination))
  expect_true(data.table::is.data.table(res$per_sd))
})

# =============================================================================
# utils_assoc.R — internal helpers for assoc_ series
# =============================================================================


# Normalise the outcome vector to integer 0/1.
# Accepts logical (T/F) or integer/numeric (0/1). Aborts on any other values.
.normalise_event <- function(x, col) {
  if (is.logical(x)) {
    cli::cli_alert_info(
      "outcome_col {.field {col}}: logical detected, converting TRUE/FALSE -> 1/0"
    )
    return(as.integer(x))
  }
  xi  <- suppressWarnings(as.integer(x))
  bad <- setdiff(stats::na.omit(unique(xi)), c(0L, 1L))
  if (length(bad) > 0L) {
    cli::cli_abort(
      "outcome_col {.field {col}}: values other than 0/1/NA found: {bad}",
      call = NULL
    )
  }
  xi
}


# Normalise logical exposure columns to integer in-place (on a data.table copy).
# Logical TRUE/FALSE -> 1/0 so coxph produces clean term names (e.g. "ad_tf"
# rather than "ad_tfTRUE").
.normalise_logical_exposures <- function(dt, exposure_cols) {
  for (col in exposure_cols) {
    if (is.logical(dt[[col]])) {
      cli::cli_alert_info(
        "exposure_col {.field {col}}: logical detected, converting TRUE/FALSE -> 1/0"
      )
      dt[, (col) := as.integer(get(col))]
    }
  }
  invisible(dt)
}


# Auto-detect the age column via UKB field 21022 (age at recruitment).
# Falls back to a grep on "^age" if the field cache is unavailable.
# Returns NULL with a warning when not found.
.detect_age_col <- function(data) {
  cols <- .detect_cols_by_field(data, 21022L)
  if (length(cols) > 0L) return(cols[1L])
  cols <- grep("^age", names(data), value = TRUE, ignore.case = TRUE)
  if (length(cols) > 0L) return(cols[1L])
  NULL
}


# Auto-detect the sex column via UKB field 31.
# Falls back to an exact "sex" match then a case-insensitive grep.
# Returns NULL with a warning when not found.
.detect_sex_col <- function(data) {
  cols <- .detect_cols_by_field(data, 31L)
  if (length(cols) > 0L) return(cols[1L])
  if ("sex" %in% names(data)) return("sex")
  cols <- grep("^sex$", names(data), value = TRUE, ignore.case = TRUE)
  if (length(cols) > 0L) return(cols[1L])
  NULL
}


# Extract tidy HR / CI / p-value rows for one exposure from a fitted coxph model.
# For factor exposures, all level rows are returned (one per non-reference level).
# For numeric / binary exposures, a single row is returned.
#
# Args:
#   model      (coxph)   Fitted model.
#   exposure   (character) Exposure variable name as it appears in the formula.
#   is_factor  (logical)   Whether the exposure is a factor in the data.
#   conf_level (numeric)   Confidence level, e.g. 0.95.
#
# Returns:
#   data.table with columns: term, HR, CI_lower, CI_upper, p_value.
#   NULL if no matching terms are found.
.extract_cox_terms <- function(model, exposure, is_factor, conf_level) {
  coef_mat  <- summary(model)$coefficients
  # confint() returns log-scale CI; exp() converts to HR scale
  ci_log    <- confint(model, level = conf_level)
  if (!is.matrix(ci_log)) {
    ci_log <- matrix(ci_log, nrow = 1L,
                     dimnames = list(exposure, names(ci_log)))
  }

  all_terms <- rownames(coef_mat)

  # Factor: terms are paste0(exposure, level) — use startsWith
  # Numeric / binary: term equals exposure name exactly
  idx <- if (is_factor) {
    which(startsWith(all_terms, exposure))
  } else {
    which(all_terms == exposure)
  }

  if (length(idx) == 0L) {
    cli::cli_alert_warning(
      "  No terms found for exposure {.field {exposure}} -- skipped."
    )
    return(NULL)
  }

  data.table::data.table(
    term     = all_terms[idx],
    HR       = exp(coef_mat[idx, "coef"]),
    CI_lower = exp(ci_log[idx, 1L]),
    CI_upper = exp(ci_log[idx, 2L]),
    p_value  = coef_mat[idx, "Pr(>|z|)"]
  )
}


# Fit one coxph model for a single exposure and return a tidy data.table.
# The event column is expected to be pre-added as ".ukb_event" in dt.
#
# n, n_events, and person_years are all extracted from the fitted model object
# so that listwise deletion (due to covariate NA) is automatically reflected —
# each model reports the statistics for its actual analysis set, not the
# full input data.
#
# Args:
#   dt           (data.table) Analysis dataset with ".ukb_event" column.
#   time_col     (character)  Follow-up time column name.
#   exposure     (character)  Single exposure variable name.
#   covariates   (character or NULL) Covariate column names.
#   strata       (character or NULL) Strata variable name.
#   model_label  (character)  Human-readable model name for the output.
#   conf_level   (numeric)    Confidence level for CI.
#
# Returns:
#   data.table or NULL on failure.
.run_one_cox_model <- function(dt, time_col, exposure, covariates,
                                strata, model_label, conf_level) {

  is_factor <- is.factor(dt[[exposure]])

  rhs_parts <- c(exposure, covariates)
  if (!is.null(strata)) {
    rhs_parts <- c(rhs_parts, paste0("strata(", strata, ")"))
  }
  rhs <- paste(rhs_parts, collapse = " + ")

  fml <- stats::as.formula(
    sprintf("Surv(%s, .ukb_event) ~ %s", time_col, rhs)
  )

  model <- tryCatch(
    survival::coxph(fml, data = dt),
    error = function(e) {
      cli::cli_alert_warning(
        "  [{model_label} | {.field {exposure}}] model failed: {conditionMessage(e)}"
      )
      NULL
    }
  )
  if (is.null(model)) return(NULL)

  terms_dt <- .extract_cox_terms(model, exposure, is_factor, conf_level)
  if (is.null(terms_dt)) return(NULL)

  # Extract n, n_events, person_years from the model itself so that listwise
  # deletion due to covariate missingness is automatically reflected.
  # model$y is the Surv matrix for the actual analysis set (after NA removal).
  # For Surv(time, event), column 1 = time, column 2 = status.
  surv_mat     <- model$y
  n_events_mod <- model$nevent
  person_years_mod <- round(sum(surv_mat[, 1L], na.rm = TRUE))

  terms_dt[, `:=`(
    exposure     = exposure,
    model        = model_label,
    n            = model$n,
    n_events     = n_events_mod,
    person_years = person_years_mod,
    HR_label     = sprintf("%.2f (%.2f-%.2f)", HR, CI_lower, CI_upper)
  )]

  data.table::setcolorder(
    terms_dt,
    c("exposure", "term", "model", "n", "n_events", "person_years",
      "HR", "CI_lower", "CI_upper", "p_value", "HR_label")
  )
  terms_dt
}


# Fit one logistic regression model for a single exposure and return a tidy
# data.table. The event column is expected to be pre-added as ".ukb_event" in dt.
#
# n and n_cases are extracted from the fitted model so that listwise deletion
# due to covariate missingness is automatically reflected.
#
# Args:
#   dt           (data.table) Analysis dataset with ".ukb_event" column.
#   exposure     (character)  Single exposure variable name.
#   covariates   (character or NULL) Covariate column names.
#   model_label  (character)  Human-readable model name for the output.
#   ci_method    (character)  "wald" (default) or "profile".
#   conf_level   (numeric)    Confidence level for CI.
#
# Returns:
#   data.table or NULL on failure.
.run_one_logistic_model <- function(dt, exposure, covariates,
                                     model_label, ci_method, conf_level) {

  is_factor <- is.factor(dt[[exposure]])

  rhs <- paste(c(exposure, covariates), collapse = " + ")
  fml <- stats::as.formula(sprintf(".ukb_event ~ %s", rhs))

  model <- tryCatch(
    stats::glm(fml, data = dt, family = stats::binomial(link = "logit")),
    error = function(e) {
      cli::cli_alert_warning(
        "  [{model_label} | {.field {exposure}}] model failed: {conditionMessage(e)}"
      )
      NULL
    }
  )
  if (is.null(model)) return(NULL)

  coef_mat  <- summary(model)$coefficients
  all_terms <- rownames(coef_mat)

  idx <- if (is_factor) {
    which(startsWith(all_terms, exposure))
  } else {
    which(all_terms == exposure)
  }

  if (length(idx) == 0L) {
    cli::cli_alert_warning(
      "  No terms found for exposure {.field {exposure}} -- skipped."
    )
    return(NULL)
  }

  # CI on log-OR scale; exp() gives OR CI
  # Reason: confint.default() = Wald (fast); confint.glm() = profile likelihood
  ci_log <- if (ci_method == "wald") {
    stats::confint.default(model, level = conf_level)
  } else {
    suppressMessages(stats::confint(model, level = conf_level))
  }
  if (!is.matrix(ci_log)) {
    ci_log <- matrix(ci_log, nrow = 1L,
                     dimnames = list(all_terms[idx], names(ci_log)))
  }

  terms_dt <- data.table::data.table(
    term     = all_terms[idx],
    OR       = exp(coef_mat[idx, "Estimate"]),
    CI_lower = exp(ci_log[idx, 1L]),
    CI_upper = exp(ci_log[idx, 2L]),
    p_value  = coef_mat[idx, "Pr(>|z|)"]
  )

  # n and n_cases from the model's actual analysis set (after listwise deletion)
  n_cases_mod <- sum(model$y, na.rm = TRUE)

  terms_dt[, `:=`(
    exposure = exposure,
    model    = model_label,
    n        = stats::nobs(model),
    n_cases  = n_cases_mod,
    OR_label = sprintf("%.2f (%.2f-%.2f)", OR, CI_lower, CI_upper)
  )]

  data.table::setcolorder(
    terms_dt,
    c("exposure", "term", "model", "n", "n_cases",
      "OR", "CI_lower", "CI_upper", "p_value", "OR_label")
  )
  terms_dt
}


# Fit one linear regression model for a single exposure and return a tidy
# data.table. The outcome column is used directly (continuous numeric).
#
# n is extracted from the fitted model so that listwise deletion due to
# covariate missingness is automatically reflected.
# SE is included alongside beta to support downstream meta-analysis.
#
# Args:
#   dt           (data.table) Analysis dataset.
#   outcome_col  (character)  Continuous outcome column name.
#   exposure     (character)  Single exposure variable name.
#   covariates   (character or NULL) Covariate column names.
#   model_label  (character)  Human-readable model name for the output.
#   conf_level   (numeric)    Confidence level for CI (t-distribution based).
#
# Returns:
#   data.table or NULL on failure.
.run_one_linear_model <- function(dt, outcome_col, exposure, covariates,
                                   model_label, conf_level) {

  is_factor <- is.factor(dt[[exposure]])

  rhs <- paste(c(exposure, covariates), collapse = " + ")
  fml <- stats::as.formula(sprintf("%s ~ %s", outcome_col, rhs))

  model <- tryCatch(
    stats::lm(fml, data = dt),
    error = function(e) {
      cli::cli_alert_warning(
        "  [{model_label} | {.field {exposure}}] model failed: {conditionMessage(e)}"
      )
      NULL
    }
  )
  if (is.null(model)) return(NULL)

  coef_mat  <- summary(model)$coefficients
  all_terms <- rownames(coef_mat)

  idx <- if (is_factor) {
    which(startsWith(all_terms, exposure))
  } else {
    which(all_terms == exposure)
  }

  if (length(idx) == 0L) {
    cli::cli_alert_warning(
      "  No terms found for exposure {.field {exposure}} -- skipped."
    )
    return(NULL)
  }

  # CI via confint.lm() uses t-distribution (exact for normal linear models)
  ci_mat <- stats::confint(model, level = conf_level)
  if (!is.matrix(ci_mat)) {
    ci_mat <- matrix(ci_mat, nrow = 1L,
                     dimnames = list(all_terms[idx], names(ci_mat)))
  }

  terms_dt <- data.table::data.table(
    term     = all_terms[idx],
    beta     = coef_mat[idx, "Estimate"],
    se       = coef_mat[idx, "Std. Error"],
    CI_lower = ci_mat[idx, 1L],
    CI_upper = ci_mat[idx, 2L],
    p_value  = coef_mat[idx, "Pr(>|t|)"]
  )

  terms_dt[, `:=`(
    exposure   = exposure,
    model      = model_label,
    n          = stats::nobs(model),
    beta_label = sprintf("%.2f (%.2f-%.2f)", beta, CI_lower, CI_upper)
  )]

  data.table::setcolorder(
    terms_dt,
    c("exposure", "term", "model", "n",
      "beta", "se", "CI_lower", "CI_upper", "p_value", "beta_label")
  )
  terms_dt
}


# Fit one coxph model and run cox.zph() PH assumption test for one exposure.
# Returns a tidy data.table with Schoenfeld residual test results.
# Term-level results (one row per exposure term) are combined with the
# global test for the whole model (global_chisq / global_df / global_p).
#
# Args:
#   dt           (data.table) Dataset with ".ukb_event" column pre-added.
#   time_col     (character)  Follow-up time column name.
#   exposure     (character)  Single exposure variable name.
#   covariates   (character or NULL) Covariate column names.
#   strata       (character or NULL) Strata variable name.
#   model_label  (character)  Human-readable model name for the output.
#
# Returns:
#   data.table or NULL on failure.
.run_one_zph_test <- function(dt, time_col, exposure, covariates,
                               strata, model_label) {

  is_factor <- is.factor(dt[[exposure]])

  rhs_parts <- c(exposure, covariates)
  if (!is.null(strata)) {
    rhs_parts <- c(rhs_parts, paste0("strata(", strata, ")"))
  }
  rhs <- paste(rhs_parts, collapse = " + ")
  fml <- stats::as.formula(
    sprintf("Surv(%s, .ukb_event) ~ %s", time_col, rhs)
  )

  model <- tryCatch(
    survival::coxph(fml, data = dt),
    error = function(e) {
      cli::cli_alert_warning(
        "  [{model_label} | {.field {exposure}}] coxph failed: {conditionMessage(e)}"
      )
      NULL
    }
  )
  if (is.null(model)) return(NULL)

  zph <- tryCatch(
    survival::cox.zph(model),
    error = function(e) {
      cli::cli_alert_warning(
        "  [{model_label} | {.field {exposure}}] cox.zph failed: {conditionMessage(e)}"
      )
      NULL
    }
  )
  if (is.null(zph)) return(NULL)

  tbl       <- zph$table           # rows = terms + "GLOBAL", cols = chisq, df, p
  all_terms <- rownames(tbl)

  # Global test row (one per model)
  global_chisq <- tbl["GLOBAL", "chisq"]
  global_df    <- tbl["GLOBAL", "df"]
  global_p     <- tbl["GLOBAL", "p"]

  # Exposure-specific term rows
  idx <- if (is_factor) {
    which(startsWith(all_terms, exposure))
  } else {
    which(all_terms == exposure)
  }

  if (length(idx) == 0L) {
    cli::cli_alert_warning(
      "  No terms found for exposure {.field {exposure}} in zph table -- skipped."
    )
    return(NULL)
  }

  terms_dt <- data.table::data.table(
    term         = all_terms[idx],
    chisq        = tbl[idx, "chisq"],
    df           = tbl[idx, "df"],
    p_value      = tbl[idx, "p"],
    ph_satisfied = tbl[idx, "p"] > 0.05,
    global_chisq = global_chisq,
    global_df    = global_df,
    global_p     = global_p
  )

  terms_dt[, `:=`(
    exposure = exposure,
    model    = model_label
  )]

  data.table::setcolorder(
    terms_dt,
    c("exposure", "term", "model",
      "chisq", "df", "p_value", "ph_satisfied",
      "global_chisq", "global_df", "global_p")
  )
  terms_dt
}


# Fit one Fine-Gray competing risks model (finegray + weighted coxph) for a
# single exposure and return a tidy data.table.
#
# The analysis dataset `dt` must already contain `.fg_status` (factor with
# levels "censor", "event", "compete") and `time_col`. Only the columns
# listed in `need_cols` are passed to finegray(), so the formula `~ .`
# carries exactly the right predictors into the weighted coxph step.
#
# SHR and CI are extracted from `summary()$conf.int` (columns 1, 3, 4):
#   1 = exp(coef)  →  SHR
#   3 = lower .xx  →  CI_lower
#   4 = upper .xx  →  CI_upper
# p-value from `summary()$coefficients[, "Pr(>|z|)"]` (robust z-based).
#
# n_compete is counted from the complete-case subset *before* finegray()
# expansion, i.e. the number of participants with a competing event who
# contributed to the analysis.
#
# Args:
#   dt           (data.table) Dataset with `.fg_status` pre-built.
#   time_col     (character)  Follow-up time column name.
#   exposure     (character)  Single exposure variable name.
#   covariates   (character or NULL) Covariate column names.
#   model_label  (character)  Human-readable model name for the output.
#   conf_level   (numeric)    Confidence level for CI.
#
# Returns:
#   data.table or NULL on failure.
.run_one_fg_model <- function(dt, time_col, exposure, covariates,
                               model_label, conf_level) {

  is_factor <- is.factor(dt[[exposure]])

  # Keep only columns needed; drop NA rows (complete-case analysis)
  need_cols <- unique(c(time_col, ".fg_status", exposure, covariates))
  need_cols <- intersect(need_cols, names(dt))
  sub <- stats::na.omit(dt[, need_cols, with = FALSE])

  if (nrow(sub) == 0L) {
    cli::cli_alert_warning(
      "  [{model_label} | {.field {exposure}}] no complete cases -- skipped."
    )
    return(NULL)
  }

  n_compete <- sum(sub$.fg_status == "compete", na.rm = TRUE)

  # Bind Surv locally so finegray() can resolve it when evaluating the formula.
  # finegray() evaluates formulas in the caller's environment, which inside a
  # package function does not automatically include survival:: exports.
  Surv <- survival::Surv  # nolint: object_usage_linter

  # Expand data with Fine-Gray inverse probability of censoring weights
  fg_fml  <- stats::as.formula(sprintf("Surv(%s, .fg_status) ~ .", time_col))
  fg_data <- tryCatch(
    survival::finegray(fg_fml, data = sub, etype = "event"),
    error = function(e) {
      cli::cli_alert_warning(
        "  [{model_label} | {.field {exposure}}] finegray() failed: {conditionMessage(e)}"
      )
      NULL
    }
  )
  if (is.null(fg_data)) return(NULL)

  # Weighted Cox on the expanded Fine-Gray dataset
  rhs     <- paste(c(exposure, covariates), collapse = " + ")
  cox_fml <- stats::as.formula(
    sprintf("Surv(fgstart, fgstop, fgstatus) ~ %s", rhs)
  )
  model <- tryCatch(
    survival::coxph(cox_fml, data = fg_data, weight = fgwt),
    error = function(e) {
      cli::cli_alert_warning(
        "  [{model_label} | {.field {exposure}}] weighted coxph failed: {conditionMessage(e)}"
      )
      NULL
    }
  )
  if (is.null(model)) return(NULL)

  s        <- summary(model)
  coef_mat <- s$coefficients  # coef | exp(coef) | se(coef) | robust se | z | Pr(>|z|)
  ci_mat   <- s$conf.int      # exp(coef) | exp(-coef) | lower .xx | upper .xx

  # Ensure matrix (single-term model returns named vector)
  if (!is.matrix(coef_mat)) {
    coef_mat <- matrix(coef_mat, nrow = 1L,
                       dimnames = list(exposure, names(coef_mat)))
    ci_mat   <- matrix(ci_mat,   nrow = 1L,
                       dimnames = list(exposure, names(ci_mat)))
  }

  all_terms <- rownames(coef_mat)
  idx <- if (is_factor) {
    which(startsWith(all_terms, exposure))
  } else {
    which(all_terms == exposure)
  }

  if (length(idx) == 0L) {
    cli::cli_alert_warning(
      "  No terms found for exposure {.field {exposure}} -- skipped."
    )
    return(NULL)
  }

  terms_dt <- data.table::data.table(
    term     = all_terms[idx],
    SHR      = ci_mat[idx, 1L],   # exp(coef)
    CI_lower = ci_mat[idx, 3L],   # lower .xx
    CI_upper = ci_mat[idx, 4L],   # upper .xx
    p_value  = coef_mat[idx, "Pr(>|z|)"]
  )

  terms_dt[, `:=`(
    exposure  = exposure,
    model     = model_label,
    n         = model$n,
    n_events  = model$nevent,
    n_compete = n_compete,
    SHR_label = sprintf("%.2f (%.2f-%.2f)", SHR, CI_lower, CI_upper)
  )]

  data.table::setcolorder(
    terms_dt,
    c("exposure", "term", "model", "n", "n_events", "n_compete",
      "SHR", "CI_lower", "CI_upper", "p_value", "SHR_label")
  )
  terms_dt
}


# Compute an LRT p-value for the exposure x by interaction on the full dataset.
# Fits a reduced model (exposure + by + covariates) and a full model
# (exposure * by + covariates) and extracts the LRT p-value via anova().
# Returns NA_real_ (with a warning) when either model fails or the p-value
# column cannot be located.
#
# For coxph: anova(mod_r, mod_f) uses the log-likelihood ratio.
# For glm:   anova(mod_r, mod_f, test = "Chisq") uses the deviance LRT.
# For lm:    anova(mod_r, mod_f) uses an F-test (equivalent to LRT for normal
#            linear models).
#
# The p-value column is identified by grepping for column names starting with
# "P" or "Pr", which covers all three anova output formats:
#   coxph   → "P(>|Chi|)"
#   glm     → "Pr(>Chi)"
#   lm      → "Pr(>F)"
#
# Args:
#   dt           (data.table) Full dataset; .ukb_event pre-added for coxph/logistic.
#   method       (character) "coxph", "logistic", or "linear".
#   outcome_col  (character) Outcome column name (continuous for linear).
#   time_col     (character or NULL) Follow-up time column (coxph only).
#   exposure     (character) Exposure variable name.
#   by           (character) Subgroup stratification variable name.
#   covariates   (character or NULL) Additional covariates.
#   model_label  (character) Label for warning messages.
#
# Returns:
#   numeric scalar p-value, or NA_real_ on failure.
.run_one_interaction_lrt <- function(dt, method, outcome_col, time_col,
                                      exposure, by, covariates, model_label) {

  rhs_base <- paste(c(exposure, by, covariates), collapse = " + ")
  rhs_full <- paste(c(paste0(exposure, " * ", by), covariates), collapse = " + ")

  lrt_p <- tryCatch({
    if (method == "coxph") {
      fml_r <- stats::as.formula(
        sprintf("Surv(%s, .ukb_event) ~ %s", time_col, rhs_base))
      fml_f <- stats::as.formula(
        sprintf("Surv(%s, .ukb_event) ~ %s", time_col, rhs_full))
      mod_r <- survival::coxph(fml_r, data = dt)
      mod_f <- survival::coxph(fml_f, data = dt)
      aov   <- anova(mod_r, mod_f)
    } else if (method == "logistic") {
      fml_r <- stats::as.formula(sprintf(".ukb_event ~ %s", rhs_base))
      fml_f <- stats::as.formula(sprintf(".ukb_event ~ %s", rhs_full))
      mod_r <- stats::glm(fml_r, data = dt, family = stats::binomial(link = "logit"))
      mod_f <- stats::glm(fml_f, data = dt, family = stats::binomial(link = "logit"))
      aov   <- stats::anova(mod_r, mod_f, test = "Chisq")
    } else {  # linear
      fml_r <- stats::as.formula(sprintf("%s ~ %s", outcome_col, rhs_base))
      fml_f <- stats::as.formula(sprintf("%s ~ %s", outcome_col, rhs_full))
      mod_r <- stats::lm(fml_r, data = dt)
      mod_f <- stats::lm(fml_f, data = dt)
      aov   <- stats::anova(mod_r, mod_f)
    }

    # Locate the p-value column (covers all three anova output formats)
    p_col <- grep("^(P|Pr)", colnames(aov), value = TRUE)[1L]
    if (is.na(p_col)) {
      cli::cli_alert_warning(
        "  [{model_label} | {.field {exposure}}] interaction LRT: p-value column not found in anova output."
      )
      return(NA_real_)
    }
    as.numeric(utils::tail(aov, 1L)[[p_col]])

  }, error = function(e) {
    cli::cli_alert_warning(
      "  [{model_label} | {.field {exposure}}] interaction LRT failed: {conditionMessage(e)}"
    )
    NA_real_
  })

  lrt_p
}

# =============================================================================
# utils_job.R — Internal helpers for job_ series
# Depends on .dx_run() defined in utils_auth.R
# =============================================================================


#' Run dx describe job-XXXX --json and return parsed R list
#'
#' @param job_id (character) Job ID, e.g. \code{"job-XXXX"}.
#' @param timeout (integer) Timeout in seconds. Default: 30.
#' @return Named R list from jsonlite::parse_json().
#'
#' @keywords internal
#' @noRd
.dx_job_describe <- function(job_id, timeout = 30) {
  result <- .dx_run(
    c("describe", job_id, "--json"),
    timeout = timeout,
    env     = c(PYTHONIOENCODING = "utf-8")
  )
  if (!result$success) {
    cli::cli_abort("Failed to describe job {.val {job_id}}: {result$stderr}", call = NULL)
  }
  jsonlite::parse_json(result$stdout)
}


#' Extract the output CSV file ID from a job describe list
#'
#' @param desc Named list returned by \code{.dx_job_describe()}.
#' @return Character string — DNAnexus file ID (e.g. \code{"file-XXXX"}).
#'
#' @keywords internal
#' @noRd
.dx_job_output_id <- function(desc) {
  # Reason: table-exporter puts CSV under output$csv (an array);
  # output is NULL for failed/running jobs so we guard explicitly
  csv <- desc$output$csv
  if (is.null(csv) || length(csv) == 0) {
    cli::cli_abort(
      c("Job has no output CSV.",
        "i" = "Check job state with {.fn job_status} before calling {.fn job_result}."),
      call = NULL
    )
  }
  csv[[1]]$`$dnanexus_link`
}


#' Run dx find jobs and return raw stdout
#'
#' @param n (integer) Maximum number of results. Default: 20.
#' @param timeout (integer) Timeout in seconds. Default: 30.
#' @return Named list with stdout, stderr, status, success.
#'
#' @keywords internal
#' @noRd
.dx_find_jobs_raw <- function(n = 20, timeout = 30) {
  .dx_run(
    c("find", "jobs", "--num-results", as.character(n)),
    timeout = timeout,
    env     = c(PYTHONIOENCODING = "utf-8")
  )
}


#' Resolve a DNAnexus file ID to its /mnt/project/ path
#'
#' @param file_id (character) DNAnexus file ID, e.g. \code{"file-XXXX"}.
#' @return Character string — absolute path under \code{/mnt/project/}.
#'
#' @keywords internal
#' @noRd
.dx_file_path <- function(file_id) {
  result <- .dx_run(c("describe", file_id, "--json"), timeout = 30,
                    env = c(PYTHONIOENCODING = "utf-8"))
  if (!result$success) {
    cli::cli_abort("Failed to describe file {.val {file_id}}: {result$stderr}", call = NULL)
  }
  desc   <- jsonlite::parse_json(result$stdout)
  folder <- sub("^/+", "", if (is.null(desc$folder)) "" else desc$folder)
  name   <- desc$name
  if (nzchar(folder)) file.path("/mnt/project", folder, name) else
                       file.path("/mnt/project", name)
}


#' Parse dx find jobs text output into a data.frame
#'
#' @param stdout (character) Raw stdout from \code{dx find jobs}.
#' @return data.frame with columns: job_id, name, state, created, runtime.
#'
#' @keywords internal
#' @noRd
.dx_parse_jobs <- function(stdout) {

  empty <- data.frame(
    job_id  = character(0),
    name    = character(0),
    state   = character(0),
    created = as.POSIXct(character(0)),
    runtime = character(0),
    stringsAsFactors = FALSE
  )

  lines <- strsplit(stdout, "\n")[[1]]
  lines <- lines[nzchar(trimws(lines))]
  if (length(lines) == 0) return(empty)

  rows <- list()
  i    <- 1L

  while (i <= length(lines)) {
    line1 <- lines[i]

    # Only process lines starting with "* " that describe a job
    if (!startsWith(line1, "*")) { i <- i + 1L; next }

    # Line 1 format: * <name> (<exec>:<func>) (<state>) <job-XXXX>
    m1   <- regexec(
      "^\\*\\s+(.+?)\\s+\\(.+?\\)\\s+\\((\\w+)\\)\\s+(job-\\S+)",
      line1, perl = TRUE
    )
    cap1 <- regmatches(line1, m1)[[1]]

    # Skip lines that don't match (e.g. "* More results not shown...")
    if (length(cap1) < 4) { i <- i + 1L; next }

    name   <- cap1[2]
    state  <- cap1[3]
    job_id <- cap1[4]

    # Line 2 format: <user> <YYYY-MM-DD HH:MM:SS> [(runtime <HH:MM:SS>)]
    created <- as.POSIXct(NA_character_)
    runtime <- NA_character_

    if (i + 1L <= length(lines)) {
      line2 <- lines[i + 1L]
      m2    <- regexec(
        "^\\s+\\S+\\s+(\\d{4}-\\d{2}-\\d{2} \\d{2}:\\d{2}:\\d{2})(?:\\s+\\(runtime (\\S+)\\))?",
        line2, perl = TRUE
      )
      cap2 <- regmatches(line2, m2)[[1]]
      if (length(cap2) >= 2) {
        created <- as.POSIXct(cap2[2], format = "%Y-%m-%d %H:%M:%S")
        if (length(cap2) >= 3 && nzchar(cap2[3])) runtime <- cap2[3]
      }
      i <- i + 2L
    } else {
      i <- i + 1L
    }

    rows[[length(rows) + 1L]] <- data.frame(
      job_id  = job_id,
      name    = name,
      state   = state,
      created = created,
      runtime = runtime,
      stringsAsFactors = FALSE
    )
  }

  if (length(rows) == 0) return(empty)
  do.call(rbind, rows)
}

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @import data.table
#' @importFrom stats confint anova relevel setNames rnorm runif
#' @importFrom utils object.size
## usethis namespace: end

utils::globalVariables(c(
  # data.table .() — function alias for list(), not exported by data.table
  ".",
  # data.table column references — assoc_*
  ".ukb_event", ".ukb_trend_score", ".fg_status",
  "HR", "CI_lower", "CI_upper", "OR", "SHR",
  "model", "term", "level", "p_interaction", "i.p_interaction",
  "subgroup_level", "fgwt",
  # data.table column references — derive_*
  "eid", "effect_allele", "..required",
  "array_idx", "text", "year_month", "baseline_date", "parsed_date",
  "instance_date", "i.date",
  "icd_code", "raw_clean", "col_name", "diag_date",
  "hist_code", "behv_code",
  "death_date",
  # data.table column references — ops_*
  "pct_na"
))

NULL

# =============================================================================
# utils_ops.R — Internal helpers for ops_ series
# =============================================================================


# ── ops_toy helpers ───────────────────────────────────────────────────────────

#' Generate toy cohort data (UKB-like phenotype table)
#'
#' @param n (integer) Number of participants.
#' @return A data.table with UKB-style column names.
#' @keywords internal
#' @noRd
.ops_toy_cohort <- function(n) {

  # ── 1. Identity ──────────────────────────────────────────────────────────────
  eid <- seq(10000001L, by = 1L, length.out = n)

  # ── 2. Demographics ──────────────────────────────────────────────────────────
  p31 <- sample(c("Female", "Male"), n, replace = TRUE, prob = c(0.54, 0.46))
  p34 <- sample(1930L:1980L, n, replace = TRUE)

  baseline_days <- as.integer(as.Date("2010-12-31") - as.Date("2006-01-01"))
  p53_i0 <- format(
    as.Date("2006-01-01") + sample(0L:baseline_days, n, replace = TRUE)
  )
  p21022 <- sample(30L:80L, n, replace = TRUE)

  # ── 3. Anthropometrics & Lifestyle ───────────────────────────────────────────
  # BMI: similar distribution but not identical to real UKB
  p21001_i0 <- round(pmax(12, pmin(65, rnorm(n, mean = 26.2, sd = 5.5))), 4L)

  # Smoking: label-valued, proportions vary with seed
  p20116_i0 <- sample(
    c("Never", "Previous", "Current", "Prefer not to answer"),
    n, replace = TRUE, prob = c(0.52, 0.31, 0.14, 0.03)
  )

  # Alcohol: label-valued
  p1558_i0 <- sample(
    c("Daily or almost daily", "Three or four times a week",
      "Once or twice a week",  "One to three times a month",
      "Special occasions only", "Never", "Prefer not to answer"),
    n, replace = TRUE, prob = c(0.08, 0.21, 0.28, 0.17, 0.14, 0.09, 0.03)
  )

  # Ethnicity: simplified broad categories
  p21000_i0 <- sample(
    c("White", "Asian", "Black", "Mixed", "Other"),
    n, replace = TRUE, prob = c(0.87, 0.06, 0.02, 0.02, 0.03)
  )

  # Townsend deprivation index: approx N(-1.3, 3.2), clipped
  p22189 <- round(pmax(-7, pmin(12, rnorm(n, mean = -1.3, sd = 3.2))), 2L)

  # Assessment centre: subset of common UK sites
  p54_i0 <- sample(
    c("Leeds", "Manchester", "Edinburgh", "Bristol", "Birmingham",
      "Oxford", "Newcastle", "Nottingham", "Liverpool", "Sheffield"),
    n, replace = TRUE
  )

  # ── 4. Genetic PCs (p22009_a1 ~ p22009_a10) ──────────────────────────────────
  pc_mat  <- matrix(round(rnorm(n * 10L), 6L), nrow = n)
  pc_cols <- stats::setNames(
    as.data.frame(pc_mat), paste0("p22009_a", 1:10)
  )

  # ── 5. Self-report disease (i0, a0~a4) ───────────────────────────────────────
  # UKB coding 6: non-cancer illness text labels (as exported from RAP)
  sr_codes <- c(
    "hypertension", "type 2 diabetes", "asthma", "back problem",
    "thyroid problem (not cancer)", "fracture", "joint disorder",
    "heart attack/myocardial infarction"
  )

  # Decreasing fill rate across array slots (a0 most populated)
  sr_fill <- c(0.30, 0.18, 0.10, 0.05, 0.02)

  sr_disease <- stats::setNames(
    lapply(sr_fill, function(p)
      ifelse(runif(n) < p, sample(sr_codes, n, replace = TRUE), NA_character_)
    ),
    paste0("p20002_i0_a", 0:4)
  )

  # Date: decimal year format e.g. 2005.5 (mid-year), NA-aligned with disease
  # Reason: keep most self-report dates post-2000 so incident cases dominate
  sr_dates <- stats::setNames(
    lapply(seq_along(sr_fill), function(i) {
      has_val <- !is.na(sr_disease[[i]])
      ifelse(has_val, sample(2000L:2015L, n, replace = TRUE) + 0.5, NA_real_)
    }),
    paste0("p20008_i0_a", 0:4)
  )

  # ── 5b. Self-report cancer (p20001 i0 a0~a4, p20006 i0 a0~a4) ───────────────
  # UKB coding 3: cancer illness text labels (as exported from RAP)
  sc_codes <- c(
    "lung cancer", "breast cancer", "bladder cancer",
    "malignant melanoma", "non-melanoma skin cancer",
    "lymphoma", "thyroid cancer", "kidney/renal cell cancer"
  )

  # Lower fill rate than non-cancer; decreasing across array slots
  sc_fill <- c(0.05, 0.03, 0.015, 0.008, 0.003)

  sc_disease <- stats::setNames(
    lapply(sc_fill, function(p)
      ifelse(runif(n) < p, sample(sc_codes, n, replace = TRUE), NA_character_)
    ),
    paste0("p20001_i0_a", 0:4)
  )

  # Decimal year format, NA-aligned with disease — same convention as p20008
  sc_dates <- stats::setNames(
    lapply(seq_along(sc_fill), function(i) {
      has_val <- !is.na(sc_disease[[i]])
      ifelse(has_val, sample(2000L:2015L, n, replace = TRUE) + 0.5, NA_real_)
    }),
    paste0("p20006_i0_a", 0:4)
  )

  # ── 6. HES (p41270 JSON + p41280_a0~a9) ──────────────────────────────────────
  icd10_hes     <- c("E11", "I10", "J45", "L20", "C44", "I25",
                     "N18", "F32", "K57", "M79", "I48", "G35")
  hes_has_rec   <- runif(n) < 0.35
  hes_date_span <- as.integer(as.Date("2022-12-31") - as.Date("2000-01-01"))

  # Reason: build p41270 and p41280 together so code count and dates are
  # index-aligned — every code at position k gets a date in p41280_ak.
  hes_records <- lapply(seq_len(n), function(i) {
    if (!hes_has_rec[i]) return(list(json = NA_character_, dates = rep(NA_character_, 9L)))
    k     <- sample(1:4, 1)
    codes <- sample(icd10_hes, k)
    dates <- format(as.Date("2000-01-01") + sample(0L:hes_date_span, k, replace = TRUE))
    list(
      json  = paste0('["', paste(codes, collapse = '","'), '"]'),
      dates = c(dates, rep(NA_character_, 9L - k))   # pad to length 9
    )
  })

  p41270 <- vapply(hes_records, `[[`, character(1), "json")

  p41280 <- stats::setNames(
    lapply(0:8, function(i)
      vapply(hes_records, function(r) r$dates[i + 1L], character(1))
    ),
    paste0("p41280_a", 0:8)
  )

  # ── 7. Cancer registry (i0~i2, 3 instances) ──────────────────────────────────
  icd10_cancer <- c("C44", "C50", "C34", "C61", "C18", "C43", "C20", "C64")
  hist_codes   <- c(8090L, 8140L, 8500L, 8070L, 8010L, 8743L, 8130L, 8520L, 8000L, 8720L)
  behv_codes   <- c(3L, 3L, 3L, 3L, 3L, 3L, 3L, 2L, 2L, 1L, 0L, 6L, 5L, 9L)  # 3=malignant ~85%
  cancer_span  <- as.integer(as.Date("2020-12-31") - as.Date("1990-01-01"))

  # Decreasing prevalence across instances
  cancer_prob <- c(0.05, 0.02, 0.008)

  cancer_cols <- lapply(seq_along(cancer_prob), function(i) {
    flag <- runif(n) < cancer_prob[i]
    list(
      icd  = ifelse(flag, sample(icd10_cancer, n, replace = TRUE), NA_character_),
      hist = ifelse(flag, sample(hist_codes,   n, replace = TRUE), NA_integer_),
      behv = ifelse(flag, sample(behv_codes,   n, replace = TRUE), NA_integer_),
      date = ifelse(flag,
        format(as.Date("1990-01-01") + sample(0L:cancer_span, n, replace = TRUE)),
        NA_character_)
    )
  })

  # ── 8. Death registry (i0, 3 contributory causes) ────────────────────────────
  icd10_death <- c("I21.9", "C34.9", "I25.9", "C50.9", "C61",
                   "I64",   "C25.9", "I48.0", "C18.9", "E11.9")
  death_span  <- as.integer(as.Date("2023-12-31") - as.Date("2011-01-01"))
  death_flag  <- runif(n) < 0.10

  p40001_i0    <- ifelse(death_flag, sample(icd10_death, n, replace = TRUE), NA_character_)
  p40002_i0_a0 <- ifelse(death_flag & runif(n) < 0.30, sample(icd10_death, n, replace = TRUE), NA_character_)
  p40002_i0_a1 <- ifelse(death_flag & runif(n) < 0.15, sample(icd10_death, n, replace = TRUE), NA_character_)
  p40002_i0_a2 <- ifelse(death_flag & runif(n) < 0.05, sample(icd10_death, n, replace = TRUE), NA_character_)
  p40000_i0 <- ifelse(
    death_flag,
    format(as.Date("2011-01-01") + sample(0L:death_span, n, replace = TRUE)),
    NA_character_
  )

  # ── 9. First occurrence (p131742) ────────────────────────────────────────────
  # Reason: representative first-occurrence date field used by derive_first_occurrence
  fo_span <- as.integer(as.Date("2022-12-31") - as.Date("1995-01-01"))
  p131742 <- ifelse(
    runif(n) < 0.08,
    format(as.Date("1995-01-01") + sample(0L:fo_span, n, replace = TRUE)),
    NA_character_
  )

  # ── 10. GRS columns (raw, unstandardised — for testing grs_standardize()) ────
  grs_bmi     <- round(rnorm(n, mean = 0.82, sd = 2.5), 6L)
  grs_raw     <- round(rnorm(n, mean = 1.54, sd = 3.5), 6L)
  grs_finngen <- round(rnorm(n, mean = 0.41, sd = 1.8), 6L)

  # ── 11. Messy columns (robustness testing) ────────────────────────────────────
  # Reason: stress-test derive_missing and decode_ against real data quality issues
  messy_allna <- rep(NA_character_, n)
  messy_empty <- ifelse(runif(n) < 0.5, "", NA_character_)
  messy_label <- sample(
    c("#N/A", "N/A", "-1", "999", ".", "unknown", "NULL", NA_character_),
    n, replace = TRUE, prob = c(0.05, 0.05, 0.08, 0.08, 0.05, 0.05, 0.04, 0.60)
  )

  # ── Assemble ──────────────────────────────────────────────────────────────────
  dt <- data.table::data.table(
    eid       = eid,
    p31       = p31,       p34       = p34,
    p53_i0    = p53_i0,    p21022    = p21022,
    p21001_i0 = p21001_i0, p20116_i0 = p20116_i0,
    p1558_i0  = p1558_i0,  p21000_i0 = p21000_i0,
    p22189    = p22189,    p54_i0    = p54_i0
  )

  dt <- cbind(dt, data.table::as.data.table(pc_cols))
  dt <- cbind(dt, data.table::as.data.table(sr_disease), data.table::as.data.table(sr_dates))
  dt <- cbind(dt, data.table::as.data.table(sc_disease), data.table::as.data.table(sc_dates))

  dt[, p41270 := p41270]
  dt <- cbind(dt, data.table::as.data.table(p41280))

  # Cancer: 3 instances × 4 fields (icd, hist, behv, date)
  for (i in seq_along(cancer_cols)) {
    idx <- i - 1L
    dt[, (paste0("p40006_i", idx)) := cancer_cols[[i]]$icd ]
    dt[, (paste0("p40011_i", idx)) := cancer_cols[[i]]$hist]
    dt[, (paste0("p40012_i", idx)) := cancer_cols[[i]]$behv]
    dt[, (paste0("p40005_i", idx)) := cancer_cols[[i]]$date]
  }

  dt[, `:=`(
    p40001_i0    = p40001_i0,
    p40002_i0_a0 = p40002_i0_a0,
    p40002_i0_a1 = p40002_i0_a1,
    p40002_i0_a2 = p40002_i0_a2,
    p40000_i0    = p40000_i0
  )]

  dt[, p131742 := p131742]

  dt[, `:=`(
    grs_bmi     = grs_bmi,
    grs_raw     = grs_raw,
    grs_finngen = grs_finngen
  )]

  dt[, `:=`(
    messy_allna = messy_allna,
    messy_empty = messy_empty,
    messy_label = messy_label
  )]

  dt
}


#' Generate analysis-ready toy dataset for assoc_* functions
#'
#' Returns a pre-derived participant-level table with covariates already
#' converted to factors, BMI/TDI binned, and two disease outcomes (DM and HTN)
#' with internally consistent status / date / timing / followup columns.
#'
#' @param n (integer) Number of participants.
#' @return A data.table ready for \code{assoc_*} functions.
#' @keywords internal
#' @noRd
.ops_toy_association <- function(n) {

  censor_date  <- as.Date("2022-10-31")
  disease_span <- as.integer(censor_date - as.Date("1990-01-01"))

  # ── 1. Identity ──────────────────────────────────────────────────────────────
  eid <- seq(10000001L, by = 1L, length.out = n)

  # ── 2. Demographics ──────────────────────────────────────────────────────────
  p31 <- factor(
    sample(c("Female", "Male"), n, replace = TRUE, prob = c(0.54, 0.46)),
    levels = c("Female", "Male")
  )

  baseline_days <- as.integer(as.Date("2010-12-31") - as.Date("2006-01-01"))
  p53_i0 <- data.table::as.IDate(
    as.Date("2006-01-01") + sample(0L:baseline_days, n, replace = TRUE)
  )

  p21022 <- sample(40L:70L, n, replace = TRUE)

  # ── 3. Covariates ─────────────────────────────────────────────────────────────
  # BMI + derived category (same breaks as derive_cut examples)
  p21001_i0 <- round(pmax(12, pmin(65, rnorm(n, mean = 26.2, sd = 5.5))), 2L)
  bmi_cat <- cut(p21001_i0,
    breaks = c(-Inf, 18.5, 25, 30, Inf),
    labels = c("Underweight", "Normal", "Overweight", "Obese"),
    right  = FALSE
  )

  # Smoking (~3% NA for realism, no "Prefer not to answer" in analysis-ready data)
  p20116_i0 <- factor(
    sample(c("Never", "Previous", "Current", NA_character_),
           n, replace = TRUE, prob = c(0.52, 0.31, 0.14, 0.03)),
    levels = c("Never", "Previous", "Current")
  )

  # Alcohol
  p1558_i0 <- factor(sample(
    c("Daily or almost daily", "Three or four times a week",
      "Once or twice a week",  "One to three times a month",
      "Special occasions only", "Never", NA_character_),
    n, replace = TRUE, prob = c(0.08, 0.21, 0.28, 0.17, 0.14, 0.09, 0.03)
  ))

  # Ethnicity
  p21000_i0 <- factor(sample(
    c("White", "Asian", "Black", "Mixed", "Other"),
    n, replace = TRUE, prob = c(0.87, 0.06, 0.02, 0.02, 0.03)
  ))

  # TDI + derived quartile category (tdi_cat derived from p22189 in this dataset)
  p22189  <- round(pmax(-7, pmin(12, rnorm(n, mean = -1.3, sd = 3.2))), 2L)
  tdi_q   <- stats::quantile(p22189, probs = c(0.25, 0.5, 0.75))
  tdi_cat <- cut(p22189,
    breaks = c(-Inf, tdi_q[[1L]], tdi_q[[2L]], tdi_q[[3L]], Inf),
    labels = c("Q1 (least deprived)", "Q2", "Q3", "Q4 (most deprived)"),
    right  = TRUE
  )

  # Assessment centre
  p54_i0 <- factor(sample(
    c("Leeds", "Manchester", "Edinburgh", "Bristol", "Birmingham",
      "Oxford", "Newcastle", "Nottingham", "Liverpool", "Sheffield"),
    n, replace = TRUE
  ))

  # ── 4. Genetic PCs ────────────────────────────────────────────────────────────
  pc_mat  <- matrix(round(rnorm(n * 10L), 6L), nrow = n)
  pc_cols <- stats::setNames(
    as.data.frame(pc_mat), paste0("p22009_a", 1:10)
  )

  # ── 5. GRS (continuous exposure example) ─────────────────────────────────────
  grs_bmi <- round(rnorm(n, mean = 0.82, sd = 2.5), 6L)

  # ── 6. Disease outcomes ───────────────────────────────────────────────────────
  # Helper: generate one disease outcome with internally consistent columns.
  # Reason: status / date / timing / followup must be derived sequentially —
  # they cannot be generated independently without breaking their relationships.
  .make_outcome <- function(pct_case) {

    has_disease <- runif(n) < pct_case

    # Disease date: only for cases, uniformly between 1990-01-01 and censor_date
    d_date <- rep(as.Date(NA), n)
    d_date[has_disease] <- as.Date("1990-01-01") +
      sample(0L:disease_span, sum(has_disease), replace = TRUE)
    d_date <- data.table::as.IDate(d_date)

    # Timing: compare disease date to individual baseline
    timing <- integer(n)
    timing[!has_disease]                                     <- 0L
    timing[has_disease & !is.na(d_date) & d_date <= p53_i0] <- 1L  # prevalent
    timing[has_disease & !is.na(d_date) & d_date >  p53_i0] <- 2L  # incident
    timing[has_disease &  is.na(d_date)]                     <- NA_integer_

    # Follow-up end: incident → disease date; all others → censor_date
    followup_end <- rep(data.table::as.IDate(censor_date), n)
    incident     <- !is.na(timing) & timing == 2L
    followup_end[incident] <- d_date[incident]

    # Follow-up years: (followup_end - baseline) / 365.25; NA for prevalent
    followup_years <- as.numeric(followup_end - p53_i0) / 365.25
    followup_years[!is.na(timing) & timing == 1L] <- NA_real_

    list(
      status         = has_disease,
      date           = d_date,
      timing         = timing,
      followup_end   = followup_end,
      followup_years = round(followup_years, 4L)
    )
  }

  dm  <- .make_outcome(0.12)   # ~12% type 2 diabetes
  htn <- .make_outcome(0.28)   # ~28% hypertension

  # ── Assemble ──────────────────────────────────────────────────────────────────
  dt <- data.table::data.table(
    eid       = eid,
    p31       = p31,
    p53_i0    = p53_i0,
    p21022    = p21022,
    p21001_i0 = p21001_i0,
    bmi_cat   = bmi_cat,
    p20116_i0 = p20116_i0,
    p1558_i0  = p1558_i0,
    p21000_i0 = p21000_i0,
    p22189    = p22189,
    tdi_cat   = tdi_cat,
    p54_i0    = p54_i0
  )

  dt <- cbind(dt, data.table::as.data.table(pc_cols))

  dt[, grs_bmi := grs_bmi]

  dt[, `:=`(
    dm_status         = dm$status,
    dm_date           = dm$date,
    dm_timing         = dm$timing,
    dm_followup_end   = dm$followup_end,
    dm_followup_years = dm$followup_years
  )]

  dt[, `:=`(
    htn_status         = htn$status,
    htn_date           = htn$date,
    htn_timing         = htn$timing,
    htn_followup_end   = htn$followup_end,
    htn_followup_years = htn$followup_years
  )]

  dt
}


#' Generate toy forest plot data (assoc_coxph-style results table)
#'
#' @param n (integer) Number of exposures. Default 8L.
#' @return A data.table matching the output structure of assoc_coxph().
#' @keywords internal
#' @noRd
.ops_toy_forest <- function(n = 8L) {

  exposures <- c(
    "bmi", "smoking_ex", "smoking_current", "alcohol_freq",
    "townsend_deprivation", "age_at_recruitment", "ethnicity_asian",
    "ethnicity_black", "physical_activity", "sleep_duration"
  )[seq_len(min(n, 10L))]

  models <- c("Unadjusted", "Age and sex adjusted", "Fully adjusted")

  # Generate one row per exposure × model
  rows <- lapply(exposures, function(exp) {
    # Each exposure gets a "true" log-HR drawn from a realistic range
    log_hr_true <- rnorm(1, mean = 0.05, sd = 0.25)

    lapply(models, function(mod) {
      # SE shrinks with adjustment (wider CI for unadjusted)
      se <- switch(mod,
        "Unadjusted"           = runif(1, 0.03, 0.10),
        "Age and sex adjusted" = runif(1, 0.03, 0.09),
        "Fully adjusted"       = runif(1, 0.03, 0.08)
      )
      hr       <- exp(log_hr_true + rnorm(1, 0, se * 0.2))
      ci_lower <- exp(log(hr) - 1.96 * se)
      ci_upper <- exp(log(hr) + 1.96 * se)
      z        <- log(hr) / se
      p_val    <- 2 * stats::pnorm(-abs(z))

      list(
        exposure     = exp,
        term         = exp,
        model        = mod,
        n            = sample(40000L:500000L, 1L),
        n_events     = sample(500L:20000L, 1L),
        person_years = round(runif(1, 3e5, 7e6), 0),
        HR           = round(hr, 6L),
        CI_lower     = round(ci_lower, 6L),
        CI_upper     = round(ci_upper, 6L),
        p_value      = round(p_val, 6L),
        HR_label     = sprintf("%.2f (%.2f-%.2f)", hr, ci_lower, ci_upper)
      )
    })
  })

  dt <- data.table::rbindlist(unlist(rows, recursive = FALSE))

  # model as ordered factor — matches assoc_coxph output
  dt[, model := factor(model,
    levels  = c("Unadjusted", "Age and sex adjusted", "Fully adjusted"),
    ordered = TRUE
  )]

  dt
}


#' Check dx CLI installation
#'
#' @return Named list: ok, path, version.
#' @keywords internal
#' @noRd
.ops_check_dx <- function() {
  path <- Sys.which("dx")
  if (!nzchar(path)) {
    return(list(ok = FALSE, path = NA_character_, version = NA_character_))
  }

  ver_res <- processx::run(
    command          = unname(path),
    args             = "--version",
    error_on_status  = FALSE
  )
  version <- trimws(ver_res$stdout)
  if (!nzchar(version)) version <- trimws(ver_res$stderr)

  list(ok = TRUE, path = unname(path), version = version)
}


#' Check RAP authentication status
#'
#' @return Named list: ok, logged_in, user, project.
#' @keywords internal
#' @noRd
.ops_check_auth <- function() {
  # Guard: if dx is not on PATH, skip silently
  if (!nzchar(Sys.which("dx"))) {
    return(list(ok = FALSE, logged_in = FALSE, user = NA_character_, project = NA_character_))
  }

  whoami <- .dx_run(c("whoami"))
  if (!whoami$success) {
    return(list(ok = FALSE, logged_in = FALSE, user = NA_character_, project = NA_character_))
  }

  project <- .dx_get_project_id()
  list(
    ok        = TRUE,
    logged_in = TRUE,
    user      = whoami$stdout,
    project   = project
  )
}


#' Check R package dependencies
#'
#' Returns a list of check results for all Imports and key Suggests.
#'
#' @return A list of named lists, each with: package, required, group,
#'   installed, version.
#' @keywords internal
#' @noRd
.ops_check_deps <- function() {
  # Each entry: package name, required (TRUE = Imports), module group label
  deps <- list(
    # Core
    list(pkg = "cli",          required = TRUE,  group = "core"),
    list(pkg = "data.table",   required = TRUE,  group = "core"),
    list(pkg = "processx",     required = TRUE,  group = "core"),
    list(pkg = "rlang",        required = TRUE,  group = "core"),
    list(pkg = "tools",        required = TRUE,  group = "core"),
    # Extract / Fetch
    list(pkg = "curl",         required = TRUE,  group = "extract / fetch"),
    list(pkg = "jsonlite",     required = TRUE,  group = "extract / fetch"),
    # Analysis
    list(pkg = "survival",     required = TRUE,  group = "assoc_coxph"),
    list(pkg = "dplyr",        required = TRUE,  group = "assoc / derive"),
    list(pkg = "tidyselect",   required = TRUE,  group = "assoc / derive"),
    # Visualisation
    list(pkg = "forestploter", required = TRUE,  group = "plot_forest"),
    list(pkg = "broom",        required = TRUE,  group = "plot_tableone"),
    list(pkg = "gt",           required = TRUE,  group = "plot_tableone"),
    list(pkg = "gtsummary",    required = TRUE,  group = "plot_tableone"),
    # Optional / Suggests
    list(pkg = "pROC",         required = FALSE, group = "assoc (optional)"),
    list(pkg = "knitr",        required = FALSE, group = "vignettes"),
    list(pkg = "rmarkdown",    required = FALSE, group = "vignettes")
  )

  lapply(deps, function(d) {
    inst    <- requireNamespace(d$pkg, quietly = TRUE)
    version <- if (inst) as.character(utils::packageVersion(d$pkg)) else NA_character_
    list(
      package   = d$pkg,
      required  = d$required,
      group     = d$group,
      installed = inst,
      version   = version
    )
  })
}

# =============================================================================
# decode.R - UKB column name and value decoding
# =============================================================================


#' Rename UKB field ID columns to human-readable snake_case names
#'
#' Renames columns from the raw UKB field ID format used by
#' \code{\link{extract_pheno}} (e.g. \code{participant.p31}) and
#' \code{\link{job_result}} (e.g. \code{p53_i0}) to human-readable
#' snake_case identifiers (e.g. \code{sex},
#' \code{date_of_attending_assessment_centre_i0}).
#'
#' Column labels are taken from the UKB field title dictionary via
#' \code{\link{extract_ls}}. Both the dataset name and field list are cached
#' after the first call, so subsequent calls to \code{decode_names()} involve
#' no network requests.
#'
#' When an auto-generated name exceeds \code{max_nchar} characters it is
#' flagged with a warning so you can decide whether to shorten it manually
#' with \code{names(data)[...] <- ...}. The function never truncates names
#' automatically, because the right short name depends on scientific context
#' that only you know.
#'
#' @param data (data.frame or data.table) Data extracted from UKB-RAP via
#'   \code{extract_pheno()} or \code{job_result()}.
#' @param max_nchar (integer) Column names longer than this value are flagged.
#'   Default: \code{60}.
#'
#' @return The input \code{data} with column names replaced by snake_case
#'   labels. Returns a \code{data.table} if the input is a \code{data.table}.
#' @export
#'
#' @examples
#' \dontrun{
#' df <- extract_pheno(c(31, 53, 21022))
#' df <- decode_names(df)
#' # participant.eid    → eid
#' # participant.p31    → sex
#' # participant.p21022 → age_at_recruitment
#' # participant.p53_i0 → date_of_attending_assessment_centre_i0  (warned if > 60)
#'
#' # Shorten a long name afterwards
#' names(df)[names(df) == "date_of_attending_assessment_centre_i0"] <- "date_baseline"
#' }
decode_names <- function(data, max_nchar = 60L) {

  if (!is.data.frame(data)) {
    cli::cli_abort("{.arg data} must be a data.frame or data.table.", call = NULL)
  }
  max_nchar <- .assert_count(max_nchar)

  col_names <- names(data)

  # Reason: delegate entirely to extract_ls() — both the dataset name and the
  # field list are session-cached, so this is a pure memory lookup on warm paths
  fields_df <- suppressMessages(extract_ls())

  # Build old → new name vector (positionally aligned with col_names)
  new_names <- .build_name_map(col_names, fields_df)

  # Apply renaming; make.unique() guards against duplicate titles in the
  # UKB field dictionary producing identical snake_case names
  names(data) <- make.unique(new_names, sep = "_")

  n_renamed <- sum(new_names != col_names, na.rm = TRUE)
  cli::cli_alert_success("Renamed {n_renamed} column{?s}.")

  # Flag names that exceed the character limit
  long_cols <- new_names[nchar(new_names) > max_nchar]
  if (length(long_cols) > 0) {
    cli::cli_alert_warning(
      "{length(long_cols)} column name{?s} longer than {max_nchar} characters - consider renaming manually:"
    )
    cli::cli_ul(long_cols)
  }

  data
}


#' Decode UKB categorical column values using Showcase metadata
#'
#' Converts raw integer codes produced by \code{\link{extract_pheno}} into
#' human-readable labels for all categorical fields
#' (\code{value_type} 21 and 22), using the UKB Showcase encoding tables.
#' Continuous, text, date, and already-decoded columns are left unchanged.
#'
#' This function requires two metadata files downloaded from the UKB Research
#' Analysis Platform:
#' \itemize{
#'   \item \code{field.tsv} - maps field IDs to encoding IDs and value types.
#'   \item \code{esimpint.tsv} - maps encoding ID + integer code to label.
#' }
#' Download them once with:
#' \preformatted{
#' fetch_metadata(dest_dir = "data/metadata")
#' }
#' Both files are cached in the session after the first read.
#'
#' \strong{Call order}: use \code{decode_values()} \emph{before}
#' \code{\link{decode_names}}, so that column names still contain the numeric
#' field ID needed to look up the encoding.
#'
#' @param data (data.frame or data.table) Data from \code{extract_pheno()},
#'   with column names in \code{participant.pXXXX} or \code{pXXXX_iX} format.
#' @param metadata_dir (character) Directory containing \code{field.tsv} and
#'   \code{esimpint.tsv}. Default: \code{"data/metadata/"}.
#'
#' @return The input \code{data} with categorical columns replaced by character
#'   labels. Returns a \code{data.table} if the input is a \code{data.table}.
#' @export
#'
#' @examples
#' \dontrun{
#' # Download metadata once
#' fetch_metadata(dest_dir = "data/metadata")
#'
#' # Recommended call order
#' df <- extract_pheno(c(31, 54, 20116, 21000))
#' df <- decode_values(df)                  # 0/1 → "Female"/"Male", etc.
#' df <- decode_names(df)                   # participant.p31 → sex
#' }
decode_values <- function(data, metadata_dir = "data/metadata/") {

  if (!is.data.frame(data)) {
    cli::cli_abort("{.arg data} must be a data.frame or data.table.", call = NULL)
  }
  .assert_scalar_string(metadata_dir)

  col_names <- names(data)

  # Auto-detect UKB field columns: must still contain numeric field ID
  target_cols <- col_names[grepl("(?<![a-z])p\\d+", col_names, perl = TRUE)]

  if (length(target_cols) == 0) {
    cli::cli_alert_warning(
      "No UKB field ID columns detected. Call {.fn decode_values} before {.fn decode_names}."
    )
    return(invisible(data))
  }

  # Load metadata (hits cache after first call)
  field_meta <- .load_field_meta(metadata_dir)
  esimpint   <- .load_esimpint(metadata_dir)

  # Subset to categorical fields with a real encoding
  cat_fields <- field_meta[
    field_meta$value_type %in% c(21L, 22L) & field_meta$encoding_id > 0L, ,
    drop = FALSE
  ]

  n_decoded <- 0L
  n_skipped <- 0L

  for (col in target_cols) {

    # Skip columns already decoded to character (e.g. job_result() output)
    if (is.character(data[[col]])) { n_skipped <- n_skipped + 1L; next }

    fid <- .extract_field_id(col)
    if (is.na(fid)) { n_skipped <- n_skipped + 1L; next }

    enc_row <- cat_fields[cat_fields$field_id == fid, , drop = FALSE]
    if (nrow(enc_row) == 0) { n_skipped <- n_skipped + 1L; next }

    enc_id  <- enc_row$encoding_id[1L]
    mapping <- esimpint[esimpint$encoding_id == enc_id, , drop = FALSE]
    if (nrow(mapping) == 0) { n_skipped <- n_skipped + 1L; next }

    # Reason: match() returns NA for any code absent from the encoding table,
    # which naturally covers UKB missing codes (-1, -3, -7, -13, -818)
    # when they are not listed as valid entries in that encoding
    data[[col]] <- mapping$meaning[match(data[[col]], mapping$value)]
    n_decoded   <- n_decoded + 1L
  }

  cli::cli_alert_success(
    "Decoded {n_decoded} categorical column{?s}; {n_skipped} non-categorical column{?s} unchanged."
  )

  data
}

# =============================================================================
# fetch.R — Remote RAP file system exploration and data retrieval
# =============================================================================


#' List files and folders at a remote RAP path
#'
#' Returns a structured data.frame describing the contents of a remote
#' DNAnexus Research Analysis Platform (RAP) directory. Analogous to
#' \code{file_info()} but for remote project storage.
#'
#' @param path (character) Remote path to list. Default: \code{"."} (project
#'   root). Both \code{"Bulk/"} and \code{"/Bulk/"} are accepted.
#' @param type (character) Filter results by entry type: \code{"all"}
#'   (default), \code{"file"}, or \code{"folder"}.
#' @param pattern (character) Optional regex to filter by name, e.g.
#'   \code{"\\.bed$"}. Default: \code{NULL}.
#'
#' @return A data.frame with columns:
#'   \describe{
#'     \item{name}{Entry name (no trailing slash for folders).}
#'     \item{type}{\code{"file"} or \code{"folder"}.}
#'     \item{size}{File size string (e.g. \code{"120.94 GB"}), \code{NA} for
#'       folders or non-file objects.}
#'     \item{modified}{Last modified time (\code{POSIXct}), \code{NA} for
#'       folders.}
#'   }
#' @export
#'
#' @examples
#' \dontrun{
#' fetch_ls()
#' fetch_ls("Showcase metadata/", type = "file")
#' fetch_ls("results/", pattern = "\\.csv$")
#' }
fetch_ls <- function(path = ".", type = "all", pattern = NULL) {
  type <- match.arg(type, c("all", "file", "folder"))

  result <- .dx_ls_raw(path)
  if (!result$success) {
    stop("Failed to list '", path, "': ", result$stderr, call. = FALSE)
  }

  df <- .dx_ls_parse(result$stdout)

  if (type != "all") {
    df <- df[df$type == type, , drop = FALSE]
  }

  if (!is.null(pattern)) {
    df <- df[grepl(pattern, df$name, perl = TRUE), , drop = FALSE]
  }

  df
}


#' Get pre-authenticated download URL(s) for a remote RAP file or folder
#'
#' Generates temporary HTTPS URLs for files on the DNAnexus Research Analysis
#' Platform. For a single file, returns one URL. For a folder, lists all files
#' inside and returns a named character vector of URLs.
#'
#' @param path (character) Remote file path or folder path, e.g.
#'   \code{"Showcase metadata/field.tsv"} or \code{"Showcase metadata/"}.
#' @param duration (character) How long the URLs remain valid. Accepts
#'   suffixes: \code{s}, \code{m}, \code{h}, \code{d}, \code{w}, \code{M},
#'   \code{y}. Default: \code{"1d"} (one day).
#'
#' @return A named character vector of pre-authenticated HTTPS URLs.
#'   Names are the file names.
#' @export
#'
#' @examples
#' \dontrun{
#' # Single file
#' fetch_url("Showcase metadata/field.tsv")
#'
#' # Entire folder
#' fetch_url("Showcase metadata/", duration = "7d")
#' }
fetch_url <- function(path, duration = "1d") {
  norm      <- .dx_normalize_path(path)
  probe     <- .dx_probe_path(path)
  is_folder <- probe$is_folder

  if (!is_folder) {
    url <- .dx_make_url(norm, duration = duration)
    return(stats::setNames(url, basename(norm)))
  }

  # Folder: list all files then loop
  files <- fetch_ls(path, type = "file")
  if (nrow(files) == 0) {
    cli::cli_inform("No files found in {.path {path}}.")
    return(invisible(character(0)))
  }

  urls <- vapply(
    file.path(norm, files$name, fsep = "/"),
    function(f) .dx_make_url(f, duration = duration),
    character(1)
  )
  stats::setNames(urls, files$name)
}


#' Download a file from RAP project storage
#'
#' Downloads one file or all files within a folder from RAP project storage
#' to the current directory or a specified destination within the RAP
#' environment. This function must be called from within RAP.
#'
#' @param path (character) Remote file or folder path.
#' @param dest_dir (character) Destination directory. Created automatically
#'   if it does not exist.
#' @param overwrite (logical) Overwrite existing local files. Default:
#'   \code{FALSE}.
#' @param resume (logical) Resume an interrupted download. Useful for large
#'   files (e.g. \code{.bed}, \code{.bgen}). Default: \code{FALSE}.
#' @param verbose (logical) Show download progress. Default: \code{TRUE}.
#'
#' @return Invisibly returns the local file path(s) as a character vector.
#' @export
#'
#' @examples
#' \dontrun{
#' # Download a single metadata file
#' fetch_file("Showcase metadata/field.tsv", dest_dir = "data/")
#'
#' # Download an entire folder
#' fetch_file("Showcase metadata/", dest_dir = "data/metadata/")
#'
#' # Resume an interrupted download
#' fetch_file("results/summary_stats.csv", dest_dir = "data/", resume = TRUE)
#' }
fetch_file <- function(path, dest_dir, overwrite = FALSE,
                       resume = FALSE, verbose = TRUE) {
  .assert_on_rap()

  if (!dir.exists(dest_dir)) {
    dir.create(dest_dir, recursive = TRUE)
  }

  is_folder <- .dx_probe_path(path)$is_folder

  if (!is_folder) {
    # Single file
    url      <- .dx_make_url(path)
    destfile <- file.path(dest_dir, basename(.dx_normalize_path(path)))
    .dx_download_file(url, destfile, overwrite = overwrite, resume = resume, verbose = verbose)
    return(invisible(destfile))

  } else {
    # Folder: list files, batch generate URLs, parallel download
    files <- fetch_ls(path, type = "file")
    if (nrow(files) == 0) {
      cli::cli_inform("No files found in {.path {path}}.")
      return(invisible(character(0)))
    }

    norm      <- .dx_normalize_path(path)
    urls      <- vapply(file.path(norm, files$name, fsep = "/"),
                        function(f) .dx_make_url(f), character(1))
    destfiles <- file.path(dest_dir, files$name)

    .dx_download_batch(urls, destfiles, overwrite = overwrite,
                       resume = resume, verbose = verbose)
    return(invisible(destfiles))
  }
}


#' Download the Showcase metadata folder
#'
#' Downloads the entire \code{Showcase metadata/} folder from the DNAnexus
#' Research Analysis Platform to a local directory. This includes
#' \code{field.tsv}, \code{encoding.tsv}, and all associated encoding tables.
#'
#' @param dest_dir (character) Destination directory. Created automatically
#'   if it does not exist.
#' @param overwrite (logical) Overwrite existing local files. Default:
#'   \code{FALSE}.
#' @param resume (logical) Resume interrupted downloads. Default: \code{FALSE}.
#' @param verbose (logical) Show download progress. Default: \code{TRUE}.
#'
#' @return Invisibly returns the local file paths as a character vector.
#' @export
#'
#' @examples
#' \dontrun{
#' fetch_metadata(dest_dir = "metadata")
#' fetch_metadata(dest_dir = "metadata", overwrite = TRUE)
#' }
fetch_metadata <- function(dest_dir, overwrite = FALSE,
                           resume = FALSE, verbose = TRUE) {
  fetch_file("Showcase metadata/", dest_dir = dest_dir,
             overwrite = overwrite, resume = resume, verbose = verbose)
}


#' Download the UKB field dictionary file
#'
#' Downloads \code{field.tsv} from the \code{Showcase metadata/} folder on the
#' DNAnexus Research Analysis Platform. This file contains the complete UKB
#' data dictionary: field IDs, titles, value types, and encoding references.
#'
#' @param dest_dir (character) Destination directory. Created automatically
#'   if it does not exist.
#' @param overwrite (logical) Overwrite existing local file. Default:
#'   \code{FALSE}.
#' @param resume (logical) Resume an interrupted download. Default:
#'   \code{FALSE}.
#' @param verbose (logical) Show download progress. Default: \code{TRUE}.
#'
#' @return Invisibly returns the local file path as a character string.
#' @export
#'
#' @examples
#' \dontrun{
#' fetch_field(dest_dir = "metadata")
#' fetch_field(dest_dir = "metadata", overwrite = TRUE)
#' }
fetch_field <- function(dest_dir, overwrite = FALSE,
                        resume = FALSE, verbose = TRUE) {
  fetch_file("Showcase metadata/field.tsv", dest_dir = dest_dir,
             overwrite = overwrite, resume = resume, verbose = verbose)
}


#' Print a remote RAP directory tree
#'
#' Displays the remote directory structure in a tree-like format by
#' recursively listing sub-folders up to \code{max_depth}. Analogous to
#' \code{file_tree()} but for remote project storage.
#'
#' @param path (character) Remote root path. Default: \code{"."} (project
#'   root). Both \code{"Bulk/"} and \code{"/Bulk/"} are accepted.
#' @param max_depth (integer) Maximum recursion depth. Default: \code{2}.
#' @param verbose (logical) Whether to print the tree to the console.
#'   Default: \code{TRUE}.
#'
#' @section Warning:
#' Each level of recursion triggers one HTTPS API call per folder. Deep trees
#' (e.g. \code{max_depth > 3}) on large UKB projects may issue 100+ network
#' requests, causing the console to hang for tens of seconds or time out.
#' Keep \code{max_depth} at 2-3 for interactive use.
#'
#' @return Invisibly returns a character vector of tree lines.
#' @export
#'
#' @examples
#' \dontrun{
#' fetch_tree()
#' fetch_tree("Bulk/", max_depth = 2)
#' fetch_tree(verbose = FALSE)
#' }
fetch_tree <- function(path = ".", max_depth = 2, verbose = TRUE) {
  norm  <- .dx_normalize_path(path)
  lines <- character(0)

  tree_chars <- list(branch = "+-- ", pipe = "|   ", space = "    ")

  traverse <- function(p, depth = 0, prefix = "") {
    if (depth >= max_depth) return()

    res <- .dx_ls_names(p)
    if (!res$success || !nzchar(res$stdout)) return()

    entries <- strsplit(res$stdout, "\n")[[1]]
    entries <- trimws(entries[nzchar(entries)])

    for (i in seq_along(entries)) {
      entry     <- entries[i]
      is_folder <- endsWith(entry, "/")
      name      <- sub("/$", "", entry)
      is_last   <- (i == length(entries))

      lines <<- c(lines, paste0(prefix, tree_chars$branch, entry))

      if (is_folder) {
        # Reason: build full sub-path by appending folder name to current path
        sub_path   <- paste0(p, name, "/")
        new_prefix <- paste0(prefix, if (is_last) tree_chars$space else tree_chars$pipe)
        traverse(sub_path, depth + 1, new_prefix)
      }
    }
  }

  traverse(norm)

  if (verbose) {
    display <- if (nzchar(norm)) norm else "/"
    cli::cli_inform("Remote: {.path {display}}")
    for (line in lines) cli::cli_inform("{line}")
  }

  invisible(lines)
}

# =============================================================================
# test-integration-fetch.R — Integration tests for fetch_ series
# Requires real dx-toolkit, token, and network connection
# Run manually before release: devtools::test(filter = "integration-fetch")
# =============================================================================

skip_on_cran()

# ===========================================================================
# fetch_ls()
# ===========================================================================

test_that("fetch_ls() returns a data.frame with correct columns", {
  .skip_if_no_rap()
  result <- fetch_ls()
  expect_s3_class(result, "data.frame")
  expect_named(result, c("name", "type", "size", "modified"))
})

test_that("fetch_ls() returns at least one entry at project root", {
  .skip_if_no_rap()
  result <- fetch_ls()
  expect_gt(nrow(result), 0)
})

test_that("fetch_ls() type column contains only 'file' or 'folder'", {
  .skip_if_no_rap()
  result <- fetch_ls()
  expect_true(all(result$type %in% c("file", "folder")))
})

test_that("fetch_ls() filter by type = 'folder' returns only folders", {
  .skip_if_no_rap()
  result <- fetch_ls(type = "folder")
  expect_true(all(result$type == "folder"))
})

# ===========================================================================
# fetch_tree()
# ===========================================================================

test_that("fetch_tree() returns a character vector invisibly", {
  .skip_if_no_rap()
  result <- suppressMessages(fetch_tree(max_depth = 1, verbose = FALSE))
  expect_type(result, "character")
})

test_that("fetch_tree() output lines contain entry names", {
  .skip_if_no_rap()
  result <- suppressMessages(fetch_tree(max_depth = 1, verbose = FALSE))
  expect_true(length(result) > 0)
  expect_true(any(grepl("\\+--", result)))
})

# ===========================================================================
# fetch_url()
# ===========================================================================

test_that("fetch_url() returns a named character vector for a single file", {
  .skip_if_no_rap()
  result <- fetch_url("Showcase metadata/field.tsv")
  expect_type(result, "character")
  expect_length(result, 1)
  expect_true(startsWith(result[[1]], "https://"))
  expect_equal(names(result), "field.tsv")
})

test_that("fetch_url() URL is accessible (HTTP 200)", {
  .skip_if_no_rap()
  url <- fetch_url("Showcase metadata/field.tsv")[[1]]
  response <- tryCatch(
    suppressWarnings(readLines(url, n = 1, warn = FALSE)),
    error = function(e) NULL
  )
  expect_false(is.null(response))
})

# =============================================================================
# test-fetch.R — Unit tests for fetch_ series (mock-based, no network required)
# =============================================================================
.skip_if_no_mockery()

# Helper: build a fake .dx_run() result
.fake_dx <- function(stdout = "", stderr = "", status = 0) {
  list(
    stdout  = stdout,
    stderr  = stderr,
    status  = status,
    success = status == 0
  )
}

# Fake dx ls -l output with folders and files
.fake_ls_stdout <- function() {
  paste(
    "Project: project-XXXXXXXXXXXX",
    "Folder : /",
    "",
    "Bulk/",
    "Showcase metadata/",
    "",
    "State   Last modified             Size          Name",
    "closed  2024-01-15 10:30:00    120.94 GB  ukb23158_c1_b0_v1.bed (file-AAAAAAAAAA)",
    "closed  2024-01-15 10:31:00      1.23 MB  field.tsv (file-BBBBBBBBBB)",
    sep = "\n"
  )
}

# Fake dx ls -l output with files only (no folders)
.fake_ls_files_only <- function() {
  paste(
    "Project: project-XXXXXXXXXXXX",
    "Folder : /Showcase metadata/",
    "",
    "State   Last modified             Size      Name",
    "closed  2024-01-15 10:30:00    1.23 MB  field.tsv (file-AAAAAAAAAA)",
    "closed  2024-01-15 10:31:00  800.00 KB  encoding.tsv (file-BBBBBBBBBB)",
    sep = "\n"
  )
}

# ===========================================================================
# .dx_ls_parse() — internal parser (pure function, no mocking needed)
# ===========================================================================

test_that(".dx_ls_parse() returns empty data.frame for empty stdout", {
  result <- ukbflow:::.dx_ls_parse("")
  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 0)
  expect_named(result, c("name", "type", "size", "modified"))
})

test_that(".dx_ls_parse() parses folders correctly", {
  result <- ukbflow:::.dx_ls_parse(.fake_ls_stdout())
  folders <- result[result$type == "folder", ]
  expect_equal(nrow(folders), 2)
  expect_equal(folders$name, c("Bulk", "Showcase metadata"))
  expect_true(all(is.na(folders$size)))
})

test_that(".dx_ls_parse() parses files with size and modified time", {
  result <- ukbflow:::.dx_ls_parse(.fake_ls_stdout())
  files <- result[result$type == "file", ]
  expect_equal(nrow(files), 2)
  expect_equal(files$name, c("ukb23158_c1_b0_v1.bed", "field.tsv"))
  expect_false(any(is.na(files$size)))
  expect_s3_class(files$modified, "POSIXct")
})

test_that(".dx_ls_parse() returns correct columns", {
  result <- ukbflow:::.dx_ls_parse(.fake_ls_stdout())
  expect_named(result, c("name", "type", "size", "modified"))
})

test_that(".dx_ls_parse() handles files-only output (no folders)", {
  result <- ukbflow:::.dx_ls_parse(.fake_ls_files_only())
  expect_equal(nrow(result), 2)
  expect_true(all(result$type == "file"))
})

test_that(".dx_ls_parse() returns empty data.frame when dx reports 'No data objects found'", {
  stdout <- paste(
    "Project: project-XXXXXXXXXXXX",
    "Folder : /empty/",
    "",
    "No data objects found.",
    sep = "\n"
  )
  result <- ukbflow:::.dx_ls_parse(stdout)
  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 0L)
  expect_named(result, c("name", "type", "size", "modified"))
})

# ===========================================================================
# fetch_ls()
# ===========================================================================

test_that("fetch_ls() returns a data.frame with correct columns", {
  mockery::stub(fetch_ls, ".dx_ls_raw",
                function(path, ...) .fake_dx(stdout = .fake_ls_stdout()))
  result <- fetch_ls()
  expect_s3_class(result, "data.frame")
  expect_named(result, c("name", "type", "size", "modified"))
})

test_that("fetch_ls() throws error when dx fails", {
  mockery::stub(fetch_ls, ".dx_ls_raw",
                function(path, ...) .fake_dx(stderr = "Not logged in", status = 1))
  expect_error(fetch_ls(), "Failed to list")
})

test_that("fetch_ls() filters by type = 'file'", {
  mockery::stub(fetch_ls, ".dx_ls_raw",
                function(path, ...) .fake_dx(stdout = .fake_ls_stdout()))
  result <- fetch_ls(type = "file")
  expect_true(all(result$type == "file"))
})

test_that("fetch_ls() filters by type = 'folder'", {
  mockery::stub(fetch_ls, ".dx_ls_raw",
                function(path, ...) .fake_dx(stdout = .fake_ls_stdout()))
  result <- fetch_ls(type = "folder")
  expect_true(all(result$type == "folder"))
})

test_that("fetch_ls() filters by pattern", {
  mockery::stub(fetch_ls, ".dx_ls_raw",
                function(path, ...) .fake_dx(stdout = .fake_ls_stdout()))
  result <- fetch_ls(pattern = "\\.bed$")
  expect_equal(nrow(result), 1)
  expect_equal(result$name, "ukb23158_c1_b0_v1.bed")
})

test_that("fetch_ls() returns empty data.frame when pattern matches nothing", {
  mockery::stub(fetch_ls, ".dx_ls_raw",
                function(path, ...) .fake_dx(stdout = .fake_ls_stdout()))
  result <- fetch_ls(pattern = "\\.bgen$")
  expect_equal(nrow(result), 0)
})

# ===========================================================================
# fetch_url()
# ===========================================================================

test_that("fetch_url() returns a named character vector for a single file", {
  mockery::stub(fetch_url, ".dx_probe_path",
                function(...) list(is_folder = FALSE, is_file = TRUE, exists = TRUE, stdout = ""))
  mockery::stub(fetch_url, ".dx_make_url",
                function(path, ...) "https://dl.dnanex.us/fake/field.tsv")
  result <- fetch_url("Showcase metadata/field.tsv")
  expect_type(result, "character")
  expect_length(result, 1)
  expect_true(startsWith(result, "https://"))
  expect_equal(names(result), "field.tsv")
})

test_that("fetch_url() returns named vector for a folder", {
  fake_files <- data.frame(
    name = c("field.tsv", "encoding.tsv"),
    type = "file", size = NA, modified = as.POSIXct(NA),
    stringsAsFactors = FALSE
  )
  mockery::stub(fetch_url, "fetch_ls", function(...) fake_files)
  mockery::stub(fetch_url, ".dx_make_url",
                function(path, ...) paste0("https://dl.dnanex.us/fake/", basename(path)))
  result <- fetch_url("Showcase metadata/")
  expect_length(result, 2)
  expect_equal(names(result), c("field.tsv", "encoding.tsv"))
})

test_that("fetch_url() returns character(0) for empty folder", {
  mockery::stub(fetch_url, "fetch_ls",
                function(...) data.frame(name = character(0), type = character(0),
                                         size = character(0),
                                         modified = as.POSIXct(character(0)),
                                         stringsAsFactors = FALSE))
  result <- suppressMessages(fetch_url("empty/"))
  expect_equal(result, character(0))
})

# ===========================================================================
# fetch_file()
# ===========================================================================

test_that("fetch_file() calls .dx_download_file for a single file", {
  downloaded <- character(0)
  local_mocked_bindings(.is_on_rap = function() TRUE, .package = "ukbflow")
  mockery::stub(fetch_file, ".dx_probe_path",
                function(...) list(is_folder = FALSE, is_file = TRUE, exists = TRUE, stdout = ""))
  mockery::stub(fetch_file, ".dx_make_url",
                function(path, ...) "https://dl.dnanex.us/fake/field.tsv")
  mockery::stub(fetch_file, ".dx_download_file",
                function(url, destfile, ...) { downloaded <<- destfile; invisible(destfile) })
  withr::with_tempdir({
    fetch_file("Showcase metadata/field.tsv", dest_dir = ".")
    expect_true(grepl("field.tsv", downloaded))
  })
})

test_that("fetch_file() returns character(0) for empty folder", {
  local_mocked_bindings(.is_on_rap = function() TRUE, .package = "ukbflow")
  mockery::stub(fetch_file, ".dx_probe_path",
                function(...) list(is_folder = TRUE, is_file = FALSE, exists = TRUE, stdout = ""))
  mockery::stub(fetch_file, "fetch_ls",
                function(...) data.frame(name = character(0), type = character(0),
                                         size = character(0),
                                         modified = as.POSIXct(character(0)),
                                         stringsAsFactors = FALSE))
  withr::with_tempdir({
    result <- suppressMessages(fetch_file("empty/", dest_dir = "."))
    expect_equal(result, character(0))
  })
})

test_that("fetch_file() creates dest_dir if it does not exist", {
  local_mocked_bindings(.is_on_rap = function() TRUE, .package = "ukbflow")
  mockery::stub(fetch_file, ".dx_probe_path",
                function(...) list(is_folder = FALSE, is_file = TRUE, exists = TRUE, stdout = ""))
  mockery::stub(fetch_file, ".dx_make_url",
                function(path, ...) "https://dl.dnanex.us/fake/field.tsv")
  mockery::stub(fetch_file, ".dx_download_file",
                function(url, destfile, ...) invisible(destfile))
  withr::with_tempdir({
    new_dir <- file.path(getwd(), "new_subdir")
    expect_false(dir.exists(new_dir))
    fetch_file("Showcase metadata/field.tsv", dest_dir = new_dir)
    expect_true(dir.exists(new_dir))
  })
})

# ===========================================================================
# fetch_metadata()
# ===========================================================================

test_that("fetch_metadata() calls fetch_file with 'Showcase metadata/' path", {
  called_with <- NULL
  mockery::stub(fetch_metadata, "fetch_file",
                function(path, ...) { called_with <<- path; invisible(path) })
  fetch_metadata()
  expect_equal(called_with, "Showcase metadata/")
})

test_that("fetch_metadata() passes dest_dir through to fetch_file", {
  args_received <- NULL
  mockery::stub(fetch_metadata, "fetch_file",
                function(path, dest_dir, ...) { args_received <<- dest_dir; invisible(path) })
  fetch_metadata(dest_dir = "custom/dir/")
  expect_equal(args_received, "custom/dir/")
})

# ===========================================================================
# fetch_field()
# ===========================================================================

test_that("fetch_field() calls fetch_file with 'Showcase metadata/field.tsv' path", {
  called_with <- NULL
  mockery::stub(fetch_field, "fetch_file",
                function(path, ...) { called_with <<- path; invisible(path) })
  fetch_field()
  expect_equal(called_with, "Showcase metadata/field.tsv")
})

test_that("fetch_field() passes dest_dir through to fetch_file", {
  args_received <- NULL
  mockery::stub(fetch_field, "fetch_file",
                function(path, dest_dir, ...) { args_received <<- dest_dir; invisible(path) })
  fetch_field(dest_dir = "custom/dir/")
  expect_equal(args_received, "custom/dir/")
})

# ===========================================================================
# fetch_tree()
# ===========================================================================

test_that("fetch_tree() returns invisible character vector", {
  mockery::stub(fetch_tree, ".dx_ls_names",
                function(path, ...) .fake_dx(stdout = "Bulk/\nShowcase metadata/\n"))
  result <- suppressMessages(fetch_tree(verbose = FALSE))
  expect_type(result, "character")
})

test_that("fetch_tree() respects max_depth = 0 (no traversal)", {
  call_count <- 0L
  mockery::stub(fetch_tree, ".dx_ls_names", function(path, ...) {
    call_count <<- call_count + 1L
    .fake_dx(stdout = "Bulk/\n")
  })
  suppressMessages(fetch_tree(max_depth = 0, verbose = FALSE))
  expect_equal(call_count, 0L)
})

test_that("fetch_tree() produces no output when verbose = FALSE", {
  mockery::stub(fetch_tree, ".dx_ls_names",
                function(path, ...) .fake_dx(stdout = "Bulk/\n"))
  expect_silent(fetch_tree(verbose = FALSE))
})


# ===========================================================================
# .dx_probe_path() — internal folder/file detector
# ===========================================================================

test_that(".dx_probe_path() detects folder by trailing slash without a network call", {
  # Reason: trailing slash is a client-side signal — no dx ls should fire
  result <- ukbflow:::.dx_probe_path("Bulk/")
  expect_true(result$is_folder)
  expect_false(result$is_file)
  expect_true(result$exists)
  expect_equal(result$stdout, "")
})

test_that(".dx_probe_path() returns is_folder = TRUE when dx reports Folder header", {
  local_mocked_bindings(
    .dx_ls_raw = function(...) .fake_dx(stdout = "Project: p\nFolder : /Bulk/\n"),
    .package = "ukbflow"
  )
  result <- ukbflow:::.dx_probe_path("Bulk")
  expect_true(result$is_folder)
  expect_false(result$is_file)
  expect_true(result$exists)
})

test_that(".dx_probe_path() returns is_file = TRUE for a file path", {
  # Reason: dx ls -l on a single file has no "Folder :" header line
  single_file_stdout <- paste(
    "Project: project-XXXXXXXXXXXX",
    "",
    "State   Last modified             Size      Name",
    "closed  2024-01-15 10:30:00    1.23 MB  field.tsv (file-AAAAAAAAAA)",
    sep = "\n"
  )
  local_mocked_bindings(
    .dx_ls_raw = function(...) .fake_dx(stdout = single_file_stdout),
    .package = "ukbflow"
  )
  result <- ukbflow:::.dx_probe_path("Showcase metadata/field.tsv")
  expect_false(result$is_folder)
  expect_true(result$is_file)
  expect_true(result$exists)
})

test_that(".dx_probe_path() returns exists = FALSE when dx fails", {
  local_mocked_bindings(
    .dx_ls_raw = function(...) .fake_dx(status = 1, stderr = "Not found"),
    .package = "ukbflow"
  )
  result <- ukbflow:::.dx_probe_path("nonexistent/path")
  expect_false(result$exists)
  expect_false(result$is_folder)
  expect_false(result$is_file)
})


# ===========================================================================
# .dx_ls_parse() — edge cases
# ===========================================================================

test_that(".dx_ls_parse() handles file name with spaces", {
  stdout <- paste(
    "Project: project-XXXXXXXXXXXX",
    "Folder : /",
    "",
    "State   Last modified             Size      Name",
    "closed  2024-01-15 10:30:00    1.23 MB  my report file.tsv (file-AAAAAAAAAA)",
    sep = "\n"
  )
  result <- ukbflow:::.dx_ls_parse(stdout)
  expect_equal(nrow(result), 1L)
  expect_equal(result$name, "my report file.tsv")
})

test_that(".dx_ls_parse() handles size reported in bytes", {
  stdout <- paste(
    "Project: project-XXXXXXXXXXXX",
    "Folder : /",
    "",
    "State   Last modified             Size      Name",
    "closed  2024-01-15 10:30:00    512 bytes  tiny.txt (file-AAAAAAAAAA)",
    sep = "\n"
  )
  result <- ukbflow:::.dx_ls_parse(stdout)
  expect_equal(nrow(result), 1L)
  expect_equal(result$size, "512 bytes")
})

test_that(".dx_ls_parse() handles missing size field (NA)", {
  stdout <- paste(
    "Project: project-XXXXXXXXXXXX",
    "Folder : /",
    "",
    "State   Last modified   Name",
    "closed  2024-01-15 10:30:00  nosize.txt (file-AAAAAAAAAA)",
    sep = "\n"
  )
  result <- ukbflow:::.dx_ls_parse(stdout)
  expect_equal(nrow(result), 1L)
  expect_true(is.na(result$size))
})


# ===========================================================================
# fetch_file() — RAP guard and return values
# ===========================================================================

test_that("fetch_file() aborts when not called from within the RAP", {
  local_mocked_bindings(.is_on_rap = function() FALSE, .package = "ukbflow")
  expect_error(fetch_file("Showcase metadata/field.tsv"), class = "rlang_error")
})

test_that("fetch_file() returns destfile invisibly for a single file", {
  local_mocked_bindings(.is_on_rap = function() TRUE, .package = "ukbflow")
  mockery::stub(fetch_file, ".dx_probe_path",
                function(...) list(is_folder = FALSE, is_file = TRUE, exists = TRUE, stdout = ""))
  mockery::stub(fetch_file, ".dx_make_url",
                function(path, ...) "https://dl.dnanex.us/fake/field.tsv")
  mockery::stub(fetch_file, ".dx_download_file",
                function(url, destfile, ...) invisible(destfile))
  withr::with_tempdir({
    result <- fetch_file("Showcase metadata/field.tsv", dest_dir = ".")
    expect_equal(result, file.path(".", "field.tsv"))
  })
})

test_that("fetch_file() returns destfiles invisibly for a folder", {
  fake_files <- data.frame(
    name = c("field.tsv", "encoding.tsv"),
    type = "file", size = NA, modified = as.POSIXct(NA),
    stringsAsFactors = FALSE
  )
  local_mocked_bindings(.is_on_rap = function() TRUE, .package = "ukbflow")
  mockery::stub(fetch_file, ".dx_probe_path",
                function(...) list(is_folder = TRUE, is_file = FALSE, exists = TRUE, stdout = ""))
  mockery::stub(fetch_file, "fetch_ls", function(...) fake_files)
  mockery::stub(fetch_file, ".dx_make_url",
                function(path, ...) paste0("https://dl.dnanex.us/fake/", basename(path)))
  mockery::stub(fetch_file, ".dx_download_batch",
                function(urls, destfiles, ...) invisible(destfiles))
  withr::with_tempdir({
    result <- fetch_file("Showcase metadata/", dest_dir = ".")
    expect_equal(result, file.path(".", c("field.tsv", "encoding.tsv")))
  })
})

# =============================================================================
# plot.R — publication-ready visualisation for ukbflow
# =============================================================================


#' Publication-ready forest plot
#'
#' Produces a publication-ready forest plot with UKB-standard styling.
#' The user supplies a data frame whose first column is the row label
#' (\code{item}), plus any additional display columns (e.g. \code{Cases/N}).
#' The gap column and the auto-formatted \code{OR (95\% CI)} text column are
#' inserted automatically at \code{ci_column}. Numeric p-value columns
#' declared via \code{p_cols} are formatted in-place.
#'
#' @param data data.frame. First column must be the label column (\code{item}).
#'   Additional columns are displayed as-is (character) or formatted if named
#'   in \code{p_cols} (must be numeric). Column order is preserved.
#' @param est Numeric vector. Point estimates (\code{NA} = no CI drawn).
#' @param lower Numeric vector. Lower CI bounds (same length as \code{est}).
#' @param upper Numeric vector. Upper CI bounds (same length as \code{est}).
#' @param ci_column Integer. Column position in the final rendered table where
#'   the gap/CI graphic is placed. Must be between \code{2} and
#'   \code{ncol(data) + 1} (inclusive). Default: \code{2L}.
#' @param ref_line Numeric. Reference line. Default: \code{1} (HR/OR).
#'   Use \code{0} for beta coefficients.
#' @param xlim Numeric vector of length 2. X-axis limits. \code{NULL} (default)
#'   uses \code{c(0, 2)}.
#' @param ticks_at Numeric vector. Tick positions. \code{NULL} (default) = 5
#'   evenly spaced ticks across \code{xlim}.
#' @param arrow_lab Character vector of length 2. Directional labels.
#'   Default: \code{c("Lower risk", "Higher risk")}. \code{NULL} = none.
#' @param header Character vector of length \code{ncol(data) + 2}. Column
#'   header labels for the final rendered table (original columns + gap_ci +
#'   OR label). \code{NULL} (default) = use column names from \code{data} plus
#'   \code{"gap_ci"} and \code{"OR (95\% CI)"}. Pass \code{""} for the gap
#'   column position.
#' @param indent Integer vector (length = \code{nrow(data)}). Indentation
#'   level of the label column: each unit adds two leading spaces.
#'   Default: all zeros.
#' @param bold_label Logical vector (length = \code{nrow(data)}). Which rows
#'   to bold in the label column. \code{NULL} (default) = auto-derive from
#'   \code{indent}: rows where \code{indent == 0} are bolded (parent rows),
#'   indented sub-rows are plain.
#' @param ci_col Character scalar or vector (length = \code{nrow(data)}).
#'   CI colour(s). \code{NA} rows are skipped automatically.
#'   Default: \code{"black"}.
#' @param ci_sizes Numeric. Point size. Default: \code{0.6}.
#' @param ci_Theight Numeric. CI cap height. Default: \code{0.2}.
#' @param ci_digits Integer. Decimal places for the auto-generated
#'   \code{OR (95\% CI)} column. Default: \code{2L}.
#' @param ci_sep Character. Separator between lower and upper CI in the label,
#'   e.g. \code{", "} or \code{" - "}. Default: \code{", "}.
#' @param p_cols Character vector. Names of numeric p-value columns in
#'   \code{data}. These are formatted to \code{p_digits} decimal places with
#'   \code{"<0.001"}-style clipping. \code{NULL} = none.
#' @param p_digits Integer. Decimal places for p-value formatting.
#'   Default: \code{3L}.
#' @param bold_p \code{TRUE} (bold all non-NA p below \code{p_threshold}),
#'   \code{FALSE} (no bolding), or a logical vector (per-row control).
#'   Default: \code{TRUE}.
#' @param p_threshold Numeric. P-value threshold for bolding when
#'   \code{bold_p = TRUE}. Default: \code{0.05}.
#' @param align Integer vector of length \code{ncol(data) + 2}. Alignment per
#'   column: \code{-1} left, \code{0} centre, \code{1} right. Must cover all
#'   final columns (original + gap_ci + OR label).
#'   \code{NULL} = auto (column 1 left, all others centre).
#' @param background Character. Row background style: \code{"zebra"},
#'   \code{"bold_label"} (shade rows where \code{bold_label = TRUE}),
#'   or \code{"none"}. Default: \code{"zebra"}.
#' @param bg_col Character. Shading colour for backgrounds (scalar), or a
#'   per-row vector of length \code{nrow(data)} (overrides style).
#'   Default: \code{"#F0F0F0"}.
#' @param border Character. Border style: \code{"three_line"} or \code{"none"}.
#'   Default: \code{"three_line"}.
#' @param border_width Numeric. Border line width(s). Scalar = all three lines
#'   same width; length-3 vector = top-of-header, bottom-of-header,
#'   bottom-of-body. Default: \code{3}.
#' @param row_height \code{NULL} (auto), numeric scalar, or numeric vector
#'   (length = total gtable rows including margins). Auto sets 8 / 12 / 10 / 15
#'   mm for top / header / data / bottom respectively.
#' @param col_width \code{NULL} (auto), numeric scalar, or numeric vector
#'   (length = total gtable columns). Auto rounds each default width up so the
#'   adjustment is in \eqn{[5, 10)} mm.
#' @param save Logical. Save output to files? Default: \code{FALSE}.
#' @param dest Character. Destination file path (extension ignored; all four
#'   formats are saved). Required when \code{save = TRUE}.
#' @param save_width Numeric. Output width in cm. Default: \code{20}.
#' @param save_height Numeric or \code{NULL}. Output height in cm.
#'   \code{NULL} = \code{nrow(data) * 0.9 + 3}.
#' @param theme Character preset (\code{"default"}) or a
#'   \code{forestploter::forest_theme} object. Default: \code{"default"}.
#'
#' @return A \pkg{forestploter} plot object (gtable), returned invisibly.
#'   Display with \code{plot()} or \code{grid::grid.draw()}.
#'
#' @importFrom forestploter forest forest_theme edit_plot add_border
#' @importFrom grid gpar unit
#' @export
#'
#' @examples
#' df <- data.frame(
#'   item      = c("Exposure vs. control", "Unadjusted", "Fully adjusted"),
#'   `Cases/N` = c("", "89/4521", "89/4521"),
#'   p_value   = c(NA_real_, 0.001, 0.006),
#'   check.names = FALSE
#' )
#'
#' p <- plot_forest(
#'   data       = df,
#'   est        = c(NA, 1.52, 1.43),
#'   lower      = c(NA, 1.18, 1.11),
#'   upper      = c(NA, 1.96, 1.85),
#'   ci_column  = 2L,
#'   indent     = c(0L, 1L, 1L),
#'   bold_label = c(TRUE, FALSE, FALSE),
#'   p_cols     = "p_value",
#'   xlim       = c(0.5, 3.0)
#' )
#' plot(p)
plot_forest <- function(data,
                        est,
                        lower,
                        upper,
                        ci_column   = 2L,
                        ref_line    = 1,
                        xlim        = NULL,
                        ticks_at    = NULL,
                        arrow_lab   = c("Lower risk", "Higher risk"),
                        header      = NULL,
                        indent      = NULL,
                        bold_label  = NULL,
                        ci_col      = "black",
                        ci_sizes    = 0.6,
                        ci_Theight  = 0.2,
                        ci_digits   = 2L,
                        ci_sep      = ", ",
                        p_cols      = NULL,
                        p_digits    = 3L,
                        bold_p      = TRUE,
                        p_threshold = 0.05,
                        align       = NULL,
                        background  = "zebra",
                        bg_col      = "#F0F0F0",
                        border      = "three_line",
                        border_width = 3,
                        row_height  = NULL,
                        col_width   = NULL,
                        save        = FALSE,
                        dest        = NULL,
                        save_width  = 20,
                        save_height = NULL,
                        theme       = "default") {

  # Validate inputs
  .assert_data_frame(data)
  n        <- nrow(data)
  nc_orig  <- ncol(data)
  nc_final <- nc_orig + 2L   # gap_ci + OR label always inserted

  .assert_length_n(est,   n)
  .assert_length_n(lower, n)
  .assert_length_n(upper, n)
  if (!is.null(indent))     .assert_length_n(indent,     n)
  if (!is.null(bold_label)) .assert_length_n(bold_label, n)
  if (length(ci_col) > 1L)  .assert_length_n(ci_col,     n)
  .assert_has_cols(data, p_cols)
  .assert_logical(bold_p)
  if (length(bold_p) == 1L) bold_p <- rep(bold_p, n)
  .assert_length_n(bold_p, n)

  if (ci_column < 2L || ci_column > nc_orig + 1L)
    cli::cli_abort(
      "{.arg ci_column} must be between 2 and {nc_orig + 1L} (ncol(data) + 1).",
      call = NULL
    )
  if (!is.null(header)) .assert_length_n(header, nc_final)
  if (!is.null(align))  .assert_length_n(align,  nc_final)

  # Defaults
  if (is.null(indent)) indent <- rep(0L, n)
  # bold_label default: bold parent rows (indent == 0), leave sub-rows plain
  if (is.null(bold_label)) bold_label <- indent == 0L

  # Pre-process data
  out        <- .fp_build_data(data, est, lower, upper, indent,
                               p_cols, p_digits, ci_digits, ci_sep, ci_column)
  data_r     <- out$data
  p_col_idxs <- out$p_col_idxs
  p_numeric  <- out$p_numeric

  nr <- nrow(data_r)

  if (!is.null(header)) names(data_r) <- header

  # Build base plot
  if (is.null(xlim)) xlim <- c(0, 2)
  if (is.null(ticks_at))
    ticks_at <- seq(xlim[1L], xlim[2L], length.out = 5L)

  p <- forestploter::forest(
    data      = data_r,
    est       = list(est),
    lower     = list(lower),
    upper     = list(upper),
    ci_column = ci_column,
    ref_line  = ref_line,
    xlim      = xlim,
    ticks_at  = ticks_at,
    arrow_lab = arrow_lab,
    sizes     = ci_sizes,
    theme     = .fp_theme(theme, ci_Theight = ci_Theight)
  )

  # Post-processing
  p <- .fp_align(p, nc_final, align)
  p <- .fp_bold_label(p, bold_label)

  if (!is.null(p_cols))
    p <- .fp_bold_p(p, p_cols, p_col_idxs, p_numeric, bold_p, p_threshold, nr)

  p <- .fp_ci_colors(p, ci_col, ci_column, est, nr)
  p <- .fp_background(p, nr, nc_final, background, bold_label, bg_col)
  p <- .fp_borders(p, nr, border, border_width)
  p <- .fp_layout(p, row_height, col_width)

  # Save
  if (isTRUE(save)) {
    if (is.null(dest))
      cli::cli_abort("{.arg dest} must be provided when {.arg save = TRUE}.", call = NULL)
    h <- if (is.null(save_height)) nr * 0.9 + 3 else save_height
    .fp_save(p, dest, save_width, h)
  }

  invisible(p)
}


#' Publication-ready Table 1 (Baseline Characteristics)
#'
#' Generates a publication-quality baseline-characteristics table (Table 1)
#' using \pkg{gtsummary}, with optional SMD column, \emph{Lancet}-style
#' theming, and automatic export to four formats.
#'
#' The following behaviours are fixed (not exposed as parameters):
#' \itemize{
#'   \item Variable labels are \strong{bold} (\code{bold_labels()}).
#'   \item P-values are formatted as \code{<0.001} or to 3 decimal places.
#'   \item Significant p-values (\eqn{p < 0.05}) are \strong{bold}.
#'   \item The p-value column header is rendered as \emph{P}-value.
#'   \item The table uses a three-line (booktabs) border.
#'   \item Saving always exports \strong{word}, \strong{html}, \strong{pdf},
#'     and \strong{png}.
#' }
#'
#' @param data data.frame. Input data containing all variables.
#' @param vars Character vector. Variable names to display in the table.
#' @param strata Character scalar. Column name for the grouping/stratification
#'   variable. \code{NULL} produces an unstratified overall summary.
#' @param type Named list or \code{NULL}. Variable type overrides passed
#'   directly to \code{\link[gtsummary]{tbl_summary}}.
#'   E.g. \code{list(age = "continuous2")}.
#' @param label Named list / formula list or \code{NULL}. Label overrides.
#'   E.g. \code{list(age ~ "Age (years)", sex ~ "Sex")}.
#' @param statistic Named list or \code{NULL}. Statistic format strings.
#'   E.g. \code{list(all_continuous() ~ "{mean} ({sd})")}.
#' @param digits Named list or \code{NULL}. Decimal place overrides.
#'   E.g. \code{list(age ~ 1)}.
#' @param percent Character. Percentage base for categorical variables:
#'   \code{"column"} (default), \code{"row"}, or \code{"cell"}.
#' @param missing Character. Display of missing counts:
#'   \code{"no"} (default), \code{"ifany"}, or \code{"always"}.
#' @param add_p Logical. Add a p-value column. Default \code{TRUE};
#'   disabled with a warning if \code{strata} is \code{NULL}.
#' @param add_smd Logical. Add a standardised mean difference (SMD) column.
#'   Continuous variables use Cohen's \emph{d}; categorical variables use
#'   root-mean-square deviation (RMSD) of group proportions. SMD is shown
#'   only on label rows. Default \code{FALSE}; disabled with a warning if
#'   \code{strata} is \code{NULL}.
#' @param overall Logical. Add an Overall column when \code{strata} is set.
#'   Default \code{FALSE}.
#' @param exclude_labels Character vector or \code{NULL}. Row labels to
#'   remove from the rendered table, matched exactly against the
#'   \code{label} column in the gtsummary table body.
#'   E.g. \code{c("Unknown", "Missing")}.
#' @param theme Character. Visual theme preset. Only \code{"lancet"}
#'   (default) is currently supported. Controls alternating row shading
#'   (\code{#f8e8e7} / white), three-line borders, and 15 px font size.
#' @param label_width Integer. Label column width in pixels. Default \code{200}.
#' @param stat_width Integer. Statistics column(s) width in pixels.
#'   Default \code{140}.
#' @param pvalue_width Integer. P-value and SMD column width in pixels.
#'   Default \code{100}.
#' @param row_height Integer. Data row padding (top + bottom) in pixels.
#'   Default \code{8}.
#' @param save Logical. Export the table to files. Default \code{FALSE}.
#' @param dest Character or \code{NULL}. File path without extension.
#'   When \code{save = TRUE}, four files are written:
#'   \code{<dest>.docx}, \code{<dest>.html}, \code{<dest>.pdf},
#'   \code{<dest>.png}. Required when \code{save = TRUE}.
#' @param png_scale Numeric. Zoom factor for PNG export via \pkg{webshot2}.
#'   Higher values produce larger, higher-resolution images. Default: \code{2}.
#' @param pdf_width Numeric or \code{NULL}. PDF paper width in inches passed to
#'   \code{pagedown::chrome_print}. A larger value increases the page size so
#'   more content fits on a single page. Default: \code{NULL} (Chrome default).
#' @param pdf_height Numeric or \code{NULL}. PDF paper height in inches.
#'   Increase if the table is cut off across pages. Default: \code{NULL}.
#'
#' @return A \pkg{gt} table object, returned invisibly.
#'
#' @importFrom broom tidy
#' @importFrom gtsummary tbl_summary bold_labels add_overall add_p bold_p
#'   modify_header modify_table_body modify_table_styling as_gt style_pvalue
#' @importFrom gt gtsave tab_options tab_style opt_css
#'   cell_borders cell_fill cells_column_labels cells_body cols_width px
#' @importFrom dplyr any_of
#' @export
#'
#' @examples
#' library(gtsummary)
#' data(trial)
#'
#' # Basic stratified table
#' plot_tableone(
#'   data   = trial,
#'   vars   = c("age", "marker", "grade"),
#'   strata = "trt",
#'   save   = FALSE
#' )
#'
#' # With SMD, custom labels, exclude Unknown level
#' plot_tableone(
#'   data           = trial,
#'   vars           = c("age", "marker", "grade", "stage"),
#'   strata         = "trt",
#'   label          = list(age ~ "Age (years)", marker ~ "Marker level"),
#'   add_smd        = TRUE,
#'   exclude_labels = "Unknown",
#'   save           = FALSE
#' )
plot_tableone <- function(
    data,
    vars,
    strata         = NULL,
    type           = NULL,
    label          = NULL,
    statistic      = NULL,
    digits         = NULL,
    percent        = "column",
    missing        = "no",
    add_p          = TRUE,
    add_smd        = FALSE,
    overall        = FALSE,
    exclude_labels = NULL,
    theme          = "lancet",
    label_width    = 200,
    stat_width     = 140,
    pvalue_width   = 100,
    row_height     = 8,
    save           = FALSE,
    dest           = NULL,
    png_scale      = 2,
    pdf_width      = NULL,
    pdf_height     = NULL
) {

  # Validate
  .assert_data_frame(data)
  data <- as.data.frame(data)
  .assert_has_cols(data, vars)
  .assert_has_cols(data, strata)

  if (!is.null(exclude_labels)) .assert_character(exclude_labels)
  .assert_not_null_if_true(save, dest)

  if (isTRUE(add_p) && is.null(strata)) {
    cli::cli_warn("{.arg add_p} requires {.arg strata}; disabling.")
    add_p <- FALSE
  }

  if (isTRUE(add_smd) && is.null(strata)) {
    cli::cli_warn("{.arg add_smd} requires {.arg strata}; disabling.")
    add_smd <- FALSE
  }

  # Build
  tbl <- .t1_build_tbl(data, vars, strata, type, label, statistic,
                        digits, percent, missing)

  if (isTRUE(overall) && !is.null(strata))
    tbl <- gtsummary::add_overall(tbl)

  if (isTRUE(add_p))
    tbl <- .t1_add_p(tbl)

  if (isTRUE(add_smd))
    tbl <- .t1_add_smd(tbl, data, vars, strata)

  if (!is.null(exclude_labels))
    tbl <- gtsummary::modify_table_body(
      tbl,
      ~ dplyr::filter(.x, !.data$label %in% exclude_labels)
    )

  # Render
  gt_tbl <- gtsummary::as_gt(tbl)
  gt_tbl <- .t1_apply_theme(gt_tbl, tbl, theme, label_width, stat_width,
                              pvalue_width, row_height)

  # Save
  if (isTRUE(save))
    .t1_save(gt_tbl, dest, png_scale, pdf_width, pdf_height)

  invisible(gt_tbl)
}

# =============================================================================
# test-integration-plot.R — Integration tests for plot_forest() and plot_tableone()
#   These tests call the full function pipeline end-to-end.
#   plot_forest() requires an open graphics device (grid layout computation).
# =============================================================================


# ===========================================================================
# Shared helpers
# ===========================================================================

skip_on_cran()

.fake_forest_df <- function() {
  data.frame(
    item    = c("AD vs control", "Crude", "Adjusted"),
    cases_n = c("", "89/4521", "89/4521"),
    stringsAsFactors = FALSE
  )
}

.fake_t1_df <- function(n = 120, seed = 42) {
  set.seed(seed)
  data.frame(
    age = round(rnorm(n, 57, 8), 1),
    sex = factor(sample(c("Male", "Female"), n, TRUE)),
    bmi = round(rnorm(n, 27, 5), 1),
    trt = factor(sample(c("Drug A", "Drug B"), n, TRUE)),
    stringsAsFactors = FALSE
  )
}

# Open a temporary PDF device and close it on exit.
# Returns the temp file path (for cleanup).
.open_pdf_device <- function() {
  tmp <- tempfile(fileext = ".pdf")
  grDevices::pdf(tmp, width = 12, height = 8)
  tmp
}

.close_pdf_device <- function(tmp) {
  try(grDevices::dev.off(), silent = TRUE)
  unlink(tmp)
}


# ===========================================================================
# plot_forest() — end-to-end
# ===========================================================================

test_that("plot_forest() returns an invisible gtable-like object", {
  df  <- .fake_forest_df()
  tmp <- .open_pdf_device()
  on.exit(.close_pdf_device(tmp), add = TRUE)

  p <- suppressMessages(
    plot_forest(
      data  = df,
      est   = c(NA, 1.52, 1.43),
      lower = c(NA, 1.18, 1.11),
      upper = c(NA, 1.96, 1.85)
    )
  )
  # forestploter::forest() returns a gtable (list-based S3)
  expect_true(is.list(p))
  expect_true(!is.null(p$grobs) || !is.null(p$heights))
})

test_that("plot_forest() accepts ci_column = 3 with extra column", {
  df <- data.frame(
    item    = c("Group", "Crude", "Adjusted"),
    cases_n = c("", "45/800", "45/800"),
    p_value = c(NA_real_, 0.031, 0.004),
    stringsAsFactors = FALSE
  )
  tmp <- .open_pdf_device()
  on.exit(.close_pdf_device(tmp), add = TRUE)

  expect_no_error(
    suppressMessages(
      plot_forest(
        data      = df,
        est       = c(NA, 1.30, 1.25),
        lower     = c(NA, 1.05, 1.01),
        upper     = c(NA, 1.60, 1.55),
        ci_column = 3L,
        p_cols    = "p_value"
      )
    )
  )
})

test_that("plot_forest() works with indent and bold_label", {
  df  <- .fake_forest_df()
  tmp <- .open_pdf_device()
  on.exit(.close_pdf_device(tmp), add = TRUE)

  expect_no_error(
    suppressMessages(
      plot_forest(
        data       = df,
        est        = c(NA, 1.52, 1.43),
        lower      = c(NA, 1.18, 1.11),
        upper      = c(NA, 1.96, 1.85),
        indent     = c(0L, 1L, 1L),
        bold_label = c(TRUE, FALSE, FALSE)
      )
    )
  )
})

test_that("plot_forest() works with xlim and ticks_at", {
  df  <- .fake_forest_df()
  tmp <- .open_pdf_device()
  on.exit(.close_pdf_device(tmp), add = TRUE)

  expect_no_error(
    suppressMessages(
      plot_forest(
        data     = df,
        est      = c(NA, 1.52, 1.43),
        lower    = c(NA, 1.18, 1.11),
        upper    = c(NA, 1.96, 1.85),
        xlim     = c(0.5, 3.0),
        ticks_at = c(0.5, 1.0, 2.0, 3.0)
      )
    )
  )
})

test_that("plot_forest() works with ref_line = 0 (beta coefficients)", {
  df  <- .fake_forest_df()
  tmp <- .open_pdf_device()
  on.exit(.close_pdf_device(tmp), add = TRUE)

  expect_no_error(
    suppressMessages(
      plot_forest(
        data     = df,
        est      = c(NA, 0.42, 0.38),
        lower    = c(NA, 0.15, 0.10),
        upper    = c(NA, 0.69, 0.66),
        ref_line = 0
      )
    )
  )
})

test_that("plot_forest() works with background = 'bold_label'", {
  df  <- .fake_forest_df()
  tmp <- .open_pdf_device()
  on.exit(.close_pdf_device(tmp), add = TRUE)

  expect_no_error(
    suppressMessages(
      plot_forest(
        data       = df,
        est        = c(NA, 1.52, 1.43),
        lower      = c(NA, 1.18, 1.11),
        upper      = c(NA, 1.96, 1.85),
        indent     = c(0L, 1L, 1L),
        background = "bold_label"
      )
    )
  )
})

test_that("plot_forest() works with background = 'none' and border = 'none'", {
  df  <- .fake_forest_df()
  tmp <- .open_pdf_device()
  on.exit(.close_pdf_device(tmp), add = TRUE)

  expect_no_error(
    suppressMessages(
      plot_forest(
        data       = df,
        est        = c(NA, 1.52, 1.43),
        lower      = c(NA, 1.18, 1.11),
        upper      = c(NA, 1.96, 1.85),
        background = "none",
        border     = "none"
      )
    )
  )
})

test_that("plot_forest() works with per-row ci_col vector", {
  df  <- .fake_forest_df()
  tmp <- .open_pdf_device()
  on.exit(.close_pdf_device(tmp), add = TRUE)

  expect_no_error(
    suppressMessages(
      plot_forest(
        data   = df,
        est    = c(NA, 1.52, 1.43),
        lower  = c(NA, 1.18, 1.11),
        upper  = c(NA, 1.96, 1.85),
        ci_col = c(NA, "steelblue", "tomato")
      )
    )
  )
})

test_that("plot_forest() works with custom row_height scalar", {
  df  <- .fake_forest_df()
  tmp <- .open_pdf_device()
  on.exit(.close_pdf_device(tmp), add = TRUE)

  expect_no_error(
    suppressMessages(
      plot_forest(
        data       = df,
        est        = c(NA, 1.52, 1.43),
        lower      = c(NA, 1.18, 1.11),
        upper      = c(NA, 1.96, 1.85),
        row_height = 10
      )
    )
  )
})

test_that("plot_forest() with p_cols bold_p = FALSE skips bolding", {
  df <- data.frame(
    item    = c("Group", "Crude", "Adjusted"),
    p_value = c(NA_real_, 0.001, 0.220),
    stringsAsFactors = FALSE
  )
  tmp <- .open_pdf_device()
  on.exit(.close_pdf_device(tmp), add = TRUE)

  expect_no_error(
    suppressMessages(
      plot_forest(
        data   = df,
        est    = c(NA, 1.52, 1.43),
        lower  = c(NA, 1.18, 1.11),
        upper  = c(NA, 1.96, 1.85),
        p_cols = "p_value",
        bold_p = FALSE
      )
    )
  )
})

test_that("plot_forest() works with bold_p as per-row logical vector", {
  df <- data.frame(
    item    = c("Group", "Crude", "Adjusted"),
    p_value = c(NA_real_, 0.001, 0.220),
    stringsAsFactors = FALSE
  )
  tmp <- .open_pdf_device()
  on.exit(.close_pdf_device(tmp), add = TRUE)

  expect_no_error(
    suppressMessages(
      plot_forest(
        data   = df,
        est    = c(NA, 1.52, 1.43),
        lower  = c(NA, 1.18, 1.11),
        upper  = c(NA, 1.96, 1.85),
        p_cols = "p_value",
        bold_p = c(FALSE, TRUE, FALSE)
      )
    )
  )
})

test_that("plot_forest() works with arrow_lab = NULL", {
  df  <- .fake_forest_df()
  tmp <- .open_pdf_device()
  on.exit(.close_pdf_device(tmp), add = TRUE)

  expect_no_error(
    suppressMessages(
      plot_forest(
        data      = df,
        est       = c(NA, 1.52, 1.43),
        lower     = c(NA, 1.18, 1.11),
        upper     = c(NA, 1.96, 1.85),
        arrow_lab = NULL
      )
    )
  )
})

test_that("plot_forest() works with custom col_width scalar", {
  df  <- .fake_forest_df()
  tmp <- .open_pdf_device()
  on.exit(.close_pdf_device(tmp), add = TRUE)

  expect_no_error(
    suppressMessages(
      plot_forest(
        data      = df,
        est       = c(NA, 1.52, 1.43),
        lower     = c(NA, 1.18, 1.11),
        upper     = c(NA, 1.96, 1.85),
        col_width = 30
      )
    )
  )
})

test_that("plot_forest() accepts header of final column count (ncol + 2)", {
  # Final table: item | gap_ci | OR (95% CI) | cases_n  → 4 columns
  df  <- .fake_forest_df()
  tmp <- .open_pdf_device()
  on.exit(.close_pdf_device(tmp), add = TRUE)

  expect_no_error(
    suppressMessages(
      plot_forest(
        data   = df,
        est    = c(NA, 1.52, 1.43),
        lower  = c(NA, 1.18, 1.11),
        upper  = c(NA, 1.96, 1.85),
        header = c("Exposure", "", "HR (95% CI)", "Cases/N")
      )
    )
  )
})

test_that("plot_forest() warns on unknown theme and falls back to default", {
  df  <- .fake_forest_df()
  tmp <- .open_pdf_device()
  on.exit(.close_pdf_device(tmp), add = TRUE)

  expect_warning(
    suppressMessages(
      plot_forest(
        data  = df,
        est   = c(NA, 1.52, 1.43),
        lower = c(NA, 1.18, 1.11),
        upper = c(NA, 1.96, 1.85),
        theme = "nonexistent_theme"
      )
    ),
    "Unknown theme"
  )
})


# ===========================================================================
# plot_tableone() — end-to-end
# ===========================================================================

test_that("plot_tableone() returns a gt table object", {
  df  <- .fake_t1_df()
  res <- suppressMessages(suppressWarnings(
    plot_tableone(df, vars = c("age", "sex", "bmi"), save = FALSE)
  ))
  expect_true(inherits(res, "gt_tbl"))
})

test_that("plot_tableone() works with strata and add_p = TRUE", {
  df  <- .fake_t1_df()
  expect_no_error(
    suppressMessages(suppressWarnings(
      plot_tableone(
        df,
        vars   = c("age", "bmi"),
        strata = "trt",
        add_p  = TRUE,
        save   = FALSE
      )
    ))
  )
})

test_that("plot_tableone() works with add_smd = TRUE", {
  df  <- .fake_t1_df()
  expect_no_error(
    suppressMessages(suppressWarnings(
      plot_tableone(
        df,
        vars    = c("age", "sex", "bmi"),
        strata  = "trt",
        add_smd = TRUE,
        save    = FALSE
      )
    ))
  )
})

test_that("plot_tableone() works with overall = TRUE", {
  df  <- .fake_t1_df()
  expect_no_error(
    suppressMessages(suppressWarnings(
      plot_tableone(
        df,
        vars    = c("age", "bmi"),
        strata  = "trt",
        overall = TRUE,
        save    = FALSE
      )
    ))
  )
})

test_that("plot_tableone() works without strata (unstratified)", {
  df  <- .fake_t1_df()
  res <- suppressMessages(suppressWarnings(
    plot_tableone(df, vars = c("age", "bmi"), strata = NULL, save = FALSE)
  ))
  expect_true(inherits(res, "gt_tbl"))
})

test_that("plot_tableone() works with custom label list", {
  df  <- .fake_t1_df()
  expect_no_error(
    suppressMessages(suppressWarnings(
      plot_tableone(
        df,
        vars   = c("age", "sex"),
        strata = "trt",
        label  = list(age ~ "Age (years)", sex ~ "Sex"),
        save   = FALSE
      )
    ))
  )
})

test_that("plot_tableone() works with exclude_labels", {
  df  <- .fake_t1_df()
  # sex has levels Male / Female; exclude "Female" level row
  expect_no_error(
    suppressMessages(suppressWarnings(
      plot_tableone(
        df,
        vars           = c("age", "sex"),
        strata         = "trt",
        exclude_labels = "Female",
        save           = FALSE
      )
    ))
  )
})

test_that("plot_tableone() works with missing = 'ifany'", {
  df      <- .fake_t1_df()
  df$age[sample(nrow(df), 10)] <- NA
  expect_no_error(
    suppressMessages(suppressWarnings(
      plot_tableone(
        df,
        vars    = c("age", "bmi"),
        strata  = "trt",
        missing = "ifany",
        save    = FALSE
      )
    ))
  )
})

test_that("plot_tableone() with all options combined does not error", {
  df  <- .fake_t1_df()
  expect_no_error(
    suppressMessages(suppressWarnings(
      plot_tableone(
        df,
        vars    = c("age", "sex", "bmi"),
        strata  = "trt",
        add_p   = TRUE,
        add_smd = TRUE,
        overall = TRUE,
        label   = list(age ~ "Age (years)", bmi ~ "BMI (kg/m\u00b2)"),
        save    = FALSE
      )
    ))
  )
})

test_that("plot_tableone() works with percent = 'row'", {
  df  <- .fake_t1_df()
  expect_no_error(
    suppressMessages(suppressWarnings(
      plot_tableone(
        df,
        vars    = c("sex"),
        strata  = "trt",
        percent = "row",
        save    = FALSE
      )
    ))
  )
})

test_that("plot_tableone() works with missing = 'always'", {
  df      <- .fake_t1_df()
  df$age[sample(nrow(df), 5)] <- NA
  expect_no_error(
    suppressMessages(suppressWarnings(
      plot_tableone(
        df,
        vars    = c("age", "bmi"),
        strata  = "trt",
        missing = "always",
        save    = FALSE
      )
    ))
  )
})


# ===========================================================================
# plot_tableone() — save branch smoke test
# ===========================================================================

test_that("plot_tableone() save = TRUE writes an HTML file", {
  df   <- .fake_t1_df()
  base <- tempfile()
  on.exit(
    unlink(paste0(base, c(".docx", ".html", ".pdf", ".png")), force = TRUE),
    add = TRUE
  )

  suppressMessages(suppressWarnings(
    plot_tableone(
      df,
      vars   = c("age", "sex"),
      strata = "trt",
      save   = TRUE,
      dest   = base
    )
  ))

  expect_true(file.exists(paste0(base, ".html")))
})

# =============================================================================
# helper-integration.R — Shared helpers for integration tests
# Loaded automatically by testthat before running tests
# =============================================================================


#' Skip integration test if prerequisites are not met
#'
#' Skips on CI, CRAN, and when DX_API_TOKEN is not set.
#' Returns the token invisibly so tests can reuse it without a second Sys.getenv().
#'
#' @keywords internal
.skip_if_no_dx_token <- function() {
  testthat::skip_on_ci()
  testthat::skip_on_cran()

  token <- Sys.getenv("DX_API_TOKEN")
  if (!nzchar(token)) {
    testthat::skip("DX_API_TOKEN not set. Set it to run integration tests.")
  }

  invisible(token)
}


#' Skip integration test if not running on RAP with a valid token
#'
#' Extends \code{.skip_if_no_dx_token()} with an additional RAP environment
#' check. Use for tests that require \code{fetch_file()} or other RAP-only
#' operations.
#'
#' @keywords internal
.skip_if_no_rap <- function() {
  .skip_if_no_dx_token()
  if (!ukbflow:::.is_on_rap()) {
    testthat::skip("Not running on RAP — requires the RAP environment.")
  }
}


#' Skip if mockery is not installed
#'
#' Used at the top of test files that rely on \code{mockery::stub()}.
#' Ensures graceful skipping when the package is unavailable (e.g. during
#' \code{_R_CHECK_DEPENDS_ONLY_} checks).
#'
#' @keywords internal
.skip_if_no_mockery <- function() {
  testthat::skip_if_not_installed("mockery")
}

# =============================================================================
# utils_auth.R — Internal helpers for auth_ series
# =============================================================================


#' @keywords internal
#' @noRd
.dx_run <- function(args, timeout = 30, env = NULL) {
  # Check dx is available
  dx_path <- Sys.which("dx")
  if (dx_path == "") {
    stop(
      "dx-toolkit not found. Please install dxpy: pip install dxpy",
      call. = FALSE
    )
  }

  # Reason: merge extra env vars into current environment so PATH and other
  # required vars are preserved (processx replaces env entirely if set)
  run_env <- if (!is.null(env)) c(Sys.getenv(), env) else NULL

  result <- processx::run(
    command = dx_path,
    args = args,
    error_on_status = FALSE,
    timeout = timeout,
    env = run_env
  )

  list(
    stdout  = trimws(result$stdout),
    stderr  = trimws(result$stderr),
    status  = result$status,
    success = result$status == 0
  )
}


#' @keywords internal
#' @noRd
.dx_get_project_id <- function() {
  result <- .dx_run(c("env", "--bash"))
  project_id <- regmatches(
    result$stdout,
    regexpr("(?<=DX_PROJECT_CONTEXT_ID=)[^\n]+", result$stdout, perl = TRUE)
  )
  if (length(project_id) == 0) NA_character_ else project_id
}

# =============================================================================
# utils_fetch.R — Internal helpers for fetch_ series
# Depends on .dx_run() defined in utils_auth.R
# =============================================================================


#' Normalize a remote RAP path to dx-compatible format
#'
#' @keywords internal
#' @noRd
.dx_normalize_path <- function(path) {
  # Reason: dx CLI uses relative paths from project root — a leading slash
  # causes ResolutionError ("Could not find project named 'C'")
  path <- sub("^/+", "", trimws(path))
  if (path == ".") path <- ""
  path
}


#' Probe a remote RAP path and return its type
#'
#' @param path (character) Remote path to probe.
#' @return A list with fields:
#'   \describe{
#'     \item{is_folder}{Logical. TRUE if the path resolves to a folder.}
#'     \item{is_file}{Logical. TRUE if the path resolves to a file.}
#'     \item{exists}{Logical. TRUE if the path exists on the platform.}
#'     \item{stdout}{Raw stdout from dx ls -l, for further inspection.}
#'   }
#'
#' @keywords internal
#' @noRd
.dx_probe_path <- function(path) {
  # Reason: trailing slash is a reliable client-side signal that the caller
  # already knows the path is a folder — skip the network round-trip.
  if (endsWith(trimws(path), "/")) {
    return(list(is_folder = TRUE, is_file = FALSE, exists = TRUE, stdout = ""))
  }

  result <- .dx_ls_raw(path)
  is_folder <- result$success && grepl("Folder\\s*:", result$stdout)
  list(
    is_folder = is_folder,
    is_file   = result$success && !is_folder,
    exists    = result$success,
    stdout    = result$stdout
  )
}


#' Run dx ls -l on a remote path and return raw output
#'
#' @keywords internal
#' @noRd
.dx_ls_raw <- function(path, timeout = 60) {
  norm <- .dx_normalize_path(path)
  args <- c("ls", "-l")
  if (nzchar(norm)) args <- c(args, norm)
  .dx_run(args, timeout = timeout)
}


#' Run dx ls (names only) on a remote path and return raw output
#'
#' @keywords internal
#' @noRd
.dx_ls_names <- function(path, timeout = 60) {
  norm <- .dx_normalize_path(path)
  args <- c("ls")
  if (nzchar(norm)) args <- c(args, norm)
  .dx_run(args, timeout = timeout)
}


#' Generate a pre-authenticated download URL for a remote RAP file
#'
#' @param path (character) Remote file path.
#' @param duration (character) URL validity duration. Default: \code{"1d"}.
#' @return Character string — the pre-authenticated HTTPS URL.
#'
#' @keywords internal
#' @noRd
.dx_make_url <- function(path, duration = "1d") {
  norm <- .dx_normalize_path(path)
  result <- .dx_run(c("make_download_url", "--duration", duration, norm), timeout = 30)
  if (!result$success) {
    stop("Failed to generate URL for '", path, "': ", result$stderr, call. = FALSE)
  }
  trimws(result$stdout)
}


#' Download a single file via pre-authenticated URL
#'
#' @param url (character) Pre-authenticated HTTPS URL.
#' @param destfile (character) Full local file path to save to.
#' @param overwrite (logical) Overwrite if file exists. Default: FALSE.
#' @param resume (logical) Resume interrupted download. Default: FALSE.
#'
#' @keywords internal
#' @noRd
.dx_download_file <- function(url, destfile, overwrite = FALSE, resume = FALSE, verbose = TRUE) {
  if (file.exists(destfile) && !overwrite && !resume) {
    cli::cli_inform("Skipping (already exists): {.file {basename(destfile)}}")
    return(invisible(destfile))
  }

  h <- curl::new_handle()
  curl::handle_setopt(h, followlocation = TRUE, noprogress = !verbose)

  # Reason: resume appends from the current file size if signatures match
  if (resume && file.exists(destfile)) {
    curl::handle_setopt(h, resume_from = file.info(destfile)$size)
  }

  curl::curl_download(url, destfile = destfile, handle = h, quiet = !verbose)
  invisible(destfile)
}


#' Download multiple files in parallel via pre-authenticated URLs
#'
#' @param urls (character) Named character vector of URLs (names = filenames).
#' @param destfiles (character) Character vector of full local file paths.
#' @param overwrite (logical) Overwrite existing files. Default: FALSE.
#' @param resume (logical) Resume interrupted downloads. Default: FALSE.
#'
#' @keywords internal
#' @noRd
.dx_download_batch <- function(urls, destfiles, overwrite = FALSE,
                                resume = FALSE, verbose = TRUE) {
  # Reason: skip files that already exist when overwrite = FALSE and resume = FALSE
  if (!overwrite && !resume) {
    skip    <- file.exists(destfiles)
    if (any(skip)) {
      cli::cli_inform("Skipping {sum(skip)} already existing file(s).")
      urls      <- urls[!skip]
      destfiles <- destfiles[!skip]
    }
  }

  if (length(urls) == 0) {
    cli::cli_inform("All files already exist, nothing to download.")
    return(invisible(destfiles))
  }

  curl::multi_download(
    urls     = urls,
    destfile = destfiles,
    resume   = resume,
    progress = verbose,
    multiplex = TRUE
  )

  invisible(destfiles)
}


#' Parse dx ls -l stdout into a data.frame
#'
#' @param stdout (character) Raw stdout from dx ls -l.
#' @return data.frame with columns: name, type, size, modified.
#'
#' @keywords internal
#' @noRd
.dx_ls_parse <- function(stdout) {

  lines <- strsplit(stdout, "\n")[[1]]
  lines <- trimws(lines)
  lines <- lines[nzchar(lines)]

  # Drop dx header lines (Project: / Folder :)
  lines <- lines[!grepl("^(Project:|Folder\\s*:)", lines)]

  folder_rows <- list()
  file_rows   <- list()
  in_files    <- FALSE

  for (line in lines) {

    # "State   Last modified ..." marks start of file section
    if (grepl("^State\\s+Last modified", line)) {
      in_files <- TRUE
      next
    }

    # dx prints this when a folder has no files
    if (grepl("^No data objects found", line)) next

    if (!in_files) {
      # Folder lines always end with "/"
      if (endsWith(line, "/")) {
        folder_rows[[length(folder_rows) + 1]] <- data.frame(
          name     = sub("/$", "", line),
          type     = "folder",
          size     = NA_character_,
          modified = as.POSIXct(NA_character_),
          stringsAsFactors = FALSE
        )
      }

    } else {
      # File line format: state  YYYY-MM-DD HH:MM:SS  [size]  name (id)
      # Step 1: strip the trailing (id) so name parsing is clean
      id_m <- regexpr("\\s+\\([a-z]+-[A-Za-z0-9]+\\)\\s*$", line, perl = TRUE)
      if (id_m < 0) next
      core <- trimws(substr(line, 1, id_m - 1))

      # Step 2: parse fixed fields with optional size
      m <- regexec(
        paste0(
          "^(\\S+)",                                      # state
          "\\s+(\\d{4}-\\d{2}-\\d{2} \\d{2}:\\d{2}:\\d{2})",  # modified
          "\\s+(?:([\\d.]+ (?:[KMGT]?B|bytes))\\s+)?",    # optional size (KB/MB/GB/bytes)
          "(.+?)\\s*$"                                    # name
        ),
        core, perl = TRUE
      )
      cap <- regmatches(core, m)[[1]]
      if (length(cap) < 5) next

      file_rows[[length(file_rows) + 1]] <- data.frame(
        name     = cap[5],
        type     = "file",
        size     = if (nzchar(cap[4])) cap[4] else NA_character_,
        modified = as.POSIXct(cap[3], format = "%Y-%m-%d %H:%M:%S"),
        stringsAsFactors = FALSE
      )
    }
  }

  empty <- data.frame(
    name     = character(0),
    type     = character(0),
    size     = character(0),
    modified = as.POSIXct(character(0)),
    stringsAsFactors = FALSE
  )

  rbind(
    if (length(folder_rows) > 0) do.call(rbind, folder_rows) else empty,
    if (length(file_rows)   > 0) do.call(rbind, file_rows)   else empty
  )
}

# =============================================================================
# utils_extract.R — Internal helpers for extract_ series
# Depends on .dx_run() defined in utils_auth.R
# =============================================================================


# Reason: session-level cache avoids repeated --list-fields network requests
# (29,951 rows, takes several seconds each time)
.ukbflow_cache <- new.env(parent = emptyenv())


#' Detect whether the current session is running inside the RAP environment
#'
#' @keywords internal
#' @noRd
.is_on_rap <- function() {
  on_cloud   <- nzchar(Sys.getenv("DX_PROJECT_CONTEXT_ID")) ||
                nzchar(Sys.getenv("DX_JOB_ID"))
  mount_exists <- dir.exists("/mnt/project")
  on_cloud || mount_exists
}


#' Find the most recent .dataset file in the project root
#'
#' @return Character string — the dataset file name.
#'
#' @keywords internal
#' @noRd
.dx_find_dataset <- function() {
  # Reason: dataset name is stable for the session — cache avoids repeated
  # dx ls network calls (e.g. from extract_ls(), decode_names(), etc.)
  if (!is.null(.ukbflow_cache$dataset)) return(.ukbflow_cache$dataset)

  result <- .dx_run("ls", timeout = 30)
  if (!result$success) {
    cli::cli_abort("Failed to list project files: {result$stderr}", call = NULL)
  }

  lines    <- strsplit(result$stdout, "\n")[[1]]
  lines    <- trimws(lines[nzchar(lines)])
  datasets <- lines[grepl("\\.dataset$", lines)]

  if (length(datasets) == 0) {
    cli::cli_abort(
      c("No {.file .dataset} file found in project root.",
        "i" = "Use {.fn fetch_ls} to verify your project contents."),
      call = NULL
    )
  }

  # Reason: multiple datasets may exist (different dates); take the latest
  dataset <- trimws(datasets[length(datasets)])
  .ukbflow_cache$dataset <- dataset
  dataset
}


#' Run dx extract_dataset --list-fields and return raw output
#'
#' @param dataset (character) Dataset file name.
#' @param timeout (integer) Timeout in seconds. Default: 120.
#' @return Named list with stdout, stderr, status, success.
#'
#' @keywords internal
#' @noRd
.dx_list_fields_raw <- function(dataset, timeout = 120) {
  # Reason: PYTHONIOENCODING=utf-8 is required on Windows — the --list-fields
  # output contains non-ASCII characters that crash colorama under GBK encoding
  .dx_run(
    c("extract_dataset", dataset, "--list-fields", "--entities", "participant"),
    timeout = timeout,
    env     = c(PYTHONIOENCODING = "utf-8")
  )
}


#' Match integer field IDs to exact column names in the dataset
#'
#' @param field_id (integer) Vector of field IDs, e.g. \code{c(31, 53, 22189)}.
#' @param fields_df (data.frame) Output of \code{extract_ls()}.
#' @return A list with:
#'   \itemize{
#'     \item \code{matched} — list of per-field info (field_id, field_names,
#'       title, n_cols)
#'     \item \code{unmatched} — integer vector of field IDs with no match
#'   }
#'
#' @keywords internal
#' @noRd
.dx_match_fields <- function(field_id, fields_df) {
  matched   <- list()
  unmatched <- integer(0)

  for (fid in field_id) {
    # Reason: anchor pattern prevents p31 matching p310xx, p311xx, etc.
    pattern <- paste0("^participant\\.p", fid, "(_|$)")
    hits    <- fields_df[grepl(pattern, fields_df$field_name, perl = TRUE), , drop = FALSE]

    if (nrow(hits) == 0) {
      unmatched <- c(unmatched, fid)
    } else {
      # Strip " | Instance X | Array Y" suffix to get base title
      base_title <- sub("\\s*\\|.*$", "", hits$title[1])
      matched[[length(matched) + 1]] <- list(
        field_id    = fid,
        field_names = hits$field_name,
        title       = trimws(base_title),
        n_cols      = nrow(hits)
      )
    }
  }

  list(matched = matched, unmatched = unmatched)
}


#' Run dx extract_dataset --fields-file and write output CSV
#'
#' @param dataset (character) Dataset file name.
#' @param fields (character) Character vector of full field names to extract,
#'   e.g. \code{c("participant.eid", "participant.p31", "participant.p53_i0")}.
#' @param out_path (character) Local path for the output CSV.
#' @param timeout (integer) Timeout in seconds. Default: 300.
#' @return Named list with stdout, stderr, status, success.
#'
#' @keywords internal
#' @noRd
.dx_extract_run <- function(dataset, fields, out_path, timeout = 300) {
  fields_file <- tempfile(fileext = ".txt")
  on.exit(unlink(fields_file))
  # Reason: same CRLF fix as extract_batch() — dx extract_dataset also rejects
  # Windows-style line endings in the fields file
  con <- file(fields_file, open = "wb")
  writeLines(fields, con = con, sep = "\n")
  close(con)

  .dx_run(
    c("extract_dataset", dataset, "--fields-file", fields_file, "-o", out_path),
    timeout = timeout,
    env     = c(PYTHONIOENCODING = "utf-8")
  )
}


#' Upload a local file to DNAnexus project root
#'
#' @param local_path (character) Path to the local file to upload.
#' @return Character string — the DNAnexus file ID (e.g. \code{"file-XXXX"}).
#'
#' @keywords internal
#' @noRd
.dx_upload_file <- function(local_path) {
  result <- .dx_run(c("upload", local_path, "--brief"), timeout = 60)
  if (!result$success) {
    stop("Failed to upload field list: ", result$stderr, call. = FALSE)
  }
  trimws(result$stdout)
}


#' Submit a table-exporter job on DNAnexus
#'
#' @param dataset (character) Dataset file name.
#' @param file_id (character) DNAnexus file ID of the uploaded fields file.
#' @param output (character) Output file name (without extension).
#' @param instance_type (character) DNAnexus instance type.
#' @param priority (character) Job priority: \code{"low"} or \code{"high"}.
#' @return Character string — the job ID (e.g. \code{"job-XXXX"}).
#'
#' @keywords internal
#' @noRd
.dx_run_table_exporter <- function(dataset, file_id, output, instance_type,
                                   priority = "low") {
  result <- .dx_run(
    c(
      "run", "table-exporter",
      paste0("-idataset_or_cohort_or_dashboard=", dataset),
      "-ientity=participant",
      paste0("-ifield_names_file_txt=", file_id),
      paste0("-ioutput=", output),
      "-ioutput_format=CSV",
      paste0("--instance-type=", instance_type),
      "--priority", priority,
      "--brief", "--yes"
    ),
    timeout = 60
  )
  if (!result$success) {
    stop("Failed to submit table-exporter job: ", result$stderr, call. = FALSE)
  }
  job_id <- trimws(result$stdout)
  # Reason: validate job_id format — a successful submission always returns
  # a "job-XXXX" string; anything else indicates a silent failure
  if (!grepl("^job-", job_id)) {
    stop("Unexpected response from table-exporter: ", job_id, call. = FALSE)
  }
  job_id
}


#' Parse --list-fields tab-separated output into a data.frame
#'
#' @param stdout (character) Raw stdout from dx extract_dataset --list-fields.
#' @return data.frame with columns: field_name, title.
#'
#' @keywords internal
#' @noRd
.dx_parse_fields <- function(stdout) {
  lines <- strsplit(stdout, "\n")[[1]]
  lines <- trimws(lines[nzchar(lines)])

  if (length(lines) == 0) {
    return(data.frame(
      field_name = character(0),
      title      = character(0),
      stringsAsFactors = FALSE
    ))
  }

  parts <- strsplit(lines, "\t")

  data.frame(
    field_name = vapply(parts, `[`, character(1), 1),
    title      = vapply(
      parts,
      function(x) if (length(x) >= 2) x[2] else NA_character_,
      character(1)
    ),
    stringsAsFactors = FALSE
  )
}

---
title: "Association Analysis for UKB Outcomes"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Association Analysis for UKB Outcomes}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = FALSE
)
```

## Overview

The `assoc_*` functions fit regression models for each exposure variable and
return tidy result tables suitable for downstream forest plots and publication
tables.

| Function | Alias | Model | Effect measure |
|---|---|---|---|
| `assoc_coxph()` | `assoc_cox()` | Cox proportional hazards | HR |
| `assoc_logistic()` | `assoc_logit()` | Logistic regression | OR |
| `assoc_linear()` | `assoc_lm()` | Linear regression | beta |
| `assoc_coxph_zph()` | `assoc_zph()` | Schoenfeld residual PH test | chisq / p |
| `assoc_subgroup()` | `assoc_sub()` | Stratified analysis + LRT interaction | HR / OR / beta |
| `assoc_trend()` | `assoc_tr()` | Dose-response trend | HR / OR / beta + p_trend |
| `assoc_competing()` | `assoc_fg()` | Fine-Gray competing risks | SHR |
| `assoc_lag()` | — | Cox lag sensitivity analysis | HR |

> **Prerequisite**: the analysis dataset should already contain derived case
> status, follow-up time, and covariates produced by the `derive_*` functions.
> See `vignette("derive")` and `vignette("derive-survival")`.

```{r setup-data}
library(ukbflow)

# ops_toy(scenario = "association") returns a pre-derived analysis-ready table:
# covariates already as factors, bmi_cat / tdi_cat binned, and two outcomes
# (dm_* and htn_*) with status / date / timing / followup columns in place.
dt <- ops_toy(scenario = "association")
dt <- dt[dm_timing != 1L]   # incident analysis: exclude prevalent DM cases
```

---

## The Three-Model Framework

All main functions automatically produce up to three adjustment levels without
requiring manual formula construction:

| Model | Covariates | When included |
|---|---|---|
| **Unadjusted** | None (crude) | Always (when `base = TRUE`) |
| **Age and sex adjusted** | Age (field 21022) + sex (field 31), auto-detected | When both columns are found |
| **Fully adjusted** | User-supplied `covariates` | When `covariates` is non-NULL |

Age and sex columns are located automatically by scanning the dataset's column
names for standard UKB naming patterns (`p21022_*` for age at recruitment,
`p31_*` for sex). As a fallback, any column starting with `"age"` or named
`"sex"` is also recognised. No prerequisite pipeline step is required — the
detection works on any data.frame whose columns follow UKB or ukbflow naming
conventions. Set `base = FALSE` to skip the first two models and run only the
Fully adjusted model.

---

## Step 1: Cox Proportional Hazards

`assoc_coxph()` is the primary function for time-to-event outcomes. It accepts
logical (`TRUE`/`FALSE`) or integer (`0`/`1`) event indicators and returns one
row per exposure x term x model combination.

```{r assoc-coxph}
# Crude + age-sex adjusted (automatic); p21022 and p31 auto-detected
res <- assoc_coxph(
  data         = dt,
  outcome_col  = "dm_status",
  time_col     = "dm_followup_years",
  exposure_col = c("p20116_i0", "bmi_cat")
)
```

```{r assoc-coxph-full}
# Add a Fully adjusted model
res <- assoc_coxph(
  data         = dt,
  outcome_col  = "dm_status",
  time_col     = "dm_followup_years",
  exposure_col = "p20116_i0",
  covariates   = c("bmi_cat", "tdi_cat", "p1558_i0",
                   paste0("p22009_a", 1:10))
)
```

Output columns:

| Column | Description |
|---|---|
| `exposure` | Exposure variable name |
| `term` | Coefficient name from `coxph` |
| `model` | Ordered factor: Unadjusted < Age and sex adjusted < Fully adjusted |
| `n` | Participants in model (after NA removal) |
| `n_events` | Events in model |
| `person_years` | Total person-years (rounded) |
| `HR` | Hazard ratio |

> `n`, `n_events`, and `person_years` all reflect the model-specific
> complete-case analysis set — participants with any missing value across
> outcome, time, exposure, or covariates are dropped before fitting. These
> counts therefore differ between models (e.g. Fully adjusted typically has
> fewer participants than Unadjusted).
| `CI_lower`, `CI_upper` | Confidence interval bounds |
| `p_value` | Wald test p-value |
| `HR_label` | Formatted string, e.g. `"1.43 (1.28-1.60)"` |

---

## Step 2: Logistic Regression

`assoc_logistic()` is for binary outcomes without a time dimension (e.g.
case-control or cross-sectional designs).

```{r assoc-logistic}
res <- assoc_logistic(
  data         = dt,
  outcome_col  = "dm_status",
  exposure_col = c("p20116_i0", "bmi_cat"),
  covariates   = c("tdi_cat", "p1558_i0", paste0("p22009_a", 1:10))
)
```

For sparse data or small samples, use profile likelihood confidence intervals:

```{r assoc-logistic-profile}
res <- assoc_logistic(
  data         = dt,
  outcome_col  = "dm_status",
  exposure_col = "grs_bmi",
  ci_method    = "profile"   # slower but more accurate for sparse data
)
```

Output is identical in structure to `assoc_coxph()` but with `OR` and
`OR_label` in place of `HR` / `HR_label`, and `n_cases` instead of
`n_events` / `person_years`.

---

## Step 3: Linear Regression

`assoc_linear()` is for continuous outcomes (e.g. biomarker levels, BMI).
The standard error of beta is included to support downstream meta-analysis.

```{r assoc-linear}
# Continuous outcome (BMI); smoking and GRS as exposures
res <- assoc_linear(
  data         = dt,
  outcome_col  = "p21001_i0",
  exposure_col = c("p20116_i0", "grs_bmi"),
  covariates   = c("tdi_cat", "p1558_i0", paste0("p22009_a", 1:10))
)
```

> **Common mistake**: passing a binary (0/1) or logical column as
> `outcome_col` is permitted (linear probability model) but triggers a
> warning. Use `assoc_logistic()` for binary outcomes — linear regression
> on a 0/1 outcome produces unbounded predictions and is rarely appropriate
> in epidemiological analyses.

---

## Step 4: Proportional Hazards Assumption Test

`assoc_coxph_zph()` re-fits the same Cox models and tests the PH assumption
via Schoenfeld residuals (`cox.zph()`). Use it alongside `assoc_coxph()` to
validate model assumptions.

```{r assoc-zph}
zph <- assoc_coxph_zph(
  data         = dt,
  outcome_col  = "dm_status",
  time_col     = "dm_followup_years",
  exposure_col = c("p20116_i0", "bmi_cat"),
  covariates   = c("tdi_cat", "p1558_i0")
)

# Identify any violations
zph[ph_satisfied == FALSE]
```

Output columns include `chisq`, `df`, `p_value`, `ph_satisfied` (logical),
and global test statistics (`global_chisq`, `global_df`, `global_p`).

When the PH assumption is violated, consider adding `strata()` to
`assoc_coxph()` or modelling time-varying effects.

---

## Step 5: Subgroup Analysis

`assoc_subgroup()` stratifies the dataset by a grouping variable and fits
the specified model within each subgroup. A likelihood ratio test (LRT) for
the exposure x subgroup interaction is computed on the full dataset and
appended as `p_interaction`.

```{r assoc-subgroup}
# Subgroup by sex; p31 is automatically excluded from within-stratum models
res <- assoc_subgroup(
  data         = dt,
  outcome_col  = "dm_status",
  time_col     = "dm_followup_years",
  exposure_col = "p20116_i0",
  by           = "p31",
  method       = "coxph",
  covariates   = c("p21022", "bmi_cat", "tdi_cat")
)
```

Output columns include `subgroup`, `subgroup_level`, and `p_interaction` in
addition to the standard effect estimate columns.

> **Note**: the `by` variable is automatically excluded from the subgroup
> models. Do not include it in `covariates` — a collinearity warning is
> issued if you do. Subgroup analysis is most interpretable when `by` has a
> small number of levels (e.g. sex, smoking status); continuous variables are
> technically permitted but per-unique-value subgrouping is rarely meaningful
> in practice — use a pre-categorised version (e.g. via `derive_cut()`)
> instead.

---

## Step 6: Dose-Response Trend Analysis

`assoc_trend()` fits categorical and trend models simultaneously for each
ordered-factor exposure, returning per-category estimates alongside a p-value
for linear trend.

```{r assoc-trend-prep}
# assoc_trend() requires an ordered factor; make bmi_cat ordered in-place
dt[, bmi_cat := factor(bmi_cat, levels = levels(bmi_cat), ordered = TRUE)]
```

```{r assoc-trend}
res <- assoc_trend(
  data         = dt,
  outcome_col  = "dm_status",
  time_col     = "dm_followup_years",
  exposure_col = "bmi_cat",
  method       = "coxph",
  covariates   = c("p21022", "p31", "tdi_cat", "p20116_i0")
)
```

Supply custom scores reflecting approximate median BMI per category:

```{r assoc-trend-scores}
res <- assoc_trend(
  data         = dt,
  outcome_col  = "dm_status",
  time_col     = "dm_followup_years",
  exposure_col = "bmi_cat",
  method       = "coxph",
  covariates   = c("p21022", "p31", "tdi_cat", "p20116_i0"),
  scores       = c(17, 22, 27, 35)   # approximate median BMI per category
)
```

The output contains a reference row (`HR = 1.00 (ref)`) followed by
non-reference rows, with `HR_per_score`, `HR_per_score_label`, and `p_trend`
appended as shared columns. These trend columns carry the same value across
every row within the same exposure × model combination — including the
reference row — so the full result table remains self-contained and easy to
filter or export.

---

## Step 7: Fine-Gray Competing Risks

`assoc_competing()` fits a Fine-Gray subdistribution hazard model via
`survival::finegray()` + inverse-probability-of-censoring-weighted Cox
regression. Use this when a competing event (e.g. death) can preclude the
outcome of interest.

Two input modes are supported:

**Mode A** — a single column encodes all event types:

```{r assoc-competing-a}
# Construct a combined event column: 0 = censored, 1 = DM, 2 = HTN (competing)
dt_cr <- dt[htn_timing != 1L]   # also exclude prevalent HTN
dt_cr[, event_type := data.table::fcase(
  dm_timing  == 2L, 1L,
  htn_timing == 2L, 2L,
  default          = 0L
)]

res <- assoc_competing(
  data         = dt_cr,
  outcome_col  = "event_type",       # 0 = censored, 1 = DM, 2 = HTN
  time_col     = "dm_followup_years",
  exposure_col = "p20116_i0",
  event_val    = 1L,
  compete_val  = 2L,
  covariates   = c("bmi_cat", "tdi_cat")
)
```

**Mode B** — separate 0/1 columns for primary and competing events:

```{r assoc-competing-b}
res <- assoc_competing(
  data         = dt_cr,
  outcome_col  = "dm_status",        # primary event
  time_col     = "dm_followup_years",
  exposure_col = c("p20116_i0", "bmi_cat"),
  compete_col  = "htn_status",       # competing event
  covariates   = c("tdi_cat", "p1558_i0")
)
```

Output uses `SHR` (subdistribution hazard ratio) and `SHR_label` in place of
HR, and adds `n_compete` (number of competing events in the analysis set).

---

## Working with Results

All functions return a `data.table` that feeds directly into `plot_forest()`:

```{r downstream}
res <- assoc_coxph(
  data         = dt,
  outcome_col  = "dm_status",
  time_col     = "dm_followup_years",
  exposure_col = "p20116_i0",
  covariates   = c("bmi_cat", "tdi_cat", "p1558_i0")
)

# Pass directly to plot_forest()
plot_forest(as.data.frame(res))

# Filter to a single model
res[model == "Fully adjusted"]

# Export
data.table::fwrite(res, "assoc_results.csv")
```

---

## Step 8: Lag Sensitivity Analysis

`assoc_lag()` re-runs the same Cox models at one or more lag periods to assess
whether associations are driven by early events (reverse causation or detection
bias). For each lag, participants whose follow-up time is less than
`lag_years` are excluded; follow-up is kept on its original scale.

```{r assoc-lag}
res <- assoc_lag(
  data         = dt,
  outcome_col  = "dm_status",
  time_col     = "dm_followup_years",
  exposure_col = "p20116_i0",
  lag_years    = c(0, 1, 2),   # 0 = full cohort reference
  covariates   = c("bmi_cat", "tdi_cat", "p1558_i0")
)
```

Setting `lag_years = 0` (or including `0` in the vector) runs the model on the
full unfiltered cohort, providing a reference row against which lagged results
can be compared.

Output adds two columns to the standard `assoc_coxph()` result:

| Column | Description |
|---|---|
| `lag_years` | Lag period applied (numeric, same units as `time_col`) |
| `n_excluded` | Participants excluded because follow-up < `lag_years` |

---

## Getting Help

- `?assoc_coxph`, `?assoc_logistic`, `?assoc_linear`
- `?assoc_coxph_zph`, `?assoc_subgroup`, `?assoc_trend`, `?assoc_competing`, `?assoc_lag`
- `vignette("derive-survival")` — follow-up time and event derivation
- `vignette("derive")` — disease phenotype derivation
- [GitHub Issues](https://github.com/evanbio/ukbflow/issues)

---
title: "Monitoring and Retrieving Extraction Jobs"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Monitoring and Retrieving Extraction Jobs}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = FALSE
)
```

## Overview

When `extract_batch()` submits a table-exporter job, it runs asynchronously on the RAP cloud. The `job_*` functions let you monitor progress, inspect job history, and load results once the job completes.

---

## Typical Workflow

```{r workflow}
library(ukbflow)

# 1. Submit extraction job
job_id <- extract_batch(c(31, 53, 21022, 22189), file = "ukb_demographics")

# 2. Wait for completion
job_wait(job_id)

# 3. Load result (RAP only)
df <- job_result(job_id)
```

---

## Monitoring a Job

### Check status

`job_status()` returns the current state of a job:

```{r job-status}
job_status(job_id)
#> job-XXXXXXXXXXXX
#>            done
```

Possible states:

| State | Meaning |
|---|---|
| `idle` | Queued, waiting to be scheduled |
| `runnable` | Resources being allocated |
| `running` | Actively executing |
| `done` | Completed successfully |
| `failed` | Failed — see failure message |
| `terminated` | Manually terminated |

For failed jobs, the error message is accessible via:

```{r job-failed}
s <- job_status(job_id)
if (s == "failed") cli::cli_inform(attr(s, "failure_message"))
```

### Wait for completion

`job_wait()` polls at regular intervals until the job reaches a terminal state:

```{r job-wait}
job_wait(job_id)                    # wait indefinitely (default)
job_wait(job_id, interval = 60)     # poll every 60 seconds
job_wait(job_id, timeout = 7200)    # give up after 2 hours
```

`job_wait()` stops with an error if the job fails or is terminated, so you can safely chain it with `job_result()`:

```{r job-wait-chain}
job_wait(job_id)
df <- job_result(job_id)
```

---

## Retrieving Results

### Get the file path

`job_path()` returns the `/mnt/project/` path of the output CSV on RAP:

```{r job-path}
path <- job_path(job_id)
#> "/mnt/project/results/ukb_demographics.csv"
```

Use this to read the file directly or pass it to other tools:

```{r job-path-read}
df <- data.table::fread(job_path(job_id))
```

### Load into R

`job_result()` combines `job_path()` and `fread()` in one step. Must be run inside the RAP environment:

```{r job-result}
df <- job_result(job_id)
# returns a data.table, e.g. 502353 rows x 5 cols (incl. eid)
```

---

## Browsing Job History

`job_ls()` returns a summary of recent jobs:

```{r job-ls}
job_ls()          # last 20 jobs
job_ls(n = 5)     # last 5 jobs

# Filter by state
job_ls(state = "failed")
job_ls(state = c("done", "failed"))
```

The result is a data.frame with columns:

| Column | Description |
|---|---|
| `job_id` | Job ID, e.g. `job-XXXXXXXXXXXX` |
| `name` | Job name (typically `Table exporter`) |
| `state` | Current state |
| `created` | Job creation time (`POSIXct`) |
| `runtime` | Runtime string, e.g. `0:04:36` (`NA` if still running) |

---

## Getting Help

- `?job_status`, `?job_wait`, `?job_path`, `?job_result`, `?job_ls`
- `vignette("extract")` — submitting extraction jobs
- [GitHub Issues](https://github.com/evanbio/ukbflow/issues)

---
title: "Genetic Risk Score (GRS) Analysis on UKB RAP"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Genetic Risk Score (GRS) Analysis on UKB RAP}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = FALSE
)
```

## Overview

The `grs_*` functions provide an end-to-end pipeline for computing and
validating polygenic / genetic risk scores within the UK Biobank Research
Analysis Platform (RAP). Because individual-level genotype data cannot be
downloaded locally, all computationally intensive steps are executed as
cloud jobs via the DNAnexus Swiss Army Knife app.

| Function | Where it runs | Purpose |
|---|---|---|
| `grs_check()` | Local or RAP | Validate and export a SNP weights file |
| `grs_bgen2pgen()` | Submits RAP jobs | Convert UKB imputed BGEN files to PGEN |
| `grs_score()` | Submits RAP jobs | Score genotypes across chromosomes with plink2 |
| `grs_standardize()` | Local or RAP | Z-score standardise GRS columns |
| `grs_zscore()` | Local or RAP | Alias for `grs_standardize()` |
| `grs_validate()` | Local or RAP | Assess predictive performance (OR/HR, AUC, C-index) |

**Typical pipeline:**

```
grs_check()  ->  grs_bgen2pgen()  ->  grs_score()  ->  grs_standardize()  ->  grs_validate()
```

> **Prerequisite**: you must be authenticated to UKB RAP and have a project
> selected. See `vignette("auth")`.

---

## Step 1: Validate the Weights File -- `grs_check()`

Before uploading to RAP, validate your SNP weights file. The function reads
the file, checks required columns, flags problems, and writes a
plink2-compatible space-delimited output.

**Required columns:**

| Column | Description |
|---|---|
| `snp` | SNP identifier (expected `rs` + digits format) |
| `effect_allele` | Effect allele: one of A / T / C / G |
| `beta` | Effect size (log-OR or beta); must be numeric |

```{r grs-check-local}
library(ukbflow)

# Local: weights file on your machine
w <- grs_check("weights.csv", dest = "weights_clean.txt")
#> Read 'weights.csv': 312 rows, 5 columns.
#> ✔ No NA values.
#> ✔ No duplicate SNPs.
#> ✔ All SNP IDs match rs[0-9]+ format.
#> ✔ All effect alleles are A/T/C/G.
#> Beta summary:
#>   Range     : -0.412 to 0.589
#>   Mean |beta|: 0.183
#>   Positive  : 187 (59.9%)
#>   Negative  : 125 (40.1%)
#>   Zero      : 0
#> ✔ Weights file passed checks: 312 SNPs ready for UKB RAP.
#> ✔ Saved: 'weights_clean.txt'
```

```{r grs-check-rap}
# On RAP (RStudio) -- use /mnt/project/ paths directly
w <- grs_check(
  file = "/mnt/project/weights/weights.csv",
  dest = "/mnt/project/weights/weights_clean.txt"
)
```

The validated weights are also returned invisibly for inspection.

---

## Step 2: Convert BGEN to PGEN -- `grs_bgen2pgen()`

UKB imputed genotype data is stored in BGEN format on RAP. `grs_bgen2pgen()`
submits one Swiss Army Knife job per chromosome to convert BGEN files to the
faster PGEN format with a MAF filter applied via plink2.

This function is called from your local machine or RAP RStudio -- the heavy
lifting runs entirely in the cloud.

```{r grs-bgen2pgen-test}
# Test on the smallest chromosome first
ids <- grs_bgen2pgen(chr = 22, dest = "/pgen/", priority = "high")
#> Uploading 'grs_bgen2pgen_std.R' to RAP ...
#> ✔ Uploaded: '/grs_bgen2pgen_std.R'
#> Submitting 1 job(s) -- mem2_ssd1_v2_x4 / priority: high
#> ✔ chr22 -> 'job-Gxxxxxxxxxxxxxxxxxxxxxxxx'
#> ✔ 1/1 job(s) submitted. Monitor with job_ls().
```

```{r grs-bgen2pgen-all}
# Full run: small chromosomes on standard, large on upgraded instance
ids_small <- grs_bgen2pgen(chr = 17:22, dest = "/pgen/")
ids_large <- grs_bgen2pgen(chr = 1:16,  dest = "/pgen/", instance = "large")

# Monitor progress (job_wait() takes one job ID at a time)
job_ls()
for (id in c(ids_small, ids_large)) job_wait(id)
```

**Instance types:**

| Preset | DNAnexus instance | Cores | RAM | SSD | Recommended for |
|---|---|---|---|---|---|
| `"standard"` | `mem2_ssd1_v2_x4` | 4 | 12 GB | 200 GB | chr 15–22 |
| `"large"` | `mem2_ssd2_v2_x8` | 8 | 28 GB | 640 GB | chr 1–16 |

**Key arguments:**

| Argument | Default | Description |
|---|---|---|
| `chr` | `1:22` | Chromosomes to process |
| `dest` | — | RAP output path for PGEN files (required) |
| `maf` | `0.01` | MAF filter passed to plink2 `--maf` |
| `instance` | `"standard"` | Instance type preset |
| `priority` | `"low"` | Job priority (`"low"` or `"high"`) |

> **Storage warning**: chromosomes 1–16 may exceed the 200 GB SSD on
> `"standard"` instances. Use `instance = "large"` for these.

---

## Step 3: Calculate GRS Scores -- `grs_score()`

Once PGEN files are ready, `grs_score()` uploads your weights file(s) and
submits one plink2 scoring job per GRS. Each job scores all 22 chromosomes
and saves a single CSV to RAP.

```{r grs-score}
ids <- grs_score(
  file     = c(
    grs_a = "weights/grs_a_weights.txt",
    grs_b = "weights/grs_b_weights.txt"
  ),
  pgen_dir = "/mnt/project/pgen",
  dest     = "/grs/",
  priority = "high"
)
#> ── Uploading 2 weight file(s) to RAP ────────────────────────────────────────
#> Uploading 'grs_a_weights.txt' ...
#> ✔ Uploaded: '/grs_a_weights.txt'
#> Uploading 'grs_b_weights.txt' ...
#> ✔ Uploaded: '/grs_b_weights.txt'
#> Submitting 2 job(s) -- mem2_ssd1_v2_x4 / priority: high
#> ✔ grs_a -> 'job-Gxxxxxxxxxxxxxxxxxxxxxxxx'
#> ✔ grs_b -> 'job-Gyyyyyyyyyyyyyyyyyyyyyyyy'
#> ✔ 2/2 job(s) submitted. Monitor with job_ls().

job_wait(ids)
```

When running from RAP RStudio with weights already at the project root,
the upload step is skipped automatically:

```{r grs-score-rap}
# On RAP: weights already at /mnt/project/grs_a_weights.txt
ids <- grs_score(
  file     = c(grs_a = "/mnt/project/grs_a_weights.txt"),
  pgen_dir = "/mnt/project/pgen",
  dest     = "/grs/"
)
#> ℹ grs_a_weights.txt already at RAP root, skipping upload.
```

> **Important**: the `maf` argument must match the value used in
> `grs_bgen2pgen()` so that the correct PGEN files are located.

Output per job: `/grs/<GRS_name>_scores.csv` with columns `IID` and the GRS
score named `GRS_<name>`.

---

## Step 4: Standardise GRS Columns -- `grs_standardize()`

After downloading the score CSVs from RAP (via `fetch_file()`) and merging
them into your analysis dataset, Z-score standardise each GRS column so that
effect estimates are interpretable as per-SD associations.

```{r grs-standardize}
# Auto-detect all columns containing "grs" (case-insensitive)
cohort <- grs_standardize(cohort)
#> Auto-detected 2 GRS column(s): 'GRS_a', 'GRS_b'
#> ✔ GRS_a -> GRS_a_z  [mean=0.0000, sd=1.0000]
#> ✔ GRS_b -> GRS_b_z  [mean=0.0000, sd=1.0000]
```

```{r grs-standardize-explicit}
# Or specify columns explicitly
cohort <- grs_standardize(cohort, grs_cols = c("GRS_a", "GRS_b"))
```

`grs_zscore()` is an alias and produces an identical result.

The original columns are preserved; `_z` columns are inserted immediately
after their source column.

---

## Step 5: Validate Predictive Performance -- `grs_validate()`

`grs_validate()` runs four complementary validation analyses for each GRS:

| Analysis | What it measures |
|---|---|
| **Per SD** | OR (logistic) or HR (Cox) per 1-SD increase in GRS |
| **High vs Low** | OR / HR comparing top 20% vs bottom 20% |
| **Trend test** | P-trend across Q1–Q4 quartiles |
| **Discrimination** | AUC (logistic) or C-index (Cox) with 95% CI |

### Logistic (cross-sectional)

```{r grs-validate-logistic}
res <- grs_validate(
  data        = cohort,
  grs_cols    = c("GRS_a_z", "GRS_b_z"),
  outcome_col = "outcome"
)
#> ── Creating GRS groups ───────────────────────────────────────────────────────
#> ── Effect per SD (OR) ───────────────────────────────────────────────────────
#> ── High vs Low ──────────────────────────────────────────────────────────────
#> ── Trend test ───────────────────────────────────────────────────────────────
#> ── AUC ──────────────────────────────────────────────────────────────────────
#> ✔ Validation complete.

res$per_sd
res$discrimination
```

### Cox (survival)

```{r grs-validate-cox}
res <- grs_validate(
  data        = cohort,
  grs_cols    = c("GRS_a_z", "GRS_b_z"),
  outcome_col = "outcome",
  time_col    = "followup_years",
  covariates  = c("age", "sex", paste0("pc", 1:10))
)

res$per_sd          # HR per SD x model
res$high_vs_low     # HR: top 20% vs bottom 20%
res$trend           # p-trend across Q1–Q4
res$discrimination  # C-index with 95% CI
```

**Return value:** a named list with four `data.table` elements.

| Element | Columns (logistic) | Columns (Cox) |
|---|---|---|
| `per_sd` | `exposure`, `term`, `model`, `OR`, `CI_lower`, `CI_upper`, `p_value` | same with `HR` |
| `high_vs_low` | same as `per_sd` | same with `HR` |
| `trend` | `exposure`, `model`, `p_trend` | same |
| `discrimination` | `GRS`, `AUC`, `CI_lower`, `CI_upper` | `GRS`, `C_index`, `CI_lower`, `CI_upper` |

> AUC calculation requires the `pROC` package.
> Install with `install.packages("pROC")` if needed.

---

## Complete Pipeline Example

```{r grs-full-pipeline}
library(ukbflow)

# 1. Validate weights (local)
grs_check("weights.csv", dest = "weights_clean.txt")

# 2. Convert BGEN -> PGEN on RAP (submit jobs)
ids_std  <- grs_bgen2pgen(chr = 17:22, dest = "/pgen/", maf = 0.01)
ids_lrg  <- grs_bgen2pgen(chr = 1:16,  dest = "/pgen/", maf = 0.01, instance = "large")
for (id in c(ids_std, ids_lrg)) job_wait(id)

# 3. Score GRS on RAP (submit jobs)
score_ids <- grs_score(
  file     = c(grs_a = "weights_clean.txt"),
  pgen_dir = "/mnt/project/pgen",
  maf      = 0.01,   # must match grs_bgen2pgen()
  dest     = "/grs/"
)
job_wait(score_ids)

# 4. Download score CSV from RAP
fetch_file("/grs/GRS_a_scores.csv", dest = "GRS_a_scores.csv")

# 5. Merge into analysis dataset and standardise
# cohort: your analysis data.table with IID and phenotype columns
scores <- data.table::fread("GRS_a_scores.csv")   # columns: IID, GRS_a
cohort <- scores[cohort, on = "IID"]               # right-join: keep all cohort rows
cohort <- grs_standardize(cohort, grs_cols = "GRS_a")

# 6. Validate
res <- grs_validate(
  data        = cohort,
  grs_cols    = "GRS_a_z",
  outcome_col = "outcome",
  time_col    = "followup_years",
  covariates  = c("age", "sex", paste0("pc", 1:10))
)

res$per_sd
res$discrimination
```

---

## Getting Help

- `?grs_check`, `?grs_bgen2pgen`, `?grs_score`
- `?grs_standardize`, `?grs_validate`
- `vignette("auth")` -- RAP authentication and project selection
- `vignette("job")` -- monitoring and managing RAP jobs
- `vignette("assoc")` -- association analysis functions used inside `grs_validate()`
- [GitHub Issues](https://github.com/evanbio/ukbflow/issues)

---
title: "Smoking and Lung Cancer Risk: A Complete Analysis Workflow"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Smoking and Lung Cancer Risk: A Complete Analysis Workflow}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = FALSE
)
```

## 1 Introduction

Cigarette smoking is the leading preventable cause of lung cancer,
accounting for approximately 85% of all cases and conferring a 15–30
times higher risk compared with never-smokers. Risk scales with
cumulative exposure and decreases — but never fully reverts — after
cessation. Sex, age, and socioeconomic deprivation are well-established
modifiers of this association in large prospective cohorts such as the
UK Biobank. This vignette uses a simulated UK Biobank-style dataset to
walk through the complete ukbflow pipeline for this canonical exposure–
outcome pair.

## 2 Data Loading

We use the package built-in function `ops_toy()` to simulate and generate
a synthetic UK Biobank-style dataset for this analysis.

```{r load-data}
library(ukbflow)

data <- ops_toy(n = 500000, seed = 2026)
#> ✔ ops_toy: 500000 participants | 75 columns | scenario = "cohort" | seed = 2026
```

## 3 Decode Column Names

In a real RAP session, `decode_names()` translates UKB field IDs into
human-readable snake_case column names (e.g. `p31` → `sex`,
`p21022` → `age_at_recruitment`).

```{r decode-names}
data <- decode_names(data)
#> Field dictionary not cached - calling `extract_ls()` to populate it.
#> Using dataset: "XXXXXXXXXXXXXXXXXXXXXXX"
#> Fetching approved fields... (cached after first call)
#> 29951 fields available. Assign to a variable or use pattern= to search.
#> ✔ Renamed 68 columns.
#> ! 5 column names longer than 60 characters - consider renaming manually:
#> • interpolated_year_when_non_cancer_illness_first_diagnosed_i0_a0
#> • interpolated_year_when_non_cancer_illness_first_diagnosed_i0_a1
#> • interpolated_year_when_non_cancer_illness_first_diagnosed_i0_a2
#> • interpolated_year_when_non_cancer_illness_first_diagnosed_i0_a3
#> • interpolated_year_when_non_cancer_illness_first_diagnosed_i0_a4
```

## 4 Derive Phenotypes

We first handle non-informative missing codes (e.g. "Do not know",
"Prefer not to answer") with `derive_missing()`.

```{r derive-missing}
data <- derive_missing(data)
#> ✔ derive_missing: replaced 279650 values across 3 columns (action = "na").
```

Next, convert categorical columns to factors with `derive_covariate()`.

```{r derive-covariate}
data <- derive_covariate(
  data,
  as_factor = c(
    "p31",        # sex
    "p20116_i0",  # smoking_status_i0
    "p1558_i0",   # alcohol_intake_frequency_i0
    "p54_i0"      # uk_biobank_assessment_centre_i0
  )
)
#> ── Factor ───────────────────────────────────────────────────────────────────
#> sex [2 levels]
#>   Female: n=270191 (54%)  |  Male: n=229809 (46%)  |  <NA>: n=0 (0%)
#> smoking_status_i0 [3 levels]
#>   Current: n=69867 (14%)  |  Never: n=260458 (52.1%)  |  Previous: n=154871 (31%)  |  <NA>: n=14804 (3%)
#> ! alcohol_intake_frequency_i0: 6 levels > max_levels (5), consider collapsing categories.
#>   Daily or almost daily: n=40113 (8%)  |  Never: n=45207 (9%)  |  ...  |  <NA>: n=14973 (3%)
#> ! uk_biobank_assessment_centre_i0: 10 levels > max_levels (5), consider collapsing categories.
#>   Birmingham: n=49709 (9.9%)  |  Bristol: n=49832 (10%)  |  ...  |  Sheffield: n=50293 (10.1%)
```

We then bin BMI and Townsend Deprivation Index (TDI) into categories
using `derive_cut()`.

```{r derive-cut}
data <- derive_cut(
  data,
  col    = "p21001_i0",                              # body_mass_index_bmi_i0
  n      = 4,
  breaks = c(18.5, 25, 30),
  labels = c("Underweight", "Normal", "Overweight", "Obese"),
  name   = "bmi_cat"
)
#> ── Source: body_mass_index_bmi_i0 ───────────────────────────────────────────
#>   mean=26.21, median=26.19, sd=5.48, Q1=22.48, Q3=29.9, NA=0% (n=0)
#> ── New column: bmi_cat ──────────────────────────────────────────────────────
#> bmi_cat [4 levels]
#>   Underweight: n=40238 (8%)  |  Normal: n=166710 (33.3%)
#>   Overweight:  n=170794 (34.2%)  |  Obese: n=122258 (24.5%)  |  <NA>: n=0 (0%)

data <- derive_cut(
  data,
  col    = "p22189",                                 # townsend_deprivation_index_at_recruitment
  n      = 4,
  labels = c("Q1 (least deprived)", "Q2", "Q3", "Q4 (most deprived)"),
  name   = "tdi_cat"
)
#> ── Source: townsend_deprivation_index_at_recruitment ────────────────────────
#>   mean=-1.25, median=-1.3, sd=3.1, Q1=-3.46, Q3=0.86, NA=0% (n=0)
#> ── New column: tdi_cat ──────────────────────────────────────────────────────
#> tdi_cat [4 levels]
#>   Q1 (least deprived): n=125290 (25.1%)  |  Q2: n=124831 (25%)
#>   Q3: n=125102 (25%)  |  Q4 (most deprived): n=124777 (25%)  |  <NA>: n=0 (0%)
```

Self-reported lung cancer (field 20001) is derived with `derive_selfreport()`,
which searches the cancer self-report columns for a matching label.

```{r derive-selfreport}
data <- derive_selfreport(
  data,
  name  = "lung_cancer",
  regex = "lung cancer",
  field = "cancer"
)
#> ✔ derive_selfreport (lung_cancer): 6700 cases, 6700 with dates.
```

ICD-10 diagnoses in the UK Biobank can be ascertained from four sources:
First Occurrence fields, HES inpatient records, death registry, and cancer
registry. For malignant neoplasms such as lung cancer, the cancer registry
provides the most complete and accurate ascertainment and is therefore our
primary source.

```{r derive-icd10}
data <- derive_icd10(
  data,
  name      = "lung",
  icd10     = "^C3[34]",
  match     = "regex",
  source    = "cancer_registry",
  behaviour = 3L
)
#> ✔ derive_cancer_registry (lung): 2449 cases, 2449 with date.
#> ✔ derive_icd10 (lung): 2449 cases across 1 source, 2449 with date.
```

`derive_case()` merges the self-report and ICD-10 flags into a single case
status and earliest date using an OR rule across sources.

```{r derive-case}
data <- derive_case(
  data,
  name                = "lung",
  selfreport_col      = "lung_cancer_selfreport",
  selfreport_date_col = "lung_cancer_selfreport_date"
)
#> ✔ derive_case (lung): 9110 cases, 9110 with date.
#> ℹ   Both sources (lung_icd10 & lung_cancer_selfreport): 39
```

Finally, `derive_timing()` classifies each case as prevalent (disease before
baseline) or incident (disease after baseline), which determines eligibility
for the survival analysis.

```{r derive-timing}
data <- derive_timing(data, name = "lung", baseline_col = "p53_i0")  # date_of_attending_assessment_centre_i0
#> ✔ derive_timing (lung_timing):
#> ℹ   0 (no disease): 490890
#> ℹ   1 (prevalent):  5056
#> ℹ   2 (incident):   4054
#> ℹ   NA (no date):   0
```

`derive_followup()` computes follow-up end date and time in years for each
participant, taking the earliest of the event date, death date, and the
administrative censoring date.

```{r derive-followup}
data <- derive_followup(
  data,
  name         = "lung",
  event_col    = "lung_date",
  baseline_col = "p53_i0",         # date_of_attending_assessment_centre_i0
  censor_date  = as.Date("2022-10-31"),
  death_col    = "p40000_i0",      # date_of_death_i0
  lost_col     = FALSE             # lost-to-follow-up not available in this dataset
)
#> ✔ derive_followup (lung):
#> ℹ   lung_followup_end: 500000 / 500000 non-missing
#> ℹ   lung_followup_years: mean=13.71, median=14.09, range=[0, 16.83]
```

## 5 Exposure Definition

Smoking status has three levels (Never, Previous, Current). For a
cleaner binary contrast, we collapse Previous and Current into a single
"Ever" category, with "Never" as the reference level.

```{r exposure}
data[, smoking_ever := factor(
  ifelse(p20116_i0 == "Never", "Never", "Ever"),
  levels = c("Never", "Ever")   # Never = reference
)]
```

## 6 Cohort Assembly

We exclude prevalent lung cancer cases and participants with missing
exposure or follow-up time to arrive at the final analysis cohort.
`ops_snapshot()` is a handy utility for recording cohort size at each
step — the snapshot history can directly inform the flow diagram in your
methods section.

```{r snapshot-raw}
ops_snapshot(data, label = "raw")
#> ── snapshot: raw ────────────────────────────────────────────────────────────
#>   rows      500,000
#>   cols           89
#>   NA cols        59
#>   size       293 MB
```

Exclude participants with prevalent lung cancer (diagnosed at or before
baseline) — they are not eligible for the incident analysis.

```{r excl-prevalent}
data <- data[lung_timing != 1 | is.na(lung_timing)]
ops_snapshot(data, label = "after excluding prevalent cases")
#> ── snapshot: after excluding prevalent cases ─────────────────────────────────
#>   rows      494,944  (-5,056)
#>   cols           89  (= 0)
#>   NA cols        58  (-1)
#>   size       290.1 MB  (-2.95 MB)
```

Exclude participants with missing values in the exposure or any covariate.

```{r excl-na}
data <- data[!is.na(smoking_ever)  &
             !is.na(p31)           &   # sex
             !is.na(p21022)        &   # age_at_recruitment
             !is.na(bmi_cat)       &   # bmi category
             !is.na(p1558_i0)      &   # alcohol_intake_frequency_i0
             !is.na(tdi_cat)       &   # townsend deprivation category
             !is.na(p54_i0)        &   # assessment_centre
             !is.na(p22009_a1)     &   # PC1
             !is.na(p22009_a2)     &   # PC2
             !is.na(p22009_a3)     &   # PC3
             !is.na(p22009_a4)     &   # PC4
             !is.na(p22009_a5)     &   # PC5
             !is.na(p22009_a6)     &   # PC6
             !is.na(p22009_a7)     &   # PC7
             !is.na(p22009_a8)     &   # PC8
             !is.na(p22009_a9)     &   # PC9
             !is.na(p22009_a10)]       # PC10
ops_snapshot(data, label = "after excluding missing covariates")
#> ── snapshot: after excluding missing covariates ──────────────────────────────
#>   rows      465,937  (-29,007)
#>   cols           89  (= 0)
#>   NA cols        55  (-3)
#>   size       273.3 MB  (-16.75 MB)
```

Review the full exclusion history.

```{r snapshot-history}
ops_snapshot()
#> ── ops_snapshot history ─────────────────────────────────────────────────────
#>    idx                              label timestamp   nrow  ncol n_na_cols size_mb
#>  1:  1                                raw  15:35:24 500000    89        59  293.03
#>  2:  2    after excluding prevalent cases  15:35:51 494944    89        58  290.08
#>  3:  3 after excluding missing covariates  15:36:18 465937    89        55  273.33
```

Before running the association analysis, we take a quick look at the
final cohort — exposure distribution, outcome ascertainment, and
follow-up time.

```{r cohort-overview}
# Exposure distribution
data[, .N, by = smoking_ever]

#>    smoking_ever      N
#>          <fctr>  <int>
#>  1:        Ever 215893
#>  2:       Never 250044

# Outcome: incident cases and timing
data[, .N, by = lung_timing]
#>    lung_timing      N
#>          <int>  <int>
#>  1:          0 462110
#>  2:          2   3827

# Follow-up time in years
data[, .(mean   = round(mean(lung_followup_years),   2),
         median = round(median(lung_followup_years), 2),
         min    = round(min(lung_followup_years),    2),
         max    = round(max(lung_followup_years),    2))]
#>     mean median min    max
#>    <num>  <num> <num> <num>
#>  1: 13.71  14.09     0 16.83
```

## 7 Association Analysis

We fit a Cox proportional hazards model with `assoc_coxph()`. The function
automatically produces three adjustment levels: unadjusted, age and sex
adjusted, and fully adjusted.

```{r assoc-coxph}
res <- assoc_coxph(
  data         = data,
  outcome_col  = "lung_status",
  time_col     = "lung_followup_years",
  exposure_col = "smoking_ever",
  covariates   = c("p21022",              # age_at_recruitment
                   "p31",                 # sex
                   "bmi_cat",
                   "tdi_cat",
                   "p1558_i0",            # alcohol_intake_frequency_i0
                   "p54_i0",              # uk_biobank_assessment_centre_i0
                   paste0("p22009_a", 1:10))  # genetic PCs 1-10
)
#> ℹ outcome_col lung_status: logical detected, converting TRUE/FALSE -> 1/0
#> ── assoc_coxph ──────────────────────────────────────────────────────────────
#> ℹ 1 exposure x 3 models = 3 Cox regressions
#> ℹ Input cohort: 465937 participants (n/n_events/person_years reflect each model's actual analysis set)
#> ── smoking_ever ─────────────────────────────────────────────────────────────
#>   ✔ Unadjusted             | smoking_everEver: HR 0.99 (0.93-1.06), p = 0.834
#>   ✔ Age and sex adjusted   | smoking_everEver: HR 0.99 (0.93-1.06), p = 0.838
#>   ✔ Fully adjusted         | smoking_everEver: HR 0.99 (0.93-1.06), p = 0.829
#> ✔ Done: 3 result rows across 1 exposure and 3 models.
```

The results reflect the absence of a built-in exposure–outcome relationship
in the simulated data. A real UK Biobank analysis may yield substantially
different estimates.

## 8 Visualisation

We first inspect the result table returned by `assoc_coxph()`.

```{r res-print}
print(res)
#>        exposure             term                model      n n_events person_years        HR  CI_lower
#>          <char>           <char>                <ord>  <int>    <num>        <num>     <num>     <num>
#>  1: smoking_ever smoking_everEver           Unadjusted 465937     3827      6388195 0.9932076 0.9320517
#>  2: smoking_ever smoking_everEver Age and sex adjusted 465937     3827      6388195 0.9933735 0.9322072
#>  3: smoking_ever smoking_everEver       Fully adjusted 465937     3827      6388195 0.9930029 0.9318576
#>     CI_upper   p_value         HR_label
#>        <num>     <num>           <char>
#>  1: 1.058376 0.8335146 0.99 (0.93-1.06)
#>  2: 1.058553 0.8375372 0.99 (0.93-1.06)
#>  3: 1.058160 0.8285626 0.99 (0.93-1.06)

class(res)
#> [1] "data.table" "data.frame"
```

`plot_forest()` relies on dplyr internally, so we convert the result to a
plain data.frame first.

```{r res-df}
res_df <- as.data.frame(res)
```

`res_df` can be used directly with `plot_forest()` to produce a forest plot.

```{r forest-plot}
p <- plot_forest(
  data       = res_df,
  est        = res_df$HR,
  lower      = res_df$CI_lower,
  upper      = res_df$CI_upper,
  ci_column  = 2L,          # forest plot rendered in column 2 (HR_label)
  p_cols     = "p_value",
  ref_line   = 1,
  xlim       = c(0, 2.0),
  ticks_at   = c(0, 0.5, 1.0, 1.5, 2.0)
)

plot(p)
```

Or we can reshape `res_df` or add equal-length vectors to produce a
publication-ready figure.

```{r forest-plot-pub}
p2 <- plot_forest(
  data       = res_df[, c("model", "n_events", "n", "p_value")],
  est        = res_df$HR,
  lower      = res_df$CI_lower,
  upper      = res_df$CI_upper,
  ci_column  = 4L,
  p_cols     = "p_value",
  ref_line   = 1,
  xlim       = c(0, 2.0),
  ticks_at   = c(0.5, 0.75, 1.0, 1.25, 1.5),
  header     = c("Model", "Cases", "N", "", "HR (95% CI)", "P value")
)

plot(p2)
```

To add a header row with an exposure label, prepend a row with `NA` estimates.

```{r forest-plot-header}
res_pub <- rbind(
  data.frame(model = "Ever vs. Never", HR_label = "", p_value = NA,
             HR = NA, CI_lower = NA, CI_upper = NA, stringsAsFactors = FALSE),
  res_df[, c("model", "HR_label", "p_value", "HR", "CI_lower", "CI_upper")]
)

p3 <- plot_forest(
  data       = res_pub[, c("model", "p_value")],
  est        = res_pub$HR,
  lower      = res_pub$CI_lower,
  upper      = res_pub$CI_upper,
  ci_column  = 2L,
  p_cols     = "p_value",
  ref_line   = 1,
  xlim       = c(0, 2.0),
  ticks_at   = c(0, 0.5, 1.0, 1.5, 2.0),
  indent     = c(0L, 1L, 1L, 1L),
  bold_label = c(TRUE, FALSE, FALSE, FALSE),
  header     = c("Model", "", "HR (95% CI)", "P value")
)

plot(p3)
```

For further customisation — colours, font sizes, borders, arrow labels,
and saving to file — see `?plot_forest`.

> **Acknowledgement**: Forest plots are one of the most widely used methods
> in epidemiology for presenting effect estimates and their uncertainty.
> `plot_forest()` is powered by the
> [forestploter](https://cran.r-project.org/package=forestploter) package.
> We thank its author for the excellent work.

The package also provides `plot_tableone()` for generating publication-ready
baseline characteristic tables. Here we show a demo — for more advanced
usage see `?plot_tableone`.

```{r tableone}
t1 <- plot_tableone(
  data    = as.data.frame(data),
  vars    = c("p21022", "p31", "bmi_cat", "tdi_cat", "p1558_i0"),
  strata  = "smoking_ever",
  label   = list(
    p21022   ~ "Age at recruitment (years)",
    p31      ~ "Sex",
    bmi_cat  ~ "BMI category",
    tdi_cat  ~ "Townsend deprivation index",
    p1558_i0 ~ "Alcohol intake frequency"
  ),
  add_p   = TRUE,
  save    = FALSE
)

print(t1)
```

## Getting Help

- `?ops_toy`, `?ops_snapshot`, `?ops_na`
- `?derive_missing`, `?derive_covariate`, `?derive_cut`, `?derive_selfreport`
- `?derive_icd10`, `?derive_case`, `?derive_timing`, `?derive_followup`
- `?assoc_coxph`, `?plot_forest`, `?plot_tableone`
- `vignette("derive")`, `vignette("derive-survival")`, `vignette("assoc")`, `vignette("plot")`
- [GitHub Issues](https://github.com/evanbio/ukbflow/issues)

## 9 Session Info

<details>
<summary>Session Info</summary>

```{r session-info}
sessionInfo()
#> R version 4.5.1 (2025-06-13 ucrt)
#> Platform: x86_64-w64-mingw32/x64
#> Running under: Windows 11 x64 (build 26200)
#>
#> Matrix products: default
#>   LAPACK version 3.12.1
#>
#> locale:
#> [1] LC_COLLATE=Chinese (Simplified)_China.utf8  LC_CTYPE=Chinese (Simplified)_China.utf8
#> [3] LC_MONETARY=Chinese (Simplified)_China.utf8 LC_NUMERIC=C
#> [5] LC_TIME=Chinese (Simplified)_China.utf8
#>
#> time zone: Asia/Shanghai
#> tzcode source: internal
#>
#> attached base packages:
#> [1] stats     graphics  grDevices utils     datasets  methods   base
#>
#> other attached packages:
#> [1] ukbflow_0.3.0  testthat_3.2.3
#>
#> loaded via a namespace (and not attached):
#>  [1] gt_1.0.0            sass_0.4.10         tidyr_1.3.1         generics_0.1.4      gtsummary_2.4.0
#>  [6] xml2_1.3.8          lattice_0.22-7      digest_0.6.37       magrittr_2.0.3      evaluate_1.0.5
#> [11] grid_4.5.1          cards_0.7.0         pkgload_1.4.0       fastmap_1.2.0       rprojroot_2.1.1
#> [16] jsonlite_2.0.0      Matrix_1.7-3        processx_3.8.6      pkgbuild_1.4.8      sessioninfo_1.2.3
#> [21] backports_1.5.0     cardx_0.3.0         brio_1.1.5          survival_3.8-3      ps_1.9.1
#> [26] gridExtra_2.3       purrr_1.0.4         cli_3.6.4           forestploter_1.1.3  rlang_1.1.6
#> [31] litedown_0.7        commonmark_2.0.0    ellipsis_0.3.2      splines_4.5.1       remotes_2.5.0
#> [36] withr_3.0.2         cachem_1.1.0        devtools_2.4.5.9000 tools_4.5.1         memoise_2.0.1
#> [41] dplyr_1.1.4.9000    broom_1.0.10        curl_7.0.0          vctrs_0.6.5         R6_2.6.1
#> [46] lifecycle_1.0.4     fs_1.6.6            usethis_3.2.1       pkgconfig_2.0.3     desc_1.4.3
#> [51] pillar_1.11.1       gtable_0.3.6        data.table_1.17.0   glue_1.8.0          xfun_0.52
#> [56] tibble_3.2.1        tidyselect_1.2.1    rstudioapi_0.17.1   knitr_1.50          htmltools_0.5.8.1
#> [61] rmarkdown_2.29      compiler_4.5.1      markdown_2.0
```

</details>

## 10 References

- Xu C (2023). *forestploter: Create Flexible Forest Plot*.
  R package version 1.1.3.
  <https://CRAN.R-project.org/package=forestploter>

- Sjoberg DD, Whiting K, Curry M, Lavery JA, Larmarange J (2021).
  Reproducible Summary Tables with the gtsummary Package.
  *The R Journal*, 13(1), 570–580.
  <https://doi.org/10.32614/RJ-2021-053>

- Iannone R, Cheng J, Schloerke B, Haughton S, Hughes E, Lauer A,
  François R, Seo J, Brevoort K, Roy O (2026).
  *gt: Easily Create Presentation-Ready Display Tables*.
  R package version 1.3.0.9000.
  <https://gt.rstudio.com>

---
title: "Publication-Ready Visualisation"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Publication-Ready Visualisation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = FALSE
)
```

## Overview

Two functions produce publication-ready figures and tables with minimal
post-processing:

| Function | Output | Typical use |
|---|---|---|
| `plot_forest()` | Forest plot (PNG / PDF / JPG / TIFF) | Regression results from `assoc_*()` |
| `plot_tableone()` | Table 1 (DOCX / HTML / PDF / PNG) | Baseline characteristics |

When `save = TRUE`, both functions write all supported formats in a single call
and return the plot/table object invisibly for further customisation.

---

## `plot_forest()` — Forest Plot

### Minimal example

`plot_forest()` takes a data frame whose **first column** is the row label,
plus any additional display columns. The CI graphic and formatted
`OR (95% CI)` text column are inserted automatically.

```{r forest-minimal}
library(ukbflow)

df <- data.frame(
  item      = c("Exposure vs. control", "Unadjusted", "Fully adjusted"),
  `Cases/N` = c("", "89 / 4 521", "89 / 4 521"),
  p_value   = c(NA_real_, 0.001, 0.006),
  check.names = FALSE
)

p <- plot_forest(
  data      = df,
  est       = c(NA,   1.52, 1.43),
  lower     = c(NA,   1.18, 1.11),
  upper     = c(NA,   1.96, 1.85),
  ci_column = 3L,
  indent    = c(0L,   1L,   1L),
  p_cols    = "p_value",
  xlim      = c(0.5,  3.0)
)
plot(p)
```

### Building the input data frame from `assoc_*()` results

The output of `assoc_coxph()` (and siblings) can be reshaped directly into
the format expected by `plot_forest()`:

```{r forest-from-assoc}
dt  <- ops_toy(scenario = "association")
dt  <- dt[dm_timing != 1L]

res <- assoc_coxph(
  data         = dt,
  outcome_col  = "dm_status",
  time_col     = "dm_followup_years",
  exposure_col = "p20116_i0",
  covariates   = c("bmi_cat", "tdi_cat", "p1558_i0")
)
res <- as.data.frame(res)

# Reshape: one row per model, label column first
df2 <- data.frame(
  item    = c("Smoking status", as.character(res$model)),
  `N`     = c("", paste0(res$n, " / ", res$n_events)),
  p_value = c(NA_real_, res$p_value),
  check.names = FALSE
)

p <- plot_forest(
  data      = df2,
  est       = c(NA,   res$HR),
  lower     = c(NA,   res$CI_lower),
  upper     = c(NA,   res$CI_upper),
  ci_column = 3L,
  indent    = c(0L,   rep(1L, nrow(res))),
  p_cols    = "p_value",
  xlim      = c(0.5,  2.5)
)
plot(p)
```

### Key parameters

**CI appearance**

```{r forest-ci}
# uses df, est, lower, upper from the minimal example above
p <- plot_forest(
  data      = df,
  est       = est, lower = lower, upper = upper,
  ci_column = 3L,
  ci_col    = c("grey50", "steelblue", "steelblue"),  # per-row colours
  ci_sizes  = 0.5,       # point size
  ci_Theight = 0.15,     # cap height
  ref_line  = 1,         # reference line (use 0 for beta coefficients)
  xlim      = c(0.2, 5), ticks_at = c(0.5, 1, 2, 3)
)
```

**Row labels and indentation**

```{r forest-indent}
# indent = 0 → bold parent row; indent >= 1 → indented sub-row (plain)
p <- plot_forest(
  data       = df,
  est        = est, lower = lower, upper = upper,
  ci_column  = 3L,
  indent     = c(0L, 1L, 1L),        # parent + 2 sub-rows
  bold_label = c(TRUE, FALSE, FALSE)  # explicit control (overrides indent default)
)
```

**P-value formatting**

```{r forest-pval}
# p_cols: column names in data that contain raw numeric p-values.
# Values < 10^(-p_digits) are displayed as e.g. "<0.001".
# bold_p = TRUE bolds all p < p_threshold (default 0.05).
p <- plot_forest(
  data        = df,
  est         = est, lower = lower, upper = upper,
  ci_column   = 3L,
  p_cols      = "p_value",
  p_digits    = 3L,
  bold_p      = TRUE,
  p_threshold = 0.05
)
```

**Column headers and alignment**

`header` renames all columns in the *final* rendered table.  The final table
always has `ncol(data) + 2` columns: the original columns, plus the `gap_ci`
graphic column and the auto-generated `OR (95% CI)` text column.  Pass `""`
for the gap column position.

```{r forest-header}
# data has 3 columns → final table has 5 columns (original 3 + gap_ci + OR label)
# Layout with ci_column = 3L: item | Cases/N | gap_ci | OR (95% CI) | p_value
p <- plot_forest(
  data      = df,
  est       = est, lower = lower, upper = upper,
  ci_column = 3L,
  header    = c("Comparison", "Cases / N", "", "HR (95% CI)", "P-value")
  #             col 1          col 2        gap  OR label       col 5
)
```

`align` controls per-column text alignment across all `ncol(data) + 2`
columns: `-1` = left, `0` = centre, `1` = right.  `NULL` (default) left-aligns
column 1 and centres the rest.

```{r forest-align}
p <- plot_forest(
  data      = df,
  est       = est, lower = lower, upper = upper,
  ci_column = 3L,
  align     = c(-1L, 0L, 0L, 0L, 1L)   # label left | Cases/N centre | gap | OR centre | p right
)
```

**Background and borders**

```{r forest-style}
p <- plot_forest(
  data       = df,
  est        = est, lower = lower, upper = upper,
  ci_column  = 3L,
  background = "zebra",       # "zebra" | "bold_label" | "none"
  bg_col     = "#F0F0F0",     # shading colour
  border     = "three_line",  # "three_line" | "none"
  border_width = 3            # scalar or length-3 vector (top / mid / bottom)
)
```

**Layout and saving**

```{r forest-save}
# uses df, est, lower, upper from the minimal example above
p <- plot_forest(
  data        = df,
  est         = est, lower = lower, upper = upper,
  ci_column   = 3L,
  row_height  = NULL,   # auto (8 / 12 / 10 / 15 mm); or scalar/vector
  col_width   = NULL,   # auto (rounds up to nearest 5 mm)
  save        = TRUE,
  dest        = "forest_main",   # extension ignored; all 4 formats saved
  save_width  = 20,              # cm
  save_height = NULL             # auto: nrow(data) * 0.9 + 3 cm
)
```

> All four formats (PNG, PDF, JPG, TIFF) are written at **300 dpi** with a
> white background. The function returns the plot object invisibly; display
> with `plot(p)` or `grid::grid.draw(p)`.

---

## `plot_tableone()` — Baseline Characteristics Table

### Minimal example

```{r tableone-minimal}
library(gtsummary)
data(trial)   # built-in gtsummary dataset

plot_tableone(
  data   = trial,
  vars   = c("age", "marker", "grade"),
  strata = "trt",
  save   = FALSE
)
```

### With SMD, custom labels, and export

```{r tableone-full}
plot_tableone(
  data    = trial,
  vars    = c("age", "marker", "grade", "stage"),
  strata  = "trt",
  label   = list(age ~ "Age (years)", marker ~ "Marker level (ng/mL)"),
  add_p   = TRUE,    # Wilcoxon / chi-squared p-values; formatted as <0.001
  add_smd = TRUE,
  overall = TRUE,
  dest    = "table1",
  save    = TRUE
)
```

### Key parameters

**Variable types and statistics**

```{r tableone-types}
dt <- as.data.frame(ops_toy(scenario = "association"))

plot_tableone(
  data      = dt,
  vars      = c("p21022", "p21001_i0", "p31", "p20116_i0"),
  strata    = "dm_status",
  type      = list(p21022 = "continuous2"),   # show median + IQR
  statistic = list(
    all_continuous()  ~ "{mean} ({sd})",
    all_categorical() ~ "{n} ({p}%)"
  ),
  digits    = list(p21022 ~ 1, p21001_i0 ~ 1),
  missing   = "ifany",   # show missing counts when present
  save      = FALSE
)
```

**SMD column**

The SMD column summarises covariate balance between groups:
- Continuous variables: Cohen's *d* (pooled-SD formula)
- Categorical variables: RMSD of group proportions

```{r tableone-smd}
plot_tableone(
  data    = dt,
  vars    = c("p21022", "p21001_i0", "p31"),
  strata  = "dm_status",
  add_smd = TRUE,
  save    = FALSE
)
```

**Excluding rows**

Use `exclude_labels` to remove specific level rows from the rendered table
(e.g. a redundant reference category or an "Unknown" level):

```{r tableone-exclude}
plot_tableone(
  data           = dt,
  vars           = c("p31", "p20116_i0"),
  strata         = "dm_status",
  exclude_labels = "Never",   # e.g. remove reference category from display
  save           = FALSE
)
```

**Export formats**

When `save = TRUE`, four files are written simultaneously:

| Format | Tool | Notes |
|---|---|---|
| `.docx` | `gt::gtsave()` | Ready for Word submission |
| `.html` | `gt::gtsave()` | Interactive preview |
| `.pdf` | `pagedown::chrome_print()` | Requires Chrome / Chromium |
| `.png` | `webshot2::webshot()` | 2x zoom, table element only |

> PDF and PNG rendering requires `pagedown` and `webshot2` respectively.
> Install with `install.packages(c("pagedown", "webshot2"))`.

---

## Getting Help

- `?plot_forest`, `?plot_tableone`
- `vignette("assoc")` — association analysis producing forest plot inputs
- [GitHub Issues](https://github.com/evanbio/ukbflow/issues)

---
title: "Deriving Disease Phenotypes from UKB Data"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Deriving Disease Phenotypes from UKB Data}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = FALSE
)
```

## Overview

The `derive_*` functions convert raw UKB columns into analysis-ready variables.
This vignette covers the disease phenotype derivation pipeline:

| Step | Function(s) | Purpose |
|---|---|---|
| 1 | `derive_missing()` | Handle "Do not know" / "Prefer not to answer" |
| 2 | `derive_covariate()` | Convert types; summarise covariates |
| 3 | `derive_cut()` | Bin continuous variables into groups |
| 4 | `derive_selfreport()` | Self-reported disease status + date |
| 5 | `derive_hes()` | HES inpatient ICD-10 status + date |
| 6 | `derive_first_occurrence()` | First Occurrence field status + date |
| 7 | `derive_cancer_registry()` | Cancer registry status + date |
| 8 | `derive_death_registry()` | Death registry ICD-10 status + date |
| 9 | `derive_icd10()` | Combine any subset of sources (wrapper) |
| 10 | `derive_case()` | Merge self-report + ICD-10 into final case definition |

All functions accept a `data.frame` or `data.table` and return a `data.table`.
For `data.table` input, new columns are added **by reference** (no copy);
`data.frame` input is converted to `data.table` internally before modification.

> **In production**, replace `ops_toy()` with `extract_batch()` followed by
> `decode_values()` and `decode_names()`. See `vignette("decode")`.
> Column names below use the RAP raw format (`p{field}_{instance}_{array}`)
> as returned by `ops_toy()` and `extract_batch()` before decoding.

---

## Setup

```{r load-data}
library(ukbflow)

df <- ops_toy(n = 500)
```

---

## Step 1: Handle Informative Missing Labels

UKB uses special labels such as `"Do not know"` and `"Prefer not to answer"`
to distinguish refusal from true missing data. `derive_missing()` converts
these to `NA` (default) or retains them as `"Unknown"` for modelling.

```{r derive-missing}
df <- derive_missing(df)
```

> **Performance**: `derive_missing()` uses `data.table::set()` for in-place
> replacement — no column copies are made regardless of dataset size.

To keep non-response as a model category:

```{r derive-missing-unknown}
df <- derive_missing(df, action = "unknown")
```

To add custom labels beyond the built-in list:

```{r derive-missing-extra}
df <- derive_missing(df, extra_labels = "Not applicable")
```

---

## Step 2: Prepare Covariates

`derive_covariate()` converts categorical columns to `factor` and prints a
distribution summary for each.

```{r derive-covariate}
df <- derive_covariate(
  df,
  as_factor = c(
    "p31",        # sex
    "p20116_i0",  # smoking_status_i0
    "p1558_i0"    # alcohol_intake_frequency_i0
  ),
  factor_levels = list(
    p20116_i0 = c("Never", "Previous", "Current")
  )
)
```

---

## Step 3: Bin Continuous Variables

`derive_cut()` creates a new factor column by binning a continuous variable
into quantile-based or custom groups.

```{r derive-cut}
df <- derive_cut(
  df,
  col    = "p21001_i0",                              # body_mass_index_bmi_i0
  n      = 4,
  breaks = c(18.5, 25, 30),
  labels = c("Underweight", "Normal", "Overweight", "Obese"),
  name   = "bmi_cat"
)

df <- derive_cut(
  df,
  col    = "p22189",                                 # townsend_deprivation_index_at_recruitment
  n      = 4,
  labels = c("Q1 (least deprived)", "Q2", "Q3", "Q4 (most deprived)"),
  name   = "tdi_cat"
)
```

---

## Step 4: Self-Reported Disease

`derive_selfreport()` searches UKB self-reported non-cancer illness (field
20002) or cancer (field 20001) columns for a disease label matching a regex,
then returns binary status and the earliest report date. Column detection
is automatic from field IDs.

```{r derive-selfreport}
# Non-cancer: type 2 diabetes (field 20002)
df <- derive_selfreport(df,
  name  = "dm",
  regex = "type 2 diabetes"
)
```

```{r derive-selfreport-cancer}
# Cancer: lung cancer (field 20001)
df <- derive_selfreport(df,
  name  = "lung_cancer",
  regex = "lung cancer",
  field = "cancer"
)
```

This adds two columns per call:

| Column | Type | Description |
|---|---|---|
| `dm_selfreport` | logical | `TRUE` if any instance matched |
| `dm_selfreport_date` | IDate | Earliest report date |

---

## Step 5: HES Inpatient Records

`derive_hes()` scans UKB Hospital Episode Statistics ICD-10 codes (field
41270, stored as a JSON array per participant) and matches the earliest
corresponding date from field 41280.

```{r derive-hes}
# Prefix match: codes starting with "I10" (hypertension)
df <- derive_hes(df, name = "htn", icd10 = "I10")

# Exact match
df <- derive_hes(df, name = "dm_hes", icd10 = "E11", match = "exact")

# Regex: E10 and E11 simultaneously
df <- derive_hes(df, name = "dm_broad", icd10 = "^E1[01]", match = "regex")
```

The `match` argument controls how codes are compared:

| `match` | Behaviour | Example |
|---|---|---|
| `"prefix"` (default) | Code starts with pattern | `"E11"` matches `"E110"`, `"E119"` |
| `"exact"` | Full 3- or 4-digit match | `"E11"` matches only `"E11"` |
| `"regex"` | Full regular expression | `"^E1[01]"` |

---

## Step 6: First Occurrence Fields

UKB First Occurrence fields (p131xxx) record the earliest date a condition
was observed across **all linked sources** — self-report, HES inpatient, GP
records, and death registry — pre-integrated by UKB. Look up your disease in the
[UKB Field Finder](https://biobank.ndph.ox.ac.uk/showcase/search.cgi).

```{r derive-fo}
# ops_toy includes p131742 as a representative First Occurrence column
df <- derive_first_occurrence(df, name = "htn", field = 131742L, col = "p131742")
```

---

## Step 7: Cancer Registry

`derive_cancer_registry()` searches the cancer registry ICD-10 field (40006)
and optionally filters by histology (field 40011) and behaviour (field 40012).

```{r derive-cancer}
# ICD-10 only
df <- derive_cancer_registry(df,
  name  = "skin_cancer",
  icd10 = "^C44"
)

# With histology and behaviour filters
df <- derive_cancer_registry(df,
  name      = "scc",
  icd10     = "^C44",
  histology = c(8070L, 8071L, 8072L),
  behaviour = 3L                        # 3 = malignant
)
```

---

## Step 8: Death Registry

`derive_death_registry()` searches primary (field 40001) and secondary (field
40002) causes of death for ICD-10 codes.

```{r derive-death}
df <- derive_death_registry(df, name = "mi",   icd10 = "I21")
df <- derive_death_registry(df, name = "dm",   icd10 = "E11")
df <- derive_death_registry(df, name = "lung", icd10 = "C34")
```

---

## Step 9: Combine Sources with `derive_icd10()`

`derive_icd10()` is a high-level wrapper that calls any combination of the
source-specific functions above and merges their outputs into a single status
column and earliest date. This is the recommended approach for multi-source
ascertainment.

```{r derive-icd10}
# Non-cancer disease: HES + death + First Occurrence
df <- derive_icd10(df,
  name   = "dm",
  icd10  = "E11",
  source = c("hes", "death", "first_occurrence"),
  fo_col = "p131742"
)

# Cancer outcome: cancer registry
df <- derive_icd10(df,
  name      = "lung",
  icd10     = "^C3[34]",
  match     = "regex",
  source    = "cancer_registry",
  behaviour = 3L
)
```

Intermediate source columns are retained alongside the combined result:

| Column | Type | Description |
|---|---|---|
| `dm_icd10` | logical | `TRUE` if positive in any specified source |
| `dm_icd10_date` | IDate | Earliest date across all sources |
| `dm_hes` | logical | HES status |
| `dm_hes_date` | IDate | HES date |
| `dm_fo` | logical | First Occurrence status |
| `dm_fo_date` | IDate | First Occurrence date |
| `dm_death` | logical | Death registry status |
| `dm_death_date` | IDate | Death registry date |

---

## Step 10: Final Case Definition

`derive_case()` merges the self-report and ICD-10 flags into a unified case
status, with the earliest date across both sources taken via `pmin()`.

```{r derive-case}
df <- derive_case(df, name = "dm")
```

Output columns:

| Column | Type | Description |
|---|---|---|
| `dm_status` | logical | `TRUE` if positive in self-report OR ICD-10 |
| `dm_date` | IDate | **Earliest** date across all sources (`pmin`) |

> **Why the earliest date matters**: `dm_date` is the direct input to
> `derive_timing()`, `derive_age()`, and `derive_followup()` — it is the
> chronological anchor of every downstream survival analysis.
> See `vignette("derive-survival")`.

---

## Getting Help

- `?derive_missing`, `?derive_covariate`, `?derive_cut`
- `?derive_selfreport`, `?derive_hes`, `?derive_first_occurrence`
- `?derive_cancer_registry`, `?derive_death_registry`
- `?derive_icd10`, `?derive_case`
- `vignette("derive-survival")` — timing, age at event, follow-up
- `vignette("decode")` — decoding column names and values
- [GitHub Issues](https://github.com/evanbio/ukbflow/issues)

---
title: "Exploring and Fetching RAP Files"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Exploring and Fetching RAP Files}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = FALSE
)
```

## Overview

The `fetch_*` functions provide a convenient R interface for exploring and downloading files from your UK Biobank RAP project. Rather than switching to the terminal and using `dx` commands directly, you can browse your remote project structure and retrieve files entirely within your R session.

> **UK Biobank Data Policy (2024+)**: Only summary-level outputs and metadata files may be downloaded locally. Individual-level phenotype and genotype data must remain within the RAP environment.

---

## Prerequisites

Ensure you are authenticated before using any `fetch_*` functions:

```{r auth}
library(ukbflow)

auth_login()
auth_select_project("project-XXXXXXXXXXXX")
```

See `vignette("auth")` for details.

---

## Exploring Remote Files

### List files and folders

`fetch_ls()` lists the contents of a remote RAP directory, returning a structured data frame:

```{r fetch-ls}
# List project root
fetch_ls()
#>                  name   type    size            modified
#> 1  Showcase metadata folder    <NA>                <NA>
#> 2             results folder    <NA>                <NA>
#> 3        analysis.log   file  4.2 KB 2024-11-01 10:22:03

# List a specific folder
fetch_ls("Showcase metadata/")
#>          name  type     size            modified
#> 1   field.tsv  file  12.3 MB 2024-10-15 08:01:44
#> 2 esimpint.tsv  file   3.1 MB 2024-10-15 08:01:50

# Filter by type
fetch_ls("results/", type = "file")

# Filter by name pattern
fetch_ls("results/", pattern = "\\.csv$")
```

The returned data frame has four columns:

| Column | Description |
|---|---|
| `name` | File or folder name |
| `type` | `"file"` or `"folder"` |
| `size` | File size (e.g. `"1.2 MB"`), `NA` for folders |
| `modified` | Last modified time (`POSIXct`), `NA` for folders |

### Browse the directory tree

`fetch_tree()` prints a tree-like view of the remote project structure:

```{r fetch-tree}
# Top-level overview
fetch_tree()

# Drill into a subfolder
fetch_tree("results/", max_depth = 2)
```

> **Note**: Each level of recursion triggers one API call per folder. Keep `max_depth` at 2–3 for interactive use to avoid long waits on large projects.

---

## Generating Download URLs

`fetch_url()` generates temporary pre-authenticated HTTPS URLs for remote files. Useful for passing to downstream tools or scripting metadata and results workflows without triggering a full download.

```{r fetch-url}
# Single file
fetch_url("Showcase metadata/field.tsv")

# Entire folder (returns a named character vector)
fetch_url("Showcase metadata/", duration = "7d")
```

URLs are valid for the specified `duration` (default: `"1d"`).

---

## Downloading Files

### Single file or folder

`fetch_file()` downloads a file or an entire folder to the current or a specified directory within the RAP environment.

> **Note**: `fetch_file()`, `fetch_metadata()`, and `fetch_field()` can only be called from within the RAP environment. Calling them locally will produce an error, as individual-level UKB data must remain on the platform.

```{r fetch-file}
# Download a single file
fetch_file("Showcase metadata/field.tsv", dest_dir = "data/")

# Download an entire folder
fetch_file("Showcase metadata/", dest_dir = "data/metadata/")

# Resume an interrupted download
fetch_file("results/summary_stats.csv", dest_dir = "data/", resume = TRUE)
```

Folders are downloaded in parallel using `curl::multi_download()` for efficiency.

### Metadata shortcuts

Two convenience wrappers are provided for commonly used UKB files:

```{r fetch-meta}
# Download all Showcase metadata files (field.tsv, encoding.tsv, etc.)
fetch_metadata(dest_dir = "data/metadata")

# Download the field dictionary only
fetch_field(dest_dir = "data/metadata")
```

---

## Common Options

`fetch_metadata()` and `fetch_field()` are thin wrappers around `fetch_file()`, so all three share the same download-control arguments:

| Argument | Default | Description |
|---|---|---|
| `dest_dir` | — | Destination directory (created if needed). Must be specified explicitly. |
| `overwrite` | `FALSE` | Overwrite existing local files |
| `resume` | `FALSE` | Resume an interrupted download |
| `verbose` | `TRUE` | Show download progress |

---

## Getting Help

- `?fetch_ls`, `?fetch_tree`, `?fetch_url`, `?fetch_file`, `?fetch_metadata`, `?fetch_field`
- [GitHub Issues](https://github.com/evanbio/ukbflow/issues)

---
title: "Survival Analysis Setup for UKB Outcomes"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Survival Analysis Setup for UKB Outcomes}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = FALSE
)
```

## Overview

After disease case definitions have been derived (see `vignette("derive")`),
three additional functions prepare the data for time-to-event analysis:

| Function | Output columns | Purpose |
|---|---|---|
| `derive_timing()` | `{name}_timing` | Classify prevalent vs. incident disease |
| `derive_age()` | `age_at_{name}` | Age at event (years) |
| `derive_followup()` | `{name}_followup_end`, `{name}_followup_years` | Follow-up end date and duration |

> **Prerequisite**: `{name}_status` and `{name}_date` must already be present —
> produced by `vignette("derive")`. The examples below assume the full
> disease derivation pipeline has been run on an `ops_toy()` dataset, so
> the baseline date column is `p53_i0` and age at recruitment is `p21022`.

---

## Step 1: Classify Timing — Prevalent vs. Incident

`derive_timing()` compares the disease date to the UKB baseline assessment
date and assigns each participant to one of four categories:

| Value | Meaning |
|---|---|
| `0` | No disease (`status` is `FALSE`) |
| `1` | **Prevalent** — disease date on or before baseline |
| `2` | **Incident** — disease date strictly after baseline |
| `NA` | Case with no recorded date; timing cannot be determined |

```{r setup-data}
library(ukbflow)

# Build on the derive pipeline from vignette("derive")
df <- ops_toy(n = 500)
df <- derive_missing(df)
df <- derive_covariate(df, as_factor = c("p31", "p20116_i0"))
df <- derive_selfreport(df, name = "dm", regex = "type 2 diabetes")
df <- derive_icd10(df, name = "dm", icd10 = "E11", source = c("hes", "death"))
df <- derive_case(df, name = "dm")
```

```{r derive-timing}
# Uses {name}_status and {name}_date by default
df <- derive_timing(df, name = "dm", baseline_col = "p53_i0")
```

Supply explicit column names when the defaults do not apply:

```{r derive-timing-explicit}
df <- derive_timing(df,
  name         = "dm",
  status_col   = "dm_status",
  date_col     = "dm_date",
  baseline_col = "p53_i0"
)
```

Call once per variable needed — for example, once for the combined case and
once per individual source (HES, self-report, etc.).

---

## Step 2: Age at Event

`derive_age()` computes age at the time of the event for cases, and returns
`NA` for non-cases and cases without a date.

$$\text{age\_at\_event} = \text{age\_at\_recruitment} + \frac{\text{event\_date} - \text{baseline\_date}}{365.25}$$

The divisor 365.25 accounts for leap years, ensuring sub-monthly precision in
age calculation across the full UKB follow-up window.

```{r derive-age}
# Auto-detects {name}_date and {name}_status; produces age_at_{name} column.
df <- derive_age(df,
  name         = "dm",
  baseline_col = "p53_i0",
  age_col      = "p21022"
)
```

Supply explicit column mappings when names do not follow the default
`{name}_date` / `{name}_status` pattern:

```{r derive-age-explicit}
df <- derive_age(df,
  name         = "dm",
  baseline_col = "p53_i0",
  age_col      = "p21022",
  date_cols    = c(dm = "dm_date"),
  status_cols  = c(dm = "dm_status")
)
```

---

## Step 3: Follow-Up Time

`derive_followup()` computes the follow-up end date as the **earliest** of:

1. The outcome event date (if the participant is a case)
2. Date of death (field 40000; competing event)
3. Date lost to follow-up (field 191)
4. The administrative censoring date

Follow-up time in years is then derived from the baseline date.

```{r derive-followup}
df <- derive_followup(df,
  name         = "dm",
  event_col    = "dm_date",
  baseline_col = "p53_i0",
  censor_date  = as.Date("2022-10-31"),   # set to your study's cut-off date
  death_col    = "p40000_i0",
  lost_col     = FALSE                    # not available in ops_toy
)
```

Output columns:

| Column | Type | Description |
|---|---|---|
| `dm_followup_end` | IDate | Earliest competing date |
| `dm_followup_years` | numeric | Years from baseline to end |

### Prevalent cases receive `NA` follow-up time

Participants whose event date falls **before or on the baseline date** (prevalent
cases, `{name}_timing == 1`) will have `followup_years` set to `NA` rather than
a zero or negative value, which has no meaning in time-to-event analysis.
Use `derive_timing()` to identify and exclude prevalent cases before fitting a
Cox model (see the full pipeline example below).

### Auto-detection of death and lost-to-follow-up columns

When `death_col` and `lost_col` are `NULL` (default), `derive_followup()`
looks them up automatically from the field cache (UKB fields 40000 and 191).
Pass `FALSE` to explicitly disable a competing event:

```{r derive-followup-nodeath}
df <- derive_followup(df,
  name         = "dm",
  event_col    = "dm_date",
  baseline_col = "p53_i0",
  censor_date  = as.Date("2022-10-31"),
  death_col    = FALSE,
  lost_col     = FALSE
)
```

---

## Full Survival-Ready Pipeline

After completing all three steps, the data contains everything needed to fit
a Cox proportional hazards model:

```{r cox-example}
library(survival)

# Incident analysis: exclude prevalent cases and those with undetermined timing
df_incident <- df[dm_timing != 1L]

fit <- coxph(
  Surv(dm_followup_years, dm_status) ~
    p20116_i0 + p21022 + p31 + p1558_i0,
  data = df_incident
)
summary(fit)
```

Column roles in the model:

| Column | Role |
|---|---|
| `dm_status` | Event indicator (logical) |
| `dm_followup_years` | Time variable |
| `dm_timing` | Filter: exclude prevalent (`== 1`) |
| `age_at_dm` | Age at diagnosis (descriptive / secondary analysis) |
| `p20116_i0` | Exposure of interest (smoking status) |

---

## Getting Help

- `?derive_timing`, `?derive_age`, `?derive_followup`
- `vignette("derive")` — disease phenotype derivation
- `vignette("decode")` — decoding column names and values
- [GitHub Issues](https://github.com/evanbio/ukbflow/issues)

---
title: "Extracting Phenotype Data"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Extracting Phenotype Data}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = FALSE
)
```

## Overview

UKB phenotype data is stored in a proprietary `.dataset` format on the RAP and cannot be read directly. The `extract_*` functions provide R interfaces for discovering approved fields and extracting phenotype data via the DNAnexus `dx extract_dataset` and `table-exporter` tools.

Two workflows are available:

| Function | Mode | Scale | Output |
|---|---|---|---|
| `extract_batch()` | Async job | Large / production (typically 50+ fields) | job ID → CSV on RAP cloud |
| `extract_pheno()` | Synchronous | Small (quick checks) | data.table in memory |

**`extract_batch()` is the recommended approach** for any serious analysis. `extract_pheno()` is provided for quick interactive inspection inside the RAP environment only.

---

## Prerequisites

Ensure you are authenticated and have selected your project:

```{r auth}
library(ukbflow)

auth_login()
auth_select_project("project-XXXXXXXXXXXX")
```

---

## Step 1: Browse Available Fields

Before extracting, use `extract_ls()` to explore what fields are approved for your project:

```{r extract-ls}
# List all approved fields (cached after first call)
extract_ls()

# Search by keyword
extract_ls(pattern = "cancer")
extract_ls(pattern = "p31|p53|p21022")

# Force refresh after switching projects or datasets
extract_ls(refresh = TRUE)
```

The result is a data.frame with two columns:

| Column | Example |
|---|---|
| `field_name` | `participant.p53_i0` |
| `title` | `Date of attending assessment centre \| Instance 0` |

> Fields reflect your project's approved data only — not all UKB fields are present.

---

## Step 2: Extract Data

### Recommended: `extract_batch()`

For large-scale or production extractions, submit an asynchronous table-exporter job on the RAP cloud:

```{r extract-batch}
# Submit extraction job
job_id <- extract_batch(c(31, 53, 21022, 22189))

# Custom output name
job_id <- extract_batch(
  field_id = c(31, 53, 21022, 22189),
  file     = "ukb_demographics"
)

# High priority (faster queue, higher cost)
job_id <- extract_batch(
  field_id = c(31, 53, 21022, 22189),
  priority = "high"
)
```

The job runs asynchronously on the RAP cloud. The output CSV is saved to your RAP project and can be monitored with the `job_` series:

```{r job-monitor}
job_status(job_id)        # check progress
job_path(job_id)          # get cloud file path once complete
job_result(job_id)        # read result as data.table (inside RAP only)
```

#### Instance type

`extract_batch()` automatically selects an appropriate instance based on the number of columns:

| Columns | Instance |
|---|---|
| ≤ 20 | `mem1_ssd1_v2_x4` |
| ≤ 100 | `mem1_ssd1_v2_x8` |
| ≤ 500 | `mem1_ssd1_v2_x16` |
| > 500 | `mem1_ssd1_v2_x36` |

You can override this with the `instance_type` argument if needed.

---

### Quick inspection: `extract_pheno()`

For small-scale interactive checks **inside the RAP RStudio environment**:

```{r extract-pheno}
df <- extract_pheno(c(31, 53, 21022))
```

> `extract_pheno()` is restricted to the RAP environment and returns data in memory only. For any analysis intended to be saved or reproduced, use `extract_batch()`.

Note: `extract_pheno()` returns **raw coded values** (e.g. `1`/`0` for Sex, numeric codes for diseases). Use the `decode_*` series to convert codes to human-readable labels.

---

## A Note on Column Names

Column naming differs between the two extraction methods:

**`extract_batch()`** — no prefix:

| Column | Meaning |
|---|---|
| `eid` | Participant ID |
| `p31` | Field 31 (Sex) |
| `p53_i0` | Field 53, Instance 0 |
| `p20002_i0_a0` | Field 20002, Instance 0, Array 0 |

**`extract_pheno()`** — `participant.` prefix:

| Column | Meaning |
|---|---|
| `participant.eid` | Participant ID |
| `participant.p31` | Field 31 (Sex) |
| `participant.p53_i0` | Field 53, Instance 0 |
| `participant.p20002_i0_a0` | Field 20002, Instance 0, Array 0 |

---

## Getting Help

- `?extract_ls`, `?extract_pheno`, `?extract_batch`
- `vignette("auth")` — authentication setup
- [GitHub Issues](https://github.com/evanbio/ukbflow/issues)

---
title: "Get Started with ukbflow"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Get Started with ukbflow}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = FALSE
)
```

## Welcome to `ukbflow`

**`ukbflow`** is an R package for UK Biobank analysis on the
[Research Analysis Platform (RAP)](https://ukbiobank.dnanexus.com).
It covers the full midstream-to-downstream pipeline — from phenotype derivation
and association analysis to publication-ready figures and genetic risk scoring —
designed for RAP-native UKB workflows, with local simulated data for development
and testing.

## Installation

```{r install}
# From CRAN (recommended)
install.packages("ukbflow")

# Latest development version from GitHub
pak::pkg_install("evanbio/ukbflow")
```

## A Quick Taste

### Load data

```{r load-data}
library(ukbflow)

df <- ops_toy()   # synthetic UKB-like cohort, no RAP connection needed

# On RAP, replace with:
# auth_login()
# auth_select_project("project-XXXXXXXXXXXX")
# df <- extract_pheno(c(31, 21022, 53, 20116)) |>
#   decode_values() |>
#   decode_names()
```

### Derive a disease phenotype

```{r derive}
df <- df |>
  derive_missing() |>                                               # recode "Prefer not to answer" → NA
  derive_selfreport(name = "t2dm", regex = "diabetes",           # T2DM self-report
                    field = "noncancer") |>
  derive_icd10(name = "t2dm", icd10 = "E11", source = "hes") |> # T2DM from HES
  derive_case(name = "t2dm") |>                                  # → t2dm_status, t2dm_date
  derive_followup(name         = "t2dm",
                  event_col    = "t2dm_date",
                  baseline_col = "p53_i0",                          # assessment centre date
                  censor_date  = as.Date("2022-06-01"))
```

### Run an association model

```{r assoc}
res <- assoc_coxph(
  data         = df,
  outcome_coll  = "t2dm_status",
  time_col     = "t2dm_followup_years",
  exposure_col = "p21001_i0",   # BMI (continuous)
  covariates   = c("p21022",    # age_at_recruitment
                   "p31")       # sex
)
```

### Plot the results

```{r plot}
# Forest plot — see vignette("plot") for full usage
res_df <- as.data.frame(res)
plot_forest(
  data      = res_df,
  est       = res_df$HR,
  lower     = res_df$CI_lower,
  upper     = res_df$CI_upper,
  ci_column = 7L   # res_df has 6 cols before HR; CI graphic goes here
)

# Table 1
plot_tableone(
  data   = as.data.frame(df),
  vars   = c("p21022",     # age_at_recruitment
             "p31",        # sex
             "p21001_i0"), # bmi
  strata = "t2dm_status"
)
```

## Full Function Overview

| Module | Key functions | Vignette |
|---|---|---|
| Auth | `auth_login()`, `auth_select_project()` | `vignette("auth")` |
| Fetch | `fetch_ls()`, `fetch_file()`, `fetch_tree()` | `vignette("fetch")` |
| Extract | `extract_pheno()`, `extract_batch()`, `extract_ls()` | `vignette("extract")` |
| Job | `job_wait()`, `job_status()`, `job_result()` | `vignette("job")` |
| Decode | `decode_values()`, `decode_names()` | `vignette("decode")` |
| Derive | `derive_missing()`, `derive_icd10()`, `derive_case()` | `vignette("derive")` |
| Survival | `derive_timing()`, `derive_age()`, `derive_followup()` | `vignette("derive-survival")` |
| Assoc | `assoc_coxph()`, `assoc_logistic()`, `assoc_subgroup()` | `vignette("assoc")` |
| Plot | `plot_forest()`, `plot_tableone()` | `vignette("plot")` |
| GRS | `grs_check()`, `grs_score()`, `grs_validate()` | `vignette("grs")` |
| Ops | `ops_setup()`, `ops_toy()`, `ops_snapshot()` | `vignette("ops")` |

## End-to-End Case Study

For a complete worked example using a simulated UK Biobank cohort — covering
data loading, phenotype derivation, cohort assembly, Cox regression, and
publication-ready visualisation — see:

`vignette("smoking_lung_cancer")` — **Smoking and Lung Cancer Risk: A Complete Analysis Workflow**

## Additional Resources

- [Documentation site](https://evanbio.github.io/ukbflow/)
- [GitHub](https://github.com/evanbio/ukbflow)
- View all functions: `?ukbflow` or `help(package = "ukbflow")`

> *"All models are wrong, but some are publishable."*
>
> — after George Box

---
title: "Decoding UKB Column Names and Values"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Decoding UKB Column Names and Values}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = FALSE
)
```

## Overview

Raw UKB phenotype data contains encoded column names and values that need to be converted before analysis.

| Source | Column names | Column values |
|---|---|---|
| `extract_pheno()` | `participant.p31` | Raw integer codes — needs `decode_values()` |
| `extract_batch()` | `p31`, `p53_i0` | Usually already decoded — `decode_values()` typically not needed |

Both outputs need `decode_names()` to convert field ID column names to human-readable snake_case.

> **Call order matters**: when using `extract_pheno()` output, always run `decode_values()` before `decode_names()`, because value decoding relies on the numeric field ID still being present in the column name.

---

## Recommended Workflow

```{r workflow}
library(ukbflow)

df <- extract_pheno(c(31, 54, 20116, 21022))
df <- decode_values(df)   # 0/1 → "Female"/"Male", etc.
df <- decode_names(df)    # participant.p31 → sex
```

---

## Step 1: Decode Values

`decode_values()` converts raw integer codes to human-readable labels for categorical fields that have UKB encoding mappings. Continuous, date, text, and already-decoded fields are left unchanged.

```{r decode-values}
df <- decode_values(df)
#> ✔ Decoded 3 categorical columns; 2 non-categorical columns unchanged.
```

It requires two metadata files from the UKB Showcase. Download them once with:

```{r fetch-meta}
fetch_metadata(dest_dir = "data/metadata")
```

Then point `decode_values()` to the same directory (default matches `fetch_metadata()`):

```{r decode-values-dir}
df <- decode_values(df, metadata_dir = "data/metadata")
```

### What gets decoded

| Column | Raw value | Decoded value |
|---|---|---|
| `p31` | `0` / `1` | `"Female"` / `"Male"` |
| `p54` | `11012` | `"Leeds"` |
| `p20116_i0` | `0` / `1` / `2` | `"Never"` / `"Previous"` / `"Current"` |

Codes absent from the encoding table (including UKB missing codes `-1`, `-3`, `-7`) are returned as `NA`.

---

## Step 2: Decode Names

`decode_names()` renames columns from field ID format to snake_case labels using the approved UKB field dictionary available to your project.

```{r decode-names}
df <- decode_names(df)
#> ✔ Renamed 5 columns.
```

### Name conversion examples

| Raw name | Decoded name |
|---|---|
| `participant.eid` | `eid` |
| `participant.p31` | `sex` |
| `participant.p21022` | `age_at_recruitment` |
| `participant.p53_i0` | `date_of_attending_assessment_centre_i0` |
| `p31` | `sex` |
| `p53_i0` | `date_of_attending_assessment_centre_i0` |

Both `extract_pheno()` format (`participant.p31`) and `extract_batch()` format (`p31`) are handled automatically.

### Long names

Some UKB field titles are verbose. Names exceeding `max_nchar` characters are flagged with a warning (default: 60). Lower the threshold to catch more aggressively:

```{r long-names}
df <- decode_names(df, max_nchar = 30)
#> ! 1 column name longer than 30 characters - consider renaming manually:
#> • date_of_attending_assessment_centre_i0
```

Rename manually to something concise:

```{r rename}
names(df)[names(df) == "date_of_attending_assessment_centre_i0"] <- "date_baseline"
```

---

## Getting Help

- `?decode_values`, `?decode_names`
- `vignette("extract")` — extracting phenotype data
- [GitHub Issues](https://github.com/evanbio/ukbflow/issues)

---
title: "Installation Guide"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Installation Guide}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = FALSE
)
```

## Overview

`ukbflow` is an R package for UK Biobank data analysis and visualization, designed to work natively with the [UK Biobank Research Analysis Platform (RAP)](https://ukbiobank.dnanexus.com). It supports two working modes:

- **Local → RAP**: Run R locally, authenticate with DNAnexus, submit jobs, and download summary results.
- **RAP → RAP**: Run R directly inside the RAP cloud environment, where authentication is handled automatically.

> **UK Biobank Data Policy (2024+)**: Individual-level phenotype and genotype data must remain within the RAP environment and cannot be downloaded locally. Only summary-level outputs (aggregated statistics, plots, reports) may be exported.

---

## Quick Install

### From CRAN *(recommended)*

```{r cran-install}
install.packages("ukbflow")
```

### From GitHub *(latest development version)*

```{r github-install}
# Using pak (faster)
install.packages("pak")
pak::pkg_install("evanbio/ukbflow")

# Using remotes
install.packages("remotes")
remotes::install_github("evanbio/ukbflow")
```

---

## System Requirements

- **R Version**: >= 4.1
- **Operating Systems**: Windows, macOS, Linux
- **dxpy**: Required for local → RAP authentication and job submission (see below)
- **RAP Account**: A valid UK Biobank RAP account and API token

---

## Dependencies

`ukbflow` dependencies are installed automatically with the package.

### Core Dependencies

- **data.table** — Fast in-memory data processing
- **cli** — Progress messages and user feedback
- **processx** — Reliable system command execution with robust error handling (used by auth and job functions)
- **jsonlite** — JSON parsing for RAP API responses
- **curl** — File downloads

### Analysis Dependencies

- **gtsummary** — Table 1 generation
- **gt** — Publication-quality table rendering
- **survival** — Survival analysis
- **dplyr** / **tidyselect** / **rlang** — Data manipulation

### Visualization Dependencies

- **forestploter** — Forest plot generation

### Optional Dependencies

These packages are not installed automatically. Install them only if you need the corresponding features:

```{r optional-deps}
# PDF export from plot_tableone() (requires Chrome / Chromium)
install.packages("pagedown")

# PNG export from plot_tableone()
install.packages("webshot2")
```

---

## Install dxpy (Local Mode Only)

The `auth_*` and `job_*` functions rely on the `dx` command-line tool from [dxpy](https://documentation.dnanexus.com/downloads). Required only when running locally.

```{bash install-dxpy}
pip install dxpy
```

Verify:

```{bash verify-dx}
dx --version
```

> Skip this step if you are running entirely within the RAP RStudio environment.

---

## Authentication Setup

### Local → RAP

Obtain your API token from the DNAnexus platform under **Account Settings > API Tokens**. Store it in your `.Renviron` file (never in your script):

```{r auth-renviron}
usethis::edit_r_environ()
# Add the following line, then save and restart R:
# DX_API_TOKEN=your_token_here
```

Then authenticate:

```{r auth-local}
library(ukbflow)

auth_login()                                 # reads DX_API_TOKEN automatically
auth_status()                                # confirm user and active project
auth_list_projects()                         # find your project ID
auth_select_project("project-XXXXXXXXXXXX")  # switch to your UKB project
```

> For full details on token management, project selection, and both authentication modes, see `vignette("auth")`.

### RAP → RAP

Authentication is automatic inside the RAP environment. Verify the session with:

```{r auth-rap}
library(ukbflow)

auth_status()  # confirms user and active project
```

---

## Verify Installation

```{r verify}
library(ukbflow)

packageVersion("ukbflow")
ops_setup()   # checks R version, dxpy, DX_API_TOKEN, and key package versions
```

---

## Update ukbflow

### From GitHub

```{r update-github}
remotes::install_github("evanbio/ukbflow", force = TRUE)
```

### From CRAN

```{r update-cran}
update.packages("ukbflow")
```

---

## Troubleshooting

### `dx` not found

**Solution**: Ensure dxpy is installed and on your PATH:

```{bash dx-path}
pip install dxpy
which dx   # macOS/Linux
where dx   # Windows
```

### Token expired or session lost

DNAnexus API tokens have a limited validity period. If authentication fails, generate a new token from the DNAnexus platform and log in again:

```{r fix-token}
auth_login("your_new_token_here")
```

### Installation fails on Windows

**Solution**: Install [Rtools](https://cran.r-project.org/bin/windows/Rtools/) for packages that require compilation.

### Network / Firewall issues

**Solution**: Configure a proxy before installing:

```{r proxy}
Sys.setenv(http_proxy  = "http://your-proxy:port")
Sys.setenv(https_proxy = "https://your-proxy:port")
```

---

## Uninstall

```{r uninstall}
remove.packages("ukbflow")
```

---

## Getting Help

- **Documentation**: <https://evanbio.github.io/ukbflow/>
- **Issues**: [GitHub Issues](https://github.com/evanbio/ukbflow/issues)

---

## Next Steps

After installation:

1. Read the [Getting Started](https://evanbio.github.io/ukbflow/articles/get-started.html) guide
2. Browse the [Function Reference](https://evanbio.github.io/ukbflow/reference/)

---
title: "Authentication and Project Setup"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Authentication and Project Setup}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = FALSE
)
```

## Overview

`ukbflow` interacts with the UK Biobank Research Analysis Platform (RAP) via the DNAnexus `dx` toolkit. Authentication is required to access your UKB project, submit jobs, and retrieve results.

Two working modes are supported:

- **Local → RAP**: Run R locally, authenticate with a DNAnexus API token, and interact with RAP remotely.
- **RAP → RAP**: Run R inside the RAP cloud environment (e.g., RStudio on DNAnexus), where authentication is handled automatically by the platform.

---

## Obtaining an API Token

1. Log in to the [DNAnexus platform](https://platform.dnanexus.com)
2. Go to **Account Settings > API Tokens**
3. Click **New Token**, set an appropriate expiry, and copy the token

> Keep your token private. Treat it like a password — do not share it or commit it to version control.

---

## Storing Your Token Securely

The recommended approach is to store your token in `~/.Renviron`, a local file that is never committed to git. This keeps the token out of your scripts entirely.

```{r renviron}
usethis::edit_r_environ()
# Add the following line, then save and restart R:
# DX_API_TOKEN=your_token_here
```

After restarting R, the token is available to `auth_login()` automatically.

---

## Logging In

### Local → RAP

```{r auth-login}
library(ukbflow)

auth_login()  # reads DX_API_TOKEN from environment
```

The token is cached locally by the dx toolkit and persists across R sessions. In most local workflows, **you only need to log in once** — subsequent R sessions usually do not require re-authentication unless the token expires, the local dx session is cleared, or you explicitly log out.

If you prefer to pass the token directly (e.g., in an interactive session), you can do so:

```{r auth-login-direct}
auth_login("your_token_here")
```

> **Security note**: Avoid saving `auth_login("token")` calls in script files. Prefer the `.Renviron` approach for any workflow you intend to save or share.

### RAP → RAP

When running inside the RAP RStudio environment, authentication is handled automatically by the platform. No login step is required:

```{r auth-rap}
library(ukbflow)

auth_status()  # verify the current session
```

---

## Checking Authentication Status

```{r auth-status}
auth_status()
#> • User:    "user-XXXXXXXXXXXX"
#> • Project: "project-XXXXXXXXXXXX"
```

`auth_status()` returns the current user and active project. Use this to confirm your session before running any analysis.

---

## Listing Available Projects

```{r list-projects}
auth_list_projects()
#> project-XXXXXXXXXXXX : My UKB Project (CONTRIBUTOR)
#> project-YYYYYYYYYYYY : Shared Analysis Project (VIEW)
```

This returns all RAP projects accessible to your account, along with their project IDs. Project IDs take the form `project-XXXXXXXXXXXX` and are required for `auth_select_project()`.

---

## Selecting a Project

```{r select-project}
auth_select_project("project-XXXXXXXXXXXX")
#> ✔ Project selected: "project-XXXXXXXXXXXX"
```

`ukbflow` uses project IDs rather than names to avoid ambiguity. Use `auth_list_projects()` to find the correct ID first.

Once selected, the project context is saved by the dx toolkit and persists across sessions.

---

## Logging Out

```{r auth-logout}
auth_logout()
#> ✔ Logged out from DNAnexus.
```

Logging out invalidates the current DNAnexus session on the remote platform. The local token file is not removed but becomes invalid. A new token must be generated from the DNAnexus platform before logging in again.

---

## Token Expiry

DNAnexus API tokens have a limited validity period set at creation time. When a token expires:

- `auth_login()` or other dx-dependent functions will fail
- Generate a new token from the DNAnexus platform and log in again:

```{r reauth}
auth_login("your_new_token_here")
```

Or update `~/.Renviron` with the new token and call `auth_login()` without arguments.

---

## Full Local → RAP Workflow

```{r full-workflow}
library(ukbflow)

auth_login()                                 # authenticate
auth_status()                                # verify session
auth_list_projects()                         # find your project ID
auth_select_project("project-XXXXXXXXXXXX")  # set active project

# ... run your analysis ...

auth_logout()                                # optional: clear session
```

---

## Getting Help

- `?auth_login`, `?auth_status`, `?auth_list_projects`, `?auth_select_project`, `?auth_logout`
- [DNAnexus documentation](https://documentation.dnanexus.com)
- [GitHub Issues](https://github.com/evanbio/ukbflow/issues)

---
title: "Operational Utilities: Setup, Diagnostics, and Pipeline Tracking"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Operational Utilities: Setup, Diagnostics, and Pipeline Tracking}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = FALSE
)
```

## Overview

The `ops_*` functions are a set of lightweight utilities that sit outside the
main analysis pipeline. They help you verify your environment before starting,
explore data quality, and track how your cohort changes at each processing step.

| Function | Purpose |
|---|---|
| `ops_setup()` | Check dx CLI, RAP authentication, and R package dependencies |
| `ops_toy()` | Generate synthetic UKB-like data for development and testing |
| `ops_na()` | Summarise missing values (NA and `""`) across all columns |
| `ops_snapshot()` | Record pipeline checkpoints and track dataset changes |

`ops_setup()` may query dx CLI and RAP authentication status as part of its
health check. All other functions operate entirely locally: `ops_toy()` and
`ops_na()` are read-only; `ops_snapshot()` and its companions track and
optionally clean up columns; `ops_withdraw()` removes withdrawn participants
in-place. None of them read from or write to RAP storage.

---

## `ops_setup()` — Environment Health Check

Run `ops_setup()` once after installing ukbflow to confirm that all required
components are in place before starting a real analysis.

```{r ops-setup}
library(ukbflow)

ops_setup()
#> ── ukbflow environment check ──────────────────────────────────────────────
#> ℹ ukbflow 0.1.0  |  R 4.4.1  |  2026-03-09
#> ── 1. dx-toolkit ──────────────────────────────────────────────────────────
#> ✔ dx: /usr/local/bin/dx  (dx-toolkit v0.375.0)
#> ── 2. RAP authentication ───────────────────────────────────────────────────
#> ✔ user: evan.zhou
#> ✔ project: project-GXk9...
#> ── 3. R packages ───────────────────────────────────────────────────────────
#> ✔ cli  3.6.3  [core]
#> ✔ data.table  1.15.4  [core]
#> ✔ survival  3.7.0  [assoc_coxph]
#> ✔ forestploter  1.1.1  [plot_forest]
#> ...
#> ───────────────────────────────────────────────────────────────────────────
#> ✔ 15 passed
#> ! 2 optional / warning
```

For programmatic use (e.g. inside scripts or CI), set `verbose = FALSE` and
inspect the returned list:

```{r ops-setup-prog}
result <- ops_setup(verbose = FALSE)
result$summary
#> $pass
#> [1] 15
#> $warn
#> [1] 2
#> $fail
#> [1] 0

# Gate the rest of your script on a clean environment
stopifnot(result$summary$fail == 0)
```

Individual checks can be disabled when only a subset is needed:

```{r ops-setup-partial}
# Check R package dependencies only (skip dx and RAP auth)
ops_setup(check_dx = FALSE, check_auth = FALSE)
```

---

## `ops_toy()` — Synthetic UKB Data

`ops_toy()` generates a realistic but entirely synthetic dataset that mimics
the structure of UKB phenotype data on the RAP. Use it to develop and test
`derive_*`, `assoc_*`, and `plot_*` functions without needing real UKB data
access.

### Cohort scenario

The default `"cohort"` scenario produces a wide participant-level table that
covers all major UKB data domains:

```{r ops-toy-cohort}
dt <- ops_toy()
#> ✔ ops_toy: 1000 participants | 75 columns | scenario = "cohort" | seed = 42

dim(dt)
#> [1] 1000   75

names(dt)
#>  [1] "eid"          "p31"          "p34"          "p53_i0"
#>  [5] "p21022"       "p21001_i0"    "p20116_i0"    "p1558_i0"
#>  ...
```

Column groups included:

| Group | Columns |
|---|---|
| Demographics | `eid`, `p31`, `p34`, `p53_i0`, `p21022` |
| Covariates | `p21001_i0`, `p20116_i0`, `p1558_i0`, `p21000_i0`, `p22189`, `p54_i0` |
| Genetic PCs | `p22009_a1` – `p22009_a10` |
| Self-report disease | `p20002_i0_a0` – `a4`, `p20008_i0_a0` – `a4` |
| Self-report cancer | `p20001_i0_a0` – `a4`, `p20006_i0_a0` – `a4` |
| HES | `p41270` (JSON array), `p41280_a0` – `a8` |
| Cancer registry | `p40006_i0` – `i2`, `p40011_i0` – `i2`, `p40012_i0` – `i2`, `p40005_i0` – `i2` |
| Death registry | `p40001_i0`, `p40002_i0_a0` – `a2`, `p40000_i0` |
| First occurrence | `p131742` |
| GRS columns | `grs_bmi`, `grs_raw`, `grs_finngen` |
| Messy columns | `messy_allna`, `messy_empty`, `messy_label` |

The messy columns deliberately stress-test `derive_missing()` and `ops_na()`
against common data quality issues (all-NA columns, empty strings, non-standard
missing labels).

Feed the output directly into the derive pipeline:

```{r ops-toy-pipeline}
dt <- ops_toy()
dt <- derive_missing(dt)
dt <- derive_covariate(dt,
  as_numeric = "p21001_i0",
  as_factor  = c("p31", "p20116_i0")
)
```

### Forest scenario

The `"forest"` scenario returns a results table matching the output of
`assoc_coxph()`, useful for developing and testing `plot_forest()` without
running a real Cox model:

```{r ops-toy-forest}
dt_forest <- ops_toy(scenario = "forest")
#> ✔ ops_toy: 24 rows | 11 columns | scenario = "forest" | seed = 42

plot_forest(
  data  = dt_forest[model == "Fully adjusted"],
  est   = dt_forest[model == "Fully adjusted", HR],
  lower = dt_forest[model == "Fully adjusted", CI_lower],
  upper = dt_forest[model == "Fully adjusted", CI_upper]
)
```

### Reproducibility

Results are reproducible by default (`seed = 42`). Pass `seed = NULL` for a
different dataset on every call:

```{r ops-toy-seed}
dt1 <- ops_toy(seed = 1)
dt2 <- ops_toy(seed = 1)
identical(dt1, dt2)   # TRUE

dt_random <- ops_toy(seed = NULL)   # different every call
```

---

## `ops_na()` — Missing Value Diagnostics

`ops_na()` scans every column for `NA` **and empty strings (`""`)**, returning
counts and percentages sorted by missingness. Counting `""` as missing is
intentional — UKB exports frequently use empty strings as placeholders for
absent text values, so `ops_na()` reports *effective* missingness rather than
a plain `is.na()` count. It is designed to be called before `derive_missing()`
to understand the data quality profile of a freshly extracted UKB dataset.

```{r ops-na-basic}
dt <- ops_toy()
ops_na(dt)
#> ── ops_na ──────────────────────────────────────────────────────────────────
#> ℹ 1000 rows | 65 columns | threshold = 0%
#> ✖ messy_allna   1000 / 1000  (100.00%)
#> ✖ p41280_a4     1000 / 1000  (100.00%)
#> ✖ p20002_i0_a4   976 / 1000  ( 97.60%)
#> ✖ p131742        916 / 1000  ( 91.60%)
#> ...
#> ────────────────────────────────────────────────────────────────────────────
#> ✖ 41 columns ≥ 10% missing
#> ✔ 24 columns complete (0% missing)
```

Columns with ≥ 10% missing are flagged in red (`✖`); those between 0% and 10%
in yellow (`!`). The summary block (totals) is always printed regardless of the
`threshold` setting.

### Controlling CLI output with `threshold`

Use `threshold` to silence low-missingness columns from the per-column listing
when the dataset has many columns. The summary block and returned data.table
are always complete.

```{r ops-na-threshold}
# Only list columns with > 50% missing in the console output
ops_na(dt, threshold = 50)

# Suppress all per-column lines — summary only
ops_na(dt, threshold = 99)
```

### Programmatic use

`ops_na()` returns a `data.table` invisibly, regardless of `threshold`:

```{r ops-na-prog}
result <- ops_na(dt, verbose = FALSE)
result
#>           column  n_na pct_na
#>           <char> <int>  <num>
#>  1:  messy_allna  1000  100.0
#>  2:    p41280_a4  1000  100.0
#>  ...

# Identify columns to drop before modelling
cols_to_drop <- result[pct_na > 90, column]
dt[, (cols_to_drop) := NULL]
```

---

## `ops_snapshot()` — Pipeline Checkpoints

`ops_snapshot()` records a lightweight summary of your dataset at each
processing step and stores it in the session cache. Each subsequent call
automatically computes deltas (Δ) against the previous snapshot, making it
easy to track how rows, columns, and missingness change through the pipeline.

### Recording snapshots

```{r ops-snapshot-record}
dt <- ops_toy()
ops_snapshot(dt, label = "raw")
#> ── snapshot: raw ───────────────────────────────────────────────────────────
#>   rows      1,000
#>   cols         65
#>   NA cols      41
#>   size       0.61 MB
#> ────────────────────────────────────────────────────────────────────────────

dt <- derive_missing(dt)
ops_snapshot(dt, label = "after_derive_missing")
#> ── snapshot: after_derive_missing ──────────────────────────────────────────
#>   rows      1,000  (= 0)
#>   cols         65  (= 0)
#>   NA cols      43  (+2)
#>   size       0.61 MB  (= 0)
#> ────────────────────────────────────────────────────────────────────────────

dt <- dt[p31 == "Female"]
ops_snapshot(dt, label = "female_only")
#> ── snapshot: female_only ───────────────────────────────────────────────────
#>   rows        570  (-430)
#>   cols         65  (= 0)
#>   NA cols      43  (= 0)
#>   size       0.36 MB  (-0.25 MB)
#> ────────────────────────────────────────────────────────────────────────────
```

When `label` is omitted, snapshots are named `snapshot_1`, `snapshot_2`, etc.
automatically. Labels should be unique within a session: if the same label is
used twice, the history row is appended again but the stored column list is
overwritten — which can cause `ops_snapshot_cols()` and `ops_snapshot_diff()`
to behave unexpectedly.

### Viewing the full history

Call `ops_snapshot()` with no arguments to print and return the complete
history data.table:

```{r ops-snapshot-history}
ops_snapshot()
#> ── ops_snapshot history ────────────────────────────────────────────────────
#>    idx                label timestamp  nrow  ncol n_na_cols size_mb
#>  1:  1                  raw  14:30:01  1000    65        41    0.61
#>  2:  2 after_derive_missing  14:30:05  1000    65        43    0.61
#>  3:  3          female_only  14:30:08   570    65        43    0.36
#> ────────────────────────────────────────────────────────────────────────────
```

### Silent recording

Set `verbose = FALSE` to record a snapshot without printing anything — useful
inside functions or automated scripts:

```{r ops-snapshot-silent}
ops_snapshot(dt, label = "pre_assoc", verbose = FALSE)
```

### Resetting history

```{r ops-snapshot-reset}
ops_snapshot(reset = TRUE)
#> ✔ Snapshot history cleared.
```

> **Session scope**: the snapshot history lives in ukbflow's session cache and
> is cleared when the R session ends or when `ops_snapshot(reset = TRUE)` is
> called. It is not written to disk.

---

## Snapshot Helpers

### `ops_snapshot_cols()` — column names at a checkpoint

Returns the column names recorded at a given snapshot label, minus protected
columns (`eid`, `sex`, `age`, `age_at_recruitment`, and any registered via
`ops_set_safe_cols()`). The primary use is building a drop vector after the
raw columns are no longer needed.

```{r ops-snapshot-cols}
raw_cols <- ops_snapshot_cols("raw")
# raw_cols is a character vector of droppable column names
```

Pass `keep` to protect additional columns beyond the defaults:

```{r ops-snapshot-cols-keep}
raw_cols <- ops_snapshot_cols("raw", keep = "p53_i0")
```

### `ops_snapshot_diff()` — compare two checkpoints

Returns lists of columns added and removed between two snapshots — useful for
auditing what `derive_*` functions produced.

```{r ops-snapshot-diff}
result <- ops_snapshot_diff("raw", "after_derive_missing")
result$added    # columns added in this step
result$removed  # columns dropped in this step
```

### `ops_snapshot_remove()` — drop raw columns after deriving

Removes the raw columns captured at a snapshot from `data`, keeping any
derived columns added since. Built-in safe columns (`eid`, etc.) and columns
supplied in `keep` are always retained.

```{r ops-snapshot-remove}
# After deriving, drop the original raw columns
dt <- ops_snapshot_remove(dt, from = "raw")
#> ✔ ops_snapshot_remove: dropped 60 raw columns, 15 remaining.
```

For `data.table` input the operation is by reference (in-place); for
`data.frame` input a new `data.table` is returned and the original is not
modified.

### `ops_set_safe_cols()` — register study-specific protected columns

Adds column names to the session safe list so they are never dropped by
`ops_snapshot_cols()` or `ops_snapshot_remove()`.

```{r ops-set-safe-cols}
ops_set_safe_cols(c("date_baseline", "age_at_recruitment"))

# Clear registered safe cols
ops_set_safe_cols(reset = TRUE)
```

---

## `ops_withdraw()` — Exclude Withdrawn Participants

UK Biobank periodically issues withdrawal files listing participants who have
revoked consent. `ops_withdraw()` reads the headerless single-column CSV
supplied by UKB and removes matching rows from your dataset. Two snapshots
(`before_withdraw` / `after_withdraw`) are recorded automatically.

```{r ops-withdraw}
dt <- ops_withdraw(dt, file = "withdraw.csv")
#> ── snapshot: before_withdraw ───────────────────────────────────────────────
#>   rows      502,492
#>   ...
#> ── snapshot: after_withdraw ────────────────────────────────────────────────
#>   rows      502,489  (-3)
#>   ...
#> ℹ Withdrawal file: w854944_20260310.csv (312 IDs)
#> ✖ Excluded: 3 participants found in data
#> ✔ Remaining: 502,489 participants
```

Run this immediately after loading your extracted dataset, before any
`derive_*` steps, so withdrawn participants never enter the analysis.

---

## Typical Workflow

The four `ops_*` functions form a natural bookend around the core pipeline:

```{r ops-workflow}
library(ukbflow)

# 1. Verify environment before starting
ops_setup()

# 2. Generate test data (or extract real data from RAP)
dt <- ops_toy()

# 3. Inspect data quality before processing
ops_na(dt)

# 4. Run pipeline with checkpoints
ops_snapshot(dt, label = "raw")

dt <- derive_missing(dt)
ops_snapshot(dt, label = "after_derive_missing")

dt <- derive_covariate(dt,
  as_numeric = "p21001_i0",
  as_factor  = c("p31", "p20116_i0")
)
ops_snapshot(dt, label = "after_derive_covariate")

# 5. Review full pipeline history
ops_snapshot()
```

---

## Getting Help

- `?ops_setup`, `?ops_toy`, `?ops_na`, `?ops_snapshot`
- `?ops_snapshot_cols`, `?ops_snapshot_diff`, `?ops_snapshot_remove`, `?ops_set_safe_cols`
- `?ops_withdraw`
- `vignette("get-started")` — end-to-end pipeline overview
- `vignette("derive")` — disease phenotype derivation
- [GitHub Issues](https://github.com/evanbio/ukbflow/issues)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fetch.R
\name{fetch_file}
\alias{fetch_file}
\title{Download a file from RAP project storage}
\usage{
fetch_file(path, dest_dir, overwrite = FALSE, resume = FALSE, verbose = TRUE)
}
\arguments{
\item{path}{(character) Remote file or folder path.}

\item{dest_dir}{(character) Destination directory. Created automatically
if it does not exist.}

\item{overwrite}{(logical) Overwrite existing local files. Default:
\code{FALSE}.}

\item{resume}{(logical) Resume an interrupted download. Useful for large
files (e.g. \code{.bed}, \code{.bgen}). Default: \code{FALSE}.}

\item{verbose}{(logical) Show download progress. Default: \code{TRUE}.}
}
\value{
Invisibly returns the local file path(s) as a character vector.
}
\description{
Downloads one file or all files within a folder from RAP project storage
to the current directory or a specified destination within the RAP
environment. This function must be called from within RAP.
}
\examples{
\dontrun{
# Download a single metadata file
fetch_file("Showcase metadata/field.tsv", dest_dir = "data/")

# Download an entire folder
fetch_file("Showcase metadata/", dest_dir = "data/metadata/")

# Resume an interrupted download
fetch_file("results/summary_stats.csv", dest_dir = "data/", resume = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/job.R
\name{job_status}
\alias{job_status}
\title{Check the current state of a DNAnexus job}
\usage{
job_status(job_id)
}
\arguments{
\item{job_id}{(character) Job ID returned by \code{extract_batch()},
e.g. \code{"job-XXXX"}.}
}
\value{
A named character string — the job state. Possible values:
\describe{
\item{\code{"idle"}}{Queued, waiting to be scheduled.}
\item{\code{"runnable"}}{Resources being allocated.}
\item{\code{"running"}}{Actively executing.}
\item{\code{"done"}}{Completed successfully.}
\item{\code{"failed"}}{Failed; see \code{attr(result, "failure_message")}.}
\item{\code{"terminated"}}{Manually terminated.}
}
}
\description{
Returns the current state of a job submitted by \code{extract_batch()}.
For failed jobs, the failure message is attached as an attribute.
}
\examples{
\dontrun{
job_id <- extract_batch(c(31, 53, 21022))
job_status(job_id)

s <- job_status(job_id)
if (s == "failed") cli::cli_inform(attr(s, "failure_message"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot_tableone}
\alias{plot_tableone}
\title{Publication-ready Table 1 (Baseline Characteristics)}
\usage{
plot_tableone(
  data,
  vars,
  strata = NULL,
  type = NULL,
  label = NULL,
  statistic = NULL,
  digits = NULL,
  percent = "column",
  missing = "no",
  add_p = TRUE,
  add_smd = FALSE,
  overall = FALSE,
  exclude_labels = NULL,
  theme = "lancet",
  label_width = 200,
  stat_width = 140,
  pvalue_width = 100,
  row_height = 8,
  save = FALSE,
  dest = NULL,
  png_scale = 2,
  pdf_width = NULL,
  pdf_height = NULL
)
}
\arguments{
\item{data}{data.frame. Input data containing all variables.}

\item{vars}{Character vector. Variable names to display in the table.}

\item{strata}{Character scalar. Column name for the grouping/stratification
variable. \code{NULL} produces an unstratified overall summary.}

\item{type}{Named list or \code{NULL}. Variable type overrides passed
directly to \code{\link[gtsummary]{tbl_summary}}.
E.g. \code{list(age = "continuous2")}.}

\item{label}{Named list / formula list or \code{NULL}. Label overrides.
E.g. \code{list(age ~ "Age (years)", sex ~ "Sex")}.}

\item{statistic}{Named list or \code{NULL}. Statistic format strings.
E.g. \code{list(all_continuous() ~ "{mean} ({sd})")}.}

\item{digits}{Named list or \code{NULL}. Decimal place overrides.
E.g. \code{list(age ~ 1)}.}

\item{percent}{Character. Percentage base for categorical variables:
\code{"column"} (default), \code{"row"}, or \code{"cell"}.}

\item{missing}{Character. Display of missing counts:
\code{"no"} (default), \code{"ifany"}, or \code{"always"}.}

\item{add_p}{Logical. Add a p-value column. Default \code{TRUE};
disabled with a warning if \code{strata} is \code{NULL}.}

\item{add_smd}{Logical. Add a standardised mean difference (SMD) column.
Continuous variables use Cohen's \emph{d}; categorical variables use
root-mean-square deviation (RMSD) of group proportions. SMD is shown
only on label rows. Default \code{FALSE}; disabled with a warning if
\code{strata} is \code{NULL}.}

\item{overall}{Logical. Add an Overall column when \code{strata} is set.
Default \code{FALSE}.}

\item{exclude_labels}{Character vector or \code{NULL}. Row labels to
remove from the rendered table, matched exactly against the
\code{label} column in the gtsummary table body.
E.g. \code{c("Unknown", "Missing")}.}

\item{theme}{Character. Visual theme preset. Only \code{"lancet"}
(default) is currently supported. Controls alternating row shading
(\code{#f8e8e7} / white), three-line borders, and 15 px font size.}

\item{label_width}{Integer. Label column width in pixels. Default \code{200}.}

\item{stat_width}{Integer. Statistics column(s) width in pixels.
Default \code{140}.}

\item{pvalue_width}{Integer. P-value and SMD column width in pixels.
Default \code{100}.}

\item{row_height}{Integer. Data row padding (top + bottom) in pixels.
Default \code{8}.}

\item{save}{Logical. Export the table to files. Default \code{FALSE}.}

\item{dest}{Character or \code{NULL}. File path without extension.
When \code{save = TRUE}, four files are written:
\code{<dest>.docx}, \code{<dest>.html}, \code{<dest>.pdf},
\code{<dest>.png}. Required when \code{save = TRUE}.}

\item{png_scale}{Numeric. Zoom factor for PNG export via \pkg{webshot2}.
Higher values produce larger, higher-resolution images. Default: \code{2}.}

\item{pdf_width}{Numeric or \code{NULL}. PDF paper width in inches passed to
\code{pagedown::chrome_print}. A larger value increases the page size so
more content fits on a single page. Default: \code{NULL} (Chrome default).}

\item{pdf_height}{Numeric or \code{NULL}. PDF paper height in inches.
Increase if the table is cut off across pages. Default: \code{NULL}.}
}
\value{
A \pkg{gt} table object, returned invisibly.
}
\description{
Generates a publication-quality baseline-characteristics table (Table 1)
using \pkg{gtsummary}, with optional SMD column, \emph{Lancet}-style
theming, and automatic export to four formats.
}
\details{
The following behaviours are fixed (not exposed as parameters):
\itemize{
\item Variable labels are \strong{bold} (\code{bold_labels()}).
\item P-values are formatted as \code{<0.001} or to 3 decimal places.
\item Significant p-values (\eqn{p < 0.05}) are \strong{bold}.
\item The p-value column header is rendered as \emph{P}-value.
\item The table uses a three-line (booktabs) border.
\item Saving always exports \strong{word}, \strong{html}, \strong{pdf},
and \strong{png}.
}
}
\examples{
library(gtsummary)
data(trial)

# Basic stratified table
plot_tableone(
  data   = trial,
  vars   = c("age", "marker", "grade"),
  strata = "trt",
  save   = FALSE
)

# With SMD, custom labels, exclude Unknown level
plot_tableone(
  data           = trial,
  vars           = c("age", "marker", "grade", "stage"),
  strata         = "trt",
  label          = list(age ~ "Age (years)", marker ~ "Marker level"),
  add_smd        = TRUE,
  exclude_labels = "Unknown",
  save           = FALSE
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract.R
\name{extract_ls}
\alias{extract_ls}
\title{List all approved fields in the UKB dataset}
\usage{
extract_ls(dataset = NULL, pattern = NULL, refresh = FALSE)
}
\arguments{
\item{dataset}{(character) Dataset file name, e.g.
\code{"app12345_20260101.dataset"}. Default: \code{NULL} (auto-detect).}

\item{pattern}{(character) Optional regex to filter results by
\code{field_name} or \code{title}. Default: \code{NULL}.}

\item{refresh}{(logical) Force re-fetch from cloud, ignoring cache.
Default: \code{FALSE}.}
}
\value{
A data.frame with columns:
\describe{
\item{field_name}{Full field name as used in extraction, e.g.
\code{"participant.p31"}, \code{"participant.p53_i0"}.}
\item{title}{Human-readable field description, e.g.
\code{"Sex"}, \code{"Date of attending assessment centre | Instance 0"}.}
}
}
\description{
Returns a data.frame of all fields available for extraction in the current
UKB project dataset. Fields reflect what has been approved for your project
— not all UKB fields are present.
}
\details{
Results are cached in the session after the first call. Subsequent calls
return instantly from cache. Use \code{refresh = TRUE} to force a new
network request (e.g. after switching projects).
}
\examples{
\dontrun{
# List all approved fields
extract_ls()

# Search by keyword
extract_ls(pattern = "cancer")
extract_ls(pattern = "p31|p53|p22009")

# Force refresh after switching projects
extract_ls(refresh = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/package.R
\docType{package}
\name{ukbflow-package}
\alias{ukbflow}
\alias{ukbflow-package}
\title{ukbflow: Streamlined Workflow for UK Biobank Data Extraction, Analysis, and Visualization}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Provides a streamlined workflow for UK Biobank cloud-based analysis on the Research Analysis Platform (RAP). Includes tools for phenotype extraction and decoding, variable derivation, survival and association analysis, genetic risk score computation, and publication-quality visualization. For details on the UK Biobank resource, see Bycroft et al. (2018) \doi{10.1038/s41586-018-0579-z}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/evanbio/ukbflow}
  \item \url{https://evanbio.github.io/ukbflow/}
  \item Report bugs at \url{https://github.com/evanbio/ukbflow/issues}
}

}
\author{
\strong{Maintainer}: Yibin Zhou \email{evanzhou.bio@gmail.com} (\href{https://orcid.org/0009-0009-4600-8175}{ORCID})

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/derive.R
\name{derive_age}
\alias{derive_age}
\title{Compute age at event for one or more UKB outcomes}
\usage{
derive_age(
  data,
  name,
  baseline_col,
  age_col,
  date_cols = NULL,
  status_cols = NULL
)
}
\arguments{
\item{data}{(data.frame or data.table) UKB phenotype data.}

\item{name}{(character) One or more output prefixes, e.g.
\code{c("disease", "disease_icd10", "outcome")}. Each produces \code{age_at_{name}}.}

\item{baseline_col}{(character) Name of the baseline date column
(e.g. \code{"date_baseline"}).}

\item{age_col}{(character) Name of the age-at-baseline column
(e.g. \code{"age_recruitment"}).}

\item{date_cols}{(character or NULL) Named character vector mapping each
name to its event date column, e.g.
\code{c(disease = "disease_date", outcome = "outcome_date")}. \code{NULL} (default)
triggers auto-detection as \code{{name}_date}.}

\item{status_cols}{(character or NULL) Named character vector mapping each
name to its status column. \code{NULL} (default) triggers auto-detection.}
}
\value{
The input \code{data} with one new \code{age_at_{name}} column
per entry in \code{name}, added in-place.
}
\description{
For each name in \code{name}, adds one column \code{age_at_{name}} (numeric,
years) computed as:
\deqn{age\_at\_event = age\_col + (event\_date - baseline\_date) / 365.25}
}
\details{
The value is \code{NA} for participants who did not experience the event
(status is \code{FALSE} / \code{0}) or who lack an event date.

\strong{Auto-detection per name} (when \code{date_cols} / \code{status_cols}
are \code{NULL}):
\itemize{
\item \code{date_col}   - looked up as \code{{name}_date}.
\item \code{status_col} - looked up first as \code{{name}_status}, then
as \code{{name}} (logical column); if neither exists all rows with a
non-\code{NA} date are treated as cases.
}

\strong{data.table pass-by-reference}: new columns are added in-place.
}
\examples{
dt <- ops_toy(scenario = "association", n = 100)
derive_age(dt, name = "dm", baseline_col = "p53_i0", age_col = "p21022")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auth.R
\name{auth_status}
\alias{auth_status}
\title{Check current DNAnexus authentication status}
\usage{
auth_status()
}
\value{
A named list with \code{user} and \code{project}.
}
\description{
Returns the current logged-in user and selected project.
}
\examples{
\dontrun{
auth_status()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract.R
\name{extract_pheno}
\alias{extract_pheno}
\title{Extract phenotype data from a UKB dataset}
\usage{
extract_pheno(field_id, dataset = NULL, timeout = 300)
}
\arguments{
\item{field_id}{(integer) Vector of UKB Field IDs to extract, e.g.
\code{c(31, 53, 22189)}. \code{eid} is always included automatically.}

\item{dataset}{(character) Dataset file name. Default: \code{NULL}
(auto-detect from project root).}

\item{timeout}{(integer) Extraction timeout in seconds. Default: \code{300}.}
}
\value{
A \code{data.table} with one row per participant. Column names
follow the \code{participant.p<id>_i<n>_a<m>} convention.
Fields not found are skipped with a warning.
}
\description{
Extracts phenotypic fields from the UKB Research Analysis Platform dataset
and returns a \code{data.table}. All instances and arrays are returned for
each requested field. Column names are kept as-is (e.g.
\code{participant.p53_i0}); use the \code{clean_} series for renaming.
}
\examples{
\dontrun{
df <- extract_pheno(c(31, 53, 21022))
df <- extract_pheno(c(31, 53, 20002), dataset = "app12345_20260101.dataset")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assoc.R
\name{assoc_subgroup}
\alias{assoc_subgroup}
\alias{assoc_sub}
\title{Subgroup association analysis with optional interaction test}
\usage{
assoc_subgroup(
  data,
  outcome_col,
  time_col = NULL,
  exposure_col,
  by,
  method = c("coxph", "logistic", "linear"),
  covariates = NULL,
  interaction = TRUE,
  conf_level = 0.95
)

assoc_sub(
  data,
  outcome_col,
  time_col = NULL,
  exposure_col,
  by,
  method = c("coxph", "logistic", "linear"),
  covariates = NULL,
  interaction = TRUE,
  conf_level = 0.95
)
}
\arguments{
\item{data}{(data.frame or data.table) Analysis dataset.}

\item{outcome_col}{(character) Outcome column name.}

\item{time_col}{(character or NULL) Follow-up time column (required when
\code{method = "coxph"}).}

\item{exposure_col}{(character) One or more exposure variable names.}

\item{by}{(character) Single stratification variable name. Its unique
non-NA values (or factor levels, in order) define the subgroups.
Should be a categorical or binary variable with a small number of
levels (e.g. sex, smoking status). Continuous variables are technically
permitted and the interaction LRT will still run, but per-unique-value
subgrouping is rarely meaningful in practice — use a pre-categorised
version (e.g. via \code{\link{derive_cut}}) instead.}

\item{method}{(character) Regression method: \code{"coxph"} (default),
\code{"logistic"}, or \code{"linear"}.}

\item{covariates}{(character or NULL) Covariate column names for the Fully
adjusted model. When \code{NULL}, only the Unadjusted model is run.}

\item{interaction}{(logical) Compute the LRT p-value for the exposure
\eqn{\times} by interaction on the full dataset. Default: \code{TRUE}.}

\item{conf_level}{(numeric) Confidence level. Default: \code{0.95}.}
}
\value{
A \code{data.table} with one row per subgroup level \eqn{\times}
exposure \eqn{\times} term \eqn{\times} model, containing:
\describe{
\item{\code{subgroup}}{Name of the \code{by} variable.}
\item{\code{subgroup_level}}{Factor: level of the \code{by} variable
(ordered by original level / sort order).}
\item{\code{exposure}}{Exposure variable name.}
\item{\code{term}}{Coefficient name from the fitted model.}
\item{\code{model}}{Ordered factor: \code{Unadjusted} <
\code{Fully adjusted}.}
\item{\code{n}}{Participants in model (after NA removal).}
\item{...}{Effect estimate columns: \code{HR}/\code{OR}/\code{beta},
\code{CI_lower}, \code{CI_upper}, \code{p_value}, and a formatted
label column. Cox models additionally include \code{n_events} and
\code{person_years}; logistic models include \code{n_cases}; linear
models include \code{se}.}
\item{\code{p_interaction}}{LRT p-value for the exposure \eqn{\times}
by interaction on the full dataset. Shared across all subgroup levels
for the same exposure \eqn{\times} model. \code{NA} when the
interaction model fails. Only present when
\code{interaction = TRUE}.}
}
}
\description{
Stratifies the dataset by a single grouping variable (\code{by}) and runs
the specified association model (\code{coxph}, \code{logistic}, or
\code{linear}) within each subgroup. Unlike \code{\link{assoc_coxph}} and
its siblings, there is no automatic age-and-sex-adjusted model: at the
subgroup level the variable that defines the stratum would have zero
variance, making the auto-detected adjustment meaningless. Accordingly,
this function does not accept a \code{base} argument. Instead, two
models are available:
}
\details{
\itemize{
\item \strong{Unadjusted} - always run (no covariates).
\item \strong{Fully adjusted} - run when \code{covariates} is non-NULL.
Users are responsible for excluding the \code{by} variable from
\code{covariates} (a warning is issued if it is included).
}

\strong{Interaction test}: when \code{interaction = TRUE} (default), a
likelihood ratio test (LRT) for the exposure \eqn{\times} by interaction is
computed on the \emph{full} dataset for each exposure \eqn{\times} model
combination and appended as \code{p_interaction}. LRT is preferred over
Wald because it handles factor, binary, and continuous \code{by} variables
uniformly without requiring the user to recode the \code{by} variable.
}
\examples{
dt <- ops_toy(scenario = "association", n = 500)
dt <- dt[dm_timing != 1L]

res <- assoc_subgroup(
  data         = dt,
  outcome_col  = "dm_status",
  time_col     = "dm_followup_years",
  exposure_col = "p20116_i0",
  by           = "p31",
  method       = "coxph",
  covariates   = c("bmi_cat", "tdi_cat")
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/derive.R
\name{derive_covariate}
\alias{derive_covariate}
\title{Prepare UKB covariates for analysis}
\usage{
derive_covariate(
  data,
  as_numeric = NULL,
  as_factor = NULL,
  factor_levels = NULL,
  max_levels = 5L
)
}
\arguments{
\item{data}{(data.frame or data.table) UKB data, typically output of
\code{\link{derive_missing}}.}

\item{as_numeric}{(character or NULL) Column names to convert to
\code{numeric}. Values that cannot be coerced (e.g. residual text) become
\code{NA} with a warning. Default: \code{NULL}.}

\item{as_factor}{(character or NULL) Column names to convert to
\code{factor}. Default levels are the sorted unique non-NA values unless
overridden by \code{factor_levels}. Default: \code{NULL}.}

\item{factor_levels}{(named list or NULL) Custom level ordering for specific
factor columns. Names must match entries in \code{as_factor}; values are
character vectors of levels in the desired order (first level = reference
group in regression). Columns not listed use default ordering.
Default: \code{NULL}.}

\item{max_levels}{(integer) Factor columns with more levels than this
threshold trigger a warning suggesting the user consider collapsing
categories. Default: \code{5L}.}
}
\value{
The input \code{data} with converted columns. Always returns a
\code{data.table}.
}
\description{
Converts decoded UKB columns to analysis-ready types: character-encoded
numeric fields to \code{numeric}, and categorical fields to \code{factor}.
Prints a concise summary for each converted column - mean / median / SD /
missing rate for numeric columns, and level counts for factor columns - so
you can verify distributions without leaving the pipeline.
}
\details{
\strong{data.table pass-by-reference}: when the input is a
\code{data.table}, modifications are made in-place. Pass
\code{data.table::copy(data)} to preserve the original.
}
\examples{
dt <- ops_toy(n = 100)
dt <- derive_missing(dt)
# messy_label is a character column with numeric-looking values ("-1", "999")
# mixed with non-parseable strings — demonstrates coercion and NA warnings
derive_covariate(dt,
  as_numeric = "messy_label",
  as_factor  = c("p31", "p20116_i0"),
  factor_levels = list(
    p20116_i0 = c("Never", "Previous", "Current")
  )
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assoc.R
\name{assoc_coxph}
\alias{assoc_coxph}
\alias{assoc_cox}
\title{Cox proportional hazards association analysis}
\usage{
assoc_coxph(
  data,
  outcome_col,
  time_col,
  exposure_col,
  covariates = NULL,
  base = TRUE,
  strata = NULL,
  conf_level = 0.95
)

assoc_cox(
  data,
  outcome_col,
  time_col,
  exposure_col,
  covariates = NULL,
  base = TRUE,
  strata = NULL,
  conf_level = 0.95
)
}
\arguments{
\item{data}{(data.frame or data.table) Analysis dataset. Must contain all
columns referenced by \code{outcome_col}, \code{time_col}, and
\code{exposure_col}.}

\item{outcome_col}{(character) Name of the event indicator column.
Accepts \code{logical} (\code{TRUE}/\code{FALSE}) or numeric/integer
(\code{0}/\code{1}).}

\item{time_col}{(character) Name of the follow-up time column (numeric,
in consistent units, e.g. years).}

\item{exposure_col}{(character) One or more exposure variable names.
Each variable is analysed separately; results are stacked row-wise.}

\item{covariates}{(character or NULL) Additional covariate column names for
the \strong{Fully adjusted} model (e.g.
\code{c("tdi", "smoking", paste0("pc", 1:10))}). When \code{NULL}
(default), the Fully adjusted model is not run.}

\item{base}{(logical) If \code{TRUE} (default), always include the
\strong{Unadjusted} and \strong{Age and sex adjusted} models in addition
to any user-specified \code{covariates} model. Set to \code{FALSE} to run
only the Fully adjusted model (requires \code{covariates} to be non-NULL).}

\item{strata}{(character or NULL) Optional stratification variable.
Passed to \code{survival::strata()} in the Cox formula.}

\item{conf_level}{(numeric) Confidence level for hazard ratio intervals.
Default: \code{0.95}.}
}
\value{
A \code{data.table} with one row per exposure \eqn{\times} term
\eqn{\times} model combination, and the following columns:
\describe{
\item{\code{exposure}}{Exposure variable name.}
\item{\code{term}}{Coefficient name as returned by \code{coxph} (e.g.
\code{"bmi_categoryObese"} for a factor, or the variable name itself
for numeric/binary exposures).}
\item{\code{model}}{Ordered factor: \code{Unadjusted} <
\code{Age and sex adjusted} < \code{Fully adjusted}.}
\item{\code{n}}{Number of participants included in the model (after
\code{NA} removal).}
\item{\code{n_events}}{Number of events in the model's analysis set
(after \code{NA} removal).}
\item{\code{person_years}}{Total person-years of follow-up in the
model's analysis set (rounded, after \code{NA} removal).}
\item{\code{HR}}{Hazard ratio (point estimate).}
\item{\code{CI_lower}}{Lower bound of the confidence interval.}
\item{\code{CI_upper}}{Upper bound of the confidence interval.}
\item{\code{p_value}}{Wald test p-value.}
\item{\code{HR_label}}{Formatted string, e.g.
\code{"1.23 (1.05-1.44)"}.}
}
}
\description{
Fits one or more Cox models for each exposure variable and returns a tidy
result table suitable for downstream forest plots. By default, two standard
adjustment models are always included alongside any user-specified model:
}
\details{
\itemize{
\item \strong{Unadjusted} - no covariates (crude).
\item \strong{Age and sex adjusted} - age + sex auto-detected from the
data via UKB field IDs (21022 and 31). Skipped with a warning if either
column cannot be found.
\item \strong{Fully adjusted} - the covariates supplied via the
\code{covariates} argument. Only run when \code{covariates} is non-NULL.
}

\strong{Outcome coding}: \code{outcome_col} may be \code{logical}
(\code{TRUE}/\code{FALSE}) or integer/numeric (\code{0}/\code{1}).
Logical values are converted to integer internally.

\strong{Exposure types supported}:
\itemize{
\item \emph{Binary} - \code{0}/\code{1} or \code{TRUE}/\code{FALSE};
produces one \code{term} row per model.
\item \emph{Factor} - produces one \code{term} row per non-reference level.
\item \emph{Numeric} (continuous) - produces one \code{term} row per model.
}
}
\examples{
dt <- ops_toy(scenario = "association", n = 500)
dt <- dt[dm_timing != 1L]

res <- assoc_coxph(
  data         = dt,
  outcome_col  = "dm_status",
  time_col     = "dm_followup_years",
  exposure_col = "p20116_i0",
  covariates   = c("bmi_cat", "tdi_cat"),
  base         = FALSE
)

# Fully adjusted only (skip Unadjusted + Age-sex)
res <- assoc_coxph(
  data         = dt,
  outcome_col  = "dm_status",
  time_col     = "dm_followup_years",
  exposure_col = "p20116_i0",
  covariates   = c("p21022", "p31", "bmi_cat", "tdi_cat"),
  base         = FALSE
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auth.R
\name{auth_select_project}
\alias{auth_select_project}
\title{Select a DNAnexus project}
\usage{
auth_select_project(project)
}
\arguments{
\item{project}{(character) Project ID in the form \code{"project-XXXXXXXXXXXX"}.}
}
\value{
Invisible TRUE on success.
}
\description{
Switches the active project context on the DNAnexus platform. Only project
IDs (e.g. \code{"project-XXXXXXXXXXXX"}) are accepted. Run
\code{\link[=auth_list_projects]{auth_list_projects()}} to find your project ID.
}
\examples{
\dontrun{
auth_select_project("project-XXXXXXXXXXXX")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decode.R
\name{decode_names}
\alias{decode_names}
\title{Rename UKB field ID columns to human-readable snake_case names}
\usage{
decode_names(data, max_nchar = 60L)
}
\arguments{
\item{data}{(data.frame or data.table) Data extracted from UKB-RAP via
\code{extract_pheno()} or \code{job_result()}.}

\item{max_nchar}{(integer) Column names longer than this value are flagged.
Default: \code{60}.}
}
\value{
The input \code{data} with column names replaced by snake_case
labels. Returns a \code{data.table} if the input is a \code{data.table}.
}
\description{
Renames columns from the raw UKB field ID format used by
\code{\link{extract_pheno}} (e.g. \code{participant.p31}) and
\code{\link{job_result}} (e.g. \code{p53_i0}) to human-readable
snake_case identifiers (e.g. \code{sex},
\code{date_of_attending_assessment_centre_i0}).
}
\details{
Column labels are taken from the UKB field title dictionary via
\code{\link{extract_ls}}. Both the dataset name and field list are cached
after the first call, so subsequent calls to \code{decode_names()} involve
no network requests.

When an auto-generated name exceeds \code{max_nchar} characters it is
flagged with a warning so you can decide whether to shorten it manually
with \code{names(data)[...] <- ...}. The function never truncates names
automatically, because the right short name depends on scientific context
that only you know.
}
\examples{
\dontrun{
df <- extract_pheno(c(31, 53, 21022))
df <- decode_names(df)
# participant.eid    → eid
# participant.p31    → sex
# participant.p21022 → age_at_recruitment
# participant.p53_i0 → date_of_attending_assessment_centre_i0  (warned if > 60)

# Shorten a long name afterwards
names(df)[names(df) == "date_of_attending_assessment_centre_i0"] <- "date_baseline"
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grs.R
\name{grs_standardize}
\alias{grs_standardize}
\alias{grs_zscore}
\title{Standardise GRS columns by Z-score transformation}
\usage{
grs_standardize(data, grs_cols = NULL)

grs_zscore(data, grs_cols = NULL)
}
\arguments{
\item{data}{A \code{data.frame} or \code{data.table} containing at least
one GRS column.}

\item{grs_cols}{Character vector of column names to standardise.
If \code{NULL} (default), all columns whose names contain \code{"grs"}
(case-insensitive) are selected automatically.}
}
\value{
The input \code{data} as a \code{data.table} with one additional
\code{_z} column per GRS column appended after its source column.
}
\description{
Adds a \code{_z} column for every selected GRS column:
\code{z = (x - mean(x)) / sd(x)}.  The original columns are kept
unchanged.  \code{\link{grs_zscore}} is an alias for this function.
}
\examples{
dt <- data.frame(
  IID   = 1:5,
  GRS_a = c(0.12, 0.34, 0.56, 0.23, 0.45),
  GRS_b = c(1.1,  0.9,  1.3,  0.8,  1.0)
)
grs_standardize(dt)
grs_zscore(dt)   # identical
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/derive.R
\name{derive_followup}
\alias{derive_followup}
\title{Compute follow-up end date and follow-up time for survival analysis}
\usage{
derive_followup(
  data,
  name,
  event_col,
  baseline_col,
  censor_date,
  death_col = NULL,
  lost_col = NULL
)
}
\arguments{
\item{data}{(data.frame or data.table) UKB phenotype data.}

\item{name}{(character) Output column prefix, e.g. \code{"outcome"} produces
\code{outcome_followup_end} and \code{outcome_followup_years}.}

\item{event_col}{(character) Name of the outcome event date column
(e.g. \code{"outcome_date"}).}

\item{baseline_col}{(character) Name of the baseline date column
(e.g. \code{"date_baseline"}).}

\item{censor_date}{(Date or character) Scalar administrative censoring date,
e.g. \code{as.Date("2022-06-01")}.  A character string in
\code{"YYYY-MM-DD"} format is also accepted.}

\item{death_col}{(character or NULL) Name of the death date column
(UKB field 40000).  \code{NULL} (default) triggers auto-detection via
the \code{\link{extract_ls}} cache; pass \code{FALSE} to explicitly
disable death as a competing end-point.}

\item{lost_col}{(character or NULL) Name of the lost-to-follow-up date
column (UKB field 191).  \code{NULL} (default) triggers auto-detection;
pass \code{FALSE} to explicitly disable.}
}
\value{
The input \code{data} with two new columns added in-place:
\code{{name}_followup_end} (IDate) and \code{{name}_followup_years}
(numeric).
}
\description{
Adds two columns to \code{data}:
\itemize{
\item \code{{name}_followup_end} (IDate) - the earliest of the outcome
event date, death date, lost-to-follow-up date, and the administrative
censoring date.
\item \code{{name}_followup_years} (numeric) - time in years from
\code{baseline_col} to \code{{name}_followup_end}.
}
}
\details{
\strong{data.table pass-by-reference}: when the input is a
\code{data.table}, new columns are added in-place via \code{:=}.
}
\examples{
\donttest{
dt <- ops_toy(n = 100)
derive_hes(dt, name = "htn", icd10 = "I10")
derive_followup(dt,
  name         = "htn_hes",
  event_col    = "htn_hes_date",
  baseline_col = "p53_i0",
  censor_date  = as.Date("2022-06-01"),
  death_col    = "p40000_i0",
  lost_col     = FALSE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/job.R
\name{job_wait}
\alias{job_wait}
\title{Wait for a DNAnexus job to finish}
\usage{
job_wait(job_id, interval = 30, timeout = Inf, verbose = TRUE)
}
\arguments{
\item{job_id}{(character) Job ID returned by \code{extract_batch()}.}

\item{interval}{(integer) Polling interval in seconds. Default: \code{30}.}

\item{timeout}{(numeric) Maximum wait time in seconds. Default: \code{Inf}
(wait indefinitely). On UKB RAP, jobs can stay in \code{"runnable"} for
several hours during peak times — set a finite value (e.g. \code{7200})
only if you need a hard deadline.}

\item{verbose}{(logical) Print state and elapsed time at each poll.
Default: \code{TRUE}.}
}
\value{
Invisibly returns the final state string (\code{"done"}).
}
\description{
Polls \code{job_status()} at regular intervals until the job reaches a
terminal state (\code{"done"}, \code{"failed"}, or \code{"terminated"}).
Stops with an error if the job fails, is terminated, or times out.
}
\examples{
\dontrun{
job_id <- extract_batch(c(31, 53, 21022))
job_wait(job_id)

# Read result immediately after completion (RAP only)
job_wait(job_id)
df <- job_result(job_id)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fetch.R
\name{fetch_tree}
\alias{fetch_tree}
\title{Print a remote RAP directory tree}
\usage{
fetch_tree(path = ".", max_depth = 2, verbose = TRUE)
}
\arguments{
\item{path}{(character) Remote root path. Default: \code{"."} (project
root). Both \code{"Bulk/"} and \code{"/Bulk/"} are accepted.}

\item{max_depth}{(integer) Maximum recursion depth. Default: \code{2}.}

\item{verbose}{(logical) Whether to print the tree to the console.
Default: \code{TRUE}.}
}
\value{
Invisibly returns a character vector of tree lines.
}
\description{
Displays the remote directory structure in a tree-like format by
recursively listing sub-folders up to \code{max_depth}. Analogous to
\code{file_tree()} but for remote project storage.
}
\section{Warning}{

Each level of recursion triggers one HTTPS API call per folder. Deep trees
(e.g. \code{max_depth > 3}) on large UKB projects may issue 100+ network
requests, causing the console to hang for tens of seconds or time out.
Keep \code{max_depth} at 2-3 for interactive use.
}

\examples{
\dontrun{
fetch_tree()
fetch_tree("Bulk/", max_depth = 2)
fetch_tree(verbose = FALSE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auth.R
\name{auth_logout}
\alias{auth_logout}
\title{Logout from DNAnexus}
\usage{
auth_logout()
}
\value{
Invisible TRUE on success.
}
\description{
Invalidates the current DNAnexus session on the remote platform. The local
token file is not removed but becomes invalid. A new token must be generated
from the DNAnexus platform before calling \code{\link[=auth_login]{auth_login()}} again.
}
\examples{
\dontrun{
auth_logout()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/derive.R
\name{derive_cut}
\alias{derive_cut}
\title{Cut a continuous UKB variable into quantile-based or custom groups}
\usage{
derive_cut(data, col, n, breaks = NULL, labels = NULL, name = NULL)
}
\arguments{
\item{data}{(data.frame or data.table) UKB data.}

\item{col}{(character) Name of the source numeric column.}

\item{n}{(integer) Number of groups. Any integer \code{>= 2} is accepted.
Commonly used values are \code{2}, \code{3}, \code{4}, \code{5}.}

\item{breaks}{(numeric vector or NULL) Interior cut points; length must
equal \code{n - 1}. When \code{NULL} (default), quantile-based equal-
frequency boundaries are computed automatically.}

\item{labels}{(character vector or NULL) Group labels of length \code{n}.
Defaults to \code{"Q1"}, \code{"Q2"}, \ldots, \code{"Qn"}.}

\item{name}{(character or NULL) Name for the new column. Defaults to
\code{"{col}_bi"} / \code{"{col}_tri"} / \code{"{col}_quad"} /
\code{"{col}_quin"} for \code{n} = 2 / 3 / 4 / 5; other values of
\code{n} produce \code{"{col}_g{n}"}.}
}
\value{
The input \code{data} with one new factor column appended. Always
returns a \code{data.table}.
}
\description{
Creates a new factor column by binning a continuous variable into \code{n}
groups. When \code{breaks} is omitted, group boundaries are derived from
quantiles of the observed data (equal-frequency binning). When
\code{breaks} is supplied, those values are used as interior cut points.
}
\details{
Before binning, a numeric summary (mean, median, SD, Q1, Q3, missing rate)
is printed for the source column. After binning, the group distribution is
printed via an internal summary helper.

Only one column can be processed per call; loop over columns explicitly
when binning multiple variables.

\strong{data.table pass-by-reference}: when the input is a
\code{data.table}, the new column is added in-place. Pass
\code{data.table::copy(data)} to preserve the original.
}
\examples{
dt <- ops_toy(n = 100)
derive_cut(dt, col = "p21001_i0", n = 3)

# Custom breaks and labels
derive_cut(dt, col = "p21001_i0", n = 3,
           breaks = c(25, 30),
           labels = c("<25", "25-30", "30+"),
           name   = "bmi_group")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grs.R
\name{grs_check}
\alias{grs_check}
\title{Check and export a GRS weights file}
\usage{
grs_check(file, dest = NULL)
}
\arguments{
\item{file}{Character scalar. Path to the input weights file.
Read via \code{data.table::fread} (format auto-detected; handles
CSV, TSV, space-delimited, etc.).}

\item{dest}{Character scalar or \code{NULL}. Output path for the validated,
space-delimited weights file. When \code{NULL} (default), no file is
written and the validated \code{data.table} is returned invisibly only.}
}
\value{
A \code{data.table} with columns \code{snp}, \code{effect_allele},
and \code{beta}, returned invisibly.
}
\description{
Reads a SNP weights file, validates its content, and writes a
plink2-compatible space-delimited output ready for upload to UKB RAP.
}
\details{
The input file must contain at least the three columns below (additional
columns are ignored):
\describe{
\item{\code{snp}}{SNP identifier, expected in \code{rs} + digits format.}
\item{\code{effect_allele}}{Effect allele; must be one of A / T / C / G.}
\item{\code{beta}}{Effect size (log-OR or beta coefficient); must be numeric.}
}

Checks performed:
\itemize{
\item Required columns present.
\item No \code{NA} values in the three required columns.
\item No duplicate \code{snp} identifiers.
\item \code{snp} matches \code{rs[0-9]+} pattern (warning if not).
\item \code{effect_allele} contains only A / T / C / G (warning if not).
\item \code{beta} is numeric (error if not).
}
}
\examples{
tmp_in <- tempfile(fileext = ".csv")
weights <- data.frame(
  snp           = c("rs1234567", "rs2345678", "rs3456789"),
  effect_allele = c("A", "T", "G"),
  beta          = c(0.12, -0.05, 0.23)
)
write.csv(weights, tmp_in, row.names = FALSE)

w <- grs_check(tmp_in)
w

# Save validated weights to a file
tmp_out <- tempfile(fileext = ".txt")
grs_check(tmp_in, dest = tmp_out)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/job.R
\name{job_ls}
\alias{job_ls}
\title{List recent DNAnexus jobs in the current project}
\usage{
job_ls(n = 20, state = NULL)
}
\arguments{
\item{n}{(integer) Maximum number of recent jobs to return. Must be a
single positive integer. Default: \code{20}.}

\item{state}{(character) Filter by state(s). Must be \code{NULL} or a
character vector of valid states:
\code{"idle"}, \code{"runnable"}, \code{"running"}, \code{"done"},
\code{"failed"}, \code{"terminated"}.
Default: \code{NULL} (return all).}
}
\value{
A data.frame with columns:
\describe{
\item{job_id}{Job ID string, e.g. \code{"job-XXXX"}.}
\item{name}{Job name (typically \code{"Table exporter"}).}
\item{state}{Current job state.}
\item{created}{Job creation time (\code{POSIXct}).}
\item{runtime}{Runtime string (e.g. \code{"0:04:36"}), \code{NA} if
still running.}
}
}
\description{
Returns a summary of the most recent jobs, optionally filtered by state.
Useful for quickly reviewing which jobs have completed, failed, or are
still running.
}
\examples{
\dontrun{
job_ls()
job_ls(n = 5)
job_ls(state = "failed")
job_ls(state = c("done", "failed"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/derive.R
\name{derive_first_occurrence}
\alias{derive_first_occurrence}
\title{Derive a binary disease flag from UKB First Occurrence fields}
\usage{
derive_first_occurrence(data, name, field, col = NULL)
}
\arguments{
\item{data}{(data.frame or data.table) UKB phenotype data.}

\item{name}{(character) Output column prefix, e.g. \code{"disease"} produces
\code{disease_fo} and \code{disease_fo_date}.}

\item{field}{(integer or character) UKB field ID of the First Occurrence
field, e.g. \code{131666} for E11 (type 2 diabetes).}

\item{col}{(character or NULL) Name of the source column in \code{data}.
When \code{NULL} (default) the column is detected automatically from
\code{field}.}
}
\value{
The input \code{data} (invisibly) with two new columns added
in-place: \code{{name}_fo} (logical) and \code{{name}_fo_date} (IDate).
}
\description{
UKB pre-computes the earliest recorded date for hundreds of ICD-10 chapters
and categories as \emph{First Occurrence} fields (\code{p131xxx}).  Each
field contains a single date per participant; no array or instance depth is
involved.  This function reads that date, converts it to \code{IDate},
and writes two analysis-ready columns:
}
\details{
\describe{
\item{\code{{name}_fo_date}}{Earliest First Occurrence date (\code{IDate}).
Values that cannot be coerced to a valid date (e.g. UKB error codes)
are silently set to \code{NA}.}
\item{\code{{name}_fo}}{Logical flag derived from
\code{{name}_fo_date}: \code{TRUE} if and only if a valid date exists.
This guarantees that every positive case has a usable date - essential
for time-to-event and prevalent/incident classification.}
}

\strong{Column detection}: the function locates the source column
automatically from \code{field}, handling both the raw format used by
\code{\link{extract_pheno}} (\code{participant.p131666}) and the
snake_case format produced by \code{\link{decode_names}}
(\code{date_e11_first_reported_type_2_diabetes}).  Supply \code{col}
to override auto-detection.

\strong{data.table pass-by-reference}: when the input is a
\code{data.table}, new columns are added in-place via \code{:=}.
The returned object and the original variable point to the same memory.
}
\examples{
dt <- ops_toy(n = 100)
derive_first_occurrence(dt, name = "outcome", field = 131742L, col = "p131742")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auth.R
\name{auth_list_projects}
\alias{auth_list_projects}
\title{List available DNAnexus projects}
\usage{
auth_list_projects()
}
\value{
A character vector of project names and IDs.
}
\description{
Returns a list of all projects accessible to the current user.
}
\examples{
\dontrun{
auth_list_projects()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/derive.R
\name{derive_cancer_registry}
\alias{derive_cancer_registry}
\title{Derive a binary disease flag from UKB cancer registry}
\usage{
derive_cancer_registry(
  data,
  name,
  icd10 = NULL,
  histology = NULL,
  behaviour = NULL,
  code_cols = NULL,
  hist_cols = NULL,
  behv_cols = NULL,
  date_cols = NULL
)
}
\arguments{
\item{data}{(data.frame or data.table) UKB phenotype data containing
cancer registry fields.}

\item{name}{(character) Output column prefix, e.g. \code{"outcome_invasive"}
produces \code{outcome_invasive_cancer} and
\code{outcome_invasive_cancer_date}.}

\item{icd10}{(character or NULL) Regular expression matched against the
ICD-10 code column (\code{p40006}).  \code{NULL} = no ICD-10 filter.
Examples: \code{"^C44"}, \code{"^(C44|D04)"}.}

\item{histology}{(integer vector or NULL) Histology codes to retain
(\code{p40011}).  \code{NULL} = no histology filter.
Example: \code{c(8070:8078, 8083, 8084)}.}

\item{behaviour}{(integer vector or NULL) Behaviour codes to retain
(\code{p40012}).  \code{NULL} = no behaviour filter.
Typical values: \code{3L} (invasive / malignant),
\code{2L} (in situ).}

\item{code_cols}{(character or NULL) Names of ICD-10 code columns
(\code{p40006_i*}).  \code{NULL} = auto-detect via field 40006.}

\item{hist_cols}{(character or NULL) Names of histology columns
(\code{p40011_i*}).  \code{NULL} = auto-detect via field 40011.}

\item{behv_cols}{(character or NULL) Names of behaviour columns
(\code{p40012_i*}).  \code{NULL} = auto-detect via field 40012.}

\item{date_cols}{(character or NULL) Names of diagnosis date columns
(\code{p40005_i*}).  \code{NULL} = auto-detect via field 40005.}
}
\value{
The input \code{data} with two new columns added in-place:
\code{{name}_cancer} (logical) and \code{{name}_cancer_date} (IDate).
Always returns a \code{data.table}.
}
\description{
The UK Biobank cancer registry links each participant's cancer diagnoses
from national cancer registries.  Each diagnosis is stored as a separate
instance with four parallel fields: ICD-10 code (\code{p40006}), histology
code (\code{p40011}), behaviour code (\code{p40012}), and diagnosis date
(\code{p40005}).  Unlike HES or self-report data, each instance holds
exactly one record - there is no array (\code{a*}) dimension.
}
\details{
All three filter arguments (\code{icd10}, \code{histology},
\code{behaviour}) are applied with AND logic: a record must satisfy
every non-\code{NULL} filter to be counted.  For OR conditions (e.g.\
D04 \emph{or} C44 with specific histology), call the function twice and
combine the resulting columns downstream.

\describe{
\item{\code{{name}_cancer}}{Logical flag: \code{TRUE} if any cancer
registry record satisfies all supplied filters.}
\item{\code{{name}_cancer_date}}{Earliest matching diagnosis date
(\code{IDate}).}
}
}
\examples{
\donttest{
dt <- ops_toy(n = 100)
derive_cancer_registry(dt, name = "skin",     icd10 = "^C44")
derive_cancer_registry(dt, name = "invasive", icd10 = "^C", behaviour = 3L)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decode.R
\name{decode_values}
\alias{decode_values}
\title{Decode UKB categorical column values using Showcase metadata}
\usage{
decode_values(data, metadata_dir = "data/metadata/")
}
\arguments{
\item{data}{(data.frame or data.table) Data from \code{extract_pheno()},
with column names in \code{participant.pXXXX} or \code{pXXXX_iX} format.}

\item{metadata_dir}{(character) Directory containing \code{field.tsv} and
\code{esimpint.tsv}. Default: \code{"data/metadata/"}.}
}
\value{
The input \code{data} with categorical columns replaced by character
labels. Returns a \code{data.table} if the input is a \code{data.table}.
}
\description{
Converts raw integer codes produced by \code{\link{extract_pheno}} into
human-readable labels for all categorical fields
(\code{value_type} 21 and 22), using the UKB Showcase encoding tables.
Continuous, text, date, and already-decoded columns are left unchanged.
}
\details{
This function requires two metadata files downloaded from the UKB Research
Analysis Platform:
\itemize{
\item \code{field.tsv} - maps field IDs to encoding IDs and value types.
\item \code{esimpint.tsv} - maps encoding ID + integer code to label.
}
Download them once with:
\preformatted{
fetch_metadata(dest_dir = "data/metadata")
}
Both files are cached in the session after the first read.

\strong{Call order}: use \code{decode_values()} \emph{before}
\code{\link{decode_names}}, so that column names still contain the numeric
field ID needed to look up the encoding.
}
\examples{
\dontrun{
# Download metadata once
fetch_metadata(dest_dir = "data/metadata")

# Recommended call order
df <- extract_pheno(c(31, 54, 20116, 21000))
df <- decode_values(df)                  # 0/1 → "Female"/"Male", etc.
df <- decode_names(df)                   # participant.p31 → sex
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fetch.R
\name{fetch_field}
\alias{fetch_field}
\title{Download the UKB field dictionary file}
\usage{
fetch_field(dest_dir, overwrite = FALSE, resume = FALSE, verbose = TRUE)
}
\arguments{
\item{dest_dir}{(character) Destination directory. Created automatically
if it does not exist.}

\item{overwrite}{(logical) Overwrite existing local file. Default:
\code{FALSE}.}

\item{resume}{(logical) Resume an interrupted download. Default:
\code{FALSE}.}

\item{verbose}{(logical) Show download progress. Default: \code{TRUE}.}
}
\value{
Invisibly returns the local file path as a character string.
}
\description{
Downloads \code{field.tsv} from the \code{Showcase metadata/} folder on the
DNAnexus Research Analysis Platform. This file contains the complete UKB
data dictionary: field IDs, titles, value types, and encoding references.
}
\examples{
\dontrun{
fetch_field(dest_dir = "metadata")
fetch_field(dest_dir = "metadata", overwrite = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/job.R
\name{job_result}
\alias{job_result}
\title{Load the result of a completed DNAnexus job into R}
\usage{
job_result(job_id)
}
\arguments{
\item{job_id}{(character) Job ID returned by \code{extract_batch()}.}
}
\value{
A \code{data.table} with one row per participant.
}
\description{
Reads the output CSV produced by \code{extract_batch()} directly from RAP
project storage and returns a \code{data.table}. Must be run inside the
RAP environment.
}
\examples{
\dontrun{
job_id <- extract_batch(c(31, 53, 21022))
job_wait(job_id)
df <- job_result(job_id)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/derive.R
\name{derive_death_registry}
\alias{derive_death_registry}
\title{Derive a binary disease flag from UKB death registry}
\usage{
derive_death_registry(
  data,
  name,
  icd10,
  match = c("prefix", "exact", "regex"),
  primary_cols = NULL,
  secondary_cols = NULL,
  date_cols = NULL
)
}
\arguments{
\item{data}{(data.frame or data.table) UKB phenotype data containing
death registry fields.}

\item{name}{(character) Output column prefix, e.g. \code{"disease"} produces
\code{disease_death} and \code{disease_death_date}.}

\item{icd10}{(character) ICD-10 code(s) to match.  For \code{"prefix"}
and \code{"exact"}, supply a vector such as \code{c("L20", "L21")}.
For \code{"regex"}, supply a single regex string.}

\item{match}{(character) Matching strategy: \code{"prefix"} (default),
\code{"exact"}, or \code{"regex"}.}

\item{primary_cols}{(character or NULL) Names of primary cause columns
(\code{p40001_i*}).  \code{NULL} = auto-detect via field 40001.}

\item{secondary_cols}{(character or NULL) Names of secondary cause
columns (\code{p40002_i*_a*}).  \code{NULL} = auto-detect via field
40002.}

\item{date_cols}{(character or NULL) Names of death date columns
(\code{p40000_i*}).  \code{NULL} = auto-detect via field 40000.}
}
\value{
The input \code{data} with two new columns added in-place:
\code{{name}_death} (logical) and \code{{name}_death_date} (IDate).
Always returns a \code{data.table}.
}
\description{
Death registry records store the underlying (primary) cause of death in
field \code{p40001} and contributory (secondary) causes in field
\code{p40002}, both coded in ICD-10.  The date of death is in field
\code{p40000}.  All three fields have an instance dimension (\code{i0},
\code{i1}) reflecting potential amendments; \code{p40002} additionally
has an array dimension (\code{a0}, \code{a1}, \ldots).
}
\details{
\describe{
\item{\code{{name}_death}}{Logical flag: \code{TRUE} if any death
registry record (primary or secondary cause) contains a matching
ICD-10 code.}
\item{\code{{name}_death_date}}{Earliest death date across matching
instances (\code{IDate}).  Note: this is the \emph{date of death},
not onset date.}
}
}
\examples{
\donttest{
dt <- ops_toy(n = 100)
derive_death_registry(dt, name = "mi",   icd10 = "I21")
derive_death_registry(dt, name = "lung", icd10 = "C34")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/derive.R
\name{derive_selfreport}
\alias{derive_selfreport}
\title{Define a self-reported phenotype from UKB touchscreen data}
\usage{
derive_selfreport(
  data,
  name,
  regex,
  field = c("noncancer", "cancer"),
  ignore_case = TRUE,
  disease_cols = NULL,
  date_cols = NULL,
  visit_cols = NULL
)
}
\arguments{
\item{data}{(data.frame or data.table) UKB data containing self-report
fields.}

\item{name}{(character) Output column name prefix, e.g. \code{"disease"} or
\code{"outcome"}.}

\item{regex}{(character) Regular expression matched against disease text
values (after \code{tolower()}), e.g. \code{"^diabetes$"}.}

\item{field}{(character) Self-report field type: \code{"noncancer"}
(p20002 / p20008) or \code{"cancer"} (p20001 / p20006).}

\item{ignore_case}{(logical) Should regex matching ignore case?
Default: \code{TRUE}.}

\item{disease_cols}{(character or NULL) Column name(s) containing disease
text (p20002 or p20001). \code{NULL} = auto-detect via
\code{extract_ls()} cache.}

\item{date_cols}{(character or NULL) Column name(s) containing the
self-report year-month date (p20008 or p20006). \code{NULL} =
auto-detect.}

\item{visit_cols}{(character or NULL) Column name(s) containing the
baseline assessment date (p53). \code{NULL} = auto-detect.}
}
\value{
The input \code{data} with two new columns appended in-place:
\code{{name}_selfreport} (logical) and
\code{{name}_selfreport_date} (IDate). Always returns a
\code{data.table}.
}
\description{
Searches UKB self-reported illness fields across all instances and arrays,
matches records against a user-supplied regex, parses the associated
year-month date, and appends two columns to the data:
\code{{name}_selfreport} (logical) and \code{{name}_selfreport_date}
(\code{IDate}, earliest matching instance).
}
\details{
The function auto-detects the relevant columns using the
\code{extract_ls()} field dictionary cache (populated by
\code{extract_ls()} or \code{extract_pheno()}). The three \code{_cols}
parameters let you override auto-detection when column names have been
customised (e.g. after \code{decode_names()}).

\strong{Field mapping by \code{field}:}
\itemize{
\item \code{"noncancer"}: disease text = p20002, date = p20008.
\item \code{"cancer"}:    disease text = p20001, date = p20006.
}
Baseline visit date (p53) is used as a fallback when no specific
diagnosis date is recorded.

\strong{data.table pass-by-reference}: new columns are added in-place.
Pass \code{data.table::copy(data)} to preserve the original.
}
\examples{
\donttest{
dt <- ops_toy(n = 100)
# disease_cols / date_cols / visit_cols can be omitted — auto-detected from
# column names matching p20002_* / p20008_* / p53_* respectively
derive_selfreport(dt, name = "htn", regex = "hypertension",
                  field = "noncancer")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/derive.R
\name{derive_missing}
\alias{derive_missing}
\title{Handle informative missing labels in UKB decoded data}
\usage{
derive_missing(
  data,
  cols = tidyselect::everything(),
  action = c("na", "unknown"),
  extra_labels = NULL
)
}
\arguments{
\item{data}{(data.frame or data.table) Decoded UKB data, typically output
of \code{\link{decode_values}} followed by \code{\link{decode_names}}.}

\item{cols}{(tidyselect) Columns to process. Default: \code{everything()}
(all columns). Non-character columns in the selection are silently skipped.}

\item{action}{(character) One of \code{"na"} (default) or \code{"unknown"}.
\itemize{
\item \code{"na"}: convert all informative missing labels to \code{NA}.
\item \code{"unknown"}: convert informative missing labels to
\code{"Unknown"}, preserving them as a distinct model category.
}
Empty strings are always converted to \code{NA} regardless of this
parameter.}

\item{extra_labels}{(character or NULL) Additional labels to treat as
informative missing, appended to the built-in list. Default: \code{NULL}.}
}
\value{
The input \code{data} with missing labels replaced in-place.
Always returns a \code{data.table}.
}
\description{
After \code{\link{decode_values}} converts categorical codes to character
labels, some values represent meaningful non-response rather than true data:
\code{"Do not know"}, \code{"Prefer not to answer"}, and
\code{"Prefer not to say"}. This function either converts them to \code{NA}
(for complete-case analysis) or retains them as \code{"Unknown"} (to
preserve the informative missingness as a model category).
}
\details{
Empty strings (\code{""}) are always converted to \code{NA} regardless of
\code{action}, as they carry no informational content.

Only character columns are modified; numeric, integer, Date, and logical
columns are silently skipped.

\strong{data.table pass-by-reference}: when the input is a
\code{data.table}, modifications are made in-place via
\code{\link[data.table]{set}}. The returned object and the original variable
point to the same memory. If you need to preserve the original, pass
\code{data.table::copy(data)} instead.
}
\examples{
dt <- ops_toy(n = 100)
derive_missing(dt)

# Retain informative non-response as a model category
derive_missing(dt, action = "unknown")

# Add a custom label
derive_missing(dt, extra_labels = "Not applicable")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assoc.R
\name{assoc_competing}
\alias{assoc_competing}
\alias{assoc_fg}
\title{Fine-Gray competing risks association analysis}
\usage{
assoc_competing(
  data,
  outcome_col,
  time_col,
  exposure_col,
  compete_col = NULL,
  event_val = 1L,
  compete_val = 2L,
  covariates = NULL,
  base = TRUE,
  conf_level = 0.95
)

assoc_fg(
  data,
  outcome_col,
  time_col,
  exposure_col,
  compete_col = NULL,
  event_val = 1L,
  compete_val = 2L,
  covariates = NULL,
  base = TRUE,
  conf_level = 0.95
)
}
\arguments{
\item{data}{(data.frame or data.table) Analysis dataset.}

\item{outcome_col}{(character) Primary event column. In Mode A: a
multi-value column (any integer or character codes). In Mode B: a 0/1
binary column.}

\item{time_col}{(character) Follow-up time column (numeric, e.g. years).}

\item{exposure_col}{(character) One or more exposure variable names.}

\item{compete_col}{(character or NULL) Mode B only: name of the 0/1
competing event column. When \code{NULL} (default), Mode A is used.}

\item{event_val}{Scalar value in \code{outcome_col} indicating the primary
event (Mode A only). Default: \code{1L}.}

\item{compete_val}{Scalar value in \code{outcome_col} indicating the
competing event (Mode A only). Default: \code{2L}.}

\item{covariates}{(character or NULL) Covariate names for the Fully
adjusted model. When \code{NULL}, only Unadjusted (and Age/sex adjusted
if \code{base = TRUE}) are run.}

\item{base}{(logical) Whether to auto-detect age and sex columns and include
an Age and sex adjusted model. Default: \code{TRUE}.}

\item{conf_level}{(numeric) Confidence level for SHR intervals.
Default: \code{0.95}.}
}
\value{
A \code{data.table} with one row per exposure \eqn{\times} term
\eqn{\times} model combination:
\describe{
\item{\code{exposure}}{Exposure variable name.}
\item{\code{term}}{Coefficient name as returned by \code{coxph}.}
\item{\code{model}}{Ordered factor: \code{Unadjusted} <
\code{Age and sex adjusted} < \code{Fully adjusted}.}
\item{\code{n}}{Participants in the model (after NA removal).}
\item{\code{n_events}}{Primary events in the analysis set.}
\item{\code{n_compete}}{Competing events in the analysis set.}
\item{\code{SHR}}{Subdistribution hazard ratio.}
\item{\code{CI_lower}}{Lower CI bound.}
\item{\code{CI_upper}}{Upper CI bound.}
\item{\code{p_value}}{Robust z-test p-value from weighted Cox.}
\item{\code{SHR_label}}{Formatted string, e.g. \code{"1.23 (1.05-1.44)"}.}
}
}
\description{
Fits a Fine-Gray subdistribution hazard model (via
\code{survival::finegray()} + weighted \code{coxph()}) for each exposure
variable and returns a tidy result table with subdistribution hazard ratios
(SHR).
}
\details{
Two input modes are supported depending on how the outcome is coded in your
dataset:

\describe{
\item{\strong{Mode A - single multi-value column}}{
\code{compete_col = NULL} (default). \code{outcome_col} contains all
event codes in one column (e.g. \code{0}/\code{1}/\code{2}/\code{3}).
Use \code{event_val} and \code{compete_val} to identify the event of
interest and the competing event; all other values are treated as
censored. Example: UKB \code{censoring_type} where 1 = event, 2 = death
(competing), 0/3 = censored.
}
\item{\strong{Mode B - dual binary columns}}{
\code{compete_col} is the name of a separate 0/1 column for the
competing event. \code{outcome_col} is a 0/1 column for the primary
event. When both are 1 for the same participant, the primary event takes
priority. Example: \code{outcome_col = "copd_status"},
\code{compete_col = "death_status"}.
}
}

Internally both modes are converted to a three-level factor
\code{c("censor", "event", "compete")} before being passed to
\code{finegray()}.

Three adjustment models are produced (where data allow):
\itemize{
\item \strong{Unadjusted} - always included.
\item \strong{Age and sex adjusted} - when \code{base = TRUE} and
age/sex columns are detected.
\item \strong{Fully adjusted} - when \code{covariates} is non-NULL.
}
}
\examples{
\donttest{
dt <- ops_toy(scenario = "association", n = 500)
dt <- dt[dm_timing != 1L & htn_timing != 1L]

res <- assoc_competing(
  data         = dt,
  outcome_col  = "dm_status",
  time_col     = "dm_followup_years",
  exposure_col = "p20116_i0",
  compete_col  = "htn_status",
  covariates   = c("bmi_cat", "tdi_cat")
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/job.R
\name{job_path}
\alias{job_path}
\title{Get the RAP file path of a completed DNAnexus job output}
\usage{
job_path(job_id)
}
\arguments{
\item{job_id}{(character) Job ID returned by \code{extract_batch()}.}
}
\value{
A character string — the absolute path to the output CSV under
\code{/mnt/project/}.
}
\description{
Returns the absolute \code{/mnt/project/} path of the CSV produced by
\code{extract_batch()}. Use this to read the file directly on the RAP
without downloading.
}
\examples{
\dontrun{
path <- job_path(job_id)
df   <- data.table::fread(path)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract.R
\name{extract_batch}
\alias{extract_batch}
\title{Submit a large-scale phenotype extraction job via table-exporter}
\usage{
extract_batch(
  field_id,
  dataset = NULL,
  file = NULL,
  instance_type = NULL,
  priority = c("low", "high")
)
}
\arguments{
\item{field_id}{(integer) Vector of UKB Field IDs to extract.
\code{eid} is always included automatically.}

\item{dataset}{(character) Dataset file name. Default: \code{NULL}
(auto-detect from project root).}

\item{file}{(character) Output file name on the cloud (without extension),
e.g. \code{"ad_cscc_pheno"}. Default: \code{NULL} (auto-generate as
\code{"ukb_pheno_YYYYMMDD_HHMMSS"} to avoid same-day collisions).}

\item{instance_type}{(character) DNAnexus instance type, e.g.
\code{"mem1_ssd1_v2_x16"}. Default: \code{NULL} (auto-select:
\code{x4} for up to 20 cols, \code{x8} for up to 100 cols,
\code{x16} for up to 500 cols, \code{x36} for more than 500 cols).}

\item{priority}{(character) Job scheduling priority. \code{"low"}
(recommended, cheaper) or \code{"high"} (faster queue). Default:
\code{"low"}.}
}
\value{
Invisibly returns the job ID string (e.g. \code{"job-XXXX"}).
}
\description{
Submits an asynchronous table-exporter job on the DNAnexus Research
Analysis Platform for large-scale phenotype extraction. Use this instead
of \code{extract_pheno()} when extracting many fields (e.g. 50+).
}
\details{
The job runs on the cloud and typically completes in 20-40 minutes.
Monitor progress and retrieve results using the \code{job_} series.
}
\examples{
\dontrun{
job_id <- extract_batch(core_field_ids)
job_id <- extract_batch(core_field_ids, file = "ad_cscc_pheno")
job_id <- extract_batch(core_field_ids, priority = "high")
# Monitor: job_status(job_id)
# Download: job_result(job_id, dest = "data/pheno.csv")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ops.R
\name{ops_toy}
\alias{ops_toy}
\title{Generate toy UKB-like data for testing and development}
\usage{
ops_toy(scenario = "cohort", n = 1000L, seed = 42L)
}
\arguments{
\item{scenario}{(character) Data structure to generate:
\itemize{
\item \code{"cohort"}: wide participant-level table with raw UKB field columns for
the full \verb{derive_*} → \verb{assoc_*} → \verb{plot_*} pipeline.
\item \code{"association"}: analysis-ready table with covariates already as factors,
BMI/TDI binned, and two pre-derived disease outcomes (\verb{dm_*}, \verb{htn_*})
including status, date, timing, and follow-up columns. Use this for
\verb{assoc_*} examples and testing without running the derive pipeline.
\item \code{"forest"}: association results table matching \code{assoc_coxph()} output,
for testing \code{plot_forest()}. \code{n} = number of exposures (default 8).
}}

\item{n}{(integer) Number of participants (or exposures for \code{"forest"}).
Default \code{1000L} for \code{"cohort"}, \code{2000L} for \code{"association"}, \code{8L} for
\code{"forest"}.}

\item{seed}{(integer or NULL) Random seed for reproducibility. Pass \code{NULL}
for a different dataset on every call. Default \code{42L}.}
}
\value{
A \code{data.table} with UKB-style column names. See Details for the
columns included in each scenario.
}
\description{
Creates a small, synthetic dataset that mimics the structure of UK Biobank
phenotype data on the RAP. Useful for developing and testing \verb{derive_*},
\verb{assoc_*}, and \verb{plot_*} functions without requiring real UKB data access.
}
\details{
This dataset is entirely synthetic. Column names follow RAP conventions
(e.g. \code{p41270}, \code{p20002_i0_a0}).
\subsection{scenario = "cohort"}{

Includes the following column groups:
\itemize{
\item \strong{Demographics}: \code{eid}, \code{p31}, \code{p34}, \code{p53_i0}, \code{p21022}
\item \strong{Covariates}: \code{p21001_i0}, \code{p20116_i0}, \code{p1558_i0}, \code{p21000_i0},
\code{p22189}, \code{p54_i0}
\item \strong{Genetic PCs}: \code{p22009_a1} – \code{p22009_a10}
\item \strong{Self-report disease}: \verb{p20002_i0_a0–a4}, \verb{p20008_i0_a0–a4}
\item \strong{Self-report cancer}: \verb{p20001_i0_a0–a4}, \verb{p20006_i0_a0–a4}
\item \strong{HES}: \code{p41270} (JSON array), \verb{p41280_a0–a8}
\item \strong{Cancer registry}: \verb{p40006_i0–i2}, \verb{p40011_i0–i2}, \verb{p40012_i0–i2},
\verb{p40005_i0–i2}
\item \strong{Death registry}: \code{p40001_i0}, \verb{p40002_i0_a0–a2}, \code{p40000_i0}
\item \strong{First occurrence}: \code{p131742}
\item \strong{GRS}: \code{grs_bmi}, \code{grs_raw}, \code{grs_finngen}
\item \strong{Messy columns}: \code{messy_allna}, \code{messy_empty}, \code{messy_label}
}
}

\subsection{scenario = "association"}{

Analysis-ready table. All derive inputs (raw arrays, HES JSON, registry
fields) are omitted; derive outputs are pre-computed with internally
consistent relationships:
\itemize{
\item \strong{Demographics}: \code{eid}, \code{p31} (factor), \code{p53_i0} (IDate), \code{p21022}
\item \strong{Covariates}: \code{p21001_i0}, \code{bmi_cat} (factor, derived from \code{p21001_i0}),
\code{p20116_i0} (factor), \code{p1558_i0} (factor), \code{p21000_i0} (factor),
\code{p22189}, \code{tdi_cat} (factor, derived from \code{p22189} quartiles), \code{p54_i0}
(factor)
\item \strong{Genetic PCs}: \code{p22009_a1} – \code{p22009_a10}
\item \strong{GRS}: \code{grs_bmi} (continuous exposure)
\item \strong{DM outcome}: \code{dm_status}, \code{dm_date}, \code{dm_timing}, \code{dm_followup_end},
\code{dm_followup_years} (type 2 diabetes, ~12\% prevalence)
\item \strong{HTN outcome}: \code{htn_status}, \code{htn_date}, \code{htn_timing},
\code{htn_followup_end}, \code{htn_followup_years} (hypertension, ~28\% prevalence)
}

Internal relationships guaranteed:
\itemize{
\item \code{bmi_cat} is cut from \code{p21001_i0} (breaks 18.5 / 25 / 30)
\item \code{tdi_cat} is cut from \code{p22189} quartiles
\item \code{dm_date} is \code{NA} iff \code{dm_status = FALSE}
\item \code{dm_timing}: 0 = no disease, 1 = prevalent, 2 = incident, \code{NA} = no date
\item \code{dm_followup_years} is \code{NA} for prevalent cases (\code{dm_timing == 1})
}
}
}
\examples{
# cohort: raw UKB-style columns, feed into derive pipeline
dt <- ops_toy(n = 100)
dt <- derive_missing(dt)

# association: analysis-ready, feed directly into assoc_* functions
dt <- ops_toy(scenario = "association", n = 500)
dt <- dt[dm_timing != 1L]   # exclude prevalent cases

# forest: results table for plot_forest()
dt <- ops_toy(scenario = "forest")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fetch.R
\name{fetch_ls}
\alias{fetch_ls}
\title{List files and folders at a remote RAP path}
\usage{
fetch_ls(path = ".", type = "all", pattern = NULL)
}
\arguments{
\item{path}{(character) Remote path to list. Default: \code{"."} (project
root). Both \code{"Bulk/"} and \code{"/Bulk/"} are accepted.}

\item{type}{(character) Filter results by entry type: \code{"all"}
(default), \code{"file"}, or \code{"folder"}.}

\item{pattern}{(character) Optional regex to filter by name, e.g.
\code{"\\.bed$"}. Default: \code{NULL}.}
}
\value{
A data.frame with columns:
\describe{
\item{name}{Entry name (no trailing slash for folders).}
\item{type}{\code{"file"} or \code{"folder"}.}
\item{size}{File size string (e.g. \code{"120.94 GB"}), \code{NA} for
folders or non-file objects.}
\item{modified}{Last modified time (\code{POSIXct}), \code{NA} for
folders.}
}
}
\description{
Returns a structured data.frame describing the contents of a remote
DNAnexus Research Analysis Platform (RAP) directory. Analogous to
\code{file_info()} but for remote project storage.
}
\examples{
\dontrun{
fetch_ls()
fetch_ls("Showcase metadata/", type = "file")
fetch_ls("results/", pattern = "\\\\.csv$")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/derive.R
\name{derive_case}
\alias{derive_case}
\title{Combine self-report and ICD-10 sources into a unified case definition}
\usage{
derive_case(
  data,
  name,
  icd10_col = NULL,
  selfreport_col = NULL,
  icd10_date_col = NULL,
  selfreport_date_col = NULL
)
}
\arguments{
\item{data}{(data.frame or data.table) UKB phenotype data.}

\item{name}{(character) Column prefix used both to locate the default
input columns and to name the output columns.  Defaults:
\code{{name}_icd10}, \code{{name}_selfreport},
\code{{name}_icd10_date}, \code{{name}_selfreport_date}.}

\item{icd10_col}{(character or NULL) Name of the ICD-10 status column.
\code{NULL} = \code{paste0(name, "_icd10")}.}

\item{selfreport_col}{(character or NULL) Name of the self-report status
column.  \code{NULL} = \code{paste0(name, "_selfreport")}.}

\item{icd10_date_col}{(character or NULL) Name of the ICD-10 date column.
\code{NULL} = \code{paste0(name, "_icd10_date")}.}

\item{selfreport_date_col}{(character or NULL) Name of the self-report date
column.  \code{NULL} = \code{paste0(name, "_selfreport_date")}.}
}
\value{
The input \code{data} with two new columns added in-place:
\code{{name}_status} (logical) and \code{{name}_date} (IDate).
Always returns a \code{data.table}.
}
\description{
Takes the self-report flag and ICD-10 flag produced by
\code{\link{derive_selfreport}} and \code{\link{derive_icd10}} (or any
pair of logical columns) and merges them into a single logical case status
and earliest date.
}
\details{
\describe{
\item{\code{{name}_status}}{Logical: \code{TRUE} if positive in at
least one source.}
\item{\code{{name}_date}}{Earliest diagnosis/report date across both
sources (\code{IDate}).}
}
}
\examples{
\donttest{
dt <- ops_toy(n = 100)
derive_selfreport(dt, name = "htn", regex = "hypertension",
                  field        = "noncancer",
                  disease_cols = paste0("p20002_i0_a", 0:4),
                  date_cols    = paste0("p20008_i0_a", 0:4),
                  visit_cols   = "p53_i0")
derive_icd10(dt, name = "htn", icd10 = "I10", source = c("hes", "death"))
derive_case(dt, name = "htn")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assoc.R
\name{assoc_logistic}
\alias{assoc_logistic}
\alias{assoc_logit}
\title{Logistic regression association analysis}
\usage{
assoc_logistic(
  data,
  outcome_col,
  exposure_col,
  covariates = NULL,
  base = TRUE,
  ci_method = c("wald", "profile"),
  conf_level = 0.95
)

assoc_logit(
  data,
  outcome_col,
  exposure_col,
  covariates = NULL,
  base = TRUE,
  ci_method = c("wald", "profile"),
  conf_level = 0.95
)
}
\arguments{
\item{data}{(data.frame or data.table) Analysis dataset.}

\item{outcome_col}{(character) Binary outcome column (\code{0}/\code{1} or
\code{TRUE}/\code{FALSE}).}

\item{exposure_col}{(character) One or more exposure variable names.}

\item{covariates}{(character or NULL) Covariate column names for the
\strong{Fully adjusted} model. Default: \code{NULL}.}

\item{base}{(logical) Include \strong{Unadjusted} and \strong{Age and sex
adjusted} models. Default: \code{TRUE}.}

\item{ci_method}{(character) CI calculation method: \code{"wald"} (default)
or \code{"profile"}.}

\item{conf_level}{(numeric) Confidence level. Default: \code{0.95}.}
}
\value{
A \code{data.table} with one row per exposure \eqn{\times} term
\eqn{\times} model combination, and columns:
\describe{
\item{\code{exposure}}{Exposure variable name.}
\item{\code{term}}{Coefficient name (e.g. \code{"bmi_categoryObese"}).}
\item{\code{model}}{Ordered factor: \code{Unadjusted} <
\code{Age and sex adjusted} < \code{Fully adjusted}.}
\item{\code{n}}{Participants in model (after NA removal).}
\item{\code{n_cases}}{Number of cases (outcome = 1) in model.}
\item{\code{OR}}{Odds ratio (point estimate).}
\item{\code{CI_lower}}{Lower confidence bound.}
\item{\code{CI_upper}}{Upper confidence bound.}
\item{\code{p_value}}{Wald test p-value.}
\item{\code{OR_label}}{Formatted string, e.g. \code{"1.23 (1.05-1.44)"}.}
}
}
\description{
Fits one or more logistic regression models for each exposure variable and
returns a tidy result table suitable for downstream forest plots. By default,
two standard adjustment models are always included:
}
\details{
\itemize{
\item \strong{Unadjusted} - no covariates (crude).
\item \strong{Age and sex adjusted} - age + sex auto-detected from the
data via UKB field IDs (21022 and 31). Skipped with a warning if either
column cannot be found.
\item \strong{Fully adjusted} - the covariates supplied via the
\code{covariates} argument. Only run when \code{covariates} is non-NULL.
}

\strong{Outcome coding}: \code{outcome_col} may be \code{logical}
(\code{TRUE}/\code{FALSE}) or integer/numeric (\code{0}/\code{1}).
Logical values are converted to integer internally.

\strong{CI methods}:
\itemize{
\item \code{"wald"} (default) - fast, appropriate for large UKB samples.
\item \code{"profile"} - profile likelihood CI via \code{confint.glm()};
slower but more accurate for small or sparse data.
}
}
\examples{
dt <- ops_toy(scenario = "association", n = 500)

res <- assoc_logistic(
  data         = dt,
  outcome_col  = "dm_status",
  exposure_col = "p20116_i0",
  covariates   = c("bmi_cat", "tdi_cat"),
  base         = FALSE
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot_forest}
\alias{plot_forest}
\title{Publication-ready forest plot}
\usage{
plot_forest(
  data,
  est,
  lower,
  upper,
  ci_column = 2L,
  ref_line = 1,
  xlim = NULL,
  ticks_at = NULL,
  arrow_lab = c("Lower risk", "Higher risk"),
  header = NULL,
  indent = NULL,
  bold_label = NULL,
  ci_col = "black",
  ci_sizes = 0.6,
  ci_Theight = 0.2,
  ci_digits = 2L,
  ci_sep = ", ",
  p_cols = NULL,
  p_digits = 3L,
  bold_p = TRUE,
  p_threshold = 0.05,
  align = NULL,
  background = "zebra",
  bg_col = "#F0F0F0",
  border = "three_line",
  border_width = 3,
  row_height = NULL,
  col_width = NULL,
  save = FALSE,
  dest = NULL,
  save_width = 20,
  save_height = NULL,
  theme = "default"
)
}
\arguments{
\item{data}{data.frame. First column must be the label column (\code{item}).
Additional columns are displayed as-is (character) or formatted if named
in \code{p_cols} (must be numeric). Column order is preserved.}

\item{est}{Numeric vector. Point estimates (\code{NA} = no CI drawn).}

\item{lower}{Numeric vector. Lower CI bounds (same length as \code{est}).}

\item{upper}{Numeric vector. Upper CI bounds (same length as \code{est}).}

\item{ci_column}{Integer. Column position in the final rendered table where
the gap/CI graphic is placed. Must be between \code{2} and
\code{ncol(data) + 1} (inclusive). Default: \code{2L}.}

\item{ref_line}{Numeric. Reference line. Default: \code{1} (HR/OR).
Use \code{0} for beta coefficients.}

\item{xlim}{Numeric vector of length 2. X-axis limits. \code{NULL} (default)
uses \code{c(0, 2)}.}

\item{ticks_at}{Numeric vector. Tick positions. \code{NULL} (default) = 5
evenly spaced ticks across \code{xlim}.}

\item{arrow_lab}{Character vector of length 2. Directional labels.
Default: \code{c("Lower risk", "Higher risk")}. \code{NULL} = none.}

\item{header}{Character vector of length \code{ncol(data) + 2}. Column
header labels for the final rendered table (original columns + gap_ci +
OR label). \code{NULL} (default) = use column names from \code{data} plus
\code{"gap_ci"} and \code{"OR (95\% CI)"}. Pass \code{""} for the gap
column position.}

\item{indent}{Integer vector (length = \code{nrow(data)}). Indentation
level of the label column: each unit adds two leading spaces.
Default: all zeros.}

\item{bold_label}{Logical vector (length = \code{nrow(data)}). Which rows
to bold in the label column. \code{NULL} (default) = auto-derive from
\code{indent}: rows where \code{indent == 0} are bolded (parent rows),
indented sub-rows are plain.}

\item{ci_col}{Character scalar or vector (length = \code{nrow(data)}).
CI colour(s). \code{NA} rows are skipped automatically.
Default: \code{"black"}.}

\item{ci_sizes}{Numeric. Point size. Default: \code{0.6}.}

\item{ci_Theight}{Numeric. CI cap height. Default: \code{0.2}.}

\item{ci_digits}{Integer. Decimal places for the auto-generated
\code{OR (95\% CI)} column. Default: \code{2L}.}

\item{ci_sep}{Character. Separator between lower and upper CI in the label,
e.g. \code{", "} or \code{" - "}. Default: \code{", "}.}

\item{p_cols}{Character vector. Names of numeric p-value columns in
\code{data}. These are formatted to \code{p_digits} decimal places with
\code{"<0.001"}-style clipping. \code{NULL} = none.}

\item{p_digits}{Integer. Decimal places for p-value formatting.
Default: \code{3L}.}

\item{bold_p}{\code{TRUE} (bold all non-NA p below \code{p_threshold}),
\code{FALSE} (no bolding), or a logical vector (per-row control).
Default: \code{TRUE}.}

\item{p_threshold}{Numeric. P-value threshold for bolding when
\code{bold_p = TRUE}. Default: \code{0.05}.}

\item{align}{Integer vector of length \code{ncol(data) + 2}. Alignment per
column: \code{-1} left, \code{0} centre, \code{1} right. Must cover all
final columns (original + gap_ci + OR label).
\code{NULL} = auto (column 1 left, all others centre).}

\item{background}{Character. Row background style: \code{"zebra"},
\code{"bold_label"} (shade rows where \code{bold_label = TRUE}),
or \code{"none"}. Default: \code{"zebra"}.}

\item{bg_col}{Character. Shading colour for backgrounds (scalar), or a
per-row vector of length \code{nrow(data)} (overrides style).
Default: \code{"#F0F0F0"}.}

\item{border}{Character. Border style: \code{"three_line"} or \code{"none"}.
Default: \code{"three_line"}.}

\item{border_width}{Numeric. Border line width(s). Scalar = all three lines
same width; length-3 vector = top-of-header, bottom-of-header,
bottom-of-body. Default: \code{3}.}

\item{row_height}{\code{NULL} (auto), numeric scalar, or numeric vector
(length = total gtable rows including margins). Auto sets 8 / 12 / 10 / 15
mm for top / header / data / bottom respectively.}

\item{col_width}{\code{NULL} (auto), numeric scalar, or numeric vector
(length = total gtable columns). Auto rounds each default width up so the
adjustment is in \eqn{[5, 10)} mm.}

\item{save}{Logical. Save output to files? Default: \code{FALSE}.}

\item{dest}{Character. Destination file path (extension ignored; all four
formats are saved). Required when \code{save = TRUE}.}

\item{save_width}{Numeric. Output width in cm. Default: \code{20}.}

\item{save_height}{Numeric or \code{NULL}. Output height in cm.
\code{NULL} = \code{nrow(data) * 0.9 + 3}.}

\item{theme}{Character preset (\code{"default"}) or a
\code{forestploter::forest_theme} object. Default: \code{"default"}.}
}
\value{
A \pkg{forestploter} plot object (gtable), returned invisibly.
Display with \code{plot()} or \code{grid::grid.draw()}.
}
\description{
Produces a publication-ready forest plot with UKB-standard styling.
The user supplies a data frame whose first column is the row label
(\code{item}), plus any additional display columns (e.g. \code{Cases/N}).
The gap column and the auto-formatted \code{OR (95\% CI)} text column are
inserted automatically at \code{ci_column}. Numeric p-value columns
declared via \code{p_cols} are formatted in-place.
}
\examples{
df <- data.frame(
  item      = c("Exposure vs. control", "Unadjusted", "Fully adjusted"),
  `Cases/N` = c("", "89/4521", "89/4521"),
  p_value   = c(NA_real_, 0.001, 0.006),
  check.names = FALSE
)

p <- plot_forest(
  data       = df,
  est        = c(NA, 1.52, 1.43),
  lower      = c(NA, 1.18, 1.11),
  upper      = c(NA, 1.96, 1.85),
  ci_column  = 2L,
  indent     = c(0L, 1L, 1L),
  bold_label = c(TRUE, FALSE, FALSE),
  p_cols     = "p_value",
  xlim       = c(0.5, 3.0)
)
plot(p)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ops.R
\name{ops_snapshot_remove}
\alias{ops_snapshot_remove}
\title{Remove raw source columns recorded at a snapshot}
\usage{
ops_snapshot_remove(data, from, keep = NULL, verbose = TRUE)
}
\arguments{
\item{data}{A data.frame or data.table.}

\item{from}{(character) Label of the snapshot whose columns should be
dropped (typically \code{"raw"}).}

\item{keep}{(character or NULL) Additional column names to protect beyond
the built-in and user-registered safe cols. Default \code{NULL}.}

\item{verbose}{(logical) Print a summary of dropped columns. Default
\code{TRUE}.}
}
\value{
A \code{data.table} with the specified columns removed. For
\code{data.table} input the operation is performed by reference (in-place);
for \code{data.frame} input the data is first converted to a new
\code{data.table} — the original \code{data.frame} is not modified.
}
\description{
Drops columns that were present at snapshot \code{from} from \code{data},
while automatically protecting built-in safe columns
(\code{"eid"}, \code{"sex"}, \code{"age"}, \code{"age_at_recruitment"}) and
any user-registered safe columns set via \code{\link{ops_set_safe_cols}}.
Columns that no longer exist in \code{data} are silently skipped.
}
\examples{
dt <- ops_toy(n = 100)
ops_snapshot(dt, label = "raw")
dt <- derive_missing(dt)
ops_snapshot(dt, label = "derived")
ops_snapshot_diff("raw", "derived")
dt <- ops_snapshot_remove(dt, from = "raw")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grs.R
\name{grs_score}
\alias{grs_score}
\title{Calculate genetic risk scores from PGEN files on RAP}
\usage{
grs_score(
  file,
  pgen_dir = NULL,
  dest = NULL,
  maf = 0.01,
  instance = "standard",
  priority = "low"
)
}
\arguments{
\item{file}{Named character vector of local weight file paths. Names become
the GRS identifiers (output column = \code{GRS_<name>}).
Example: \code{c(grs_a = "weights_a.txt")}.}

\item{pgen_dir}{Character scalar. Path to PGEN files on RAP
(e.g. \code{"/mnt/project/pgen"}). Must be specified explicitly.}

\item{dest}{Character scalar. RAP destination path for output CSV files
(e.g. \code{"/grs/"}). Must be specified explicitly.}

\item{maf}{Numeric scalar. MAF filter threshold used when locating PGEN
files. Must match the value used in \code{\link{grs_bgen2pgen}}.
Default: \code{0.01}.}

\item{instance}{Character scalar. Instance type preset: \code{"standard"}
or \code{"large"}. Default: \code{"standard"}.}

\item{priority}{Character scalar. Job priority: \code{"low"} or
\code{"high"}. Default: \code{"low"}.}
}
\value{
A named character vector of job IDs (one per GRS), returned
invisibly. Failed submissions are \code{NA}. Use \code{\link{job_ls}} to
monitor progress.
}
\description{
Uploads local SNP weight files to the RAP project root, then submits one
Swiss Army Knife job per GRS. Each job runs plink2 \code{--score} across
all 22 chromosomes and saves a single CSV to \code{dest} on completion.
Jobs run in parallel; use \code{\link{job_ls}} to monitor progress.
}
\details{
Weight files should have three columns (any delimiter, header required):
\describe{
\item{Column 1}{Variant ID (e.g. \code{rs} IDs).}
\item{Column 2}{Effect allele (A1).}
\item{Column 3}{Effect weight (beta / log-OR).}
}
This matches the output format of \code{\link{grs_check}}.

\strong{Output per job:} \code{dest/<score_name>_scores.csv} with columns
\code{IID} and the GRS score (named \code{GRS_<name>}).

\strong{Instance types:}
\describe{
\item{\code{"standard"}}{\code{mem2_ssd1_v2_x4}: 4 cores, 12 GB RAM.}
\item{\code{"large"}}{\code{mem2_ssd2_v2_x8}: 8 cores, 28 GB RAM.}
}
}
\examples{
\dontrun{
ids <- grs_score(
  file = c(
    grs_a = "weights/grs_a_weights.txt",
    grs_b = "weights/grs_b_weights.txt"
  ),
  pgen_dir = "/mnt/project/pgen",
  dest     = "/grs/",
  priority = "high"
)

job_ls()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assoc.R
\name{assoc_trend}
\alias{assoc_trend}
\alias{assoc_tr}
\title{Dose-response trend analysis}
\usage{
assoc_trend(
  data,
  outcome_col,
  time_col = NULL,
  exposure_col,
  method = c("coxph", "logistic", "linear"),
  covariates = NULL,
  base = TRUE,
  scores = NULL,
  conf_level = 0.95
)

assoc_tr(
  data,
  outcome_col,
  time_col = NULL,
  exposure_col,
  method = c("coxph", "logistic", "linear"),
  covariates = NULL,
  base = TRUE,
  scores = NULL,
  conf_level = 0.95
)
}
\arguments{
\item{data}{(data.frame or data.table) Analysis dataset.}

\item{outcome_col}{(character) Outcome column name.}

\item{time_col}{(character or NULL) Follow-up time column (required when
\code{method = "coxph"}).}

\item{exposure_col}{(character) One or more exposure column names. Each
must be a \code{factor} (ordered or unordered). The first level is used
as the reference group.}

\item{method}{(character) Regression method: \code{"coxph"} (default),
\code{"logistic"}, or \code{"linear"}.}

\item{covariates}{(character or NULL) Covariates for the Fully adjusted
model. Default: \code{NULL}.}

\item{base}{(logical) Include Unadjusted and Age-and-sex-adjusted models.
Default: \code{TRUE}.}

\item{scores}{(numeric or NULL) Numeric scores assigned to factor levels in
level order. Length must equal \code{nlevels} of every exposure.
\code{NULL} (default) uses \code{0, 1, 2, \ldots}}

\item{conf_level}{(numeric) Confidence level. Default: \code{0.95}.}
}
\value{
A \code{data.table} with one row per exposure \eqn{\times} level
\eqn{\times} model, containing:
\describe{
\item{\code{exposure}}{Exposure variable name.}
\item{\code{level}}{Factor level (ordered factor preserving original
level order).}
\item{\code{term}}{Coefficient name as returned by the model (reference
row uses \code{paste0(exposure, ref_level)}).}
\item{\code{model}}{Ordered factor: \code{Unadjusted} <
\code{Age and sex adjusted} < \code{Fully adjusted}.}
\item{\code{n}}{Participants in model (after NA removal).}
\item{...}{Effect estimate columns from the categorical model:
\code{HR}/\code{OR}/\code{beta}, \code{CI_lower}, \code{CI_upper},
\code{p_value}, and a formatted label. Reference row has
\code{HR = 1} / \code{beta = 0} and \code{NA} for CI and p.
Cox models additionally include \code{n_events} and
\code{person_years}; logistic includes \code{n_cases}; linear
includes \code{se}.}
\item{\code{HR_per_score} / \code{OR_per_score} /
\code{beta_per_score}}{Per-score-unit effect estimate from the trend
model. Shared across all levels within the same exposure \eqn{\times}
model.}
\item{\code{HR_per_score_label} / \code{OR_per_score_label} /
\code{beta_per_score_label}}{Formatted string for the per-score
estimate.}
\item{\code{p_trend}}{P-value for linear trend from the trend model.
Shared across all levels within the same exposure \eqn{\times}
model.}
}
}
\description{
Fits categorical and trend models simultaneously for each ordered-factor
exposure, returning per-category effect estimates alongside a p-value for
linear trend. Two models are run internally per exposure \eqn{\times}
adjustment combination:
}
\details{
\enumerate{
\item \strong{Categorical model} - exposure treated as a factor; produces
one row per non-reference level (HR / OR / \eqn{\beta} vs reference).
\item \strong{Trend model} - exposure recoded as numeric scores (default:
0, 1, 2, \ldots); produces the per-score-unit effect estimate
(\code{*_per_score}) and \code{p_trend}.
}

Both results are merged: the output contains a reference row (effect = 1 /
0, CI = NA) followed by non-reference rows, with \code{*_per_score} and
\code{p_trend} appended as shared columns (same value within each
exposure \eqn{\times} model combination).

\strong{Scores}: By default levels are scored 0, 1, 2, \ldots so the
reference group = 0 and each step = 1 unit. Supply \code{scores} to use
meaningful units (e.g. median years per category) - only \code{p_trend} and
the per-score estimate change; per-category HRs are unaffected.

\strong{Adjustment models}: follows the same logic as
\code{\link{assoc_coxph}} - Unadjusted and Age-and-sex-adjusted models are
included by default (\code{base = TRUE}); a Fully adjusted model is added
when \code{covariates} is non-NULL.
}
\examples{
dt <- ops_toy(scenario = "association", n = 500)
dt <- dt[dm_timing != 1L]

res <- assoc_trend(
  data         = dt,
  outcome_col  = "dm_status",
  time_col     = "dm_followup_years",
  exposure_col = "bmi_cat",
  method       = "coxph",
  covariates   = c("tdi_cat", "p20116_i0"),
  base         = FALSE
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assoc.R
\name{assoc_linear}
\alias{assoc_linear}
\alias{assoc_lm}
\title{Linear regression association analysis}
\usage{
assoc_linear(
  data,
  outcome_col,
  exposure_col,
  covariates = NULL,
  base = TRUE,
  conf_level = 0.95
)

assoc_lm(
  data,
  outcome_col,
  exposure_col,
  covariates = NULL,
  base = TRUE,
  conf_level = 0.95
)
}
\arguments{
\item{data}{(data.frame or data.table) Analysis dataset.}

\item{outcome_col}{(character) Name of the continuous numeric outcome column.}

\item{exposure_col}{(character) One or more exposure variable names.}

\item{covariates}{(character or NULL) Covariate column names for the
\strong{Fully adjusted} model. Default: \code{NULL}.}

\item{base}{(logical) Include \strong{Unadjusted} and \strong{Age and sex
adjusted} models. Default: \code{TRUE}.}

\item{conf_level}{(numeric) Confidence level. Default: \code{0.95}.}
}
\value{
A \code{data.table} with one row per exposure \eqn{\times} term
\eqn{\times} model combination, and columns:
\describe{
\item{\code{exposure}}{Exposure variable name.}
\item{\code{term}}{Coefficient name (e.g. \code{"bmi_categoryObese"}).}
\item{\code{model}}{Ordered factor: \code{Unadjusted} <
\code{Age and sex adjusted} < \code{Fully adjusted}.}
\item{\code{n}}{Participants in model (after NA removal).}
\item{\code{beta}}{Regression coefficient (\eqn{\beta}).}
\item{\code{se}}{Standard error of \eqn{\beta}.}
\item{\code{CI_lower}}{Lower confidence bound.}
\item{\code{CI_upper}}{Upper confidence bound.}
\item{\code{p_value}}{t-test p-value.}
\item{\code{beta_label}}{Formatted string, e.g.
\code{"0.23 (0.05-0.41)"}.}
}
}
\description{
Fits one or more linear regression models for each exposure variable and
returns a tidy result table. By default, two standard adjustment models are
always included:
}
\details{
\itemize{
\item \strong{Unadjusted} - no covariates (crude).
\item \strong{Age and sex adjusted} - age + sex auto-detected from the
data via UKB field IDs (21022 and 31). Skipped with a warning if either
column cannot be found.
\item \strong{Fully adjusted} - the covariates supplied via the
\code{covariates} argument. Only run when \code{covariates} is non-NULL.
}

\strong{Outcome}: intended for continuous numeric variables. Passing a
binary (0/1) or logical column is permitted (linear probability model) but
will trigger a warning recommending \code{\link{assoc_logistic}} instead.

\strong{CI method}: based on the t-distribution via \code{confint.lm()},
which is exact under the normal linear model assumption. There is no
\code{ci_method} argument (unlike \code{\link{assoc_logistic}}) as profile
likelihood does not apply to \code{lm}.

\strong{SE column}: the standard error of \eqn{\beta} is included to
support downstream meta-analysis and GWAS-style summary statistics.
}
\examples{
dt <- ops_toy(scenario = "association", n = 500)

res <- assoc_linear(
  data         = dt,
  outcome_col  = "p21001_i0",
  exposure_col = "p20116_i0",
  covariates   = c("bmi_cat", "tdi_cat"),
  base         = FALSE
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auth.R
\name{auth_login}
\alias{auth_login}
\title{Login to DNAnexus with a token}
\usage{
auth_login(token = NULL)
}
\arguments{
\item{token}{(character) DNAnexus API token. If NULL, reads from the
environment variable \code{DX_API_TOKEN}.}
}
\value{
Invisible TRUE on success.
}
\description{
Authenticates with the DNAnexus Research Analysis Platform using an API
token. Equivalent to running \verb{dx login --token} on the command line.
}
\examples{
\dontrun{
# Supply token directly
auth_login(token = "your_token_here")

# Or store token in ~/.Renviron (recommended):
# usethis::edit_r_environ()
# Add: DX_API_TOKEN=your_token_here
# Save and restart R, then call:
auth_login()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ops.R
\name{ops_snapshot}
\alias{ops_snapshot}
\title{Record and review dataset pipeline snapshots}
\usage{
ops_snapshot(
  data = NULL,
  label = NULL,
  reset = FALSE,
  verbose = TRUE,
  check_na = TRUE
)
}
\arguments{
\item{data}{A data.frame or data.table to snapshot. Pass \code{NULL} (or omit)
to print the full snapshot history without recording a new entry.}

\item{label}{(character) A short name for this pipeline stage, e.g.
\code{"raw"}, \code{"after_derive_missing"}. Defaults to \code{"snapshot_N"} where N
is the sequential index.}

\item{reset}{(logical) If \code{TRUE}, clears the entire snapshot history and
returns invisibly. Default \code{FALSE}.}

\item{verbose}{(logical) Print the CLI report. Default \code{TRUE}.}

\item{check_na}{(logical) Whether to count columns with any \code{NA} or blank
string values and include the delta in the report. Set to \code{FALSE} to skip
the NA scan (useful for large datasets or when NA tracking is not needed).
Default \code{TRUE}.}
}
\value{
When \code{data} is supplied, returns the new snapshot row invisibly
(a one-row data.table). When called with no \code{data}, returns the full
history data.table invisibly.
}
\description{
Captures a lightweight summary of a data.frame at a given pipeline stage
and stores it in the session cache. Subsequent calls automatically compute
deltas against the previous snapshot, making it easy to track how data
changes through \verb{derive_*}, \verb{assoc_*}, and other processing steps.
}
\examples{
dt <- ops_toy(n = 100)
ops_snapshot(dt, label = "raw")

dt <- derive_missing(dt)
ops_snapshot(dt, label = "after_derive_missing")

# View full history
ops_snapshot()

# Reset history
ops_snapshot(reset = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ops.R
\name{ops_snapshot_cols}
\alias{ops_snapshot_cols}
\title{Retrieve column names recorded at a snapshot}
\usage{
ops_snapshot_cols(label, keep = NULL)
}
\arguments{
\item{label}{(character) Snapshot label passed to \code{ops_snapshot()}.}

\item{keep}{(character or NULL) Column names to exclude from the returned
vector (i.e. columns to retain in the data even if they were present at
that snapshot). Default \code{NULL}.}
}
\value{
A character vector of column names.
}
\description{
Returns the column names stored by a previous \code{\link{ops_snapshot}}
call, optionally excluding columns you wish to keep.
}
\examples{
dt <- ops_toy(n = 100)
ops_snapshot(dt, label = "raw")
ops_snapshot_cols("raw")
ops_snapshot_cols("raw", keep = "eid")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ops.R
\name{ops_setup}
\alias{ops_setup}
\title{Check the ukbflow operating environment}
\usage{
ops_setup(
  check_dx = TRUE,
  check_auth = TRUE,
  check_deps = TRUE,
  verbose = TRUE
)
}
\arguments{
\item{check_dx}{(logical) Check dx CLI installation (dxpy is implied by dx).}

\item{check_auth}{(logical) Check RAP login status.}

\item{check_deps}{(logical) Check R package dependencies.}

\item{verbose}{(logical) Print the formatted report. Set to \code{FALSE} for
programmatic use (results are still returned invisibly).}
}
\value{
An invisible named list with elements \code{dx}, \code{dxpy}, \code{auth}, \code{deps},
and \code{summary}. Each element reflects the result of its respective check
block and can be inspected programmatically.
}
\description{
Runs a four-block health check covering the dx CLI, dxpy (Python), RAP
authentication, and R package dependencies. Designed to be the first
function a new user runs after installation.
}
\details{
The function is \strong{read-only}: it never modifies system state, installs
packages, or authenticates. Auth failures are reported as warnings, not
errors, because the check itself does not require a live RAP connection.
}
\examples{
ops_setup(check_dx = FALSE, check_auth = FALSE)

result <- ops_setup(check_dx = FALSE, check_auth = FALSE, verbose = FALSE)
result$summary$fail == 0
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fetch.R
\name{fetch_metadata}
\alias{fetch_metadata}
\title{Download the Showcase metadata folder}
\usage{
fetch_metadata(dest_dir, overwrite = FALSE, resume = FALSE, verbose = TRUE)
}
\arguments{
\item{dest_dir}{(character) Destination directory. Created automatically
if it does not exist.}

\item{overwrite}{(logical) Overwrite existing local files. Default:
\code{FALSE}.}

\item{resume}{(logical) Resume interrupted downloads. Default: \code{FALSE}.}

\item{verbose}{(logical) Show download progress. Default: \code{TRUE}.}
}
\value{
Invisibly returns the local file paths as a character vector.
}
\description{
Downloads the entire \code{Showcase metadata/} folder from the DNAnexus
Research Analysis Platform to a local directory. This includes
\code{field.tsv}, \code{encoding.tsv}, and all associated encoding tables.
}
\examples{
\dontrun{
fetch_metadata(dest_dir = "metadata")
fetch_metadata(dest_dir = "metadata", overwrite = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ops.R
\name{ops_set_safe_cols}
\alias{ops_set_safe_cols}
\title{Register additional safe columns protected from snapshot-based drops}
\usage{
ops_set_safe_cols(cols = NULL, reset = FALSE)
}
\arguments{
\item{cols}{(character) One or more column names to protect.}

\item{reset}{(logical) If \code{TRUE}, clear the current user-registered
safe list before adding. Default \code{FALSE}.}
}
\value{
Invisibly returns the updated safe cols vector.
}
\description{
Adds column names to the session-level safe list. Columns in this list are
automatically excluded when \code{\link{ops_snapshot_cols}} is used to
build a drop vector, in addition to the built-in protected columns
(\code{"eid"}, \code{"sex"}, \code{"age"}, \code{"age_at_recruitment"}).
}
\examples{
ops_set_safe_cols(c("date_baseline", "townsend_index"))
ops_set_safe_cols(reset = TRUE)  # clear user-registered safe cols
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ops.R
\name{ops_snapshot_diff}
\alias{ops_snapshot_diff}
\title{Compare column names between two snapshots}
\usage{
ops_snapshot_diff(label1, label2)
}
\arguments{
\item{label1}{(character) Label of the earlier snapshot.}

\item{label2}{(character) Label of the later snapshot.}
}
\value{
A named list with two character vectors: \code{added} (columns
present in \code{label2} but not \code{label1}) and \code{removed}
(columns present in \code{label1} but not \code{label2}).
}
\description{
Returns lists of columns added and removed between two recorded snapshots.
}
\examples{
dt <- ops_toy(n = 100)
ops_snapshot(dt, label = "raw")
dt <- derive_missing(dt)
ops_snapshot(dt, label = "derived")
ops_snapshot_diff("raw", "derived")
# $added   — newly derived columns
# $removed — columns dropped between snapshots
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ops.R
\name{ops_na}
\alias{ops_na}
\title{Summarise missing values by column}
\usage{
ops_na(data, threshold = 0, verbose = TRUE)
}
\arguments{
\item{data}{A data.frame or data.table to scan.}

\item{threshold}{(numeric) Columns with \code{pct_na <= threshold} are silenced
from the per-column CLI output. The summary block is always shown.
Default \code{0}: every column with any missing value is listed.}

\item{verbose}{(logical) Print the CLI report. Default \code{TRUE}.}
}
\value{
An invisible data.table with columns \code{column}, \code{n_na}, and \code{pct_na},
sorted by \code{pct_na} descending. \code{n_na} counts both \code{NA} and \code{""}.
Always contains all columns regardless of \code{threshold} (which only affects
CLI output).
}
\description{
Scans each column of a data.frame or data.table and returns the count and
percentage of missing values. Results are sorted by missingness in
descending order. Columns above 10\\% are flagged in red; those between 0\\%
and 10\\% in yellow. A summary block is always printed regardless of
\code{threshold}.
}
\details{
\strong{Missing value definition}: a value is counted as missing if it is \code{NA}
\emph{or} an empty string (\code{""}). Empty strings are treated as missing because
UKB exports frequently use \code{""} as a placeholder for absent text values.
This means \code{n_na} and \code{pct_na} reflect \emph{effective} missingness, not just
\code{is.na()}. Numeric and logical columns are not affected (they cannot hold
\code{""}).
}
\examples{
dt <- ops_toy(n = 100)

# Show all columns with any missing value
ops_na(dt)

# Only list columns with > 10\% missing in the CLI output
ops_na(dt, threshold = 10)

# Programmatic use — retrieve result silently
result <- ops_na(dt, verbose = FALSE)
result[pct_na > 50]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/derive.R
\name{derive_hes}
\alias{derive_hes}
\title{Derive a binary disease flag from UKB HES inpatient diagnoses}
\usage{
derive_hes(
  data,
  name,
  icd10,
  match = c("prefix", "exact", "regex"),
  disease_cols = NULL,
  date_cols = NULL
)
}
\arguments{
\item{data}{(data.frame or data.table) UKB phenotype data containing HES
fields (\code{p41270} and \code{p41280_a*}).}

\item{name}{(character) Output column prefix, e.g. \code{"disease"} produces
\code{disease_hes} and \code{disease_hes_date}.}

\item{icd10}{(character) ICD-10 code(s) to match.  For \code{"prefix"} and
\code{"exact"}, supply a vector such as \code{c("L20", "L21")}.  For
\code{"regex"}, supply a single regex string.}

\item{match}{(character) Matching strategy: \code{"prefix"} (default)
matches any code starting with the supplied string; \code{"exact"}
requires a full match; \code{"regex"} uses \code{icd10} directly.}

\item{disease_cols}{(character or NULL) Name of the \code{p41270} column.
\code{NULL} = auto-detect.}

\item{date_cols}{(character or NULL) Names of \code{p41280_a*} columns.
\code{NULL} = auto-detect.}
}
\value{
The input \code{data} with two new columns added in-place:
\code{{name}_hes} (logical) and \code{{name}_hes_date} (IDate).
Always returns a \code{data.table}.
}
\description{
Hospital Episode Statistics (HES) inpatient records store ICD-10 diagnosis
codes in field \code{p41270} (single JSON-array column on UKB RAP) and
corresponding first-diagnosis dates in field \code{p41280}
(\code{p41280_a0}, \code{p41280_a1}, \ldots).  The array index in
\code{p41270} and \code{p41280} are aligned: the \emph{N}-th code in the
JSON array corresponds to \code{p41280_aN} (date of first in-patient
diagnosis for that code).
}
\details{
\describe{
\item{\code{{name}_hes}}{Logical flag: \code{TRUE} if any HES record
contains a matching ICD-10 code.}
\item{\code{{name}_hes_date}}{Earliest first-diagnosis date across all
matching codes (\code{IDate}).  \code{NA} if no date is available.}
}
}
\examples{
\donttest{
dt <- ops_toy(n = 100)
derive_hes(dt, name = "htn",      icd10 = "I10")
derive_hes(dt, name = "diabetes", icd10 = c("E10", "E11"))
derive_hes(dt, name = "asthma",   icd10 = "^J4", match = "regex")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grs.R
\name{grs_bgen2pgen}
\alias{grs_bgen2pgen}
\title{Convert UKB imputed BGEN files to PGEN on RAP}
\usage{
grs_bgen2pgen(
  chr = 1:22,
  dest = NULL,
  maf = 0.01,
  instance = "standard",
  priority = "low"
)
}
\arguments{
\item{chr}{Integer vector. Chromosomes to process. Default: \code{1:22}.}

\item{dest}{Character scalar. RAP destination path for output PGEN files
(e.g. \code{"/pgen/"}). Must be specified explicitly.}

\item{maf}{Numeric scalar. Minor allele frequency filter passed to plink2
\code{--maf}. Variants with MAF below this threshold are excluded.
Default: \code{0.01}. Must be in \code{(0, 0.5)}.}

\item{instance}{Character scalar. Instance type preset: \code{"standard"}
or \code{"large"}. See Details. Default: \code{"standard"}.}

\item{priority}{Character scalar. Job priority: \code{"low"} or
\code{"high"}. Default: \code{"low"}.}
}
\value{
A character vector of job IDs (one per chromosome), returned
invisibly. Failed submissions are \code{NA}. Use \code{\link{job_ls}},
\code{\link{job_status}}, or \code{\link{job_wait}} to monitor progress.
}
\description{
Submits one Swiss Army Knife job per chromosome to the DNAnexus Research
Analysis Platform, each converting a UKB imputed BGEN file to PGEN format
with a MAF > 0.01 filter applied via plink2. Jobs run in parallel across
chromosomes.
}
\details{
The function auto-generates the plink2 driver script, uploads it once to
the RAP project root (\code{/}) on RAP, then loops over \code{chr} submitting
one job per chromosome. A 500 ms pause between submissions prevents API
rate-limiting.

\strong{Output path is critical.} The driver script writes plink2 output to
\code{/home/dnanexus/out/out/} - the fixed path that Swiss Army Knife
auto-uploads to \code{dest} on completion. Output files per chromosome:
\code{chr\{N\}_maf001.pgen/.pvar/.psam/.log}.

\strong{Instance types:}
\describe{
\item{\code{"standard"}}{\code{mem2_ssd1_v2_x4}: 4 cores, 12 GB RAM.
Suitable for smaller chromosomes (roughly chr 17–22).}
\item{\code{"large"}}{\code{mem2_ssd2_v2_x8}: 8 cores, 28 GB RAM,
640 GB SSD. Required for large chromosomes (roughly chr 1–16) where
standard storage is insufficient.}
}
}
\examples{
\dontrun{
# Test with chr22 first (smallest chromosome)
ids <- grs_bgen2pgen(chr = 22, dest = "/pgen/", priority = "high")

# Small chromosomes - standard instance
ids_small <- grs_bgen2pgen(chr = 15:22, dest = "/pgen/")

# Large chromosomes - upgrade instance to handle storage
ids_large <- grs_bgen2pgen(chr = 1:16, dest = "/pgen/", instance = "large")

# Monitor
job_ls()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assoc.R
\name{assoc_coxph_zph}
\alias{assoc_coxph_zph}
\alias{assoc_zph}
\title{Proportional hazards assumption test for Cox regression}
\usage{
assoc_coxph_zph(
  data,
  outcome_col,
  time_col,
  exposure_col,
  covariates = NULL,
  base = TRUE,
  strata = NULL
)

assoc_zph(
  data,
  outcome_col,
  time_col,
  exposure_col,
  covariates = NULL,
  base = TRUE,
  strata = NULL
)
}
\arguments{
\item{data}{(data.frame or data.table) Analysis dataset.}

\item{outcome_col}{(character) Binary event indicator (\code{0}/\code{1}
or \code{TRUE}/\code{FALSE}).}

\item{time_col}{(character) Follow-up time column name.}

\item{exposure_col}{(character) One or more exposure variable names.}

\item{covariates}{(character or NULL) Covariates for the Fully adjusted
model. Default: \code{NULL}.}

\item{base}{(logical) Include Unadjusted and Age and sex adjusted models.
Default: \code{TRUE}.}

\item{strata}{(character or NULL) Optional stratification variable.}
}
\value{
A \code{data.table} with one row per exposure \eqn{\times} term
\eqn{\times} model combination, and columns:
\describe{
\item{\code{exposure}}{Exposure variable name.}
\item{\code{term}}{Coefficient name.}
\item{\code{model}}{Ordered factor: \code{Unadjusted} <
\code{Age and sex adjusted} < \code{Fully adjusted}.}
\item{\code{chisq}}{Schoenfeld residual chi-squared statistic.}
\item{\code{df}}{Degrees of freedom.}
\item{\code{p_value}}{P-value for the PH test (term-level).}
\item{\code{ph_satisfied}}{Logical; \code{TRUE} if \code{p_value > 0.05}.}
\item{\code{global_chisq}}{Global chi-squared for the whole model.}
\item{\code{global_df}}{Global degrees of freedom.}
\item{\code{global_p}}{Global p-value for the whole model.}
}
}
\description{
Tests the proportional hazards (PH) assumption using Schoenfeld residuals
via \code{\link[survival]{cox.zph}}. Re-fits the same models as
\code{\link{assoc_coxph}} (same interface) and returns a tidy result table
with term-level and global test statistics.
}
\details{
A non-significant p-value (p > 0.05) indicates the PH assumption is
satisfied for that term. The global test (\code{global_p}) reflects the
overall PH assumption for the whole model.
}
\examples{
dt <- ops_toy(scenario = "association", n = 500)
dt <- dt[dm_timing != 1L]

zph <- assoc_coxph_zph(
  data         = dt,
  outcome_col  = "dm_status",
  time_col     = "dm_followup_years",
  exposure_col = "p20116_i0",
  covariates   = c("bmi_cat", "tdi_cat"),
  base         = FALSE
)
zph[ph_satisfied == FALSE]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grs.R
\name{grs_validate}
\alias{grs_validate}
\title{Validate GRS predictive performance}
\usage{
grs_validate(
  data,
  grs_cols = NULL,
  outcome_col,
  time_col = NULL,
  covariates = NULL
)
}
\arguments{
\item{data}{A \code{data.frame} or \code{data.table}.}

\item{grs_cols}{Character vector of GRS column names to validate.
If \code{NULL} (default), all columns whose names contain \code{"grs"}
(case-insensitive) are selected automatically. All specified columns
must be numeric.}

\item{outcome_col}{Character scalar. Name of the outcome column
(\code{0}/\code{1} or \code{TRUE}/\code{FALSE}).}

\item{time_col}{Character scalar or \code{NULL}. Name of the follow-up
time column. When \code{NULL} (default), logistic regression is used;
when supplied, Cox regression is used.}

\item{covariates}{Character vector or \code{NULL}. Covariates for the
fully adjusted model. When \code{NULL}, only unadjusted and age-sex
adjusted models are run.}
}
\value{
A named \code{list} with four \code{data.table} elements:
\itemize{
\item \code{per_sd}: OR / HR per 1-SD increase in GRS.
\item \code{high_vs_low}: OR / HR for High vs Low extreme tertile.
\item \code{trend}: P-trend across Q1–Q4 quartiles.
\item \code{discrimination}: AUC (logistic) or C-index (Cox) with 95\\% CI.
}
}
\description{
For each GRS column, computes four sets of validation metrics:
\enumerate{
\item \strong{Per SD} - OR (logistic) or HR (Cox) per 1-SD increase.
\item \strong{High vs Low} - OR / HR comparing top 20\\% vs bottom 20\\%
(extreme tertile grouping: Low / Mid / High).
\item \strong{Trend test} - P-trend across quartiles (Q1–Q4).
\item \strong{Discrimination} - AUC (logistic) or C-index (Cox).
}
}
\details{
GRS grouping columns are created internally via \code{\link{derive_cut}}
and are not added to the user's data.  When \code{time_col} is
\code{NULL}, logistic regression is used throughout; when supplied, Cox
proportional hazards models are used.

Models follow the same adjustment logic as \code{\link{assoc_logistic}} /
\code{\link{assoc_coxph}}: unadjusted and age-sex adjusted models are
always included; a fully adjusted model is added when \code{covariates}
is non-\code{NULL}.
}
\examples{
\donttest{
dt <- ops_toy(scenario = "association", n = 500)
dt <- grs_standardize(dt, grs_cols = "grs_bmi")

res <- grs_validate(
  data        = dt,
  grs_cols    = "grs_bmi_z",
  outcome_col = "dm_status",
  time_col    = "dm_followup_years"
)
res$per_sd

if (requireNamespace("pROC", quietly = TRUE)) {
  res_logit <- grs_validate(
    data        = dt,
    grs_cols    = "grs_bmi_z",
    outcome_col = "dm_status"
  )
  res_logit$discrimination
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fetch.R
\name{fetch_url}
\alias{fetch_url}
\title{Get pre-authenticated download URL(s) for a remote RAP file or folder}
\usage{
fetch_url(path, duration = "1d")
}
\arguments{
\item{path}{(character) Remote file path or folder path, e.g.
\code{"Showcase metadata/field.tsv"} or \code{"Showcase metadata/"}.}

\item{duration}{(character) How long the URLs remain valid. Accepts
suffixes: \code{s}, \code{m}, \code{h}, \code{d}, \code{w}, \code{M},
\code{y}. Default: \code{"1d"} (one day).}
}
\value{
A named character vector of pre-authenticated HTTPS URLs.
Names are the file names.
}
\description{
Generates temporary HTTPS URLs for files on the DNAnexus Research Analysis
Platform. For a single file, returns one URL. For a folder, lists all files
inside and returns a named character vector of URLs.
}
\examples{
\dontrun{
# Single file
fetch_url("Showcase metadata/field.tsv")

# Entire folder
fetch_url("Showcase metadata/", duration = "7d")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/derive.R
\name{derive_timing}
\alias{derive_timing}
\title{Classify disease timing relative to UKB baseline assessment}
\usage{
derive_timing(data, name, baseline_col, status_col = NULL, date_col = NULL)
}
\arguments{
\item{data}{(data.frame or data.table) UKB phenotype data.}

\item{name}{(character) Output column prefix.  The new column is named
\code{{name}_timing}.  Also used to derive default \code{status_col}
and \code{date_col} when those are \code{NULL}.}

\item{baseline_col}{(character) Name of the baseline date column in
\code{data} (e.g. \code{"date_baseline"} or \code{"p53_i0"}).}

\item{status_col}{(character or NULL) Name of the logical disease flag.
\code{NULL} = \code{paste0(name, "_status")}.}

\item{date_col}{(character or NULL) Name of the disease date column
(\code{IDate} or \code{Date}).
\code{NULL} = \code{paste0(name, "_date")}.}
}
\value{
The input \code{data} with one new integer column
\code{{name}_timing} (0/1/2/\code{NA}) added in-place.
Always returns a \code{data.table}.
}
\description{
Assigns each participant an integer timing category based on whether their
disease date falls before or after the baseline visit date:
}
\details{
\describe{
\item{\code{0}}{No disease (\code{status_col} is \code{FALSE}).}
\item{\code{1}}{Prevalent - disease date on or before baseline.}
\item{\code{2}}{Incident - disease date strictly after baseline.}
\item{\code{NA}}{Case with missing date; timing cannot be determined.}
}

Call once per timing variable needed (e.g. once for the combined case,
once per individual source).
}
\examples{
\donttest{
dt <- ops_toy(n = 100)
derive_hes(dt, name = "htn", icd10 = "I10")
derive_timing(dt, name = "htn_hes",
              status_col   = "htn_hes",
              date_col     = "htn_hes_date",
              baseline_col = "p53_i0")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assoc.R
\name{assoc_lag}
\alias{assoc_lag}
\title{Cox regression lag sensitivity analysis}
\usage{
assoc_lag(
  data,
  outcome_col,
  time_col,
  exposure_col,
  lag_years = c(1, 2),
  covariates = NULL,
  base = TRUE,
  strata = NULL,
  conf_level = 0.95
)
}
\arguments{
\item{data}{(data.frame or data.table) Analysis dataset.}

\item{outcome_col}{(character) Event indicator column (0/1 or logical).}

\item{time_col}{(character) Follow-up time column (numeric, e.g. years).}

\item{exposure_col}{(character) One or more exposure variable names.}

\item{lag_years}{(numeric) One or more lag periods in the same units as
\code{time_col}. Default: \code{c(1, 2)}. Use \code{0} to include the
unfiltered full-cohort result as a reference.}

\item{covariates}{(character or NULL) Covariates for the Fully adjusted
model. When \code{NULL}, only Unadjusted (and Age and sex adjusted if
\code{base = TRUE}) are run.}

\item{base}{(logical) Auto-detect age and sex and include an Age and sex
adjusted model. Default: \code{TRUE}.}

\item{strata}{(character or NULL) Optional stratification variable passed to
\code{survival::strata()}.}

\item{conf_level}{(numeric) Confidence level for HR intervals.
Default: \code{0.95}.}
}
\value{
A \code{data.table} with one row per lag \eqn{\times} exposure
\eqn{\times} term \eqn{\times} model combination, containing all columns
produced by \code{\link{assoc_coxph}} plus:
\describe{
\item{\code{lag_years}}{The lag period applied (numeric).}
\item{\code{n_excluded}}{Number of participants excluded because their
follow-up time was less than \code{lag_years}.}
}
\code{lag_years} and \code{n_excluded} are placed immediately after
\code{model} in the column order.
}
\description{
Runs Cox proportional hazards models at one or more lag periods to assess
whether associations are robust to the exclusion of early events. For each
lag, participants whose follow-up time is less than \code{lag_years} are
removed from the analysis dataset; follow-up time is kept on its original
scale (not shifted). This mirrors the approach used in UK Biobank sensitivity
analyses to address reverse causation and detection bias.
}
\details{
Setting \code{lag_years = 0} (or including \code{0} in the vector) runs the
model on the full unfiltered cohort, providing a reference against which
lagged results can be compared.

The same three adjustment models produced by \code{\link{assoc_coxph}} are
available here (\strong{Unadjusted}, \strong{Age and sex adjusted},
\strong{Fully adjusted}).
}
\examples{
dt <- ops_toy(scenario = "association", n = 500)
dt <- dt[dm_timing != 1L]

res <- assoc_lag(
  data         = dt,
  outcome_col  = "dm_status",
  time_col     = "dm_followup_years",
  exposure_col = "p20116_i0",
  lag_years    = c(0, 1, 2),
  covariates   = c("bmi_cat", "tdi_cat"),
  base         = FALSE
)
res[, .(lag_years, n, n_excluded, HR, CI_lower, CI_upper, p_value)]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/derive.R
\name{derive_icd10}
\alias{derive_icd10}
\title{Derive a unified ICD-10 disease flag across multiple UKB data sources}
\usage{
derive_icd10(
  data,
  name,
  icd10,
  source = c("hes", "death", "first_occurrence", "cancer_registry"),
  match = c("prefix", "exact", "regex"),
  fo_field = NULL,
  fo_col = NULL,
  histology = NULL,
  behaviour = NULL,
  hes_code_col = NULL,
  hes_date_cols = NULL,
  primary_cols = NULL,
  secondary_cols = NULL,
  death_date_cols = NULL,
  cr_code_cols = NULL,
  cr_hist_cols = NULL,
  cr_behv_cols = NULL,
  cr_date_cols = NULL
)
}
\arguments{
\item{data}{(data.frame or data.table) UKB phenotype data.}

\item{name}{(character) Output column prefix, e.g. \code{"disease"} produces
\code{disease_icd10} and \code{disease_icd10_date}, plus intermediate columns
such as \code{disease_hes}, \code{disease_hes_date}, etc.}

\item{icd10}{(character) ICD-10 code(s) to match.  For \code{"prefix"}
and \code{"exact"}, supply a vector such as \code{c("L20", "L21")}.
For \code{"regex"}, supply a single regex string.  When
\code{"cancer_registry"} is included in \code{source}, \code{icd10}
and \code{match} are automatically converted to a regex and passed to
\code{\link{derive_cancer_registry}}.}

\item{source}{(character) One or more of \code{"hes"},
\code{"death"}, \code{"first_occurrence"}, \code{"cancer_registry"}.
Defaults to all four.}

\item{match}{(character) Matching strategy passed to \code{derive_hes}
and \code{derive_death_registry}: \code{"prefix"} (default),
\code{"exact"}, or \code{"regex"}.}

\item{fo_field}{(integer or character or NULL) UKB field ID for the
First Occurrence column (e.g. \code{131666L} for E11).  Required when
\code{"first_occurrence"} is in \code{source} and \code{fo_col} is
\code{NULL}.}

\item{fo_col}{(character or NULL) Column name of the First Occurrence
field in \code{data}.  Alternative to \code{fo_field}.}

\item{histology}{(integer vector or NULL) Passed to
\code{\link{derive_cancer_registry}}.  Ignored for other sources.}

\item{behaviour}{(integer vector or NULL) Passed to
\code{\link{derive_cancer_registry}}.  Ignored for other sources.}

\item{hes_code_col}{(character or NULL) Passed as \code{disease_cols}
to \code{\link{derive_hes}}.}

\item{hes_date_cols}{(character or NULL) Passed as \code{date_cols}
to \code{\link{derive_hes}}.}

\item{primary_cols}{(character or NULL) Passed to
\code{\link{derive_death_registry}}.}

\item{secondary_cols}{(character or NULL) Passed to
\code{\link{derive_death_registry}}.}

\item{death_date_cols}{(character or NULL) Passed as \code{date_cols}
to \code{\link{derive_death_registry}}.}

\item{cr_code_cols}{(character or NULL) Passed as \code{code_cols}
to \code{\link{derive_cancer_registry}}.}

\item{cr_hist_cols}{(character or NULL) Passed as \code{hist_cols}
to \code{\link{derive_cancer_registry}}.}

\item{cr_behv_cols}{(character or NULL) Passed as \code{behv_cols}
to \code{\link{derive_cancer_registry}}.}

\item{cr_date_cols}{(character or NULL) Passed as \code{date_cols}
to \code{\link{derive_cancer_registry}}.}
}
\value{
The input \code{data} with \code{{name}_icd10} (logical) and
\code{{name}_icd10_date} (IDate) added in-place, plus all intermediate
source columns.  Always returns a \code{data.table}.
}
\description{
A high-level wrapper that calls one or more of \code{\link{derive_hes}},
\code{\link{derive_death_registry}}, \code{\link{derive_first_occurrence}},
and \code{\link{derive_cancer_registry}} according to the \code{source}
argument, then combines their results into a single status flag and
earliest-date column.
}
\details{
All intermediate source columns (\code{{name}_hes}, \code{{name}_death},
\code{{name}_fo}, \code{{name}_cancer} and their \code{_date} counterparts)
are retained in \code{data} so that per-source contributions remain
traceable.

\describe{
\item{\code{{name}_icd10}}{Logical flag: \code{TRUE} if any selected
source contains a matching record.}
\item{\code{{name}_icd10_date}}{Earliest matching date across all
selected sources (\code{IDate}).}
}
}
\examples{
\donttest{
dt <- ops_toy(n = 100)
derive_icd10(dt, name = "htn",
             icd10  = "I10",
             source = c("hes", "death"))
derive_icd10(dt, name = "mi",
             icd10  = "I21",
             source = c("hes", "death", "first_occurrence"),
             fo_col = "p131742")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ops.R
\name{ops_withdraw}
\alias{ops_withdraw}
\title{Exclude withdrawn participants from a dataset}
\usage{
ops_withdraw(data, file, eid_col = "eid", verbose = TRUE)
}
\arguments{
\item{data}{A data.frame or data.table containing a participant ID column.}

\item{file}{(character) Path to the UKB withdrawal CSV file. The file must
be a single-column, \strong{header-free} CSV as supplied by UK Biobank
(e.g. \code{w854944_20260310.csv}).}

\item{eid_col}{(character) Name of the participant ID column in \code{data}.
Default \code{"eid"}.}

\item{verbose}{(logical) Print the CLI report. Default \code{TRUE}.}
}
\value{
A data.table with withdrawn participants removed.
}
\description{
Reads a UK Biobank withdrawal list (a headerless single-column CSV of
anonymised participant IDs) and removes the corresponding rows from
\code{data}. A pair of \code{\link[=ops_snapshot]{ops_snapshot()}} calls is made automatically so the
before/after row counts are recorded in the session snapshot history.
}
\examples{
dt <- ops_toy(n = 100)
withdraw_file <- tempfile(fileext = ".csv")
writeLines(as.character(dt$eid[1:5]), withdraw_file)
dt <- ops_withdraw(dt, file = withdraw_file)
}
